/**
 * Pedestrian behavior (see entities/pedestrian)
 * @moduleName PedestrianBehavior
 */
bento.define('components/ai/pedestrian', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'states/pedestrian'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    PedestrianStates
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var multiScale;
        var movable;
        var punchable;
        var stateMachine;
        var walkSpeed = settings.walkSpeed || 0.85;

        var onHit = function (data) {
            var punch = data.punch;
            movable.velocityZ = -punch.pushBack.y;
            movable.velocity.x = punch.pushBack.x * punch.direction;


            stateMachine.setState('hurt');
            EventSystem.off('punchHit', onPunchConnected);
        };

        var onPunchConnected = function (data) {
            if (Math.random() < settings.intimidationChance) {
                var dist = entity.position.distance(data.punch.position);

                if (data.victim === entity) {
                    //already handling this punch in the [onHit] function
                    return;
                }

                if (data.victim.name.indexOf('prop') === 0) {
                    //this is just an inanimate object. Those get punched in Beatstreetville all the time
                    return;
                }

                if (dist < 80) { //TODO
                    stateMachine.setState('shocked');
                    EventSystem.off('punchHit', onPunchConnected);

                    var diff = data.punch.position.subtract(entity.position);
                    multiScale.direction.x = Utils.sign(diff.x);
                }
            }
        };

        var component = {
            name: 'pedestrianBehavior',
            start: function (data) {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(PedestrianStates);

                stateMachine.data.rootPath = settings.rootPath;
                stateMachine.setState('idle');

                multiScale = entity.getComponent('multiScale');
                multiScale.addScale('hurt', new Vector2(1, 1));

                movable = entity.getComponent('movable');

                punchable = entity.getComponent('punchable');
                punchable.onHit(onHit);

                stateMachine.data.fleeSpeed = settings.fleeSpeed;
                stateMachine.data.maxFlee = settings.maxFlee;
                stateMachine.data.startFlee = settings.startFlee;

                //listen for a punch
                EventSystem.on('punchHit', onPunchConnected);
            },
            destroy: function (data) {
                EventSystem.off('punchHit', onPunchConnected);
            },
            update: function (data) {
                if (movable.grounded) {
                    if (stateMachine.stateName === 'walk') {
                        //walk at a certain speed
                        movable.velocity.scalarMultiplyWith(walkSpeed);
                    } else if (stateMachine.stateName !== 'flee') {
                        //apply brake speed when not fleeing or walking
                        movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.15);
                    }
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };
        return component;
    };
});