/**
 * Standard walking enemy behavior AI
 * Also has the states
 * @moduleName WalkingEnemy
 */
bento.define('components/ai/walkingenemy', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/dead',
    'entities/punch',
    'components/punchable',
    'managers/ai',
    'modules/tapticengine',
    'entities/effect',
    'modules/matchmanager',
    'states/standardenemy',
    'managers/save',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Dead,
    Punch,
    Punchable,
    AiManager,
    TapticEngine,
    Effect,
    MatchManager,
    StandardEnemyStates,
    SaveManager,
    Globals
) {
    'use strict';
    var emergencyLoad = function (rootPath, onComplete) {
        var assetGroups = Bento.assets.getAssetGroups();
        var city = Utils.assetGroups.city;
        var emergencyGroup = assetGroups['emergency'] || {
            spritesheets: {},
            path: city.path
        };
        assetGroups['emergency'] = emergencyGroup;

        // which spritesheets are in the rootPath?
        // we can tell by inspecting the deleted city assetgroup
        Utils.forEach(city.spritesheets, function (citySpriteSheet) {
            if (citySpriteSheet.indexOf(rootPath) === 0) {
                if (city.spritesheets[citySpriteSheet]) {
                    emergencyGroup.spritesheets[citySpriteSheet] = citySpriteSheet;
                }
            }
        });
        Bento.assets.load('emergency', function () {
            onComplete();
        }, function (assetsLoaded, assetCount, name) {

        });
    };

    var WalkingEnemy = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var multiScale;
        var movable;
        var punchable;
        var healthBar;
        var stateMachine;
        var enemyController;
        var isDead = false;
        var difficulty = 1 - (settings.difficulty || SaveManager.load('difficulty'));
        var component = {
            name: 'walkingEnemyBehavior',
            /**
             * AI parameters
             */
            // hesitation: settings.hesitation || [10, 40], // range of frames to hesitate
            // hesitation: settings.hesitation || [10 + 30 * difficulty, 15 + 30 * difficulty], // range of frames to hesitate
            hesitation: settings.hesitation || [5 + 30 * difficulty, 10 + 30 * difficulty], // range of frames to hesitate
            hurry: settings.hurry ? settings.hurry : 0, // additional walk speed when targeting
            jitter: new Vector2(0, 0), // add random positioning?
            forceBlock: false,
            noHurt: false,

            // sets friction on the ground
            friction: 0.15,
            /**
             * Punchable callbacks
             * (Exposed here so they can be potentially be removed by other components)
             */
            // TODO: allow removal of punchable callbacks
            onHit: function (data) {
                var punch = data.punch;
                var getHurt = true;
                var stats = entity.getComponent('stats');
                var grabable = entity.getComponent('grabable');
                var pushBackFactor = 1;
                var staminaFactor = 1;
                var staminaSkillFactor = 1;
                var isBlocking = false;
                var blockChance = 0;
                var stunChance = 1;
                // compare skills (unless skill was not defined)
                if (punch.skill && stats) {
                    if (punch.skill < stats.skill) {
                        // decrease 10% with each skill difference
                        stunChance = 1 - Math.abs(punch.skill - stats.skill) * 0.1;
                        stunChance = Utils.clamp(0, stunChance, 1);

                        // decrease stamina 20% with each skill difference
                        // staminaSkillFactor = 1 - Math.abs(punch.skill - stats.skill) * 0.2;
                        // staminaSkillFactor = Utils.clamp(0.01, staminaSkillFactor, 1);

                        // roll dice
                        if (Utils.getRandomFloat(1) > stunChance) {
                            getHurt = false;
                        }
                    }
                    // force
                    if (component.forceBlock || (component.noHurt && punch.skill <= stats.skill)) {
                        getHurt = false;
                        pushBackFactor = 0.75;
                        // staminaFactor = 0.5;
                    }

                    blockChance = 1 - punch.skill * 7 / (punch.skill * 7 + stats.skill);

                    if (blockChance > 0.5) {
                        blockChance = 0.5;
                    }
                    // cannot block charge attacks
                    if (data && data.punch && data.punch.type === 'specialAttack') {
                        blockChance = 0;
                    }

                    if (Utils.getRandomFloat(1) < blockChance || component.forceBlock) {
                        isBlocking = true;
                    }
                    // turn off blocking for now
                    // if (!component.forceBlock) {
                    //     isBlocking = false;
                    // }
                }

                //faint is calculated _after_ hit connects, so this should only happen on punches that hit an enemy that has already fainted
                if (punchable.hasFainted) {
                    stateMachine.setState('faint', {}, true); //"re-faint"

                    // cant block in the air
                    isBlocking = false;

                    // half the usual damage
                    if (!punch.emulated) {
                        // punch.emulated comes from punchsyncer, in that case we ignore this and do the damage that was reported by the syncer
                        punch.attack = Math.ceil(punch.attack / 4);
                    }
                }

                // NOTE: because this is moved after the setState('faint'), the refaint is much lower
                // push back
                movable.velocityZ = -punch.pushBack.y * pushBackFactor;
                movable.velocity.x = punch.pushBack.x * punch.direction * pushBackFactor;

                // ignore block
                if (stateMachine.state.ignoreBlock) {
                    isBlocking = false;
                }

                if (!punchable.isDead && !isBlocking) {
                    // deplete health and stamina
                    punchable.depleteHealth(punch);

                    if (!punchable.hasFainted) { // don't deplete extra stamina in air
                        // deplete stamina

                        // remove this for now because it makes enemies really easy to defeat
                        // if (punch.stamina) {
                        //     punch.stamina *= staminaFactor;
                        // }

                        // remove this for now because it makes enemies really easy to defeat
                        // stamina is lower when skill is lower
                        // exception for grabbAttacks
                        // if (!punch.emulated && punch.type !== 'grabAttack') {
                        //     punch.stamina *= staminaSkillFactor;
                        // }

                        if (!stateMachine.state.ignoreFaint) {
                            punchable.depleteStamina(punch);
                        }
                        // at this point, the enemy could be fainted

                        // should enemy go into hurt state?
                        if (!punchable.hasFainted) {
                            if (!getHurt && punchable.getHealth() > 0) {
                                // shake?
                            } else if (!stateMachine.state.ignoreStun) {
                                if (grabable && grabable.isGrabbed() !== null) {
                                    stateMachine.setState('grabbedHurt', {}, true);
                                } else {
                                    stateMachine.setState('hurt', {}, true);
                                }
                            }
                        }
                    }
                    EventSystem.fire('punchHit', {
                        punch: punch,
                        victim: entity,
                        attacker: punch.attacker
                    });
                } else if (isBlocking) {
                    // go to block state
                    stateMachine.setState('block', {}, true);
                    if (punch.parentEntity) {
                        // push player back instead
                        punch.parentEntity.getComponent('movable', function (playerMovable) {
                            playerMovable.velocity.x = -punch.pushBack.x * punch.direction;
                        });
                    }
                    return 'blocked';
                }
            },
            onDeath: function (data) {
                isDead = true;
                stateMachine.setState('die', {
                    direction: data.punch ? data.punch.direction : 1
                }, true);
                entity.attach(new Dead({}));

                if (data.punch &&
                    !data.punch.emulated &&
                    Utils.isDefined(data.punch.type)
                ) {
                    EventSystem.fire('quest-' + data.punch.type, {
                        count: 1
                    });
                }
                if (settings.onDeath) {
                    settings.onDeath();
                }
            },
            onFaint: function (data) {
                stateMachine.setState('faint', {});

                // release grabable if needed
                entity.getComponent('grabable', function (grabable) {
                    grabable.setGrabbed(null);
                });
            },
            //
            start: function (data) {
                enemyController = entity.getComponent('enemyController');

                // setup state machine
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(StandardEnemyStates);
                // overwrite idle, hurt and walk states
                stateMachine.addStates(WalkingEnemy.states);

                stateMachine.data.attack = settings.attack || 1;
                stateMachine.data.rootPath = settings.rootPath || stateMachine.data.rootPath;
                stateMachine.data.behavior = 'idle';
                // rootPath check
                if (!Utils.hasAsset('spritesheets', settings.rootPath + 'idle')) {
                    stateMachine.deferStates = true;
                    Utils.log(
                        'WARNING: rootPath ' + 
                        settings.rootPath + 
                        ' was never loaded. Performing an emergency load call. ' + 
                        'Mission: ' + SaveManager.load('currentMission') + '-' + SaveManager.load('currentSection')
                    );
                    // start doing an emergency load of assets
                    emergencyLoad(settings.rootPath, function () {
                        stateMachine.resolve();
                    });
                }
                stateMachine.setState('idle');

                // setup punchable
                punchable = entity.getComponent('punchable');
                punchable.onHit(component.onHit);
                punchable.onDeath(component.onDeath);
                punchable.onFaint(component.onFaint);
                if (entity.family.indexOf('enemy') < 0) {
                    entity.family.push('enemy');
                }

                // setup movable
                movable = entity.getComponent('movable');
                movable.boundingBox = new Rectangle(-12, -6, 24, 8); // TODO: move this to entity
                movable.onLanded = function () {
                    // sound effect when landing
                    if (punchable.hasFainted || stateMachine.stateName === 'grabbedThrow') {
                        Bento.audio.stopSound('sfx_body-land');
                        Bento.audio.playSound('sfx_body-land');

                        var effect = new Effect({
                            z: Globals.Layers.ENEMIES,
                            position: entity.position.add(new Vector2(0, 0)),
                            spriteSheet: 'effects/body-land'
                        });
                        Bento.objects.attach(effect);
                        new Tween({
                            from: 0.25,
                            to: 0,
                            in: 15,
                            ease: Tween.ELASTIC,
                            alpha: -5,
                            beta: 1,
                            onUpdate: function (v, t) {
                                multiScale.hurt.x = 1 + v;
                                multiScale.hurt.y = 1 - v;
                            }
                        });
                    }
                };

                // add scale
                multiScale = entity.getComponent('multiScale');
                multiScale.addScale('hurt', new Vector2(1, 1));
            },
            update: function (data) {
                if (movable.grounded) {
                    movable.velocity.x = Utils.approach(movable.velocity.x, 0, component.friction);
                    movable.velocity.y = Utils.approach(movable.velocity.y, 0, component.friction);
                }

                // let the controlling player decide the direction the enemy should face
                if (!stateMachine.data.online.isRemote()) {
                    if (enemyController.selectedTarget && !stateMachine.state.ignoreDirection) {
                        var sign = Utils.sign(enemyController.selectedTarget.position.x - entity.position.x);
                        if (sign !== 0) {
                            multiScale.direction.x = sign;
                        }
                    }
                }

                // bug fix (not pretty)
                // during multiplayer, the death state may be delayed
                if (isDead && stateMachine.stateName !== 'die') {
                    stateMachine.setState('die', {}, true);
                }

            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                entity.family.push('enemy');
            },
            setBehavior: function (behaviorType) {
                stateMachine.data.behavior = behaviorType;
            }
        };
        return component;
    };

    // walking enemy specialized states
    WalkingEnemy.states = {
        idle: {
            start: function (data, transitionData) {
                var walkingEnemy = data.walkingEnemyBehavior;
                data.sprite.setSpriteSheet(data.rootPath + 'idle');

                data.idleTime = Utils.getRandomRange(walkingEnemy.hesitation[0], walkingEnemy.hesitation[1]);
            },
            update: function (data) {
                data.idleTime--;
            },
            move: function (data, direction) {
                if (data.idleTime > 0) {
                    return;
                }
                data.setState('walk');
            },
            action: function (data, action) {
                if (data.idleTime > 0) {
                    return;
                }
                if (action.data) {
                    data.setState('punch', {
                        target: action.data.target.name
                    });
                }
            },
            end: function (data) {

            },
            canGrab: true
        },
        walk: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'walk');
                data.walkTime = 60 + Math.random() * 30;
                data.velocityX = 0;
            },
            update: function (data) {
                // add a small percentage of the new velocity so the animation will only switch when the enemy is actually walking in a different direction
                data.velocityX *= 0.8;
                data.velocityX += data.movable.velocity.x * 0.2;

                data.walkTime--;
                if (data.walkTime <= 0 && data.behavior === 'idle' && !data.isRemote) {
                    data.setState('idle');
                } else if (Utils.sign(data.velocityX) === data.multiScale.direction.x) {
                    data.sprite.setSpriteSheet(data.rootPath + 'run');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'walk');
                }
            },
            move: function (data, direction) {
                var walkingEnemy = data.walkingEnemyBehavior;
                data.movable.velocity = direction.scalarMultiply(1.5 + walkingEnemy.hurry);
            },
            action: function (data, action) {
                //ai controller wants to punch, cancel walk and switch to idle
                data.setState('idle');
            },
            canGrab: true
        }
    };

    return WalkingEnemy;
});