/**
 * Anchor component.
 * Uses 2 positions, 1 relative in screen space <(0,0) to (1,1)>, and one absolute. Whenever the viewport changes it dimensions, it recalculates the desired position for its parent
 *
 * If something should have a (-40, 20) offset from the middle-right of the screen, pass
 *     {
 *         screenPosition : new Vector2(1, 0.5),
 *         offset : new Vector2(-40, 20)
 *     }
 *
 * @moduleName Anchor
 * @snippet Anchor.single
Anchor({
    screenPosition: new Vector2(${1:0.5}, ${2:0.5}),
    offset: new Vector2(${3:0}, ${4:0}),
    onComplete: function () {}
})
 * @snippet Anchor.multiple
Anchor({
    portrait: {
        screenPosition: new Vector2(${1:0.5}, ${2:0.5}),
        offset: new Vector2(${3:0}, ${4:0}),
    },
    landscape: {
        screenPosition: new Vector2(${5:0.5}, ${6:0.5}),
        offset: new Vector2(${7:0}, ${8:0}),
    },
    onComplete: function () {}
})
 * @snippet Anchor.isPortrait
Anchor.isPortrait()
 * @snippet Anchor.isWideScreen
Anchor.isWideScreen()
 */
bento.define('components/anchor', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'managers/resolution'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    ResolutionManager
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var Anchor = function (settings) {
        var offset = settings.offset || new Vector2(0, 0);
        var screenPosition = settings.screenPosition || new Vector2(0, 0);
        var landscape = {
            offset: settings.landscape ? (settings.landscape.offset || offset) : offset,
            screenPosition: settings.landscape ? (settings.landscape.screenPosition || screenPosition) : screenPosition
        };
        var portrait = {
            offset: settings.portrait ? (settings.portrait.offset || offset) : offset,
            screenPosition: settings.portrait ? (settings.portrait.screenPosition || screenPosition) : screenPosition
        };
        var resetPosition = function () {
            var orientation = ResolutionManager.isPortrait() ? portrait : landscape;
            entity.position.x = orientation.screenPosition.x * viewport.width + orientation.offset.x;
            entity.position.y = orientation.screenPosition.y * viewport.height + orientation.offset.y;

            if (settings.onComplete) {
                settings.onComplete(entity, ResolutionManager.isPortrait(), component);
            }
        };
        var viewport = Bento.getViewport();
        var entity;
        var component = {
            name: 'anchor',
            start: function (data) {
                EventSystem.on('orientationChanged', resetPosition);
                resetPosition();
            },
            destroy: function (data) {
                EventSystem.off('orientationChanged', resetPosition);
            },
            attached: function (data) {
                entity = data.entity;
            },
            resetPosition: resetPosition,
            getPosition: function () {
                var orientation = ResolutionManager.isPortrait() ? portrait : landscape;
                return new Vector2(
                    orientation.screenPosition.x * viewport.width + orientation.offset.x,
                    orientation.screenPosition.y * viewport.height + orientation.offset.y
                );
            }
        };
        return component;
    };
    Anchor.isPortrait = ResolutionManager.isPortrait;
    Anchor.isWideScreen = ResolutionManager.isWideScreen;
    Anchor.getWideScreenSize = ResolutionManager.getWideScreenSize;
    Anchor.getOffset = ResolutionManager.getOffset;

    return Anchor;
});