/**
 * Attaching a boss component makes the entity act like a boss
 * If the punchable's health reaches 0, the game enters slowmo mode
 * @moduleName Boss
 */
bento.define('components/boss', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'gui/menus/levelend',
    'managers/save',
    'managers/currency',
    'entities/props/lootdrop',
    'managers/buff',
    'modules/matchmanager',
    'gui/worldselect/topbarsimple',
    'globals',
    'managers/coincollector',
    'gui/currencycounter',
    'bento/components/sprite',
    'gui/shadowtext',
    'managers/inventory',
    'gui/levelupnotification',
    'managers/time',
    'managers/receiveitem',
    'managers/reward',
    'managers/item'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    LevelEndMenu,
    SaveManager,
    CurrencyManager,
    LootDrop,
    BuffManager,
    MatchManager,
    TopBarSimple,
    Globals,
    CoinCollector,
    CurrencyCounter,
    Sprite,
    ShadowText,
    InventoryManager,
    LevelUpNotification,
    TimeManager,
    ReceiveItem,
    RewardManager,
    ItemManager
) {
    'use strict';
    // initialized inside Boss module
    var missionsJson;

    var Boss = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var punchable;
        var slowmo = 0;
        var rewards;
        var component = {
            name: 'boss',
            skipIntro: settings.skipIntro || false,
            onDeath: function () {
                victorySequence();
            },
            start: function (data) {
                punchable = entity.getComponent('punchable');
                punchable.onDeath(component.onDeath);

                // remove coin dropper
                entity.getComponent('coinDropper', function (coinDropper) {
                    coinDropper.cancelCoinDrop();
                });
                Boss.currentBosses.push(entity);
            },
            destroy: function (data) {
                var index = Boss.currentBosses.indexOf(entity);
                if (index > -1) {
                    Boss.currentBosses.splice(index, 1);
                }
            },
            update: function (data) {

            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };
        var victorySequence = function () {
            var bossCount = Boss.currentBosses.length;
            var currentMission = SaveManager.load('currentMission');
            var missionStars = SaveManager.load('missionStars-' + currentMission);
            var obtained = SaveManager.load('rewards-' + currentMission);
            var isEvent = SaveManager.load('isEvent');
            var rewardsArray = [];

            var dropLoot = function () {
                // drop physical loot
                var rewardString;
                var amount;
                var randomReward;
                if (rewards.drop !== 3) {
                    rewardString = rewards['reward' + rewards.drop];
                    amount = parseInt(rewards['amount' + rewards.drop]);
                } else {
                    randomReward = RewardManager.getReward(currentMission);
                    rewardString = randomReward.reward;
                    amount = randomReward.amount;
                }

                if (isEvent) {
                    rewardString = 'playertokens/rad';
                    amount = 1;
                }

                var itemJson = Bento.assets.getJson(rewardString);
                var lootDrop = new LootDrop({
                    spriteSheet: itemJson.inventory.image,
                    speed: 2,
                    position: new Vector2(entity.position.x, entity.position.y)
                });

                Bento.objects.attach(lootDrop);
                // save drop as early as possible
                // get reward as item
                ReceiveItem({
                    key: rewardString,
                    amount: amount
                });

                // show the reward in level end
                rewardsArray.push({
                    key: rewardString,
                    amount: amount
                });

                lootDrop.setAmount(amount);

                // obtain level reward
                var earlyDrop = [obtained[0], obtained[1], obtained[2]];
                earlyDrop[rewards.drop - 1] = true;
                SaveManager.save('rewards-' + currentMission, earlyDrop);

                // save refresh 6 hours into the future
                // once you cleared all difficulties
                // (note: at this point you havent received the 3rd star)
                if (missionStars >= 3) {
                    SaveManager.save('rewardCooldown-' + currentMission, TimeManager.getTime() + 1000 * 60 * 60 * 6);
                }
            };
            var dropExtraLoot = function (array) {
                // get extra loot array
                var extraRewards = array || missionsJson[currentMission].extraRewards || [];

                Utils.forEach(extraRewards, function (reward) {
                    if (!reward || reward.key === 'nothing/nothing') {
                        return;
                    }
                    var rewardString = reward.key;
                    var amount = reward.amount;
                    var lootDrop = new LootDrop({
                        sprite: ItemManager.getSprite(rewardString),
                        speed: 2,
                        position: new Vector2(entity.position.x, entity.position.y),
                        dontShow: true
                    });

                    Bento.objects.attach(lootDrop);

                    // get reward as item
                    ReceiveItem({
                        key: rewardString,
                        amount: amount
                    });

                    // show the reward in level end
                    rewardsArray.push({
                        key: rewardString,
                        amount: amount
                    });

                    lootDrop.setAmount(amount);
                });
            };
            var doSlowmo = function () {
                if (Bento.isDev() && window.skipBossIntro) {
                    // faster bosses during dev
                    doVictory();
                    return;
                }
                new Tween({
                    from: 0.1,
                    to: 0.05,
                    in: 120,
                    ignoreGameSpeed: true,
                    ease: Tween.LINEAR,
                    onUpdate: function (v, t) {
                        Bento.setGameSpeed(v);
                    },
                    onComplete: function () {
                        restoreSlowmo();
                    }
                });
            };
            var restoreSlowmo = function () {
                // restore speed
                new Tween({
                    from: 0.05,
                    to: 1,
                    'in': 60,
                    ease: Tween.LINEAR,
                    ignoreGameSpeed: true,
                    onUpdate: function (v, t) {
                        Bento.setGameSpeed(v);
                    },
                    onComplete: function () {
                        slowmo = 2;
                        doVictory();
                        EventSystem.fire('slowmoEnd', {});
                    }
                });
            };
            var countCoins = function () {
                var coinsCollected = Math.ceil(CoinCollector.getCollected()) + BuffManager.current.extraCoins;
                var coinText = new ShadowText({
                    position: new Vector2(12, 6),
                    text: coinsCollected.toString(),
                    font: 'font',
                    fontSize: 16,
                    fontColor: '#ffffff',
                    align: 'left',
                    textBaseline: 'bottom'
                });
                var coinIcon = new Entity({
                    z: Globals.Layers.VICTORY,
                    name: 'coin',
                    components: [
                        new Sprite({
                            originRelative: new Vector2(0.5, 0.5),
                            imageName: 'gui/coin'
                        }),
                        coinText
                    ]
                });
                var topBar = showTopBar();
                var tweenInCoins = function () {
                    // find local player
                    var localPlayer;
                    Utils.forEach(Bento.objects.getByFamily('player'), function (player, i, l, breakLoop) {
                        if (!player.isRemote) {
                            localPlayer = player;
                        }
                    });
                    new Tween({
                        from: 0,
                        to: 1,
                        in: 20,
                        ease: 'exponential',
                        decay: 5,
                        onStart: function () {
                            Bento.objects.attach(coinIcon);
                            coinIcon.position = localPlayer.position.add(new Vector2(0, -16));
                        },
                        onUpdate: function (v, t) {
                            coinIcon.scale.x = v;
                            coinIcon.scale.y = v;
                        },
                        onComplete: function () {
                            // move towards center
                            new Tween({
                                from: coinIcon.position,
                                to: viewport.getCenter().add(new Vector2(0, 32)),
                                in: 30,
                                ease: 'exponential',
                                decay: 5,
                                onStart: function () {},
                                onUpdate: function (v, t) {
                                    coinIcon.position = v;
                                },
                                onComplete: function () {
                                    tweenCoinsCollected();
                                }
                            });
                            // scale up
                            new Tween({
                                from: 1,
                                to: 2,
                                in: 20,
                                ease: 'exponential',
                                decay: 5,
                                onStart: function () {
                                    coinIcon.position = localPlayer.position.add(new Vector2(0, -16));
                                },
                                onUpdate: function (v, t) {
                                    coinIcon.scale.x = v;
                                    coinIcon.scale.y = v;
                                }
                            });
                        }
                    });
                };
                var tweenCoinsCollected = function () {
                    new Tween({
                        from: 0,
                        to: 1,
                        in: Math.min(coinsCollected, 120),
                        delay: 20,
                        ease: 'linear',
                        onStart: function () {},
                        onUpdate: function (v, t) {
                            // tween counter
                            coinText.setText(Math.floor((1 - v) * coinsCollected).toString());
                            topBar.setCoins((CurrencyManager.getCoins() + Math.floor(v * coinsCollected)).toString());

                            if (t % 2 < 1) {
                                Bento.audio.stopSound('sfx_coin');
                                Bento.audio.playSound('sfx_coin');
                                topBar.flashCoin();
                            }
                        },
                        onComplete: function () {
                            // add the actual coin amount
                            CurrencyManager.addCoins(coinsCollected);

                            showLevelEnd();
                        }
                    });
                };
                // show coins collected
                tweenInCoins();

                // save amount of coins (used later, to double coins for ad)
                SaveManager.save('coinsCollected', coinsCollected);
            };
            var showTopBar = function () {
                var topbar = new TopBarSimple({
                    z: Globals.Layers.VICTORY + 1
                });
                var offset = 16;

                topbar.float = true;
                topbar.updateWhenPaused = 1;

                new Tween({
                    from: topbar.position.y - offset,
                    to: topbar.position.y,
                    in: 30,
                    ease: 'exponential',
                    decay: 5,
                    onStart: function () {
                        Bento.objects.attach(topbar);
                    },
                    onUpdate: function (v, t) {
                        topbar.position.y = v;
                    }
                });
                // move healthbar too
                Bento.objects.get('healthBar', function (healthBar) {
                    new Tween({
                        from: healthBar.position.y,
                        to: healthBar.position.y + offset,
                        in: 30,
                        ease: 'exponential',
                        decay: 5,
                        onUpdate: function (v, t) {
                            healthBar.position.y = v;
                        }
                    });
                });
                // move equipment too
                Bento.objects.get('equipped', function (equipped) {
                    new Tween({
                        from: equipped.position.y,
                        to: equipped.position.y + offset,
                        in: 30,
                        ease: 'exponential',
                        decay: 5,
                        onUpdate: function (v, t) {
                            equipped.position.y = v;
                        }
                    });
                });
                return topbar;
            };
            var showLevelEnd = function () {
                // show level end
                Utils.timeout(60 * 1, function () {
                    var levelEnd = new LevelEndMenu({
                        rewards: rewardsArray
                    });
                    Bento.objects.attach(levelEnd);
                    Bento.audio.playMusic('bgm_victory');
                });
            };
            var doVictory = function () {
                EventSystem.fire('bossWin');

                new Tween({
                    // give one second before doing the final pose
                    in: 60,
                    onComplete: function () {
                        EventSystem.fire('disableControls');
                        Bento.audio.playSound('sfx_victory');

                        // play victory
                        Utils.forEach(Bento.objects.getByFamily('player'), function (player) {
                            player.getComponent('stateMachine', function (stateMachine) {
                                if (stateMachine.stateName !== 'dead') {
                                    stateMachine.setState('victory');
                                }
                            });
                        });

                        // show topbar to prepare for coincounting
                        Utils.timeout(60 * 1.5, function () {
                            countCoins();
                        });
                    }
                });
            };

            Utils.forEach(Boss.currentBosses, function (boss, i, l, breakLoop) {
                boss.getComponent('punchable', function (punchableComponent) {
                    if (punchableComponent.isDead) {
                        bossCount -= 1;
                    }
                });
            });
            if (bossCount === 0 && slowmo === 0) {
                // complete mission
                EventSystem.fire('quest-mission', {
                    count: 1,
                    mission: currentMission
                });

                slowmo = 1;
                Bento.audio.stopSound('sfx_finalpunch');
                Bento.audio.playSound('sfx_finalpunch');

                // drop a reward for defeating the boss
                rewards = SaveManager.load('possibleRewards');
                rewards.drop = -1;

                // new reward system: drop the 1st reward if that's not collected yet
                // if there is no cooldown
                var cooldown = SaveManager.load('rewardCooldown-' + currentMission);
                if (cooldown > TimeManager.getTime()) {
                    rewards.drop = -1;
                } else {
                    var has2nd = (rewards.reward2 !== 'nothing/nothing');
                    var has1st = (rewards.reward1 !== 'nothing/nothing');
                    var collectedRewards = SaveManager.load('rewards-' + currentMission);
                    if (has2nd && !collectedRewards[1]) {
                        // force the 2nd item to be dropped
                        rewards.drop = 2;
                    } else if (has1st && !collectedRewards[0]) {
                        // force the 1st item to be dropped
                        rewards.drop = 1;
                    } else {
                        // drop random
                        rewards.drop = 3;
                    }
                }

                // a reward only drops if player didn't clear the difficulty yet
                // or it's the hardest difficulty
                var chosenDifficulty = SaveManager.load('chosenDifficulty');
                var hasClearedDifficultyBefore = chosenDifficulty < missionStars;
                if (chosenDifficulty < 2 && hasClearedDifficultyBefore) {
                    rewards.drop = -1;
                }
                // another exception: people who join games online do no get rewards
                var isJoining = SaveManager.load('isJoining');
                if (isJoining) {
                    rewards.drop = -1;
                }

                // Event: always drop Rad token
                if (isEvent) {
                    rewards.drop = 1;
                    // get ready for next mission
                    SaveManager.add('eventMissionsCleared', 1);
                }

                // make something drop
                if (rewards.drop !== -1) {
                    dropLoot();
                }
                if (rewards.drop === 1) {
                    // going to drop the 1st reward, also drop other rewards
                    dropExtraLoot();
                } else {
                    // how to randomize extra rewards?
                    // could be coins, consumables, gems, keys
                    // for now: drop 50, 60, 70 or 80 coins
                    dropExtraLoot([{
                        key: 'currency/coin',
                        amount: Globals.getExtraLootCoins(currentMission)
                    }]);
                }

                // fade out music
                new Tween({
                    from: 1,
                    to: 0,
                    in: 60,
                    ease: 'linear',
                    ignoreGameSpeed: true,
                    onUpdate: function (v, t) {
                        Bento.audio.setVolume(v, Globals.currentMusic);
                    },
                    onComplete: function () {
                        Bento.audio.stopAllMusic();
                    }
                });

                EventSystem.fire('lastPunch');
                // slowmo
                doSlowmo();

                // remove the pause button
                Bento.objects.get('pauseButton', function (pauseButton) {
                    pauseButton.removeSelf();
                });

                // make all players invincible
                Utils.forEach(Bento.objects.getByFamily('player'), function (player) {
                    if (player) {
                        player.getComponent('punchable', function (punchableComponent) {
                            punchableComponent.setInvincible(true);
                        });
                    }
                });
            }
        };

        if (!missionsJson) {
            missionsJson = Bento.assets.getJson('missions/missions');
        }

        return component;
    };
    Boss.currentBosses = []; // currentBosses is updated on destroy

    return Boss;
});