/**
 * Player Behavior
 * @moduleName PlayerBehavior
 */
bento.define('components/characters/player', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'entities/punch',
    'bento/components/sprite',
    'components/controllers/touchcontroller',
    'components/punchable',
    'modules/tapticengine',
    'globals',
    'entities/effect',
    'components/sort',
    'modules/matchmanager',
    'states/standardplayer',
    'managers/ai',
    'managers/inventory',
    'gui/consumablepopup',
    'managers/buff',
    'managers/item'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Punch,
    Sprite,
    TouchController,
    Punchable,
    TapticEngine,
    Globals,
    Effect,
    Sort,
    MatchManager,
    StandardPlayerStates,
    AiManager,
    InventoryManager,
    ConsumablePopup,
    BuffManager,
    ItemManager
) {
    'use strict';
    /*
     * Player module
     */
    var Player = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var movable;
        var punchable;
        var multiScale;
        var stateMachine;
        var isDead = false;

        var createPunch = function (punchData) {
            var tmp;
            var bb = punchData.boundingBox || new Rectangle(0, 0, 1, 1);
            var offset = punchData.offset || new Vector2(0, 0);
            var mirroredOffset = offset.clone();
            var ignoreFlip = punchData.ignoreFlip;
            var oldPos = entity.position.clone();

            mirroredOffset.x *= -1;
            mirroredOffset.x -= bb.width;

            if (multiScale.direction.x < 0) {
                tmp = offset;
                offset = mirroredOffset;
                mirroredOffset = tmp;
            }

            var pos = entity.position.add(offset);
            var mirroredPos = entity.position.add(mirroredOffset);

            var projectedBb = bb.offset(pos);
            var mirroredBb = bb.offset(mirroredPos);

            //check if any enemies are hit with this punch, and if not, if any enemies would be hit with a punch in the other direction
            var enemies = Bento.objects.getByFamily('enemy');
            var hit = false;
            var mirroredHit = false;
            var ii;
            for (ii = 0; ii < enemies.length; ii++) {
                enemies[ii].getComponent('punchable', function (pb) {
                    if (pb.collidesWith(projectedBb)) {
                        hit = true;
                    }
                    if (pb.collidesWith(mirroredBb)) {
                        mirroredHit = true;
                    }
                });
                if (hit) {
                    break;
                }
            }

            //flip player around if neccesary
            if (!hit && mirroredHit && !ignoreFlip) {
                pos = mirroredPos;
                multiScale.direction.x *= -1;
                movable.velocity.x = multiScale.direction.x;
            }

            if (!component.canPunch) {
                return new Entity({
                    name: 'punchStub'
                }).attach({
                    name: 'removeSelf',
                    start: function (data) {
                        Bento.objects.remove(data.entity);
                    }
                });
            }

            var punch = new Punch({
                target: punchData.target,
                attack: punchData.attack * BuffManager.current.attackBonus || 1,
                multiplier: punchData.multiplier,
                bonusAttack: punchData.bonusAttack,
                stamina: punchData.stamina,
                duration: punchData.duration,
                skill: Math.ceil(punchData.skill + BuffManager.current.skillBonus),
                parent: entity,
                direction: punchData.direction || multiScale.direction.x,
                freeze: punchData.freeze || 0,
                height: punchData.height,
                pushBack: punchData.pushBack || new Vector2(0, 0),
                boundingBox: bb,
                position: pos,
                onHit: punchData.onHit,
                onMiss: punchData.onMiss,
                debug: true,
                crush: BuffManager.current.crushDestructables || false,
                type: punchData.type,
            });
            punch.attach({
                name: 'repositionPunch',
                start: function () {
                    // reposition punch at the position of the player
                    // because the punch may be created at a later point and the player
                    // could have been moved in the mean time
                    var movedPosition = entity.position.subtract(oldPos);
                    // add this to the offset
                    punch.getComponent('behaviorComponent', function (behavior) {
                        behavior.offset.addTo(movedPosition);
                    });
                }
            });
            return punch;
        };

        var onHit = function (data) {
            var punch = data.punch;
            var getHurt = true;
            var stats = entity.getComponent('stats');
            var pushBackFactor = 1;
            var staminaFactor = 1;
            var cancelDamageBuff = BuffManager.current.cancelDamage;

            // enemies need only quarter the skill to knock player back
            /*if (punch.skill && stats) {
                if (punch.skill < stats.skill / 4) {
                    getHurt = false;
                    pushBackFactor = 0.75;
                    staminaFactor = 0.5;
                }
            }*/

            movable.velocityZ = -punch.pushBack.y * pushBackFactor;
            movable.velocity.x = punch.pushBack.x * punch.direction * pushBackFactor;

            if (cancelDamageBuff) {
                BuffManager.current.cancelDamage -= 1;
                return;
            }

            if (!punchable.isDead) {
                // deplete health and stamina
                punchable.depleteHealth(punch);
                // deplete stamina
                if (punch.stamina) {
                    punch.stamina *= staminaFactor;
                }

                if (BuffManager.current.knockDown) {
                    punchable.depleteStamina(punch);
                }

                if (getHurt && !punchable.hasFainted) {
                    stateMachine.setState('hurt', {
                        punch: {
                            pushBack: punch.pushBack,
                            direction: punch.direction
                        }
                    }, true);
                }
            }

            entity.getComponent('weapon', function (weapon) {
                weapon.getComponent('weapon').drop(0);
            });
        };

        var onDeath = function (data) {
            var items = InventoryManager.getEquippedConsumables();
            var equipped = Bento.objects.get('equipped');
            var die = true;
            var newHealth;

            // check items to revive

            if (BuffManager.rollCritical('revive') > 1) {
                die = false;
                newHealth = 1;
                BuffManager.current.revive = 0;
            } else if (items[0] && items[0].key === 'consumables/revive') {
                die = false;
                newHealth = 0;
                // spawn popup above the player
                entity.attach(new ConsumablePopup({
                    sprite: ItemManager.getSprite(items[0].key)
                }));
                // remove consumable
                equipped.slot0.removeItem();
            } else if (items[1] && items[1].key === 'consumables/revive') {
                die = false;
                newHealth = 0;
                // spawn popup above the player
                entity.attach(new ConsumablePopup({
                    sprite: ItemManager.getSprite(items[1].key)
                }));
                // remove consumable
                equipped.slot1.removeItem();
            }

            // die
            if (die) {
                EventSystem.fire('playerDeath', {
                    player: entity,
                });
                //TODO play dying animation
                stateMachine.setState('dead', {
                    direction: data.punch.direction
                }, true);
            } else {
                // revive
                data.punch.stamina = 1;
                punchable.resetHealth();
                punchable.depleteStamina(data.punch);
                stateMachine.setState('faint', {
                    punch: {
                        direction: data.punch.direction
                    }
                }, true);
                Bento.audio.stopSound(settings.rootPath + 'sfx_death');
                Bento.audio.playSound(settings.rootPath + 'sfx_death');

                if (newHealth) {
                    punchable.setHealth(newHealth);
                }
            }
        };

        var onFaint = function (data) {
            stateMachine.setState('faint', {
                punch: {
                    direction: data.punch.direction
                }
            });
        };

        var component = {
            name: 'playerBehavior',
            isDashing: false,
            canPunch: true,
            start: function (data) {
                // setup state machine
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(StandardPlayerStates);
                stateMachine.data.createPunch = createPunch;
                stateMachine.data.rootPath = settings.rootPath || '';
                stateMachine.setState('idle');

                // standard punch data
                stateMachine.data.punchLength = [18, 18, 18, 18];
                stateMachine.data.punchDistance = [8, 8, 8, 8];

                // setup punchable
                punchable = entity.getComponent('punchable');
                if (punchable) {
                    punchable.onHit(onHit);
                    punchable.onDeath(onDeath);
                    punchable.onFaint(onFaint);
                    punchable.boundingBox = new Rectangle(-8, -6, 16, 12);
                    if (entity.family.indexOf('player') < 0) {
                        entity.family.push('player');
                    }

                    // Buff stats
                    // stateMachine.data.stats.health += BuffManager.current.health;
                    // stateMachine.data.stats.attack += BuffManager.current.attack;
                    // stateMachine.data.stats.skill += BuffManager.current.skill;
                    // punchable.setMaxHealth(stateMachine.data.stats.health);
                }

                // setup movable
                movable = entity.getComponent('movable');
                movable.boundingBox = new Rectangle(-8, -6, 16, 8);
                // player also collides with solidPlayer family
                movable.familiesToCollide.push('solidPlayer');
                movable.onLanded = function () {
                    // sound effect when landing
                    if (punchable && punchable.hasFainted) {
                        Bento.audio.stopSound('sfx_body-land');
                        Bento.audio.playSound('sfx_body-land');

                        var effect = new Effect({
                            z: Globals.Layers.ENEMIES,
                            position: entity.position.add(new Vector2(0, 0)),
                            spriteSheet: 'effects/body-land'
                        });
                        Bento.objects.attach(effect);
                        new Tween({
                            from: 0.25,
                            to: 0,
                            in: 15,
                            ease: Tween.ELASTIC,
                            alpha: -5,
                            beta: 1,
                            onUpdate: function (v, t) {
                                multiScale.hurt.x = 1 + v;
                                multiScale.hurt.y = 1 - v;
                            }
                        });
                    }
                };
                // setup entity
                entity.boundingBox = movable.boundingBox.clone();
                entity.isDown = false;

                // setup scales
                multiScale = entity.getComponent('multiScale');
                multiScale.addScale('hurt', new Vector2(1, 1));
                multiScale.addScale('charge', new Vector2(1, 1));

                // when an player is spawned, reset ai tasks
                AiManager.distributeAiTasks();
            },
            destroy: function (data) {},
            update: function (data) {
                if (movable.grounded && !stateMachine.data.isDashing) {
                    movable.velocity.x = Utils.approach(movable.velocity.x, 0, component.decelaration);
                    movable.velocity.y = Utils.approach(movable.velocity.y, 0, component.decelaration);
                }

                // save entity attack bonus for use in punches and then reset entity attack bonus
                BuffManager.current.attackBonus = entity.attackBonus;
                entity.attackBonus = 1;

                // save entity skill bonus for use in punches and then reset entity skill bonus
                BuffManager.current.skillBonus = entity.skillBonus;
                entity.skillBonus = 1;

                // save entity doublecoins
                BuffManager.current.doubleCoins = entity.doubleCoins;
                entity.doubleCoins = false;

                // save entity doublecoins
                BuffManager.current.knockDown = entity.knockDown;
                entity.knockDown = true;

                // set chargeduration in statemachine data and then reset to standard value
                stateMachine.data.chargeDuration = entity.chargeDuration;
                entity.chargeDuration = BuffManager.current.chargeSpeed;
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            },
            decelaration: 0.2
        };

        return component;
    };
    return Player;
});