/**
 * An enemy controller tries to move an enemy by calling move() in the state
 * It is up to the character state itself how to interpret these commands.
 *
 * @moduleName EnemyController
 */
bento.define('components/controllers/enemycontroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'managers/ai',
    'modules/text',
    'modules/matchmanager',
    'entities/triggers/blockade'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    AiManager,
    Text,
    MatchManager,
    Blockade
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var EnemyController = function (settings) {
        var viewport = Bento.getViewport();
        var walkable;
        var solids;
        var tileColliders;
        var avoiders;
        var entity;
        var stateMachine;
        var debugText;
        var inActive = true;
        var forceSide = true;
        var punchBB = new Rectangle(0, 0, 1, 10);
        var attackQueue = false;
        var attackSide = null;
        var camaraderie = 1;
        var bail = false;
        var ignorePunch = 0;

        /*
         * Idle and chase behaviors
         */

        //kept outside of update function for debug purposes
        var diff = new Vector2(0, 0);
        var dir = new Vector2(0, 0);
        var target = null;
        /*
         * Check if enemyController should activate
         * Happens if enemy comes into view by the camera
         */
        var checkActive = function () {
            // colliding with the viewport activates an enemy
            if (viewport.hasPosition(entity.position)) {
                component.active = true;
            }
            // colliding with a camera locker activates an enemy
            entity.collidesWith({
                family: 'cameraLocker',
                onCollide: function () {
                    component.active = true;
                }
            });

            if (component.active) {
                // just became active: redistribute AI so it can potentially target the player
                AiManager.distributeAiTasks();
            }
        };

        var checkPunchCollision = function () {
            var noObstruction = true;
            punchBB.width = component.attackDistance;
            var sign = Utils.sign(target.position.x - entity.position.x);
            punchBB.y = entity.position.y - 5;
            if (sign === 1) {
                punchBB.x = entity.position.x + 4;
            } else {
                punchBB.x = entity.position.x - component.attackDistance - 4;
            }
            Utils.forEach(solids, function (other, a, l, breakLoop) {
                if (other.collidesWith({
                        rectangle: punchBB
                    })) {
                    //collided.push(other);
                    noObstruction = false;
                    breakLoop();
                }
            });
            if (noObstruction && tileColliders.collidesWith(punchBB, new Vector2(0, 0))) {
                noObstruction = false;
            }

            return noObstruction;
        };

        var applyForAttackSide = function (data) {
            AiManager.applyForAttackSide({
                target: data.target,
                id: data.id,
                side: data.side,
                succes: data.succes,
                fail: data.fail,
            });
        };

        var forceAttackSide = function (data) {
            AiManager.forceAttackSide({
                target: data.target,
                id: data.id,
                side: data.side
            });
            // since it's being forced set attackSide right away
            attackSide = {
                target: data.target,
                id: data.id,
                side: data.side
            };
        };

        var checkAttackSide = function (data) {
            AiManager.getLocalAttackSide({
                target: attackSide.target,
                id: attackSide.id,
                side: attackSide.side,
                succes: attackSide.succes,
                fail: function () {
                    attackSide = null;
                }
            });
        };

        var revokeAttackSide = function () {
            AiManager.revokeAttackSide(attackSide);
            attackSide = null;
        };

        var checkDistance = function (a) {
            if (a.top <= component.targetWalkPosition.y &&
                component.targetWalkPosition.y <= a.bottom
            ) {
                return 0;
            }
            return Math.min(Math.abs(component.targetWalkPosition.y - a.top),
                Math.abs(component.targetWalkPosition.y - a.bottom));
        };
        
        var chaseTarget = function (data) {
            if (!target) {
                return;
            }
            var chased = target;

            // when a citizen is active choose citizen, supports one citizen right now.
            // when more citizens are being used TODO: choose closest one
            var citizen = Bento.getByFamily('citizen');
            if (citizen && citizen.length > 0) {
                chased = citizen[0];
            }
            var compareDistance;
            Utils.forEach(Bento.getByFamily('player'), function (player, i, l, breakLoop) {
                // if a player is within punching range always target this
                // unless they're dead
                var isDead = false;
                player.getComponent('stateMachine', function (component) {
                    isDead = (component.stateName === 'dead');
                });
                if (isDead) {
                    return;
                }
                diff = player.position.subtract(entity.position);
                if (
                    Math.abs(diff.x) < component.attackDistance &&
                    Math.abs(diff.y) < 8 &&
                    !player.isDown
                ) {
                    if (!Utils.isDefined(compareDistance) || compareDistance > Math.abs(diff.x)) {
                        chased = player;
                        compareDistance = Math.abs(diff.x);
                    }

                    if (player === target) {
                        chased = target;
                        breakLoop();
                    }
                }
            });

            //check if close to other enemies
            var sign = Utils.sign(entity.position.x - chased.position.x);
            var bailCount = 0;

            if (!attackSide) {
                Utils.forEach(EnemyController.activeControllers, function (ec, i, l, breakLoop) {
                    var otherSign = Utils.sign(ec.parent.position.x - chased.position.x);
                    if (ec.parent !== entity) {
                        if (!bail) {
                            if (ec.parent.position.distance(entity.position) < 50 &&
                                otherSign === sign
                            ) {
                                bail = true;
                                component.targetWalkPosition = entity.position.clone();
                                breakLoop();
                                bailCount = 1;
                            }
                        } else {
                            if (ec.parent.position.distance(entity.position) < 70) {
                                ++bailCount;
                            }
                        }
                    }
                });
            } else {
                // check if attackSide is valid locally, AiManager syncs every 10 frames. Should be accurate enough
                checkAttackSide();
            }

            if (bail && !bailCount) {
                bail = false;
                bailCount = 0;
            }

            if (attackSide) {
                if (attackSide.side !== sign) {
                    revokeAttackSide();
                }
            }

            if (stateMachine.stateName.indexOf('grab') !== -1) {
                if (attackSide) {
                    revokeAttackSide();
                }
            }

            if (stateMachine.stateName === 'hurt') {
                if (!attackSide) {
                    forceAttackSide({
                        target: chased.name,
                        id: entity.name,
                        side: sign
                    });
                }
            }

            // component.target = chased;
            component.selectedTarget = chased;


            entity.setBehavior('target');
            diff = chased.position.subtract(entity.position);
            var viewBB = viewport.clone().grow(5);
            // TODO: 8 is a random number
            // can only attack when inside viewBB

            if (
                Math.abs(diff.x) < component.attackDistance &&
                Math.abs(diff.y) < 8 &&
                checkPunchCollision() &&
                viewBB.hasPosition(entity.position) &&
                !chased.isDown && !bail && !ignorePunch
            ) {
                if (attackSide) {
                    entity.processAction({
                        type: 'attack',
                        data: {
                            target: chased
                        }
                    });
                } else {
                    // check if attackSide is valid locally, AiManager syncs every 10 frames. Should be accurate enough
                    attackSide = {
                        target: chased.name,
                        id: entity.name,
                        side: sign
                    };
                    checkAttackSide({
                        succes: function () {
                            attackSide = {
                                target: chased.name,
                                id: entity.name,
                                side: sign
                            };
                        }
                    });
                    if (!attackSide) {
                        ignorePunch = 30;
                    }
                }

                component.targetWalkPosition.x = entity.position.x;
                component.targetWalkPosition.y = entity.position.y;
            } else if (!attackQueue) {
                if (entity.position.distance(component.targetWalkPosition) <= 3 || component.bump > 5) {
                    if (attackSide && Math.abs(diff.x) > component.attackDistance + 40) {
                        revokeAttackSide();
                    }

                    if (!walkable) {
                        walkable = Bento.objects.get('walkableArea');
                        return;
                    }

                    var action = Math.random();
                    var aggression = getAggression() * camaraderie;
                    if (action <= aggression) {
                        // attack mode
                        var attackAssigned = function () {
                            component.targetWalkPosition.x = chased.position.x + sign * (component.attackDistance - 8);
                            component.targetWalkPosition.y = chased.position.y + Utils.getRandomRange(-4, 4);
                            component.tension = 0;
                            attackQueue = false;
                            attackSide = {
                                target: chased.name,
                                id: entity.name,
                                side: sign
                            };
                            camaraderie = 1;
                        };
                        var attackDenied = function () {
                            // try other side?
                            attackSide = null;
                            component.tension = 0;
                            attackQueue = false;
                            attackSide = null;
                            camaraderie = 1;
                        };
                        attackQueue = true;
                        applyForAttackSide({
                            target: chased.name,
                            id: entity.name,
                            side: sign,
                            succes: attackAssigned,
                            fail: attackDenied,
                        });
                    } else {
                        // walk mode
                        entity.processAction('idle');
                        action -= aggression;
                        action /= 1 - aggression;
                        if (component.firstRandom) {
                            action = component.firstRandom;
                            component.firstRandom = undefined;
                        }
                        if (Math.abs(entity.position.y - chased.position.y) > 20 && action < 0.4) {
                            component.targetWalkPosition.x = chased.position.x + -sign * Math.max(component.attackDistance, 50);
                            component.targetWalkPosition.y = entity.position.y;
                        } else {
                            component.targetWalkPosition.x = chased.position.x + sign * Utils.getRandomRange(Math.min(component.attackDistance + 40, 80), 120);
                            // clamp position y in viewport

                            // switch position to other side of player if not natural side or outside of viewport
                            if (!viewport.hasPosition(component.targetWalkPosition) ||
                                (sign !== Utils.sign(component.targetOffset.x) && (!Utils.getRandom(3) || forceSide))
                            ) {
                                component.targetWalkPosition.x = -(component.targetWalkPosition.x - chased.position.x) + chased.position.x;
                                // set y to make sure enemy doesnt walk through player.position
                                component.targetWalkPosition.y = chased.position.y + Utils.sign(entity.position.y - chased.position.y) * Utils.getRandomRange(40, 80);
                            } else {
                                component.targetWalkPosition.y = chased.position.y + Utils.getRandomRange(-48, 48);
                            }
                        }
                        ++component.tension;

                        forceSide = false;

                        // try to avoid things like cars when there is no reason (no player) to go there
                        Utils.forEach(avoiders, function (avoid, a, l, breakLoop) {
                            if (Math.abs(avoid.position.y - target.position.y) > 28) {
                                var min = Math.min(entity.position.y, component.targetWalkPosition.y);
                                var max = Math.max(entity.position.y, component.targetWalkPosition.y);
                                if ((min < avoid.position.y && avoid.position.y < max) ||
                                    Math.abs(avoid.position.y - component.targetWalkPosition.y) < 28
                                ) {
                                    component.targetWalkPosition.y = avoid.position.y + Utils.sign(entity.position.y - avoid.position.y) * 28;
                                }
                            }
                        });

                        // try to avoid other enemies
                        Utils.forEach(Bento.objects.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                            enemy.getComponent('enemyController', function (controller) {
                                if (controller.active || !controller.isLocallyControlled) {
                                    if (component.targetWalkPosition.distance(enemy.position) < 24) {
                                        component.targetWalkPosition = entity.position.clone();
                                        breakLoop();
                                    }
                                }
                            });
                        });
                    }
                    var area = walkable.checkVertical(component.targetWalkPosition.x);
                    // select closest area
                    if (area[0]) {
                        var closestArea = area[0];
                        var distance = checkDistance(area[0]);
                        var newDistance;
                        var i;
                        for (i = 1; i < area.length && distance > 0; ++i) {
                            newDistance = checkDistance(area[i]);
                            if (newDistance < distance) {
                                closestArea = area[i];
                                distance = newDistance;
                            }
                        }
                        // clamp targetWalkPosition.y in walkable area
                        component.targetWalkPosition.y = Utils.clamp(closestArea.top, component.targetWalkPosition.y, closestArea.bottom);
                    }
                    // reset collision bumping
                    component.bump = 0;
                } else {
                    if (!viewport.hasPosition(component.targetWalkPosition)) {
                        component.targetWalkPosition.x = entity.position.x;
                        component.targetWalkPosition.y = entity.position.y;
                        return;
                    }

                    dir = component.targetWalkPosition.subtract(entity.position);
                    dir.normalize();
                    dir.scalarMultiplyWith(2);
                    // check for collisions, maybe in player itself
                    entity.getComponent('movable', function (movable) {
                        var bb = movable.boundingBox.offset(entity.position.add(dir));
                        Utils.forEach(solids, function (other, a, l, breakLoop) {
                            if (other.collidesWith({
                                    rectangle: bb
                                })) {
                                //collided.push(other);
                                ++component.bump;
                                breakLoop();
                            }
                        });
                        if (tileColliders.collidesWith(bb, new Vector2(0, 0))) {
                            ++component.bump;
                        }
                    });

                    dir.normalize();
                    entity.moveInDirection(dir);
                }
            }


            if (ignorePunch) {
                --ignorePunch;
            }
        };

        var enemyHurt = function (name) {
            if (name) {
                if (name === entity.name) {
                    return;
                }
            }
            camaraderie += 0.3;
        };

        var getAggression = function () {
            if (component.aggression < 1) {
                for (var i = 0; i < EnemyController.activeControllers.length; ++i) {
                    if (EnemyController.activeControllers[i].targetPlayerId === component.targetPlayerId &&
                        EnemyController.activeControllers[i] !== component) {
                        return component.aggression;
                    }
                }
                return component.alone;
            }
            return component.aggression;
        };

        /*
         * Assign an enemyController to target a playerId
         */
        var assignTargetPlayerId = function () {
            if (!MatchManager.isOnline || !MatchManager.localPlayer) {
                component.targetPlayerId = -1;
                component.isLocallyControlled = true;
                return;
            }
            var entityName = entity.name;
            var entityNumbers = entityName.split('-') || [0];
            var entityNumber = parseInt(entityNumbers[entityNumbers.length - 1]) + entity.name.length;
            var otherPlayers = MatchManager.otherPlayers;
            var localPlayer = MatchManager.localPlayer;
            var playerIds = [localPlayer.playerId];
            var addGap = {
                "0": 0,
                "1": 0,
                "2": 0,
                "3": 0
            };

            Utils.forEach(otherPlayers, function (playerData) {
                playerIds.push(playerData.playerId);
            });
            playerIds.sort();

            // check for any gaps
            Utils.forEach(playerIds, function (playerId, i, l, breakLoop) {
                if (playerId !== i) {
                    addGap[i] = playerId - i;
                }
            });

            // distribute number
            component.targetPlayerId = entityNumber % playerIds.length;
            component.targetPlayerId += addGap[component.targetPlayerId];

            if (component.targetPlayerId === localPlayer.playerId) {
                component.isLocallyControlled = true;
            } else {
                component.isLocallyControlled = false;
            }
        };
        var component = {
            name: 'enemyController',
            attackDistance: settings.attackDistance || 48,
            // offset from the target (left/ride side)
            targetOffset: settings.targetOffset || null,
            // suggested walking target
            targetWalkPosition: new Vector2(0, 0),
            targetPlayerId: -1,
            isLocallyControlled: true,
            // suggested targeted player
            target: null,
            selectedTarget: null,
            // uninterrupted attack
            //uninterrupted: 0,
            // how much tension can build up:
            aggression: 0.5,
            // how much tension can build up:
            alone: 0.8,
            // tension to attack
            tension: 0,
            // amount of times collided with level
            bump: 0,
            // suggested state
            aiState: AiManager.AiStates.CHASE,
            // suggested attack state
            debug: settings.debug || false,
            // whether the enemyController should be included in the AI manager, defaults to false
            active: settings.active || false,
            start: function (data) {
                walkable = Bento.objects.get('walkableArea');
                if (!this.targetOffset) {
                    this.targetOffset = new Vector2(0, 0);
                    this.targetOffset.x = (Math.round(Math.random()) * 2 - 1) * this.attackDistance / 2;
                    // this.targetOffset.y = Math.random() * this.attackDistance / 2 - this.attackDistance / 4;
                }

                AiManager.registerEnemy(component);

                stateMachine = entity.getComponent('stateMachine');

                // register a punch: this "wakes" up an enemy from inactive state
                entity.getComponent('punchable', function (punchable) {
                    punchable.onHit(function () {
                        if (!component.active) {
                            component.active = true;
                            AiManager.distributeAiTasks();
                        }
                    });
                });

                assignTargetPlayerId();

                // when an enemy is spawned, reset ai tasks
                AiManager.distributeAiTasks();

                //set listener for enemies being hurt to add aggression
                EventSystem.on('enemyHurt', enemyHurt);

                // a player disconnects: reassign target
                EventSystem.on(MatchManager.Events.PLAYERDISCONNECTS, assignTargetPlayerId);

                // set target walk position to enemy position
                component.targetWalkPosition.x = entity.position.x;
                component.targetWalkPosition.y = entity.position.y;

                solids = Bento.objects.getByFamily('solid');
                tileColliders = Bento.objects.get('tileColliders');
                avoiders = Bento.objects.getByFamily('avoider');
                // EnemyController.paths = true;
            },
            destroy: function (data) {
                EventSystem.off(MatchManager.Events.PLAYERDISCONNECTS, assignTargetPlayerId);
                EventSystem.off('enemyHurt', enemyHurt);
                AiManager.unregisterEnemy(component);
                // the enemy needs to be actually dead before sending this event
                entity.getComponent('punchable', function (punchable) {
                    if (punchable.isDead) {
                        EventSystem.fire('destroyEnemyController', {
                            entity: entity
                        });
                    }
                });
                var index = EnemyController.activeControllers.indexOf(component);
                if (index > -1) {
                    EnemyController.activeControllers.splice(index, 1);
                }
                if (attackSide) {
                    revokeAttackSide();
                }
            },
            update: function (data) {
                if (EnemyController.debug) {
                    if (!debugText) {
                        // init debug text
                        debugText = new Text({
                            position: new Vector2(0, 0),
                            font: 'font',
                            fontSize: 8,
                            fontColor: '#ffffff',
                            align: 'center',
                            textBaseline: 'top',
                        });
                        debugText.attach({
                            name: 'inverse',
                            update: function () {
                                debugText.scale.x = entity.scale.x;
                            }
                        });
                        entity.attach(debugText);
                    }
                    // show current state of enemy
                    debugText.setText(
                        entity.name + '\n' +
                        'state: ' + stateMachine.stateName + '\n' +
                        'ignore: ' + stateMachine.state.ignoreSync
                    );
                }

                if (component.target) {
                    target = component.target;
                }

                if (component.active && component.isLocallyControlled) {
                    if (inActive) {
                        EnemyController.activeControllers.push(component);
                        inActive = false;
                        component.firstRandom = 1;
                    }
                    component.aiState = AiManager.AiStates.CHASE;
                    component.ai(data, target, chaseTarget);
                }
            },
            attached: function (data) {
                entity = data.entity;
            },
            setTarget: function (t, side) {
                component.target = t;

                if (side && Utils.sign(this.targetOffset.x) !== Utils.sign(side)) {
                    this.targetOffset.x *= -1;
                }
            },
            ai: function (data, t, chaseBehavior) {
                // overridable ai behavior, used by bosses
                chaseBehavior(data);
            },
            draw: function (data) {
                if (EnemyController.paths && component.targetWalkPosition) {
                    var ctx = data.renderer.getContext('2d');
                    var targ = component.targetWalkPosition.subtract(entity.position);
                    targ.x *= stateMachine.data.multiScale.direction.x;
                    ctx.beginPath();
                    ctx.moveTo(0, 0);
                    ctx.lineTo(targ.x, targ.y);
                    ctx.stroke();
                }
            }
        };
        return component;
    };

    EnemyController.debug = false;
    EnemyController.paths = false;
    EnemyController.activeControllers = [];

    return EnemyController;
});