/**
 * Keyboard Controller. Used to provide player input to a character.
 * @moduleName KeyboardController
 */
bento.define('components/controllers/keyboardcontroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/clickable'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Clickable
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var holdTimer = 0;

        var isEnabled = true;

        var enableControls = function () {
            isEnabled = true;
        };

        var disableControls = function () {
            isEnabled = false;
        };

        var leftCount = 0;
        var leftPressed = 0;
        var leftFlick = false;
        var rightCount = 0;
        var rightPressed = 0;
        var rightFlick = false;
        var MAXCOUNT = 30;

        var component = {
            name: 'keyboardController',
            start: function (data) {
                EventSystem.on('enableControls', enableControls);
                EventSystem.on('disableControls', disableControls);
                EventSystem.on('startCutscene', disableControls);
                EventSystem.on('endCutscene', enableControls);
            },
            destroy: function (data) {
                EventSystem.off('enableControls', enableControls);
                EventSystem.off('disableControls', disableControls);
                EventSystem.off('startCutscene', disableControls);
                EventSystem.off('endCutscene', enableControls);
            },
            update: function (data) {
                if (!isEnabled) {
                    return;
                }

                var joystick = new Vector2(0, 0);
                var flick = false;

                //Left key is down
                if (Bento.input.isKeyDown('left')) {
                    //do this only when not running because of left
                    if (!leftFlick) {
                        //if button is pressed down this frame
                        if (leftPressed % 2 === 0) {
                            // if second press
                            if (leftPressed === 2) {
                                //if below double press threshold
                                if (leftCount <= MAXCOUNT) {
                                    leftFlick = true;
                                    flick = true;
                                } else {
                                    leftCount = 0;
                                }
                            }
                            //set to non pressed state
                            leftPressed++;
                        }
                        //when above threshold set press state to second press
                        if (leftCount > MAXCOUNT) {
                            leftPressed = 3;
                        }
                        //move player normally
                        joystick.x -= 1;
                    }
                } else {
                    //when not down

                    //if released
                    leftFlick = false;
                    if (leftPressed % 2 === 1) {
                        leftPressed++;
                    }

                    // if above threshold reset both
                    if (leftCount > MAXCOUNT) {
                        leftPressed = 0;
                        leftCount = 0;
                    }

                    //reset count
                    if (leftPressed === 4) {
                        leftCount = 0;
                    }
                }

                leftPressed = leftPressed % 4;

                //when double press eligable reset or start counting
                if (leftPressed > 0) {
                    leftCount++;
                }

                //Right key is down
                if (Bento.input.isKeyDown('right')) {
                    //do this only when not running because of left
                    if (!rightFlick) {
                        //if button is pressed down this frame
                        if (rightPressed % 2 === 0) {
                            // if second press
                            if (rightPressed === 2) {
                                //if below double press threshold
                                if (rightCount <= MAXCOUNT) {
                                    flick = true;
                                    rightFlick = true;
                                } else {
                                    rightCount = 0;
                                }
                            }
                            //set to non pressed state
                            rightPressed++;
                        }
                        //when above threshold set press state to second press
                        if (rightCount > MAXCOUNT) {
                            rightPressed = 3;
                        }
                        //move player normally
                        joystick.x += 1;
                    }
                } else {
                    //when not down

                    //if released
                    rightFlick = false;
                    if (rightPressed % 2 === 1) {
                        rightPressed++;
                    }

                    // if above threshold reset both
                    if (rightCount > MAXCOUNT) {
                        rightPressed = 0;
                        rightCount = 0;
                    }

                    //reset count
                    if (rightPressed === 4) {
                        rightCount = 0;
                    }
                }

                rightPressed = rightPressed % 4;

                //when double press eligable start counting
                if (rightPressed > 0) {
                    rightCount++;
                }

                if (Bento.input.isKeyDown('up')) {
                    joystick.y -= 1;
                }
                if (Bento.input.isKeyDown('down')) {
                    joystick.y += 1;
                }

                if (Bento.input.isKeyDown('z')) {
                    holdTimer++;
                    entity.processAction({
                        type: 'hold',
                        data: {
                            duration: holdTimer
                        }
                    });
                } else if (holdTimer !== 0) {
                    entity.processAction({
                        type: 'tap',
                        data: {
                            duration: holdTimer
                        }
                    });
                    holdTimer = 0;
                }

                if (Bento.input.isKeyDown('x')) {
                    // get direction player is facing
                    var direction = 1;
                    entity.getComponent('multiScale', function (multiScale) {
                        direction = multiScale.direction.x;
                    });

                    // simulate swipe movement to jump
                    entity.processAction({
                        type: 'swipe',
                        data: {
                            direction: (new Vector2(0, -1)),
                            cacheDirection: direction
                        }
                    });
                    entity.moveInDirection(joystick);
                }

                if (Bento.input.isKeyDown('c')) {
                    // initiate super attack if available
                    Bento.objects.get('superAttackButton', function (superAttackButton) {
                        superAttackButton.getComponent('button', function (atkBtn) {
                            atkBtn.doCallback();
                        });
                    });
                }

                joystick.normalize();

                if (Bento.input.isKeyDown('shift')) {
                    joystick.scalarMultiplyWith(0.5);
                }

                if (joystick.magnitude() > 0) {
                    if (flick && joystick.x !== 0) {
                        entity.processAction({
                            type: 'swipe',
                            data: {
                                direction: (new Vector2(joystick.x, 0).normalize())
                            }
                        });
                    } else {
                        entity.moveInDirection(joystick);
                    }
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            },
        };
        return component;
    };
});