/**
 * Multi touch controller. Used to provide player input to a character.
 * Active in both potrait and landscape
 * @moduleName MultiTouchController
 */
bento.define('components/controllers/multitouchcontroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/clickable',
    'globals',
    'bento/components/sprite',
    'managers/resolution',
    'managers/save'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Clickable,
    Globals,
    Sprite,
    ResolutionManager,
    SaveManager
) {
    'use strict';
    var sensitivityDefaults = function () {
        // this is a temporary solution until we create a menu where the user can change sensitivity
        // for now we set higher sensitivities for iPads
        var deviceModel;
        var numbers;
        var generation;
        var version;
        if (Utils.isCocoonJs() && window.device) {
            deviceModel = window.device.model;
            if (deviceModel.indexOf('iPad') < 0) {
                return;
            }
            // get the 2 numbers behind iPad
            // see https://www.theiphonewiki.com/wiki/Models
            deviceModel = deviceModel.replace('iPad', '');
            numbers = deviceModel.split(',');
            generation = parseInt(numbers[0]);
            version = parseInt(numbers[1]);

            SaveManager.save('sensitivityFactor', 0.75);
            SaveManager.save('flickSensitivity', 1.25);

            // ipad pro: extra swipe sensitivity
            if (
                generation === 7 ||
                (generation === 6 && version < 11)
            ) {
                SaveManager.save('sensitivityFactor', 0.5);
            }
        }
    };

    return function (settings) {
        sensitivityDefaults();
        var viewport = Bento.getViewport();
        var entity;

        var sensitivityFactor = SaveManager.load('sensitivityFactor'); // lower is more sensitive
        var flickSensitivity = SaveManager.load('flickSensitivity'); // higher is more sensitive

        var maxRadius = (settings.maxRadius || 32) * sensitivityFactor;
        var deadZone = (settings.deadZone || 8) * sensitivityFactor;

        var touchData = null;
        var joystickPosition = new Vector2(0, 0);
        var oldPos = new Vector2(0, 0);

        var holdTimer = 0;
        var flickY = 0;
        var flickX = 0;
        var touchTime = 0;
        var cacheDirection = 0;

        var secondTouch = null;
        var secondFlickY = 0;
        var secondFlickX = 0;
        var secondTouchTime = 0;
        var secondOldPos = new Vector2(0, 0);
        var secondCacheDirection = 0;
        var isAndroid = Utils.isNativeAndroid();

        var cursorDistance = settings.cursorDistance || 20;

        var isActive = true;

        var isEnabled = true;

        var moveEnabled = true;
        var attackEnabled = true;
        var chargeEnabled = true;
        var dashEnabled = true;
        var jumpEnabled = true;

        var enableControls = function () {
            isEnabled = true;
        };

        var disableControls = function () {
            isEnabled = false;
            joystickPosition = new Vector2(0, 0);
            touchData = null;
        };

        var cursor = new Entity({
            z: Globals.Layers.GUI,
            name: 'cursor',
            position: new Vector2(0, 0),
            alpha: 0.5,
            family: ['gui'],
            components: [
                new Sprite({
                    originRelative: new Vector2(0.5, 0.5),
                    imageName: 'arrow'
                }), {
                    name: 'visualize',
                    update: function () {
                        var magnitude = limitJoystick();
                        var pos = joystickPosition.clone().scalarMultiply(cursorDistance / maxRadius);
                        cursor.position = entity.position.add(pos);
                        cursor.rotation = joystickPosition.angle();

                        if (magnitude < deadZone) {
                            cursor.visible = false;
                        } else {
                            cursor.visible = true;
                        }
                    }
                }
            ]
        });

        var limitJoystick = function () {
            var magnitude = joystickPosition.magnitude();
            if (joystickPosition.magnitude() > maxRadius) {
                joystickPosition.scalarMultiplyWith(maxRadius / magnitude);
            }
            return Math.min(magnitude, maxRadius);
        };

        var clickable = new Clickable({
            updateWhenPaused: Globals.PauseLevels.FREEZE,
            pointerDown: function (data) {

                //check if there is a super attack button in the way
                var pressedSuper = false;
                Bento.objects.get('superAttackButton', function (atkButton) {
                    atkButton.getComponent('button', function (btn) {
                        var bb = btn.dimension.offset(atkButton.position);
                        var offset = ResolutionManager.getOffset().clone(); 
                        offset.y = 0; 
                        if (bb.hasPosition(data.position.subtract(offset))) {
                            pressedSuper = true;
                        }
                    });
                });

                if (!isActive || !isEnabled || pressedSuper) {
                    return;
                }

                if (touchData) {
                    if (!secondTouch && touchData.id != data.id) {
                        secondTouch = data.id;
                        secondTouchTime = 0;
                        secondOldPos = data.position.clone();
                        entity.getComponent('multiScale', function (multiScale) {
                            secondCacheDirection = multiScale.direction.x;
                        });
                    }
                    return;
                }

                if (secondTouch) {
                    return;
                }

                touchData = data.id;
                startedWalking = false;

                joystickPosition = new Vector2(0, 0);
                oldPos = data.position.clone();
                holdTimer = 0;
                touchTime = 0;
                flickY = 0;
                flickX = 0;
                entity.getComponent('multiScale', function (multiScale) {
                    cacheDirection = multiScale.direction.x;
                });
            },
            pointerMove: function (data) {
                if ((touchData !== data.id && secondTouch !== data.id) || !isActive || !isEnabled || !moveEnabled) {
                    return;
                }
                var difference;
                if (touchData === data.id) {
                    difference = data.position.substract(oldPos);

                    joystickPosition.addTo(difference);

                    // limit
                    limitJoystick();

                    flickY = difference.y;
                    flickX = difference.x;

                    // move with touch
                    oldPos = data.position.clone();
                } else {
                    // save positions for flicking
                    difference = data.position.substract(secondOldPos);
                    secondFlickY = difference.y;
                    secondFlickX = difference.x;
                    secondOldPos = data.position.clone();
                }
            },
            pointerUp: function (data) {
                if ((touchData !== data.id && secondTouch !== data.id) ||
                    !isActive || !isEnabled
                ) {
                    return;
                }

                if (touchData === data.id) {
                    if (!secondTouch) {
                        if (!startedWalking && attackEnabled) {
                            entity.processAction({
                                type: 'tap',
                                data: {
                                    duration: holdTimer
                                }
                            });
                            holdTimer = 0;
                        } else if (
                            flickX * flickX + flickY * flickY > (isAndroid ? 0.4 * 0.4 : 0.8 * 0.8) * flickSensitivity &&
                            touchTime < (isAndroid ? 15 : 10) * flickSensitivity
                        ) {
                            // dash only in x or y direction
                            if (Math.abs(flickX) >= Math.abs(flickY)) {
                                flickY = 0;
                            } else {
                                flickX = 0;
                            }

                            if ((dashEnabled && flickX) ||
                                (jumpEnabled && flickY)
                            ) {
                                entity.processAction({
                                    type: 'swipe',
                                    data: {
                                        direction: (new Vector2(flickX, flickY)).normalize(),
                                        cacheDirection: cacheDirection
                                    }
                                });
                            }
                        } else {
                            entity.processAction({
                                type: 'release',
                                data: {
                                    joystickPosition: joystickPosition.scalarMultiply(1 / maxRadius)
                                }
                            });
                        }
                    }

                    joystickPosition = new Vector2(0, 0);

                    if (secondTouch) {
                        touchData = secondTouch;
                        secondTouch = null;
                        touchTime = secondTouchTime;
                        secondTouchTime = 0;
                        oldPos = secondOldPos.clone();
                        secondOldPos.x = 0;
                        secondOldPos.y = 0;
                        cacheDirection = secondCacheDirection;
                    } else {
                        touchData = null;
                    }
                } else {
                    if (
                        secondFlickX * secondFlickX + secondFlickY * secondFlickY > (isAndroid ? 0.4 * 0.4 : 0.8 * 0.8) * flickSensitivity &&
                        secondTouchTime < (isAndroid ? 15 : 10) * flickSensitivity
                    ) {
                        // dash only in x or y direction
                        if (Math.abs(secondFlickX) >= Math.abs(secondFlickY)) {
                            secondFlickY = 0;
                        } else {
                            secondFlickX = 0;
                        }

                        if ((dashEnabled && secondFlickX) ||
                            (jumpEnabled && secondFlickY)
                        ) {
                            entity.processAction({
                                type: 'swipe',
                                data: {
                                    direction: (new Vector2(secondFlickX, secondFlickY)).normalize(),
                                    cacheDirection: secondCacheDirection
                                }
                            });
                        }
                    } else if (attackEnabled) {
                        entity.processAction({
                            type: 'tap',
                            data: {
                                duration: holdTimer
                            }
                        });
                        holdTimer = 0;
                    } else {
                            entity.processAction({
                                type: 'release',
                                data: {
                                    joystickPosition: joystickPosition.scalarMultiply(1 / maxRadius)
                                }
                            });
                        }

                    secondTouch = null;
                    secondFlickX = 0;
                    secondFlickY = 0;
                }
            }
        });
        var startedWalking = false;

        var resumeEvent = function () {
            touchData = null;
            secondTouch = null;
            Bento.input.resetPointers();
        };

        var component = {
            name: 'touchController',
            onSensitivityChange: function (sensitivity) {
                sensitivityFactor = sensitivity;
                maxRadius = (settings.maxRadius || 32) * sensitivityFactor;
                deadZone = (settings.deadZone || 8) * sensitivityFactor;
            },
            onFlickSensitivityChange: function (sensitivity) {
                flickSensitivity = sensitivity;
            },
            release: function () {
                touchData = null;
                secondTouch = null;
                Bento.input.resetPointers();
            },
            start: function (data) {
                Bento.objects.attach(cursor);
                EventSystem.on('enableControls', enableControls);
                EventSystem.on('disableControls', disableControls);
                EventSystem.on('startCutscene', disableControls);
                EventSystem.on('endCutscene', enableControls);
                EventSystem.on('onResume', resumeEvent);
                EventSystem.on('releaseTouches', component.release);
                EventSystem.on('touchcancel', component.release);
                EventSystem.on('sensitivityChanged', component.onSensitivityChange);
                EventSystem.on('flickSensitivityChanged', component.onFlickSensitivityChange);
            },
            destroy: function (data) {
                EventSystem.off('enableControls', enableControls);
                EventSystem.off('disableControls', disableControls);
                EventSystem.off('startCutscene', disableControls);
                EventSystem.off('endCutscene', enableControls);
                EventSystem.off('onResume', resumeEvent);
                EventSystem.off('releaseTouches', component.release);
                EventSystem.off('touchcancel', component.release);
                EventSystem.off('sensitivityChanged', component.onSensitivityChange);
                EventSystem.off('flickSensitivityChanged', component.onFlickSensitivityChange);
            },
            update: function (data) {
                // console.log('☞ TOUCHES', touchData, secondTouch);

                if (!isActive || !isEnabled) {
                    return;
                }

                if (secondTouch) {
                    startedWalking = false;
                }

                if (touchData || secondTouch) {
                    if (touchData) {
                        ++touchTime;
                    }
                    if (secondTouch) {
                        ++secondTouchTime;
                    }
                    if (limitJoystick() > deadZone && !secondTouch) {
                        holdTimer = 0;
                    } else if (Math.abs(secondFlickY) > 5 * (Utils.isNativeAndroid() ? 0.5 : 1.0)) {
                        holdTimer = 0;
                    } else {
                        if (!startedWalking && chargeEnabled) {
                            holdTimer++;
                        } else {
                            holdTimer = 0;
                        }
                    }

                    if (holdTimer > 0) {
                        entity.processAction({
                            type: 'hold',
                            data: {
                                duration: holdTimer
                            }
                        });
                    }

                    if (joystickPosition.magnitude() > deadZone && !secondTouch) {
                        entity.moveInDirection(joystickPosition.scalarMultiply(1 / maxRadius));
                        startedWalking = true;
                    }
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                entity.attach(clickable);
            },
            isActive: function () {
                return isActive;
            },
            canAttack: function (value) {
                attackEnabled = value;
            },
            canCharge: function (value) {
                chargeEnabled = value;
            },
            canDash: function (value) {
                dashEnabled = value;
            },
            canJump: function (value) {
                jumpEnabled = value;
            },
            canMove: function (value) {
                moveEnabled = value;
                if (!value) {
                    joystickPosition = new Vector2(0, 0);
                    startedWalking = false;
                }
            }
        };
        return component;
    };
});