/**
 * Touch Controller. Used to provide player input to a character.
 * Only active in portrait mode
 * @moduleName TouchController
 */
bento.define('components/controllers/touchcontroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/clickable',
    'globals',
    'bento/components/sprite',
    'managers/resolution'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Clickable,
    Globals,
    Sprite,
    ResolutionManager
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var maxRadius = settings.maxRadius || 32;
        var tapDistance = settings.tapDistance || 2;
        var deadZone = settings.deadZone || 8;

        var touchData = null;
        var joystickPosition = new Vector2(0, 0);
        var oldPos = new Vector2(0, 0);

        var holdTimer = 0;
        var flickY = 0;
        var flickX = 0;
        var touchTime = 0;

        var cursorDistance = settings.cursorDistance || 20;

        var isActive = true;

        var isEnabled = true;

        var moveEnabled = true;
        var attackEnabled = true;
        var chargeEnabled = true;
        var dashEnabled = true;

        var enableControls = function () {
            isEnabled = true;
        };

        var disableControls = function () {
            isEnabled = false;
            joystickPosition = new Vector2(0, 0);
            touchData = null;
        };

        var cursor = new Entity({
            z: Globals.Layers.GUI,
            name: 'cursor',
            position: new Vector2(0, 0),
            alpha: 0.5,
            components: [
                new Sprite({
                    originRelative: new Vector2(0.5, 0.5),
                    imageName: 'arrow'
                }), {
                    name: 'visualize',
                    update: function () {
                        var magnitude = limitJoystick();
                        var pos = joystickPosition.clone().scalarMultiply(cursorDistance / maxRadius);
                        cursor.position = entity.position.add(pos);
                        cursor.rotation = joystickPosition.angle();

                        if (magnitude < deadZone) {
                            cursor.visible = false;
                        } else {
                            cursor.visible = true;
                        }
                    }
                }
            ]
        });

        var limitJoystick = function () {
            var magnitude = joystickPosition.magnitude();
            if (joystickPosition.magnitude() > maxRadius) {
                joystickPosition.scalarMultiplyWith(maxRadius / magnitude);
            }
            return Math.min(magnitude, maxRadius);
        };

        var clickable = new Clickable({
            pointerDown: function (data) {
                // touch must be below the gui area
                if (data.position.y < 64) {
                    return;
                }

                if (touchData || !isActive || !isEnabled) {
                    return;
                }
                touchData = data;
                startedWalking = false;

                joystickPosition = new Vector2(0, 0);
                oldPos = data.position.clone();
                holdTimer = 0;
                touchTime = 0;
                flickY = 0;
                flickX = 0;
            },
            pointerMove: function (data) {
                if (!touchData || touchData.id !== data.id || !isActive || !isEnabled || !moveEnabled) {
                    return;
                }
                var difference = data.position.substract(oldPos);

                joystickPosition.addTo(difference);

                // limit
                limitJoystick();

                flickY = difference.y;
                flickX = difference.x;

                // move with touch
                oldPos = data.position.clone();
            },
            pointerUp: function (data) {
                if (!touchData || touchData.id !== data.id || !isActive || !isEnabled) {
                    return;
                }

                if (!startedWalking && attackEnabled) {
                    entity.processAction({
                        type: 'tap',
                        data: {
                            duration: holdTimer
                        }
                    });
                } else if (dashEnabled && flickX * flickX + flickY * flickY > (Utils.isNativeAndroid() ? 0.5 * 0.5 : 1.0 * 1.0) && touchTime < 10) {
                    // dash only in x or y direction
                    if (Math.abs(flickX) >= Math.abs(flickY)) {
                        flickY = 0;
                    } else {
                        flickX = 0;
                    }

                    entity.processAction({
                        type: 'swipe',
                        data: {
                            direction: (new Vector2(flickX, flickY)).normalize()
                        }
                    });
                }

                joystickPosition = new Vector2(0, 0);

                touchData = null;
            }
        });
        var startedWalking = false;

        var onOrientationChanged = function (e) {
            isActive = e.orientation === 'portrait';
            cursor.visible = isActive;
        };

        var component = {
            name: 'touchController',
            start: function (data) {
                Bento.objects.attach(cursor);
                EventSystem.on('orientationChanged', onOrientationChanged);
                onOrientationChanged({
                    orientation: ResolutionManager.getOrientation()
                });

                EventSystem.on('enableControls', enableControls);
                EventSystem.on('disableControls', disableControls);
                EventSystem.on('startCutscene', disableControls);
                EventSystem.on('endCutscene', enableControls);
            },
            destroy: function (data) {
                EventSystem.off('orientationChanged', onOrientationChanged);

                EventSystem.off('enableControls', enableControls);
                EventSystem.off('disableControls', disableControls);
                EventSystem.off('startCutscene', disableControls);
                EventSystem.off('endCutscene', enableControls);
            },
            update: function (data) {
                if (!isActive || !isEnabled) {
                    return;
                }
                if (touchData) {
                    touchTime++;
                    if (limitJoystick() > deadZone) {
                        holdTimer = 0;
                    } else {
                        if (!startedWalking && chargeEnabled) {
                            holdTimer++;
                        } else {
                            holdTimer = 0;
                        }
                    }

                    if (holdTimer > 0) {
                        entity.processAction({
                            type: 'hold',
                            data: {
                                duration: holdTimer
                            }
                        });
                    }

                    if (joystickPosition.magnitude() > deadZone) {
                        entity.moveInDirection(joystickPosition.scalarMultiply(1 / maxRadius));
                        startedWalking = true;
                    }
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                entity.attach(clickable);
            },
            isActive: function () {
                return isActive;
            },
            canAttack: function (value) {
                attackEnabled = value;
            },
            canCharge: function (value) {
                chargeEnabled = value;
            },
            canDash: function (value) {
                dashEnabled = value;
            },
            canMove: function (value) {
                moveEnabled = value;
                if (!value) {
                    joystickPosition = new Vector2(0, 0);
                    startedWalking = false;
                }
            }
        };
        return component;
    };
});