/**
 * 'Traditional' Controller. Used to provide player input to a character.
 * Only active in landscape mode
 * @moduleName TraditionalController
 */
bento.define('components/controllers/traditionalcontroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/clickable',
    'globals',
    'bento/components/sprite',
    'bento/gui/clickbutton',
    'managers/resolution'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Clickable,
    Globals,
    Sprite,
    ClickButton,
    ResolutionManager
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var maxRadius = settings.maxRadius || 32;
        var deadZone = settings.deadZone || 8;

        var touchData = null;
        var joystickPosition = new Vector2(0, 0);
        var oldPos = new Vector2(0, 0);

        var holdTimer = 0;
        var flickY = 0;
        var flickX = 0;
        var touchTime = 0;

        var cursorDistance = settings.cursorDistance || 20;

        var isHolding = false;

        var isActive = false;

        var isEnabled = true;

        var moveEnabled = true;
        var attackEnabled = true;
        var chargeEnabled = true;
        var dashEnabled = true;

        var enableControls = function () {
            isEnabled = true;
        };

        var disableControls = function () {
            isEnabled = false;
            joystickPosition = new Vector2(0, 0);
            touchData = null;
        };

        var cursor = new Entity({
            z: Globals.Layers.GUI,
            name: 'cursor',
            position: new Vector2(0, 0),
            alpha: 0.5,
            components: [
                new Sprite({
                    originRelative: new Vector2(0.5, 0.5),
                    imageName: 'arrow'
                }), {
                    name: 'visualize',
                    update: function () {
                        var magnitude = limitJoystick();
                        var pos = joystickPosition.clone().scalarMultiply(cursorDistance / maxRadius);
                        cursor.position = entity.position.add(pos);
                        cursor.rotation = joystickPosition.angle();

                        if (magnitude < deadZone) {
                            cursor.visible = false;
                        } else {
                            cursor.visible = true;
                        }
                    }
                }
            ]
        });

        var attackButton = new Entity({
            z: Globals.Layers.GUI,
            name: 'attackButton',
            position: new Vector2(0, 0),
            float: true,
            alpha: 0.75,
            boundingBox: new Rectangle(-32, -32, 64, 64),
            components: [
                new Sprite({
                    imageName: 'gui/fistbutton',
                    frameCountX: 1,
                    frameCountY: 1,
                    originRelative: new Vector2(0.5, 0.5)
                }),
                new Clickable({
                    onHold: function () {
                        attackButton.alpha = 1;
                    },
                    onClick: function (data) {
                        if (touchData && touchData.id === data.id) {
                            // should not be the touch that is used to walk
                            return;
                        }
                        isHolding = true;
                        attackButton.alpha = 1;
                    },
                    onHoldEnter: function (data) {
                        if (touchData && touchData.id === data.id) {
                            // should not be the touch that is used to walk
                            return;
                        }
                        isHolding = true;
                        attackButton.alpha = 1;
                    },
                    onHoldLeave: function (data) {
                        if (touchData && touchData.id === data.id) {
                            // should not be the touch that is used to walk
                            return;
                        }

                        if (isHolding) {
                            attackButton.alpha = 0.75;
                            isHolding = false;
                            entity.processAction({
                                type: 'tap',
                                data: {
                                    duration: holdTimer
                                }
                            });

                            holdTimer = 0;
                        }
                    },
                    onHoldEnd: function (data) {
                        if (touchData && touchData.id === data.id) {
                            // should not be the touch that is used to walk
                            return;
                        }
                        if (isHolding) {
                            attackButton.alpha = 0.75;
                            isHolding = false;
                            entity.processAction({
                                type: 'tap',
                                data: {
                                    duration: holdTimer
                                }
                            });

                            holdTimer = 0;
                        }
                    }
                })
            ]
        });

        var limitJoystick = function () {
            var magnitude = joystickPosition.magnitude();
            if (joystickPosition.magnitude() > maxRadius) {
                joystickPosition.scalarMultiplyWith(maxRadius / magnitude);
            }
            return Math.min(magnitude, maxRadius);
        };

        var clickable = new Clickable({
            pointerDown: function (data) {
                // touch must be below the gui area
                if (data.position.y < 32) {
                    return;
                }
                if (data.position.x > viewport.width - 64 && data.position.y > viewport.height - 64) {
                    return;
                }

                if (touchData || !isActive) {
                    return;
                }
                touchData = {
                    id: data.id
                };

                joystickPosition = new Vector2(0, 0);
                oldPos = data.position.clone();
                touchTime = 0;
                flickY = 0;
                flickX = 0;

                //if not attacking, send an 'empty' move action to cancel movement like automatic running
                if (!attackButton.getBoundingBox().hasPosition(data.position)) {
                    entity.moveInDirection(new Vector2(0, 0));
                }
            },
            pointerMove: function (data) {
                if (isHolding) {
                    return;
                }
                if (!touchData || touchData.id !== data.id || !isActive || !moveEnabled) {
                    return;
                }
                var difference = data.position.substract(oldPos);

                joystickPosition.addTo(difference);

                // limit
                limitJoystick();

                flickY = difference.y;
                flickX = difference.x;

                // move with touch
                oldPos = data.position.clone();
            },
            pointerUp: function (data) {
                if (!touchData || touchData.id !== data.id || !isActive) {
                    return;
                }

                if (dashEnabled && flickX * flickX + flickY * flickY > (Utils.isNativeAndroid() ? 1 * 1 : 4 * 4) && touchTime < 10) {
                    // dash only in x or y direction
                    if (Math.abs(flickX) >= Math.abs(flickY)) {
                        flickY = 0;
                    } else {
                        flickX = 0;
                    }

                    entity.processAction({
                        type: 'swipe',
                        data: {
                            direction: (new Vector2(flickX, flickY)).normalize()
                        }
                    });
                }

                joystickPosition = new Vector2(0, 0);

                touchData = null;
            }
        });

        var onOrientationChanged = function (e) {
            isActive = e.orientation === 'landscape';
            cursor.visible = isActive;
            attackButton.visible = isActive;

            if (isActive) {
                attackButton.position = new Vector2(viewport.width - 32, viewport.height - 32);
            } else {
                attackButton.position = new Vector2(-999, -999);
            }
        };

        var component = {
            name: 'traditionalController',
            start: function (data) {
                Bento.objects.attach(cursor);
                Bento.objects.attach(attackButton);

                EventSystem.on('orientationChanged', onOrientationChanged);
                onOrientationChanged({
                    orientation: ResolutionManager.getOrientation()
                });

                EventSystem.on('enableControls', enableControls);
                EventSystem.on('disableControls', disableControls);
                EventSystem.on('startCutscene', disableControls);
                EventSystem.on('endCutscene', enableControls);
            },
            destroy: function (data) {
                EventSystem.off('orientationChanged', onOrientationChanged);

                EventSystem.off('enableControls', enableControls);
                EventSystem.off('disableControls', disableControls);
                EventSystem.off('startCutscene', disableControls);
                EventSystem.off('endCutscene', enableControls);
            },
            update: function (data) {
                if (!isActive) {
                    return;
                }

                if (touchData) {
                    touchTime++;

                    if (!isHolding && joystickPosition.magnitude() > deadZone) {
                        entity.moveInDirection(joystickPosition.scalarMultiply(1 / maxRadius));
                    }
                }

                if (isHolding && chargeEnabled) {
                    holdTimer++;

                    if (holdTimer > 0) {
                        entity.processAction({
                            type: 'hold',
                            data: {
                                duration: holdTimer
                            }
                        });
                    }
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                entity.attach(clickable);
            },
            canAttack: function (value) {
                attackEnabled = value;
            },
            canCharge: function (value) {
                chargeEnabled = value;
            },
            canDash: function (value) {
                dashEnabled = value;
            },
            canMove: function (value) {
                moveEnabled = value;
                if(!value) {
                    joystickPosition = new Vector2(0, 0);
                }
            }
        };
        return component;
    };
});