/**
 * Damage counter module
 * Spawns numbers reflecting the damage a character receives
 *
 * @moduleName DamageCounter
 */
bento.define('components/damagecounter', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/text',
    'globals',
    'components/leftshadow',
    'modules/localization',
    'gui/counter'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Text,
    Globals,
    LeftShadow,
    Localization,
    Counter
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var DamageFloater = function () {
        var viewport = Bento.getViewport();

        var tween = function () {
            new Tween({
                from: entity.position.y + 10,
                to: entity.position.y - 10,
                'in': 20,
                ease: Tween.EASEOUTCUBIC,
                onUpdate: function (v, t) {
                    entity.position.y = v;
                }
            });
            new Tween({
                from: entity.position.y - 10,
                to: entity.position.y - 30,
                'in': 30,
                delay: 30,
                ease: Tween.EASEINCUBIC,
                onUpdate: function (v, t) {
                    entity.position.y = v;
                },
                onComplete: function () {
                    Bento.objects.remove(entity);
                    entity.active = false;
                }
            });


            new Tween({
                from: 0,
                to: 1,
                'in': 10,
                ease: Tween.LINEAR,
                onUpdate: function (v, t) {
                    entity.alpha = v;
                }
            });

            new Tween({
                from: 1,
                to: 0,
                'in': 20,
                delay: 40,
                ease: Tween.EASEINQUAD,
                onUpdate: function (v, t) {
                    entity.alpha = v;
                }
            });

            new Tween({
                from: 0,
                to: 1,
                'in': 40,
                ease: Tween.EASEOUTELASTIC,
                onUpdate: function (v, t) {
                    entity.scale.x = v;
                    entity.scale.y = v;
                }
            });


            //check if the floater is outside of the players visible area, and move it to the players viewport
            //however if the floater is too far away, don't bother

            var ww = (text.dimension.width - 8);
            var tx = Utils.clamp(viewport.x + ww, entity.position.x, viewport.x + viewport.width - ww);

            if (Math.abs(tx - entity.position.x) < 50 && (tx - entity.position.x) !== 0) {

                new Tween({
                    from: entity.position.x,
                    to: tx,
                    'in': 12,
                    ease: Tween.EASEOUTBACK,
                    onUpdate: function (v, t) {
                        entity.position.x = v;
                    }
                });
            }
        };

        /*var text = new Text({
            z: 0,
            position: new Vector2(0, 0),
            text: '',
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            lineWidth: 1,
            strokeStyle: ['#160d32'],
            innerStroke: false,
            pixelStroke: true // workaround for Cocoon bug
        }).attach(new LeftShadow({}));*/

        var text = new Counter({
            name: 'damageCounter',
            image: Bento.assets.getImage('gui/counters/damage'),
            frameWidth: 12,
            frameHeight: 14,
            spacing: new Vector2(-4, 0),
            align: 'center',
            animations: {
                '0': {
                    frames: [0]
                },
                '1': {
                    frames: [1]
                },
                '2': {
                    frames: [2]
                },
                '3': {
                    frames: [3]
                },
                '4': {
                    frames: [4]
                },
                '5': {
                    frames: [5]
                },
                '6': {
                    frames: [6]
                },
                '7': {
                    frames: [7]
                },
                '8': {
                    frames: [8]
                },
                '9': {
                    frames: [9]
                },
                '-': {
                    frames: [10]
                },
                'M': {
                    frames: [11]
                },
                'A': {
                    frames: [12]
                },
                'X': {
                    frames: [13]
                }
            }
        });
        var textHurt = new Counter({
            name: 'damageCounter',
            image: Bento.assets.getImage('gui/counters/hurt'),
            frameWidth: 12,
            frameHeight: 14,
            spacing: new Vector2(-4, 0),
            align: 'center',
            animations: {
                '0': {
                    frames: [0]
                },
                '1': {
                    frames: [1]
                },
                '2': {
                    frames: [2]
                },
                '3': {
                    frames: [3]
                },
                '4': {
                    frames: [4]
                },
                '5': {
                    frames: [5]
                },
                '6': {
                    frames: [6]
                },
                '7': {
                    frames: [7]
                },
                '8': {
                    frames: [8]
                },
                '9': {
                    frames: [9]
                },
                '-': {
                    frames: [10]
                },
                'M': {
                    frames: [11]
                },
                'A': {
                    frames: [12]
                },
                'X': {
                    frames: [13]
                }
            }
        });

        var entity = new Entity({
            z: Globals.Layers.DAMAGE_FLOATERS,
            name: 'damageFloater',
            position: new Vector2(0, 0),
            family: ['gui'],
            components: [
                text,
                textHurt
            ]
        }).extend({
            show: function (position, damage, type) {
                if (type === 'hurt') {
                    textHurt.setValue(damage);
                    textHurt.visible = true;
                    text.visible = false;
                } else {
                    text.setValue(damage);
                    text.visible = true;
                    textHurt.visible = false;
                }
                entity.position = position.clone();

                Bento.objects.attach(entity);
                tween();
            }
        });

        return entity;
    };

    var __damageFloaterPool = [];

    var getDamageFloater = function () {
        var floater = new DamageFloater({});

        floater.active = true;

        return floater;
    };

    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var movable;

        var onHurt = function (data) {
            var dmg = data.damage;
            if (dmg <= 0) {
                return;
            }

            if (data.maxDamage) {
                // dmg = Localization.getText('maximumDamage');
                dmg = 'MAX'; // for translation we do need actual text objects...
            }

            var floater = getDamageFloater();
            var position = entity.position.clone();
            position.addTo(new Vector2(Math.random() * 16 - 8, -entity.dimension.height));
            if (movable) {
                position.addTo(new Vector2(movable.velocity.x * 1.5, (movable.velocity.y + movable.velocityZ) * 1.5));
            }

            // TODO: custom colours which get sent in with data
            floater.show(position, dmg, settings.type);
        };

        var component = {
            name: 'damageCounter',
            start: function (data) {
                entity.getComponent('punchable', function (punchable) {
                    punchable.onHurt(onHurt);
                });

                entity.getComponent('movable', function (mvb) {
                    movable = mvb;
                });
            },
            destroy: function (data) {},
            update: function (data) {},
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            },
            block: function () {
                // var dmg = 'BLOCK';
                // var floater = getDamageFloater();
                // var position = entity.position.clone();
                // position.addTo(new Vector2(Math.random() * 16 - 8, -entity.dimension.height));
                // if (movable) {
                //     position.addTo(new Vector2(movable.velocity.x * 1.5, (movable.velocity.y + movable.velocityZ) * 1.5));
                // }

                // floater.show(position, dmg, '#26bfff', '#1db8f8');
            }
        };
        return component;
    };
});