/**
 * Entities with this component are able to be destroyed
 * @moduleName Destructible
 */
bento.define('components/destructible', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'entities/consumable',
    'entities/coin',
    'modules/matchmanager',
    'entities/props/particle',
    'components/coindropper',
    'components/itemdropper'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Consumable,
    Coin,
    MatchManager,
    Particle,
    CoinDropper,
    ItemDropper
) {
    'use strict';
    /**
     * Entities with this component are able to be destroyed
     * @param  {Object} settings            Options for setting up this component.
     * @param  {Object} [settings.health=1]
     */
    return function (settings) {
        settings = settings || {};
        var viewport = Bento.getViewport();
        var screenMid = new Vector2(viewport.width / 2, viewport.height / 2);
        var entity;
        var punchable;
        var health = Utils.isDefined(settings.health) ? parseInt(settings.health) : 1;
        var coins = Utils.isDefined(settings.coins) ? parseInt(settings.coins) : 0;
        var currentPunch = null;
        var hits = 1;

        var item = settings.item;

        var onHit = function (data) {
            new Tween({
                from: 2,
                to: 0,
                'in': 15,
                ease: Tween.LINEAR,
                onUpdate: function (v, t) {
                    entity.transform.x = Utils.getRandomRangeFloat(-v, v);
                    entity.transform.y = Utils.getRandomRangeFloat(-v, v);
                },
                onComplete: function () {
                    entity.transform.x = 0;
                    entity.transform.y = 0;
                }
            });
            punchable.depleteHealth(data.punch);

            // sync hit
            EventSystem.fire('punchHit', {
                punch: data.punch,
                victim: entity,
                attacker: data.punch.attacker
            });

            hits -= data.punch.multiplier;
            if (hits <= 0) {
                punchable.kill(data.punch);
            }
        };

        var onDeath = function (data) {
            var punch = data.punch;
            Bento.objects.remove(entity);

            // spawn a particle
            var particleSpritesheet = settings.spriteSheet;
            var particle = new Particle({
                punch: punch,
                position: entity.position.clone(),
                spriteSheet: particleSpritesheet
            });
            if (!particleSpritesheet) {
                // reuse sprite?
                var sprite = entity.getComponent('sprite');
                entity.remove(sprite);
                particle.attach(sprite);
            }
            Bento.objects.attach(particle);

            if (!punch.emulated) {
                EventSystem.fire('quest-destruction', {
                    count: 1
                });
            }
        };

        var component = {
            name: 'destructible',
            attached: function (data) {
                entity = data.entity;

                if (entity.family.indexOf('destructible') < 0) {
                    entity.family.push('destructible');
                }

                // attach a coindropper if needed
                if (coins) {
                    entity.attach(new CoinDropper({
                        coins: coins,
                        downOnly: true
                    }));
                } else if (item) {
                    entity.attach(new ItemDropper({
                        item: item
                    }));
                }
            },
            start: function (data) {
                punchable = entity.getComponent('punchable');

                punchable.onHit(onHit);
                punchable.onDeath(onDeath);
                punchable.setMaxHealth(health);
                punchable.boundingBox = entity.boundingBox;

            },
            destroy: function (data) {},
            // update: function (data) {},
            // draw: function (data) {},
        };

        return component;
    };
});