/**
 * Grabbing enemies is something a player should always be able to do, but remotes won't listen to
 * grab states automatically (because it's controlled by 1 local player)
 * @moduleName Grabable
 */
bento.define('components/grabable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'states/standardenemy',
    'modules/matchmanager',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    StandardEnemyStates,
    MatchManager,
    Punch
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var Grabable = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var enemyController;
        var stateMachine;
        var grabbed = null;
        var duration = settings.grabDuration || 60 * 3;
        var component = {
            name: 'grabable',
            //
            start: function (data) {
                enemyController = entity.getComponent('enemyController');

                // setup state machine
                stateMachine = entity.getComponent('stateMachine');
                // overwrite idle, hurt and walk states
                stateMachine.addStates(Grabable.states);

                stateMachine.data.grabDuration = duration;
                stateMachine.data.grabable = component;

                // grab listener for non-remotes
                if (enemyController) {
                    if (enemyController.isLocallyControlled) {
                        MatchManager.onMessage('grabbed-' + entity.name, component.onGrabMessage);
                    } else {
                        MatchManager.onMessage('grabbedThrow-' + entity.name, component.onGrabMessage);
                    }
                }
            },
            onGrabMessage: function (msg) {
                // console.log(onGrab)
                if (msg.grabbed) {
                    // component.setGrabbed(msg.grabber);
                    stateMachine.setState('grabbed', {
                        grabber: msg.grabber
                    });
                } else if (msg.grabbedHurt) {
                    stateMachine.setState('grabbedHurt');
                } else if (msg.grabbedThrow) {
                    component.setGrabbed(null);
                    stateMachine.setState('grabbedThrow', JSON.parse(msg.transitionData));
                } else if (msg.release) {
                    component.release(null);
                }
                return true;
            },
            destroy: function (data) {
                MatchManager.offMessage('grabbed-' + entity.name, component.onGrabMessage);
                MatchManager.offMessage('grabbedThrow-' + entity.name, component.onGrabMessage);
            },
            attached: function (data) {
                entity = data.entity;
            },
            setGrabbed: function (grabberId) {
                var oldGrabber;
                // was there someone else previously grabbing?
                if (grabbed !== null && grabbed !== grabberId) {
                    // the previous person should release
                    oldGrabber = Utils.getPlayerById(grabbed);
                    if (oldGrabber) {
                        oldGrabber.getComponent('stateMachine', function (playerStateMachine) {
                            playerStateMachine.data.currentlyGrabbed = null;
                            if (playerStateMachine.stateName === 'grab' || playerStateMachine.stateName === 'grabAttack') {
                                playerStateMachine.setState('idle');
                            }
                        });
                    }
                }
                // update new id
                grabbed = grabberId;

                if (grabbed === null) {
                    // check if not being grabbed still
                    if (
                        stateMachine.stateName === 'grabbed' ||
                        stateMachine.stateName === 'grabbedHurt'
                    ) {
                        stateMachine.setState('idle');
                    }
                }
            },
            release: function (player) {
                if (enemyController && !enemyController.isLocallyControlled && player && !player.isRemote) {
                    MatchManager.sendData({
                        release: true
                    }, 'grabbed-' + entity.name);
                }
                component.setGrabbed(null);
            },
            isGrabbed: function () {
                return grabbed;
            },
            update: function () {
                // bit of a hack for now, but need to overwrite the standard enemy states and this start comes later
                stateMachine.addStates(Grabable.states);
                component.update = null;
            }
        };
        return component;
    };
    // specialized states
    Grabable.states = {
        grabbed: {
            start: function (data, transitionData) {
                StandardEnemyStates.grabbed.start(data, transitionData);

                var grabber = data.lastGrabbed;
                var player = Utils.getPlayerById(grabber);

                data.grabable.setGrabbed(grabber);

                // sync grabbed state
                if (data.enemyController && !data.enemyController.isLocallyControlled && player && !player.isRemote) {
                    MatchManager.sendData({
                        grabbed: true,
                        grabber: transitionData.grabber,
                    }, 'grabbed-' + data.entity.name);
                }
                EventSystem.fire('grabbed', {
                    player: player,
                    entity: data.entity
                });
            },
            update: StandardEnemyStates.grabbed.update,
            end: function (data) {
                StandardEnemyStates.grabbed.end(data);
                var nextState = data.stateMachine.stateName;
                var playerMultiScale;
                var grabber = data.grabable.isGrabbed();
                if (nextState !== 'grabbed' && nextState !== 'grabbedHurt' && nextState !== 'grabbedThrow') {
                    // the enemy was released by something else
                    var player = Utils.getPlayerById(grabber);
                    data.grabable.release(player);

                    // blocked
                    if (nextState === 'block') {
                        // make a little jump
                        if (player) {
                            playerMultiScale = player.getComponent('multiScale');
                            data.movable.velocity.x = playerMultiScale.direction.x * 4;
                            data.movable.velocityZ = -4;
                        }
                    }
                }
            }
        },
        grabbedHurt: {
            start: function (data, transitionData) {
                if (data.grabable && data.grabable.isGrabbed() !== null) {
                    data.movable.velocity.x = 0;
                }

                data.punchDirection = Utils.sign(data.movable.velocity.x) * -1 || data.multiScale.direction.x * -1;

                if (Utils.hasAsset('spritesheets', data.rootPath + 'grabbed')) {
                    data.sprite.setSpriteSheet(data.rootPath + 'grabbed');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'flyup');
                }

                new Tween({
                    from: 0.5,
                    to: 0,
                    'in': 30,
                    ease: Tween.ELASTIC,
                    alpha: -5,
                    beta: 2,
                    onUpdate: function (v, t) {
                        data.multiScale.hurt.x = 1 + v;
                        data.multiScale.hurt.y = 1 - v;
                    },
                    onComplete: function () {}
                });
            },
            update: StandardEnemyStates.grabbed.update
        },
        // grabbed throw is a copy of the faint state
        grabbedThrow: {
            ignoreSync: true, // since this is similar to faint, don't sync it because it's very glitchy, instead we will sync this manually
            start: function (data, transitionData) {
                StandardEnemyStates.faint.start(data, transitionData);
                // sync grabbed throw state
                var grabber = data.lastGrabbed;
                var player = Utils.getPlayerById(grabber);
                if (data.enemyController && !data.enemyController.isLocallyControlled && player && !player.isRemote) {
                    MatchManager.sendData({
                        grabbedThrow: true,
                        transitionData: JSON.stringify(transitionData)
                    }, 'grabbed-' + data.entity.name);
                } else if (data.enemyController && data.enemyController.isLocallyControlled && player && !player.isRemote) {
                    // manually send grabbedThrow because this state is ignoring sync
                    MatchManager.sendData({
                        grabbedThrow: true,
                        transitionData: JSON.stringify(transitionData)
                    }, 'grabbedThrow-' + data.entity.name);
                }
                // invincible while flying
                data.punchable.isInvincible = true;

                // damage other enemies by throwing into them
                // create a punch -> only for the local player, the rest will sync the hits
                if (player && !player.isRemote) {
                    data.sync({
                        throwDamage: Math.round(player.getComponent('stateMachine').data.stats.attack / 4)
                    });
                    data.grabbedPunch = new Punch({
                        target: ['enemy'],
                        attack: data.syncData.throwDamage,
                        attacker: player,
                        multiplier: 1,
                        bonusAttack: 0,
                        stamina: 1, // knocks down other enemies
                        parent: data.entity,
                        direction: data.multiScale.direction.x,
                        pushBack: new Vector2(data.multiScale.direction.x, 1.5),
                        boundingBox: new Rectangle(-16, -6, 32, 12),
                        position: data.entity.position.clone(),
                        duration: 60,
                        freeze: 3,
                        height: 32,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3-small',
                            sfx: 'sfx_punch',
                            onComplete: function () {
                                // turn around // NOTE: syncing issues
                                // data.movable.velocity.x *= -0.25;
                                // data.movable.velocityZ -= 1;
                            }
                        })
                    }).attach({
                        name: 'cleanUp',
                        destroy: function (d) {
                            if (data.grabbedPunch === d.entity) {
                                data.grabbedPunch = null;
                            }
                        }
                    });
                    Bento.objects.attach(data.grabbedPunch);
                }
                data.grabable.setGrabbed(null);
            },
            update: StandardEnemyStates.faint.update,
            end: function (data) {
                data.punchable.isInvincible = false;
                if (data.grabbedPunch) {
                    Bento.objects.remove(data.grabbedPunch);
                }
            }
        }
    };

    return Grabable;
});