/**
 * Health bar for enemies
 * @moduleName HealthBar
 */
bento.define('components/healthbar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/components/nineslice',
    'bento/tween',
    'gui/main/levelcount'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Entity,
    EventSystem,
    Utils,
    NineSlice,
    Tween,
    LevelCount
) {
    'use strict';

    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var healthBar;
        var healthBarImage;
        var barPositioning;
        var healthBarSpacing = 4;
        var frameWidth = 27;
        var frameHeight = 7;
        var alphaTween;
        var currentHealth = 0;
        var alwaysVisible = settings.alwaysVisible || false;

        var addHealthBar = function () {
            frameHeight = settings.height || entity.dimension.height;

            var healthBarBehavior = {
                follow: entity,
                offsetY: frameHeight + healthBarSpacing,
                offsetX: frameWidth / 2,
                update: function (data) {
                    var parent = data.entity;

                    parent.position.x = this.follow.position.x - this.offsetX;
                    parent.position.y = this.follow.position.y - this.offsetY + 8;
                    parent.transform.y = this.follow.transform.y;

                    healthBar.z = entity.z;
                }
            };
            healthBarImage = new Sprite({
                imageName: 'gui/main/hp'
            });

            barPositioning = new Entity({
                name: 'barpositioning',
                position: new Vector2(2, 2),
                components: [
                    healthBarImage
                ]
            });


            var levelIcon = new Entity({
                name: 'levelIcon',
                position: new Vector2(0, -7),
                components: [
                    new Sprite({
                        imageName: 'gui/main/lvl'
                    })
                ]
            });

            var lvlCount = new LevelCount({
                level: settings.level
            });

            healthBar = new Entity({
                name: 'healthBarEntity',
                z: entity.z,
                position: new Vector2(0, -frameHeight),
                family: ['gui'],
                alpha: settings.alwaysVisible ? 1 : 0,
                components: [
                    new Sprite({
                        imageName: 'gui/main/hpbar'
                    }),
                    barPositioning,
                    levelIcon,
                    lvlCount,
                    healthBarBehavior
                ]
            });

            // Health bar can't be attached to enemy entity, because some weird things
            // happen with its x-scale after getting hit and the health bar glitches out
            Bento.objects.attach(healthBar);
        };

        var component = {
            name: 'healthBar',
            /**
             * Adds the healthBar entity above the character
             * @param  {Object} opts                            Options for settings up the health bar.
             * @param  {String} opts.rootPath                   Root path of the character's sprites. Needed to get idle frame height to calculate where to put the health bar.
             * @param  {Number} [opts.spacing=healthBarSpacing] How much vertical spacing there should be between the health bar and the character.
             * @param  {Number} [opts.width=frameWidth]         The width of the health bar.
             */
            createHealthBar: function (opts) {
                opts = opts || {};
                healthBarSpacing = Utils.isNotEmpty(opts.spacing) ? opts.spacing : healthBarSpacing;
                frameWidth = Utils.isNotEmpty(opts.width) ? opts.width : frameWidth;
                addHealthBar();
            },
            attached: function (data) {
                entity = data.entity;
            },
            start: function (data) {
                var punchable = entity.getComponent('punchable');
                if (punchable) {
                    currentHealth = punchable.getHealth();

                    // register listener to the punchable
                    punchable.onHealthChange(function (params) {
                        var didIncrease = params.health > currentHealth;
                        currentHealth = params.health;
                        component.setHealth(params.relativeHealth, didIncrease);
                    });
                }

                component.createHealthBar(settings);

                // hide lvl if not defined
                if (!Utils.isDefined(settings.level)) {
                    healthBar.getComponent('levelIcon', function (levelIcon) {
                        levelIcon.visible = false;
                    });
                    healthBar.getComponent('levelCount', function (lvlCount) {
                        lvlCount.visible = false;
                    });
                }

                if (settings.health) {
                    currentHealth = settings.health;
                    component.setHealth(currentHealth / settings.maxHealth);

                }
            },
            update: function () {
                healthBar.visible = (healthBar.alpha > 0);
            },
            destroy: function (data) {
                Bento.objects.remove(healthBar);
            },
            setFill: function (healthRatio) {
                healthRatio = Utils.clamp(0, healthRatio, 1);
                var width = 22 * healthRatio;
                width = Math.ceil(width);
                barPositioning.scale.x = width;

            },
            setHealth: function (healthRatio, ignoreTween) {
                if (alphaTween) {
                    Bento.objects.remove(alphaTween);
                    alphaTween = null;
                }

                this.setFill(healthRatio);

                if (healthRatio <= 0) {
                    healthBar.alpha = 0;
                    return;
                }

                if (!ignoreTween && !alwaysVisible) {
                    healthBar.alpha = 1;
                    alphaTween = new Tween({
                        from: 1,
                        to: 0,
                        delay: 60,
                        'in': 60,
                        ease: Tween.EASEINQUAD,
                        onUpdate: function (v, t) {
                            healthBar.alpha = v;
                        }
                    });
                }
                if (alwaysVisible) {
                    healthBar.alpha = 1;
                }
            },
            // only used for players
            maxHealth: settings.maxHealth
        };

        return component;
    };
});