/**
 * Stops drawing objects below the parent entity
 * Its position must be the topleft corner and the alpha 1
 * Doesn't check if the object dimension itself fills up the viewport
 * @moduleName HideBelow
 */
bento.define('components/hidebelow', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var parent;
        var isHiding = false;
        // some objects could be centered instead of the topleft corner
        var centered = settings.centered || 0;
        var originalVisibleState = {};
        var hide = function () {
            // console.log('hide');
            isHiding = true;
            Utils.forEach(Bento.objects.getObjects(), function (object, i, l, breakLoop) {
                var isBelow;
                if (object === parent) {
                    breakLoop();
                    return;
                }
                if (object && object.isEntity && object.name) {
                    if (object.z < entity.z) {
                        isBelow = true;
                    } else if (object.z === entity.z && object.rootIndex < entity.rootIndex) {
                        isBelow = true;
                    }

                    if (isBelow) {
                        originalVisibleState[object.id] = {
                            entity: object,
                            visible: object.visible
                        };
                        object.visible = false;
                    }
                }
            });
        };
        var show = function () {
            isHiding = false;
            Utils.forEach(originalVisibleState, function (object, i, l, breakLoop) {
                object.entity.visible = object.visible;
            });
            originalVisibleState = {};
            // console.log('show');
        };
        var hideChecker = {
            z: 0,
            updateWhenPaused: Infinity,
            // using draw function because the update function causes android to show black flashes
            draw: function (data) {
                var worldPosition = entity.getWorldPosition();
                var shouldHide = (
                    Math.floor(worldPosition.x) === Math.floor(viewport.x + centered * viewport.width / 2) &&
                    Math.floor(worldPosition.y) === Math.floor(viewport.y + centered * viewport.height / 2) &&
                    parent.alpha === 1 &&
                    entity.alpha === 1
                );

                if (!isHiding && shouldHide) {
                    hide();
                } else if (isHiding && !shouldHide) {
                    show();
                }
            }
        };
        var component = {
            name: 'hideBelow',
            start: function (data) {
                // find root parent
                while (entity.parent) {
                    entity = entity.parent;
                }
                hideChecker.z = entity.z - 1;
                Bento.objects.attach(hideChecker);
            },
            destroy: function (data) {
                if (isHiding) {
                    show();
                }
                Bento.objects.remove(hideChecker);
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                parent = entity;
            },
            unHide: function () {
                // temporarily unhide
                Utils.forEach(originalVisibleState, function (object, i, l, breakLoop) {
                    object.entity.visible = object.visible;
                });
            },
            reHide: function () {
                // undo unhide
                Utils.forEach(originalVisibleState, function (object, i, l, breakLoop) {
                    object.entity.visible = false;
                });
            }
        };
        return component;
    };
});