/**
 * Movable component
 * TODO collisions
 */
bento.define('components/movable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var shouldIgnore = function (ignoredArray, familyArray) {
        var i, family;

        if (!familyArray) {
            return false;
        }

        for (i = 0; i < familyArray.length; ++i) {
            family = familyArray[i];
            if (ignoredArray.indexOf(family) >= 0) {
                return true;
            }
        }
        return false;
    };
    var VEC_NULL = new Vector2(0, 0); //prevent creating 2 new vectors per frame per movable

    var Movable = function (settings) {
        this.name = 'movable';
        this.paused = false;
        this.entity = null;

        this.boundingBox = settings.boundingBox || new Rectangle(0, 0, 1, 1);

        // colliding families
        this.solids = Bento.objects.getByFamily('solid');
        this.familiesToCollide = settings.familiesToCollide || [];

        this.height = settings.height || 0;
        this.gravity = Utils.isDefined(settings.gravity) ? settings.gravity : 0.15;
        this.grounded = true;
        this.velocityZ = settings.velocityZ || 0;

        this.friction = settings.friction || new Vector2(1, 1);
        this.bounce = settings.bounce || new Vector2(0, 0);
        this.frictionZ = settings.frictionZ || 1;
        this.bounceZ = settings.bounceZ || 0;
        this.canBounce = settings.canBounce || false;


        this.velocity = settings.velocity || new Vector2(0, 0);
        this.maxVelocity = settings.maxVelocity || new Vector2(Infinity, Infinity);

        this.collided = [];
        this.tileColliders = null;
        this.collisionListeners = [];

        this.ignoreCollisions = false;

        this.settings = settings || {};
        this.settings.ignoreSolids = this.settings.ignoreSolids || false;
        this.settings.onCollide = this.settings.onCollide || function () {};
        this.settings.useHshg = this.settings.useHshg || false;
        this.settings.ignoredFamilies = this.settings.ignoredFamilies || [];

        this.onLanded = settings.onLanded;
    };

    Movable.prototype.onCollide = function () {
        if (this.ignoreCollisions) {
            return;
        }

        var collided = this.collided;
        var entity = this.entity;
        var velocity = this.velocity;
        var settings = this.settings;
        var boundingBox = this.boundingBox;
        var self = this;

        var ii;
        var other;
        var otherBox;
        var diffBox;

        var direction = new Vector2(Utils.sign(velocity.x), Utils.sign(velocity.y));
        var normalizedSpeed = velocity.clone().normalize();

        var pushX = function (dir) {
            // pushes out entity in x
            // dir = -1 or 1
            entity.position.x += diffBox.width * dir;
        };
        var pushY = function (dir) {
            // pushes out entity in y
            // dir = -1 or 1
            entity.position.y += diffBox.height * dir;
        };
        var isColliding = function () {
            return other.collidesWith({
                rectangle: boundingBox.offset(entity.position)
            });
        };
        var newSpeed = function (dir) {
            // calculates new speed: bounce in normal direction, friction in perpendicular
            // direction should be 'x' or 'y'
            var perpendicular = (dir === 'x' ? 'y' : 'x');
            velocity[dir] *= -self.bounce[dir];
            velocity[perpendicular] *= self.friction[perpendicular];

        };

        for (ii = 0; ii < collided.length; ii++) {
            other = collided[ii];
            otherBox = other.getBoundingBox();

            // polygon?
            if (otherBox.isPolygon) {
                // push out in reverse direction until not colliding
                entity.position.x -= normalizedSpeed.x;
                entity.position.y -= normalizedSpeed.y;
                while (isColliding()) {
                    entity.position.x -= normalizedSpeed.x;
                    entity.position.y -= normalizedSpeed.y;
                }
                // temp: reverse direction
                // todo: find out which direction the bounce really is
                velocity.x *= -settings.bounce.x;
                velocity.y *= -settings.bounce.y;
            } else if (other.isTriangle && other.collisionData) {
                // push out of a topright triangle or bottomleft

                // only x or y direction (pick the shortest path)
                var entityBox = boundingBox.offset(entity.position);
                var collDataX = other.collisionData.x;
                var collDataY = other.collisionData.y;
                var diffX = Math.min(collDataX - collDataY, collDataX);
                var diffYBottom;
                var diffYTop;
                if (other.isTriangle < 0) {
                    diffYBottom = Math.abs(otherBox.height - collDataY);
                    diffYTop = Math.abs(collDataY + boundingBox.height);
                } else {
                    diffYBottom = Math.abs(boundingBox.height - (collDataY - otherBox.height));
                    diffYTop = Math.abs(collDataY);
                }
                var diffY = Math.min(diffYBottom, diffYTop);
                diffBox = new Rectangle(0, 0, diffX, diffY);
                if (Math.abs(diffX) < Math.abs(diffY)) {
                    pushX(-1);
                } else {
                    if (diffY === diffYBottom) {
                        pushY(1);
                    } else {
                        pushY(-1);
                    }
                }

                if (this.canBounce) {
                    // reflect
                    velocity.x *= -self.bounce.x;
                    velocity.y *= -self.bounce.y;
                    velocity.x *= self.friction.y;
                    velocity.y *= self.friction.x;
                } else {
                    velocity.y = 0;
                    velocity.x = 0;
                }
            } else {
                // normal rectangle collision
                diffBox = boundingBox.offset(entity.position).intersection(otherBox);
                if (diffBox.width === 0 && diffBox.height === 0) {
                    // not colliding?
                    // hshg is wrong apparently? or collision already resolved because of another collision object
                    continue;
                }

                // push out of x or y?
                // push out based on box size
                if (diffBox.width < diffBox.height) {
                    pushX(-direction.x);
                    // check
                    if (isColliding()) {
                        // failed!
                        // revert x and try y instead
                        pushX(direction.x);
                        pushY(-direction.y);
                        //
                        if (this.canBounce) {
                            newSpeed('y');
                        } else {
                            velocity.y = 0;
                        }
                    } else {
                        if (this.canBounce) {
                            newSpeed('x');
                        } else {
                            velocity.x = 0;
                        }
                    }
                } else {
                    pushY(-direction.y);
                    if (isColliding()) {
                        pushY(direction.y);
                        pushX(-direction.x);
                        if (this.canBounce) {
                            newSpeed('x');
                        } else {
                            velocity.x = 0;
                        }
                    } else {
                        if (this.canBounce) {
                            newSpeed('y');
                        } else {
                            velocity.y = 0;
                        }
                    }
                }
            }
        }

        //send collision data to listeners
        settings.onCollide({
            entity: entity,
            others: collided,
            other: other
        });

        for (ii = 0; ii < this.collisionListeners.length; ii++) {
            this.collisionListeners[ii]({
                entity: entity,
                others: collided,
                other: other
            });
        }
    };

    Movable.prototype.update = function (data) {
        var entity = this.entity;
        var velocity = this.velocity;
        var normalizedSpeed = velocity.clone().normalize();
        var settings = this.settings;
        var collided = this.collided;
        var self = this;

        if (this.paused) {
            return;
        }

        // cap velocity
        if (this.maxVelocity) { //TODO? default maxVelocity is Infinity, which still causes 2 clamps to be used every frame
            velocity.x = Utils.clamp(-this.maxVelocity.x, velocity.x, this.maxVelocity.x);
            velocity.y = Utils.clamp(-this.maxVelocity.y, velocity.y, this.maxVelocity.y);
        }

        // move entity
        entity.position.x += velocity.x * data.speed;
        entity.position.y += velocity.y * data.speed;

        //update 'height'
        this.height += this.velocityZ * data.speed;
        if (this.grounded && this.height !== 0) {
            this.grounded = false;
        }
        if (!this.grounded) {
            this.velocityZ += this.gravity * data.speed;
        }

        if (this.height >= 0 && this.velocityZ > 0) {
            this.height = 0;
            // TODO: this part is a bit wonky...
            this.velocityZ *= -this.bounceZ;
            this.velocity.x *= this.frictionZ;
            this.velocity.y *= this.frictionZ;

            if (!this.grounded && this.onLanded) {
                this.onLanded();
            }
            if (this.bounceZ === 0) {
                this.grounded = true;
            }
            // set to 0 under a certain threshold to stop the shaking
            else if (this.bounceZ !== 0 && Math.abs(this.velocityZ) < 0.25) {
                this.velocityZ = 0;
                this.grounded = true;
                if (this.frictionZ !== 1) {
                    this.velocity.x = 0;
                    this.velocity.y = 0;
                }
            }
        } else if (this.height > 0 || this.velocityZ !== 0) {
            this.grounded = false;
        }
        entity.transform.y = this.height;


        //check collisions
        if (this.tileColliders && !settings.ignoreSolids) {
            this.tileColliders.collidesWith(this.boundingBox.offset(entity.position), VEC_NULL, function (otherEntity) {
                collided.push(otherEntity);
                if (settings.onCollideHash) {
                    settings.onCollideHash(otherEntity);
                }
            });
        }

        // collision with solids
        // TODO use boundingBox instead of entity!
        if (!this.settings.useHshg && !settings.ignoreSolids) {
            var bb = this.boundingBox.offset(entity.position);
            var families = this.solids;
            Utils.forEach(this.familiesToCollide, function (family) {
                families = families.concat(Bento.objects.getByFamily(family));
            });
            Utils.forEach(families, function (other, i, l, breakLoop) {
                if (shouldIgnore(settings.ignoredFamilies, other.family)) {
                    return;
                }
                var otherBBox = other.getBoundingBox();
                if (otherBBox.isPolygon) {
                    // default behavior: ignore polygons because they cause bugs
                    if (self.ignorePolygons) {
                        return;
                    }
                    if (!otherBBox.intersect(bb)) {
                        return;
                    }
                }
                if (other.collidesWith({
                        rectangle: bb
                    })) {
                    collided.push(other);
                }
            });
        }

        if (collided.length) {
            this.onCollide();
        }

        // reset collided
        this.collided.length = 0;
    };

    Movable.prototype.attached = function (data) {
        var self = this;
        this.entity = data.entity;
        if (!this.tileColliders) {
            Bento.objects.get('tileColliders', function (obj) {
                self.tileColliders = obj;
            });
        }
    };

    Movable.prototype.draw = function (data) {
        if (Movable.debug) {
            data.renderer.fillRect(
                [0, 0, 1, 0.4],
                this.boundingBox.x,
                this.boundingBox.y - data.entity.transform.y,
                this.boundingBox.width,
                this.boundingBox.height
            );
        }
    };

    Movable.prototype.serialize = function () {
        return {
            x: this.velocity.x.toFixed(2),
            y: this.velocity.y.toFixed(2),
            z: this.velocityZ.toFixed(2),
            h: this.height.toFixed(2)
        };
    };

    Movable.prototype.parse = function (data) {
        this.velocity.x = parseFloat(data.x);
        this.velocity.y = parseFloat(data.y);
        this.velocityZ = parseFloat(data.z);
        this.height = parseFloat(data.h);
    };

    // static values
    Movable.debug = false;

    return Movable;
});