/**
 * Online Component. Used for syncing characters across mutliple instances of the game
 */
bento.define('components/online', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    MatchManager
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var clamp = function (min, value, max) {
        var temp;
        if (min > max) {
            min = temp;
            min = max;
            max = temp;
        }
        return Math.max(min, Math.min(value, max));
    };
    return function (settings) {
        var entity;

        var key;

        var stateMachine;
        var movable;
        var enemyController;
        var multiScale;
        var isRemote = settings.isRemote;

        var updateInterval = settings.updateInterval || 5;
        var updatePosMultiplier = 20;
        var updateTimer = 0;
        var timer = 0;
        var lastTimeSent = 0;

        var messageQueue;

        var lastSyncedPosition = new Vector2(0, 0);
        var targetPosition = new Vector2(0, 0);
        var currentVelocity = new Vector2(0, 0);
        var resolveDt = 0;

        var handleMessages = function () {
            if (stateMachine.state.ignoreSync || MatchManager.noSync) {
                messageQueue.length = 0;
                return;
            }

            var ii;
            var data;
            var type;

            var handledStateChange = false;
            var handledPositioning = false;

            for (ii = messageQueue.length - 1; ii >= 0; ii--) {
                data = messageQueue[ii].data;
                type = messageQueue[ii].type;

                if (type === 'stateChange' && !handledStateChange) {
                    stateMachine.data.sync(data.syncData, true);
                    stateMachine.setState(data.stateName, data.transitionData, true);
                    handledStateChange = true;
                    break; 
                }

                if (type === 'positioning' && !handledPositioning) {
                    targetPosition = Utils.toVector2(data.position);
                    // move the target position back by dx, dy
                    if (data.dx) {
                        targetPosition.x -= data.dx;
                    }
                    if (data.dy) {
                        targetPosition.y -= data.dy;
                    }
                    // get the diff time to update the targetpos and velocity it moved in
                    if (data.dt) {
                        currentVelocity = new Vector2(data.dx / data.dt, data.dy / data.dt);
                        resolveDt = data.dt;
                    }
                    if (data.dir && multiScale) {
                        multiScale.direction.x = data.dir;
                    }
                    movable.parse(data.movable);
                    handledPositioning = true;

                    if (data.force) {
                        entity.position = Utils.toVector2(data.position);
                        targetPosition = Utils.toVector2(data.position);
                        resolveDt = 0;
                    }
                }

                if (handledStateChange && handledPositioning) {
                    break;
                }
            }

            messageQueue.length = 0;
        };

        var startSync = function (m) {
            if (key !== entity.name) {
                key = entity.name;
                messageQueue = MatchManager.getMessages(key);
            }

            MatchManager.sendData({
                type: 'positioning',
                data: {
                    force: true,
                    position: entity.position,
                    movable: movable.serialize()
                }
            }, key, MatchManager.SendModes.UNRELIABLE);

            EventSystem.off('startRoom', startSync);
        };
        var setupLocal = function () {
            // note by Hernan: this seems unneeded?
            // EventSystem.on('startRoom', startSync);

            stateMachine.data.onSetState = function (state, transitionData, stateName) {
                if (state.ignoreSync) {
                    return;
                }
                MatchManager.sendData({
                    type: 'stateChange',
                    data: {
                        stateName: stateName,
                        transitionData: transitionData,
                        syncData: stateMachine.data.syncData
                    }
                }, key);
                MatchManager.sendData({
                    type: 'positioning',
                    data: {
                        position: entity.position,
                        movable: movable.serialize(),
                    }
                }, key, MatchManager.SendModes.UNRELIABLE);
            };

            movable.ignoreCollisions = false;
            stateMachine.data.isRemote = false;
        };
        var setupRemote = function () {
            stateMachine.data.onSetState = function () {};
            stateMachine.data.isRemote = true;
        };

        var component = {
            name: 'online',
            start: function (data) {
                enemyController = entity.getComponent('enemyController');
                multiScale = entity.getComponent('multiScale');

                targetPosition = entity.position.clone();
                // console.log('start target', targetPosition);

                if (!MatchManager.isOnline) {
                    isRemote = false;
                    setupLocal();
                } else if (Utils.isDefined(isRemote)) {
                    if (isRemote) {
                        setupRemote();
                    } else {
                        setupLocal();
                    }
                }
            },
            destroy: function (data) {},
            update: function (data) {
                if (key !== entity.name) {
                    // console.log('key changed from ' + key + ' to ' + entity.name);
                    key = entity.name;
                    messageQueue = MatchManager.getMessages(key);
                }

                if (isRemote) {
                    handleMessages();

                    if (resolveDt > 0) {
                        resolveDt -= data.speed;
                        targetPosition.x += currentVelocity.x * data.speed;
                        targetPosition.y += currentVelocity.y * data.speed;
                        if (!stateMachine.state.ignoreSync) { // not while the state is supposed to be ingored
                            entity.position.x += (targetPosition.x - entity.position.x) * 0.1;
                            entity.position.y += (targetPosition.y - entity.position.y) * 0.1;
                        }
                    }

                } else if (updateTimer % updateInterval === 0) {
                    if (enemyController && enemyController.active === false) {
                        // don't send positioning data for inactive enemies
                        return;
                    }
                    if (MatchManager.noSync) {
                        // don't send positioning data during cutscenes
                        return;
                    }
                    if (
                        Math.floor(lastSyncedPosition.x) !== Math.floor(entity.position.x) ||
                        Math.floor(lastSyncedPosition.y) !== Math.floor(entity.position.y) ||
                        updateTimer % (updateInterval * updatePosMultiplier) === 0
                    ) {
                        MatchManager.sendData({
                            type: 'positioning',
                            data: {
                                dt: Math.floor(timer - lastTimeSent),
                                dx: Math.floor(entity.position.x - lastSyncedPosition.x),
                                dy: Math.floor(entity.position.y - lastSyncedPosition.y),
                                position: {
                                    x: Math.floor(entity.position.x),
                                    y: Math.floor(entity.position.y)
                                },
                                movable: movable.serialize(),
                                dir: multiScale ? multiScale.direction.x : 1
                            }
                        }, key, MatchManager.SendModes.UNRELIABLE);

                        lastTimeSent = timer;

                        lastSyncedPosition.x = entity.position.x;
                        lastSyncedPosition.y = entity.position.y;
                    }
                    // updateTimer = 0;
                }

                ++updateTimer;
                timer += data.speed;

                // when ignoring sync, the movable should collide and vice versa
                if (isRemote) {
                    movable.ignoreCollisions = !stateMachine.state.ignoreSync;
                }
            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
                key = entity.name;
                messageQueue = MatchManager.getMessages(key);

                stateMachine = entity.getComponent('stateMachine');
                movable = entity.getComponent('movable');
            },
            isRemote: function () {
                return isRemote;
            },
            setRemote: function (newValue) {
                isRemote = newValue;
                if (isRemote) {
                    setupRemote();
                } else {
                    setupLocal();
                }
            }
        };
        return component;
    };
});