/**
 * Punchable component. Checks for collisions between entities in the 'punches' family and the enity this component is attached too.
 */
bento.define('components/punchable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    MatchManager,
    Punch
) {
    'use strict';
    var Punchable = function (settings) {
        var viewport = Bento.getViewport();
        var entity;

        var maxHealth = settings.maxHealth || 1;
        var health = settings.health || maxHealth;
        var defense = settings.defense || 1;
        var defenseBonus = 1;
        // flag to explicitly synchronize deaths, purely used as a backup against desyncs
        var syncDeath = settings.syncDeath || false;

        // minimum amount of hits to take if hitting with baseAttack
        var minHits = settings.minHits || 0;
        // Math.ceil to prevent cases where the entity still has 0.000001 health after the minimum amount of hits
        var damagePerHit = minHits ? Math.ceil(maxHealth / minHits) : Infinity;

        var maxStamina = 1; //settings.maxStamina || settings.stamina || 1;
        var stamina = 1; //settings.stamina || maxStamina;

        var movable;
        var enemyController;
        var hitResponses = [];

        var breakable = settings.breakable || false;

        var onPunchHit = function (punch) {
            var ii = 0;
            var iil = 0;
            var foundTarget = false;
            if (component.isDead || component.isInvincible) {
                return;
            }
            // for enemies only: should not be able to get hit while enemyController is not active
            // this is to prevent any stray obstacle from suddenly hitting an enemy
            if (enemyController && !enemyController.active) {
                return;
            }

            for (iil = punch.target.length; ii < iil; ++ii) {
                if (entity.family.indexOf(punch.target[ii]) > -1) {
                    foundTarget = true;
                }
            }

            if (!foundTarget) {
                return;
            }

            // before hit starts
            if (punch.preHit) {
                punch.preHit(entity);
            }

            hitResponses = [];
            onHit(punch);
            if (hitResponses.indexOf('blocked') < 0) {
                punch.onHit(entity);
            }

            if (health === 0 || (punch.crush && breakable)) {
                component.isDead = true;
                onDeath(punch);
                // broadcast death
                if (syncDeath && MatchManager.isOnline) {
                    MatchManager.sendData({
                        type: 'punchableDeath',
                        data: {
                            n: entity.name
                        }
                    }, 'general');
                }
                return;
            }
        };

        /**
         * Listeners
         */
        var onHitListeners = [];
        var onHurtListeners = [];
        var onDeathListeners = [];
        var onFaintListeners = [];
        var onHealthChangeListeners = [];

        var onHit = function (punch) {
            Utils.forEach(onHitListeners, function (listener) {
                var response = listener({
                    punch: punch,
                    health: health,
                    maxHealth: maxHealth
                });
                if (response) {
                    hitResponses.push(response);
                }
            });
        };
        var onHurt = function (punch, damage, maxDamage) {
            Utils.forEach(onHurtListeners, function (listener) {
                listener({
                    punch: punch,
                    health: health,
                    maxHealth: maxHealth,
                    damage: damage,
                    maxDamage: maxDamage
                });
            });
        };
        var onDeath = function (punch) {
            Utils.forEach(onDeathListeners, function (listener) {
                if (listener) {
                    listener({
                        punch: punch
                    });
                }
            });
        };
        var onFaint = function (punch) {
            Utils.forEach(onFaintListeners, function (listener) {
                listener({
                    punch: punch
                });
            });
        };
        var onHealthChange = function () {
            Utils.forEach(onHealthChangeListeners, function (listener) {
                listener({
                    health: health,
                    maxHealth: maxHealth,
                    relativeHealth: Utils.clamp(0, health / maxHealth, 1)
                });
            });
        };

        var component = {
            name: 'punchable',
            debug: settings.debug || false,
            boundingBox: settings.boundingBox || new Rectangle(-8, -8, 16, 16),
            type: settings.type || '',
            punches: Bento.objects.getByFamily('punches'),
            isDead: false,
            hasFainted: false,
            isInvincible: false,
            invincibleTime: 0,
            onDeathMessage: function () {},
            start: function (data) {
                movable = entity.getComponent('movable');
                enemyController = entity.getComponent('enemyController');

                if (syncDeath) {
                    // add to the listeners with entity name (and currentSection?) as id
                    Punchable.deathListeners[entity.name] = component;
                }
            },
            destroy: function (data) {
                if (syncDeath) {
                    Punchable.deathListeners[entity.name] = undefined;
                }
            },
            update: function (data) {
                var bb = this.boundingBox.offset(entity.position);

                Utils.forEach(this.punches, function (punch, i, l, breakLoop) {
                    if (!punch) {
                        // Utils.log('WARNING: punch is ' + punch);
                        return;
                    }
                    if (punch.hasAlreadyPunched(entity)) {
                        return;
                    }
                    if (movable && punch.height < -movable.height) {
                        return;
                    }
                    punch.collidesWith({
                        rectangle: bb,
                        onCollide: function (p) {
                            onPunchHit(punch);
                        }
                    });
                });

                defense = 1;

                if (this.invincibleTime > 0) {
                    --this.invincibleTime;
                    if (this.invincibleTime % 2 === 0) {
                        entity.visible = !entity.visible;
                    }
                    if (this.invincibleTime <= 0) {
                        this.setInvincible(false);
                    }
                }
            },
            draw: function (data) {
                if (Punchable.debug) {
                    data.renderer.fillRect(
                        [1, 0, 0, 0.4],
                        this.boundingBox.x,
                        this.boundingBox.y - data.entity.transform.y,
                        this.boundingBox.width,
                        this.boundingBox.height
                    );
                }
            },
            attached: function (data) {
                entity = data.entity;
            },
            collidesWith: function (other) {
                if (this.isDead || this.hasFainted) {
                    return;
                }

                if (!other.isRectangle) {
                    other = other.getBoundingBox();
                }
                return (this.boundingBox.offset(entity.position).intersect(other));
            },
            setStamina: function (st) {
                stamina = Utils.clamp(0, st, maxStamina);

                if (stamina !== 0) {
                    this.hasFainted = false;
                }
            },
            setMaxStamina: function (st) {
                maxStamina = st;
            },
            addStamina: function (st) {
                component.setStamina(stamina + st);
            },
            resetStamina: function () {
                component.setStamina(maxStamina);
            },
            setMaxHealth: function (mh) {
                var healthPercentage = health / maxHealth;
                maxHealth = mh;
                component.setHealth(maxHealth * healthPercentage);
            },
            setHealth: function (h) {
                health = Math.min(maxHealth, h);
                onHealthChange();
            },
            resetHealth: function () {
                component.isDead = false;
                health = maxHealth;
                onHealthChange();
            },
            addHealth: function (h) {
                component.setHealth(health + h);
            },
            getHealth: function () {
                return health;
            },
            getMaxHealth: function () {
                return maxHealth;
            },
            getRelativeHealth: function () {
                return health / maxHealth;
            },
            hit: function (punch) {
                onPunchHit(punch);
            },
            kill: function (punch) {
                component.isDead = true;
                onDeath(punch);
                // broadcast death
                if (syncDeath && MatchManager.isOnline) {
                    MatchManager.sendData({
                        type: 'punchableDeath',
                        data: {
                            n: entity.name
                        }
                    }, 'general');
                }
            },
            setInvincible: function (invincible) {
                invincible = invincible ? true : false;

                this.isInvincible = invincible;
                // this.invincibleTime = invincible ? 10000000 : 0;

                if (!invincible) {
                    entity.visible = true;
                }
            },
            setInvincibleTime: function (time) {
                time = time > 0 ? time : 0;
                var invincible = time ? true : false;

                this.setInvincible(invincible);

                this.invincibleTime = time;
            },
            /*
             * Adds a callback to when entity gets hit
             * Listener parameters:
             * settings.punch - Punch object that hit the entity
             * settings.health - Current health
             * settings.maxHealth - Max health
             */
            onHit: function (listener) {
                onHitListeners.push(listener);
            },
            /*
             * Adds a callback to when entity gets hurt (health depleted by damage)
             * Listener parameters:
             * settings.punch - Punch object that hit the entity
             * settings.health - Current health
             * settings.maxHealth - Max health
             */
            onHurt: function (listener) {
                onHurtListeners.push(listener);
            },
            /*
             * Adds a callback to when entity dies
             */
            onDeath: function (listener) {
                onDeathListeners.push(listener);
            },
            offDeath: function (listener) {
                var index = onDeathListeners.indexOf(listener);
                if (onDeathListeners[index]) {
                    onDeathListeners[index] = null;
                }
            },
            /*
             * Adds a callback to when entity faints
             */
            onFaint: function (listener) {
                onFaintListeners.push(listener);
            },
            /**
             * Adds a callback to whenever the entity's health changes
             */
            onHealthChange: function (listener) {
                onHealthChangeListeners.push(listener);
            },
            /**
             * Intended for multiplayer, to increase an enemy's health
             */
            increaseHealth: function (percentage) {
                health *= 1 + (percentage / 100);
                maxHealth *= 1 + (percentage / 100);
                health = Math.floor(health);
                maxHealth = Math.floor(maxHealth);
            },
            // deplete health when hit by punch
            depleteHealth: function (punch) {
                // the damage that would be dealt if not accounting for minHits
                var baseDamage = Math.min(punch.attack, damagePerHit);
                // current damage formula: (attack * multiplier) / defense
                var damage = baseDamage * punch.multiplier * defense;
                damage = Math.ceil(damage);
                health = Math.max(0, health - damage);
                onHealthChange();

                onHurt(punch, damage, baseDamage === damagePerHit);
            },
            // deplete stamina when hit by punch
            depleteStamina: function (punch) {
                if (!component.hasFainted) {
                    var punchStamina = punch.stamina;
                    if (!punchStamina) {
                        punchStamina = 0.334;
                    }

                    stamina -= punchStamina;
                    // if (stamina <= 0 || punch.faintTime) {
                    if (stamina <= 0) {
                        stamina = 0;
                        onFaint(punch);
                        component.hasFainted = true;
                    }
                }
            },
            addDefenseBonus: function (value) {
                // multiply (otherwise defense could become negative)
                defense *= value;
            },
            forceDeath: function () {
                if (component.isDead) {
                    // already dead, everything is fine!
                    return;
                }
                // having to call forceDeath is a sign that something desynced
                console.log('WARNING: forced death');
                component.isDead = true;
                onDeath(new Punch({}));
            }
        };
        return component;
    };

    Punchable.debug = false;

    // there is no listener defined by the component itself, because this it's too
    // complicated to define a clear sender/listener and can cause memory leaks
    // instead we use the "general" message event listener in screens/main
    Punchable.deathListeners = {};

    return Punchable;
});