/**
 * Punch Syncer
 * Every player can ony punch and be punched on their own client. However, we want each client to have some authority over their experience, regardless of what the master says.
 * Hence, every connecting punch that the player makes should be broadcast to the other players as well.
 *
 * @moduleName PunchSyncer
 */
bento.define('components/punchsyncer', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    MatchManager,
    Punch
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var messageQueue = MatchManager.getMessages('punches');
        var onlineKey = '';

        var onPunchHit = function (e) {
            if (e.attacker !== entity) {
                return;
            }
            var punch = e.punch;
            var m;
            var s;
            e.victim.getComponent('movable', function (movable) {
                m = movable.serialize();
            });
            e.victim.getComponent('stateMachine', function (stateMachine) {
                s = stateMachine.stateName;
            });

            var data = {
                t: {
                    n: e.victim.name,
                    x: e.victim.position.x,
                    y: e.victim.position.y,
                    m: m,
                    s: s
                },
                p: punch.serialize()
            };

            MatchManager.sendData({
                type: 'punch',
                data: data
            }, onlineKey);
        };

        var emulatePunch = function (msg) {
            var data = msg.data;
            var target = Bento.objects.getByName(data.t.n)[0];
            if (!target) {
                // target does not exist anymore
                return true;
            }
            var punchable = target.getComponent('punchable');
            if (!punchable) {
                // target does is not a punchable?
                return;
            }

            var defaults = Punch.getDefaults();

            // If a property uses its default value, that property is not sent but
            // is instead re-created from the default value on the receiving end.
            var punch = {
                onHit: function () {},
                target: target.family.slice(),
                type: Utils.isDefined(data.p.t) ? data.p.t : defaults.type,
                attack: Utils.isDefined(data.p.a) ? data.p.a : defaults.attack,
                multiplier: Utils.isDefined(data.p.m) ? data.p.m : defaults.multiplier,
                bonusAttack: Utils.isDefined(data.p.b) ? data.p.b : defaults.bonusAttack,
                pushBack: new Vector2(
                    Utils.isDefined(data.p.x) ? data.p.x : defaults.pushBack.x,
                    Utils.isDefined(data.p.y) ? data.p.y : defaults.pushBack.y
                ),
                stamina: Utils.isDefined(data.p.s) ? data.p.s : defaults.stamina,
                direction: Utils.isDefined(data.p.d) ? data.p.d : defaults.direction,
                freeze: 0, // TODO: does this need to be synced as well?
                crush: Utils.isDefined(data.p.c) ? data.p.c : defaults.crush,
                emulated: true
            };

            //set the target to the same position and state it had when it was punched
            target.getComponent('punchable').isInvincible = false;
            if (data.t.s) {
                target.getComponent('stateMachine', function (stateMachine) {
                    stateMachine.setState(data.t.s, {}, true);
                });
            }
            target.position.x = data.t.x;
            target.position.y = data.t.y;
            target.getComponent('movable', function (movable) {
                movable.parse(data.t.m);
            });

            punchable.hit(punch);

            return true;
        };

        var component = {
            name: 'punchSyncer',
            start: function (data) {
                onlineKey = settings.onlineKey || 'punches-' + entity.name;
                EventSystem.on('punchHit', onPunchHit);
                MatchManager.onMessage(onlineKey, emulatePunch);
            },
            destroy: function (data) {
                EventSystem.off('punchHit', onPunchHit);
                MatchManager.offMessage(onlineKey, emulatePunch);
            },
            update: function (data) {

            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity;
            }
        };
        return component;
    };
});