/**
 * State machine manages developer defined states
 * @moduleName StateMachine
 */
bento.define('components/statemachine', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween
) {
    'use strict';
    var StateMachine = function (settings) {
        var states = {};
        var stateMachine = {
            name: 'stateMachine',
            state: {}, // curent state
            stateName: '',
            data: {
                previousState: '',
                entity: null,
                rootPath: '',
                behavior: '',
                syncData: {},
                setState: function () {},
                processAction: function () {},
                moveInDirection: function () {},
                // overwrite this callback
                onSetState: function () {}
            },
            start: function (data) {
                // add every component of the parent entity to stateMachine.data for easy access
                var entity = data.entity;
                Utils.forEach(entity.components, function (component, i, l, breakLoop) {
                    if (component && component.name) {
                        stateMachine.data[component.name] = component;
                    }
                });
                stateMachine.data.entity = entity;

                // access functions through stateData
                stateMachine.data.setState = stateMachine.setState;
                stateMachine.data.processAction = stateMachine.processAction;
                stateMachine.data.moveInDirection = stateMachine.moveInDirection;
                stateMachine.data.sync = stateMachine.sync;
            },
            /**
             * Define a new state
             */
            addState: function (name, newState) {
                if (states[name]) {
                    // overwriting states may be desirable
                    // Utils.log('WARNING: state ' + name + ' already exists');
                }
                states[name] = newState;
            },
            /**
             * Define new states
             */
            addStates: function (objLiteral) {
                Utils.extend(states, objLiteral);
            },
            /**
             * Set current state
             */
            setState: function (newState, transitionData, force) {
                // set state by name
                var key;
                var previousState = stateMachine.stateName;

                // assets were not loaded yet, this could crash the game
                if (stateMachine.deferStates) {
                    // save for later
                    stateMachine.deferredState = {
                        newState: newState,
                        transitionData: transitionData,
                        force: force
                    };
                    return;
                }
                if (Utils.isString(newState)) {
                    stateMachine.stateName = newState;
                    newState = states[stateMachine.stateName];
                    if (!newState) {
                        Utils.log("WARNING: state " + stateMachine.stateName + " does not exist!");
                        return;
                    }
                } else {
                    for (key in states) {
                        if (states[key] === newState) {
                            stateMachine.stateName = key;
                            break;
                        }
                    }
                }

                if (stateMachine.state === newState && !force) {
                    return;
                }
                if (stateMachine.state && stateMachine.state.end) {
                    stateMachine.state.end(stateMachine.data, transitionData, newState);
                }
                stateMachine.data.previousState = previousState;
                stateMachine.state = newState;
                stateMachine.state.start(stateMachine.data, transitionData);

                if (stateMachine.data.onSetState) {
                    stateMachine.data.onSetState(newState, transitionData, stateMachine.stateName);
                }
            },
            /**
             * Perform action in current state
             */
            processAction: function (actionData) {
                if (stateMachine.state.action) {
                    stateMachine.state.action(stateMachine.data, actionData);
                }
            },
            /**
             * Perform move in current state
             */
            moveInDirection: function (normalizedVector) {
                if (stateMachine.state.move) {
                    stateMachine.state.move(stateMachine.data, normalizedVector);
                }
            },
            update: function (data) {
                if (stateMachine.state.update) {
                    stateMachine.state.update(stateMachine.data);
                }
            },
            /**
             * Prepare to sync data
             * This function should only be called during state.start (this is called before the online component will sync)
             */
            sync: function (obj, force) {
                var canSync = true;
                // remotes should ignore this sync
                if (!force) {
                    stateMachine.data.entity.getComponent('online', function (online) {
                        if (online.isRemote()) {
                            canSync = false;
                        }
                    });
                }

                if (canSync) {
                    Utils.extend(stateMachine.data.syncData, obj, true);
                }
            },
            hasState: function (name) {
                return states[name];
            },
            /**
             * Deferring states happens when a rootPath is missing in the assets
             * We cannot set the state until this is resolved
             */
            deferStates: false,
            deferredState: {},
            resolve: function () {
                var deffered = stateMachine.deferredState;
                stateMachine.deferStates = false;
                // continue with state change
                stateMachine.setState(deffered.newState, deffered.transitionData, deffered.force);
            }
        };

        return stateMachine;
    };

    return StateMachine;
});