/**
 * Allows entities to be picked up as a weapon
 * @moduleName Weapon
 */
bento.define('components/weapon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager',
    'components/movable',
    'entities/punch',
    'managers/save',
    'managers/buff',
    'entities/triggers/blockade'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    MatchManager,
    Movable,
    Punch,
    SaveManager,
    BuffManager,
    Blockade
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    return function (settings) {
        var viewport = Bento.getViewport();
        var screenDimension = Bento.screens.getCurrentScreen().dimension;

        var entity;
        var sprite;
        var parentCharacter = null;
        var spriteCharacter = null;
        var emptyData = {
            frames: []
        };
        var weaponData;
        var rotationSpeed = 0.3;

        var punch;

        var durability = settings.durability;

        // if in active area infront of blockade or not
        var activated = false;

        var movable = new Movable({
            onLanded: function () {
                if (punch) {
                    punch.disable();
                    punch = null;
                }
                movable.velocity.scalarMultiplyWith(0.5);
                movable.gravity = 0.15;
            }
        });

        var onlineKey = '';

        var positionToFrame = function () {
            var currentFrame;
            var frames;
            var spriteData;
            // update position and rotation based on weapon data of the character sprite

            weaponData = (spriteCharacter.animationSettings.data || {}).weapon || emptyData;
            frames = weaponData.frames;
            currentFrame = Math.floor(spriteCharacter.getCurrentFrame());

            if (frames.length && currentFrame < frames.length) {
                // Note: the weapon position in the json is defined as the position relative to the left upper corner
                // this should have been relative the origin, but I can do that here by subtracting
                entity.position.x = frames[currentFrame].x - spriteCharacter.animationSettings.origin.x;
                entity.position.y = frames[currentFrame].y - spriteCharacter.animationSettings.origin.y;
                entity.rotation = frames[currentFrame].rotation;
            }
        };

        var handleMessage = function (m) {
            var character;
            if (m.type === 'pickUp') {
                if (!MatchManager.isMaster && !m.data.v) {
                    //event is not verified yet, ignore it
                    return true;
                }

                character = Bento.objects.getByName(m.data.c)[0];
                if (!component.pickUp(character, true)) {
                    if (MatchManager.isMaster && parentCharacter) {
                        MatchManager.sendData({
                            type: 'changeOwner',
                            data: {
                                w: m.data.c,
                                c: parentCharacter.name
                            }
                        });
                    } else {
                        //TODO what to do if this happens?
                    }
                } else if (MatchManager.isMaster) {
                    //send a verified event to all clients
                    MatchManager.sendData({
                        type: 'pickUp',
                        data: {
                            c: m.data.c,
                            v: true
                        }
                    }, onlineKey);
                }
                return true;
            }

            if (m.type === 'drop') {
                //TODO check if the droppee is actually the client we believe to be holding the item. If so, detach the item and place it correctly
                if (parentCharacter && parentCharacter.name === m.data.c) {
                    //TODO drop it
                    component.drop(0, true);

                    entity.position.x = m.data.x;
                    entity.position.y = m.data.y;
                    movable.height = m.data.h;
                }
                return true;
            }

            if (m.type === 'move') {
                //TODO check if the droppee is actually the client we believe to be holding the item. If so, detach the item and place it correctly

                entity.position.x = m.data.x;
                entity.position.y = m.data.y;
                movable.height = m.data.h;

                movable.velocity.x = m.data.vx;
                movable.velocity.y = m.data.vy;
                movable.velocityZ = m.data.vz;

                movable.gravity = m.data.g;

                return true;
            }

            if (m.type === 'changeOwner') {
                if (parentCharacter && parentCharacter.name === m.data.w) { //the wrong character is holding this weapon
                    component.drop(0, true);
                    character = Bento.objects.getByName(m.data.c)[0];
                    component.pickUp(character, true);
                }
                return true;
            }

            if (m.type === 'durability') {
                component.removeDurability(m.data.d, true);
            }

            //no one else needs to handle this message so tell MatchManager not to store it
            return true;
        };

        var onThrowHit = function () {
            if (!punch) {
                return;
            }

            Bento.audio.stopSound('sfx_punch1');
            Bento.audio.playSound('sfx_punch1');

            punch.disable();
            movable.velocity.x *= -0.55;
            movable.velocityZ -= 0.25;
            punch = null;

            movable.gravity = 0.15;

            MatchManager.sendData({
                type: 'move',
                data: {
                    x: entity.position.x,
                    y: entity.position.y,
                    h: movable.height,
                    vx: movable.velocity.x,
                    vy: movable.velocity.y,
                    vz: movable.velocityZ,
                    g: movable.gravity
                }
            }, onlineKey);

            component.removeDurability(settings.throw.durabilityCost);
        };

        var component = {
            name: 'weapon',
            start: function (data) {
                sprite = entity.getComponent('sprite');

                onlineKey = 'weapon-' + entity.weaponId;

                MatchManager.onMessage(onlineKey, handleMessage);
            },
            destroy: function (data) {
                MatchManager.offMessage(onlineKey, handleMessage);

                // the weapon may be carried over on the next screen
                // only for local player: save it so it can be restored later
                if (parentCharacter && parentCharacter.isPlayer && !parentCharacter.isRemote) {
                    SaveManager.save('currentWeapon', {
                        type: entity.getType(),
                        durability: durability
                    });
                }
            },
            update: function (data) {
                // update position and rotation based on weapon data of the character sprite
                if (spriteCharacter) {
                    positionToFrame();
                } else {
                    movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.1);
                    movable.velocity.y = Utils.approach(movable.velocity.y, 0, 0.1);

                    if (!movable.grounded) {
                        entity.rotation += rotationSpeed * data.speed;
                        entity.scale.x = Utils.sign(movable.velocity.x) || 1;
                    } else {
                        entity.rotation = 0;
                    }

                    // make sure the weapon can't fly out of the level
                    var rightSide;
                    if (Blockade.active) {
                        if (!activated) {
                            if (Blockade.active.position.x >= entity.position.x) {
                                activated = true;
                                rightSide = Blockade.active.position.x;
                            } else {
                                rightSide = screenDimension.width;
                            }
                        } else {
                            rightSide = Blockade.active.position.x;
                        }
                    } else {
                        rightSide = screenDimension.width;
                    }
                    entity.position.x = Utils.clamp(0, entity.position.x, rightSide);
                }
            },
            draw: function (data) {
                /*
                var offset = (new Vector2(0, -entity.transform.y)).rotateRadian(-entity.rotation);
                data.renderer.fillCircle([1, 0, 0, 1], offset.x, offset.y, 5);
                */
            },
            attached: function (data) {
                entity = data.entity;

                entity.attach(movable);
            },
            /**
             * Weapon is picked up by a character
             */
            pickUp: function (character, dontSync) {
                if (parentCharacter || !movable.grounded || durability <= 0 || !character) {
                    // already picked up
                    return false;
                }
                parentCharacter = character;
                // get reference to sprite
                spriteCharacter = parentCharacter.getComponent('sprite');

                //TODO
                sprite.setSpriteSheet(settings.activeSprite);

                // detach from world
                Bento.objects.remove(entity);
                entity.scale.x = 1;

                // attach to character
                character.attach(entity);

                entity.remove(movable);

                if (!dontSync) {
                    MatchManager.sendData({
                        type: 'pickUp',
                        data: {
                            c: character.name,
                            v: MatchManager.isMaster
                        }
                    }, onlineKey);
                }

                return true;
            },
            /**
             * Weapon is lost by character
             */
            drop: function (force, dontSync) {
                if (!parentCharacter) {
                    return;
                }

                var name = parentCharacter.name;
                var parentStats = entity.parent.getComponent('stats') || {
                    attack: 1,
                    skill: 0
                };

                sprite.setSpriteSheet(settings.passiveSprite);
                entity.rotation = 0;

                entity.attach(movable);

                //TODO
                if (force !== 0) {
                    movable.velocity.x = force;
                    movable.velocityZ = -0.25;
                    movable.gravity = 0.10;
                    rotationSpeed = 0.45 * Utils.sign(movable.velocity.x);

                    punch = new Punch({
                        boundingBox: entity.boundingBox.clone(),
                        parent: entity,
                        duration: 999,
                        target: ['enemy'], //TODO
                        direction: Utils.sign(force),
                        pushBack: new Vector2(2, 0.5),
                        skill: parentStats.skill + settings.throw.damage,
                        stamina: settings.throw.stamina || settings.hit.stamina || 0.4,
                        attack: parentStats.attack + BuffManager.current.throwStrength,
                        multiplier: 1 + settings.throw.damage,
                        onHit: onThrowHit,
                        position: entity.position.clone(),
                        type: 'throwingWeapon'
                    });

                    Bento.objects.attach(punch);
                } else {
                    movable.velocityZ = -2;
                    movable.gravity = 0.20;
                    rotationSpeed = 0.3;
                }

                positionToFrame();

                //TODO calculate position in world space;

                parentCharacter.remove(entity);

                entity.position.x = parentCharacter.position.x + entity.position.x * Utils.sign(parentCharacter.scale.x);
                movable.height = entity.position.y; //TODO origin differences
                entity.position.y = parentCharacter.position.y;
                Bento.objects.attach(entity);

                if (parentCharacter && parentCharacter.isPlayer && !parentCharacter.isRemote) {
                    SaveManager.save('currentWeapon', {});
                }

                parentCharacter = null;
                spriteCharacter = null;

                if (!dontSync) {
                    MatchManager.sendData({
                        type: 'drop',
                        data: {
                            c: name,
                            x: entity.position.x,
                            y: entity.position.y,
                            h: movable.height
                        }
                    }, onlineKey);

                    if (force !== 0) {
                        MatchManager.sendData({
                            type: 'move',
                            data: {
                                x: entity.position.x,
                                y: entity.position.y,
                                h: movable.height,
                                vx: movable.velocity.x,
                                vy: movable.velocity.y,
                                vz: movable.velocityZ,
                                g: movable.gravity
                            }
                        }, onlineKey);
                    }
                }
            },
            /**
             *
             */
            removeDurability: function (d, dontSync) {
                if (durability <= 0) {
                    //already broken!
                    return;
                }

                d *= BuffManager.current.weaponDurability;

                durability -= d;

                if (durability <= 0) {
                    console.log('weapon breaks!');

                    if (parentCharacter) {
                        component.drop(0, false);
                    }

                    new Tween({
                        from: 1,
                        to: 0,
                        'in': 20,
                        ease: Tween.EASEINQUAD,
                        onUpdate: function (v, t) {
                            entity.alpha = v;
                        },
                        onComplete: function () {
                            Bento.objects.remove(entity);
                        }
                    });
                }
                if (!dontSync) {
                    MatchManager.sendData({
                        type: 'durability',
                        data: {
                            d: d
                        }
                    }, onlineKey);
                }
            },
            setDurability: function (value) {
                durability = value;
            }
        };
        return component;
    };
});