/**
 * Train for use in the subway missions
 * @moduleName SubwayTrain
 */
bento.define('entities/background/subwaytrain', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var camera = Bento.objects.get('camera');
        var screenDimension = Bento.screens.getCurrentScreen().dimension;
        var tiled = (settings.tiled || {});
        var object = tiled.object || {};
        var nameProperties = Utils.nameToJson(object.name);
        var length = nameProperties.length || 2;
        var direction = nameProperties.direction || 1;
        var waitTime = (nameProperties.time || 60) * 60;
        var speed = 10;
        var goingPast = false;
        var playingSound = false;

        var leftEnd = new Entity({
            name: 'leftEnd',
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    spriteSheet: 'background/subwaytrain/leftend'
                })
            ]
        });

        var cars = new Entity({
            name: 'car',
            position: new Vector2(leftEnd.dimension.width, 0),
            components: [
                new Sprite({
                    spriteSheet: 'background/subwaytrain/car'
                })
            ]
        });

        var rightEnd = new Entity({
            name: 'rightEnd',
            position: new Vector2(leftEnd.dimension.width + length * cars.dimension.width, 0),
            components: [
                new Sprite({
                    spriteSheet: 'background/subwaytrain/rightend'
                })
            ]
        });

        var pixelLength = leftEnd.dimension.width + length * cars.dimension.width + rightEnd.dimension.width;

        var resetPosition = function () {
            if (direction === 1) {
                entity.position.x = -pixelLength;
            } else {
                entity.position.x = screenDimension.width;
            }
            goingPast = false;
        };

        var checkDone = function () {
            if (direction === 1) {
                if (entity.position.x > screenDimension.width) {
                    resetPosition();
                }
            } else {
                if (entity.position.x < -pixelLength) {
                    resetPosition();
                }
            }
        };

        var checkCamera = function () {
            // shake the camera when the train is on the screen
            if ((entity.position.x < viewport.x && viewport.x < entity.position.x + pixelLength) ||
                (entity.position.x < viewport.x + viewport.width && viewport.x + viewport.width < entity.position.x + pixelLength)) {
                if (entity.timer % 2 === 0 && camera) {
                    camera.shake(2, 2);
                }

                if (!playingSound) {
                    Bento.audio.setVolume(1, 'sfx_subway');
                    Bento.audio.playSound('sfx_subway', true);
                    playingSound = true;
                }
            } else if (playingSound) {
                new Tween({
                    from: 1,
                    to: 0,
                    in: 80,
                    delay: 0,
                    ease: 'easeOutCubic',
                    onUpdate: function (v, t) {
                        Bento.audio.setVolume(v, 'sfx_subway');
                    },
                    onComplete: function () {
                        Bento.audio.stopSound('sfx_subway');
                    }
                });
                playingSound = false;
            }
        };

        var behavior = {
            name: 'behaviorComponent',
            timer: nameProperties.start * 60 || 0,
            start: function (data) {
                resetPosition();
            },
            destroy: function (data) {
                Bento.audio.stopSound('sfx_subway');
            },
            update: function (data) {
                if (!goingPast) {
                    // wait till timer is over
                    behavior.timer++;
                    if (behavior.timer >= waitTime) {
                        goingPast = true;
                        behavior.timer = 0;
                    }
                } else {
                    entity.position.x += speed * direction;
                    checkDone();
                    checkCamera();
                }
            },
            draw: function (data) {
                // draw cars at all locations
                var i = length - 1;
                while (i > 0) {
                    cars.position.x = leftEnd.dimension.width + cars.dimension.width * i;
                    cars.draw(data);
                    --i;
                }
                cars.position.x = leftEnd.dimension.width;
            }
        };
        var entity = new Entity({
            z: Globals.Layers.SUBWAY,
            name: 'subwayTrain',
            family: ['background'],
            position: settings.position,
            updateWhenPaused: 0,
            components: [
                leftEnd,
                cars,
                rightEnd,
                behavior
            ]
        });
        return entity;
    };
});