/**
 * Pauses the gameplay to show a boss intro,
 * @moduleName BossIntro
 */
bento.define('entities/bossintro', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/boss',
    'globals',
    'components/fillscreen',
    'components/anchor',
    'managers/character',
    'components/orientationlock'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Boss,
    Globals,
    FillScreen,
    Anchor,
    CharacterManager,
    OrientationLock
) {
    'use strict';
    /*
     * Wraps position of an entity
     */
    var WrapBehavior = function (settings) {
        var entity;
        var speed = settings.speed || 1;
        var component = {
            name: 'wrapBehavior',
            update: function () {
                entity.position.x -= speed;
                entity.position.x = entity.position.x % entity.dimension.width;
            },
            attached: function (data) {
                entity = data.entity;
            }
        };
        return component;
    };
    /*
     * Moves in character during boss intro
     */
    var PortraitBehavior = function (settings) {
        var viewport = Bento.getViewport();
        var entity;
        var delay = settings.delay || 0;
        var offset = settings.offset || 0;
        var offsetY = settings.offsetY || 0;
        var top = settings.top; // top moves from right to left
        var component = {
            name: 'portraitBehavior',
            start: function (data) {
                if (top) {
                    // note: 64 is a pretty random number, what if there's a character that's taller...?
                    entity.position = new Vector2(viewport.width / 3, 64 + offsetY);
                } else {
                    entity.position = new Vector2(-viewport.width / 3, 64 + offsetY);
                }
                entity.visible = false;
                new Tween({
                    from: entity.position.x + viewport.width * (top ? 1 : -1),
                    to: entity.position.x + offset,
                    in: 120,
                    ease: Tween.EXPONENTIAL,
                    delay: delay,
                    alpha: -20,
                    onStart: function () {
                        entity.visible = true;
                    },
                    onUpdate: function (v, t) {
                        entity.position.x = v;
                    }
                });
            },
            update: function () {},
            attached: function (data) {
                entity = data.entity;
            }
        };
        return component;
    };

    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var overlay = new Entity({
            name: 'overlay',
            position: new Vector2(0, 0),
            alpha: 0.5,
            components: [
                new FillScreen({})
            ]
        });
        var overlayFlash = new Entity({
            name: 'overlayFlash',
            alpha: 0,
            components: [
                new FillScreen({
                    color: [1, 1, 1, 1]
                })
            ]
        });
        var boss;
        var state = 'wait';
        var camera = settings.camera || undefined;
        var cameraBehavior;
        var oldCameraTarget = null;
        var showIntro = function () {
            var getPortraits = function (isBoss) {
                var entities = [];
                var characters = isBoss ? Boss.currentBosses : Bento.objects.getByFamily('player');
                Utils.forEach(characters, function (character, i, l, breakLoop) {
                    var skip = false;
                    if (isBoss) {
                        character.getComponent('boss', function (bossComponent) {
                            skip = bossComponent.skipIntro;
                        });
                    }
                    if (skip) {
                        return;
                    }
                    // get their spritesheet name
                    character.getComponent('sprite', function (sprite) {
                        var spriteSheetName = sprite.currentSpriteSheet;
                        var portraitSprite;
                        var portraitEntity;

                        // reconstruct
                        if (spriteSheetName) {
                            portraitSprite = new Sprite({
                                spriteSheet: spriteSheetName
                            });
                            portraitSprite.setCurrentSpeed(0);
                            portraitEntity = new Entity({
                                name: 'portrait',
                                scale: new Vector2(2 * (isBoss ? -1 : 1), 2),
                                components: [
                                    portraitSprite,
                                    new PortraitBehavior({
                                        top: isBoss,
                                        delay: 10 * entities.length,
                                        offset: 40 * entities.length * (isBoss ? -1 : 1),
                                        offsetY: -8 * entities.length,
                                        scale: 1 - 0.1 * entities.length
                                    })
                                ]
                            });
                            entities.push(portraitEntity);
                        }
                    });
                });
                return entities;
            };
            var getPlayerPortraits = function () {
                var entities = [];
                var currentCharacter = CharacterManager.getSelected();
                var largePortraitSprite = new Sprite({
                    spriteSheet: 'characterchange/' + currentCharacter,
                });
                var largePortraitBackSprite = new Sprite({
                    spriteSheet: 'characterchange/splash_' + currentCharacter,
                });
                var playerEntity = new Entity({
                    name: 'playerPortrait',
                    position: new Vector2(0, 0),
                    components: [
                        largePortraitBackSprite,
                        largePortraitSprite,
                        new PortraitBehavior({
                            top: false,
                            delay: 10 * entities.length,
                            offset: 64 * entities.length,
                            offsetY: -32 - (Anchor.isPortrait() ? 0 : 32)
                        })
                    ]
                });
                entities.push(playerEntity);
                return entities;
            };
            var topBackground = new Entity({
                name: 'topBackground',
                position: new Vector2(viewport.width / 4, 32),
                components: [
                    new Sprite({
                        spriteSheet: 'bossintro/vs-enemysplash'
                    })
                ]
            });
            var bottomBackground = new Entity({
                name: 'bottomBackground',
                position: new Vector2(-viewport.width / 4, 32),
                scale: new Vector2(-1, 1),
                components: [
                    new Sprite({
                        spriteSheet: 'bossintro/vs-enemysplash'
                    })
                ]
            });
            var portraitContainerTop = new Entity({
                name: 'portraitContainerTop',
                components: [
                    new Anchor({
                        portrait: {
                            screenPosition: new Vector2(0.5, 1 / 4),
                            offset: new Vector2(0, -32)
                        },
                        landscape: {
                            screenPosition: new Vector2(0.5, 1 / 4),
                            offset: new Vector2(16, 0)
                        }
                    }),
                    topBackground
                ]
            });
            var portraitContainerBottom = new Entity({
                name: 'portraitContainerTop',
                components: [
                    new Anchor({
                        screenPosition: new Vector2(0.5, 3 / 4),
                        offset: new Vector2(0, 0)
                    })
                ]
            });
            var oldPauseLevel = Bento.objects.isPaused();

            // pause the game
            Bento.objects.pause(Globals.PauseLevels.BOSSINTRO);

            // fade in
            new Tween({
                from: 0,
                to: 0.7,
                in: 10,
                ease: Tween.LINEAR,
                onStart: function () {
                    var topPortraits = getPortraits(true);
                    var bottomPortraits = getPlayerPortraits(false);
                    topPortraits.reverse();
                    bottomPortraits.reverse();
                    entity.attach(overlay);
                    // add top portraits
                    entity.attach(portraitContainerTop);
                    Utils.forEach(topPortraits, function (portraitEntity, i, l, breakLoop) {
                        portraitContainerTop.attach(portraitEntity);
                    });
                    // add bottom portraits
                    entity.attach(portraitContainerBottom);
                    Utils.forEach(bottomPortraits, function (portraitEntity, i, l, breakLoop) {
                        portraitContainerBottom.attach(portraitEntity);
                    });

                },
                onUpdate: function (v, t) {
                    overlay.alpha = v;
                },
                onComplete: function () {}
            });

            // vs button
            Utils.timeout(60, function () {
                var vsSprite = new Sprite({});
                var vs = new Entity({
                    name: 'vs',
                    components: [
                        vsSprite,
                        new Anchor({
                            screenPosition: new Vector2(0.5, 0.5),
                            offset: new Vector2(0, 0)
                        })
                    ]
                });
                vsSprite.setSpriteSheet('bossintro/vs-big', function () {
                    var vsSplash = new Sprite({});
                    var vsSplashEntity = new Entity({
                        name: 'vsSplash',
                        components: [
                            vsSplash
                        ]
                    });
                    vs.attach(vsSplashEntity);
                    vs.moveComponentTo(vsSplash, 0);
                    vsSprite.setSpriteSheet('bossintro/vs-small');
                    Bento.audio.stopSound('sfx_vsdrop');
                    Bento.audio.playSound('sfx_vsdrop');
                    vsSplash.setSpriteSheet('bossintro/vs-flash', function () {
                        vs.remove(vsSplashEntity);
                    });
                    if (Anchor.isWideScreen() && Anchor.isPortrait()) {
                        vsSplashEntity.scale.y = (viewport.height + Anchor.getOffset().y * 2) / 320;
                    }
                    new Tween({
                        from: 1,
                        to: 0,
                        in: 30,
                        ease: 'linear',
                        onUpdate: function (v, t) {
                            overlayFlash.alpha = v;
                        }
                    });
                });
                entity.attach(vs);
            }, Globals.PauseLevels.BOSSINTRO);

            // end after 3 seconds
            Utils.timeout(180, function () {
                new Tween({
                    from: 1,
                    to: 0,
                    in: 30,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        entity.alpha = v;
                    },
                    onComplete: function () {
                        cameraBehavior.target = oldCameraTarget;
                        Bento.objects.pause(oldPauseLevel);
                        Bento.objects.remove(entity);
                    }
                });
            });

            // tap to increase speed
            // entity.attach(new Clickable({
            //     pointerDown: function (data) {
            //         Bento.setGameSpeed(3);
            //     }
            // }));
        };
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                if (Bento.isDev() && window.skipBossIntro) {
                    // faster bosses during dev
                    entity.removeSelf();
                    return;
                }

                // find the camera object
                if (!camera) {
                    camera = Bento.objects.get('camera');
                }
                // cache the old target
                cameraBehavior = camera.getComponent('cameraBehavior');
                oldCameraTarget = cameraBehavior.target;

                // mute music
                new Tween({
                    from: 1,
                    to: 0,
                    in: 60,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        Bento.audio.setVolume(v, Globals.currentMusic);
                    },
                    onComplete: function () {
                        Bento.audio.stopAllMusic();
                        var musicKey = 'sfx_' + Globals.currentMusic.split('_')[1] + '-appear';

                        if (Utils.hasAsset('audio', musicKey)) {
                            Bento.audio.stopSound(musicKey);
                            Bento.audio.playSound(musicKey);
                        } else {
                            Bento.audio.stopSound('sfx_world1-appear');
                            Bento.audio.playSound('sfx_world1-appear');
                        }
                    }
                });
            },
            destroy: function (data) {
                // restore music
                if (Utils.hasAsset('audio', Globals.currentMusic + '-boss')) {
                    Bento.audio.playMusic(Globals.currentMusic + '-boss');
                    Bento.audio.setVolume(1, Globals.currentMusic + '-boss');
                    Globals.currentMusic += '-boss';
                } else {
                    console.log('Warning: no boss music found for ' + Globals.currentMusic);
                    Bento.audio.playMusic('bgm_world1-boss');
                    Bento.audio.setVolume(1, 'bgm_world1-boss');
                    Globals.currentMusic = 'bgm_world1-boss';
                }

                // restore speed
                Bento.setGameSpeed(1);
            },
            update: function (data) {
                if (!boss) {
                    // find the boss, there should be at least one entity with the component Boss
                    Utils.forEach(Boss.currentBosses, function (bossEntity, i, l, breakLoop) {
                        bossEntity.getComponent('boss', function (bossComponent) {
                            if (!bossComponent.skipIntro) {
                                boss = bossEntity;
                            }
                        });
                    });
                    state = 'scrollToBoss';
                    return;
                }
                if (state === 'scrollToBoss') {
                    // set the target to the boss
                    cameraBehavior.target = boss;

                    // after 60 frames, start intro cut
                    Utils.timeout(60, showIntro);

                    state = 'wait';
                }
            }
        };
        var entity = new Entity({
            z: Globals.Layers.BOSSINTRO,
            name: 'bossIntro',
            position: settings.position,
            updateWhenPaused: Globals.PauseLevels.BOSSINTRO,
            float: true,
            components: [
                behavior,
                overlayFlash
            ]
        });
        return entity;
    };
});