/**
 * Coin
 * @moduleName Coin
 */
bento.define('entities/coin', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/sort',
    'components/movable',
    'components/shadow',
    'entities/effect',
    'managers/inventory',
    'managers/coincollector',
    'modules/matchmanager',
    'components/drawinviewport',
    'modules/bezier'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Sort,
    Movable,
    Shadow,
    Effect,
    InventoryManager,
    CoinCollector,
    MatchManager,
    DrawInViewport,
    Bezier
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var angle = settings.angle;
        var velocityZ = settings.velocityZ;
        var speed = settings.speed;
        var name = settings.name || settings.uniqueName;
        var ignoredFamilies = settings.ignoredFamilies || [];
        var double = 1;
        if (!name) {
            Utils.log('WARNING: coin has no unique id');
        }

        var sprite = new Sprite({
            imageName: 'coin',
            originRelative: new Vector2(0.5, 1),
            frameCountX: 4,
            frameCountY: 3,
            padding: 1,
            animations: {
                default: {
                    speed: 0.2,
                    frames: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                }
            }
        });
        var movable = new Movable({
            bounceZ: 0.5,
            frictionZ: 0.5,
            boundingBox: new Rectangle(-8, -8, 16, 16),
            ignoredFamilies: ignoredFamilies
        });
        var grabbingPlayers = [];

        var behavior = {
            name: 'behaviorComponent',
            onPickUp: function (msg) {
                // find corresponding player
                var otherPlayer = Bento.objects.get(msg.id);
                if (!otherPlayer) {
                    // failed getting the player: just remove the coin and move on
                    Bento.objects.remove(entity);
                } else {
                    behavior.pickUp(otherPlayer);
                }
                return true;
            },
            start: function (data) {
                movable.velocityZ = velocityZ;
                movable.grounded = false;

                var direction = angle;
                movable.velocity.x = speed;
                movable.velocity.y = 0;
                movable.velocity.rotateRadian(direction);

                behavior.collectPlayers();

                // clean up the message list before listening
                var messages = MatchManager.getMessages(name);
                messages.length = 0;
                MatchManager.onMessage(name, behavior.onPickUp);

                EventSystem.on('playerSpawned', behavior.collectPlayers);
            },
            collectPlayers: function () {
                var players = Bento.objects.getByFamily('player');

                // only local player can pickup the coin
                Utils.forEach(players, function (p, i, l, breakLoop) {
                    if (!p.isRemote) {
                        grabbingPlayers.push(p);
                        breakLoop();
                    }
                });
            },
            destroy: function (data) {
                EventSystem.off('playerSpawned', behavior.collectPlayers);

                // it's still possible for 2 player to pick up a coin at the same time
                // in that case the message that gets sent lingers around in the matchmanager
                // because the message listener is turned off on destroy
                MatchManager.offMessage(name, behavior.onPickUp);

                // other options:
                // allow remote players to pick up coins by setting grabbingPlayers = Bento.objects.getByFamily('player')
                // and don't send any message for picking up coins
                // however this causes coins to be desynced
            },
            update: function (data) {
                var dx;
                var dy;
                var ii;
                if (entity.timer > 30 && movable.height < 32 && movable.velocityZ >= 0) {
                    Utils.forEach(grabbingPlayers, function (player, i, l, breakLoop) {
                        player.getComponent('movable', function (playerMovable) {
                            if (!playerMovable.grounded) {
                                return;
                            }
                            dx = entity.position.x - player.position.x;
                            dy = entity.position.y - player.position.y;

                            if (dx * dx + dy * dy < 48 * 48) {
                                behavior.pickUp(player);
                            }
                        });
                    });
                }

                // remove double
                double = 1;
            },
            pickUp: function (target) {
                var hasCollectedBefore = false;
                behavior.tweenCoin(target, function () {
                    if (!hasCollectedBefore && !target.isRemote) {
                        target.getComponent('coinCounterComponent', function (coinCounter) {
                            coinCounter.addCoin(1);
                        });
                    }
                });
                movable.gravity = 0;

                Bento.objects.remove(entity);

                if (!target.isRemote) {
                    // collect coins
                    hasCollectedBefore = CoinCollector.collect(entity.name, 1);
                    // send message that coin was picked up to be sure that it will disappear on everyone's screen
                    MatchManager.sendData({
                        id: target.name
                    }, name);
                }
            },
            tweenCoin: function (target, onComplete) {
                var targetPos = target.position;
                var dist = entity.position.subtract(targetPos);
                var distMagnitude = dist.sqrMagnitude();
                var duration = distMagnitude / 100;
                var curve = new Bezier(
                    entity.position,
                    entity.position.add(new Vector2(
                        Utils.getRandomRange(-distMagnitude / 32, distMagnitude / 32),
                        Utils.getRandomRange(-distMagnitude / 32, distMagnitude / 32)
                    )),
                    targetPos.add(new Vector2(0, -16))
                );
                var showEffect = function () {
                    var effect = new Effect({
                        spriteSheet: 'effects/coin',
                        position: clone.position.clone()
                    });
                    // offset with the same height
                    effect.transform.y = entity.transform.y;
                    Bento.objects.attach(effect);
                    Bento.audio.stopSound('sfx_coin');
                    Bento.audio.playSound('sfx_coin');

                    if (onComplete) {
                        onComplete();
                    }
                };
                var clone;

                // remove sprite
                if (sprite && sprite.parent && sprite.parent.remove) {
                    sprite.parent.remove(sprite);
                }

                // make a similar coin
                clone = new Entity({
                    name: 'coinTween',
                    position: entity.position.clone(),
                    components: [
                        sprite,
                        new Sort({})
                    ]
                });
                // copy transform
                new Tween({
                    from: entity.transform.y,
                    to: 0,
                    in: duration,
                    onUpdate: function (v, t) {
                        clone.transform.y = v;
                    }
                });

                // tween towards target
                new Tween({
                    from: 0,
                    to: 1,
                    in: duration,
                    delay: 0,
                    ease: 'quadratic',
                    decay: 1,
                    oscillations: 1,
                    onStart: function () {
                        Bento.objects.attach(clone);
                    },
                    onUpdate: function (v, t) {
                        var curvePos = Utils.toVector2(curve.get(v));
                        // interpolate towards target position, because this position is moving
                        clone.position.x = curvePos.x * (1 - v) + targetPos.x * v;
                        clone.position.y = curvePos.y * (1 - v) + targetPos.y * v;
                    },
                    onComplete: function () {
                        showEffect();
                        clone.removeSelf();
                    }
                });

                // scale down
                // new Tween({
                //     from: 1,
                //     to: 0.5,
                //     in: duration,
                //     ease: 'quadratic',
                //     onUpdate: function (v, t) {
                //         clone.scale.x = v;
                //         clone.scale.y = v;
                //     },
                //     onComplete: function () {

                //     }
                // });
            }
        };

        var entity = new Entity({
            name: name,
            family: ['coins', 'pickups'],
            position: settings.position || new Vector2(0, 0),
            updateWhenPaused: 0,
            components: [
                sprite,
                behavior,
                movable,
                new Sort({}),
                new Shadow({
                    imageName: 'shadow16'
                }),
                new DrawInViewport({})
            ]
        });
        return entity;
    };
});