/**
 * Consumable item. 'Consumed' by the player when picked up (effects are applied).
 * @moduleName Consumable
 */
bento.define('entities/consumable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/sort',
    'components/shadow',
    'components/movable',
    'entities/effect',
    'managers/item'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Sort,
    Shadow,
    Movable,
    Effect,
    ItemManager
) {
    'use strict';
    var Consumable = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var item = 'consumables/' + settings.item;
        if (!Utils.hasAsset('json', item)) {
            Utils.log('ERROR: item consumables/' + item + ' does not exist');
            return;
        }
        var json = Bento.assets.getJson(item);
        var sprite = new Sprite({
            spriteSheet: json.sprite
        });
        var ignoredFamilies = settings.ignoredFamilies || [];
        var movable = new Movable({
            bounceZ: 0.5,
            frictionZ: 0.5,
            boundingBox: Utils.toRectangle(json.boundingBox),
            ignoredFamilies: ignoredFamilies
        });
        var players = Bento.objects.getByFamily('player');
        var pickUp = function (target) {
            var itemHandlers = ItemManager.getItemHandlers();
            var handler = itemHandlers[item];

            movable.gravity = 0;
            Bento.objects.remove(entity);
            Bento.audio.stopSound('sfx_consumable');
            Bento.audio.playSound('sfx_consumable');

            // apply effect to local player
            if (!target.isRemote) {
                handler(item, target, entity.position.clone());
            }
        };

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                var angle = settings.angle;
                var velocityZ = settings.velocityZ;
                var speed = settings.speed;
                movable.velocityZ = velocityZ;
                movable.grounded = false;

                var direction = angle;
                movable.velocity.x = speed;
                movable.velocity.y = 0;
                movable.velocity.rotateRadian(direction);

                movable.grounded = false;
            },
            destroy: function (data) {},
            update: function (data) {
                var dx;
                var dy;
                var ii;
                if (movable.height < 32 && movable.velocityZ >= 0 && entity.timer > 60) {
                    Utils.forEach(players, function (player, i, l, breakLoop) {
                        // all players can pick up an item, even if it's a remote
                        player.getComponent('movable', function (playerMovable) {
                            if (!playerMovable.grounded) {
                                return;
                            }
                            dx = entity.position.x - player.position.x;
                            dy = entity.position.y - player.position.y;

                            if (dx * dx + dy * dy < 16 * 16) {
                                pickUp(player);
                            }
                        });
                    });

                }
            },
            draw: function (data) {}
        };


        var entity = new Entity({
            z: 0,
            name: 'consumable',
            family: ['consumables', 'pickups'],
            position: settings.position || new Vector2(0, 0),
            updateWhenPaused: 0,
            components: [
                sprite,
                behavior,
                movable,
                new Sort({}),
                new Shadow({
                    imageName: 'shadow16'
                })
            ]
        });
        return entity;
    };

    return Consumable;
});