/**
 * Bat enemy
 * @moduleName Bat
 */
bento.define('entities/enemies/bat', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'components/ai/walkingenemy',
    'entities/effect',
    'components/sort',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    WalkingEnemy,
    Effect,
    Sort,
    Punch
) {
    'use strict';
    var Bat = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 2,
            "skill": -2,
            "health": 4,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'bat/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            hurry: 0.2
        });
        // overwrite states
        var stateMachine;
        var batBehavior = {
            name: 'batBehavior',
            start: function (data) {
                stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addStates(Bat.flyStates);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 64,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                batBehavior
            ]
        });

        return entity;
    };
    Bat.flyStates = {
        // idle can be on the ground or flying
        idle: {
            start: function (data, transitionData) {
                // attack type is randomized, sometimes bat will spin attack on the ground
                data.sync({
                    spin: Utils.getRandom(2)
                });

                var walkingEnemy = data.walkingEnemyBehavior;
                data.idleTime = Utils.getRandomRange(walkingEnemy.hesitation[0], walkingEnemy.hesitation[1]);

                if (data.isFlying) {
                    data.sprite.setSpriteSheet(data.rootPath + 'idle-flying');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                }
                data.movable.gravity = 0;

                // take off asap
                if (!data.isFlying && !data.isRemote) {
                    // keep this after any sprite change
                    data.setState('takeOff');
                }
            },
            action: function (data, action) {
                if (data.isFlying && data.idleTime > 0) {
                    return;
                }
                if (action.data) {
                    if (data.isFlying) {
                        // 50% chance of dropping down to do a spin attack
                        if (data.syncData.spin === 0) {
                            // land before doing a spin attack
                            data.setState('landSpinAttack');
                        } else {
                            // dive attack
                            data.setState('diveAttack');
                        }
                    } else {
                        // no need to land first, do the spin attack right away
                        data.setState('spinAttack');
                    }
                }
            },
            update: function (data) {
                var movable = data.movable;
                WalkingEnemy.states.idle.update(data);

                // stop moving during idle
                if (data.isFlying) {
                    // brake slowly in the air
                    movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.1);
                    movable.velocity.y = Utils.approach(movable.velocity.y, 0, 0.1);

                    // max height
                    if (movable.height < -30) {
                        movable.height = -30;
                        movable.velocityZ = 0;
                    }
                }

            },
            move: WalkingEnemy.states.idle.move,
            end: WalkingEnemy.states.idle.end,
            canGrab: true
        },
        // "walk" is now a flying state
        // check if bat needs to fly first
        walk: {
            start: function (data, transitionData) {
                // bat can't walk on the ground, so it needs to take off first
                if (!data.isFlying && !data.isRemote) {
                    data.setState('takeOff');
                    return;
                }
                data.movable.gravity = 0;
                WalkingEnemy.states.walk.start(data, transitionData);
            },
            update: function (data) {
                data.walkTime--;
                if (data.walkTime <= 0 && data.behavior === 'idle' && !data.isRemote) {
                    data.setState('idle');
                }
                // max height
                if (data.isFlying && data.movable.height < -30) {
                    data.movable.height = -30;
                    data.movable.velocityZ = 0;
                }
            },
            move: function (data, direction) {
                // lerp towards new direction instead of setting new direction immediately
                var walkingEnemy = data.walkingEnemyBehavior;
                var target = direction.scalarMultiply(1.5 + walkingEnemy.hurry);
                data.movable.velocity.x += (target.x - data.movable.velocity.x) * 0.1;
                data.movable.velocity.y += (target.y - data.movable.velocity.y) * 0.1;
            },
            action: WalkingEnemy.states.walk.action,
            canGrab: false
        },
        // start flying
        takeOff: {
            start: function (data) {
                data.movable.gravity = 0;
                data.isFlying = true;

                data.sprite.setSpriteSheet(data.rootPath + 'jump-start', function () {
                    data.movable.velocityZ = -1;
                    data.sprite.setSpriteSheet(data.rootPath + 'flying-start', function () {
                        data.sprite.setSpriteSheet(data.rootPath + 'idle-flying');
                    });
                });
            },
            update: function (data) {
                var movable = data.movable;
                // max height
                if (movable.height < -30) {
                    movable.height = -30;
                    movable.velocityZ = 0;
                    data.setState('idle');
                }

                // don't move
                movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.05);
                movable.velocity.y = Utils.approach(movable.velocity.y, 0, 0.05);
            }
        },
        // bat has no punch
        punch: {
            start: function (data) {
                data.setState('idle');
            }
        },
        diveAttack: {
            ignoreStun: true,
            ignoreDirection: true,
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'swoopdown-start', function () {
                    data.swoopDown = 1;
                    data.movable.gravity = -0.1;
                    data.movable.velocity.x = data.multiScale.direction.x * 2.5;
                    data.movable.velocityZ = 2.5;
                    data.sprite.setSpriteSheet(data.rootPath + 'swoopdown');
                    Bento.audio.stopSound('sfx_throw');
                    Bento.audio.playSound('sfx_throw');
                });
            },
            update: function (data) {
                var movable = data.movable;
                if (data.movable.velocityZ < -1) {
                    data.sprite.setSpriteSheet(data.rootPath + 'swoopup');
                }
                // attach punch object when low enough and diving down
                if (movable.height > -16 && data.movable.velocityZ > 0) {
                    if (data.punchObject) {
                        data.punchObject.removeSelf();
                        data.punchObject = null;
                    }
                    // attach punch object
                    data.punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: data.attack || 1,
                        stamina: 1,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        duration: Infinity,
                        //pushBack: new Vector2(0.5, 1.5),
                        pushBack: new Vector2(0.2, 2),
                        parent: data.entity,
                        boundingBox: new Rectangle(-24, -8, 48, 16),
                        position: data.entity.position.add(new Vector2(0, 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx2-medium',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        })
                    });
                    Bento.objects.attach(data.punchObject);
                }
                // back to ascending
                if (movable.height < -24 && data.movable.velocityZ < 0) {
                    // remove punch
                    if (data.punchObject) {
                        data.punchObject.removeSelf();
                        data.punchObject = null;
                    }

                    // start slowing down
                    movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.05);
                    // data.sprite.setSpriteSheet(data.rootPath + 'swoopup-turn', function () {});
                }
                // max height
                if (movable.height < -30 && data.movable.velocityZ < 0) {
                    movable.height = -30;
                    movable.velocityZ = 0;

                    // end of attack
                    data.setState('idle');
                }

                // touched the ground should not ground the bat
                if (movable.height >= 0) {
                    data.movable.velocity.x = data.multiScale.direction.x * 2.5;
                    movable.grounded = false;
                }
            },
            end: function (data) {
                data.movable.gravity = 0;
                if (data.punchObject) {
                    data.punchObject.removeSelf();
                    data.punchObject = null;
                }
            }
        },
        landSpinAttack: {
            ignoreStun: true,
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data) {
                // land to do a spin attack
                data.movable.gravity = 1;
            },
            update: function (data) {
                if (data.movable.grounded) {
                    data.isFlying = false;
                    data.sprite.setSpriteSheet(data.rootPath + 'land', function () {
                        // start the actual spin attack
                        data.setState('spinAttack');
                    });
                }
            },
            end: function (data) {
                // normal gravity
                data.movable.gravity = 0.15;
            }
        },
        spinAttack: {
            ignoreDirection: true,
            start: function (data) {
                data.spinDuration = 0;
                data.isSpinning = false;
                data.sprite.setSpriteSheet(data.rootPath + 'spin-start', function () {
                    data.sprite.setSpriteSheet(data.rootPath + 'spin');
                    data.isSpinning = true;

                    Bento.audio.stopSound('sfx_startdash');
                    Bento.audio.playSound('sfx_startdash');

                    if (data.punchObject) {
                        data.punchObject.removeSelf();
                        data.punchObject = null;
                    }
                    // attach punch object
                    data.punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: data.attack || 1,
                        stamina: 1,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        duration: Infinity,
                        //pushBack: new Vector2(0.5, 1.5),
                        pushBack: new Vector2(0.2, 2),
                        parent: data.entity,
                        boundingBox: new Rectangle(-38, -12, 76, 24),
                        position: data.entity.position.add(new Vector2(0, 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx2-medium',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        })
                    });
                    Bento.objects.attach(data.punchObject);

                });
            },
            update: function (data) {
                data.spinDuration += 1;

                if (data.spinDuration === 180) {
                    data.setState('dizzy');
                }
                // slowly move
                if (data.isSpinning) {
                    data.movable.velocity.x = data.multiScale.direction.x * 0.5;
                    if (data.spinDuration % 4 === 0) {
                        var puff = new Effect({
                            position: data.entity.position.add(new Vector2(0, -1)),
                            spriteSheet: 'effects/walkcloud'
                        });
                        puff.scale.x = Utils.sign(data.movable.velocity.x);
                        puff.attach(new Sort({}));
                        Bento.objects.attach(puff);
                    }
                }
            },
            end: function (data) {
                if (data.punchObject) {
                    data.punchObject.removeSelf();
                    data.punchObject = null;
                }
            }
        },
        dizzy: {
            ignoreStun: true,
            ignoreDirection: true,
            start: function (data) {
                data.dizzyDuration = 0;
                data.sprite.setSpriteSheet(data.rootPath + 'dizzy-start', function () {
                    data.sprite.setSpriteSheet(data.rootPath + 'dizzy');
                });
            },
            update: function (data) {
                data.dizzyDuration += 1;

                if (data.dizzyDuration === 120) {
                    data.sprite.setSpriteSheet(data.rootPath + 'dizzy-end', function () {
                        data.setState('idle');
                    });
                }
            }
        }
    };
    return Bat;
});