/**
 * Cat enemy
 * @moduleName Cat
 */
bento.define('entities/enemies/cat', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    Punch
) {
    'use strict';
    return function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var jumpDistance = 40;
        var jumpHeight = 1.2;
        var punchObject;
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 2,
            "skill": 2,
            "health": 10,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));
        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'cat/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            hurry: 0.6,
        });
        // overwrite states
        var addStates = {
            name: 'extraStates',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addStates(states);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 26,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                addStates
            ]
        });

        var states = {
            punch: {
                ignoreStun: true,
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    var target = Bento.objects.getByName(transitionData.target)[0];
                    if (target) {
                        data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                    }

                    data.sync({
                        attackType: Utils.getRandom(2)
                    });

                    var comboAttack = function () {
                        data.sprite.setSpriteSheet(data.rootPath + 'attack1', function () {
                            data.punchState = 3;
                            data.sprite.setSpriteSheet(data.rootPath + 'attack2', function () {
                                data.setState('idle');
                            });
                        });
                    };
                    data.punchState = 1;


                    data.attackType = data.syncData.attackType;
                    if (data.attackType === 0) {
                        comboAttack();
                    } else {
                        data.distance = jumpDistance;
                        data.jumpHeight = -jumpHeight;

                        data.sprite.setSpriteSheet(data.rootPath + 'catjump', function () {
                            data.punchState = 2;
                            data.jumpHeight = -jumpHeight;
                            data.distance = jumpDistance;
                            //hacky set spritesheet name to something else so that the spritesheet switches
                            data.sprite.currentSpriteSheet = 'none';
                            data.sprite.setSpriteSheet(data.rootPath + 'catjump', function () {
                                data.punchState = 1;
                                data.attackType = 0;
                                comboAttack();
                            });
                        });
                    }
                    data.punchable.isInvincible = false;
                },
                update: function (data) {
                    if (data.attackType === 0) {
                        // combo punch
                        if ((data.punchState === 1 || data.punchState === 3) && data.sprite.currentFrame >= 3) {
                            punchObject = new Punch({
                                target: ['player', 'citizen'],
                                attack: Math.ceil(data.attack / 2) || 1,
                                stamina: data.stamina / 2 || 0.334,
                                skill: data.stats.skill,
                                direction: data.multiScale.direction.x,
                                duration: data.duration,
                                pushBack: new Vector2(data.punchState === 1 ? 0.2 : 1.5, data.punchState === 1 ? 2 : 1.4),
                                parent: data.entity,
                                boundingBox: new Rectangle(0, -5, 30, 10),
                                position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -30 : 0), 0)),
                                onHit: Punch.getDefaultCallback({
                                    entity: data.entity,
                                    effect: 'effects/hitfx3-small',
                                    sfx: 'sfx_punch1',
                                    taptic: 'medium'
                                }),
                                debug: true
                            });
                            Bento.objects.attach(punchObject);
                            ++data.punchState;
                        }
                    } else {
                        if (data.sprite.currentFrame >= 1) {
                            data.movable.velocityZ += data.jumpHeight;
                            data.jumpHeight = 0;

                            // jump back and forward
                            var sign = data.multiScale.direction.x;
                            if (data.punchState === 1) {
                                sign = -sign;
                            }

                            // move when trying to tackle
                            data.movable.velocity.x = data.distance * 0.1 * sign;
                            data.movable.velocity.y = 0;
                            data.distance *= 0.9;
                        }
                    }
                }
            }
        };

        return entity;
    };
});