/**
 * Gangsta Boss, Fights with his fists. Does a spinning attack every once in a while.
 * @moduleName GangstaBoss
 */
bento.define('entities/enemies/gangstaboss', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'entities/punch',
    'globals',
    'states/bossenemy'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Punch,
    Globals,
    BossEnemyStates
) {
    'use strict';
    return function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": -2,
            "skill": 5,
            "health": 30,
            "coins": 0,
            "stamina": 0.5,
            "minHits": 15
        }));
        var stateMachine;
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'gangstaboss/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            onDeath: function () {
                // fire the 'Some Kind of Rapper' Achievement
                Utils.postAchievement("CgkI9NqdkoAGEAIQBw");
            }
        });

        var addedCollisionEvents = false;
        var currentScreen = Bento.screens.getCurrentScreen();
        var insideArea = true;

        // var jumpHeight = 10;
        // var secondJumpHeight = 3;
        // var jumpSpeed = 1;
        // var jumpTarget;

        var attackState = 0;
        var attacksTilSpin = 2;
        var spinTimer = 0;
        var whenToSpin = 40;

        var spinMoveSpeed = 3;

        var punchObject;

        var onFaint = function (data) {
            if (stateMachine.data.enemyController.isLocallyControlled) {
                ++attackState;
            }
        };

        var manageCollision = function () {
            if (!addedCollisionEvents) {
                stateMachine.data.movable.collisionListeners.push(onCollision);
                addedCollisionEvents = true;
            }
        };

        var onCollision = function (colData) {
            // .push(function (colData) {
            //     if(colData.other.family[0] === 'solid') {
            if (stateMachine.stateName === 'spinAttack') {
                if (colData.other.name === 'gridEntry' ||
                    colData.other.name === 'solid' ||
                    colData.other.name.split('-')[0] === 'prop'
                ) {

                    var currentVelocity = degreesToVector(stateMachine.data.spinRotation);
                    var coordKey = xOrYCollision(colData.other);
                    currentVelocity[coordKey] *= -1;
                    stateMachine.data.spinRotation = vectorToDegrees(currentVelocity);
                }
            }
        };

        var insideLevel = function () {
            var bb = entity.boundingBox.offset(entity.position);
            insideArea = true;
            if (bb.getX2() > currentScreen.dimension.width || bb.x < 0) {
                insideArea = false;
            }
            return insideArea;
        };

        var xOrYCollision = function (other) {
            // when hitboxes overlap horizontally the collision happened on the Y axis
            if (entity.position.x < other.position.x + other.dimension.width &&
                entity.position.x + stateMachine.data.movable.boundingBox.x + stateMachine.data.movable.boundingBox.width > other.position.x
            ) {
                return 'y';
            }
            return 'x';
        };

        var behavior = {
            name: 'gangstaBossBehavior',
            start: function () {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);

                // boss should be aggressive
                stateMachine.data.enemyController.aggression = 0.7;

                entity.getComponent('punchable', function (punchable) {
                    punchable.onFaint(onFaint);
                });

                // because grabbed states are being added later we need to add our own grabstates later too
                Utils.timeout(2, function () {
                    stateMachine.addStates(grabState);
                });

                if (stateMachine.data.enemyController.isLocallyControlled) {
                    manageCollision();
                }
            },
            update: function () {
                if (stateMachine.data.enemyController.isLocallyControlled) {
                    if (attackState >= attacksTilSpin) {
                        if (insideLevel()) {
                            var punchable = entity.getComponent('punchable');
                            if (stateMachine.stateName === 'idle' &&
                                !punchable.isDead
                            ) {
                                ++spinTimer;
                                if (spinTimer >= whenToSpin) {
                                    stateMachine.setState('spinAttack');
                                }
                            }
                        } else {
                            // set spintimer to spin immediately when done walking back in
                            spinTimer = whenToSpin;
                        }
                    }
                }
            },
            destroy: function () {}
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 50,
            boundingBox: new Rectangle(-18, -6, 36, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 72,
            components: [
                walkingEnemy,
                behavior
            ]
        });

        var degreesToVector = function (value) {
            var rads = Utils.toRadian(value);
            return new Vector2(Math.cos(rads), Math.sin(rads));
        };

        var vectorToDegrees = function (vector) {
            return Utils.toDegree(Math.atan2(vector.y, vector.x));
        };

        // var gangstaAI = function (data, t, chaseBehavior) {
        //     chaseBehavior(data);
        // };

        var states = {
            idle: {
                start: StandardEnemyStates.idle.start,
                move: function (data, direction) {
                    if (attackState < attacksTilSpin || !insideArea) {
                        data.setState('walk');
                    }
                },
                action: function (data, action) {
                    // punch when not waiting for spin
                    if (attackState < attacksTilSpin) {
                        StandardEnemyStates.idle.action(data, action);
                    }
                },
                canGrab: true
            },
            punch: {
                ignoreStun: true,
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.punchAnim = (data.punchAnim || 0) + 1;
                    data.sprite.setSpriteSheet(data.rootPath + 'punch' + (data.punchAnim % 2 + 1), function () {
                        data.setState('idle');
                    });

                    data.punchState = 1;

                    //turn towards player
                    var t = Bento.objects.get(transitionData.target);

                    if (t) {
                        data.multiScale.direction.x = Utils.sign(t.position.x - data.entity.position.x);
                    }
                    data.punchable.isInvincible = false;
                },
                update: function (data) {
                    if (data.punchState === 1 && data.sprite.currentFrame >= 3) {
                        punchObject = new Punch({
                            target: ['player', 'citizen'],
                            attack: data.attack * 0.5 || 1,
                            stamina: data.stamina || 0.334,
                            skill: data.stats.skill,
                            direction: data.multiScale.direction.x,
                            pushBack: new Vector2(2, 1.5),
                            parent: data.entity,
                            boundingBox: new Rectangle(0, -5, data.entity.punchRange || 50, 10),
                            position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -50 : 0), 0)),
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx3',
                                sfx: 'sfx_punch1',
                                taptic: 'medium'
                            }),
                            debug: true
                        });
                        Bento.objects.attach(punchObject);
                        data.punchState = 2;
                    }
                },
                end: function (data) {
                    ++attackState;
                }
            },
            dizzy: {
                ignoreBlock: true,
                ignoreDirection: true,
                canGrab: false,
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'dizzy');
                    data.punchable.resetStamina();
                    data.punchable.isInvincible = false;
                    data.walkingEnemyBehavior.noHurt = true;
                    data.dizzyTimer = 180;
                },
                update: function (data) {
                    if (--data.dizzyTimer < 0) {
                        data.setState('idle');
                        // data.punchable.hasFainted = true;
                    }
                },
                end: function (data) {
                    data.walkingEnemyBehavior.noHurt = false;
                }
            },
            spinAttack: {
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.movable.familiesToCollide.push('solidPlayer');
                    data.sprite.setSpriteSheet(data.rootPath + 'spinattack');
                    data.spinRotation = Utils.getRandom(360);
                    data.spinAmount = 5;
                    data.momentum = 0;
                    data.spinDirection = data.multiScale.direction.x;
                    data.punchable.isInvincible = true;

                    punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: data.attack * 1.5 || 1,
                        stamina: 1,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        duration: Infinity,
                        //pushBack: new Vector2(0.5, 1.5),
                        pushBack: new Vector2(0.2, 3),
                        parent: data.entity,
                        boundingBox: new Rectangle(0, -10, 60, 20),
                        position: data.entity.position.add(new Vector2(-30, 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3',
                            sfx: 'sfx_punch3',
                            taptic: 'medium'
                        }),
                        debug: true
                    });
                    punchObject.rememberPunched = false;
                    Bento.objects.attach(punchObject);

                    Bento.audio.stopSound('sfx_gangstaboss-spinattack');
                    Bento.audio.playSound('sfx_gangstaboss-spinattack');
                },
                update: function (data) {
                    if (data.enemyController.isLocallyControlled) {
                        // spin position gets decided by the player that controls him, others just sync position
                        // add spinAmount to spinRotation and decrease spinAmount each time, this makes the circle bigger and bigger
                        data.spinRotation += data.spinAmount * data.spinDirection;
                        data.spinRotation += 360;
                        data.spinRotation = data.spinRotation % 360;
                        data.spinAmount *= 0.995;

                        // increase momentum
                        if (data.momentum < 1) {
                            data.momentum += 0.02;
                        }

                        // convert degrees into vector and move gangstaboss according to spinMoveSpeed
                        var moveAngle = degreesToVector(data.spinRotation);
                        data.movable.velocity.x = moveAngle.x * (spinMoveSpeed * data.momentum);
                        data.movable.velocity.y = moveAngle.y * (spinMoveSpeed * data.momentum);

                        if (data.spinAmount < 0.8) {
                            data.setState('dizzy');
                            attackState = 0;
                            data.punchable.isInvincible = false;
                        }
                    }
                },
                end: function (data) {
                    var colIndex = data.movable.familiesToCollide.indexOf('solidPlayer');
                    if (colIndex !== -1) {
                        data.movable.familiesToCollide.splice(colIndex, 1);
                    }
                    data.punchable.isInvincible = false;
                    spinTimer = 0;
                    punchObject.removeSelf();
                },
                canGrab: false
            },
            die: StandardEnemyStates.die,
            getUp: StandardEnemyStates.getUp,
            jumpBack: {
                ignoreDirection: true,
                start: StandardEnemyStates.jumpBack.start,
                update: function (data) {
                    if (data.movable.grounded) {
                        data.setState('idle');
                        data.punchable.isInvincible = false;
                        if (attackState < attacksTilSpin) {
                            attackState = attacksTilSpin;
                        }
                    }
                }
            }
        };

        var grabState = {
            grabbed: BossEnemyStates.grabbed,
            grabbedThrow: BossEnemyStates.grabbedThrow
        };

        return entity;
    };
});