/**
 * mr pigeon enemy, can jump around, shoots his hand. (long range attack)
 * @moduleName MrPigeon
 */
bento.define('entities/enemies/mrpigeon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'entities/punch',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Punch,
    Globals
) {
    'use strict';
    return function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 0,
            "skill": 1,
            "health": 3,
            "coins": 0,
            "stamina": 0.5,
            "minHits": 5
        }));
        var stateMachine;
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'mrpigeon/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
        });

        var behavior = {
            name: 'mrPigeonBehavior',
            start: function () {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 90,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                behavior
            ]
        });

        var states = {
            walk: {
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.punchable.isInvincible = false;
                    // if local player is the target decide if should hop or walk
                    var hop;
                    if (data.enemyController.targetPlayerId === -1 ||
                        (MatchManager.localPlayer && MatchManager.localPlayer.playerId === data.enemyController.targetPlayerId)) {
                        if (data.enemyController.target &&
                            data.entity.position.distance(data.enemyController.target.position
                                .add(data.enemyController.targetOffset)) > 40) {
                            hop = true;
                        } else {
                            hop = false;
                        }
                    } else {
                        hop = false;
                    }

                    // sync decision
                    data.sync({
                        hop: hop
                    });

                    // set animation
                    if (data.syncData.hop) {
                        data.sprite.setSpriteSheet(data.rootPath + 'pigeonjump');
                        data.hasToHop = true;
                    } else {
                        data.sprite.setSpriteSheet(data.rootPath + 'walk');
                    }
                },
                update: function (data) {
                    if (data.movable.velocity.x > 0) {
                        data.multiScale.direction.x = 1;
                    } else if (data.movable.velocity.x < 0) {
                        data.multiScale.direction.x = -1;
                    }

                    // set to idle if player is too far when walking
                    if (!data.syncData.hop) {
                        if (data.enemyController.targetPlayerId === -1 ||
                            (MatchManager.localPlayer && MatchManager.localPlayer.playerId === data.enemyController.targetPlayerId)) {
                            if (data.enemyController.target &&
                                data.entity.position.distance(data.enemyController.target.position
                                    .add(data.enemyController.targetOffset)) > 50) {
                                data.setState('idle');
                            }
                        }
                    } else {
                        if (data.movable.grounded && !data.hasToHop) {
                            data.setState('idle');
                            data.sprite.setSpriteSheet(data.rootPath + 'pigeonland', function () {
                                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                            });
                        }
                    }
                },
                move: function (data, direction) {
                    var viewport = Bento.getViewport().clone();
                    if (data.syncData.hop) {
                        if(data.movable.grounded) {
                            data.movable.velocity = direction.scalarMultiply(1.8);
                            data.movable.velocityZ = -2;
                            data.hasToHop = false;
                        }
                    } else {
                        data.movable.velocity = direction.scalarMultiply(1.5);

                        data.sprite.setSpriteSheet(data.rootPath + 'walk');
                    }
                },
                action: function (data, action) {
                    //ai controller wants to punch, cancel walk and switch to idle
                    data.setState('idle');
                }
            },
            punch: {
                ignoreStun: true,
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.punchAnim = (data.punchAnim || 0) + 1;
                    data.sprite.setSpriteSheet(data.rootPath + 'attack', function () {
                        data.setState('idle');
                    });

                    data.punchState = 1;

                    //turn towards player
                    var target = Bento.objects.getByName(transitionData.target)[0];
                    if (target) {
                        data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                    }
                    data.punchable.isInvincible = false;
                },
                update: function (data) {
                    if (data.punchState === 1 && data.sprite.currentFrame >= 4) {
                        var punchObject = new Punch({
                            target: ['player', 'citizen'],
                            attack: data.attack || 1,
                            stamina: data.stamina || 0.334,
                            skill: data.stats.skill,
                            direction: data.multiScale.direction.x,
                            pushBack: new Vector2(0.5, 1.5),
                            parent: data.entity,
                            boundingBox: new Rectangle(0, -5, data.entity.punchRange || 32, 10),
                            position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -data.entity.punchRange : 0), 0)),
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx3-medium',
                                sfx: 'sfx_punch1',
                                taptic: 'medium'
                            }),
                            debug: true
                        });
                        Bento.objects.attach(punchObject);

                        data.punchState = 2;
                    }
                }
            },
        };

        return entity;
    };
});