/**
 * Pig enemy
 * @moduleName Pig
 */
bento.define('entities/enemies/pig', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'components/ai/walkingenemy',
    'entities/effect',
    'components/sort',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    WalkingEnemy,
    Effect,
    Sort,
    Punch
) {
    'use strict';
    var Pig = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 2,
            "skill": -2,
            "health": 4,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'pig/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            hurry: 0.2
        });
        // overwrite states
        var stateMachine;
        var pigBehavior = {
            name: 'pigBehavior',
            start: function (data) {
                stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addStates(Pig.rollStates);
                stateMachine.data.cooldown = 3;
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 24,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                pigBehavior
            ]
        });

        return entity;
    };
    Pig.rollStates = {
        // overwrite the WalkingEnemy idle state
        idle: {
            start: function (data, transitionData) {
                // 50% chance of starting a roll when inside level
                if (!data.cooldown &&
                    data.entity.position.x > 16 &&
                    data.entity.position.x < Bento.screens.getCurrentScreen().dimension.width - 16
                ) {
                    data.sync({
                        roll: Utils.getRandom(3)
                    });
                } else {
                    --data.cooldown;
                    data.sync({
                        roll: 1
                    });
                }

                WalkingEnemy.states.idle.start(data, transitionData);

                if (data.syncData.roll === 0 && !data.isRemote) {
                    data.cooldown = 2;
                    data.setState('startRoll');
                }
            },
            action: WalkingEnemy.states.idle.action,
            update: WalkingEnemy.states.idle.update,
            move: WalkingEnemy.states.idle.move,
            end: WalkingEnemy.states.idle.end,
            canGrab: true
        },
        startRoll: {
            ignoreStun: true,
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'roll-start', function () {
                    data.setState('roll');
                });
            }
        },
        roll: {
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data) {
                data.sync({
                    angle: Utils.getRandomRange(0, 90),
                    sign: Utils.getRandom(2) * 2 - 1
                });

                data.sprite.setSpriteSheet(data.rootPath + 'roll');

                // allow collisions with solid players?
                data.movable.familiesToCollide.push('solidPlayer');

                // set movable to bounce
                data.movable.canBounce = true;
                data.movable.friction = new Vector2(1, 1); // friction is a multiplier
                data.movable.bounce = new Vector2(1, 1); // bounce factor

                // disable braking by the walking enemy behavior
                data.walkingEnemyBehavior.friction = 0;

                // pick a random direction
                data.movable.velocity.x = data.multiScale.direction.x * 2;
                data.movable.velocity.y = 0;
                data.movable.velocity.rotateDegree(data.syncData.angle * data.syncData.sign);

                data.rollTimer = 0;

                // create punch
                if (data.punchObject) {
                    data.punchObject.removeSelf();
                    data.punchObject = null;
                }
                data.punchObject = new Punch({
                    target: ['player', 'citizen'],
                    attack: data.attack || 1,
                    stamina: 1,
                    skill: data.stats.skill,
                    direction: data.multiScale.direction.x,
                    duration: Infinity,
                    //pushBack: new Vector2(0.5, 1.5),
                    pushBack: new Vector2(0.2, 2),
                    parent: data.entity,
                    boundingBox: new Rectangle(-16, -8, 32, 16),
                    position: data.entity.position.add(new Vector2(0, 0)),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx2-medium',
                        sfx: 'sfx_punch1',
                        taptic: 'medium'
                    })
                });
                Bento.objects.attach(data.punchObject);
            },
            update: function (data) {
                data.rollTimer += 1;
                if (data.rollTimer % 4 === 0) {
                    var puff = new Effect({
                        position: data.entity.position.add(new Vector2(0, -1)),
                        spriteSheet: 'effects/walkcloud'
                    });
                    puff.scale.x = Utils.sign(data.movable.velocity.x);
                    puff.attach(new Sort({}));
                    Bento.objects.attach(puff);
                }

                if (data.rollTimer > 90) {
                    // quit rolling after a while
                    data.setState('endRoll');
                }
            },
            end: function (data) {
                // revert
                data.walkingEnemyBehavior.friction = 0.15;
                data.movable.canBounce = false;

                if (data.punchObject) {
                    data.punchObject.removeSelf();
                    data.punchObject = null;
                }
            }
        },
        endRoll: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'roll-end', function () {
                    data.setState('idle');
                });
            }
        }
    };
    return Pig;
});