/**
 * Pig Chef Boss, Fights with a cleaver. Sometimes jumps and tries to land on the player or does a tackle.
 * @moduleName PigChef
 */
bento.define('entities/enemies/pigchef', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'entities/punch',
    'globals',
    'states/bossenemy'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Punch,
    Globals,
    BossEnemyStates
) {
    'use strict';
    return function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": -2,
            "skill": 5,
            "health": 30,
            "coins": 0,
            "stamina": 0.5,
            "minHits": 15
        }));
        var stateMachine;
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'pigchef/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            onDeath: function () {
                // fire the 'MEEEEAAT!' Achievement
                Utils.postAchievement("CgkI9NqdkoAGEAIQBg");
            }
        });

        var camera;

        var jumpHeight = 10;
        var secondJumpHeight = 3;
        var jumpSpeed = 1;
        var jumpTarget;

        var attackState = 0;
        var jumpTimer = 0;
        var whenToJump = 180;

        // find closest player and set its position as the target
        var setJumpTarget = function () {
            var players = Bento.objects.getByFamily('player');
            var closest = players[0];
            var distance = entity.position.distance(players[0].position);
            var d;
            var sMachine;
            Utils.forEach(players, function (p, i, l, breakLoop) {
                sMachine = p.getComponent('stateMachine');
                if (sMachine && sMachine.stateName !== 'dead') {
                    d = entity.position.distance(p.position);
                    if (d < distance) {
                        closest = p;
                        distance = d;
                    }
                }
            });
            jumpTarget = closest.position;
        };

        var onFaint = function (data) {
            if (stateMachine.data.enemyController.isLocallyControlled) {
                ++attackState;
            }
        };

        var behavior = {
            name: 'pigChefBehavior',
            start: function () {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);

                stateMachine.data.enemyController.ai = pigAI;

                // boss should be aggressive
                stateMachine.data.enemyController.aggression = 0.7;

                camera = Bento.objects.getByFamily('camera')[0];

                EventSystem.on('startRoom', onRoomStart);

                entity.getComponent('punchable', function (punchable) {
                    punchable.onFaint(onFaint);
                });

                // because grabbed states are being added later we need to add our own grabstates later too
                Utils.timeout(2, function () {
                    stateMachine.addStates(grabState);
                });
            },
            destroy: function () {
                EventSystem.off('startRoom', onRoomStart);
            },
            update: function () {
                if (stateMachine.data.enemyController.isLocallyControlled) {
                    if (attackState >= 3) {
                        var punchable = entity.getComponent('punchable');
                        ++jumpTimer;
                        if (jumpTimer >= whenToJump &&
                            stateMachine.stateName !== 'jumpAttack' &&
                            stateMachine.stateName !== 'faint' &&
                            stateMachine.stateName !== 'getUp' &&
                            !punchable.isDead
                        ) {
                            stateMachine.setState('jumpAttack');
                            setJumpTarget();
                        }
                    }
                }
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 40,
            boundingBox: new Rectangle(-18, -6, 36, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 72,
            components: [
                walkingEnemy,
                behavior
            ]
        });

        var pigAI = function (data, t, chaseBehavior) {
            if (stateMachine.stateName !== 'jumpAttack') {
                if (attackState < 2) {
                    stateMachine.data.enemyController.attackDistance = 40;
                } else {
                    stateMachine.data.enemyController.attackDistance = 80;
                }
                chaseBehavior(data);
            } else {
                // entity.moveInDirection(jumpTarget.subtract(entity.position));
                entity.moveInDirection();
            }
        };

        var states = {
            punch: {
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.punchState = 1;

                    // sync attackstate
                    data.sync({
                        attackState: attackState
                    });

                    if (data.enemyController.isLocallyControlled) {
                        ++attackState;
                    }

                    // synced attack state determines the attack, cleaver or tackle

                    var attackString;
                    if (data.syncData.attackState > 2) {
                        data.setState('idle');
                        return;
                    } else if (data.syncData.attackState === 2) {
                        attackString = 'tackle';
                        data.duration = 18;
                        data.attackFrame = 4;
                        data.slideTime = 1.5;
                        data.slide = 50;
                        data.punchable.isInvincible = true;
                    } else {
                        attackString = 'attack1';
                        data.duration = 10;
                        data.attackFrame = 3;
                        data.slide = 0;
                        data.punchable.isInvincible = false;
                    }

                    data.sprite.setSpriteSheet(data.rootPath + attackString, function () {
                        data.setState('idle');
                    });


                },
                update: function (data) {
                    if (data.punchState === 1 && data.sprite.currentFrame >= data.attackFrame) {
                        data.punchable.isInvincible = false;
                        var punchObject = new Punch({
                            target: ['player', 'citizen'],
                            attack: !data.slide ? data.attack + 2 : data.attack || 1,
                            stamina: data.slide ? 1 : data.stamina || 0.334,
                            skill: data.stats.skill,
                            direction: data.multiScale.direction.x,
                            duration: data.duration,
                            pushBack: data.slide ? new Vector2(0.5, 2) : new Vector2(2.2, 1.4),
                            parent: data.entity,
                            boundingBox: new Rectangle(0, -5, 40, 10),
                            position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -40 : 0), 0)),
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx3',
                                sfx: 'sfx_punch1',
                                taptic: 'medium'
                            }),
                            debug: true
                        });
                        Bento.objects.attach(punchObject);
                        data.punchState = 2;
                    }

                    // move when sliding
                    if (data.slide && data.sprite.currentFrame >= data.attackFrame) {
                        var magnitude = data.slideTime * data.slideTime;
                        data.slideTime -= 1 / 28;
                        magnitude -= data.slideTime * data.slideTime;

                        data.movable.velocity.x = magnitude * data.slide * data.multiScale.direction.x;
                        data.movable.velocity.y = 0;
                    }
                },
                end: function (data) {
                    data.punchable.isInvincible = false;
                }
            },
            laugh: {
                ignoreBlock: true,
                ignoreDirection: true,
                canGrab: true,
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'laugh', function () {
                        data.setState('idle');
                    });
                },
            },
            jumpAttack: {
                ignoreBlock: true,
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'jump-windup', function () {
                        data.sprite.setSpriteSheet(data.rootPath + 'jump-air');
                        data.movable.velocityZ = -jumpHeight;
                    });
                    data.punchState = 1;
                    data.punchable.isInvincible = true;
                },
                update: function (data) {
                    // if (jumpTarget.x !== data.entity.position.x) {
                    //     data.multiScale.direction.x = Utils.sign(jumpTarget.x - data.entity.position.x);
                    // }

                    if (data.sprite.currentSpriteSheet === data.rootPath + 'jump-air') {
                        if (data.movable.velocityZ >= 0) {
                            data.sprite.setSpriteSheet(data.rootPath + 'jump-fall');
                        }
                    } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jump-fall') {
                        if (data.movable.grounded) {
                            var shakepower = 8;
                            if (data.punchState === 2) {
                                shakepower = 3;
                            }
                            camera.shake(20, shakepower);
                            Bento.audio.stopSound('sfx_varkastomp');
                            Bento.audio.playSound('sfx_varkastomp');

                            var punchObject = new Punch({
                                target: ['player', 'citizen'],
                                attack: data.attack + 3 || 1,
                                stamina: 1,
                                skill: data.stats.skill,
                                direction: data.multiScale.direction.x,
                                duration: 5,
                                pushBack: data.slide ? new Vector2(0.5, 2) : new Vector2(2.2, 1.4),
                                parent: data.entity,
                                boundingBox: new Rectangle(-22, -5, 44, 10),
                                position: data.entity.position.clone(),
                                onHit: Punch.getDefaultCallback({
                                    entity: data.entity,
                                    effect: 'effects/hitfx3',
                                    sfx: 'sfx_punch3',
                                    taptic: 'medium'
                                }),
                            });

                            Bento.objects.attach(punchObject);
                            data.sprite.setSpriteSheet(data.rootPath + 'jump-land', function () {
                                if (data.punchState === 2) {
                                    data.setState('idle');
                                } else {
                                    setJumpTarget();
                                    data.sprite.setSpriteSheet(data.rootPath + 'jump-air');
                                    data.movable.velocityZ = -secondJumpHeight;
                                    ++data.punchState;
                                }
                            });
                        }
                    }
                },
                move: function (data, direction) {
                    if (!data.movable.grounded) {
                        var distance = jumpTarget.distance(entity.position);
                        if (distance > 0) {
                            var velocity = jumpTarget.subtract(entity.position);
                            velocity.normalize();
                            if (distance > jumpSpeed) {
                                velocity.scalarMultiplyWith(jumpSpeed);
                            } else if (distance > 0) {
                                velocity.scalarMultiplyWith(distance);
                            }

                            data.movable.velocity = velocity;
                        } else {
                            data.movable.velocity.x = 0;
                            data.movable.velocity.y = 0;
                        }
                    }
                },
                end: function (data) {
                    data.punchable.isInvincible = false;
                    attackState = 0;
                    jumpTimer = 0;
                }
            },
            getUp: StandardEnemyStates.getUp,
            jumpBack: {
                ignoreDirection: true,
                start: StandardEnemyStates.jumpBack.start,
                update: function (data) {
                    if (data.movable.grounded) {
                        data.setState('idle');
                        data.punchable.isInvincible = false;
                        if (attackState < 2) {
                            attackState = 2;
                        }
                    }
                }
            }
        };

        var grabState = {
            grabbed: BossEnemyStates.grabbed,
            grabbedThrow: BossEnemyStates.grabbedThrow
        };

        var onRoomStart = function () {
            Utils.timeout(50, function () {
                stateMachine.setState('laugh');
            });
            EventSystem.off('startRoom', onRoomStart);
        };

        return entity;
    };
});