/**
 * punker enemy, tries to tackle the player
 * @moduleName Punker
 */
bento.define('entities/enemies/punker', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'entities/punch',
    'entities/triggers/blockade',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Punch,
    Blockade,
    Globals
) {
    'use strict';
    return function (settings) {
        var screenDimension = Bento.screens.getCurrentScreen().dimension;
        var view = Bento.getViewport();
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var punchDistance = 32;
        var tackleDistance = 80;
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 4,
            "skill": 0,
            "health": 4,
            "coins": 0,
            "stamina": 0.5,
            "minHits": 5
        }));
        var stateMachine;
        var difficulty = nameProperties.difficulty || settings.difficulty || SaveManager.load('difficulty');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'punker/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
        });
        var goTackle;
        var cooldown = 0;
        var cdAmount = 300;

        var punkAI = function (data, t, chaseBehavior) {
            if (--cooldown > 0) {
                // set attack distance to punchdistance because he cant tackle
                stateMachine.data.enemyController.attackDistance = punchDistance;
            } else {
                // set attack distance to tackle distance
                stateMachine.data.enemyController.attackDistance = tackleDistance;
            }

            if (!goTackle && cooldown <= 0) {
                if (view.hasPosition(entity.position)) {
                    var players = Bento.objects.getByFamily('player');
                    var tackle = Math.random() * 2;
                    var playerDistance;
                    if (tackle < difficulty) {
                        Utils.forEach(players, function (p, i, l, breakLoop) {
                            playerDistance = Math.abs(p.position.x - entity.position.x);
                            if (playerDistance > punchDistance + 8 && playerDistance < 80) {
                                if (Math.abs(p.position.y - entity.position.y) < 10 + (1 - playerDistance / 80) * 38) {
                                    goTackle = p;
                                    breakLoop();
                                }
                            }
                        });
                    }
                }
            }
            if (goTackle) {
                stateMachine.processAction({
                    type: 'attack',
                    data: {
                        target: goTackle
                    }
                });
            } else {
                chaseBehavior(data);
            }
        };

        var behavior = {
            name: 'punkerBehavior',
            start: function () {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);
                stateMachine.data.enemyController.ai = punkAI;
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: tackleDistance,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                behavior
            ]
        });

        var punchObject;

        var getRightSide = function () {
            if (Blockade.active) {
                return Blockade.active.position.x;
            }
            return screenDimension.width;
        };

        var states = {
            punch: {
                ignoreStun: true,
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    //turn towards player and check distance
                    var target = Bento.objects.getByName(transitionData.target)[0];
                    if (target) {
                        data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                        data.sync({
                            distance: goTackle ? punchDistance + 1 : (cooldown > 0 ? punchDistance : Math.abs(target.position.x - data.entity.position.x)),
                            target: target.name.split('-')[1]
                        });
                    } else {
                        data.sync({
                            distance: punchDistance,
                            target: '0'
                        });
                        target = Utils.getPlayerById(0);
                    }

                    // depending on distance use tackle or punch
                    var attackString;
                    if (data.syncData.distance > punchDistance) {
                        data.attackType = 1;
                        data.punchState = 0;
                        data.sprite.setSpriteSheet(data.rootPath + 'attackcharge', function () {
                            data.sprite.setSpriteSheet(data.rootPath + 'attack1');
                            // TODO: maybe able to hit him when skill level higher
                            data.punchState = 1;
                            data.walkingEnemyBehavior.noHurt = false;
                        });
                        data.distance = Math.abs(target.position.x - data.entity.position.x) + 80;
                        // cant get hurt state during tackle charge when hit by a lower skill
                        data.walkingEnemyBehavior.noHurt = true;
                    } else {
                        data.sprite.setSpriteSheet(data.rootPath + 'attack2', function () {
                            data.setState('idle');
                        });
                        data.attackType = 2;
                        data.punchState = 1;
                    }
                    data.punchable.isInvincible = false;
                },
                update: function (data) {
                    if (data.attackType === 1) {
                        if (data.punchState === 1) {
                            // tackle
                            punchObject = new Punch({
                                target: ['player', 'citizen'],
                                attack: data.attack || 1,
                                stamina: data.stamina || 0.334,
                                skill: data.stats.skill,
                                direction: data.multiScale.direction.x,
                                duration: Infinity,
                                pushBack: new Vector2(2, 1.5),
                                parent: data.entity,
                                boundingBox: new Rectangle(0, -8, 24, 16),
                                position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -24 : 0), 0)),
                                onHit: Punch.getDefaultCallback({
                                    entity: data.entity,
                                    effect: 'effects/hitfx3-medium',
                                    sfx: 'sfx_punch3',
                                    taptic: 'medium'
                                }),
                                debug: true
                            });
                            Bento.objects.attach(punchObject);

                            data.punchState = 2;
                        } else if (data.punchState > 1) {
                            // get vertical position
                            var target = Utils.getPlayerById(parseInt(data.syncData.target));
                            var vertical = 0;
                            if (target) {
                                vertical = target.position.y;
                                vertical -= entity.position.y;
                                vertical *= 0.04;
                            }

                            // move when trying to tackle
                            data.movable.velocity.x = 2 * data.multiScale.direction.x;
                            data.movable.velocity.y = vertical;
                            data.distance -= 2;
                            if (data.distance <= 0) {
                                Bento.objects.remove(punchObject);
                                punchObject = null;
                                data.setState('idle');
                            }
                        }
                    } else if (data.attackType === 2) {
                        // normal punch
                        if (data.punchState === 1 && data.sprite.currentFrame >= 3) {
                            punchObject = new Punch({
                                target: ['player', 'citizen'],
                                attack: !data.slide ? data.attack + 2 : data.attack || 1,
                                stamina: data.slide ? 1 : data.stamina || 0.334,
                                skill: data.stats.skill,
                                direction: data.multiScale.direction.x,
                                duration: data.duration,
                                pushBack: data.slide ? new Vector2(0.5, 2) : new Vector2(2.2, 1.4),
                                parent: data.entity,
                                boundingBox: new Rectangle(0, -5, 24, 10),
                                position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -24 : 0), 0)),
                                onHit: Punch.getDefaultCallback({
                                    entity: data.entity,
                                    effect: 'effects/hitfx3-small',
                                    sfx: 'sfx_punch1',
                                    taptic: 'medium'
                                }),
                                debug: true
                            });
                            Bento.objects.attach(punchObject);
                            data.punchState = 2;
                        }
                    }
                },
                end: function (data) {
                    // after tackle set cooldown, else reset cooldown
                    if (data.syncData.distance > punchDistance) {
                        cooldown = cdAmount;
                    } else {
                        cooldown = 0;
                    }

                    goTackle = null;
                    data.walkingEnemyBehavior.noHurt = false;
                    data.punchable.isInvincible = false;
                    Bento.objects.remove(punchObject);
                }
            },
        };

        return entity;
    };
});