/**
 * Raccoon enemy
 * @moduleName Raccoon
 */
bento.define('entities/enemies/raccoon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'components/ai/walkingenemy',
    'entities/punch',
    'entities/effect'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    WalkingEnemy,
    Punch,
    Effect
) {
    'use strict';
    var Raccoon = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 0,
            "skill": -1,
            "health": 0,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'raccoon/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            hurry: 0.3
        });
        // overwrite states
        var raccoonBehavior = {
            name: 'raccoonBehavior',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addStates(Raccoon.spinKickStates);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 24,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                raccoonBehavior
            ]
        });

        return entity;
    };
    Raccoon.spinKickStates = {
        // overwrite the WalkingEnemy idle state
        idle: {
            start: function (data, transitionData) {
                // attack type is randomized, sometimes raccoon will do a spinkick
                data.sync({
                    spin: Utils.getRandom(2),
                    change: Utils.getRandom(2)
                });
                var doSpinKick = data.syncData.spin;
                if (data.syncData.change > 0) {
                    // also 50% chance to change attack type
                    if (doSpinKick > 0) {
                        data.enemyController.attackDistance = 64;
                    } else {
                        data.enemyController.attackDistance = 24;
                    }
                    data.attackType = doSpinKick ? 'spinKick' : 'normal';
                }
                WalkingEnemy.states.idle.start(data, transitionData);
            },
            action: function (data, action) {
                if (data.idleTime > 0) {
                    return;
                }
                // perform
                if (data.attackType === 'spinKick') {
                    data.setState('spinKickPrepare');
                } else {
                    if (action.data) {
                        data.setState('punch', {
                            target: action.data.target.name
                        });
                    }
                }
            },
            update: WalkingEnemy.states.idle.update,
            move: WalkingEnemy.states.idle.move,
            end: WalkingEnemy.states.idle.end,
            canGrab: true
        },
        spinKickPrepare: {
            ignoreStun: true,
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'spinkick-start', function () {
                    data.setState('spinKick');
                });
            }
        },
        spinKick: {
            ignoreDirection: true,
            start: function (data) {
                // fly
                var effect = new Effect({
                    z: Globals.Layers.ENEMIES,
                    position: data.entity.position.add(new Vector2(0, -8)),
                    spriteSheet: 'effects/jump-dust'
                });
                Bento.objects.attach(effect);
                Bento.audio.stopSound('sfx_special');
                Bento.audio.playSound('sfx_special');

                data.attackState = 0;
                data.sprite.setSpriteSheet(data.rootPath + 'spinkick');
                data.movable.velocity.x = data.multiScale.direction.x * 2;
                data.movable.velocityZ = -2;
                data.movable.gravity = 0.1;
                data.cachedOnLanded = data.movable.onLanded;
                data.movable.onLanded = function () {
                    data.sprite.setSpriteSheet(data.rootPath + 'spinkick-land', function () {
                        data.setState('idle');
                    });
                    data.movable.onLanded = data.cachedOnLanded;
                };
                data.punchable.isInvincible = true;
            },
            update: function (data) {
                var punchObject;
                if (data.attackState === 0 && data.sprite.getCurrentFrame() < 3) {
                    data.attackState = 1;
                    punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: Math.ceil(data.attack / 4) || 1,
                        stamina: data.stamina || 0.334,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        pushBack: new Vector2(0.5, 1.5),
                        parent: data.entity,
                        freeze: 5,
                        boundingBox: new Rectangle(0, -5, 24, 10),
                        position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -24 : 0), 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3-medium',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        }),
                        debug: true
                    });
                    Bento.objects.attach(punchObject);
                }
                if (data.attackState === 1 && data.sprite.getCurrentFrame() >= 3) {
                    data.attackState = 0;
                    punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: Math.ceil(data.attack / 4) || 1,
                        stamina: data.stamina || 0.334,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        pushBack: new Vector2(0.5, 1.5),
                        parent: data.entity,
                        freeze: 5,
                        boundingBox: new Rectangle(0, -5, 24, 10),
                        position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -24 : 0), 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3-medium',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        }),
                        debug: true
                    });
                    Bento.objects.attach(punchObject);
                }
            },
            end: function (data) {
                data.movable.gravity = 0.14;
                data.punchable.isInvincible = false;
            }
        },
    };
    return Raccoon;
});