/**
 * Rat enemy
 * @moduleName Rat
 */
bento.define('entities/enemies/rat', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    Punch
) {
    'use strict';
    var Rat = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var rootPath = settings.rootPath || tileProperties.rootPath || 'rat/';
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 0,
            "skill": -2,
            "health": -7,
            "coins": 0 + Rat.getXmasCoins(lv, rootPath),
            "stamina": 0.2,
            "minHits": 3
        }));

        var onDeath = function () {
            EventSystem.fire('quest-rats', {
                count: 1
            });
        };

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: rootPath,
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
        });
        // overwrite states
        var ratComponent = {
            name: 'ratComponent',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');

                // use kick if it's ratgreen
                if (rootPath === 'ratgreen/') {
                    stateMachine.addState('punch', Rat.States.punch);
                }
                
                entity.getComponent('punchable', function (punchable) {
                    punchable.onDeath(onDeath);
                });
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 24,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                ratComponent
            ]
        });

        return entity;
    };
    Rat.getXmasCoins = function (lv, rootPath) {
        // was the rootPath swapped with xmas?
        var cityAssetGroups = Utils.assetGroups.city; 
        var spriteSheet = cityAssetGroups.spritesheets[rootPath + 'idle'] || '';
        var coins = 0;
        if (spriteSheet.indexOf('xmas') >= 0) {
            coins = Math.round(1 + Math.sqrt(1 + Math.floor(lv / 5) * 5));
        }
        return coins;
    };

    Rat.States = {
        // kick
        punch: {
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                data.punchAnim = (data.punchAnim || 0) + 1;
                data.sprite.setSpriteSheet(data.rootPath + 'kick', function () {
                    data.setState('idle');
                });

                data.punchState = 1;

                //turn towards player
                var target = Bento.objects.getByName(transitionData.target)[0];
                if (target) {
                    data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                }
                data.punchable.isInvincible = false;
            },
            update: function (data) {
                if (data.punchState === 1 && data.sprite.currentFrame >= 2) {
                    var punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: data.attack || 1,
                        stamina: data.stamina || 0.334,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        pushBack: new Vector2(0.5, 1.5),
                        parent: data.entity,
                        boundingBox: new Rectangle(0, -5, data.entity.punchRange || 32, 10),
                        position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -data.entity.punchRange - 4 : 4), 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3-small',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        })
                    });
                    Bento.objects.attach(punchObject);

                    data.punchState = 2;
                }
            }
        }
    };
    return Rat;
});