/**
 * Ratgirl enemy
 * @moduleName RatGirl
 */
bento.define('entities/enemies/ratgirl', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'entities/punch',
    'entities/effect',
    'modules/tapticengine',
    'components/debug',
    'managers/save',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    Punch,
    Effect,
    TapticEngine,
    Debug,
    SaveManager,
    Globals
) {
    'use strict';
    return function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 1,
            "skill": 2,
            "health": 3,
            "coins": 0,
            "stamina": 0.5,
            "minHits": 5
        }));
        var stateMachine;
        var normalRange = 24;
        var whipRange = 48;
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'ratgirl/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
        });

        var behavior = {
            name: 'ratGirlBehavior',
            start: function () {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: whipRange,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            sfxBase: 'female/',
            components: [
                walkingEnemy,
                behavior
            ]
        });

        var states = {
            punch: {
                ignoreBlock: true,
                ignoreDirection: true,
                start: function (data, transitionData) {
                    data.sync({
                        punchOrWhip: Utils.getRandom(3)
                    });
                    var target = Bento.objects.getByName(transitionData.target)[0];
                    if (target) {
                        //determine which attack to use based on distance
                        var xOffset = target.position.x - data.entity.position.x;
                        var yOffset = target.position.y - data.entity.position.y;
                        if (xOffset * xOffset + yOffset * yOffset > normalRange * normalRange) {
                            //whip attack
                            data.punchAnim = 2;
                        } else {
                            //close range, small chance to do a whip attack instead of a punch
                            if (data.punchAnim === 2) {
                                data.punchAnim = 1;
                                data.entity.punchRange = normalRange;
                            } else {
                                if (data.syncData.punchOrWhip !== 0) {
                                    data.punchAnim = 1;
                                    data.entity.punchRange = normalRange;
                                } else {
                                    data.punchAnim = 2;
                                }
                            }
                        }

                        //turn towards player
                        data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                    } else {
                        data.punchAnim = 2;
                    }

                    data.sprite.setSpriteSheet(data.rootPath + 'punch' + data.punchAnim, function () {
                        data.setState('idle');
                    });

                    data.punchState = 1;
                    data.punchable.isInvincible = false;
                },
                update: function (data) {
                    var frameCheck;
                    if (data.punchAnim === 1) {
                        frameCheck = 3;
                    } else {
                        frameCheck = 6;
                    }
                    if (data.punchState === 1 && data.sprite.currentFrame >= frameCheck) {
                        var punchObject = new Punch({
                            target: ['player', 'citizen'],
                            attack: data.attack || 1,
                            stamina: data.punchAnim ? 0.25 : 0.6,
                            skill: data.stats.skill,
                            direction: data.multiScale.direction.x,
                            pushBack: new Vector2(0.5, 1.5),
                            parent: data.entity,
                            boundingBox: new Rectangle(0, -5, data.entity.punchRange || 32, 10),
                            position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -data.entity.punchRange : 0), 0)),
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx3-medium',
                                sfx: 'sfx_punch1',
                                taptic: 'medium'
                            }),
                            debug: true
                        });
                        Bento.objects.attach(punchObject);

                        data.punchState = 2;
                    }
                },
                end: function (data) {
                    data.entity.punchRange = whipRange;
                }
            }
        };

        return entity;
    };
});