/**
 * Twerka enemy
 * @moduleName Twerka
 */
bento.define('entities/enemies/twerka', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'entities/punch'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    Punch
) {
    'use strict';
    var Twerka = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stateMachine;
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 0,
            "skill": -2,
            "health": -7,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: settings.rootPath || tileProperties.rootPath || 'twerka/',
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
        });
        // overwrite states
        var twerkaBehavior = {
            name: 'twerkaBehavior',
            start: function (data) {
                stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addStates(states);
                stateMachine.data.enemyController.ai = twerkAI;
            }
        };

        var kneeTimer = 0;
        var kneeHesitation = 20;

        // players that stand diagonally will be attack after x amount of time with a kneeattack
        var twerkAI = function (data, t, chaseBehavior) {
            if (!t) {
                return;
            }
            var yDistance = Math.abs(t.position.y - entity.position.y);
            var xDistance = Math.abs(t.position.x - entity.position.x);
            var difference = Math.abs(yDistance - xDistance);
            var maxDistance = 100;
            var minDistance = 16;
            if (difference < 16 &&
                xDistance < maxDistance &&
                yDistance > minDistance &&
                !t.isDown
            ) {
                --kneeTimer;
                if (kneeTimer <= 0) {
                    // stateMachine.data.setState('kneeAttack');
                    stateMachine.processAction({
                        type: 'kneeAttack',
                        data: {
                            target: t
                        }
                    });
                    kneeTimer = kneeHesitation;
                } else if (stateMachine.stateName != 'kneeAttack') {
                    stateMachine.processAction('idle');
                } else {
                    kneeTimer = kneeHesitation;
                }
            } else {
                kneeTimer = kneeHesitation;
                chaseBehavior(data);
            }
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 24,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            sfxBase: 'female/',
            components: [
                walkingEnemy,
                twerkaBehavior
            ]
        });

        var states = {
            idle: {
                start: StandardEnemyStates.idle.start,
                move: StandardEnemyStates.idle.move,
                action: function (data, action) {
                    if (action.type === 'kneeAttack') {
                        data.setState('kneeAttack', {
                            target: action.data.target.name
                        });
                    } else if (action.type) {
                        StandardEnemyStates.idle.action(data, action);
                    }
                },
                canGrab: true
            },
            kneeAttack: {
                ignoreDirection: true,
                start: function (data, transitionData) {
                    var target = Bento.objects.getByName(transitionData.target)[0];
                    if (target) {
                        data.multiScale.direction.x = Utils.sign(target.position.x - entity.position.x);
                        data.sync({
                            kneeDirection: Utils.sign(target.position.y - entity.position.y)
                        });
                    }
                    data.sprite.setSpriteSheet(data.rootPath + 'kneeattack', function () {
                        data.setState('idle');
                    });
                    data.attackState = 1;
                },
                update: function (data) {
                    if (data.sprite.currentFrame > 11) {

                        Bento.audio.stopSound('sfx_twerka-knee');
                        Bento.audio.playSound('sfx_twerka-knee');
                        if (data.attackState === 1) {
                            var punchObject = new Punch({
                                target: ['player', 'citizen'],
                                attack: data.attack || 1,
                                stamina: data.stamina || 0.334,
                                skill: data.stats.skill,
                                direction: data.multiScale.direction.x,
                                duration: 30,
                                pushBack: new Vector2(2.2, 1.4),
                                parent: data.entity,
                                boundingBox: new Rectangle(0, -5, 24, 10),
                                position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -24 : 0), 0)),
                                onHit: Punch.getDefaultCallback({
                                    entity: data.entity,
                                    effect: 'effects/hitfx3-medium',
                                    sfx: 'sfx_punch1',
                                    taptic: 'medium'
                                }),
                                debug: true
                            });
                            Bento.objects.attach(punchObject);

                            data.slide = 3.5;
                            data.attackState = 2;
                            data.movable.velocityZ -= 1;
                        }
                        data.movable.velocity.x = 0.7 * data.multiScale.direction.x * data.slide;
                        data.movable.velocity.y = 0.7 * data.syncData.kneeDirection * data.slide;
                        data.slide *= 0.97;
                    }
                }
            }
        };

        return entity;
    };
    return Twerka;
});