/**
 * Weasel enemy
 * @moduleName Weasel
 */
bento.define('entities/enemies/weasel', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/enemycontroller',
    'components/ai/walkingenemy',
    'modules/matchmanager',
    'components/online',
    'components/healthbar',
    'components/punchable',
    'components/coindropper',
    'components/damagecounter',
    'entities/enemy',
    'components/stats',
    'states/standardenemy',
    'managers/save',
    'globals',
    'entities/obstacles/tnt'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    EnemyController,
    WalkingEnemyBehavior,
    MatchManager,
    Online,
    HealthBar,
    Punchable,
    CoinDropper,
    DamageCounter,
    Enemy,
    Stats,
    StandardEnemyStates,
    SaveManager,
    Globals,
    Tnt
) {
    'use strict';
    var Weasel = function (settings) {
        var tiled = (settings.tiled || {});
        var objectProperties = tiled.objectProperties || {};
        var object = tiled.object || {};
        var tileProperties = tiled.tileProperties || {};
        var nameProperties = Utils.nameToJson(object.name);
        var hasTnt = Utils.isDefined(nameProperties['tnt']) ? nameProperties['tnt'] : 1; // TODO: add new weasel tile
        var level = nameProperties.lv || settings.lv || SaveManager.load('enemyLevel') + (nameProperties.addLv || 0) || 1;
        var lv = level - 1; // start with 0 for calculation
        var stats = new Stats(Globals.getDefaultEnemyStats({
            "level": level,
            "attack": 0,
            "skill": -1,
            "health": 0,
            "coins": 0,
            "stamina": 0.2,
            "minHits": 5
        }));

        var tutorial = SaveManager.load('tutorialWorldSelect');
        // main behavior
        var rootPath = settings.rootPath || tileProperties.rootPath || 'weasel/';
        var walkingEnemy = new WalkingEnemyBehavior({
            rootPath: rootPath,
            attack: stats.attack,
            difficulty: nameProperties.difficulty || settings.difficulty,
            hurry: 0.2
        });
        var enemyController;
        var stateMachine;
        // overwrite states
        var weaselBehavior = {
            name: 'weaselBehavior',
            tntThrowTime: 60,
            start: function (data) {
                entity.getComponent('stateMachine', function (component) {
                    stateMachine = component;
                });
                entity.getComponent('enemyController', function (component) {
                    enemyController = component;
                });

                // overwrite with tnt states
                stateMachine.addStates(Weasel.tntStates);


                if (hasTnt) {
                    if (!Utils.hasAsset('spritesheets', 'obstacles/tnt/tnt')) {
                        Utils.log('WARNING: tnt was not loaded while trying to give weasel some');
                        return;
                    }
                    if (!Utils.hasAsset('spritesheets', rootPath + 'idle')) {
                        Utils.log('WARNING: weasel spriteSheet was not loaded yet');
                        entity.canGetTnt = true;
                        return;
                    }
                    // give tnt
                    Weasel.addTnt({
                        entity: entity,
                        stateMachine: stateMachine
                    });
                    entity.canGetTnt = true;
                }
            },
            update: function () {}
        };

        var entity = new Enemy({
            tiled: settings.tiled,
            position: settings.position,
            stats: stats,
            punchRange: 24,
            boundingBox: new Rectangle(-16, -6, 32, 12),
            maxHealth: stats.health,
            health: objectProperties.health,
            healthbarSpacing: 52,
            components: [
                walkingEnemy,
                weaselBehavior
            ]
        });

        return entity;
    };
    // add tnt to weasel
    Weasel.addTnt = function (data) {
        // increase punch range
        data.entity.getComponent('enemyController', function (enemyController) {
            enemyController.attackDistance = 64;
        });

        data.entity.setSpriteSheet(data.stateMachine.data.rootPath + 'idle-tnt');

        data.stateMachine.data.hasTnt = true;
    };
    // remove tnt from weasel
    Weasel.removeTnt = function (data) {
        // data.stateMachine.addStates(WalkingEnemyBehavior.states);
        // data.stateMachine.addState('hurt', StandardEnemyStates.hurt);
        data.enemyController.attackDistance = 24;
        data.stateMachine.data.hasTnt = false;
        data.timeSinceTnt = 0;
    };


    Weasel.tntStates = {
        idle: {
            start: function (data, transitionData) {
                var walkingEnemy = data.walkingEnemyBehavior;
                var tnt = data.hasTnt ? '-tnt' : '';
                data.sprite.setSpriteSheet(data.rootPath + 'idle' + tnt);

                data.idleTime = Utils.getRandomRange(walkingEnemy.hesitation[0], walkingEnemy.hesitation[1]);

                // random chance the weasel will grab a new tnt while standing still
                var grabTnt = (Utils.getRandom(2) === 0);
                data.timeSinceTnt = data.timeSinceTnt || 0; // needs at least 1 second inbetween
                if (grabTnt && !data.isRemote && data.entity.canGetTnt && !data.hasTnt && data.timeSinceTnt > 60) {
                    data.setState('getNewTnt');
                }
            },
            update: function (data) {
                data.idleTime--;

                if (!data.hasTnt) {
                    data.timeSinceTnt += 1;
                }
            },
            move: function (data, direction) {
                if (data.idleTime > 0) {
                    return;
                }
                data.setState('walk');
            },
            action: function (data, action) {
                if (data.idleTime > 0) {
                    return;
                }
                // throw tnt
                if (data.hasTnt) {
                    data.setState('throwTnt');
                } else {
                    if (action.data) {
                        data.setState('punch', {
                            target: action.data.target.name
                        });
                    }
                }
            },
            end: function (data) {

            },
            canGrab: true
        },
        walk: {
            start: function (data, transitionData) {
                var tnt = data.hasTnt ? '-tnt' : '';
                data.sprite.setSpriteSheet(data.rootPath + 'run' + tnt);
                data.walkTime = 60 + Math.random() * 30;
                data.velocityX = 0;
            },
            update: function (data) {
                var tnt = data.hasTnt ? '-tnt' : '';
                // add a small percentage of the new velocity so the animation will only switch when the enemy is actually walking in a different direction
                data.velocityX *= 0.8;
                data.velocityX += data.movable.velocity.x * 0.2;

                data.walkTime--;
                if (data.walkTime <= 0 && data.behavior === 'idle' && !data.isRemote) {
                    data.setState('idle');
                } else if (Utils.sign(data.velocityX) === data.multiScale.direction.x) {
                    data.sprite.setSpriteSheet(data.rootPath + 'run' + tnt);
                } else {
                    if (data.hasTnt) {
                        data.sprite.setSpriteSheet(data.rootPath + 'run' + tnt);
                    } else {
                        data.sprite.setSpriteSheet(data.rootPath + 'walk');
                    }
                }
            },
            move: function (data, direction) {
                var walkingEnemy = data.walkingEnemyBehavior;
                data.movable.velocity = direction.scalarMultiply(1.5 + walkingEnemy.hurry);
            },
            action: function (data, action) {
                //ai controller wants to punch, cancel walk and switch to idle
                data.setState('idle');
            },
            canGrab: true
        },
        getNewTnt: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'get-tnt', function () {
                    Weasel.addTnt(data);
                    data.setState('idle');
                });
            }
        },
        hurt: {
            ignoreSync: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                // drop tnt
                if (data.hasTnt) {
                    var tnt = new Tnt({
                        position: data.entity.position.clone(),
                        direction: data.multiScale.direction.x,
                        velocity: new Vector2(data.multiScale.direction.x * 0.5, Utils.getRandomRangeFloat(-0.1, 0.1))
                    });
                    Bento.objects.attach(tnt);
                }

                StandardEnemyStates.hurt.start(data, transitionData);
                Weasel.removeTnt(data);
            },
            update: StandardEnemyStates.hurt.update,
            end: StandardEnemyStates.hurt.end
        },
        throwTnt: {
            canStun: false,
            canBlock: false,
            canFaint: false,
            ignoreDirection: true,
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'throw-tnt', function () {
                    data.setState('idle');
                });
                // spawn tnt object
                var tnt = new Tnt({
                    position: data.entity.position.clone(),
                    direction: data.multiScale.direction.x,
                    velocity: new Vector2(data.multiScale.direction.x * 1.5, Utils.getRandomRangeFloat(-0.5, 0.5))
                });
                Bento.objects.attach(tnt);
                Weasel.removeTnt(data);
            }
        },
    };
    return Weasel;
});