/**
 * Interstitial
 * @moduleName StaticInterstitial
 */
bento.define('entities/interstitial', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/components/fill',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'bento/packedimage',
    'modules/localization',
    'modules/firebase',
    'modules/analytics',
    'components/anchor',
    'components/modal',
    'gui/shadowtext',
    'managers/time'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Fill,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    PackedImage,
    Localization,
    Firebase,
    Analytics,
    Anchor,
    Modal,
    ShadowText,
    TimeManager
) {
    'use strict';
    var interstitialData = {
        //Javascript Month is zero indexed! WTF! (day/year aren't)
        'radbeach': {
            isActive: true,
            imageName: 'interstitial/radbeach',
            textTop: Localization.getText('radEvent').toUpperCase(),
            eventStart: new Date(2017, 12 - 1, 21),
            eventEnd: new Date(2017, 12 - 1, 26),
            textBottomOverride: null
        }
    };

    var width = Math.min(Bento.getViewport().width - 20, 160); // desired width of interstitial image
    var isShowing = false;

    var Interstitial = function (settings) {
        /*settings = {
            z: Number
            pauseLevel: Number
        }*/
        var viewport = Bento.getViewport();
        var oldPauseLevel = Bento.objects.isPaused();
        var pauseLevel = settings.pauseLevel || 1;
        var z = settings.z || 1;

        var fill = {
            name: 'fillScreen',
            draw: function (data) {
                var origin = new Vector2(0, 0);
                data.renderer.fillRect([0, 0, 0, 0.75], -1, -1, viewport.width + 2, viewport.height + 2);
            }
        };
        var buttonClose = new ClickButton({
            name: 'closeInterstitial',
            imageName: 'crosspromo/buttonclose',
            frameCountX: 1,
            frameCountY: 2,
            position: new Vector2(viewport.width / 2 - width / 2 + 8, viewport.height / 2 - width / 2 - 10),
            updateWhenPaused: pauseLevel,
            onClick: function () {
                // this button does nothing actually
                // clicking outside the interstitial will dismiss it
            }
        });

        var timeNow = TimeManager.getTime();
        var availableInterstitials = [];
        for (var index in interstitialData) {
            if (interstitialData.hasOwnProperty(index)) {
                var currentInterstitial = interstitialData[index];
                var hasEndDate = currentInterstitial.hasOwnProperty('eventEnd') && typeof currentInterstitial.eventEnd == "object" && currentInterstitial.eventEnd !== null;
                var dateHasPassed = currentInterstitial.eventEnd.getTime() < timeNow;
                if (((hasEndDate && !dateHasPassed) || (!hasEndDate)) && currentInterstitial.isActive) {
                    availableInterstitials.push(currentInterstitial);
                }
            }
        }
        if (!availableInterstitials.length) {
            //no interstitials available
            return;
        }
        var comingSoonText;
        var useInterstitial = availableInterstitials[Math.floor(Math.random() * availableInterstitials.length)];
        var topText = useInterstitial.textTop;
        var topTextOffset = 0;
        var bottomText = "";
        if (useInterstitial.hasOwnProperty("textBottomOverride") && useInterstitial.textBottomOverride !== null) {
            bottomText = useInterstitial.textBottomOverride;
        } else if (useInterstitial.hasOwnProperty('eventStart') && typeof useInterstitial.eventStart == "object") {
            if (useInterstitial.eventStart.getTime() > timeNow) {
                var lang = (typeof window.navigator.languages == "object" && window.navigator.languages.length) ? window.navigator.languages[0] : "";
                bottomText = useInterstitial.eventStart.toLocaleDateString(lang || window.navigator.userLanguage || window.navigator.language || 'en-US');

                comingSoonText = new Text({
                    text: Localization.getText('comingsoon'),
                    font: 'font',
                    fontSize: 12,
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: false,
                    fontColor: '#ffffff',
                    maxWidth: 150,
                    maxHeight: 12,
                    lineWidth: 1.5,
                    strokeStyle: '#45316a',
                    innerStroke: false,
                    pixelStroke: true,
                    components: [
                        new Anchor({
                            screenPosition: new Vector2(0, 0),
                            offset: new Vector2(0, -65 - 10),
                            onComplete: function (parent) {}
                        })
                    ]
                });
                topTextOffset = 8;
            } else {
                bottomText = "LIVE NOW";
            }
        }

        var sprite = new Sprite({});
        sprite.setup({
            imageName: useInterstitial.imageName
        });
        sprite.setOriginRelative(new Vector2(0.5, 0.5));

        var isTweening = false;
        /**
         * Interstitial image entity
         */
        var resizeInterstitial = function () {
            var scale;
            scale = width / interstitialContainer.dimension.width;
            interstitialContainer.scale.x = scale;
            interstitialContainer.scale.y = scale;
            scale = (width - 4) / interstitialEntity.dimension.width / scale;
            interstitialEntity.scale.x = scale;
            interstitialEntity.scale.y = scale;
        };
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                // resize interstitial image
                resizeInterstitial();

                // drop down
                isShowing = true;
                isTweening = true;

                new Tween({
                    from: interstitialContainer.position.y,
                    to: viewport.height / 2,
                    'in': 40,
                    ease: 'elastic',
                    alpha: -8,
                    beta: 1,
                    updateWhenPaused: pauseLevel,
                    onUpdate: function (v, t) {
                        interstitialContainer.position.y = v;
                    },
                    onComplete: function () {
                        isTweening = false;
                        entity.attach(buttonClose);
                    }
                });

                if (Analytics) {
                    Analytics.trackView('interstitial');
                    Analytics.event('interstitial', 'viewed');
                }
                if (Firebase) {
                    Firebase.logEvent('interstitial', {
                        type: 'shown'
                    });
                }
            }
        };
        var clickable = new Clickable({
            onClick: function (data) {
                closeInterstitial();
            },
            onClickMiss: function (data) {
                closeInterstitial();
            }
        });
        var closeInterstitial = function () {
            if (isTweening) {
                return;
            }

            if (settings.sfx) {
                Bento.audio.playSound(settings.sfx);
            }

            close();
        };
        var interstitialEntity = new Entity({
            name: 'interstitialImage',
            position: new Vector2(0, 0),
            updateWhenPaused: settings.pauseLevel,
            components: [
                sprite,
                behavior,
                clickable
            ]
        });
        /**
         * Extra interstitial container to show edge
         */
        var nineSlice = new Sprite({
            imageName: 'crosspromo/edge',
            originRelative: new Vector2(0.5, 0.5)
        });
        var interstitialContainer = new Entity({
            name: 'interstitialContainer',
            position: new Vector2(viewport.width / 2, -width),
            components: [
                nineSlice,
                interstitialEntity,
                new ShadowText({
                    text: topText,
                    font: 'font',
                    fontSize: 48,
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: false,
                    fontColor: '#ffffff',
                    maxWidth: 150,
                    maxHeight: 25,
                    lineWidth: 1.5,
                    strokeStyle: '#45316a',
                    innerStroke: false,
                    pixelStroke: true,
                    components: [
                        new Anchor({
                            screenPosition: new Vector2(0, 0),
                            offset: new Vector2(0, -65 + topTextOffset),
                            onComplete: function (parent) {}
                        })
                    ]
                }),
                new ShadowText({
                    text: bottomText,
                    font: 'font',
                    fontSize: 48,
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: false,
                    fontColor: '#ffffff',
                    maxWidth: 150,
                    maxHeight: 25,
                    lineWidth: 1.5,
                    strokeStyle: '#45316a',
                    innerStroke: false,
                    pixelStroke: true,
                    components: [
                        new Anchor({
                            screenPosition: new Vector2(0, 0),
                            offset: new Vector2(0, 60),
                            onComplete: function (parent) {}
                        })
                    ]
                }),
                new Anchor({
                    screenPosition: new Vector2(0.5, 0.5),
                    onComplete: function () {
                        buttonClose.position.x = viewport.width / 2 - width / 2 + 8;
                        buttonClose.position.y = viewport.height / 2 - width / 2 - 10;
                        if (!Anchor.isPortrait()) {
                            buttonClose.position.x -= 32;
                            buttonClose.position.y += 32;
                        }
                    }
                })
            ]
        });
        if (comingSoonText) {
            interstitialContainer.attach(comingSoonText);
        }
        /**
         * Container
         */
        var entity = new Entity({
            z: z,
            name: 'interstitial',
            position: new Vector2(0, 0),
            updateWhenPaused: settings.pauseLevel,
            float: true,
            components: [
                new Modal({}),
                fill,
                interstitialContainer
            ]
        });
        var close = function () {
            // tween out and remove
            entity.remove(buttonClose);
            isTweening = true;
            new Tween({
                from: interstitialContainer.position.y,
                to: -width,
                'in': 25,
                ease: 'easeInBack',
                updateWhenPaused: pauseLevel,
                onUpdate: function (v, t) {
                    interstitialContainer.position.y = v;
                },
                onComplete: function () {
                    Bento.objects.remove(entity);
                    // Bento.objects.pause(oldPauseLevel);
                    // not using pause for compatibility with old games
                    // Bento.objects.resume();
                    isShowing = false;
                    // coming back from cross promo
                    Analytics.trackView('title');
                }
            });
        };

        // Bento.objects.pause(pauseLevel);

        // adds self
        Bento.objects.attach(entity);

        return entity;
    };

    Interstitial.isShowing = function () {
        return isShowing;
    };

    return Interstitial;
});