/**
 * Citizen for protection missions
 * @moduleName Citizen
 */
bento.define('entities/npc/citizen', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/healthbar',
    'components/punchable',
    'components/damagecounter',
    'components/online',
    'entities/character',
    'modules/matchmanager',
    'managers/enemyactivator',
    'gui/menus/gameover',
    'managers/ai',
    'modules/analytics',
    'globals',
    'managers/save'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    HealthBar,
    Punchable,
    DamageCounter,
    Online,
    Character,
    MatchManager,
    EnemyActivator,
    GameOverMenu,
    AiManager,
    Analytics,
    Globals,
    SaveManager
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var citizenView = new Rectangle(-90, -320, 180, 640);
        var extraComponents = settings.components || [];
        var tiled = settings.tiled || {};
        var object = tiled.object || {};
        var nameProperties = Utils.nameToJson(object.name);
        var objectProperties = tiled.objectProperties || {};
        var tileProperties = tiled.tileProperties || {};
        var rootPath = objectProperties.rootPath || tileProperties.rootPath;
        var stateMachine;
        var multiScale;
        var isMaster;
        var invincible = 0;
        var health = nameProperties.health || SaveManager.load('citizenHealth');
        var maxHealth = nameProperties.health || SaveManager.load('citizenMaxHealth');

        var punchable = new Punchable({
            boundingBox: settings.boundingBox || new Rectangle(-16, -6, 32, 12),
            health: health - 1,
            minHits: 1,
            maxHealth: maxHealth - 1,
            syncDeath: true
        });

        // healthbar
        var healthBar = new HealthBar({
            punchable: punchable,
            spacing: settings.healthbarSpacing || 0
        });

        var onlineSync = new Online({});

        var damageCounter = new DamageCounter({});

        var onBlockadeRemoval = function () {
            stateMachine.setState('walk');
        };

        var attachGameOverMenu = function () {
            var gameOver = new GameOverMenu({
                reviveCost: Math.ceil(Globals.gemToRevive(SaveManager.load('currentMission')) * 0.5)
            });
            // set player sprite in the menu to the citizen and remove the energy text
            gameOver.getComponent('faintedPlayer', function (gameOverCitizen) {
                gameOverCitizen.getComponent('sprite').setSpriteSheet(rootPath + 'hurt');
                gameOverCitizen.getComponent('shadowText').visible = false;
                gameOverCitizen.getComponent('icon').visible = false;
            });

            var localPlayer;
            Utils.forEach(Bento.objects.getByFamily('player'), function (player, i, l, breakLoop) {
                if (!player.isRemote) {
                    // make local player dead so that he doesnt get attacked during multiplayer
                    localPlayer = player;
                    breakLoop();
                }
            });

            // change revive behavior
            gameOver.getComponent('behaviorComponent', function (gameOverBehavior) {
                gameOverBehavior.revive = function () {
                    Analytics.event('Economy', 'spendGemsForEscortMission', 1, 'yes');

                    // revive citizen and send message
                    behavior.reviveCitizen();
                    MatchManager.offMessage(entity.name, behavior.onMessage);
                    MatchManager.sendData({
                        type: 'revive'
                    }, entity.name);

                    localPlayer.getComponent('punchable', function (playerPunchable) {
                        playerPunchable.isDead = false;
                        playerPunchable.hasFainted = false;
                    });

                    localPlayer.getComponent('stateMachine', function (playerStateMachine) {
                        playerStateMachine.setState('idle');
                    });
                    EventSystem.fire('enableControls');
                    AiManager.distributeAiTasks();
                    Bento.audio.stopSound('sfx_revive');
                    Bento.audio.playSound('sfx_revive');

                    SaveManager.save('citizenFail', false);
                };
            });

            Bento.objects.attach(gameOver);
        };

        var behavior = {
            name: 'citizenBehavior',
            onHit: function (data) {
                punchable.depleteHealth({
                    name: 'punch',
                    attack: 1,
                    multiplier: 1,
                });
                SaveManager.save('citizenHealth', punchable.getHealth());
                stateMachine.setState('hurt', {}, true);
            },
            reviveCitizen: function () {
                // reset citizen health and set invincible
                if (punchable.isDead) {
                    punchable.resetHealth();
                    SaveManager.save('citizenHealth', punchable.getHealth());
                    punchable.setInvincibleTime(60);
                    stateMachine.setState('walk', {}, true);
                    invincible = 60;
                }
            },
            onMessage: function (msg) {
                // revive for now
                if (msg.type === 'revive') {
                    behavior.reviveCitizen();
                    MatchManager.offMessage(entity.name, behavior.onMessage);
                }
                // always return true if the message is handled
                return true;
            },
            onDeath: function (data) {
                // start listening for multiplayer msgs
                MatchManager.onMessage(entity.name, behavior.onMessage);
                stateMachine.setState('death');
                Utils.forEach(Bento.objects.getByFamily('player'), function (player, i, l, breakLoop) {
                    if (!player.isRemote) {
                        // make local player dead so that he doesnt get attacked during multiplayer
                        player.getComponent('punchable', function (playerPunchable) {
                            playerPunchable.isDead = true;
                            playerPunchable.hasFainted = true;
                        });
                        // set state to escortFailed, this will visually look like the idle state but mimic the dead state for trigger purposes
                        player.getComponent('stateMachine', function (playerStateMachine) {
                            playerStateMachine.setState('escortFailed');
                        });
                    }
                });
                EventSystem.fire('disableControls');

                // save failure
                SaveManager.save('citizenFail', true);

                // stop if one already exists
                if (Bento.objects.get('gameOverMenu')) {
                    return;
                }

                // show custom gameover menu after 60 seconds
                Utils.timeout(60, function () {
                    attachGameOverMenu();
                });
            },
            start: function (data) {
                stateMachine = entity.getComponent('stateMachine');
                stateMachine.addStates(states);
                stateMachine.data.rootPath = rootPath;

                stateMachine.data.behavior = 'walk';
                stateMachine.setState('walk');

                // add scale
                multiScale = entity.getComponent('multiScale');
                multiScale.addScale('hurt', new Vector2(1, 1));

                // add event when blockade is being destroyed
                EventSystem.on('removeBlockade', onBlockadeRemoval);

                // setup punchable
                punchable.onHit(behavior.onHit);
                punchable.onDeath(behavior.onDeath);

                if (entity.family.indexOf('citizen') < 0) {
                    entity.family.push('citizen');
                }
                entity.name = 'citizen';

                isMaster = MatchManager.isMaster;
                if (isMaster) {
                    onlineSync.setRemote(false);
                } else {
                    onlineSync.setRemote(true);
                }

                if (!nameProperties.health) {
                    // check for faliure with a timeout so that everything is loaded
                    Utils.timeout(60, function () {
                        if (SaveManager.load('citizenFail')) {
                            Utils.forEach(Bento.objects.getByFamily('player'), function (player, i, l, breakLoop) {
                                if (!player.isRemote) {
                                    // make local player dead so that he doesnt get attacked during multiplayer
                                    player.getComponent('punchable', function (playerPunchable) {
                                        playerPunchable.isDead = true;
                                        playerPunchable.hasFainted = true;
                                    });
                                    // set state to escortFailed, this will visually look like the idle state but mimic the dead state for trigger purposes
                                    player.getComponent('stateMachine', function (playerStateMachine) {
                                        playerStateMachine.setState('escortFailed');
                                    });
                                }
                            });
                            EventSystem.fire('disableControls');
                            attachGameOverMenu();
                        }
                    });
                } else {
                	console.log(health);
                	SaveManager.save('citizenHealth', nameProperties.health);
                }
            },
            destroy: function (data) {
                EventSystem.off('removeBlockade', onBlockadeRemoval);
                MatchManager.offMessage(entity.name, behavior.onMessage);
            },
            update: function (data) {
                // we always check if local player is master
                isMaster = MatchManager.isMaster;
                if (isMaster) {
                    if (invincible > 0) {
                        --invincible;
                    }
                    if (stateMachine.stateName === 'walk' && invincible === 0) {
                        // when walking check if enemies are within view of citizen
                        var viewBox = citizenView.offset(entity.position);
                        Utils.forEach(Bento.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                            if (viewBox.hasPosition(enemy.position)) {
                                EnemyActivator.activateEnemies();
                                stateMachine.setState('shock', {}, true);
                                breakLoop();
                            }
                        });
                    }
                }
                    console.log(SaveManager.load('citizenHealth'));

            },
            draw: function (data) {},
            attached: function (data) {
                entity = data.entity.family.push('citizen');
            },
        };

        var states = {
            idle: {
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                },
                update: function (data) {

                    if (data.movable.grounded) {
                        data.movable.velocity.x = Utils.approach(data.movable.velocity.x, 0, 0.2);
                        data.movable.velocity.y = Utils.approach(data.movable.velocity.y, 0, 0.2);
                    }
                }
            },
            walk: {
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'walk');
                    data.multiScale.direction.x = 1;
                },
                update: function (data) {
                    data.movable.velocity.x = 0.8;
                    data.multiScale.direction.x = 1;

                }
            },
            shock: {
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'shocked');
                    data.movable.velocityZ = -2;
                    data.movable.velocity.x *= -0.5;
                },
                update: function (data) {
                    if (data.movable.grounded) {
                        stateMachine.setState('idle');
                    }
                }
            },
            hurt: {
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'hurt');
                    data.hurt = 30;

                    new Tween({
                        from: 0.5,
                        to: 0,
                        'in': 30,
                        ease: Tween.ELASTIC,
                        alpha: -5,
                        beta: 2,
                        onUpdate: function (v, t) {
                            data.multiScale.hurt.x = 1 + v;
                            data.multiScale.hurt.y = 1 - v;
                        }
                    });
                },
                update: function (data) {
                    data.hurt--;
                    data.multiScale.direction.x = data.punchDirection;
                    if (data.hurt <= 0) {
                        data.setState('idle');
                    }
                },
                end: function (data) {

                }
            },
            death: {
                start: function (data, transitionData) {
                    data.sprite.setSpriteSheet(data.rootPath + 'shocked');
                    data.movable.grounded = false;
                    data.movable.velocityZ = -2;
                    data.punchable.setHealth(0);
                    SaveManager.save('citizenHealth', 0);
                },
                update: function (data) {
                    if (data.movable.grounded) {
                        data.setState('flee');
                    }
                },
            },
            flee: {
                ignoreSync: true,
                start: function (data, transitionData) {
                    multiScale.direction.x = -1;
                    data.sprite.setSpriteSheet(data.rootPath + 'flee');

                },
                update: function (data) {
                    data.movable.velocity.x = Utils.approach(data.movable.velocity.x, -3.8, 0.2);
                }
            }
        };

        var entity = new Character({
            spriteSheet: settings.spriteSheet,
            position: settings.position,
            components: [
                behavior,
                punchable,
                healthBar,
                onlineSync,
                damageCounter,
            ].concat(extraComponents)
        });

        return entity;
    };
});