/**
 * Spawns barrels at the right side of the screen
 * Syncs online
 * @moduleName BarrelSpawner
 */
bento.define('entities/obstacles/barrelspawner', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/obstacles/barrel',
    'modules/matchmanager',
    'entities/triggers/blockade'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Barrel,
    MatchManager,
    Blockade
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var id = settings.tiled.object.id;
        var name = settings.tiled.object.name;
        var properties = Utils.nameToJson(name);
        var timeUntilNextBarrel = properties.time || 60 * 4;
        var startTime = properties.delay ? -parseInt(properties.delay) : 0;
        var start = false;
        var currentScreen = Bento.screens.getCurrentScreen();

        var isRemote;
        var isListening = false;
        var inactive = true;

        var done = false;

        var amount = properties.amount || 1;
        // var direction = -1;
        // var offScreen = 256;
        var behavior = {
            name: 'behaviorComponent',
            timer: 0,
            start: function (data) {
                EventSystem.on('startRoom', behavior.onStartRoom);
                behavior.timer = timeUntilNextBarrel + startTime;
                if (!isListening) {
                    behavior.listenToIncoming();
                }
            },
            destroy: function (data) {
                behavior.stopListening();
                EventSystem.off('startRoom', behavior.onStartRoom);
            },
            update: function (data) {
                if (done) {
                    return;
                }
                if (inactive) {
                    // check to see if in screen
                    if (viewport.hasPosition(entity.position)) {
                        inactive = false;
                        MatchManager.sendData({}, entity.name);
                    }

                } else {
                    // check if local player is master
                    isRemote = !MatchManager.isMaster;


                    if (!start) {
                        // wait for the event "startRoom"
                        return;
                    }

                    if (!isRemote) {
                        // local: behavior will spawn barrel
                        behavior.timer += data.speed;
                        if (behavior.timer > timeUntilNextBarrel) {
                            // spawn barrel and let remotes know
                            behavior.spawnBarrel();
                            behavior.timer = 0;

                            var msg = {};
                            MatchManager.sendData(msg, entity.name);
                        }
                    }

                    // isRemote changed value
                    if (isRemote && !isListening) {
                        behavior.listenToIncoming();
                    } else if (!isRemote && isListening) {
                        behavior.stopListening();
                    }
                }
            },
            spawnBarrel: function () {
                // spawn a barrel
                var spawnX;
                if (Blockade.active) {
                    spawnX = Blockade.active.position.x;
                } else {
                    spawnX = currentScreen.width;
                }
                var barrel = new Barrel({
                    id: id + '-' + amount,
                    position: new Vector2(spawnX + 128, entity.position.y + 8),
                    isRemote: isRemote
                });
                Bento.objects.attach(barrel);
                --amount;
                if (amount <= 0) {
                    done = true;
                    Utils.timeout(100, function () {
                        behavior.removeSelf();
                    });
                }
            },
            // listen to incoming messages
            listenToIncoming: function () {
                MatchManager.onMessage(entity.name, behavior.onMessage);
                isListening = true;
            },
            stopListening: function () {
                MatchManager.offMessage(entity.name, behavior.onMessage);
                isListening = false;
            },
            onMessage: function (msg) {
                if (inactive) {
                    inactive = false;
                } else {
                    behavior.spawnBarrel();
                }

                // always return true if the message is handled
                return true;
            },
            onStartRoom: function () {
                start = true;
            },
            removeSelf: function () {
                //remove this
                entity.removeSelf();
                EventSystem.fire('destroyEnemyController');
            }
        };

        var entity = new Entity({
            z: 0,
            name: 'barrelSpawner-' + id,
            family: ['obstacles', 'enemy', 'ignoreEnemy'],
            position: settings.position,
            components: [
                behavior
            ]
        });
        return entity;
    };
});


//.getEnemiesLeft()