/**
 * Car obstacle
 * @moduleName Car
 */
bento.define('entities/obstacles/car', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/sort',
    'entities/punch',
    'globals',
    'components/drawinviewport',
    'managers/buff',
    'components/punchsyncer',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Sort,
    Punch,
    Globals,
    DrawInViewport,
    BuffManager,
    PunchSyncer,
    MatchManager
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var currentScreen = Bento.screens.getCurrentScreen();
        var offScreen = settings.offScreen || 256;
        var isRemote = settings.isRemote || false;
        var isListening = false;
        var sprite = new Sprite({
            spriteSheet: 'obstacles/car/car'
        });
        var shadowSprite = new Sprite({
            spriteSheet: 'obstacles/car/car-shadow'
        });
        var sort = new Sort({});

        var attack = 0.1;

        var behavior = {
            name: 'behaviorComponent',
            speed: settings.speed || -2,
            punch: null,
            notHonked: true,
            honkDistance: 64,
            start: function (data) {
                var defaultCallback = Punch.getDefaultCallback({
                    entity: entity,
                    effect: 'effects/hitfx3-medium',
                    sfx: 'sfx_punch3',
                    taptic: 'medium'
                });
                behavior.punch = new Punch({
                    target: isRemote ? ['player'] : ['player', 'enemy'], // can hit both enemies and player
                    attack: 2,
                    multiplier: BuffManager.current.enviromentDamage,
                    stamina: 1,
                    direction: Utils.sign(behavior.speed),
                    pushBack: new Vector2(0.1, 3),
                    duration: Infinity,
                    // parent: entity,
                    boundingBox: new Rectangle(-44, -11, 88, 22),
                    position: entity.position.clone(),
                    onHit: function (other) {
                        defaultCallback(other);
                        if (other.family.indexOf('player') === -1) {
                            behavior.carQuest();
                            MatchManager.sendData({}, entity.name);
                        }
                    },
                    attacker: entity,
                    preHit: function (other) {
                        other.getComponent('punchable', function (punchable) {
                            behavior.punch.attack = punchable.getMaxHealth() * attack;
                        });
                    }
                });
                Bento.objects.attach(behavior.punch);

                Bento.objects.attach(warning);
                warning.visible = false;

                if (behavior.speed > 0) {
                    entity.scale.x = -1;
                }

                if (!isListening) {
                    behavior.listenToIncoming();
                }
            },
            destroy: function (data) {
                behavior.stopListening();
                Bento.objects.remove(behavior.punch);
                Bento.objects.remove(warning);
            },
            update: function (data) {
                entity.position.x += behavior.speed * data.speed;
                behavior.punch.position.x = entity.position.x;

                // show a warning when getting close to the screen
                if (
                    (behavior.speed > 0 && entity.position.x < viewport.x && entity.position.x > viewport.x - offScreen) ||
                    (behavior.speed < 0 && entity.position.x > viewport.x + viewport.width && entity.position.x < viewport.x + viewport.width + offScreen)
                ) {
                    warning.visible = Math.floor(entity.timer / 10) % 2;
                    // position the warning so it's always on screen
                    warning.position.y = Utils.clamp(viewport.y + 20, entity.position.y, viewport.y + viewport.height - 20);
                    if (behavior.speed > 0) {
                        warning.position.x = viewport.x + 20;
                    } else {
                        warning.position.x = viewport.x + viewport.width - 20;
                    }
                } else {
                    warning.visible = false;
                }

                var carBB = behavior.punch.boundingBox.offset(entity.position);

                // inside screen
                if (viewport.intersect(carBB)) {
                    if (behavior.notHonked) {
                        var players = Bento.objects.getByFamily('player');
                        var sign = Utils.sign(behavior.speed);
                        var characterBB;
                        carBB.x += behavior.honkDistance * sign;
                        var inPath = function (c) {
                            //return false if no boundingbox (bikerspawner)
                            if (!c.boundingBox) {
                                return false;
                            }
                            characterBB = c.boundingBox.offset(c.position);
                            if (carBB.intersect(characterBB)) {
                                return true;
                            }
                            return false;
                        };
                        var honk = function (v) {
                            Bento.audio.setVolume(v || 0.8, 'sfx_carhorn');
                            Bento.audio.stopSound('sfx_carhorn');
                            Bento.audio.playSound('sfx_carhorn');
                            behavior.notHonked = false;
                        };
                        // check players
                        Utils.forEach(players, function (player, i, l, breakLoop) {
                            if (inPath(player)) {
                                honk();
                                breakLoop();
                            }
                        });

                        // check enemies
                        if (behavior.notHonked) {
                            var enemies = Bento.objects.getByFamily('enemy');
                            Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                                if (inPath(enemy)) {
                                    honk(0.6);
                                    breakLoop();
                                }
                            });
                        }
                    }
                }

                // moving out of screen
                if (behavior.speed < 0 && entity.position.x < -64) {
                    entity.removeSelf();
                } else if (behavior.speed < 0 && entity.position.x > currentScreen.width + 64) {
                    entity.removeSelf();
                }
            },
            draw: function (data) {},
            carQuest: function () {
                EventSystem.fire('quest-car', {
                    count: 1
                });
            },
            // listen to incoming messages
            listenToIncoming: function () {
                MatchManager.onMessage(entity.name, behavior.onMessage);
                isListening = true;
            },
            stopListening: function () {
                MatchManager.offMessage(entity.name, behavior.onMessage);
                isListening = false;
            },
            onMessage: function (msg) {
                behavior.carQuest();
                // always return true if the message is handled
                return true;
            },
        };
        var warning = new Entity({
            z: Globals.Layers.EFFECTS,
            name: 'warning',
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'city/car-warning',
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });
        var entity = new Entity({
            name: 'car-' + settings.id,
            family: ['obstacles'],
            position: settings.position,
            components: [
                shadowSprite,
                sprite,
                behavior,
                sort,
                new DrawInViewport({}),
                new PunchSyncer({})
            ]
        });
        return entity;
    };
});