/**
 * Rod that arcs lightning to another rod
 * Syncs online
 * @moduleName Electric
 */
bento.define('entities/obstacles/electric', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager',
    'components/sort',
    'entities/punch',
    'bento/math/polygon',
    'components/drawinviewport',
    'managers/buff'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    MatchManager,
    Sort,
    Punch,
    Polygon,
    DrawInViewport,
    BuffManager
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var id = settings.tiled.object.id;
        var name = settings.tiled.object.name;
        var properties = Utils.nameToJson(name);
        var onTime = properties.time || 60 * 2;
        var offTime = properties.interval || 60 * 2;
        var timeOffset = properties.offset || 0;
        var chargeTime = 40;
        var start = false;

        var isRemote;
        var isListening = false;
        var powered = false;

        var attack = 0.1;

        var sprite = new Sprite({
            spriteSheet: 'obstacles/electric/rod'
        });

        var players = Bento.objects.getByFamily('player');

        var collisionPolygon;

        var electricity;

        var punchObject;
        var connectedRod;


        var behavior = {
            name: 'behaviorComponent',
            timer: timeOffset,
            start: function (data) {
                EventSystem.on('startRoom', behavior.onStartRoom);
                entity.position.x += 16;

                behavior.punch = new Punch({
                    target: ['player'],
                    attack: 2,
                    multiplier: BuffManager.current.enviromentDamage,
                    stamina: 1,
                    skill: Infinity,
                    direction: 1,
                    duration: Infinity,
                    pushBack: new Vector2(0, 1.5),
                    // parent: entity,
                    boundingBox: new Rectangle(-16, -12, 33, 12),
                    position: entity.position.clone(),
                    onHit: Punch.getDefaultCallback({
                        entity: entity,
                        effect: 'effects/hitfx3-medium',
                        sfx: 'sfx_punch1',
                        taptic: 'medium'
                    })
                });
                behavior.punch.rememberPunched = false;
            },
            destroy: function (data) {
                behavior.stopListening();
                EventSystem.off('startRoom', behavior.onStartRoom);
            },
            update: function (data) {
                // only connected rods can update
                if (!connectedRod) {
                    return;
                }

                // we always check if local player is master
                isRemote = !MatchManager.isMaster;

                if (!start) {
                    // wait for the event "startRoom"
                    return;
                }

                if (!isRemote) {
                    // local: behavior will set the animation
                    behavior.timer += data.speed;
                    if ((behavior.timer > onTime && powered) ||
                        (behavior.timer > offTime && !powered)) {
                        // switch power and let others know
                        behavior.switchPower();
                        behavior.timer = 0;

                        MatchManager.sendData({}, entity.name);
                    } else if (!powered && behavior.timer === offTime - chargeTime) {
                        // start to charge and let others know
                        behavior.charge();
                        MatchManager.sendData({
                            c: 1
                        }, entity.name);
                    }
                }

                // hit player when powered
                if (powered) {
                    Utils.forEach(players, function (player, i, l, breakLoop) {
                        if (!player.isRemote) {
                            player.getComponent('movable', function (movable) {
                                if (movable.boundingBox.offset(player.position).intersect(collisionPolygon)) {
                                    player.getComponent('punchable', function (punchable) {
                                        behavior.punch.onHit = Punch.getDefaultCallback({
                                            entity: player,
                                            effect: 'effects/hitfx1-medium',
                                            sfx: 'sfx_punch1',
                                            taptic: 'medium'
                                        });
                                        behavior.punch.attack = punchable.getMaxHealth() * attack;
                                        punchable.hit(behavior.punch);
                                    });
                                }
                            });
                        }
                    });
                }

                // isRemote changed value
                if (isRemote && !isListening) {
                    behavior.listenToIncoming();
                } else if (!isRemote && isListening) {
                    behavior.stopListening();
                }
            },
            charge: function () {
                entity.setSprite('obstacles/electric/rod-start');
                connectedRod.setSprite('obstacles/electric/rod-start');
            },
            power: function () {
                entity.setSprite('obstacles/electric/rod-electric');
                connectedRod.setSprite('obstacles/electric/rod-electric');
            },
            off: function () {
                entity.setSprite('obstacles/electric/rod');
                connectedRod.setSprite('obstacles/electric/rod');
            },
            switchPower: function () {
                powered = !powered;
                var action;
                if (powered) {
                    electricity.visible = true;
                    behavior.power();
                } else {
                    electricity.visible = false;
                    behavior.power();
                    Utils.timeout(8, function () {
                        behavior.off();
                    });
                }
            },
            // listen to incoming messages
            listenToIncoming: function () {
                MatchManager.onMessage(entity.name, behavior.onMessage);
                isListening = true;
            },
            stopListening: function () {
                MatchManager.offMessage(entity.name, behavior.onMessage);
                isListening = false;
            },
            onMessage: function (msg) {
                // add different msg
                if (msg.c) {
                    behavior.charge();
                } else {
                    behavior.switchPower();
                }

                // always return true if the message is handled
                return true;
            },
            onStartRoom: function () {
                start = true;

                // check for connection with other rod
                connectedRod = Bento.objects.get('electric-' + properties.connect);
                if (connectedRod) {
                    var distance = entity.position.distance(connectedRod.position);
                    var spriteString;
                    var spriteSize;
                    if (distance > 132) {
                        spriteString = 'large';
                        spriteSize = 164;
                    } else if (distance > 74) {
                        spriteString = 'medium';
                        spriteSize = 112;
                    } else {
                        spriteString = 'small';
                        spriteSize = 54;
                    }
                    electricity = new Entity({
                        z: 0,
                        name: 'electricity',
                        position: new Vector2(0, -13),
                        components: [
                            new Sprite({
                                spriteSheet: 'obstacles/electric/electric-' + spriteString
                            })
                        ]
                    });
                    entity.attach(electricity);
                    electricity.visible = false;
                    electricity.scale.x = distance / spriteSize;
                    electricity.rotation = Math.atan2(connectedRod.position.y - entity.position.y, connectedRod.position.x - entity.position.x);

                    // get the side coordinates of the rod
                    var side1 = connectedRod.position.subtract(entity.position);
                    side1.normalize();
                    side1.scalarMultiplyWith(8);

                    // rotate coordinate 45 degrees
                    side1.x += side1.y;
                    side1.y = side1.x - side1.y;
                    side1.x -= side1.y;
                    side1.y = -side1.y;

                    var side2 = new Vector2(-side1.x, -side1.y);

                    //make polygon for collision
                    collisionPolygon = new Polygon([
                        new Vector2(entity.position.x + side1.x, entity.position.y + side1.y),
                        new Vector2(entity.position.x + side2.x, entity.position.y + side2.y),
                        new Vector2(connectedRod.position.x + side2.x, connectedRod.position.y + side2.y),
                        new Vector2(connectedRod.position.x + side1.x, connectedRod.position.y + side1.y),
                    ]);
                }
            }
        };
        var entity = new Entity({
            z: 0,
            name: 'electric-' + id,
            family: ['obstacles'],
            position: settings.position,
            components: [
                behavior,
                sprite,
                new Sort({}),
                new DrawInViewport({
                    dimension: new Rectangle(0, 0, 64, 64)
                })
            ]
        }).extend({
            setSprite: function (key) {
                sprite.setSpriteSheet(key);
            },
        });
        return entity;
    };
});