/**
 * Player
 * @moduleName Player
 */
bento.define('entities/player', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/character',
    'components/controllers/touchcontroller',
    'components/controllers/traditionalcontroller',
    'components/controllers/keyboardcontroller',
    'components/characters/player',
    'components/online',
    'components/punchsyncer',
    'components/name',
    'components/punchable',
    'managers/stats',
    'components/coincounter',
    'components/damagecounter',
    'components/shake',
    'components/stats',
    'components/controllers/multitouchcontroller',
    'components/misstext',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Character,
    TouchController,
    TraditionalController,
    KeyboardController,
    PlayerBehavior,
    Online,
    PunchSyncer,
    Name,
    Punchable,
    StatsManager,
    CoinCounter,
    DamageCounter,
    Shake,
    Stats,
    MultiTouchController,
    MissText,
    MatchManager
) {
    'use strict';
    return function (settings) {
        var behavior = new PlayerBehavior({
            rootPath: settings.rootPath || 'boy/'
        });

        var stats = settings.stats || StatsManager.getAllStats('boy');

        var punchable = new Punchable({
            health: settings.health, // current health is loaded in screens/main
            maxHealth: stats.health,
            defense: stats.defense,
            stamina: parseFloat(stats.stamina)
        });
        var onlineSync = new Online({
            updateInterval: 5,
            isRemote: settings.online
        });
        var extraComponents = settings.components || [];
        var extraFamily = settings.family || [];
        var player = new Character({
            position: settings.position,
            family: ['player'].concat(extraFamily),
            components: [
                stats,
                behavior,
                onlineSync,
                punchable,
                new CoinCounter(),
                new DamageCounter({
                    type: 'hurt'
                }),
                new Shake({})
            ].concat(extraComponents)
        }).extend({
            // attackbonus consumable adds to this attackbonus
            attackBonus: 1,
            // skillbonus consumable adds to this skillbonus
            skillBonus: 1,
            // charge duration, consumable removes from this
            chargeDuration: 30,
            // Double coins, if double the amount of coins should drop
            doubleCoins: false,
            // Double coins, if double the amount of coins should drop
            knockDown: true,
        });

        player.name = 'player-' + settings.id;
        player.isPlayer = true;
        player.isRemote = settings.online || false;
        player.playerId = settings.id || 0;
        // able to grab enemies, set in screens/main by checking for scrolls/grab item
        player.canDash = true;
        player.canCharge = true;
        player.canJump = true;

        player.getComponent('stateMachine', function (stateMachine) {
            stateMachine.data.stats = stats;
        });

        if (!settings.online) {
            player.attach(new MultiTouchController({}));
            if (Utils.isDev) {
                player.attach(new KeyboardController({}));
            }

            player.attach(new PunchSyncer({
                onlineKey: 'players'
            }));

            // add miss text to local player
            player.attach(new MissText({}));

            // sync health with other players
            if (MatchManager.isOnline) {
                player.attach({
                    name: 'healthSyncer',
                    start: function () {
                        punchable.onHealthChange(function (data) {                    
                            MatchManager.sendData({
                                health: data.health,
                                maxHealth: data.maxHealth
                            }, 'updatePlayerHealth-' + player.playerId);
                        });
                    }
                });
            }

            // local player starts dead
            if (settings.health === 0) {
                // this is kind of hacky, but we have to manually set all component values
                player.getComponent('stateMachine', function (stateMachine) {
                    player.attach({
                        name: 'startDead',
                        start: function () {
                            punchable.setHealth(0);
                            stateMachine.setState('dead');
                            EventSystem.fire('playerDeath', {
                                player: player
                            });
                            // high probability this will not be synced
                            // see screens/main where the player health is sent during onRoomStart
                        }
                    });
                });
            }

        } else {
            player.remove(player.getComponent('punchable'));
            behavior.canPunch = false;

            player.attach(new Name({
                name: settings.name
            }));

            // we do attach the punchable to the stateMachine, to prevent any crashes
            player.getComponent('stateMachine', function (stateMachine) {
                stateMachine.data.punchable = punchable;
            });
        }
        // console.log('player spawned: ', settings.online, player.name);

        return player;
    };
});