/**
 * Playable character: Pat
 * @moduleName PlayerPat
 */
bento.define('entities/players/pat', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/tapticengine',
    'entities/effect',
    'components/sort',
    'entities/player',
    'managers/stats',
    'components/stats',
    'entities/punch',
    'managers/buff',
    'gui/super/superattackbars',
    'gui/super/superattackspeedlines',
    'entities/triggers/blockade',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    TapticEngine,
    Effect,
    Sort,
    Player,
    StatsManager,
    Stats,
    Punch,
    BuffManager,
    SuperAttackBars,
    SuperAttackSpeedlines,
    Blockade,
    Globals
) {
    'use strict';
    var Pat = function (settings) {
        // WARNING: it is assumed this component comes AFTER PlayerBehavior,
        // because the playerBehavior adds the standard player states
        var addStates = {
            name: 'extraStates',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addState('runAttack', Pat.States.rocketFist);
                stateMachine.addState('chargeAttack', Pat.States.jetStream);
                stateMachine.addState('comboEnd', Pat.States.highKick);
                stateMachine.addState('jumpAttack', Pat.States.jumpAttack);
                stateMachine.addState('superAttack', Pat.States.superAttack);
            }
        };

        var behavior = {
            name: 'patBehavior',
            start: function () {
                pat.getComponent('stateMachine', function (stateMachine) {
                    stateMachine.data.punchReset = [3, 2, 2, 2];
                    stateMachine.data.punchLength = [20, 20, 20, 20];
                });
            }
        };

        var pat = new Player({
            rootPath: 'pat/',
            healthBar: settings.healthBar,
            position: settings.position,
            family: ['player'],
            online: settings.online,
            name: settings.name,
            id: settings.id,
            stats: StatsManager.getAllStats('pat'),
            health: settings.health,
            components: [
                addStates,
                behavior
            ]
        });

        return pat;
    };

    // var stats = StatsManager.getAllStats('boy');
    /**
     * Pluggable states
     */
    Pat.States = {
        rocketFist: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'rocketpunch', function () {
                    data.setState('idle');
                });
                data.specialStage = transitionData.specialStage;
                data.hasPunched = false;

                Bento.audio.setVolume(0.25, 'sfx_dashattack2');
                Bento.audio.stopSound('sfx_dashattack2');
                Bento.audio.playSound('sfx_dashattack2');

                Bento.audio.stopSound('pat/sfx_dash');
                Bento.audio.playSound('pat/sfx_dash');
                data.playerBehavior.decelaration = 0.04;

            },
            update: function (data) {
                var speed = [1, 0];

                //punches are created first for auto-target purposes, and added later for animation purposes
                var baseAttack = data.stats.attack;
                var multiplier = [1.5, 2];
                var pushHeight = [1.5, 2.5];

                if (!data.hasPunched && data.sprite.getCurrentFrame() > 3) {
                    var punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: baseAttack,
                        stamina: 0.334,
                        skill: data.stats.skill,
                        multiplier: multiplier[data.specialStage - 1] * BuffManager.rollCritical('dashAttack'),
                        freeze: 8,
                        pushBack: new Vector2(2 * BuffManager.rollCritical('knockback'), pushHeight[data.specialStage - 1]),
                        boundingBox: new Rectangle(0, -8, 36, 16),
                        offset: new Vector2(0, 0),
                        ignoreFlip: true,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1',
                            sfx: 'sfx_punch3',
                            taptic: 'heavy'
                        }),
                        type: 'dashAttack'
                    });
                    // data.movable.velocity.x = 2 * data.multiScale.direction.x;
                    Bento.objects.attach(punch);
                    data.hasPunched = true;
                }
            },
            end: function (data) {
                data.specialStage = 0;
                // reset deceleration
                data.playerBehavior.decelaration = 0.2;

            }
        },
        jetStream: {
            start: function (data, transitionData) {
                data.isLanding = false;
                // TODO
                data.sprite.setSpriteSheet(data.rootPath + 'jetstream', function () {
                    data.isLanding = true;
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    // data.sprite.setAnimation('shoryukenEnd');
                });
                data.specialStage = transitionData.specialStage;

                var baseAttack = data.stats.attack;
                var multiplier = 2 * BuffManager.rollCritical('chargeAttack');
                var bonusAttack = [0, 2];
                var pushHeight = [3, 4];

                Bento.audio.stopSound('sfx_shoot');
                Bento.audio.playSound('sfx_shoot');

                Bento.audio.stopSound('pat/sfx_special');
                Bento.audio.playSound('pat/sfx_special');

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    stamina: 1,
                    skill: data.stats.skill,
                    multiplier: multiplier,
                    ignoreFlip: true,
                    bonusAttack: bonusAttack[data.specialStage - 1],
                    freeze: 8,
                    pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), pushHeight[data.specialStage - 1]),
                    boundingBox: new Rectangle(0, -5, 28, 10),
                    offset: new Vector2(8, 0),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1',
                        sfx: 'sfx_special-hit',
                        taptic: 'heavy'
                    }),
                    type: 'specialAttack'
                });
            },
            update: function (data) {
                if (data.punch && data.sprite.getCurrentFrame() > 2) {
                    data.movable.velocity.x = 2 * -data.multiScale.direction.x;
                    // data.movable.velocityZ = jump[data.specialStage - 1];

                    Bento.objects.attach(data.punch);
                    data.punch = null;
                }

                if (data.isLanding && data.movable.grounded) {
                    data.setState('idle');
                }
            },
            action: function (data, action) {
                switch (action.type) {
                case 'tap':
                    if (data.isLanding) {
                        // data.setState('airPunch');
                    }
                    break;
                }
            },
            end: function (data) {
                data.isLanding = false;
                data.specialStage = 0;
            }
        },
        highKick: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'highkick', function () {
                    data.setState('idle');
                });

                data.combo = 0;

                Bento.audio.stopSound('sfx_kick');
                Bento.audio.playSound('sfx_kick');
                Bento.audio.stopSound('pat/sfx_dash');
                Bento.audio.playSound('pat/sfx_dash');

                data.attackState = 0;

            },
            update: function (data) {
                var baseAttack = data.stats.attack;
                if (data.attackState === 0 && data.sprite.getCurrentFrame() > 4) {
                    data.punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: Math.ceil(baseAttack / 2),
                        stamina: 0,
                        skill: data.stats.skill,
                        multiplier: BuffManager.rollCritical('comboEnd'),
                        freeze: 5,
                        pushBack: new Vector2(0, 1),
                        boundingBox: new Rectangle(0, -5, 16, 10),
                        offset: new Vector2(8, 0),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-small',
                            sfx: 'sfx_punch3',
                            taptic: 'medium',
                            onComplete: function () {
                                data.movable.velocity.x = 1 * data.multiScale.direction.x;
                            }
                        })
                    });
                    Bento.objects.attach(data.punch);
                    data.movable.velocity.x = 1.5 * data.multiScale.direction.x;
                    data.attackState = 1;
                } else if (data.attackState === 1 && data.sprite.getCurrentFrame() > 5) {
                    data.punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: Math.ceil(baseAttack / 2),
                        stamina: 1,
                        skill: data.stats.skill,
                        multiplier: BuffManager.rollCritical('comboEnd'),
                        freeze: 5,
                        pushBack: new Vector2(0.75 * BuffManager.rollCritical('knockback'), 2.0),
                        boundingBox: new Rectangle(0, -5, 19, 10),
                        offset: new Vector2(8, 0),
                        ignoreFlip: true,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-medium',
                            sfx: 'sfx_punch3',
                            taptic: 'medium',
                            onComplete: function () {
                                data.movable.velocity.x = 1 * data.multiScale.direction.x;
                            }
                        })
                    });
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                    data.attackState = 2;
                }
            }
        },
        jumpAttack: {
            start: function (data, transitionData) {
                EventSystem.fire('playerJumpAttack');
                data.movable.velocity.x = 0;
                data.movable.velocity.y = 0;
                var direction = transitionData.cacheDirection;
                data.multiScale.direction.x = direction;
                data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-windup', function () {
                    Bento.audio.stopSound('sfx_dashattack');
                    Bento.audio.playSound('sfx_dashattack');
                    Bento.audio.stopSound('pat/sfx_throw');
                    Bento.audio.playSound('pat/sfx_throw');
                    data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-loop');
                    data.movable.velocityZ = -2.5;
                    data.movable.velocity.x = direction * 3;
                    data.punchable.setInvincible(true);
                    var effect = new Effect({
                        z: Globals.Layers.ENEMIES,
                        position: data.entity.position.add(new Vector2(0, -8)),
                        spriteSheet: 'effects/jump-dust'
                    });
                    Bento.objects.attach(effect);
                });
                data.attackState = 0;
                data.punchable.setInvincible(false);
            },
            update: function (data) {
                var invincible = true;
                if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-windup') {
                    if (data.attackState === 0 && data.sprite.getCurrentFrame() >= 1) {
                        data.punch = data.createPunch({
                            target: ['enemy', 'destructible'],
                            attack: Math.ceil(data.stats.attack / 2),
                            stamina: 0.5,
                            skill: data.stats.skill,
                            multiplier: data.stats.multiplier,
                            bonusAttack: 0,
                            freeze: 5,
                            duration: 30,
                            height: 32,
                            pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), 0.8),
                            boundingBox: new Rectangle(0, -5, 16, 10),
                            offset: new Vector2(8, 0),
                            ignoreFlip: true,
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx1-small',
                                sfx: 'sfx_punch0',
                                taptic: 'medium'
                            }),
                            type: 'jumpAttack'
                        });
                        Bento.objects.attach(data.punch);
                        data.punch = data.createPunch({
                            target: ['enemy', 'destructible'],
                            attack: Math.ceil(data.stats.attack / 2),
                            stamina: 0.5,
                            skill: data.stats.skill,
                            multiplier: data.stats.multiplier,
                            bonusAttack: 0,
                            freeze: 5,
                            duration: 30,
                            height: 32,
                            pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), 0.8),
                            boundingBox: new Rectangle(0, -5, 16, 10),
                            offset: new Vector2(-24, 0),
                            ignoreFlip: true,
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx1-small',
                                sfx: 'sfx_punch0',
                                taptic: 'medium'
                            }),
                            type: 'jumpAttack'
                        });
                        Bento.objects.attach(data.punch);

                        Bento.audio.stopSound('sfx_shoot');
                        Bento.audio.playSound('sfx_shoot');
                        data.attackState = 1;
                    }
                } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-loop') {
                    // check if about to land
                    if (data.movable.velocityZ > 0 && data.movable.height > -10) {
                        data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-land', function () {
                            data.setState('idle');
                        });
                    }
                } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-land' &&
                    data.sprite.getCurrentFrame() >= 1
                ) {
                    invincible = false;
                }
                data.punchable.setInvincible(invincible);
            }
        },
        superAttack: {
            start: function (data, transitionData) {
                if (!data.isRemote) {
                    EventSystem.fire('quest-doSuperAttack', {
                        count: 1
                    });
                }
                if (!data.isRemote) {
                    var enemies = [];
                    // check for enemies on screen
                    var screenBB = Bento.getViewport().clone();
                    //shrink screen bounding box by players boundingbox
                    screenBB.grow(-data.movable.boundingBox.width, false, true);
                    Utils.forEach(Bento.objects.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                        var ignoreEnemy = enemy.family.indexOf('ignoreEnemy');

                        // check if dead
                        enemy.getComponent('punchable', function (punchable) {
                            if (punchable.isDead || punchable.getHealth() <= 0) {
                                ignoreEnemy = 1;
                            }
                        });

                        enemy.getComponent('movable', function (movable) {
                            if (!movable.grounded) {
                                ignoreEnemy = 1;
                            }
                        });

                        if (Math.abs(enemy.position.x - data.entity.position.x) < 90 &&
                            screenBB.hasPosition(enemy.position) &&
                            ignoreEnemy === -1
                        ) {
                            enemies.push(enemy);
                        }
                    });

                    // target enemies in front, with highest health
                    var health = 0;
                    var hp = 0;
                    var target;
                    var sign = data.multiScale.direction.x;
                    var range = 40;
                    var distance = 90;
                    Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                        if (Utils.sign(enemy.position.x - data.entity.position.x) === sign &&
                            Math.abs(enemy.position.y - data.entity.position.y) < range
                        ) {
                            if (Math.abs(data.entity.position.x - enemy.position.x) < distance) {
                                target = enemy;
                                distance = Math.abs(data.entity.position.x - enemy.position.x);
                            }
                        }
                    });

                    // sync data for remotes
                    data.sync({
                        target: target ? target.name : '',
                        superAttack: target ? true : false
                    });

                    if (target) {
                        Bento.objects.get('healthBar', function (healthBar) {
                            healthBar.depleteComboBar();
                        });
                    }

                    data.entity.getComponent('weapon', function (weapon) {
                        weapon.getComponent('weapon').drop(0);
                    });
                }

                if (data.syncData.superAttack) {
                    data.attackState = 0;
                    if (!data.isRemote) {
                        data.punchable.setInvincible(true);
                        EventSystem.fire('superAttack');
                    }
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    data.movable.settings.ignoreSolids = true;
                } else {
                    data.setState('superAttackDud');
                }
                data.currentTarget = null;
            },
            update: function (data) {
                var target = data.currentTarget;
                if (!target) {
                    // find and cache the target
                    data.currentTarget = Bento.objects.get(data.syncData.target);
                    target = data.currentTarget;
                }
                if (!target) {
                    // target could not be found
                    data.setState('idle');
                    return;
                }

                //set invincible every update
                if (!data.isRemote) {
                    data.punchable.setInvincible(true);
                }

                // align with target
                if (data.attackState === 0) {
                    target.getComponent('stateMachine', function (stateMachine) {
                        stateMachine.setState('targetSuperAttack');
                    });

                    data.sprite.setSpriteSheet(data.rootPath + 'walk');

                    var xSign = Utils.sign(target.position.x - data.entity.position.x);
                    if (xSign === 0) {
                        xSign = 1 + Utils.getRandom(2) * -2;
                    }

                    if (!data.isRemote) {
                        data.blackBar = new SuperAttackBars({});
                        Bento.objects.attach(data.blackBar);

                        //lower music volume
                        new Tween({
                            from: 1,
                            to: 0.3,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    // clamp position inside playable area
                    var targetPoint = target.position.add(new Vector2(-xSign * 100, 0));
                    var rightSide = Bento.screens.getCurrentScreen().dimension.width - 10;
                    if (Blockade.active) {
                        rightSide = Blockade.active.position.x;
                    }
                    targetPoint.x = Utils.clamp(10, targetPoint.x, rightSide);

                    new Tween({
                        from: data.entity.position,
                        to: targetPoint,
                        in: 30,
                        delay: 0,
                        ease: 'linear',
                        onStart: function () {},
                        onUpdate: function (v, t) {
                            data.entity.position = v;
                            data.multiScale.direction.x = xSign;
                        },
                        onComplete: function () {
                            data.sprite.setSpriteSheet(data.rootPath + 'idle');
                            // Utils.timeout(30, function () {
                            data.attackState = 2;

                            if (!data.isRemote) {
                                // set layers for player and target to be above everything else
                                data.entity.removeByName('sort');
                                data.entity.z = Globals.Layers.SUPERATTACK;
                                target.removeByName('sort');
                                target.z = Globals.Layers.SUPERATTACK - 2;
                                // force camera between player and enemy
                                Bento.objects.get('camera', function (camera) {
                                    camera.getComponent('cameraBehavior', function (camBehavior) {
                                        camBehavior.target = {
                                            position: new Vector2(data.entity.position.x * 0.5 + target.position.x * 0.5, data.entity.position.y * 0.5 + target.position.y * 0.5)
                                        };
                                    });
                                });
                            }

                            var effect = new Effect({
                                position: data.entity.position.add(new Vector2(-xSign * 4, -16)),
                                spriteSheet: 'effects/superattack-chargefx'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);

                            // set player sprite and add fx
                            Utils.timeout(2, function () {
                                //play supercharge sound
                                Bento.audio.stopSound('sfx_super-charge');
                                Bento.audio.playSound('sfx_super-charge');
                                data.sprite.setSpriteSheet(data.rootPath + 'superattack', function () {
                                    data.sprite.setSpriteSheet(data.rootPath + 'victory');
                                });
                            });
                        }
                    });
                    data.attackState = 1;
                } else if (data.attackState === 2) {
                    if (data.sprite.getCurrentFrame() >= 14) {
                        data.attackState = 3;
                    }
                } else if (data.attackState === 3) {
                    // dash towards target
                    data.attackState = 4;

                    data.rocketFist = new Entity({
                        z: Globals.Layers.SUPERATTACK + 1,
                        name: 'rocketFist',
                        position: data.entity.position.clone(),
                        components: [
                            new Sprite({
                                spriteSheet: data.rootPath + 'rocketfist'
                            })
                        ]
                    });

                    data.rocketFist.scale.x = Utils.sign(target.position.x - data.entity.position.x);

                    Bento.objects.attach(data.rocketFist);

                    var enemyRange = 24;
                    target.getComponent('punchable', function (punchable) {
                        enemyRange = punchable.boundingBox.width / 2 + 14;
                    });
                    new Tween({
                        from: data.entity.position.x + Utils.sign(target.position.x - data.entity.position.x) * 24,
                        to: target.position.x + enemyRange * Utils.sign(data.entity.position.x - target.position.x),
                        in: 4,
                        delay: 0,
                        ease: 'easeInQuad',
                        onStart: function () {
                            var effect = new Effect({
                                position: data.entity.position.add(new Vector2(Utils.sign(target.position.x - data.entity.position.x) * 40, 0)),
                                spriteSheet: 'effects/superattack-dashfx'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);
                        },
                        onUpdate: function (v, t) {
                            data.rocketFist.position.x = v;
                        },
                        onComplete: function () {
                            data.attackState = 5;
                            // arrived at target, show hit fx
                            Utils.timeout(5, function () {
                                var effect = new Effect({
                                    position: target.position.add(new Vector2(0, -12)),
                                    spriteSheet: 'effects/superattack-hitfx'
                                });
                                if (!data.isRemote) {
                                    effect.z = Globals.Layers.SUPERATTACK - 1;
                                    TapticEngine.impact({
                                        style: 'heavy'
                                    });
                                }
                                Bento.objects.add(effect);

                                if (!data.isRemote) {
                                    data.blackBar.flash();
                                }
                            });
                        }
                    });
                } else if (data.attackState === 5) {
                    var combo = 0;
                    var targetX = target.position.x;
                    var rocketX = data.rocketFist.position.x;
                    var sign = Utils.sign(target.position.x - data.entity.position.x);
                    var doPunch = function () {
                        var targetStateMachine;
                        target.getComponent('stateMachine', function (stateMachine) {
                            targetStateMachine = stateMachine;
                        });

                        Utils.timeout(5, function () {
                            var effect = new Effect({
                                position: target.position.add(new Vector2((Math.abs(data.rocketFist.position.x - target.position.x) - 30) * -sign, -16)),
                                spriteSheet: 'effects/hitfx1-small'
                            });
                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                                TapticEngine.impact({
                                    style: 'medium'
                                });
                            }
                            Bento.objects.add(effect);

                            if (targetStateMachine) {
                                targetStateMachine.data.sprite.setSpriteSheet(targetStateMachine.data.rootPath + 'hurt' + (combo % 2 + 1));
                            }
                            new Tween({
                                from: targetX,
                                to: targetX + 10 * sign,
                                in: 10,
                                delay: 0,
                                ease: 'easeOutQuad',
                                onStart: function () {},
                                onUpdate: function (v, t) {
                                    target.position.x = v;
                                },
                            });
                            new Tween({
                                from: rocketX,
                                to: rocketX + 10 * sign,
                                in: 5,
                                delay: 0,
                                ease: 'easeOutQuad',
                                onStart: function () {},
                                onUpdate: function (v, t) {
                                    data.rocketFist.position.x = v;
                                },
                            });
                        });

                        var dynaSpeed = 1 - combo / 12;
                        dynaSpeed *= dynaSpeed * dynaSpeed;
                        dynaSpeed = 1 - dynaSpeed;
                        dynaSpeed *= 10;
                        Utils.timeout(16 - Math.floor(dynaSpeed), function () {
                            if (combo < 12) {
                                doPunch();
                            } else {
                                data.attackState = 7;
                            }
                        });
                        Bento.audio.stopSound('sfx_punch');
                        Bento.audio.playSound('sfx_punch');
                        combo += 1;
                    };
                    doPunch();
                    data.attackState = 6;
                } else if (data.attackState === 7) {
                    Utils.timeout(5, function () {

                        new Tween({
                            from: data.rocketFist.position.x,
                            to: data.rocketFist.position.x + Utils.sign(target.position.x - data.rocketFist.position.x) * 600,
                            in: 60,
                            delay: 0,
                            ease: 'linear',
                            onUpdate: function (v, t) {
                                data.rocketFist.position.x = v;
                            },
                            onComplete: function () {
                                Bento.objects.remove(data.rocketFist);
                            }
                        });
                        if (!data.isRemote) {
                            Bento.objects.attach(new SuperAttackSpeedlines({
                                actors: [
                                    data.entity,
                                    target
                                ],
                                rootPath: data.rootPath
                            }));
                            data.blackBar.removeSelf();
                            data.blackBar = undefined;

                            Bento.audio.stopSound('pat/sfx_throw');
                            Bento.audio.playSound('pat/sfx_throw');

                            Bento.objects.get('camera', function (camera) {
                                camera.shake(20, 10);
                            });
                        }

                        var effect = new Effect({
                            position: target.position.add(new Vector2(0, -12)),
                            spriteSheet: 'effects/hitfx1'
                        });
                        if (!data.isRemote) {
                            effect.z = Globals.Layers.SUPERATTACK - 1;
                        }
                        Bento.objects.add(effect);
                        effect = new Effect({
                            position: target.position.add(new Vector2(0, -12)),
                            spriteSheet: 'effects/superattack-hitfx'
                        });
                        if (!data.isRemote) {
                            effect.z = Globals.Layers.SUPERATTACK - 1;
                        }
                        Bento.objects.add(effect);

                        Utils.timeout(10, function () {
                            if (!data.isRemote) {
                                target.getComponent('stateMachine', function (stateMachine) {
                                    stateMachine.data.punchable.setInvincible(false);
                                    stateMachine.data.punchable.hit(data.createPunch({
                                        target: ['enemy', 'destructible'],
                                        attack: data.stats.attack,
                                        stamina: 1,
                                        skill: Infinity,
                                        multiplier: 15,
                                        freeze: 5,
                                        pushBack: new Vector2(3, 2.0),
                                        boundingBox: new Rectangle(0, -5, 16, 10),
                                        offset: new Vector2(8, 0),
                                        onHit: Punch.getDefaultCallback({
                                            entity: data.entity,
                                            effect: 'effects/hitfx1',
                                            sfx: 'sfx_finalpunch',
                                            taptic: 'heavy',
                                            onComplete: function () {
                                                data.movable.velocity.x = 1 * data.multiScale.direction.x;
                                            }
                                        }),
                                        type: 'superAttack'
                                    }));
                                });

                                TapticEngine.impact({
                                    style: 'heavy'
                                });
                            }
                        });
                    });

                    new Tween({
                        from: data.entity.position.x,
                        to: target.position.x,
                        in: 20,
                        delay: 0,
                        ease: 'easeOutQuad',
                        onUpdate: function (v, t) {
                            data.entity.position.x = v;
                        }
                    });

                    Bento.audio.stopSound('pat/sfx_special');
                    Bento.audio.playSound('pat/sfx_special');

                    data.sprite.setSpriteSheet(data.rootPath + 'punch2', function () {
                        data.setState('idle');

                        if (!data.isRemote) {
                            Bento.objects.get('camera', function (camera) {
                                camera.getComponent('cameraBehavior', function (camBehavior) {
                                    camBehavior.target = data.entity;
                                });
                            });
                        }
                    });

                    if (!data.isRemote) {
                        new Tween({
                            from: 0.3,
                            to: 1,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    data.attackState = 8;
                }
            },
            end: function (data) {
                if (!data.isRemote) {
                    Utils.timeout(80, function () {
                        data.punchable.setInvincible(false);
                    });
                }
                if (data.rocketfist && data.rocketfist.isAdded) {
                    data.rocketfist.removeSelf();
                }
                data.movable.settings.ignoreSolids = false;
            }
        }
    };
    return Pat;
});