/**
 * Playable character: Rad
 * @moduleName PlayerRad
 */
bento.define('entities/players/rad', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/tapticengine',
    'entities/effect',
    'components/sort',
    'entities/player',
    'managers/stats',
    'components/stats',
    'entities/punch',
    'managers/buff',
    'gui/super/superattackbars',
    'globals',
    'gui/super/superattackspeedlines',
    'entities/triggers/blockade',
    'states/standardplayer'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    TapticEngine,
    Effect,
    Sort,
    Player,
    StatsManager,
    Stats,
    Punch,
    BuffManager,
    SuperAttackBars,
    Globals,
    SuperAttackSpeedlines,
    Blockade,
    StandardPlayerStates
) {
    'use strict';
    var Rad = function (settings) {
        // WARNING: it is assumed this component comes AFTER PlayerBehavior,
        // because the playerBehavior adds the standard player states

        var addStates = {
            name: 'extraStates',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addState('runAttack', Rad.States.torpedo);
                stateMachine.addState('chargeAttack', Rad.States.hookshot);
                stateMachine.addState('comboEnd', Rad.States.bite);
                stateMachine.addState('superAttack', Rad.States.superAttack);
                stateMachine.addState('jumpAttack', Rad.States.jumpAttack);
            }
        };

        var behavior = {
            name: 'radBehavior',
            start: function () {
                rad.getComponent('stateMachine', function (stateMachine) {
                    // frame after which taps get cached
                    stateMachine.data.punchReset = [3, 3, 3, 3];
                    stateMachine.data.punchLength = [24, 24, 24, 24];
                    stateMachine.data.punchDistance = [16, 16, 16, 16];
                    stateMachine.data.grabDistance = 38;
                });
            }
        };

        var rad = new Player({
            rootPath: 'rad/',
            healthBar: settings.healthBar,
            position: settings.position,
            family: ['player'],
            online: settings.online,
            name: settings.name,
            id: settings.id,
            stats: StatsManager.getAllStats('rad'),
            health: settings.health,
            components: [
                addStates,
                behavior
            ]
        });

        return rad;
    };

    /**
     * Pluggable states
     */
    Rad.States = {
        dead: {
            start: function (data, transitionData) {
                Bento.audio.stopSound('rad/sfx_death');
                Bento.audio.playSound('rad/sfx_death');

                data.sprite.setSpriteSheet(data.rootPath + 'liedown');
                data.movable.velocity.x = transitionData.direction || 0;
                data.movable.velocity.y = 0;

                data.multiScale.direction.x = transitionData.direction * -1;
                // to ensure that entity.visible is set to true
                if (!data.isRemote) {
                    data.punchable.setInvincible(false);
                }
            }
        },
        torpedo: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'torpedo', function () {
                    data.setState('idle');
                });
                data.hasPunched = 0;

                Bento.audio.stopSound('sfx_dashattack');
                Bento.audio.playSound('sfx_dashattack');

                Bento.audio.stopSound('rad/sfx_dash');
                Bento.audio.playSound('rad/sfx_dash');

            },
            update: function (data) {
                var speed = [1, 0];
                data.movable.velocity.x = 2 * data.multiScale.direction.x;

                //punches are created first for auto-target purposes, and added later for animation purposes
                var baseAttack = data.stats.attack;
                var punch;

                if (data.hasPunched < 1 && data.sprite.getCurrentFrame() > 0) {
                    punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: baseAttack / 3,
                        stamina: 0.334 / 2,
                        skill: data.stats.skill,
                        duration: 2,
                        multiplier: 1.5 * BuffManager.rollCritical('dashAttack'),
                        freeze: 5,
                        pushBack: new Vector2(0.75 * BuffManager.rollCritical('knockback'), 2),
                        boundingBox: new Rectangle(0, -5, 28, 10),
                        offset: new Vector2(0, 0),
                        ignoreFlip: true,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-medium',
                            sfx: 'sfx_dashattack-hit',
                            taptic: 'heavy',
                            onComplete: function () {
                                data.movable.velocity.x *= 0.5;
                            }
                        }),
                        type: 'dashAttack'
                    });
                    data.movable.velocity.x = 0.5 * data.multiScale.direction.x;
                    Bento.objects.attach(punch);
                    data.hasPunched = 1;
                } else if (data.hasPunched < 2 && data.sprite.getCurrentFrame() > 2) {
                    punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: baseAttack / 3,
                        stamina: 0.334 / 2,
                        skill: data.stats.skill,
                        duration: 2,
                        multiplier: 1.5 * BuffManager.rollCritical('dashAttack'),
                        freeze: 5,
                        pushBack: new Vector2(1 * BuffManager.rollCritical('knockback'), 1.5),
                        boundingBox: new Rectangle(0, -5, 28, 10),
                        offset: new Vector2(0, 0),
                        ignoreFlip: true,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-medium',
                            sfx: 'sfx_dashattack-hit',
                            taptic: 'heavy',
                            onComplete: function () {
                                data.movable.velocity.x *= 0.5;
                            }
                        }),
                        type: 'dashAttack'
                    });
                    data.movable.velocity.x = Utils.approach(data.movable.velocity.x, 0, 0.1);
                    data.hasPunched = 2;
                    data.movable.velocity.x = 0.5 * data.multiScale.direction.x;
                    Bento.objects.attach(punch);

                } else if (data.sprite.getCurrentFrame() > 0) {
                    data.movable.velocity.x = 3 * data.multiScale.direction.x;
                }
                if (data.entity.timer % 4 === 0) {
                    var puff = new Effect({
                        position: data.entity.position.add(new Vector2(
                            data.multiScale.direction.x * -16,
                            0
                        )),
                        spriteSheet: 'effects/walkcloud'
                    });
                    puff.scale.x = data.multiScale.direction.x;
                    puff.attach(new Sort({}));
                    Bento.objects.attach(puff);
                }
            }
        },
        hookshot: {
            start: function (data, transitionData) {
                data.reelIn = false;
                data.sprite.setSpriteSheet(data.rootPath + 'chargeup-hookshot', function () {
                    data.reelIn = true;
                    data.sprite.setSpriteSheet(data.rootPath + 'chargeup-reelin', function () {
                        data.sprite.setSpriteSheet(data.rootPath + 'idle');
                        data.setState('idle');
                    });
                });

                var baseAttack = data.stats.attack;

                Bento.audio.stopSound('sfx_special');
                Bento.audio.playSound('sfx_special');

                Bento.audio.stopSound('rad/sfx_special');
                Bento.audio.playSound('rad/sfx_special');

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    stamina: 0.5,
                    skill: data.stats.skill,
                    multiplier: BuffManager.rollCritical('chargeAttack'),
                    bonusAttack: 0,
                    pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), 1),
                    boundingBox: new Rectangle(0, -5, 28, 10),
                    offset: new Vector2(16, 0),
                    freeze: 5,
                    duration: 16,
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1',
                        sfx: 'sfx_special-hit',
                        taptic: 'heavy'
                    }),
                    type: 'specialAttack'
                });
                data.punchState = 0;
            },
            update: function (data) {
                if (data.punchState === 0) {
                    data.movable.velocity.x = 0.2 * data.multiScale.direction.x;
                    data.movable.grounded = false;
                    Bento.objects.attach(data.punch);
                    data.punchState = 1;
                } else if (data.punchState === 1) {
                    new Tween({
                        from: data.punch.position.x,
                        to: data.punch.position.x + 80 * data.multiScale.direction.x,
                        in: 16,
                        delay: 0,
                        ease: 'easeOutQuad',
                        onUpdate: function (v, t) {
                            if (data.punch) {
                                data.punch.position.x = v;
                            }
                        }
                    });
                    data.punchState = 2;
                } else if (data.reelIn && data.punchState === 2) {
                    data.punchState = 3;
                } else if (data.punchState === 3) {
                    // go back
                    var baseAttack = data.stats.attack;
                    data.punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: baseAttack,
                        stamina: 1,
                        skill: data.stats.skill,
                        multiplier: 1.5 * BuffManager.rollCritical('chargeAttack'),
                        bonusAttack: 0,
                        pushBack: new Vector2(-2 * BuffManager.rollCritical('knockback'), 2),
                        boundingBox: new Rectangle(0, -5, 28, 10),
                        offset: new Vector2(16, 0),
                        freeze: 6,
                        duration: 16,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1',
                            sfx: 'sfx_special-hit',
                            taptic: 'heavy'
                        }),
                        type: 'specialAttack',
                        ignoreFlip: true
                    });
                    Bento.objects.attach(data.punch);

                    new Tween({
                        from: data.entity.position.x + 96 * data.multiScale.direction.x,
                        to: data.entity.position.x + 16 * data.multiScale.direction.x,
                        in: 16,
                        delay: 0,
                        ease: 'easeOutQuad',
                        onUpdate: function (v, t) {
                            if (data.punch) {
                                data.punch.position.x = v;
                            }
                        },
                        onComplete: function () {
                            data.punch = undefined;
                        }
                    });
                    data.punchState = 4;
                }
            },
            end: function (data) {
                data.isLanding = false;
                if (data.punch && data.punch.isAdded) {
                    data.punch.removeSelf();
                }
                data.punch = undefined;
            }
        },
        jumpAttack: {
            start: function (data, transitionData) {
                EventSystem.fire('playerJumpAttack');
                data.movable.velocity.x = 0;
                data.movable.velocity.y = 0;
                data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-windup', function () {
                    data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-loop');
                });
                data.attackState = 0;
                data.punchable.setInvincible(false);

                data.sync({
                    cacheDirection: transitionData.cacheDirection
                });
            },
            update: function (data) {
                var invincible = true;
                if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-windup') {
                    if (data.attackState === 0 && data.sprite.getCurrentFrame() > 2) {
                        var direction = data.syncData.cacheDirection;
                        data.multiScale.direction.x = direction;
                        data.movable.velocityZ = -2.5;
                        data.movable.velocity.x = direction * 3;
                        data.attackState = 1;
                        data.punchable.setInvincible(true);
                        Bento.audio.stopSound('sfx_dashattack');
                        Bento.audio.playSound('sfx_dashattack');
                        Bento.audio.stopSound('rad/sfx_throw');
                        Bento.audio.playSound('rad/sfx_throw');
                        var effect = new Effect({
                            z: Globals.Layers.ENEMIES,
                            position: data.entity.position.add(new Vector2(0, -8)),
                            spriteSheet: 'effects/jump-dust'
                        });
                        Bento.objects.attach(effect);
                    }
                } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-loop') {
                    var punch;
                    if (data.attackState === 1 && data.sprite.getCurrentFrame() >= 1) {
                        // add punch
                        punch = data.createPunch({
                            target: ['enemy', 'destructible'],
                            attack: Math.ceil(data.stats.attack / 2),
                            stamina: 0.5,
                            skill: data.stats.skill,
                            multiplier: data.stats.multiplier,
                            bonusAttack: 0,
                            duration: 10,
                            freeze: 4,
                            height: 32,
                            pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), 0.8),
                            boundingBox: new Rectangle(0, -5, 44, 10),
                            offset: new Vector2(-8, 0),
                            ignoreFlip: true,
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx1-medium',
                                sfx: 'sfx_punch3',
                                taptic: 'heavy'
                            }),
                            type: 'jumpAttack'
                        });
                        Bento.objects.attach(punch);
                        data.attackState = 2;
                    } else if (data.attackState === 2 && data.sprite.getCurrentFrame() >= 3) {
                        // add punch
                        punch = data.createPunch({
                            target: ['enemy', 'destructible'],
                            attack: Math.ceil(data.stats.attack / 2),
                            stamina: 0.5,
                            skill: data.stats.skill,
                            multiplier: data.stats.multiplier,
                            bonusAttack: 0,
                            duration: 8,
                            freeze: 4,
                            height: 32,
                            pushBack: new Vector2(-0.5 * BuffManager.rollCritical('knockback'), 0.8),
                            boundingBox: new Rectangle(0, -5, 44, 10),
                            offset: new Vector2(-36, 0),
                            ignoreFlip: true,
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx1-medium',
                                sfx: 'sfx_punch3',
                                taptic: 'heavy'
                            }),
                            type: 'jumpAttack'
                        });
                        Bento.objects.attach(punch);
                        data.attackState = 3;
                    }
                    // check if about to land
                    if (data.movable.velocityZ > 0 && data.movable.height > -10) {
                        data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-land', function () {
                            data.setState('idle');
                        });
                    }
                } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-land' &&
                    data.sprite.getCurrentFrame() >= 1
                ) {
                    invincible = false;
                }
                data.punchable.setInvincible(invincible);
            }
        },
        bite: {
            start: function (data) {
                var baseAttack = data.stats.attack;
                data.sprite.setSpriteSheet(data.rootPath + 'bite', function () {
                    data.setState('idle');
                });

                data.combo = 0;

                Bento.audio.stopSound('sfx_kick');
                Bento.audio.playSound('sfx_kick');
                Bento.audio.stopSound('rad/sfx_special');
                Bento.audio.playSound('rad/sfx_special');
                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    stamina: 0.334,
                    skill: data.stats.skill,
                    multiplier: BuffManager.rollCritical('comboEnd'),
                    pushBack: new Vector2(0.75 * BuffManager.rollCritical('knockback'), 2.0),
                    boundingBox: new Rectangle(0, -5, 28, 10),
                    offset: new Vector2(16, 0),
                    freeze: 5,
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1-medium',
                        sfx: 'sfx_punch3',
                        taptic: 'medium',
                        onComplete: function () {
                            data.movable.velocity.x = 1 * data.multiScale.direction.x;
                        }
                    })
                });
            },
            update: function (data) {
                if (data.punch && data.sprite.getCurrentFrame() > 7) {
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                    data.movable.velocity.x = 1.5 * data.multiScale.direction.x;
                }
            }
        },
        superAttack: {
            start: function (data, transitionData) {
                if (!data.isRemote) {
                    EventSystem.fire('quest-doSuperAttack', {
                        count: 1
                    });
                }
                if (!data.isRemote) {
                    var enemies = [];
                    // check for enemies on screen
                    var screenBB = Bento.getViewport().clone();
                    //shrink screen bounding box by players boundingbox
                    screenBB.grow(-data.movable.boundingBox.width, false, true);
                    Utils.forEach(Bento.objects.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                        var ignoreEnemy = enemy.family.indexOf('ignoreEnemy');

                        // check if dead
                        enemy.getComponent('punchable', function (punchable) {
                            if (punchable.isDead || punchable.getHealth() <= 0) {
                                ignoreEnemy = 1;
                            }
                        });

                        enemy.getComponent('movable', function (movable) {
                            if (!movable.grounded) {
                                ignoreEnemy = 1;
                            }
                        });

                        if (Math.abs(enemy.position.x - data.entity.position.x) < 90 &&
                            screenBB.hasPosition(enemy.position) &&
                            ignoreEnemy === -1
                        ) {
                            enemies.push(enemy);
                        }
                    });

                    // target enemies in front, with highest health
                    var health = 0;
                    var hp = 0;
                    var target;
                    var sign = data.multiScale.direction.x;
                    var range = 40;
                    var distance = 90;
                    Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                        if (Utils.sign(enemy.position.x - data.entity.position.x) === sign &&
                            Math.abs(enemy.position.y - data.entity.position.y) < range
                        ) {
                            if (Math.abs(data.entity.position.x - enemy.position.x) < distance) {
                                target = enemy;
                                distance = Math.abs(data.entity.position.x - enemy.position.x);
                            }
                        }
                    });

                    // sync data for remotes
                    data.sync({
                        target: target ? target.name : '',
                        superAttack: target ? true : false
                    });

                    if (target) {
                        Bento.objects.get('healthBar', function (healthBar) {
                            healthBar.depleteComboBar();
                        });
                    }

                    data.entity.getComponent('weapon', function (weapon) {
                        weapon.getComponent('weapon').drop(0);
                    });
                }

                if (data.syncData.superAttack) {
                    data.attackState = 0;
                    if (!data.isRemote) {
                        data.punchable.setInvincible(true);
                        EventSystem.fire('superAttack');
                    }
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    data.movable.settings.ignoreSolids = true;
                    Bento.objects.get('exit', function (exit) {
                        exit.getComponent('trigger', function (trigger) {
                            trigger.active = false;
                        });
                    });
                } else {
                    data.setState('superAttackDud');
                }
                data.currentTarget = null;
            },
            update: function (data) {
                var target = data.currentTarget;
                if (!target) {
                    // find and cache the target
                    data.currentTarget = Bento.objects.get(data.syncData.target);
                    target = data.currentTarget;
                }
                if (!target) {
                    // target could not be found
                    data.setState('idle');
                    return;
                }

                //set invincible every update
                if (!data.isRemote) {
                    data.punchable.setInvincible(true);
                }

                // align with target
                if (data.attackState === 0) {
                    target.getComponent('stateMachine', function (stateMachine) {
                        stateMachine.setState('targetSuperAttack');
                    });

                    data.sprite.setSpriteSheet(data.rootPath + 'walk');

                    var xSign = Utils.sign(target.position.x - data.entity.position.x);
                    if (xSign === 0) {
                        xSign = 1 + Utils.getRandom(2) * -2;
                    }

                    if (!data.isRemote) {
                        data.blackBar = new SuperAttackBars({});
                        Bento.objects.attach(data.blackBar);

                        //lower music volume
                        new Tween({
                            from: 1,
                            to: 0.3,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    // clamp position inside playable area
                    var targetPoint = target.position.add(new Vector2(-xSign * 100, 0));
                    var rightSide = Bento.screens.getCurrentScreen().dimension.width - 10;
                    if (Blockade.active) {
                        rightSide = Blockade.active.position.x;
                    }
                    targetPoint.x = Utils.clamp(10, targetPoint.x, rightSide);

                    new Tween({
                        from: data.entity.position,
                        to: targetPoint,
                        in: 30,
                        delay: 0,
                        ease: 'linear',
                        onStart: function () {},
                        onUpdate: function (v, t) {
                            data.entity.position = v;
                            data.multiScale.direction.x = xSign;
                        },
                        onComplete: function () {
                            data.sprite.setSpriteSheet(data.rootPath + 'idle');
                            // Utils.timeout(30, function () {
                            data.attackState = 2;

                            if (!data.isRemote) {
                                // set layers for player and target to be above everything else
                                data.entity.removeByName('sort');
                                data.entity.z = Globals.Layers.SUPERATTACK;
                                target.removeByName('sort');
                                target.z = Globals.Layers.SUPERATTACK - 2;
                                // force camera between player and enemy
                                Bento.objects.get('camera', function (camera) {
                                    camera.getComponent('cameraBehavior', function (camBehavior) {
                                        camBehavior.target = {
                                            position: new Vector2(data.entity.position.x * 0.5 + target.position.x * 0.5, data.entity.position.y * 0.5 + target.position.y * 0.5)
                                        };
                                    });
                                });
                            }

                            var effect = new Effect({
                                position: data.entity.position.add(new Vector2(-xSign * 4, -16)),
                                spriteSheet: 'effects/superattack-chargefx'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);

                            // set player sprite and add fx
                            Utils.timeout(2, function () {
                                //play supercharge sound
                                Bento.audio.stopSound('sfx_super-charge');
                                Bento.audio.playSound('sfx_super-charge');


                                data.sprite.setSpriteSheet(data.rootPath + 'chargeup', function () {
                                    data.sprite.setSpriteSheet(data.rootPath + 'chargeup-loop');
                                });
                            });

                            Utils.timeout(42, function () {
                                // Rad dives underground
                                Bento.audio.stopSound('sfx_dive');
                                Bento.audio.playSound('sfx_dive');
                                data.sprite.setSpriteSheet(data.rootPath + 'superattack-dive', function () {
                                    data.sprite.setSpriteSheet(data.rootPath + 'superattack-fin');
                                    data.attackState = 3;
                                    var deg = 0;
                                    if (data.multiScale.direction.x > 0) {
                                        deg = 180;
                                    }
                                    data.circleAngle = Utils.toRadian(deg);
                                    data.circleSpeed = 0.4;
                                    data.circlePosition = new Vector2(0, 0);
                                    data.circleTravelled = 0;
                                });
                            });
                        }
                    });
                    data.attackState = 1;
                } else if (data.attackState === 3) {
                    // Rad's fin starts circling around the enemy
                    var distanceTravelled = data.circleSpeed * 5 / 100;
                    // add traveled distance to current angle
                    data.circleAngle += distanceTravelled;
                    data.circleTravelled += distanceTravelled;
                    // calculate new circle position
                    data.circlePosition.x = Math.cos(data.circleAngle);
                    data.circlePosition.y = Math.sin(data.circleAngle);
                    // shrink y position
                    data.circlePosition.y *= 0.2;
                    var diff = data.circlePosition.scalarMultiplyWith(100);
                    diff.addTo(target.position);
                    // add to speed
                    data.circleSpeed = Math.min(data.circleSpeed + 0.05, 1);
                    // lerp position to circle position
                    diff.subtractFrom(data.entity.position);
                    data.entity.position.x += 0.8 * data.circleSpeed * diff.x;
                    data.entity.position.y += 0.8 * data.circleSpeed * diff.y;

                    // set layers
                    if (!data.isRemote) {
                        if (data.entity.position.y < target.position.y) {
                            data.entity.z = Globals.Layers.SUPERATTACK - 2;
                            target.z = Globals.Layers.SUPERATTACK;
                        } else {
                            data.entity.z = Globals.Layers.SUPERATTACK;
                            target.z = Globals.Layers.SUPERATTACK - 2;
                        }
                    }

                    // set multiscale
                    var sign = Utils.sign(diff.x);
                    if (sign !== 0) {
                        data.multiScale.direction.x = sign;
                    }

                    // after circling around the enemy 1 time move in to attack
                    if (data.circleTravelled >= 2 * Math.PI) {
                        data.attackState = 4;
                        // set layers
                        data.entity.z = Globals.Layers.SUPERATTACK;
                        target.z = Globals.Layers.SUPERATTACK - 2;
                        // align player and target
                        data.entity.position.y = target.position.y;
                        Utils.timeout(8, function () {

                            if (!data.isRemote) {
                                Bento.objects.get('camera', function (camera) {
                                    camera.shake(5, 10);
                                });
                            }

                            Bento.audio.stopSound('sfx_boulder');
                            Bento.audio.playSound('sfx_boulder');
                            data.sprite.setSpriteSheet(data.rootPath + 'superattack-breach', function () {
                                data.sprite.setSpriteSheet(data.rootPath + 'superattack-fin');
                                data.attackState = 7;
                            });
                        });
                        data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                        Utils.timeout(20, function () {
                            data.attackState = 5;
                        });
                        // save target position so that we can return to it after the breach
                        data.circlePosition = target.position.clone();
                        new Tween({
                            from: data.entity.position.x,
                            to: data.entity.position.x + (target.position.x - data.entity.position.x) * 2,
                            in: 52,
                            delay: 0,
                            ease: 'easeInOutQuad',
                            onUpdate: function (v, t) {
                                data.entity.position.x = v;
                            }
                        });
                    }
                } else if (data.attackState === 5) {
                    // breach and hit target
                    if (!data.isRemote) {
                        Bento.objects.attach(new SuperAttackSpeedlines({
                            actors: [
                                data.entity,
                                target
                            ],
                            rootPath: data.rootPath
                        }));
                        data.blackBar.removeSelf();
                        data.blackBar = undefined;

                        Bento.objects.get('camera', function (camera) {
                            camera.shake(20, 10);
                        });
                    }

                    var effect = new Effect({
                        position: target.position.add(new Vector2(0, -12)),
                        spriteSheet: 'effects/hitfx1'
                    });
                    if (!data.isRemote) {
                        effect.z = Globals.Layers.SUPERATTACK - 1;
                    }
                    Bento.objects.add(effect);
                    effect = new Effect({
                        position: target.position.add(new Vector2(0, -12)),
                        spriteSheet: 'effects/superattack-hitfx'
                    });
                    if (!data.isRemote) {
                        effect.z = Globals.Layers.SUPERATTACK - 1;
                    }
                    Bento.objects.add(effect);

                    Utils.timeout(10, function () {
                        if (!data.isRemote) {
                            target.getComponent('stateMachine', function (stateMachine) {
                                stateMachine.data.punchable.setInvincible(false);
                                stateMachine.data.punchable.hit(data.createPunch({
                                    target: ['enemy', 'destructible'],
                                    attack: data.stats.attack,
                                    stamina: 1,
                                    skill: Infinity,
                                    multiplier: 15,
                                    pushBack: new Vector2(3, 2.0),
                                    boundingBox: new Rectangle(0, -5, 16, 10),
                                    offset: new Vector2(8, 0),
                                    onHit: Punch.getDefaultCallback({
                                        entity: data.entity,
                                        effect: 'effects/hitfx1',
                                        sfx: 'sfx_finalpunch',
                                        taptic: 'heavy',
                                        onComplete: function () {
                                            data.movable.velocity.x = 1 * data.multiScale.direction.x;
                                        }
                                    }),
                                    type: 'superAttack',
                                    ignoreFlip: true
                                }));
                            });

                            TapticEngine.impact({
                                style: 'heavy'
                            });
                        }
                    });

                    if (!data.isRemote) {
                        new Tween({
                            from: 0.3,
                            to: 1,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    Bento.audio.stopSound('rad/sfx_special');
                    Bento.audio.playSound('rad/sfx_special');


                    data.attackState = 6;
                } else if (data.attackState === 7) {
                    // return to target position and emerge
                    new Tween({
                        from: data.entity.position.x,
                        to: data.circlePosition.x,
                        in: 30,
                        delay: 2,
                        ease: 'easeInOutQuad',
                        onStart: function () {
                            data.multiScale.direction.x = -data.multiScale.direction.x;
                        },
                        onUpdate: function (v, t) {
                            data.entity.position.x = v;
                        },
                        onComplete: function () {
                            Bento.audio.stopSound('sfx_dive');
                            Bento.audio.playSound('sfx_dive');
                            data.sprite.setSpriteSheet(data.rootPath + 'superattack-emerge', function () {
                                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                                data.setState('idle');

                                if (!data.isRemote) {
                                    Bento.objects.get('camera', function (camera) {
                                        camera.getComponent('cameraBehavior', function (camBehavior) {
                                            camBehavior.target = data.entity;
                                        });
                                    });
                                }
                            });
                        }
                    });
                    data.attackState = 8;
                }
            },
            end: function (data) {
                if (!data.isRemote) {
                    Utils.timeout(80, function () {
                        data.punchable.setInvincible(false);
                    });
                }
                data.movable.settings.ignoreSolids = false;

                Bento.objects.get('exit', function (exit) {
                    exit.getComponent('trigger', function (trigger) {
                        trigger.active = true;
                    });
                });
            }
        },
    };
    return Rad;
});