/**
 * Playable character: Val
 * @moduleName PlayerVal
 */
bento.define('entities/players/val', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/tapticengine',
    'entities/effect',
    'components/sort',
    'entities/player',
    'managers/stats',
    'components/stats',
    'entities/punch',
    'managers/buff',
    'states/standardplayer',
    'gui/super/superattackbars',
    'gui/super/superattackspeedlines',
    'entities/triggers/blockade',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    TapticEngine,
    Effect,
    Sort,
    Player,
    StatsManager,
    Stats,
    Punch,
    BuffManager,
    StandardPlayerStates,
    SuperAttackBars,
    SuperAttackSpeedlines,
    Blockade,
    Globals
) {
    'use strict';
    var Val = function (settings) {
        // WARNING: it is assumed this component comes AFTER PlayerBehavior,
        // because the playerBehavior adds the standard player states
        var addStates = {
            name: 'extraStates',
            start: function (data) {
                var stateMachine = data.entity.getComponent('stateMachine');
                stateMachine.addState('runAttack', Val.States.lariat);
                stateMachine.addState('chargeAttack', Val.States.whirlwind);
                stateMachine.addState('comboEnd', Val.States.uppercut);
                stateMachine.addState('jumpAttack', Val.States.jumpAttack);
                stateMachine.addState('jump', Val.States.jump);
                stateMachine.addState('superAttack', Val.States.superAttack);
                stateMachine.addState('grab', Val.States.grab);
                stateMachine.addState('specialGrab', Val.States.specialGrab);
            }
        };

        var behavior = {
            name: 'valBehavior',
            start: function () {
                val.getComponent('stateMachine', function (stateMachine) {
                    stateMachine.data.punchReset = [2, 2, 3, 2];
                    stateMachine.data.punchLength = [22, 22, 22, 22];
                });
            }
        };

        var val = new Player({
            rootPath: 'val/',
            healthBar: settings.healthBar,
            position: settings.position,
            family: ['player'],
            online: settings.online,
            name: settings.name,
            id: settings.id,
            stats: StatsManager.getAllStats('val'),
            health: settings.health,
            components: [
                addStates,
                behavior
            ]
        });

        return val;
    };

    /**
     * Pluggable states
     */
    Val.States = {
        grab: {
            start: StandardPlayerStates.grab.start,
            update: StandardPlayerStates.grab.update,
            action: function (data, action) {
                var grabbed = data.currentlyGrabbed;
                // we won't make any punch object because this is a special case
                // -the enemy should not go into another state or move
                // -other enemies should not be able to get hit
                if (action.type === 'tap' && !data.punch && grabbed) {
                    data.setState('grabAttack', {
                        grabbed: grabbed.name
                    });
                } else if (action.type === 'swipe' && !data.punch) {
                    data.setState('specialGrab', {
                        direction: Utils.sign(action.data.direction.x || data.multiScale.direction.x),
                        grabbed: grabbed.name
                    });
                }

            },
            end: StandardPlayerStates.grab.end
        },
        specialGrab: {
            start: function (data, transitionData) {
                data.currentlyGrabbed = Bento.objects.get(transitionData.grabbed);
                data.sprite.setSpriteSheet(data.rootPath + 'grabthrow', function () {
                    data.setState('idle');
                });
                data.sync({
                    throwDirection: transitionData.direction
                });
                data.throwDirection = data.syncData.throwDirection;
                data.attackState = 0;
            },
            update: function (data) {
                data.multiScale.direction.x = data.throwDirection;
                var grabbed = data.currentlyGrabbed;
                if (!grabbed) {
                    return;
                }

                var grabData = data.sprite.animationSettings.data;
                var frames = data.sprite.animationSettings.data.weapon.frames;
                var currentFrame = Math.floor(data.sprite.getCurrentFrame());
                if (data.sprite.getCurrentFrame() < 4.5) {
                    // set enemy position according to animation
                    if (frames.length && currentFrame < frames.length) {
                        grabbed.position.x = data.entity.position.x + (frames[currentFrame].x - data.sprite.animationSettings.origin.x) * data.multiScale.direction.x;
                    }
                    if (currentFrame < 3) {
                        grabbed.position.y = data.entity.position.y - 1;
                    } else {
                        grabbed.position.y = data.entity.position.y + 1;
                    }
                } else if (data.attackState === 0) {
                    Bento.audio.stopSound('sfx_kick');
                    Bento.audio.playSound('sfx_kick');
                    Bento.audio.stopSound(data.rootPath + 'sfx_attack');
                    Bento.audio.playSound(data.rootPath + 'sfx_attack');

                    Bento.audio.stopSound(data.rootPath + 'sfx_throw');
                    Bento.audio.playSound(data.rootPath + 'sfx_throw');
                    grabbed.position = data.entity.position.clone();

                    if (grabbed && !data.entity.isRemote) {
                        data.currentlyGrabbed = null; // TODO use release?
                        grabbed.getComponent('stateMachine', function (stateMachine) {
                            stateMachine.setState('grabbedThrow', {
                                faintTime: 10,
                                velocity: {
                                    x: 0.1 * data.throwDirection,
                                    y: 0,
                                    z: -2
                                },
                            });
                        });

                        grabbed.getComponent('grabable', function (grabable) {
                            grabable.release(data.entity);
                        });
                        data.setState('grabThrow');
                    }
                }
            }
        },
        grabThrow: {
            start: function () {},
            update: function (data) {
                data.multiScale.direction.x = data.throwDirection;
            }
        },
        lariat: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'lariat', function () {
                    data.setState('idle');
                });
                data.specialStage = transitionData.specialStage;
                data.punchState = 0;

                Bento.audio.stopSound('sfx_dashattack');
                Bento.audio.playSound('sfx_dashattack');

                Bento.audio.stopSound('val/sfx_dash');
                Bento.audio.playSound('val/sfx_dash');
            },
            update: function (data) {
                var speed = [1, 0];
                data.movable.velocity.x = 2 * data.multiScale.direction.x;

                //punches are created first for auto-target purposes, and added later for animation purposes
                var baseAttack = data.stats.attack;
                var pushHeight = [1.5, 2.5];
                var freeze = [7, 10];
                var punch;

                if (data.punchState === 0 && data.sprite.getCurrentFrame() > 2) {
                    punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: baseAttack,
                        stamina: 1,
                        skill: data.stats.skill,
                        duration: 10,
                        multiplier: 1.5 * BuffManager.rollCritical('dashAttack'),
                        freeze: 5,
                        pushBack: new Vector2(1.2 * BuffManager.rollCritical('knockback'), 1),
                        boundingBox: new Rectangle(0, -5, 20, 10),
                        offset: new Vector2(-10, 0),
                        ignoreFlip: true,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1',
                            sfx: 'sfx_special-hit',
                            taptic: 'heavy',
                            onComplete: function () {
                                data.movable.velocity.x *= 0.5;
                            }
                        }),
                        type: 'dashAttack'
                    });
                    data.movable.velocity.x = 0.5 * data.multiScale.direction.x;
                    Bento.objects.attach(punch);
                    data.punchState = true;
                } else if (data.sprite.getCurrentFrame() > 6) {
                    data.movable.velocity.x = 0.5 * data.multiScale.direction.x;
                }
                if (data.entity.timer % 4 === 0) {
                    var puff = new Effect({
                        position: data.entity.position.add(new Vector2(
                            data.multiScale.direction.x * -16,
                            0
                        )),
                        spriteSheet: 'effects/walkcloud'
                    });
                    puff.scale.x = data.multiScale.direction.x;
                    puff.attach(new Sort({}));
                    Bento.objects.attach(puff);
                }
            },
            end: function (data) {
                data.specialStage = 0;
            }
        },
        whirlwind: {
            start: function (data, transitionData) {
                data.isLanding = false;
                // TODO
                data.sprite.setSpriteSheet(data.rootPath + 'whirlwind', function () {

                });
                data.specialStage = transitionData.specialStage;



                Bento.audio.stopSound('sfx_special');
                Bento.audio.playSound('sfx_special');

                Bento.audio.stopSound('val/sfx_special');
                Bento.audio.playSound('val/sfx_special');

                data.punchState = 0;
                data.wwGravity = 0.08 * data.multiScale.direction.x;
                data.playerBehavior.decelaration = 0;
                data.movable.velocity.x = 3 * data.multiScale.direction.x;
            },
            update: function (data) {
                if (data.punchState % 40 === 0) {
                    var baseAttack = data.stats.attack;
                    var multiplier = 2 * BuffManager.rollCritical('chargeAttack');
                    var bonusAttack = [0, 2];
                    var pushHeight = [0.8, 1];
                    var freeze = [7, 10];
                    //create punch
                    var punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: Math.ceil(baseAttack / 2),
                        stamina: 0.334,
                        skill: data.stats.skill,
                        duration: 40,
                        ignoreFlip: true,
                        multiplier: multiplier,
                        bonusAttack: bonusAttack[data.specialStage - 1],
                        pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), pushHeight[data.specialStage - 1]),
                        boundingBox: new Rectangle(0, -5, 28, 10),
                        freeze: 4,
                        offset: new Vector2(0, 0),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-medium',
                            sfx: 'sfx_punch0',
                            taptic: 'medium'
                        }),
                        type: 'specialAttack'
                    });
                    Bento.objects.attach(punch);
                    //create second punch
                    punch = data.createPunch({
                        target: ['enemy', 'destructible'],
                        attack: Math.ceil(baseAttack / 2),
                        stamina: 0.334,
                        skill: data.stats.skill,
                        duration: 40,
                        ignoreFlip: true,
                        multiplier: multiplier,
                        bonusAttack: bonusAttack[data.specialStage - 1],
                        pushBack: new Vector2(-0.5 * BuffManager.rollCritical('knockback'), pushHeight[data.specialStage - 1]),
                        boundingBox: new Rectangle(0, -5, 28, 10),
                        offset: new Vector2(-28, 0),
                        freeze: 4,
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx1-medium',
                            sfx: 'sfx_punch0',
                            taptic: 'medium'
                        }),
                        type: 'specialAttack'
                    });
                    Bento.objects.attach(punch);
                }

                data.movable.velocity.x -= data.wwGravity;
                if (Math.abs(data.movable.velocity.x) > 3.4) {
                    data.movable.velocity.x = Utils.sign(data.movable.velocity.x) * 3.4;
                }

                ++data.punchState;
                if (data.punchState >= 80) {
                    data.setState('idle');
                }
            },
            action: function (data, action) {
                switch (action.type) {
                case 'tap':
                    if (data.isLanding) {
                        // data.setState('airPunch');
                    }
                    break;
                }
            },
            end: function (data) {
                data.isLanding = false;
                data.specialStage = 0;
                data.playerBehavior.decelaration = 0.2;
            }
        },
        jumpAttack: {
            start: function (data, transitionData) {
                EventSystem.fire('playerJumpAttack');
                data.movable.velocity.x = 0;
                data.movable.velocity.y = 0;
                var direction = transitionData.cacheDirection;
                data.multiScale.direction.x = direction;
                data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-windup', function () {
                    Bento.audio.stopSound('sfx_dashattack');
                    Bento.audio.playSound('sfx_dashattack');
                    Bento.audio.stopSound('val/sfx_throw');
                    Bento.audio.playSound('val/sfx_throw');
                    data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-loop');
                    data.movable.velocityZ = -2.5;
                    data.movable.velocity.x = direction * 3;
                    data.attackState = 1;
                    data.punchable.setInvincible(true);
                    var effect = new Effect({
                        z: Globals.Layers.ENEMIES,
                        position: data.entity.position.add(new Vector2(0, -8)),
                        spriteSheet: 'effects/jump-dust'
                    });
                    Bento.objects.attach(effect);
                });
                data.attackState = 0;
                data.punchable.setInvincible(false);
            },
            update: function (data) {
                var invincible = true;
                if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-loop') {
                    if (data.attackState === 1 && data.sprite.getCurrentFrame() >= 0) {
                        // add punch
                        data.punch = data.createPunch({
                            target: ['enemy', 'destructible'],
                            attack: data.stats.attack,
                            stamina: 0.5,
                            skill: data.stats.skill,
                            multiplier: data.stats.multiplier,
                            bonusAttack: 0,
                            freeze: 7,
                            duration: 25,
                            height: 32,
                            pushBack: new Vector2(0.5 * BuffManager.rollCritical('knockback'), 0.8),
                            boundingBox: new Rectangle(0, -5, 28, 10),
                            offset: new Vector2(0, 0),
                            ignoreFlip: true,
                            onHit: Punch.getDefaultCallback({
                                entity: data.entity,
                                effect: 'effects/hitfx1-medium',
                                sfx: 'sfx_punch3',
                                taptic: 'medium'
                            }),
                            type: 'jumpAttack'
                        });
                        Bento.objects.attach(data.punch);
                        data.attackState = 2;
                    }
                    // check if about to land
                    if (data.movable.velocityZ > 0 && data.movable.height > -10) {
                        data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-land', function () {
                            data.setState('idle');
                        });
                    }
                } else if (data.sprite.currentSpriteSheet === data.rootPath + 'jumpattack-land' &&
                    data.sprite.getCurrentFrame() >= 1
                ) {
                    invincible = false;
                }
                data.punchable.setInvincible(invincible);
            }
        },
        jump: {
            start: StandardPlayerStates.jump.start,
            update: function (data) {
                if (data.movable.grounded) {
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    data.setState('idle');
                } else {

                    if (data.movable.velocityZ > 0) {
                        if (data.movable.velocityZ < 0.8) {
                            data.sprite.setSpriteSheet(data.rootPath + 'jump-top');
                        } else {
                            data.sprite.setSpriteSheet(data.rootPath + 'jump-down');
                        }
                    }
                }
            }
        },
        uppercut: {
            start: function (data) {
                var baseAttack = data.stats.attack;
                data.sprite.setSpriteSheet(data.rootPath + 'uppercut', function () {
                    data.setState('idle');
                });

                data.combo = 0;

                Bento.audio.stopSound('sfx_kick');
                Bento.audio.playSound('sfx_kick');
                Bento.audio.stopSound('val/sfx_dash');
                Bento.audio.playSound('val/sfx_dash');

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    stamina: 1,
                    skill: data.stats.skill,
                    multiplier: 1.25 + BuffManager.rollCritical('comboEnd'),
                    pushBack: new Vector2(0.75 * BuffManager.rollCritical('knockback'), 2.0),
                    boundingBox: new Rectangle(0, -5, 24, 10),
                    freeze: 5,
                    offset: new Vector2(8, 0),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1-medium',
                        sfx: 'sfx_punch3',
                        taptic: 'medium',
                        onComplete: function () {
                            data.movable.velocity.x = 1 * data.multiScale.direction.x;
                        }
                    })
                });
            },
            update: function (data) {
                if (data.punch && data.sprite.getCurrentFrame() > 3) {
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                    data.movable.velocity.x = 1.3 * data.multiScale.direction.x;
                }
            }
        },
        superAttack: {
            start: function (data, transitionData) {
                if (!data.isRemote) {
                    EventSystem.fire('quest-doSuperAttack', {
                        count: 1
                    });
                }
                if (!data.isRemote) {
                    var enemies = [];
                    // check for enemies on screen
                    var screenBB = Bento.getViewport().clone();
                    //shrink screen bounding box by players boundingbox
                    screenBB.grow(-data.movable.boundingBox.width, false, true);
                    Utils.forEach(Bento.objects.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                        var ignoreEnemy = enemy.family.indexOf('ignoreEnemy');

                        // check if dead
                        enemy.getComponent('punchable', function (punchable) {
                            if (punchable.isDead || punchable.getHealth() <= 0) {
                                ignoreEnemy = 1;
                            }
                        });

                        enemy.getComponent('movable', function (movable) {
                            if (!movable.grounded) {
                                ignoreEnemy = 1;
                            }
                        });

                        if (Math.abs(enemy.position.x - data.entity.position.x) < 90 &&
                            screenBB.hasPosition(enemy.position) &&
                            ignoreEnemy === -1
                        ) {
                            enemies.push(enemy);
                        }
                    });

                    // target enemies in front, with highest health
                    var health = 0;
                    var hp = 0;
                    var target;
                    var sign = data.multiScale.direction.x;
                    var range = 40;
                    var distance = 90;
                    Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                        if (Utils.sign(enemy.position.x - data.entity.position.x) === sign &&
                            Math.abs(enemy.position.y - data.entity.position.y) < range
                        ) {
                            if (Math.abs(data.entity.position.x - enemy.position.x) < distance) {
                                target = enemy;
                                distance = Math.abs(data.entity.position.x - enemy.position.x);
                            }
                        }
                    });

                    // sync data for remotes
                    data.sync({
                        target: target ? target.name : '',
                        superAttack: target ? true : false
                    });

                    if (target) {
                        Bento.objects.get('healthBar', function (healthBar) {
                            healthBar.depleteComboBar();
                        });
                    }

                    data.entity.getComponent('weapon', function (weapon) {
                        weapon.getComponent('weapon').drop(0);
                    });
                }

                if (data.syncData.superAttack) {
                    data.attackState = 0;
                    data.spawnFx = 0;
                    if (!data.isRemote) {
                        data.punchable.setInvincible(true);
                        EventSystem.fire('superAttack');
                    }
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    data.movable.settings.ignoreSolids = true;
                } else {
                    data.setState('superAttackDud');
                }
                data.currentTarget = null;
            },
            update: function (data) {
                var target = data.currentTarget;
                if (!target) {
                    // find and cache the target
                    data.currentTarget = Bento.objects.get(data.syncData.target);
                    target = data.currentTarget;
                }
                if (!target) {
                    // target could not be found
                    data.setState('idle');
                    return;
                }

                //set invincible every update
                if (!data.isRemote) {
                    data.punchable.setInvincible(true);
                }

                //piledriver enemy position
                var pileDriver = function () {
                    var grabData = data.sprite.animationSettings.data;
                    var frames = data.sprite.animationSettings.data.weapon.frames;
                    var currentFrame = Math.floor(data.sprite.getCurrentFrame());

                    if (frames.length && currentFrame < frames.length) {
                        target.position.x = data.entity.position.x + (frames[currentFrame].x - data.sprite.animationSettings.origin.x) * data.multiScale.direction.x;
                    }

                    if (currentFrame < 3) {
                        target.z = Globals.Layers.SUPERATTACK - 2;
                    } else {
                        target.z = Globals.Layers.SUPERATTACK + 1;
                    }
                };

                // align with target
                if (data.attackState === 0) {
                    target.getComponent('stateMachine', function (stateMachine) {
                        stateMachine.setState('targetSuperAttack');
                    });

                    data.sprite.setSpriteSheet(data.rootPath + 'walk');

                    var xSign = Utils.sign(target.position.x - data.entity.position.x);
                    if (xSign === 0) {
                        xSign = 1 + Utils.getRandom(2) * -2;
                    }

                    if (!data.isRemote) {
                        data.blackBar = new SuperAttackBars({});
                        Bento.objects.attach(data.blackBar);

                        //lower music volume
                        new Tween({
                            from: 1,
                            to: 0.3,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    // clamp position inside playable area
                    var targetPoint = target.position.add(new Vector2(-xSign * 100, 0));
                    var rightSide = Bento.screens.getCurrentScreen().dimension.width - 10;
                    if (Blockade.active) {
                        rightSide = Blockade.active.position.x;
                    }
                    targetPoint.x = Utils.clamp(10, targetPoint.x, rightSide);

                    new Tween({
                        from: data.entity.position,
                        to: targetPoint,
                        in: 30,
                        delay: 0,
                        ease: 'linear',
                        onStart: function () {},
                        onUpdate: function (v, t) {
                            data.entity.position = v;
                            data.multiScale.direction.x = xSign;
                        },
                        onComplete: function () {
                            data.sprite.setSpriteSheet(data.rootPath + 'idle');
                            // Utils.timeout(30, function () {
                            data.attackState = 2;

                            if (!data.isRemote) {
                                // set layers for player and target to be above everything else
                                data.entity.removeByName('sort');
                                data.entity.z = Globals.Layers.SUPERATTACK;
                                target.removeByName('sort');
                                target.z = Globals.Layers.SUPERATTACK - 2;
                                // force camera between player and enemy
                                Bento.objects.get('camera', function (camera) {
                                    camera.getComponent('cameraBehavior', function (camBehavior) {
                                        camBehavior.target = {
                                            position: new Vector2(data.entity.position.x * 0.5 + target.position.x * 0.5, data.entity.position.y * 0.5 + target.position.y * 0.5)
                                        };
                                    });
                                });
                            }

                            var effect = new Effect({
                                position: data.entity.position.add(new Vector2(-xSign * 4, -16)),
                                spriteSheet: 'effects/superattack-chargefx'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);

                            // set player sprite and add fx
                            Utils.timeout(2, function () {
                                //play supercharge sound
                                Bento.audio.stopSound('sfx_super-charge');
                                Bento.audio.playSound('sfx_super-charge');
                                data.sprite.setSpriteSheet(data.rootPath + 'jumpattack-windup');
                            });

                            Utils.timeout(42, function () {
                                data.sprite.setSpriteSheet(data.rootPath + 'grabbing');
                                Bento.audio.stopSound('sfx_special-hit');
                                Bento.audio.playSound('sfx_special-hit');
                                data.attackState = 3;
                            });
                            // });
                        }
                    });
                    data.attackState = 1;
                } else if (data.attackState === 3) {
                    // dash towards target
                    data.attackState = 4;

                    var enemyRange = 24;
                    target.getComponent('punchable', function (punchable) {
                        enemyRange = punchable.boundingBox.width / 2 + 14;
                    });
                    enemyRange *= 0.7;
                    new Tween({
                        from: data.entity.position.x,
                        to: target.position.x + enemyRange * Utils.sign(data.entity.position.x - target.position.x),
                        in: 10,
                        delay: 0,
                        ease: 'linear',
                        onStart: function () {
                            var effect = new Effect({
                                position: data.entity.position.add(new Vector2(Utils.sign(target.position.x - data.entity.position.x) * 40, 0)),
                                spriteSheet: 'effects/superattack-dashfx'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);
                        },
                        onUpdate: function (v, t) {
                            data.entity.position.x = v;
                        },
                        onComplete: function () {
                            Utils.timeout(5, function () {
                                // arrived at target, show hit fx
                                var effect = new Effect({
                                    position: target.position.add(new Vector2(0, -12)),
                                    spriteSheet: 'effects/superattack-hitfx'
                                });
                                if (!data.isRemote) {
                                    effect.z = Globals.Layers.SUPERATTACK - 1;
                                    TapticEngine.impact({
                                        style: 'heavy'
                                    });
                                }
                                Bento.objects.add(effect);

                                if (!data.isRemote) {
                                    data.blackBar.flash();
                                }
                            });

                            var targetStateMachine;
                            target.getComponent('stateMachine', function (stateMachine) {
                                stateMachine.data.sprite.setSpriteSheet(stateMachine.data.rootPath + 'flyup');
                            });

                            Utils.timeout(20, function () {
                                data.attackState = 5;
                                Bento.audio.stopSound('sfx_fist');
                                Bento.audio.playSound('sfx_fist');

                                Bento.audio.stopSound('val/sfx_throw');
                                Bento.audio.playSound('val/sfx_throw');

                                data.sprite.setSpriteSheet(data.rootPath + 'superattack-piledriver');
                                data.entity.position.x = target.position.x;
                                data.entity.position.y = target.position.y;
                            });
                        }
                    });
                } else if (data.attackState === 5) {
                    // piledriver jump upwards
                    data.attackState = 6;

                    new Tween({
                        from: 0,
                        to: 1,
                        in: 100,
                        delay: 0,
                        ease: 'easeOutQuad',
                        onStart: function () {
                            var effect = new Effect({
                                position: data.entity.position.clone(),
                                spriteSheet: 'effects/superattack-dust-up'
                            });

                            if (!data.isRemote) {
                                effect.z = Globals.Layers.SUPERATTACK - 1;
                            }
                            effect.scale.x = data.multiScale.direction.x;
                            Bento.objects.add(effect);
                        },
                        onUpdate: function (v, t) {
                            data.movable.height = v * (-100 - data.entity.position.y);

                            data.shadow.shadow.alpha = 1 - v;
                            target.getComponent('stateMachine', function (targetStateMachine) {
                                targetStateMachine.data.movable.height = data.movable.height + 20;
                                targetStateMachine.data.shadow.shadow.alpha = 1 - v;
                            });

                        },
                        onComplete: function () {
                            //set to true to make sure other enemies dont walk to an empty position and start punching
                            data.entity.isDown = true;
                            // set player almost on top of target, if on top enemy multiscale becomes 0;
                            data.entity.position.x = target.position.x + Utils.sign(data.entity.position.x - target.position.x);
                            data.attackState = 7;
                            data.attackState = 8;
                            data.movable.velocityZ = 30;
                        }
                    });
                } else if (data.attackState === 6) {
                    // set target position according to piledriver animation, based on weapon animation
                    pileDriver();
                } else if (data.attackState === 7) {
                    data.movable.height = -100 - data.entity.position.y;
                    target.getComponent('stateMachine', function (targetStateMachine) {
                        targetStateMachine.data.movable.height = data.movable.height + 20;
                    });
                    pileDriver();
                } else if (data.attackState === 8) {
                    pileDriver();
                    var maxHeight = -100 - data.entity.position.y;
                    data.shadow.shadow.alpha = data.movable.height / maxHeight;

                    target.getComponent('stateMachine', function (targetStateMachine) {
                        targetStateMachine.data.movable.height = data.movable.height + 20;
                        targetStateMachine.data.shadow.shadow.alpha = data.shadow.shadow.alpha;
                    });

                    if (data.movable.height > -50 && data.spawnFx === 0) {
                        var dustGround = new Effect({
                            position: target.position.add(new Vector2(0, 0)),
                            spriteSheet: 'effects/superattack-dust-ground'
                        });
                        if (!data.isRemote) {
                            dustGround.z = Globals.Layers.SUPERATTACK - 1;
                        }
                        Bento.objects.add(dustGround);

                        data.spawnFx = 1;
                    }

                    if (data.movable.grounded) {
                        data.attackState = 9;
                        data.shadow.shadow.alpha = 1;
                        target.getComponent('stateMachine', function (targetStateMachine) {
                            targetStateMachine.data.shadow.shadow.alpha = 1;
                        });
                        data.sprite.setSpriteSheet(data.rootPath + 'getup');
                        target.position = data.entity.position.clone();
                        Bento.objects.get('camera', function (camera) {
                            camera.shake(40, 10);
                        });

                        Bento.audio.stopSound('val/sfx_special');
                        Bento.audio.playSound('val/sfx_special');
                    }
                } else if (data.attackState === 9) {
                    Utils.timeout(5, function () {

                        if (!data.isRemote) {
                            Bento.objects.attach(new SuperAttackSpeedlines({
                                actors: [
                                    data.entity,
                                    target
                                ],
                                rootPath: data.rootPath
                            }));
                            data.blackBar.removeSelf();
                            data.blackBar = undefined;


                        }

                        var effect = new Effect({
                            position: target.position.add(new Vector2(0, -12)),
                            spriteSheet: 'effects/hitfx1'
                        });
                        if (!data.isRemote) {
                            effect.z = Globals.Layers.SUPERATTACK - 1;
                        }
                        Bento.objects.add(effect);

                        //maybe damage other enemies too?

                        // Utils.timeout(10, function () {
                        if (!data.isRemote) {
                            target.getComponent('stateMachine', function (stateMachine) {
                                stateMachine.data.punchable.setInvincible(false);
                                stateMachine.data.punchable.hit(data.createPunch({
                                    target: ['enemy', 'destructible'],
                                    attack: data.stats.attack,
                                    stamina: 1,
                                    skill: Infinity,
                                    multiplier: 15,
                                    freeze: 5,
                                    pushBack: new Vector2(3, 2.0),
                                    boundingBox: new Rectangle(0, -5, 16, 10),
                                    offset: new Vector2(8, 0),
                                    onHit: Punch.getDefaultCallback({
                                        entity: data.entity,
                                        effect: 'effects/hitfx1',
                                        sfx: 'sfx_finalpunch',
                                        taptic: 'heavy',
                                        onComplete: function () {
                                            data.movable.velocity.x = 1 * data.multiScale.direction.x;
                                        }
                                    }),
                                    type: 'superAttack'
                                }));
                            });

                            TapticEngine.impact({
                                style: 'heavy'
                            });
                        }
                        // });
                    });
                    data.attackState = 10;

                    if (!data.isRemote) {
                        new Tween({
                            from: 0.3,
                            to: 1,
                            in: 20,
                            delay: 0,
                            ease: 'easeInQuad',
                            onUpdate: function (v, t) {
                                Bento.audio.setVolume(v, Globals.currentMusic);
                            },
                        });
                    }

                    Utils.timeout(40, function () {
                        if (!data.isRemote) {
                            Bento.objects.get('camera', function (camera) {
                                camera.getComponent('cameraBehavior', function (camBehavior) {
                                    camBehavior.target = data.entity;
                                });
                            });
                        }
                        data.sprite.setSpriteSheet(data.rootPath + 'getup');
                        Utils.timeout(30, function () {
                            data.sprite.setSpriteSheet(data.rootPath + 'idle');
                            data.setState('idle');
                            //set to false to make sure enemies punch again
                            data.entity.isDown = false;
                        });
                    });
                }
            },
            end: function (data) {
                if (!data.isRemote) {
                    Utils.timeout(80, function () {
                        data.punchable.setInvincible(false);
                    });
                }
                data.movable.settings.ignoreSolids = false;
            }
        }
    };
    return Val;
});