/**
 * Breakable prop, doesn't get destroyed when punched, but shows a "broken" fraae
 * @moduleName BreakableProp
 */
bento.define('entities/props/breakable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/sort',
    'components/punchable',
    'components/movable',
    'components/coindropper',
    'components/debug',
    'managers/buff',
    'components/itemdropper',
    'components/drawinviewport',
    'modules/matchmanager',
    'modules/randomizer'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Sort,
    Punchable,
    Movable,
    CoinDropper,
    Debug,
    BuffManager,
    ItemDropper,
    DrawInViewport,
    MatchManager,
    Randomizer
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var id = settings.tiled.object.id;
        var tileProperties = settings.tiled.tileProperties || {};
        var rootPath = tileProperties.rootPath || '';
        var properties = settings.tiled.object.properties || {};
        var coins = parseInt(properties.coins) || 0;
        var destroy = tileProperties.destroy || false;
        var item;
        var empty = false;
        // extra properties extracted from name
        var objectName = settings.tiled.object.name;
        var nameProperties = objectName.split(',');
        // get coins from name
        Utils.forEach(nameProperties, function (prop) {
            var values = prop.split(':');
            var key = values[0];
            var value = parseFloat(values[1]);
            if (isNaN(value)) {
                // probably a string
                value = values[1];
            }

            // coins
            if (key === 'coins') {
                coins = value;
            } else if (key === 'item') {
                item = 'consumables/' + value;
            } else if (key === 'weapon') {
                item = 'weapons/' + value;
            } else if (key === 'empty') {
                empty = true;
            }
        });
        var boundingBox = (function () {
            var coords;
            if (!tileProperties || !tileProperties.boundingBox) {
                return null;
            }
            coords = tileProperties.boundingBox.split(',');
            return new Rectangle(
                parseInt(coords[0]),
                parseInt(coords[1]),
                parseInt(coords[2]),
                parseInt(coords[3])
            );
        })();
        var sprite = new Sprite({
            spriteSheet: rootPath + 'normal'
        });
        var punchable = new Punchable({
            boundingBox: boundingBox,
            health: Infinity,
            maxHealth: Infinity,
            breakable: true
        });
        var hits = 3;
        var coinDropBehavior = {
            start: function () {
                // drop some coins based on tiled value
                var dropper;

                if (!item) {
                    if (!MatchManager.isOnline) {
                        // randomize coins locally
                        if (!coins && Utils.getRandom(3) === 0) {
                            coins = 3 + Utils.getRandom(5);
                        }
                    } else {
                        // online: deterministic random seed based on tiled name and id
                        var seed = Utils.checksum(objectName + '-' + settings.tiled.object.id);
                        var randomizer = new Randomizer({
                            seed: seed
                        });
                        if (!coins && randomizer.getRandom(3) === 0) {
                            coins = 3 + randomizer.getRandom(5);
                        }
                    }
                }
                dropper = new CoinDropper({
                    coins: coins,
                    downOnly: true,
                    ignoredFamilies: ['breakables'] // doesn't bounce against own entity
                });

                if (!coins && item) {
                    dropper = new ItemDropper({
                        item: item,
                        ignoredFamilies: ['breakables'] // doesn't bounce against own entity
                    });
                }

                if (!empty) {
                    entity.attach(dropper);
                }
            }
        };

        var entity = new Entity({
            z: 0,
            name: 'breakableProp-' + id,
            family: ['solid', 'destructible', 'breakables'],
            boundingBox: boundingBox,
            components: [
                sprite,
                punchable,
                coinDropBehavior,
                new Sort({}),
                new DrawInViewport({})
            ]
        });
        var createBroken = function () {
            // start flashing and remove entity
            entity.remove(sprite);
            entity.removeSelf();
            var removalBehavior = {
                name: 'removalBehavior',
                update: function (data) {
                    if (broken.timer % 2 >= 1) {
                        broken.visible = false;
                    } else {
                        broken.visible = true;
                    }

                    if (broken.timer > 2 * 60) {
                        broken.removeSelf();
                    }
                }
            };
            var broken = new Entity({
                name: 'brokenProp',
                position: entity.position,
                components: [
                    new Sort({}),
                    sprite,
                    removalBehavior
                ]
            });
            Bento.objects.attach(broken);
        };


        punchable.onHit(function (data) {
            // shake
            new Tween({
                from: 2,
                to: 0,
                'in': 15,
                ease: Tween.LINEAR,
                onUpdate: function (v, t) {
                    entity.transform.x = Utils.getRandomRangeFloat(-v, v);
                    entity.transform.y = Utils.getRandomRangeFloat(-v, v);
                },
                onComplete: function () {
                    entity.transform.x = 0;
                    entity.transform.y = 0;
                }
            });
            punchable.depleteHealth(data.punch);

            // sync hit
            EventSystem.fire('punchHit', {
                punch: data.punch,
                victim: entity,
                attacker: data.punch.attacker
            });

            hits -= data.punch.multiplier;
            if (hits <= 0) {
                punchable.kill(data.punch);
            }
        });
        punchable.onDeath(function (data) {
            // remove punchable
            entity.remove(punchable);
            sprite.setSpriteSheet(rootPath + 'broken');

            if (destroy && destroy !== "false") {
                createBroken();
            }

            if (!data.punch.emulated) {
                EventSystem.fire('quest-destruction', {
                    count: 1
                });
            }
        });

        return entity;
    };
});