/**
 * Punch entity for collision purposes
 * @moduleName Punch
 */
bento.define('entities/punch', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/debug',
    'entities/freezecontroller',
    'entities/effect',
    'modules/tapticengine'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Debug,
    FreezeController,
    Effect,
    TapticEngine
) {
    'use strict';
    var Punch = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var duration = settings.duration || 10;
        var hasPunched = [];

        var disabled = false;

        var defaults = Punch.getDefaults();

        var behavior = {
            name: 'behaviorComponent',
            offset: new Vector2(0, 0),
            start: function (data) {
                // keep track of parent
                if (settings.parent) {
                    this.offset = entity.position.subtract(settings.parent.position);
                }

            },
            destroy: function (data) {
                if (!hasHit && settings.onMiss) {
                    settings.onMiss();
                    EventSystem.fire('punchMiss', {
                        punch: entity,
                        attacker: settings.attacker || settings.parent
                    });
                }
            },
            update: function (data) {
                if (entity.timer > duration || disabled) {
                    Bento.objects.remove(entity);
                }
                if (settings.parent) {
                    entity.position = settings.parent.position.add(this.offset);
                }
            },
            draw: function (data) {}
        };
        var entity = new Entity({
            z: 0,
            name: 'punch',
            family: ['punches'],
            position: settings.position,
            boundingBox: settings.boundingBox,
            scale: settings.scale,
            components: [
                behavior
            ]
        });
        var hasHit = false;
        entity.extend({
            type: settings.type || '',
            attacker: settings.attacker || settings.parent,
            parentEntity: settings.parent,
            target: settings.target || defaults.target,
            attack: settings.attack || defaults.attack,
            multiplier: settings.multiplier || defaults.multiplier,
            bonusAttack: settings.bonusAttack || defaults.bonusAttack,
            stamina: settings.stamina || defaults.stamina,
            skill: settings.skill || defaults.skill,
            direction: settings.direction || defaults.direction,
            freeze: settings.freeze,
            height: settings.height || defaults.height,
            pushBack: settings.pushBack || defaults.pushBack,
            crush: settings.crush || defaults.crush,
            rememberPunched: true,
            // pass callback to settings
            onHit: function (other) {
                hasHit = true;
                if (entity.rememberPunched) {
                    hasPunched.push(other);
                }
                if (settings.onHit) {
                    settings.onHit(other);
                }
                if (entity.freeze) {
                    FreezeController.freeze(entity.freeze, 2); //adding a small delay allows multiple hits to stack their freeze instead of creating multiple freezes
                }
            },
            preHit: function (other) {
                if (settings.preHit) {
                    settings.preHit(other);
                }
            },
            hasAlreadyPunched: function (e) {
                return hasPunched.indexOf(e) > -1;
            },
            disable: function () {
                disabled = true;
            },
            /**
             * Serializes the punch data for multiplayer. If a property uses
             * its default value, that property is not sent but is instead
             * re-created from the default value on the receiving end.
             */
            serialize: function () {
                return {
                    t: this.type !== defaults.type ? this.type : undefined,
                    d: this.direction !== defaults.direction ? this.direction : undefined,
                    x: this.pushBack.x !== defaults.pushBack.x ? this.pushBack.x : undefined,
                    y: this.pushBack.y !== defaults.pushBack.y ? this.pushBack.y : undefined,
                    h: this.height !== defaults.height ? this.height : undefined,
                    a: this.attack !== defaults.attack ? this.attack : undefined,
                    m: this.multiplier !== defaults.multiplier ? this.multiplier : undefined,
                    b: this.bonusAttack !== defaults.bonusAttack ? this.bonusAttack : undefined,
                    s: this.stamina !== defaults.stamina ? this.stamina : undefined,
                    sk: this.skill !== defaults.skill ? this.skill : undefined,
                    c: this.crush !== defaults.crush ? this.crush : undefined
                };
            }
        });

        if (Punch.debug) {
            entity.attach(new Debug({}));
        }

        return entity;
    };

    Punch.getDefaults = function () {
        return {
            type: '',
            target: ['enemy', 'destructible'],
            attack: 1,
            multiplier: 1,
            bonusAttack: 0,
            direction: 1,
            height: 15,
            stamina: 0,
            skill: 0,
            // The default value for push back is set to the push back value of
            // the normal punch. This way we can cut down on the amount of data
            // sent as the value for the most common punch doesn't have to be sent.
            pushBack: new Vector2(0.5, 1.25),
            crush: false
        };
    };

    // onHit callback for punches
    Punch.getDefaultCallback = function (settings) {
        var entity = settings.entity;
        var effectName = settings.effect || 'effects/hitfx3-small';
        var sfx = settings.sfx || 'sfx_punch0';
        var taptic = settings.taptic || 'light';
        var onComplete = settings.onComplete || function () {};

        return function (other) {
            var sign = Utils.sign(entity.position.x - other.position.x);
            var distanceToOther = Math.abs(other.position.x - entity.position.x);
            if (distanceToOther < other.boundingBox.width * 0.5) {
                sign *= distanceToOther / (other.boundingBox.width * 0.5);
            }
            var randomPos = new Vector2(Utils.getRandomRange(-5, 5), Utils.getRandomRange(-10, 10));
            var effect = new Effect({
                position: new Vector2(
                    other.position.x + sign * other.boundingBox.width * 0.4 + randomPos.x,
                    entity.position.y - 16 + randomPos.y
                ),
                spriteSheet: effectName,
                family: ['hitEffects']
            });
            effect.scale.x = Utils.sign(entity.scale.x);

            Bento.objects.attach(effect);

            Bento.audio.stopSound(sfx);
            Bento.audio.playSound(sfx);
            TapticEngine.impact({
                style: taptic
            });

            onComplete();
        };
    };

    return Punch;
});