/**
 * base entity that spawns enemies
 * @moduleName Spawner
 */
bento.define('entities/spawner', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/sort',
    'components/movebackground',
    'managers/ai',
    'managers/enemyactivator',
    'entities/bossintro',
    'components/boss',
    'modules/matchmanager',
    'modules/randomizer',
    'entities/triggers/blockade'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Sort,
    MoveBackground,
    AiManager,
    EnemyActivator,
    BossIntro,
    Boss,
    MatchManager,
    Randomizer,
    Blockade
) {
    'use strict';
    var Spawner = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var tiled = settings.tiled || {};
        var properties = (tiled.object || {}).properties || {};
        var spawnProperties = Utils.nameToJson(tiled.object.name);
        var enemyQueue = {
            // "1": [enemyEntity, enemyEntity, etc..]
        };

        // deterministic randomizer
        var randomizer = new Randomizer({
            seed: tiled.object.id
        });

        var ready = false;
        var currentWave = 0;
        var bossWave = -1;
        var isRemote;
        var isListening = false;

        var readyToSpawn = false;
        var spawnQueue = 0;

        var initialized = false;
        var enemyModules = {
            length: 0
        };
        var enemyId = 0;
        var spawnId = tiled.object.id;

        var spawnEnemyModules = function (Module, nameProperties, rootPath, wave) {
            // spawn it
            var enemy = new Module({
                // we fake a tiled object to pass the name properties
                tiled: {
                    object: {
                        name: nameProperties,
                        tileProperties: properties,
                    }
                },
                rootPath: rootPath
            });

            var readName = Utils.nameToJson(nameProperties);

            enemy.name += '-spwn-' + spawnId + '-' + enemyId;
            ++enemyId;

            // bosses will get the boss component attach, any enemy could be a boss
            if (readName.boss) {
                bossWave = wave;
                enemy.attach(new Boss({
                    skipIntro: readName.boss > 1
                }));
            }

            // queue the enemy
            enemyQueue[wave] = enemyQueue[wave] || []; // initialize array if it doesnt exist
            enemyQueue[wave].push(enemy);
        };

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                // the object properties should look like this:
                // wave-[number]:[enemy module path],[rootPath],[other name properties]

                // example:
                // wave-0:entities/enemies/rat,addLv:1
                // wave-0:entities/enemies/rat
                // wave-1:entities/enemies/bigrat,rootPath:bigratviolet/,addLv:1
                // etc.

                queueUpEnemies(properties);

                if (settings.movePosition) {
                    //change position according to function
                    settings.movePosition();
                }

                // position to activate, with this you can make a previous door spawn for a blockade further ahead
                spawner.activateX = spawner.activateX || spawnProperties.activateX || spawner.position.x;

                // add event to catch if there are still enemies left
                EventSystem.on('destroyEnemyController', spawnEnemyWave);
                EventSystem.on('startRoom', onStartRoom);
                if (spawnProperties.onActivate) {
                    EventSystem.on('activateEnemies', spawnEnemyWave);
                }
            },
            destroy: function (data) {
                // remove event
                behavior.stopListening();
                EventSystem.off('destroyEnemyController', spawnEnemyWave);
                EventSystem.off('startRoom', onStartRoom);
                if (spawnProperties.onActivate) {
                    EventSystem.off('activateEnemies', spawnEnemyWave);
                }
            },
            update: function (data) {
                // always check if local player is master
                isRemote = !MatchManager.isMaster;

                if (!isRemote) {
                    if (initialized && readyToSpawn) {
                        // spawn wave and let others know
                        spawnEnemyWave();

                        MatchManager.sendData({}, spawner.name);
                    }
                }

                // isRemote changed value
                if (isRemote && !isListening) {
                    if (initialized) {
                        behavior.listenToIncoming();
                    }
                } else if (!isRemote && isListening) {
                    behavior.stopListening();
                }
            },
            // listen to incoming messages
            listenToIncoming: function () {
                MatchManager.onMessage(spawner.name, behavior.onMessage);
                isListening = true;
            },
            stopListening: function () {
                MatchManager.offMessage(spawner.name, behavior.onMessage);
                isListening = false;
            },
            onMessage: function (msg) {
                spawnEnemyWave(msg.noUpdate);
                // always return true if the message is handled
                return true;
            }

        };

        var queueUpEnemies = function (data) {
            var toLoad = 0;
                var loaded = 0;
                var checkAllLoaded = function () {
                    var i = 0;
                    var moduleData;

                    if (loaded === toLoad) {
                        if (!ready) {
                            ready = true;
                        }
                        // spawn enemies in deterministic order!
                        for (i = 0; i < enemyModules.length; ++i) {
                            moduleData = enemyModules[i];
                            spawnEnemyModules(
                                moduleData.Module,
                                moduleData.nameProperties,
                                moduleData.rootPath,
                                moduleData.wave
                            );

                            // //pre emptively fill up previous enemy wave with an array to fill up the gaps because of asynchronous spawning
                            // if(moduleData.wave > 0 && Utils.isUndefined(enemyQueue[moduleData.wave - 1])) {
                            //     enemyQueue[moduleData.wave - 1] = [];
                            // }
                        }
                        initialized = true;
                    }
                };
                var index = 0;
                Utils.forEach(data, function (propertyString, i) {
                    // split the string into the correct variables
                    var namePropertiesIndex = propertyString.search(/[,]/g);
                    var nameProperties;
                    var moduleArray;
                    var rootPath;
                    var currentIndex = index; // mind the closures! local copy of index
                    index += 1;

                    if (namePropertiesIndex > 0) {
                        moduleArray = propertyString.slice(0, namePropertiesIndex).split(':');
                        nameProperties = propertyString.slice(namePropertiesIndex + 1);

                        // get rootPath
                        var rootPathIndex = nameProperties.search('rootPath');
                        if (rootPathIndex > -1) {
                            var rootPathString = nameProperties.split(',')[0];
                            rootPath = rootPathString.split(':')[1];
                            nameProperties = nameProperties.replace('rootPath:' + rootPath, '');
                        }

                        if (!nameProperties) {
                            nameProperties = undefined;
                        }
                    } else {
                        moduleArray = propertyString.split(':');
                    }

                    var wave = moduleArray[0].replace('wave-', '');
                    var module = moduleArray[1];

                    // add to total to load number
                    toLoad += 1;

                    bento.require([module], function (Module) {
                        // collect the modules before spawning them
                        // because require is asynchronous and the order might be different
                        enemyModules[currentIndex] = {
                            Module: Module,
                            nameProperties: nameProperties,
                            rootPath: rootPath,
                            wave: wave
                        };
                        enemyModules.length += 1;
                        loaded += 1;
                        // check if everything is loaded
                        checkAllLoaded();
                    });
                });
        };

        var checkSpawn = function (data) {
            // checks if spawner is active
            // the spawner should be left of the current blockade unless the activateX value overwrites it
            var rightSide = Bento.screens.getCurrentScreen().dimension.width;
            if (Blockade.active) {
                rightSide = Blockade.active.position.x;
            }

            if (spawner.force) {
                return true;
            }

            if (spawner.activateX > rightSide) {
                return false;
            } else if (!data && !spawnProperties.onActivate) {
                // if data is undefined the destroycontroller event is fired by a barrel or some other non enemy entity
                return false;
            }

            spawnProperties.onActivate = undefined;

            // check threshold
            if (Utils.isDefined(spawnProperties.threshold)) {
                var activeEnemies = 0;
                Utils.forEach(Bento.objects.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                    enemy.getComponent('enemyController', function (enemyController) {
                        if (enemyController.active) {
                            ++activeEnemies;
                        }
                    });
                });
                if (activeEnemies > spawnProperties.threshold) {
                    return false;
                }
            }
            return true;
        };
        // calls the next group of enemies to spawn
        var spawnEnemyWave = function (data) {
            var time = 0;
            var queue = enemyQueue[currentWave];
            var canSpawn = queue ? true : false;

            if (!checkSpawn(data)) {
                return;
            }
            // spawn enemies
            if (spawner.preSpawn) {
                spawner.preSpawn(canSpawn);
            }
            if (queue) {
                Utils.forEach(queue, function (enemy, i, l, breakLoop) {
                    if (time) {
                        Utils.timeout(time, function () {
                            spawner.spawnEnemy(enemy);
                        });
                    } else {
                        // spawn asap
                        spawner.spawnEnemy(enemy);
                    }
                    time += 10;
                });
                // erase queue
                enemyQueue[currentWave] = null;
            }

            Utils.timeout(time, function () {
                if (spawner.postSpawn) {
                    spawner.postSpawn(canSpawn, currentWave);
                }
            });

            // whether it spawned enemies or not, increment the wave
            ++currentWave;

            return canSpawn;
        };
        var onStartRoom = function () {
            if (!ready) {
                ready = true;
            }
            EventSystem.off('startRoom', onStartRoom);
        };

        var spawner = new Entity({
            z: 0,
            name: 'spawner-' + spawnId,
            family: ['spawners'],
            position: settings.position,
            updateWhenPaused: 0,
            components: [
                behavior,
                new Sort({})
            ]
        }).extend({
            spawnY: 12,
            activateEnemy: true,
            spawnEnemy: function (enemy) {
                enemy.position.x = spawner.position.x;
                enemy.position.y = spawner.position.y + spawner.spawnY;
                enemy.getComponent('movable', function (movable) {
                    //ignore solids
                    movable.settings.ignoreSolids = true;
                    movable.velocityZ = -1.5;
                    movable.height = -spawner.spawnY;
                });

                Bento.objects.attach(enemy);

                enemy.getComponent('enemyController', function (enemyController) {
                    if (!enemyController.active) {
                        enemyController.active = spawner.activateEnemy;
                    }
                });

                // set jump after attaching enemy to stage otherwise state doesn't exist
                var jump = new Vector2(randomizer.getRandomRangeFloat(-2, 2), randomizer.getRandomRangeFloat(0.5, 1));
                jump.normalize();
                jump.scalarMultiplyWith(randomizer.getRandomRangeFloat(1.2, 1.6));
                if (Utils.isDefined(settings.jumpDirectionX)) {
                    jump.x = settings.jumpDirectionX;
                }
                if (Utils.isDefined(settings.jumpDirectionY)) {
                    jump.y = settings.jumpDirectionY;
                }
                enemy.getComponent('stateMachine', function (stateMachine) {
                    stateMachine.setState('jump', {
                        jumpX: jump.x,
                        jumpY: jump.y
                    });
                });
                Utils.timeout(25, function () {
                    enemy.getComponent('movable', function (movable) {
                        //jump lasts about 27 frames, stop ignoring solids after 25
                        movable.settings.ignoreSolids = false;
                        if (MatchManager.isOnline) {
                            // bug: when playing online, the enemy has the tendency to freeze here
                            AiManager.distributeAiTasks();
                        }
                    });
                });
            },
            randomizer: randomizer,
            forceSpawner: function () {
                spawnEnemyWave();
            },
            getBossWave: function() {
                return parseInt(bossWave);
            },
            addWavesAtStart: function (data) {
                queueUpEnemies(data);
            }
        });
        return spawner;
    };

    return Spawner;
});