/**
 * Door in a level that opens and spawns new enemies
 * @moduleName SpawnElevator
 */
bento.define('entities/spawners/spawnelevator', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/spawner',
    'components/movebackground',
    'entities/bossintro',
    'modules/matchmanager',
    'managers/ai'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Spawner,
    MoveBackground,
    BossIntro,
    MatchManager,
    AiManager
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var tiled = (settings.tiled || {});
        var object = tiled.object || {};
        var spawnProperties = Utils.nameToJson(object.name);
        var imageName = spawnProperties.door || 'subwaydoor';

        var entity;

        //elevator specific stuff
        var elevatorDirection = parseInt(spawnProperties.door) || 1;
        var elevatorType = settings.elevator || tiled.tileProperties.elevator || 'door';
        var sprite = new Sprite({
            spriteSheet: 'props/elevator/' + elevatorType + '-closed'
        });
        var walls;
        var backWalls;
        var shaker;
        var spawnDelay = 80;

        settings.movePosition = function () {
            entity.position.x += Math.floor((96 - sprite.frameWidth) / 2);
            entity.position.y -= 5;
        };

        var startElevator = function () {
            new Tween({
                from: 0,
                to: 1,
                in: 60,
                ease: 'easeInQuad',
                onStart: function () {
                    Bento.audio.stopSound('sfx_elevator-loop');
                    Bento.audio.playSound('sfx_elevator-loop', true);
                },
                onUpdate: function (v, t) {
                    if (walls) {
                        walls.move.y = 3 * v * elevatorDirection;
                    }
                    if (backWalls) {
                        backWalls.move.y = 2.8 * v * elevatorDirection;
                    }
                    if (shaker) {
                        shaker.power = v;
                    }
                    Bento.audio.setVolume(v, 'sfx_elevator-loop');
                }
            });
        };

        var stopElevator = function () {
            new Tween({
                from: 1,
                to: 0,
                in: 60,
                ease: 'easeOutQuad',
                onStart: function () {},
                onUpdate: function (v, t) {
                    if (walls) {
                        walls.move.y = 3 * v * elevatorDirection;
                    }
                    if (backWalls) {
                        backWalls.move.y = 2.8 * v * elevatorDirection;
                    }
                    if (shaker) {
                        shaker.power = v;
                    }
                    Bento.audio.setVolume(v, 'sfx_elevator-loop');
                },
                onComplete: function () {
                    Bento.audio.stopSound('sfx_elevator-loop');
                    // beep sound
                    Bento.audio.stopSound('sfx_elevator-floor');
                    Bento.audio.playSound('sfx_elevator-floor');
                    // open elevator door
                    Bento.audio.stopSound('sfx_elevator-door');
                    Bento.audio.playSound('sfx_elevator-door');
                    sprite.setSpriteSheet('props/elevator/' + elevatorType + '-open', function () {
                        sprite.setSpriteSheet('props/elevator/' + elevatorType + '-opened');
                    });
                }
            });
        };

        var elevatorBehavior = {
            name: 'elevatorBehavior',
            start: function () {
                // add component to control the wall layer
                walls = new MoveBackground({
                    move: new Vector2(0, 3 * elevatorDirection)
                });
                // add component to control the wall layer
                backWalls = new MoveBackground({
                    move: new Vector2(0, 3 * elevatorDirection)
                });
                Bento.objects.getByName('foreground', function (foreground) {
                    if (foreground.length) {
                        foreground[0].attach(walls);
                    } else {
                        console.log('Warning: no foreground layer found');
                    }
                });

                Bento.objects.getByName('background', function (background) {
                    if (background.length) {
                        background[0].attach(backWalls);
                    } else {
                        console.log('Warning: no background layer found');
                    }
                });

                // get shaker to control amount of screen offset;
                shaker = Bento.objects.getByFamily('shaker')[0];

                entity.spawnY = 26;

                // start spawning
                Utils.timeout(60, function () {
                    entity.force = true;
                    entity.forceSpawner();
                    entity.force = undefined;
                });
            }


        };

        entity = new Spawner(settings)
            .attach(sprite)
            .attach(elevatorBehavior)
            .extend({
                preSpawn: function (canSpawn) {
                    if (!canSpawn) {
                        return;
                    }
                    stopElevator();
                },
                postSpawn: function (canSpawn, wave) {
                    if (!canSpawn) {
                        return;
                    }
                    // delay the start of the elevator with the delay that it takes for the elevator to stop and doors to open
                    Utils.timeout(spawnDelay, function () {
                        if (wave === entity.getBossWave() + 1) {
                            // boss intro should be spawned
                            if (!entity.force) {
                                Bento.objects.attach(new BossIntro({}));
                            }
                        }
                        sprite.setSpriteSheet('props/elevator/' + elevatorType + '-close', function () {
                            sprite.setSpriteSheet('props/elevator/' + elevatorType + '-closed');
                            startElevator();
                        });
                    });
                }
            });


        entity.spawnEnemy = function (enemy) {
            Utils.timeout(spawnDelay, function () {
                enemy.position.x = entity.position.x;
                enemy.position.y = entity.position.y + 30;
                enemy.getComponent('movable', function (movable) {
                    movable.velocityZ = -1.5;
                    movable.height = -30;
                });

                Bento.objects.attach(enemy);

                enemy.getComponent('enemyController', function (enemyController) {
                    if (!enemyController.active) {
                        enemyController.active = entity.activateEnemy;
                    }
                });

                // set jump after attaching enemy to stage otherwise state doesn't exist
                var jump = new Vector2(entity.randomizer.getRandomRangeFloat(-2, 2), entity.randomizer.getRandomRangeFloat(0.5, 1));
                jump.normalize();
                jump.scalarMultiplyWith(entity.randomizer.getRandomRangeFloat(1.2, 1.6));
                enemy.getComponent('stateMachine', function (stateMachine) {
                    stateMachine.setState('jump', {
                        jumpX: jump.x,
                        jumpY: jump.y
                    });
                });
                Utils.timeout(27, function () {
                    if (MatchManager.isOnline) {
                        // bug: when playing online, the enemy has the tendency to freeze here
                        AiManager.distributeAiTasks();
                    }
                });
            });
        };
        return entity;
    };
});