/**
 * Blockades prevent further movement to the right until all enemies to the left are cleared
 *
 * @moduleName Blockade
 */
bento.define('entities/triggers/blockade', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/cameralocker',
    'components/debug',
    'modules/matchmanager',
    'entities/hand'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    CameraLocker,
    Debug,
    MatchManager,
    Hand
) {
    'use strict';
    var Blockade = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();

        var locker = null;
        var spawners = [];

        var activate = function () {
            if (locker) {
                Utils.log('blockade activated a second time');
                return;
            }

            var screen = Bento.screens.getCurrentScreen();
            locker = new CameraLocker({
                rectangle: new Rectangle(0, 0, entity.position.x, screen.dimension.height)
            });

            Bento.objects.attach(locker);
            locker.enclose();

            Utils.forEach(Bento.objects.getByFamily('triggers'), function (spawner, i, l, breakLoop) {
                if (spawner.name !== 'spawn') {
                    return;
                }

                if (spawner.position.x <= entity.position.x) {
                    spawners.push(spawner);
                }
            });
        };

        var onEnemyDestroyed = function (e) {
            // wait a few frames to give spawners a chance to spawn
            Utils.timeout(10, function () {
                // count enemies left of current blockade
                var enemyCount = 0;
                Utils.forEach(Bento.getByFamily('enemy'), function (enemy, i, l, breakLoop) {
                    var enemyController = enemy.getComponent('enemyController');
                    if (Blockade.active) {
                        // check for active controllers and obstacles to the left of the blockade
                        if (enemyController && enemyController.active) {
                            enemyCount += 1;
                        } else if (enemy.position.x < Blockade.active.position.x) {
                            enemyCount += 1;
                        }
                    } else {
                        enemyCount += 1;
                    }
                });

                if (locker !== null && enemyCount === 0 && MatchManager.isMaster) {

                    EventSystem.fire('removeBlockade');

                    Bento.objects.remove(entity);
                    Blockade.activateNext();
                    MatchManager.sendData({
                        type: 'removal'
                    }, 'blockade');
                }
            });
        };

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                entity.dimension.width = 16;
                entity.dimension.height = 16;
                Blockade.addBlockade(entity);
                EventSystem.on('destroyEnemyController', onEnemyDestroyed);

                // warn if a blockade is placed < 320 pixels (landscape resolution)
                if (entity.position.x < 320) {
                    entity.position.x = 320;
                    Utils.log('WARNING: a blockade is placed < 320 pixels');
                }
            },
            destroy: function (data) {
                if (locker) {
                    Bento.objects.remove(locker);
                }
                EventSystem.off('destroyEnemyController', onEnemyDestroyed);
            },
            update: function (data) {
                if (!MatchManager.isMaster && locker !== null) {
                    var messageQueue = MatchManager.getMessages('blockade');
                    //NOTE: this is a bit dirty but I do not expect blockades to have to send any other messages than their own destruction
                    //The only problem that can arise is if multiple blockades get destroyed in a very small timeframe
                    if (messageQueue.length > 0) {
                        messageQueue.length = 0;
                        Bento.objects.remove(entity);
                        Blockade.activateNext();
                    }
                }
            },
            draw: function (data) {}
        };
        var entity = new Entity({
            z: 0,
            name: 'blockade',
            family: ['blockade'],
            position: settings.position || new Vector2(0, 0),
            updateWhenPaused: 0,
            components: [
                behavior
            ]
        }).extend({
            activate: activate
        });
        return entity;
    };

    Blockade.blockades = [];

    Blockade.reset = function () {
        Blockade.active = null;
        Blockade.done = 0;
        Blockade.blockades = [];
    };

    Blockade.addBlockade = function (blockade) {
        Blockade.blockades.push(blockade);

        Blockade.blockades.sort(function (a, b) {
            //sort so the left-most blockade is popped first
            return b.position.x - a.position.x;
        });
    };

    Blockade.activateNext = function () {
        if (Blockade.blockades.length === 0) {
            // last blockade was removed, show hand
            if (Blockade.active) {
                Bento.objects.attach(new Hand({}));
            }
            Blockade.done = 1;
            Blockade.active = null;
            return;
        }

        var next = Blockade.blockades.pop();
        next.activate();
        Blockade.active = next;
    };

    Blockade.active = null;

    Blockade.done = 0;

    return Blockade;
});