/**
 * Trigger to locks the camera to an area until some condition has been met
 * @moduleName Lock
 */
bento.define('entities/triggers/lock', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/trigger',
    'components/debug',
    'entities/cameralocker',
    'entities/hand'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Trigger,
    Debug,
    CameraLocker,
    Hand
) {
    'use strict';
    var Lock = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2,
            // size of trigger
            rectangle: new Rectangle(),
            toSpawn: [],
        }*/
        var viewport = Bento.getViewport();
        var rectangle = settings.rectangle || new Rectangle(0, 0, 0, 0);
        var spawnLeft = new Rectangle(0, rectangle.y, 1, rectangle.height);
        var spawnRight = new Rectangle(0, rectangle.y, 1, rectangle.height);
        var toSpawn = settings.toSpawn || [];
        var trigger = new Trigger({
            onTrigger: function () {
                // get the camera lock that is touching the trigger
                entity.collidesWith({
                    entities: Lock.cameraLockers,
                    onCollide: function (cameraLocker) {
                        // count enemies (only the ones inside the camera locker)
                        var enemies = Bento.objects.getByFamily('enemy');
                        enemyCountDown.enemyCount = 0;
                        cameraLocker.collidesWith({
                            entities: enemies,
                            firstOnly: false,
                            onCollide: function () {
                                enemyCountDown.enemyCount += 1;
                            }
                        });
                        if (enemyCountDown.enemyCount > 0) {
                            cameraLocker.attach(enemyCountDown);

                            // add the lock
                            Bento.objects.attach(cameraLocker);
                            // spawn solids around the cameraLock
                            cameraLocker.enclose();
                        }

                        CameraLocker.currentLock = cameraLocker;

                        // project spawn points left and right
                        spawnLeft.x = cameraLocker.boundingBox.x - 32;
                        spawnRight.x = cameraLocker.boundingBox.x + cameraLocker.boundingBox.width + 64; // TODO: why are the spawned slightly to the left?
                    }
                });
                Bento.objects.remove(entity);
            }
        });
        var enemyCountDown = {
            name: 'enemyCountDown',
            minimumToDefeat: 0,
            enemyCount: 0,
            toSpawn: [],
            start: function () {
                EventSystem.on('destroyEnemyController', this.onDestroyEnemy);
            },
            destroy: function () {
                EventSystem.off('destroyEnemyController', this.onDestroyEnemy);
            },
            onDestroyEnemy: function () {
                enemyCountDown.enemyCount -= 1;

                // spawn a new enemy?
                // TODO: this system is very flawed
                if (toSpawn.length) {
                    enemyCountDown.spawn();
                } else if (enemyCountDown.enemyCount <= 0) {
                    Utils.removeFromArray(Lock.cameraLockers, CameraLocker.currentLock);
                    Bento.objects.remove(CameraLocker.currentLock);
                }
            },
            spawn: function () {
                // spawn a new enemy
                var module = 'entities/enemies/' + toSpawn.pop();
                bento.require([module], function (Enemy) {
                    var left = Utils.getRandom(2);
                    var position = new Vector2(0, 0);
                    var enemy;

                    if (left) {
                        position = spawnLeft.getCenter();
                        position.y += -32 + Utils.getRandom(64);
                    } else {
                        position = spawnRight.getCenter();
                        position.y += -32 + Utils.getRandom(64);
                    }
                    enemyCountDown.enemyCount += 1;

                    enemy = new Enemy({
                        position: position
                    });
                    // dont start deactivated
                    enemy.removeByName('deactivate');

                    Bento.objects.attach(enemy);
                });
            }
        };
        var entity = new Entity({
            z: 1,
            name: 'lock',
            family: ['triggers'],
            position: new Vector2(rectangle.x, rectangle.y),
            boundingBox: new Rectangle(0, 0, rectangle.width, rectangle.height),
            components: [
                trigger,
                // new Debug({})
            ]
        });

        return entity;
    };

    // available camera locks
    Lock.cameraLockers = [];

    return Lock;
});