/**
 * Viewport scroller for world select
 * @moduleName ViewportScroller
 */
bento.define('entities/viewportscroller', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'managers/save',
    'components/anchor',
    'globals',
    'components/eventlistener'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    SaveManager,
    Anchor,
    Globals,
    EventListener
) {
    'use strict';
    return function (settings) {
        var viewport = Bento.getViewport();
        var worldScroller = settings.worldScroller;
        var mapManager = settings.mapManager;
        var hashGrid = settings.hashGrid;
        var indexToKey = function (index) {
            return index.x + '_' + index.y;
        };
        var viewportCameraTarget = new Vector2(0, 0);
        var viewportBehavior = {
            name: 'behavior',
            lastApprovedTarget: new Vector2(0, 0),
            saveScroll: function () {
                // save position
                SaveManager.save('worldmapPositionVertical', {
                    x: viewport.x,
                    y: viewport.y,
                    portrait: Anchor.isPortrait()
                });
            },
            start: function () {
                viewportController.loadPosition();
            },
            destroy: function () {
                // viewportBehavior.saveScroll();
            },
            onScrollEnd: function () {
                // force back to last approved target
                viewportBehavior.lastApprovedTarget.x = mapManager.mapSize / 2 - viewport.width / 2;
                worldScroller.forceTo(viewportBehavior.lastApprovedTarget);
            },
            getIndices: function (rectangle) {
                var mapSize = mapManager.mapSize;
                return {
                    topLeft: new Vector2(
                        Math.floor(rectangle.x / mapSize),
                        Math.floor(rectangle.y / mapSize)
                    ),
                    topRight: new Vector2(
                        Math.floor(rectangle.getX2() / mapSize),
                        Math.floor(rectangle.y / mapSize)
                    ),
                    bottomRight: new Vector2(
                        Math.floor(rectangle.getX2() / mapSize),
                        Math.floor(rectangle.getY2() / mapSize)
                    ),
                    bottomLeft: new Vector2(
                        Math.floor(rectangle.x / mapSize),
                        Math.floor(rectangle.getY2() / mapSize)
                    ),
                    isAnyOutside: function () {
                        // one of the corners is outside the borders
                        var corners = !hashGrid[indexToKey(this.topLeft)] +
                            !hashGrid[indexToKey(this.topRight)] +
                            !hashGrid[indexToKey(this.bottomRight)] +
                            !hashGrid[indexToKey(this.bottomLeft)];
                        return corners;
                    },
                    isTopOutside: function () {
                        return !hashGrid[indexToKey(this.topLeft)] &&
                            !hashGrid[indexToKey(this.topRight)] &&
                            hashGrid[indexToKey(this.bottomRight)] &&
                            hashGrid[indexToKey(this.bottomLeft)];
                    },
                    isBottomOutside: function () {
                        return hashGrid[indexToKey(this.topLeft)] &&
                            hashGrid[indexToKey(this.topRight)] &&
                            !hashGrid[indexToKey(this.bottomRight)] &&
                            !hashGrid[indexToKey(this.bottomLeft)];
                    },
                    isLeftOutside: function () {
                        return !hashGrid[indexToKey(this.topLeft)] &&
                            hashGrid[indexToKey(this.topRight)] &&
                            hashGrid[indexToKey(this.bottomRight)] &&
                            !hashGrid[indexToKey(this.bottomLeft)];
                    },
                    isRightOutside: function () {
                        return hashGrid[indexToKey(this.topLeft)] &&
                            !hashGrid[indexToKey(this.topRight)] &&
                            !hashGrid[indexToKey(this.bottomRight)] &&
                            hashGrid[indexToKey(this.bottomLeft)];
                    }
                };
            },
            update: function () {
                var scrollPosition = worldScroller.getScrollPosition();
                var viewportTarget = new Rectangle(scrollPosition.x, scrollPosition.y, viewport.width, viewport.height);
                var mapSize = mapManager.mapSize;
                var targetIndices = viewportBehavior.getIndices(viewportTarget);
                var lastApprovedTarget = viewportBehavior.lastApprovedTarget;
                var lastApprovedIndex = new Vector2(
                    Math.floor(lastApprovedTarget.x / mapSize),
                    Math.floor(lastApprovedTarget.y / mapSize)
                );
                var indicesOutside = targetIndices.isAnyOutside();
                var diff = Infinity;
                var alignTop = function () {
                    return new Vector2(
                        viewportTarget.x,
                        lastApprovedIndex.y * mapSize + 1
                    );
                };
                var alignBottom = function () {
                    return new Vector2(
                        viewportTarget.x,
                        (lastApprovedIndex.y + 1) * mapSize - viewport.height - 1
                    );
                };
                var alignLeft = function () {
                    return new Vector2(
                        lastApprovedIndex.x * mapSize + 1,
                        viewportTarget.y
                    );
                };
                var alignRight = function () {
                    return new Vector2(
                        (lastApprovedIndex.x + 1) * mapSize - viewport.width - 1,
                        viewportTarget.y
                    );
                };
                var moveToScrollPosition = function () {
                    // move to new scroll position
                    scrollPosition.x = mapManager.mapSize / 2 - viewport.width / 2;
                    worldScroller.forceTo(scrollPosition);
                    viewportCameraTarget.x = scrollPosition.x;
                    viewportCameraTarget.y = scrollPosition.y;
                    viewportBehavior.lastApprovedTarget.x = viewportCameraTarget.x;
                    viewportBehavior.lastApprovedTarget.y = viewportCameraTarget.y;
                };

                if (indicesOutside === 2) {
                    // is touching a side

                    if (targetIndices.isTopOutside()) {
                        scrollPosition = alignTop();
                        worldScroller.stopSlidingY();
                    } else if (targetIndices.isBottomOutside()) {
                        scrollPosition = alignBottom();
                        worldScroller.stopSlidingY();
                    } else if (targetIndices.isLeftOutside()) {
                        scrollPosition = alignLeft();
                        worldScroller.stopSlidingX();
                    } else if (targetIndices.isRightOutside()) {
                        scrollPosition = alignRight();
                        worldScroller.stopSlidingX();
                    }
                    moveToScrollPosition();

                } else if (indicesOutside === 1) {
                    // find out if x or y needs to be aligned
                    if (!hashGrid[indexToKey(targetIndices.topLeft)]) {
                        // try the shortest distance
                        if (alignLeft().subtract(scrollPosition).sqrMagnitude() < alignTop().subtract(scrollPosition).sqrMagnitude()) {
                            scrollPosition = alignLeft();
                            worldScroller.stopSlidingX();
                        } else {
                            scrollPosition = alignTop();
                            worldScroller.stopSlidingY();
                        }
                    } else if (!hashGrid[indexToKey(targetIndices.topRight)]) {
                        // try the shortest distance
                        if (alignRight().subtract(scrollPosition).sqrMagnitude() < alignTop().subtract(scrollPosition).sqrMagnitude()) {
                            scrollPosition = alignRight();
                            worldScroller.stopSlidingX();
                        } else {
                            scrollPosition = alignTop();
                            worldScroller.stopSlidingY();
                        }
                    } else if (!hashGrid[indexToKey(targetIndices.bottomRight)]) {
                        // try the shortest distance
                        if (alignRight().subtract(scrollPosition).sqrMagnitude() < alignBottom().subtract(scrollPosition).sqrMagnitude()) {
                            scrollPosition = alignRight();
                            worldScroller.stopSlidingX();
                        } else {
                            scrollPosition = alignBottom();
                            worldScroller.stopSlidingY();
                        }
                    } else if (!hashGrid[indexToKey(targetIndices.bottomLeft)]) {
                        // try the shortest distance
                        if (alignLeft().subtract(scrollPosition).sqrMagnitude() < alignBottom().subtract(scrollPosition).sqrMagnitude()) {
                            scrollPosition = alignLeft();
                            worldScroller.stopSlidingX();
                        } else {
                            scrollPosition = alignBottom();
                            worldScroller.stopSlidingY();
                        }
                    }
                    moveToScrollPosition();

                } else if (indicesOutside === 3) {
                    // is touching a corner
                    worldScroller.stopSlidingX();
                    worldScroller.stopSlidingY();

                    // align to the corner
                    if (hashGrid[indexToKey(targetIndices.topLeft)]) {
                        scrollPosition = new Vector2(
                            (targetIndices.topLeft.x + 1) * mapSize - viewport.width - 1,
                            (targetIndices.topLeft.y + 1) * mapSize - viewport.height - 1
                        );
                    } else if (hashGrid[indexToKey(targetIndices.topRight)]) {
                        scrollPosition = new Vector2(
                            targetIndices.topRight.x * mapSize + 1,
                            (targetIndices.topRight.y + 1) * mapSize - viewport.height - 1
                        );
                    } else if (hashGrid[indexToKey(targetIndices.bottomRight)]) {
                        scrollPosition = new Vector2(
                            targetIndices.bottomRight.x * mapSize + 1,
                            targetIndices.bottomRight.y * mapSize + 1
                        );
                    } else if (hashGrid[indexToKey(targetIndices.bottomLeft)]) {
                        scrollPosition = new Vector2(
                            (targetIndices.bottomLeft.x + 1) * mapSize - viewport.width - 1,
                            targetIndices.bottomLeft.y * mapSize + 1
                        );
                    }
                    moveToScrollPosition();

                } else if (indicesOutside === 4) {
                    // TODO
                    // console.log('WARNING: viewport scrolled out of bounds');
                } else {
                    // move freely
                    viewportCameraTarget.x = viewportTarget.x;
                    viewportCameraTarget.y = viewportTarget.y;

                    viewportBehavior.lastApprovedTarget.x = viewportCameraTarget.x;
                    viewportBehavior.lastApprovedTarget.y = viewportCameraTarget.y;
                }

                // smoothen out
                viewport.x += (viewportCameraTarget.x - viewport.x) / 2;
                viewport.y += (viewportCameraTarget.y - viewport.y) / 2;

                // save every few seconds
                if (viewportController.timer % 300 === 299) {
                    viewportBehavior.saveScroll();
                }
            }
        };
        var rotate = function (isPortrait) {
            var scrollPosition = worldScroller.getScrollPosition().clone();
            var h = Math.max(viewport.width, viewport.height);
            var w = Math.min(viewport.width, viewport.height);
            var diff = (h - w) / 2;
            if (isPortrait) {
                scrollPosition.x += diff;
                scrollPosition.y -= diff;
            } else {
                scrollPosition.x -= diff;
                scrollPosition.y += diff;
            }
            scrollPosition.x = mapManager.mapSize / 2 - viewport.width / 2;
            worldScroller.forceTo(scrollPosition);
            viewportCameraTarget = scrollPosition;
            // force an update
            viewportBehavior.update();
            viewport.x = viewportCameraTarget.x;
            viewport.y = viewportCameraTarget.y;
        };
        var anchor = new Anchor({
            onComplete: function (parent, isPortrait) {
                rotate(isPortrait);
            }
        });
        var viewportController = new Entity({
            z: 0,
            name: 'viewportController',
            components: [
                anchor,
                viewportBehavior,
                new EventListener({
                    name: 'saveViewportPosition',
                    onEvent: function (data) {
                        viewportBehavior.saveScroll();
                    }
                })
            ]
        });

        viewportController.extend({
            forceTo: function (position) {
                // keep centered
                position.x = mapManager.mapSize / 2 - viewport.width / 2;

                worldScroller.forceTo(position);
                viewportCameraTarget = position;
                // force an update
                viewportBehavior.update();
                viewport.x = viewportCameraTarget.x;
                viewport.y = viewportCameraTarget.y;
            },
            loadPosition: function () {
                // load the last known position
                var lastKnownPosition = SaveManager.load('worldmapPositionVertical');
                var lastPosition = Utils.toVector2(lastKnownPosition);
                viewportController.forceTo(lastPosition);
                viewportCameraTarget = lastPosition;

                // did the orientation change since last time?
                if (lastKnownPosition.portrait !== Anchor.isPortrait()) {
                    rotate(Anchor.isPortrait());
                }
                return lastKnownPosition;
            }
        });
        return viewportController;
    };
});