/**
 * A hashgrid containing the tiles the AI can walk on
 * @moduleName WalkableArea
 */
bento.define('entities/walkable', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var screenDimension = Bento.screens.getCurrentScreen().dimension;
        var floor = Math.floor;
        var grid = {};
        var cellSize = 16;
        var invCellSize = 1 / cellSize;
        var verticalArea = {};

        // fill grid with walkable tiles
        var behavior = {
            name: 'fillBehavior',
            start: function () {
                Bento.objects.get('tileColliders', function (tileColliders) {
                    // node functions
                    var checkNode = function (n) {
                        // check left
                        checkSide(n.left);
                        // check right
                        checkSide(n.right);
                        // check up
                        checkSide(n.up);
                        // check down
                        checkSide(n.down);
                    };

                    var checkSide = function (side) {
                        // check tileColliders
                        if (screenDimension.hasPosition(side) && !tileColliders.checkCell(side)) {
                            // inside walkable area
                            if (!entity.checkCell(side)) {
                                // not yet checked, add to open list
                                open.push(makeNode(side));
                                entity.insert(side.x, side.y);
                            }
                        }
                    };

                    // list with open nodes
                    var open = [];
                    // check starting node
                    var node = makeNode(settings.start);
                    entity.insert(node.node.x, node.node.y);
                    checkNode(node);

                    // check all tiles
                    while (open.length > 0) {
                        node = open.shift();
                        checkNode(node);
                    }
                });

            }
        };

        var makeNode = function (n) {
            return {
                node: n,
                left: {
                    x: n.x - cellSize,
                    y: n.y,
                },
                right: {
                    x: n.x + cellSize,
                    y: n.y,
                },
                up: {
                    x: n.x,
                    y: n.y - cellSize,
                },
                down: {
                    x: n.x,
                    y: n.y + cellSize,
                },
            };
        };

        var entity = new Entity({
            z: 0,
            name: 'walkableArea',
            family: ['walkableArea'],
            components: [
                behavior
            ]
        });
        var posToKey = function (x, y) {
            return floor(x * invCellSize).toString() + ':' + floor(y * invCellSize);
        };

        entity.extend({
            insert: function (x, y) {
                var key = posToKey(x, y);
                // insert make a dummy entity for collision purposes
                var dummy = new Entity({
                    name: 'gridEntry',
                    position: new Vector2(x, y)
                });
                var dimension = dummy.dimension;
                dimension.width = cellSize;
                dimension.height = cellSize;

                grid[key] = dummy;
            },
            checkCell: function (position) {
                var key = posToKey(position.x, position.y);
                return grid[key];
            },
            checkVertical: function (xPosition) {
                var checkX = Utils.clamp(8, xPosition, screenDimension.width - 8);
                var key = floor(checkX * invCellSize).toString();
                if (verticalArea[key]) {
                    return verticalArea[key];
                } else {
                    var area = [];
                    var walker = new Vector2(checkX, 8);
                    var start;
                    var total = 0;
                    //store all walkable areas
                    while (walker.y < screenDimension.height) {
                        if (!start) {
                            if (entity.checkCell(walker)) {
                                start = walker.y - 2;
                            }
                        } else {
                            if (!entity.checkCell(walker)) {
                                area.push({
                                    top: start,
                                    bottom: walker.y - 14
                                });
                                total += walker.y - 14 - start;
                                start = undefined;
                            }
                        }
                        walker.y += 16;
                    }

                    // if there is a start it means something is wrong with the level
                    if (start) {
                        console.log('warning, walkable area open to bottom of level at tile', key);
                        area.push({
                            top: start,
                            bottom: screenDimension.height - 6
                        });
                    }

                    verticalArea[key] = area;
                    return area;
                }
            }
        });
        return entity;
    };
});