/**
 * Blurred background
 * @moduleName BlurBackground
 * @snippet BlurBackground.snippet
BlurBackground({
    darken: 0.3,
    blurriness: 15
});
 */
bento.define('gui/blurbackground', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/snapshot',
    'modules/stackblur',
    'bento/packedimage',
    'components/anchor',
    'components/fillscreen',
    'components/hidebelow',
    'managers/resolution',
    'gui/backscroller'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    SnapShot,
    StackBlur,
    PackedImage,
    Anchor,
    FillScreen,
    HideBelow,
    ResolutionManager,
    BackScroller
) {
    'use strict';

    // consideration: use for all devices or only android?
    if (Utils.isNativeAndroid() || Utils.isOldIphone()) {
        // return a backscroller instead of a blurbackground
        return function (settings) {
            var background = new BackScroller({
                name: 'background',
                family: ['background'],
                position: new Vector2(0, 0),
                imageName: 'splash/scrollingbackground-luckyred',
                velocity: new Vector2(0.25, 0.25),
                spacing: new Vector2(-1, -1)
            });
            background.attach(new HideBelow({}));
            background.attach(new FillScreen({
                color: [0, 0, 0, Utils.isDefined(settings.darken) ? settings.darken : 0.2]
            }));
            return background;
        };
    }
    return function (settings) {
        var viewport = Bento.getViewport();
        var blur = null;
        var blurriness = settings.blurriness || 15;
        var createBlurredSprite = function (exclusions) {
            var blurredSprite;
            var canvas;
            var snapshot = new SnapShot({
                scale: 1,
            });

            // take snapshot
            snapshot.snap(exclusions);
            canvas = snapshot.getCanvas();

            // blur snapshot
            StackBlur.canvasRGBA(canvas, 0, 0, canvas.width, canvas.height, blurriness);
            blur = new PackedImage(canvas);
            blurredSprite = new Sprite({
                image: blur
            });

            return blurredSprite;
        };

        var anchor = new Anchor({
            onComplete: function () {
                var fullSize = viewport.clone();
                var offset;
                if (ResolutionManager.isWideScreen()) {
                    fullSize = ResolutionManager.getWideScreenSize();
                    offset = ResolutionManager.getOffset();
                    entity.transform.x = -offset.x;
                    entity.transform.y = -offset.y;
                }

                // unhide to take snapshot again
                hideBelow.unHide();
                // (re)take snapshot
                var blurredSprite;
                if (!Utils.isNativeAndroid()) {
                    blurredSprite = createBlurredSprite([entity, entity.parent, Bento.objects.get('senseiPopup')]);
                }
                hideBelow.reHide();

                // swap with old sprite
                entity.removeByName('sprite');
                if (blurredSprite) {
                    entity.attach(blurredSprite);
                }

                // fit to screen (the snapshot may not be the same size as the screensize)
                entity.scale.x = fullSize.width / entity.dimension.width;
                entity.scale.y = fullSize.height / entity.dimension.height;

                if (settings.darken) {
                    entity.removeByName('fillScreen');
                    entity.attach(new FillScreen({
                        color: [0, 0, 0, settings.darken]
                    }));
                }
            }
        });
        var hideBelow = new HideBelow({});
        var cleanUp = {
            name: 'cleanUp',
            destroy: function () {
                if (blur && blur.image && blur.image.dispose) {
                    blur.image.dispose();
                }
            }
        };
        var entity = new Entity({
            name: 'blurBackground',
            components: [
                anchor,
                hideBelow
            ]
        });

        return entity;
    };
});