/**
 * Coin reward is a popup that shows how many coins the player received
 * It does not actually add any coins (assumed it's already added in fact)
 * @moduleName CoinReward
 */
bento.define('gui/coinreward', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'components/modal',
    'components/fillscreen',
    'gui/shadowtext',
    'components/anchor',
    'gui/blurbackground',
    'gui/worldselect/topbar'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    Modal,
    FillScreen,
    ShadowText,
    Anchor,
    BlurBackground,
    Topbar
) {
    'use strict';
    var targets = {
        coin: 36,
        gem: 0,
        star: -36,
        energy: -72
    };
    var spriteSheets = {
        coin: 'coin/coin',
        gem: 'currency/gem',
        energy: 'energy/energy'
    };
    var big = {
        coin: {
            imageName: 'gui/coin',
            originRelative: new Vector2(0.5, 0.5)
        },
        gem: {
            spriteSheet: 'currency/gem',
            originRelative: new Vector2(0.5, 0.5)
        },
        energy: {
            spriteSheet: 'other/energy',
            originRelative: new Vector2(0.5, 0.5)
        }
    };
    var flash = {
        coin: 'flashCoin',
        gem: 'flashGems',
        energy: 'flashEnergy'
    };
    var rotate = {
        coin: false,
        gem: true,
        energy: false
    };
    // coin that will fly towards the topbar
    var Coin = function (settings) {
        var viewport = Bento.getViewport();
        var topbar = settings.topbar;
        var type = settings.type;
        var spriteSheet = spriteSheets[type];
        var sprite = new Sprite({
            spriteSheet: spriteSheet
        });
        var gravity = 0.3;
        var rotationSpeed = 0;
        var behavior = {
            name: 'behavior',
            gravity: gravity,
            velocity: settings.velocity || new Vector2(0, 0),
            update: function (data) {
                behavior.velocity.y += behavior.gravity;
                entity.position.x += behavior.velocity.x;
                entity.position.y += behavior.velocity.y;

                if (rotationSpeed) {
                    entity.rotation += rotationSpeed;
                }
            },
            start: function () {
                if (rotate[type]) {
                    rotationSpeed = Utils.getRandomRangeFloat(-0.1, 0.1);
                }
                // after a few frames of flying, fly towards coin counter in topbar
                new Tween({
                    from: 1,
                    to: 0,
                    in: 20,
                    delay: 50,
                    ease: 'exponential',
                    alpha: 5,
                    onUpdate: function (v, t) {
                        behavior.gravity = gravity * v;
                        behavior.velocity.x = behavior.velocity.x * v;
                        behavior.velocity.y = behavior.velocity.y * v;
                        entity.position.x += (targets[type] - entity.position.x) * (1 - v);
                        entity.position.y += (-viewport.height / 2 - entity.position.y) * (1 - v);
                    },
                    onComplete: function () {
                        topbar[flash[type]]();
                        entity.removeSelf();
                        Bento.audio.stopSound('sfx_coin');
                        Bento.audio.playSound('sfx_coin');
                    }
                });
            }
        };
        var entity = new Entity({
            name: 'coin',
            position: settings.position,
            components: [
                sprite,
                behavior
            ]
        });

        return entity;
    };

    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var type = settings.type || 'coin';
        var coins = settings.coins || settings.amount || 0;
        var spriteSheet = spriteSheets[type];
        var onComplete = settings.onComplete;
        var blurBackground = new BlurBackground({
            darken: 0.5,
            blurriness: 2
        });
        var overlay = new FillScreen({
            color: [0, 0, 0, 0.5]
        });
        var text = new ShadowText({
            text: '+' + coins,
            font: 'font',
            fontSize: 20,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle'
        });
        var topbar = new Topbar({
            preventClick: true
        });
        var coin = new Entity({
            name: 'coin',
            position: new Vector2(text.dimension.width / 2 + 10, 2),
            components: [
                new Sprite(big[type])
            ]
        });
        var coinRainContainer = new Entity({
            name: 'coinRainContainer',
            components: [
                new Anchor({
                    screenPosition: new Vector2(0.5, 0.5)
                })
            ]
        });
        var coinContainer = new Entity({
            name: 'container',
            components: [
                text,
                new Anchor({
                    screenPosition: new Vector2(0.5, 0.5)
                })
            ]
        });
        var godrays1 = new Entity({
            name: 'godrays',
            alpha: 0.25,
            components: [
                new Anchor({
                    screenPosition: new Vector2(0.5, 0.5)
                }),
                new Sprite({
                    imageName: 'gui/godrays1',
                    originRelative: new Vector2(0.5, 0.5)
                }), {
                    update: function (data) {
                        data.entity.rotation += 0.01;
                    }
                }
            ]
        });
        var godrays2 = new Entity({
            name: 'godrays',
            alpha: 0.25,
            components: [
                new Anchor({
                    screenPosition: new Vector2(0.5, 0.5)
                }),
                new Sprite({
                    imageName: 'gui/godrays2',
                    originRelative: new Vector2(0.5, 0.5)
                }), {
                    update: function (data) {
                        data.entity.rotation -= 0.005;
                    }
                }
            ]
        });
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                Bento.audio.stopSound('sfx_coinreward');
                Bento.audio.playSound('sfx_coinreward');

                Bento.audio.stopSound('sfx_itemget');
                Bento.audio.playSound('sfx_itemget');

                text.attach(coin);
                text.position.x -= 10; // half the coin size

                new Tween({
                    from: 0,
                    to: 2,
                    in: 60,
                    delay: 0,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 3,
                    onUpdate: function (v, t) {
                        coinContainer.scale.x = v;
                        coinContainer.scale.y = v;
                    }
                });

                // remove self after 2 seconds
                new Tween({
                    from: 1,
                    to: 0,
                    in: 30,
                    delay: 150,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        entity.alpha = v;
                    },
                    onComplete: function () {
                        entity.removeSelf();
                        if (onComplete) {
                            onComplete();
                        }
                    }
                });
            },
            destroy: function (data) {},
            update: function (data) {
                // rain some coins
                if (entity.timer % 4 === 0 && entity.timer <= 4 * 10) {
                    coinRainContainer.attach(new Coin({
                        topbar: topbar,
                        type: type,
                        position: new Vector2(
                            Utils.getRandomRangeFloat(-8, 8),
                            Utils.getRandomRangeFloat(-8, 8)
                        ),
                        velocity: new Vector2(
                            Utils.getRandomRangeFloat(-1.5, 1.5),
                            Utils.getRandomRangeFloat(-8, -6)
                        )
                    }));
                }
            },
            draw: function (data) {}
        };
        var entity = new Entity({
            z: settings.z || Globals.Layers.GUI,
            name: 'coinReward',
            position: settings.position,
            float: true,
            components: [
                blurBackground,
                new Modal({}),
                behavior,
                godrays1,
                godrays2,
                coinRainContainer,
                coinContainer,
                topbar
            ]
        });
        return entity;
    };
});