/**
 * Base module for a (modal) dialog window
 * @moduleName Dialog
 * @snippet Dialog.snippet
Dialog({
    name: '${1:dialog}',
    title: '$2',
    text: '$3',
    canClose: ${4:true},
    type: '${5:yesno}',
    onYes: function () {},
    onNo: function () {},
    onClose: function () {},
    onComplete: function () {}
});
 */
bento.define('gui/dialog', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'bento/components/nineslice',
    'components/modal',
    'components/fillscreen',
    'globals',
    'components/anchor',
    'gui/blurbackground',
    'gui/headerbox',
    'gui/shadowtext',
    'modules/analytics',
    'components/orientationlock',
    'gui/iconbutton',
    'components/debug',
    'components/backbutton'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    NineSlice,
    Modal,
    FillScreen,
    Globals,
    Anchor,
    BlurBackground,
    HeaderBox,
    ShadowText,
    Analytics,
    OrientationLock,
    IconButton,
    Debug,
    BackButton
) {
    'use strict';
    var Dialog = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        /* 
         * Settings
         */
        var z = settings.z;
        var name = settings.name || 'dialog';
        var ignoreAnalytics = settings.ignoreAnalytics || false;
        // size of dialog
        var size = settings.size || new Rectangle(0, 0, 180, 180);
        // inner edge for text
        var padding = settings.padding || new Vector2(16, 48);
        var text = settings.text || '';
        var modal = Utils.isDefined(settings.modal) ? settings.modal : true;
        var buttons = settings.buttons || settings.type || Dialog.Buttons.NONE;
        var windowImage = settings.windowImage || 'gui/dialog/window';
        var yesImage = settings.yesImage || (buttons === 'yes' ? 'gui2/misc/button_small_primary' : 'gui2/misc/button_large_primary');
        var noImage = settings.noImage || 'gui2/misc/button_small_secondary';
        var iconYesImage = settings.yesIconImage || 'gui2/misc/button_icon_okay';
        var iconNoImage = settings.noIconImage || 'gui2/misc/button_icon_back';
        var iconYes = settings.yesIcon;
        var iconNo = settings.noIcon;
        var icon = settings.icon;
        var iconSize = (icon ? Utils.clamp(0, icon.dimension.height + 8, 32) : 32);
        var preventTweens = settings.preventTweens || false;
        var heightOffset = (buttons ? -4 : 0) + (icon ? iconSize / 2 + 8 : 0);
        var heightPadding = (buttons ? -8 : 0) + (icon ? -iconSize : 0);
        var textSettings = Utils.extend({
            text: text,
            position: new Vector2(0, -8 + heightOffset),
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            maxWidth: size.width - padding.x * 2,
            maxHeight: size.height - padding.y * 2 + heightPadding
        }, settings.textSettings);
        var canClose = Utils.isDefined(settings.canClose) ? settings.canClose : false;
        /* 
         * Set up entities
         */
        var background = new BlurBackground({
            blurriness: settings.blurriness || 15
        });
        var darken = new Entity({
            name: 'darken',
            components: [
                new FillScreen({
                    color: [0, 0, 0, 0.5]
                })
            ]
        });
        //  buttons
        var buttonClose = new ClickButton({
            name: 'btnClose',
            position: new Vector2(-size.width / 2 + 20, -size.height / 2 + 24),
            sfx: 'sfx_button',
            imageName: 'gui2/misc/close_popup_x',
            frameCountX: 1,
            frameCountY: 2,
            onClick: function () {
                buttonYes.setActive(false);
                buttonNo.setActive(false);
                buttonClose.setActive(false);
                if (!isTweening) {
                    moveOut();
                }
                if (settings.onClose) {
                    settings.onClose();
                }
                if (!ignoreAnalytics) {
                    Analytics.event('Dialog', name, 0, 'close');
                }
            },
            components: [
                new BackButton({
                    removeSelf: true,
                    onPressed: function (data) {
                        data.entity.doCallback();
                    }
                })
            ]
        });
        var buttonYes = new IconButton({
            name: 'buttonYes',
            sfx: 'sfx_button',
            imageName: yesImage,
            iconName: iconYesImage,
            icon: iconYes,
            frameCountX: settings.yesFrameCountX || 1,
            frameCountY: settings.yesFrameCountY || 3,
            position: new Vector2(24, size.height / 2 - 32),
            animations: settings.yesAnimations || {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [0]
                }
            },
            onClick: function () {
                if (isTweening) {
                    return;
                }
                buttonYes.setActive(false);
                buttonNo.setActive(false);
                buttonClose.setActive(false);
                if (settings.onYes) {
                    settings.onYes();
                }
                if (!ignoreAnalytics) {
                    Analytics.event('Dialog', name, 1, 'yes');
                }
                moveOut();
            }
        });
        var buttonNo = new IconButton({
            name: 'buttonNo',
            sfx: 'sfx_button',
            imageName: noImage,
            iconName: iconNoImage,
            icon: iconNo,
            frameCountX: settings.noFrameCountX || 1,
            frameCountY: settings.noFrameCountY || 3,
            animations: settings.noAnimations || {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [0]
                }
            },
            position: new Vector2(-52, size.height / 2 - 32),
            onClick: function () {
                if (isTweening) {
                    return;
                }
                buttonYes.setActive(false);
                buttonNo.setActive(false);
                buttonClose.setActive(false);
                if (settings.onNo) {
                    settings.onNo();
                }
                if (!ignoreAnalytics) {
                    Analytics.event('Dialog', name, 0, 'no');
                }
                moveOut();
            }
        });
        // text
        var dialogText = new ShadowText(textSettings);
        // window
        var boxAnchor = new Anchor({
            screenPosition: new Vector2(0.5, 0.5),
            offset: new Vector2(0, 0)
        });
        var windowBox = new HeaderBox({
                name: 'windowBox',
                size: size,
                title: settings.title,
                useInnerSlice: true,
                windowImage: 'gui2/misc/splash_background_9slice',
                maxWidth: size.width - 48
            })
            .attach(dialogText)
            .attach(boxAnchor);

        var behavior = {
            name: 'behavior',
            start: function () {
                if (preventTweens) {
                    return;
                }
                isTweening = true;

                buttonYes.setActive(false);
                buttonNo.setActive(false);
                buttonClose.setActive(false);

                // fade in background
                new Tween({
                    from: 0,
                    to: 1,
                    in: 20,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        background.alpha = v;
                        darken.alpha = v;
                    }
                });

                // tween in box
                new Tween({
                    from: 0,
                    to: 1,
                    in: 30,
                    ease: 'elastic',
                    decay: 8,
                    onUpdate: function (v, t) {
                        windowBox.scale.x = v;
                        windowBox.scale.y = v;
                    },
                    onComplete: function () {
                        isTweening = false;

                        buttonYes.setActive(true);
                        buttonNo.setActive(true);
                        buttonClose.setActive(true);

                    }
                }).attach(new OrientationLock({}));
            }
        };
        var entity = new Entity({
            z: z || Globals.Layers.DIALOGS,
            name: name,
            float: true,
            components: [
                background,
                darken,
                windowBox
            ]
        });
        // tweening out
        var isTweening = false;
        var moveOut = function () {
            if (preventTweens) {
                cleanUp();
                return;
            }
            // fade in background
            new Tween({
                from: 1,
                to: 0,
                in: 15,
                ease: 'linear',
                onUpdate: function (v, t) {
                    background.alpha = v;
                    darken.alpha = v;
                }
            });

            // tween in box
            new Tween({
                from: 1,
                to: 0,
                in: 20,
                ease: 'easeInBack',
                onUpdate: function (v, t) {
                    windowBox.scale.x = v;
                    windowBox.scale.y = v;
                },
                onComplete: function () {
                    cleanUp();
                }
            }).attach(new OrientationLock({}));
        };
        var cleanUp = function () {
            Bento.objects.remove(entity);

            if (settings.onComplete) {
                settings.onComplete();
            }

        };
        if (modal) {
            entity.attach(new Modal({}));
        }
        entity.attach(behavior);

        // attach buttons
        if (canClose) {
            windowBox.attach(buttonClose);
        } else {
            // android back button defaults to the no button
            buttonNo.attach(new BackButton({
                removeSelf: true,
                onPressed: function (data) {
                    data.entity.doCallback();
                }
            }));
        }
        if (buttons === Dialog.Buttons.YES) {
            windowBox.attach(buttonYes);
            buttonYes.position.x = 0;
            // android back button defaults to the yes button
            buttonYes.attach(new BackButton({
                removeSelf: true,
                onPressed: function (data) {
                    data.entity.doCallback();
                }
            }));
        } else if (buttons === Dialog.Buttons.YESNO) {
            windowBox.attach(buttonYes);
            windowBox.attach(buttonNo);
        }

        // attach icon
        if (icon) {
            windowBox.attach(icon);
            icon.position.y = -44 + iconSize / 2;
            icon.position.x = 0;
        }

        // tween close button
        var windowBoxClickable = new Clickable({
            onClickMiss: function (data) {
                new Tween({
                    from: 0.5,
                    to: 1,
                    in: 30,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 2,
                    onUpdate: function (v, t) {
                        buttonClose.scale.x = v;
                        buttonClose.scale.y = v;
                    }
                });
            }
        });
        windowBox.attach(windowBoxClickable);
        windowBox.boundingBox = new Rectangle(-size.width / 2, -size.height / 2, size.width, size.height);

        // public
        entity.extend({
            close: function () {
                // close the dialog
                moveOut();
            }
        });

        return entity;
    };

    // Dialog button type enums
    Dialog.Buttons = {
        NONE: '',
        YES: 'yes',
        YESNO: 'yesno'
    };

    return Dialog;
});