/**
 * Paramaters are passed as properties of the settings object.
 * @param  {Number} pause           Pause the game while fading
 * @param  {Boolean} wait           Wait for the fade to start
 * @param  {Boolean}  fadeIn        True for fading in screen, defaults to false
 * @param  {Function} onFade        Called every update while fading
 * @param  {Function} onComplete    Called when fade is completed
 * @param  {String} type            Type of fade: 'drop' = curtain drop, 'slide' = slide in from side, default = opacity
 */
bento.define('gui/fade', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'components/fillrect',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween,
    FillRect,
    Globals
) {
    'use strict';
    var Fade = function (settings) {
        var viewport = Bento.getViewport();
        var padding = 64;
        var fillRect = new FillRect({
            originRelative: new Vector2(0.5, 0.5),
            size: viewport.getSize().grow(padding)
        });
        // start and end values for fade
        var start = 0;
        var end = 1;
        var fadeIn = settings.fadeIn;
        var callback = settings.callback || settings.onComplete || null;
        var z = settings.z || Globals.Layers.FADE;
        var type = settings.type || Fade.defaultType;
        var pause = settings.pause || 0;
        var duration = settings.duration || 15;
        var wait = settings.wait || false;
        var oldPause = Bento.objects.isPaused();
        var entity = new Entity({
            z: z,
            name: 'fade',
            position: viewport.getExtents().add(new Vector2(padding / 2, padding / 2)),
            dimension: viewport.getSize().grow(padding),
            components: [fillRect],
            updateWhenPaused: pause || Globals.PauseLevels.ALWAYS,
            float: true
        }).extend({
            startFade: function () {
                if (wait) {
                    wait = false;
                    startFade();
                }
            }
        });
        var setFade = function (v, t) {
            if (type === 'drop') {
                fillRect.setOriginRelative(new Vector2(0.5, 1));
                entity.position.y = v * viewport.height;
            } else if (type === 'slide') {
                fillRect.setOriginRelative(new Vector2(1, 0.5));
                entity.position.x = v * viewport.width;
            } else if (type === 'none') {
                duration = 0;
            } else {
                entity.alpha = (v);
            }
            if (settings.onFade) {
                settings.onFade(v, t);
            }
        };
        var startFade = function () {
            new Tween({
                from: start,
                to: end,
                'in': duration,
                ease: 'easeOutCubic',
                updateWhenPaused: pause || Globals.PauseLevels.ALWAYS,
                onUpdate: setFade,
                onComplete: endFade
            });
        };
        var endFade = function () {
            // restore pause level
            if (pause) {
                Bento.objects.pause(oldPause);
            }
            if (callback) {
                callback();
            }
            Bento.objects.remove(entity);
        };

        // fade adds itself!
        Bento.objects.add(entity);

        // pause game?
        if (pause) {
            Bento.objects.pause(pause);
        }

        // swap start and end for fading in
        if (!fadeIn) {
            start = 0;
            end = 1;
        } else {
            start = 1;
            end = 0;
        }
        // init
        setFade(start, 0);

        if (duration) {
            if (!wait) {
                startFade();
            }
        } else {
            // cancel fade
            endFade();
        }

        return entity;
    };

    Fade.defaultType = 'opacity';
    return Fade;
});