/**
 * Popup that asks the player to buy missing coins with gems
 * @moduleName BuyCoinsPopup
 */
bento.define('gui/iap/buycoinspopup', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'gui/dialog',
    'modules/localization',
    'gui/shadowtext',
    'gui/worldselect/topbarsimple',
    'globals',
    'managers/currency',
    'gui/coinreward',
    'gui/menus/iap',
    'modules/analytics',
    'gui/iconbutton'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Dialog,
    Localization,
    ShadowText,
    TopBarSimple,
    Globals,
    CurrencyManager,
    CoinReward,
    IAPMenu,
    Analytics,
    IconButton
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var coinsToBuy = settings.coinsToBuy; // will be recalculated to fit amount of gems needed
        var buttonToClick = settings.buttonToClick; // button to click after buying coins
        var onComplete = settings.onComplete;
        var popup = new Dialog({
            z: settings.z,
            name: 'buyCoinsWithGems',
            ignoreAnalytics: true,
            blurriness: 4,
            title: Localization.getText('notEnough'),
            canClose: true,
            onComplete: function () {
                if (hasBought && buttonToClick) {
                    // do callback of button
                    buttonToClick.doCallback();
                }
                if (onComplete) {
                    onComplete(hasBought);
                }
            },
            onClose: function () {
                Analytics.event('Economy', 'buyCoinsWithGems', 0, 'yes');
            }
        });
        var hasBought = false;
        var gemsNeeded = (function () {
            // calculate the best deal for user
            var i;
            var gemToCoins = Globals.gemToCoins;
            var needed = 0;
            var coinsNeeded = coinsToBuy;
            var lowestCoinsPerGem = Math.ceil(gemToCoins[0].coins / gemToCoins[0].gems);

            // try to get the coins from the higher tiers assuming those get more coins per gem
            for (i = gemToCoins.length - 1; i >= 0; --i) {
                while (coinsNeeded > gemToCoins[i].coins) {
                    coinsNeeded -= gemToCoins[i].coins;
                    needed += gemToCoins[i].gems;
                }
            }
            // remains
            needed += Math.round(coinsNeeded / lowestCoinsPerGem);

            // possible the price is 0, player must pay at least 1 gem
            return needed || 1;
        })();
        var topbar = new TopBarSimple({});
        var cost = new ShadowText({
            position: new Vector2(8, -1),
            text: gemsNeeded.toString(),
            font: 'font',
            fontSize: 22,
            fontColor: '#ffffff',
            align: 'left',
            textBaseline: 'middle',
            maxWidth: 48,
            linebreaks: false
        });
        var gem = new Entity({
            name: 'gem',
            position: new Vector2(-16, 0),
            components: [
                new Sprite({
                    imageName: 'gui/gem',
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });
        var icon = new Entity({
            name: 'icon',
            components: [
                gem,
                cost
            ]
        });
        var buttonBuy = new IconButton({
            name: 'buttonBuy',
            sfx: 'sfx_button',
            position: new Vector2(0, 68),
            imageName: 'gui2/misc/button_large_primary_excited',
            frameCountX: 6,
            frameCountY: 7,
            padding: 1,
            icon: icon,
            iconAlpha: 0.5,
            animations: {
                up: {
                    speed: 0.34,
                    frames: [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35]
                },
                down: {
                    frames: [37]
                },
                inactive: {
                    frames: [38]
                }
            },
            onClick: function () {
                var showCoinReward = function () {
                    var coinReward = new CoinReward({
                        z: settings.z + 1,
                        coins: coinsToBuy,
                        onComplete: function () {
                            // close popup
                            popup.close();
                        }
                    });
                    Bento.objects.attach(coinReward);
                    buttonBuy.setActive(false);
                };
                // buy coins with the gems needed
                if (CurrencyManager.canSpendGems(gemsNeeded)) {
                    CurrencyManager.spendGems(gemsNeeded, 'coins');
                    CurrencyManager.addCoins(coinsToBuy);
                    Analytics.event('SourceSinks', 'gemsSinkedToCoin', gemsNeeded);
                    Analytics.event('SourceSinks', 'coinsSourcedFromGem', coinsToBuy);
                    showCoinReward();
                    hasBought = true;
                    Bento.audio.stopSound('sfx_buy');
                    Bento.audio.playSound('sfx_buy');
                    Analytics.event('Economy', 'buyCoinsWithGems', 1, 'yes');
                } else {
                    // not enough gems???? go to shop
                    var iapMenu = new IAPMenu({
                        z: settings.z + 1,
                        scrollTo: 'gems'
                    });
                    Bento.objects.attach(iapMenu);
                    Analytics.event('Economy', 'buyCoinsWithGems', 1, 'notEnough');
                }
            },
            onButtonDown: function () {
                EventSystem.fire('preventScroll');
                gem.position.y = 0;
                cost.position.y = -1;
            },
            onButtonUp: function () {
                gem.position.y = -1;
                cost.position.y = -2;
            }
        });
        var coinSackSprite = new Sprite({
            spriteSheet: 'coin/sack'
        });
        var coinSack = new Entity({
            name: 'coinSack',
            position: new Vector2(0, 4),
            components: [
                coinSackSprite
            ]
        });
        var coinText = new ShadowText({
            position: new Vector2(8, 0),
            text: coinsToBuy.toString(),
            font: 'font',
            fontSize: 16,
            fontColor: '#ffff00',
            align: 'left',
            textBaseline: 'middle',
            margin: new Vector2(0, 8),
            maxWidth: 648,
            maxHeight: 16,
            linebreaks: false
        });
        var text = new ShadowText({
            position: new Vector2(0, 24),
            text: Localization.getText('buyMissingCoins'),
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            maxWidth: 160 - 16,
            maxHeight: 16
        });

        popup.getComponent('windowBox', function (windowBox) {
            windowBox.attach(buttonBuy);
            windowBox.attach(text);
            windowBox.attach(coinSack);
            windowBox.attach(coinText);
        });
        coinSackSprite.setCurrentSpeed(0.22);

        popup.attach(topbar);

        // try to get constructor asap
        if (!IAPMenu) {
            bento.require(['gui/menus/iap'], function (Module) {
                IAPMenu = Module;
            });
        }

        return popup;
    };
});