/**
 * frame containing an item you can purchase with gems
 * @moduleName GemPurchase
 */
bento.define('gui/iap/gempurchase', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'gui/shadowtext',
    'components/glow',
    'managers/currency',
    'components/eventlistener'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    ShadowText,
    Glow,
    CurrencyManager,
    EventListener
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
            title: String // text to display at top of purchase frame
            item: String // which image to load
            cost: Int / Float / String // what this purchase will cost
            onClick: Function // action that happens when button is pressed
            big: bool // purchase size
        }*/
        var type = settings.type || 'gem';
        var frame = new Entity({
            z: 0,
            name: 'frame',
            components: [
                new Sprite({
                    imageName: !settings.big ? 'gui2/iap/panel_item' : 'gui/iap/framebig',
                    originRelative: new Vector2(0.5, 0)
                })
            ]
        });

        var title = new Text({
            z: 0,
            position: new Vector2(0, 15),
            text: settings.title.toString(),
            font: 'font',
            fontSize: 16,
            fontColor: '#271a2d',
            align: 'center',
            textBaseline: 'bottom',
            maxWidth: 48,
            maxHeight: 16,
            linebreaks: false
        });

        var symbol = new Entity({
            name: 'symbol',
            position: new Vector2(title.dimension.width / 2, 12),
            components: [
                new Sprite({
                    imageName: settings.symbol,
                    originRelative: new Vector2(0.5, 1),
                })
            ]
        });

        // portrait of item
        var assetName = 'gui2/iap/' + settings.item.toString();
        var hasAnimatedSprite = Utils.hasAsset('spritesheets', assetName);
        var spriteName = hasAnimatedSprite ? assetName : undefined;
        var imageName = hasAnimatedSprite ? undefined : assetName;
        var item = new Entity({
            z: 0,
            name: 'itemPortrait',
            position: new Vector2(0, 16),
            components: [
                new Sprite({
                    imageName: imageName,
                    spriteSheet: spriteName,
                    originRelative: new Vector2(0.5, 0)
                })
            ]
        });

        var glow = new Glow({
            average: 0.2,
            amplitude: 0.2,
            update: function () {
                if (type === 'gem') {
                    var gems = CurrencyManager.getGems();
                    glow.visible = gems >= settings.cost && settings.item;
                } else {
                    // keys
                    var keys = CurrencyManager.getKeys();
                    glow.visible = keys >= settings.cost && settings.item;
                }
            }
        });

        var button = new ClickButton({
            z: 0,
            name: 'item',
            sfx: 'sfx_button',
            imageName: 'gui2/misc/buttonmini_large_secondary',
            // imageName: 'gui/buttons/' + type,
            frameCountX: 1,
            frameCountY: 3,
            animations: {
                'up': {
                    frames: [0]
                },
                'down': {
                    frames: [1]
                },
                'inactive': {
                    frames: [2]
                },
            },
            position: new Vector2(0, 76),
            float: false,
            updateWhenPaused: 0,
            onClick: settings.onClick,
            onButtonDown: function () {
                gemText.position.y = 0;
                EventSystem.fire('preventScroll');
            },
            onButtonUp: function () {
                gemText.position.y = -1;
            }
        }).attach(glow);

        var gemText;
        // swap with keys text
        if (type === 'key') {
            // amount / cost
            var amountText = new ShadowText({
                position: new Vector2(-3, 0),
                text: CurrencyManager.getKeys().toString(),
                font: 'font',
                fontSize: 16,
                fontColor: CurrencyManager.getKeys() < settings.cost ? '#ff0000' : '#00ff00',
                align: 'right',
                textBaseline: 'middle',
                maxHeight: 16
            }).attach(new EventListener({
                name: 'updateKeys',
                onEvent: function (keys) {
                    amountText.setText(CurrencyManager.getKeys().toString(), {
                        fontColor: CurrencyManager.getKeys() < settings.cost ? '#ff0000' : '#00ff00'
                    });
                }
            }));
            var keyIcon = new Entity({
                name: 'key',
                position: new Vector2(-24, 0),
                components: [
                    new Sprite({
                        imageName: 'gui/key2',
                        originRelative: new Vector2(0.5, 0.5),
                    })
                ]
            });
            var totalText = new ShadowText({
                position: new Vector2(-3, 0),
                text: '/' + settings.cost.toString(),
                font: 'font',
                fontSize: 16,
                fontColor: '#ffffff',
                align: 'left',
                textBaseline: 'middle',
                maxWidth: 16,
                maxHeight: 16
            });
            gemText = new Entity({
                name: 'gemText',
                position: new Vector2(7, -1),
                components: [
                    keyIcon,
                    amountText,
                    totalText
                ]
            });
        } else {
            gemText = new ShadowText({
                z: 0,
                position: new Vector2(8, -1),
                text: settings.cost.toString(),
                font: 'font',
                fontSize: 16,
                fontColor: '#ffffff',
                align: 'center',
                textBaseline: 'middle',
                maxWidth: 24,
                linebreaks: false,
                components: [
                    new Entity({
                        name: 'gem',
                        position: new Vector2(-24, 1),
                        components: [
                            new Sprite({
                                imageName: 'gui/gem',
                                originRelative: new Vector2(0.5, 0.5),
                            })
                        ]
                    })
                ]
            });
        }

        var entity = new Entity({
            z: 0,
            name: 'gemPurchase-' + settings.title.toString(),
            position: settings.position,
            updateWhenPaused: 0,
            components: [
                frame,
                title,
                item,
                button.attach(gemText)
            ]
        }).extend({
            setActive: function (active) {
                button.setActive(active);
                if (active) {
                    gemText.alpha = 1;
                } else {
                    gemText.alpha = 0.5;
                }
            },
            isActive: function () {
                return button.isActive();
            },
            disable: function () {
                if (button.parent) {
                    entity.remove(button);
                }
            },
            enable: function () {
                if (!button.parent) {
                    entity.attach(button);
                }
            }
        });

        if (settings.symbol) {
            entity.attach(symbol);
            title.position.x -= 4;
        }
        return entity;
    };
});