/**
 * Overlay that waits for IAP callback
 * Contains general code flow for making a purchase
 * Advised to include this module asap, as it contains global event listeners also used for restoring purchases 
 * @moduleName IapOverlay
 */
bento.define('gui/iapoverlay', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/store',
    'globals',
    'modules/localization',
    'modules/analytics',
    'managers/currency',
    'components/fillscreen',
    'modules/tenjin',
    'modules/firebase'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    ClickButton,
    Counter,
    Text,
    EventSystem,
    Utils,
    Tween,
    Store,
    Globals,
    Localization,
    Analytics,
    CurrencyManager,
    FillScreen,
    Tenjin,
    FireBase
) {
    'use strict';
    var IapOverlay = function (settings) {
        /*settings = {
            // describe your settings object parameters
            timeOutTime: Number // frames until timeout, defaults to 600 (10 seconds)
            onTimeOut: Function // what should happen if the IAP timed out
        }*/
        var viewport = Bento.getViewport();
        // timeout
        var timeOutTime = settings.timeOutTime || 600;
        var onTimeOut = settings.onTimeOut;
        // dark overlay
        var overlay = new Entity({
            name: 'overlay',
            components: [new FillScreen({
                color: [0, 0, 0, 0.5],
                originRelative: new Vector2(0.5, 0.5),
            })]
        });
        var text = new Text({
            position: new Vector2(0, 0),
            text: '...',
            // TODO: localizations
            // fontSettings: Utils.FontSettings.get('font').gameover,
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            maxWidth: viewport.width - 16,
            textBaseline: 'middle'
        });
        // pauselevel is 1 higher than current pauselevel
        var pauseLevel = 0;
        var pauseBehavior = {
            oldPauseLevel: 0,
            start: function () {
                this.oldPauseLevel = Bento.objects.isPaused();
                pauseLevel = this.oldPauseLevel + 1;
                Bento.objects.pause(pauseLevel);
            },
            destroy: function () {
                Bento.objects.pause(this.oldPauseLevel);
            }
        };
        var timeOutBehavior = {
            hasTimedOut: false,
            update: function () {
                if (this.hasTimedOut === false && entity.timer > timeOutTime) {
                    this.hasTimedOut = true;
                    if (onTimeOut) {
                        onTimeOut();
                    }
                }
            }
        };
        var entity = new Entity({
            z: settings.z,
            name: 'iapOverlay',
            position: new Vector2(viewport.width / 2, viewport.height / 2),
            updateWhenPaused: pauseLevel,
            float: true,
            components: [
                overlay,
                text,
                pauseBehavior,
                timeOutBehavior
            ]
        });

        // public
        entity.extend({
            /**
             * Close overlay (with a delay)
             */
            exit: function (delay, callback) {
                Utils.timeout(delay || 0, function () {
                    Bento.objects.remove(entity);
                    if (callback) {
                        callback();
                    }
                }, pauseLevel);
            },
            /**
             * Set overlay text
             */
            setText: function (str) {
                text.setText(str);
            }
        });

        return entity;
    };
    /*
     * Global purchase events (game specific)
     */
    // global event listener (place non-consumable purchases here!!)
    EventSystem.on('purchase_success', function (productInfo) {
        var productId = productInfo.productId;

        if (productId === '???') {
            // give the product to user
        }
    });
    // analytics + tenjin
    EventSystem.on('purchase_success', function (productInfo) {
        var productId = productInfo.productId;
        var product = productInfo.product;
        var transactionId = productInfo.transactionId;
        var trackPurchase = true;
        var lastReceipt = productInfo.lastReceipt;

        if (!lastReceipt) {
            // validation failed, don't send anything to tenjin
            return;
        }

        if (product && trackPurchase) {
            // Tenjin
            if (Utils.isNativeIos()) {
                var priceString = product.price.toString();
                Tenjin.iosTransaction(
                    productInfo.productId, // sku 
                    product.currency || '', // currencycode
                    1, // quantity
                    priceString, // price
                    productInfo.transactionId, // transaction id
                    productInfo.lastReceipt // receipt in base64
                );
            } else if (Utils.isNativeAndroid()) {
                Tenjin.androidTransaction(
                    productInfo.productId, // sku 
                    product.currency || '', // currencycode
                    1, // quantity
                    product.price, // price
                    productInfo.lastReceipt.purchaseData, // purchase data
                    productInfo.lastReceipt.dataSignature // data signature
                );
            }

            Tenjin.sendEvent("validatedPayment");

            // check receipt on iOS for sandbox value
            if (!isSandbox(lastReceipt)) {
                // google analytics
                Analytics.trackPurchase(
                    productInfo.transactionId, // id
                    product.title, // title
                    productInfo.productId, // sku
                    'IAP', // category
                    product.price, // price
                    1, // quantity
                    product.currency || '' // currencycode
                );
            } else {
                Analytics.event('IAP', 'sandbox', product.price, productInfo.productId);
            }
        }

    });
    var isSandbox = function (receipt) {
        // this check is only for ios
        if (!Utils.isNativeIos()) {
            return false;
        }
        try {
            var receiptStr;
            if (window.decodeB64) {
                receiptStr = window.decodeB64(receipt);
            } else {
                // atob is not so reliable in cocoon
                receiptStr = window.atob(receipt);
            }
            // check if the receipt contains the string '"environment" = "Sandbox"'
            return receiptStr.indexOf('"environment" = "Sandbox"') >= 0;
        } catch (e) {
            Utils.log('ERROR: malformed receipt');
            return false;
        }
    };
    /*
     * Purchase flow
     */
    IapOverlay.Purchase = function (settings) {
        var overlayZ = settings.z;
        var id = settings.id; // product id
        var onPurchaseComplete = settings.onPurchaseComplete;
        var iapOverlay = new IapOverlay({
            z: overlayZ
        });
        var purchase = function () {
            var success;

            // show overlay
            Bento.objects.attach(iapOverlay);
            iapOverlay.setText(Localization.getText('purchasing'));

            // wait for callback
            EventSystem.on('purchase_success', onPurchaseSuccess);
            EventSystem.on('purchase_error', onPurchaseError);

            // purchase!
            success = Store.purchase(id);

            if (!success) {
                if (!Utils.isCocoonJS()) {
                    onPurchaseError('Only available in the App version!');
                } else {
                    onPurchaseError(Localization.getText('errorstore'));
                }
            }
        };
        var onPurchaseSuccess = function (purchaseInfo) {
            var productId = purchaseInfo.productId;
            var reward = Utils.rewardsIAP[productId];
            var iapAmount = Bento.saveState.load('iapAmount', 0);
            var highSpender = Bento.saveState.load('highSpender', false);
            var productAmount = Bento.saveState.load('productAmount-' + productId, 0);

            // increment the amount of IAP's purchased
            iapAmount++;
            Bento.saveState.save('iapAmount', iapAmount);
            productAmount++;
            Bento.saveState.save('productAmount-' + productId, productAmount);

            iapOverlay.setText(Localization.getText('thanks'));
            iapOverlay.exit(60, function () {
                if (onPurchaseComplete) {
                    onPurchaseComplete(true, id, purchaseInfo);
                }
            });
            // clean up event listeners
            cleanUp();

            // which IAP was bought
            Analytics.event('IAP', 'bought_' + productId);
            // amount of IAP's purchased along with the name of the IAP being purchased
            Analytics.event('IAP', 'iapAmount_' + iapAmount, undefined, productId);
            Analytics.event('UI', 'bought_' + productId);

            FireBase.setUserProperty('iapAmount', iapAmount);
            // FireBase.setUserProperty(productId, productAmount);
        };
        var onPurchaseError = function (err) {
            console.log('purchase failed!', JSON.stringify(err));
            Analytics.event('UI', 'purchaseError', undefined, JSON.stringify(err));
            iapOverlay.setText(Localization.getText('error') + '\n' + JSON.stringify(err));
            iapOverlay.exit(120, function () {
                if (onPurchaseComplete) {
                    onPurchaseComplete(false, id, null);
                }
            });
            cleanUp();
        };
        var cleanUp = function () {
            EventSystem.off('purchase_success', onPurchaseSuccess);
            EventSystem.off('purchase_error', onPurchaseError);
        };

        purchase();
    };

    return IapOverlay;
});