/**
 * Scrolling inventory. Items can be selected (and deselected)
 */
bento.define('gui/inventory', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'gui/scrollinglist',
    'gui/itemslot',
    'gui/centeredlist',
    'managers/inventory',
    'bento/components/nineslice',
    'components/anchor'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    ScrollingList,
    ItemSlot,
    CenteredList,
    InventoryManager,
    NineSlice,
    Anchor
) {
    'use strict';

    var Page = function (settings) {
        var slots = settings.slots;
        var columns = settings.columns;
        var rows = settings.rows;
        var page = {};
        var margin = 3;
        var spacing = 40 + margin; //40 (width of itemslot graphic) + 6 pixels margin

        var ii;
        var col;
        var row;

        page.container = new Entity({
            components: [
                new NineSlice({
                    // imageName: 'gui2/misc/panel_highlight_background_9slice',
                    imageName: 'gui2/misc/splash_background_9slice',
                    originRelative: new Vector2(0.5, 0.5),
                    width: 148 + 16,
                    height: 148 + 16
                })
            ]
        });
        page.slots = [].concat(slots);

        for (ii = 0; ii < page.slots.length; ii++) {
            col = ii % columns;
            row = Math.floor(ii / columns);

            page.slots[ii].position.x = (col - (columns - 1) / 2) * spacing;
            page.slots[ii].position.y = (row - (rows - 1) / 2) * spacing;

            page.container.attach(page.slots[ii]);
        }

        return page;
    };

    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();

        var filter = settings.filter || '';

        var items = InventoryManager.getAllItems(filter);

        var rows = settings.rows || 3;
        var columns = settings.columns || 3;
        var itemsPerPage = rows * columns;
        var maxItems = settings.maxItems;

        if (maxItems) {
            itemsPerPage = Utils.clamp(0, itemsPerPage, maxItems);
        }

        //TEMP
        (function () {
            var ii;
            for (ii = 0; ii < 12; ii++) {
                items.push(undefined);
            }
        })();


        var itemSlots = [];

        var pagesDots = new CenteredList({
            spacing: 3,
            position: new Vector2(0, 80)
        });

        var scrollingList = new ScrollingList({
            // area: new Rectangle(-72, -72, 144, 144), //TODO
            damping: 0.9,
            // spacing: Utils.clamp(180, viewport.width, 240),
            spacing: Anchor.isPortrait() ? 8 : 72,
            snap: true,
            direction: 'x',
            onSnap: function (item, index) {
                pagesDots.iterate(function (dot, i) {
                    if (i === index) {
                        dot.alpha = (1);
                    } else {
                        dot.alpha = (0.5);
                    }
                });
            }
        });

        var onSelected = function (slot) {
            Utils.forEach(itemSlots, function (s, i, l, breakLoop) {
                if (slot !== s && s.isSelected()) {
                    s.deselect();
                    if (settings.onDeselected) {
                        settings.onDeselected(s);
                    }
                }
            });

            if (settings.onSelected) {
                settings.onSelected(slot);
            }
        };

        var onDeselected = function (slot) {
            if (settings.onDeselected) {
                settings.onDeselected(slot);
            }
        };

        var addPage = function (page) {
            var idx = pagesDots.length();
            scrollingList.addItem(page.container);

            pagesDots.addItem(new Entity({
                alpha: (idx === 0) ? 1 : 0.5,
                components: [
                    new Sprite({
                        originRelative: new Vector2(0.5, 0.5),
                        imageName: 'gui/pagedot'
                    })
                ]
            }));
        };

        var createPages = function () {
            var ii;
            var page;

            var slotSet = [];
            var remainder;
            var makeEmptySlot = function () {
                var emptySlot = new ItemSlot({
                    slotSize: 'new',
                    onClick: function () {
                        // shake indicate to user that slot is empty
                        new Tween({
                            from: 2,
                            to: 0,
                            in: 40,
                            ease: 'elastic',
                            decay: 2, // for 'exponential' and 'elastic'
                            oscillations: 5, // for 'elastic', 'sin' and 'cos'
                            onUpdate: function (v, t) {
                                emptySlot.transform.x = v;
                            },
                            onComplete: function () {
                                emptySlot.transform.x = 0;
                            }
                        });
                    }
                });
                itemSlots.push(emptySlot);
            };

            pagesDots.clear();
            for (ii = 0; ii < items.length; ii++) {
                if (!items[ii]) {
                    continue;
                }
                itemSlots.push(new ItemSlot({
                    item: items[ii],
                    onSelected: onSelected,
                    onDeselected: onDeselected,
                    // slotSize: 'inset'
                    slotSize: 'new'
                }));
            }

            //fill with empty slots to complete a page;
            remainder = itemSlots.length % itemsPerPage;
            // only fill up if needed
            // in the special case of 0 items, fill up at least one page
            if (remainder || itemSlots.length === 0) {
                for (ii = 0; ii < itemsPerPage - remainder; ii++) {
                    makeEmptySlot();
                }
            }

            for (ii = 0; ii < itemSlots.length; ii += itemsPerPage) {
                slotSet = itemSlots.slice(ii, Math.min(ii + itemsPerPage, itemSlots.length));

                page = new Page({
                    slots: slotSet,
                    rows: rows,
                    columns: columns
                });
                addPage(page);
            }

            if (pagesDots.length() <= 1) {
                pagesDots.visible = false;
            }
        };

        var entity = new Entity({
            z: 0,
            name: 'inventory',
            position: settings.position || new Vector2(0, 0),
            updateWhenPaused: 0,
            components: [
                scrollingList,
                pagesDots
            ]
        }).extend({
            addItem: function (key, amount) {
                if (!key) {
                    return;
                }
                var ii;
                var slot;
                for (ii = 0; ii < itemSlots.length; ii++) {
                    if (itemSlots[ii].key === key) {
                        slot = itemSlots[ii];
                        break;
                    } else if (itemSlots[ii].key === undefined) {
                        //this is the first empty slot, getting here means there is currently no slot with the requested key.
                        slot = itemSlots[ii];
                        slot.setItem({
                            key: key,
                            amount: 0 //we add the amount later
                        });
                        break;
                    }
                }

                if (!slot) {
                    for (ii = 0; ii < itemsPerPage; ii++) {
                        itemSlots.push(new ItemSlot({}));
                    }
                    var newPage = new Page(itemSlots.slice(-itemsPerPage, itemsPerPage), rows, columns);
                    scrollingList.addItem(newPage);
                    pagesDots.addItem(new Entity({
                        alpha: 0.5,
                        components: [
                            new Sprite({
                                originRelative: new Vector2(0.5, 0.5),
                                imageName: 'gui/pagedot'
                            })
                        ]
                    }));

                    slot = itemSlots[itemSlots.length - itemsPerPage - 1];

                    //TODO scroll to the new page?
                }

                slot.setAmount(slot.amount + amount);
            },
            removeItem: function (key, amount) {
                if (!key) {
                    return;
                }
                var ii;
                var slot;

                for (ii = 0; ii < itemSlots.length; ii++) {
                    if (itemSlots[ii].key === key) {
                        slot = itemSlots[ii];
                        break;
                    }
                }

                if (!slot) {
                    return;
                }
                slot.setAmount(slot.amount - amount);
                //TODO if slot is now empty, remove it?
            }
        });

        createPages();

        return entity;
    };
});