/**
 * Inventory Popup
 * popup for inventory that waits for a single item selection
 *
 * @moduleName InventoryPopup
 */
bento.define('gui/inventorypopup', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'gui/inventory',
    'components/anchor',
    'components/fillscreen',
    'components/modal',
    'gui/iteminfo',
    'gui/frogpopup',
    'gui/blurbackground',
    'modules/localization',
    'gui/shadowtext',
    'managers/save',
    'components/backbutton'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Inventory,
    Anchor,
    FillScreen,
    Modal,
    ItemInfo,
    FrogPopup,
    BlurBackground,
    Localization,
    ShadowText,
    SaveManager,
    BackButton
) {
    'use strict';
    return function (settings) {
        var inventory;
        var hideShopKeeper = settings.hideShopKeeper || false;
        var filter = settings.filter || '';
        var onComplete = settings.onComplete;
        var titleId = settings.title || ''; // id of localization text
        var skipItemInfo = settings.skipItemInfo || false;
        var maxItems = settings.maxItems;
        var makeInventory = function () {
            inventory = new Inventory({
                rows: settings.rows,
                columns: settings.columns,
                maxItems: settings.maxItems,
                filter: filter,
                onSelected: function (slot) {
                    // show item info
                    var itemInfo = new ItemInfo({
                        item: slot.key,
                        onYes: function () {
                            // don't wait for the tween
                            Bento.objects.remove(itemInfo);

                            // close popup immediately
                            Bento.objects.remove(container);

                            // callback
                            if (onComplete) {
                                onComplete(slot);
                            }
                        }
                    });

                    // deselect immediately
                    slot.deselect();

                    if (skipItemInfo) {
                        if (onComplete) {
                            onComplete(slot);
                        }
                    } else {
                        Bento.objects.attach(itemInfo);
                    }
                }
            }).attach(new Anchor({
                screenPosition: new Vector2(0.5, 0.5)
            }));
            return inventory;
        };
        // temporary background: fills screen
        var background = new BlurBackground({
            darken: 0.3,
            blurriness: 7,
        });
        var title = new ShadowText({
            text: Localization.getText(titleId),
            font: 'font',
            fontSize: 24,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
        }).attach(new Anchor({
            screenPosition: new Vector2(0.5, 0),
            offset: new Vector2(0, 16 + 32),
            onComplete: function () {
                if (!Anchor.isPortrait()) {
                    title.position.y -= 8 + 32;
                    title.setText(title.getText(), {
                        fontSize: 16
                    });
                } else {
                    title.setText(title.getText(), {
                        fontSize: 24
                    });
                }
            }
        }));
        var buttonClose = new ClickButton({
            name: 'btnClose',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs__back',
            frameCountX: 2,
            frameCountY: 1,
            onClick: function () {
                new Tween({
                        from: 1,
                        to: 0,
                        in: 10,
                        ease: 'linear',
                        onUpdate: function (v, t) {
                            container.alpha = v;
                        },
                        onComplete: function () {
                            // remove own modal before removing container
                            this.removeByName('modal');
                            Bento.objects.remove(container);
                            // callback
                            if (onComplete) {
                                onComplete(null);
                            }
                        }
                    })
                    // pause to prevent pressing any buttons during tween
                    .attach(new Modal({}));
            },
            components: [
                new BackButton({
                    removeSelf: true,
                    onPressed: function (data) {
                        data.entity.doCallback();
                    }
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(-48 + (Anchor.isWideScreen() ? 4 : 0), -14 + (Anchor.isWideScreen() ? Anchor.getOffset().y : 0))
                    },
                    landscape: {
                        screenPosition: new Vector2(0.25, 1),
                        offset: new Vector2(-44, -12)
                    }
                })
            ]
        });
        var btnClose = new ClickButton({
            name: 'btnClose',
            sfx: 'sfx_button',
            imageName: 'gui/buttons/exit',
            frameCountX: 2,
            frameCountY: 1,
            onClick: function () {
                new Tween({
                        from: 1,
                        to: 0,
                        in: 10,
                        ease: 'linear',
                        onUpdate: function (v, t) {
                            container.alpha = v;
                        },
                        onComplete: function () {
                            // remove own modal before removing container
                            this.removeByName('modal');
                            Bento.objects.remove(container);
                            // callback
                            if (onComplete) {
                                onComplete(null);
                            }
                        }
                    })
                    // pause to prevent pressing any buttons during tween
                    .attach(new Modal({}));
            }
        }).attach(new Anchor({
            screenPosition: new Vector2(0, 0),
            offset: new Vector2(20, 20)
        }));
        // show shop keeper
        var shopKeeper = new FrogPopup({
            z: settings.z + 1,
            onClosedShop: function () {
                // the user may have bought new items
                var oldIndex = container.components.indexOf(inventory);
                // refresh the inventory
                inventory.removeSelf();
                // make new inventory
                makeInventory();
                // add and move back to old position
                container.attach(inventory);
                container.moveComponentTo(inventory, oldIndex);
            }
        });
        var container = new Entity({
            z: settings.z || 0,
            name: 'inventoryPopup',
            float: true,
            components: [
                new Modal({
                    pauseLevel: settings.pauseLevel 
                }),
                background,
                title,
                makeInventory(),
                // shopKeeper,
                // btnClose,
                buttonClose
            ]
        });

        // don't show shopkeeper until tutorial is over
        var unlocked = SaveManager.load('unlockedLevels');
        if (unlocked.length < 5 || hideShopKeeper) {
            container.remove(shopKeeper);
        }

        return container;
    };
});