/**
 * Item Slot
 *
 * @moduleName ItemSlot
 */
bento.define('gui/itemslot', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'managers/item',
    'components/debug',
    'globals',
    'gui/shadowtext',
    'gui/tier'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    ItemManager,
    Debug,
    Globals,
    ShadowText,
    TierStars
) {
    'use strict';

    return function (settings) {

        var isSelected = false;
        var item = settings.item;
        var touchId = null;
        var touchPos = new Vector2(0, 0);
        var stopClick = false;
        var onClick = settings.onClick;

        var alwaysClickable = settings.clickable || false;
        var slotSize = settings.slotSize || 'large';
        var getImageName = function () {
            if (slotSize === 'large') {
                return 'gui/itemslot';
            } else if (slotSize === 'inset') {
                return 'gui/itemslot-inset';
            } else if (slotSize === 'new') {
                return 'gui2/loadout/item_box';
            } else {
                return 'gui/itemslot-small';
            }
        };
        var slotSprite = new Sprite({
            imageName: getImageName(),
            frameCountX: 2,
            originRelative: new Vector2(0.5, 0.5),
            animations: {
                default: {
                    speed: 0,
                    frames: [0]
                },
                selected: {
                    speed: 0,
                    frames: [1]
                }
            }
        });
        var itemImage = new Entity({
            z: 0,
            position: new Vector2(-1, 0),
            components: [

            ]
        });

        var clickable = new Clickable({
            pointerUp: function (data) {
                if (touchId !== data.id) {
                    return;
                }

                stopClick = true;
            },
            pointerMove: function (data) {
                if (touchId !== data.id) {
                    return;
                }

                if (touchPos.distance(data.position) > 10) {
                    stopClick = true;
                }
            },
            // when clicking on the object
            onClick: function (data) {
                if (onClick) {
                    onClick();
                }
                if (!item && !alwaysClickable) {
                    return;
                }
                if (touchId !== null) {
                    return;
                }
                touchId = data.id;
                touchPos = data.position.clone();

                Bento.audio.stopSound('sfx_button');
                Bento.audio.playSound('sfx_button');

                if (slotSize === 'inset') {
                    return;
                }
                new Tween({
                    from: 1,
                    to: 0.95,
                    'in': 8,
                    ease: Tween.EASEOUTBACK,
                    onUpdate: function (v, t) {
                        slot.scale.x = v;
                        slot.scale.y = v;
                    }
                });
            },
            onClickUp: function (data) {
                if (touchId !== data.id) {
                    return;
                }
                isSelected = !isSelected;

                slotSprite.setAnimation(isSelected ? 'selected' : 'default');

                stopClick = true;

                if (isSelected && settings.onSelected) {
                    settings.onSelected(slot);
                } else if (!isSelected && settings.onDeselected) {
                    settings.onDeselected(slot);
                }
            }
        });

        var behavior = {
            name: 'behavior',
            update: function (data) {
                if (stopClick) {
                    touchId = null;

                    stopClick = false;

                    if (slotSize === 'inset') {
                        return;
                    }

                    new Tween({
                        from: 0.95,
                        to: 1,
                        'in': 16,
                        ease: Tween.EASEOUTBOUNCE,
                        onUpdate: function (v, t) {
                            slot.scale.x = v;
                            slot.scale.y = v;
                        }
                    });
                }
            }
        };

        var itemCount = new ShadowText({
            z: 0,
            position: new Vector2(12, 8),
            text: '0',
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle'
        });

        var slot = new Entity({
            z: 0,
            name: 'itemSlot',
            position: new Vector2(0, 0),
            components: [
                slotSprite,
                clickable,
                behavior
                // new Debug({})
            ]
        }).extend({
            key: undefined,
            amount: 0,
            setItem: function (newItem) {
                item = newItem;

                if (!item || !item.key) {
                    if (itemImage.parent) {
                        slot.remove(itemImage);
                        slot.remove(itemCount);
                    }
                    slot.getComponent('tierStars', function (tierStars) {
                        slot.remove(tierStars);
                    });
                    slot.key = undefined;
                    slot.amount = 0;
                    return;
                }

                itemImage.getComponent('sprite', function (sprite) {
                    itemImage.remove(sprite);
                });

                itemImage.attach(ItemManager.getSprite(item.key));
                itemImage.getComponent('sprite', function (sprite) {
                    sprite.setOriginRelative(new Vector2(0.5, 0.5));
                });

                if (!itemImage.parent) {
                    slot.attach(itemImage);
                    slot.attach(itemCount);
                }

                slot.key = item.key;
                slot.setAmount(newItem.amount);

                // remove old and add new tier
                if (slotSize !== 'small') {
                    slot.getComponent('tierStars', function (tierStars) {
                        slot.remove(tierStars);
                    });

                    slot.attach(new TierStars({
                        position: new Vector2(-16, -10),
                        tier: ItemManager.getTier(item.key),
                        align: 'left',
                    }));
                }
            },
            getItem: function () {
                return slot.key;
            },
            setAmount: function (count) {
                itemCount.setText(count);
                slot.amount = count;

                itemCount.visible = (count > 1);
                itemImage.alpha = (count <= 0) ? 0 : 1;
            },
            isSelected: function () {
                return isSelected;
            },
            select: function () {
                isSelected = true;
                slotSprite.setAnimation('selected');
            },
            deselect: function () {
                isSelected = false;
                slotSprite.setAnimation('default');
            }
        });

        if (item) {
            slot.setItem(item);
        }

        return slot;
    };
});