/**
 * The dialog containing players currently in connected in the online room
 * Used by gui/menus/onlinejoin and gui/menus/onlinecreate
 * @moduleName LobbyWindow
 */
bento.define('gui/lobbywindow', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/localization',
    'gui/headerbox',
    'components/modal',
    'gui/blurbackground',
    'globals',
    'components/anchor',
    'gui/loadingspinner',
    'gui/loadingdialog',
    'gui/dialog',
    'modules/matchmanager',
    'gui/shadowtext',
    'managers/save'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Localization,
    HeaderBox,
    Modal,
    BlurBackground,
    Globals,
    Anchor,
    LoadingSpinner,
    LoadingDialog,
    Dialog,
    MatchManager,
    ShadowText,
    SaveManager
) {
    'use strict';

    var PlayerContainer = function (settings) {
        var index = settings.index;
        var create = settings.create; // create adds an offset to make room for the share button
        var loadingSpinner = new LoadingSpinner({
            position: new Vector2(-48, 2),
            balls: 6,
            radius: 7
        });
        var playerNameText = new ShadowText({
            position: new Vector2(-24, 0),
            // text: Localization.getText('onlineWaiting'),
            text: '',
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'left',
            textBaseline: 'middle',
            maxWidth: 100,
            maxHeight: 16
        });
        var entity = new Entity({
            name: 'container',
            position: new Vector2(0, create ? -80 + 32 * index : -80 + 32 * index),
            components: [
                playerNameText,
                loadingSpinner
            ]
        });

        entity.extend({
            uniqueId: null,
            reset: function () {
                // remove old icon if any
                entity.removeByName('playerIcon');
                // add loading spinner
                entity.attach(loadingSpinner);
                // set name to waiting
                // playerNameText.setText(Localization.getText('onlineWaiting'));
                playerNameText.setText('');
                // reset uniqueId
                entity.uniqueId = null;
            },
            setCharacter: function (character, name, level) {
                var playerIcon = new Entity({
                    name: 'playerIcon',
                    position: new Vector2(-48, 2),
                    scale: new Vector2(0.75, 0.75),
                    components: [
                        new Sprite({
                            originRelative: new Vector2(0.5, 0.5),
                            imageName: 'gui2/characterchange/hud_' + character
                        })
                    ]
                });

                // remove old icon if any
                entity.removeByName('playerIcon');
                // remove loading spinner
                entity.remove(loadingSpinner);

                entity.attach(playerIcon);
                playerNameText.setText(name + (level ? ' ' + Localization.getText('lv.') + level : ''));

                // tween in
                new Tween({
                    from: 0,
                    to: 0.75,
                    in: 40,
                    ease: 'elastic',
                    decay: 7,
                    oscillations: 3, // for 'elastic', 'sin' and 'cos'
                    onUpdate: function (v, t) {
                        playerIcon.scale.x = v;
                        playerIcon.scale.y = v;
                    }
                });

                new Tween({
                    from: 0,
                    to: 1,
                    in: 40,
                    delay: 10,
                    ease: 'elastic',
                    decay: 7,
                    oscillations: 3, // for 'elastic', 'sin' and 'cos'
                    onCreate: function () {
                        playerNameText.scale.x = 0;
                        playerNameText.scale.y = 0;
                    },
                    onUpdate: function (v, t) {
                        playerNameText.scale.x = v;
                        playerNameText.scale.y = v;
                    }
                });
            }
        });
        return entity;
    };

    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var create = settings.create;
        var roomCode = settings.roomCode;
        // callbacks
        var onRoomCodeChanged = settings.onRoomCodeChanged;
        var onRoomJoined = settings.onRoomJoined;
        var onRoomList = settings.onRoomList;
        var onMatchStart = settings.onMatchStart;
        var onPlayerCountChanged = settings.onPlayerCountChanged;
        // window stuff
        var maxPlayers = 2;
        var currentPlayers = 1;
        var maxPlayersPhoton = 4;
        var playerCountText = new ShadowText({
            position: new Vector2(0, -60),
            text: Localization.getText('onlinePlayerCount').replace('{CURRENT}', currentPlayers).replace('{TOTAL}', maxPlayers),
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle'
        });
        var p1Container = new PlayerContainer({
            index: 1,
            create: create
        });
        var p2Container = new PlayerContainer({
            index: 2,
            create: create
        });
        var p3Container = new PlayerContainer({
            index: 3,
            create: create
        });
        var p4Container = new PlayerContainer({
            index: 4,
            create: create
        });
        var containers = [p1Container, p2Container /*, p3Container, p4Container*/ ];
        var headerBox = new HeaderBox({
                title: Localization.getText('onlinePlayerCount').replace('{CURRENT}', currentPlayers).replace('{TOTAL}', maxPlayers),
                size: new Rectangle(0, 0, 180, 196),
                windowImage: 'gui2/misc/splash_background_9slice',
            })
            // .attach(playerCountText)
            .attach(p1Container)
            .attach(p2Container)
            // .attach(p3Container)
            // .attach(p4Container)
            .attach(new Anchor({
                screenPosition: new Vector2(0.5, 0.5),
                offset: new Vector2(0, 0)
            }));
        var getContainerByActorNr = function (actorNr) {
            var current = null;
            Utils.forEach(containers, function (container, i, l, breakLoop) {
                if (container.uniqueId === actorNr) {
                    current = container;
                    breakLoop();
                    return;
                }
            });
            return current;
        };
        var countPlayers = function () {
            currentPlayers = 0;
            Utils.forEach(containers, function (container, i, l, breakLoop) {
                if (container.uniqueId) {
                    currentPlayers += 1;
                }
            });
            // update text
            // playerCountText.setText(Localization.getText('onlinePlayerCount').replace('{CURRENT}', currentPlayers).replace('{TOTAL}', maxPlayers));
            headerBox.setTitle(Localization.getText('onlinePlayerCount').replace('{CURRENT}', currentPlayers).replace('{TOTAL}', maxPlayers));
            return currentPlayers;
        };
        var behavior = {
            name: 'behaviorComponent',
            start: function () {
                var randomizeRoomCode = function () {
                    if (settings.roomCode) {
                        // already has roomCode
                        roomCode = settings.roomCode;
                        return settings.roomCode;
                    } 
                    // randomize a room number
                    roomCode = Utils.getRandom(10000);
                    // always 4 digits
                    roomCode = ('0000' + roomCode).slice(-4);
                    // headerBox.setTitle(Localization.getText('roomCode') + ' ' + roomCode);
                    if (onRoomCodeChanged) {
                        onRoomCodeChanged(roomCode);
                    }
                    return roomCode;
                };
                var onLogin = function (status) {
                    var roomList = status.roomList;
                    var joinMatch = function () {
                        // join the roomCode provided
                        entity.visible = true;
                        if (settings.mission === 'tower') {
                            maxPlayersPhoton = 2;
                        }
                        MatchManager.findMatch(maxPlayersPhoton, undefined, onStart, {
                            roomCode: roomCode
                        });
                    };
                    var createMatch = function () {
                        if (settings.mission === 'tower') {
                            maxPlayersPhoton = 2;
                        }
                        MatchManager.findMatch(maxPlayersPhoton, undefined, onStart, {
                            create: true,
                            roomCode: roomCode,
                            enemyLevel: settings.enemyLevel,
                            chosenDifficulty: settings.chosenDifficulty,
                            cost: settings.cost,
                            rewards: settings.rewards,
                            unlocks: settings.unlocks,
                            mission: settings.mission,
                            time: Date.now(),
                            reputation: SaveManager.load('reputationPoints')
                        });
                    };
                    var onStart = function (data) {
                        if (data.success) {
                            onRoomCreated(data);
                            if (onRoomJoined) {
                                onRoomJoined(MatchManager.currentMatch.room);
                            }
                        } else if (create) {
                            // failed, try again with new code
                            randomizeRoomCode();
                            createMatch();
                        } else {
                            // tried to join but failed
                            // possible scenarios:
                            // -host closed the game
                            // -host timed out while tru
                            // -room got filled by others

                            var dialog = new Dialog({
                                name: 'error_connectFail',
                                title: 'Error',
                                text: data.error,
                                canClose: false,
                                type: 'yes',
                                onClose: function () {
                                    // close menu
                                    entity.removeSelf();
                                }
                            });
                            Bento.objects.attach(dialog);
                        }
                    };
                    Bento.objects.remove(loadingLogin);

                    if (!status.success) {
                        var dialog = new Dialog({
                            name: 'error_connectFail',
                            title: 'Error',
                            text: Localization.getText('connectFail'),
                            canClose: false,
                            type: 'yes',
                            onClose: function () {
                                // close menu
                                entity.removeSelf();
                            }
                        });
                        Bento.objects.attach(dialog);
                    } else if (create) {
                        // randomize room code
                        randomizeRoomCode();

                        // try creating this room
                        createMatch();
                    } else {
                        // before joining room: player must have enough energy to start this game!!
                        onRoomList(roomList, joinMatch);
                        entity.visible = false;
                    }
                };
                var onRoomCreated = function (data) {
                    // inspect the actors currently in the room
                    Utils.forEach(data.match.players, function (playerData, i, l, breakLoop) {
                        var name = playerData.userName;
                        // photon only: character is added to playerData
                        var character = playerData.character;
                        var level = playerData.level;
                        if (containers[i]) {
                            containers[i].setCharacter(character, name, level);
                            containers[i].uniqueId = playerData.uniqueId;
                        }
                    });

                    // wait for other players
                    EventSystem.on(MatchManager.Events.PLAYERCONNECTS, behavior.onPlayerConnects);
                    EventSystem.on(MatchManager.Events.PLAYERDISCONNECTS, behavior.onPlayerDisconnects);

                };
                var loadingLogin = new LoadingDialog({
                    text: Localization.getText('connecting')
                });
                // connect first
                if (!MatchManager.isLoggedIn()) {
                    Bento.objects.attach(loadingLogin);
                    MatchManager.login(onLogin, {
                        name: SaveManager.load('localPlayerName'),
                        character: SaveManager.load('selectedCharacter'),
                        level: SaveManager.load('level-' + SaveManager.load('selectedCharacter')),
                        seed: Globals.towerInfo.seed
                    });
                } else {

                }
                // wait for startMatch message
                MatchManager.onMessage('startMatch', behavior.onStartMessage);
            },
            onStartMessage: function (data) {
                var playerIdMap = data.msg;
                // photon only: determine playerId
                if (MatchManager.currentMatch) {
                    MatchManager.currentMatch.initPlayerIds(playerIdMap);
                } else {
                    Utils.log("Match not initialized");
                    return;
                }

                if (onMatchStart) {
                    onMatchStart();
                }

                // quest event for other players
                EventSystem.fire('quest-online', {
                    count: 1
                });

                return true;
            },
            onPlayerConnects: function (stateObj) {
                var playerData = stateObj.playerData;
                var name = playerData.userName;

                // photon only: need character data and actor data
                var character = playerData.character;
                var level = playerData.level;

                // find an empty container and add player to that
                var actorNr = playerData.uniqueId;
                // should not already be added
                if (getContainerByActorNr(actorNr)) {
                    // Utils.log('player already added');
                    return;
                }
                Utils.forEach(containers, function (container, i, l, breakLoop) {
                    if (container.uniqueId === null) {
                        container.setCharacter(character, name, level);
                        container.uniqueId = actorNr;
                        breakLoop();
                        return;
                    }
                });
                if (onPlayerCountChanged) {
                    onPlayerCountChanged(countPlayers(), maxPlayers);
                }

            },
            onPlayerDisconnects: function (stateObj) {
                var playerData = stateObj.playerData;
                var actorNr = playerData.uniqueId;
                var container = getContainerByActorNr(actorNr);

                if (container) {
                    container.reset();
                }

                if (onPlayerCountChanged) {
                    onPlayerCountChanged(countPlayers(), maxPlayers);
                }
            },
            destroy: function (data) {
                EventSystem.off(MatchManager.Events.PLAYERCONNECTS, behavior.onPlayerConnects);
                EventSystem.off(MatchManager.Events.PLAYERDISCONNECTS, behavior.onPlayerDisconnects);
                MatchManager.offMessage('startMatch', behavior.onStartMessage);
            }
        };
        var entity = new Entity({
            z: settings.z,
            name: 'lobbyWindow',
            position: settings.position,
            components: [
                headerBox,
                behavior
            ]
        });

        entity.extend({
            setRoomCode: function (code) {
                roomCode = code;
            },
            // marks the beginning of a match, should only be called by the master
            // e.g. only the person who is hosting should have a button that will call this function
            startMatch: function () {
                var players;
                var playerIdMap = {};
                var room;
                // photon only: determine playerId
                if (MatchManager.currentMatch) {
                    players = MatchManager.currentMatch.initPlayerIds();
                    Utils.forEach(players, function (playerData, i, l, breakLoop) {
                        playerIdMap[playerData.uniqueId] = playerData.playerId;
                    });

                    // close the room
                    room = MatchManager.currentMatch.room;
                    room.setIsOpen(false);
                    room.setIsVisible(false);

                } else {
                    Utils.log("Match not initialized");
                    return;
                }

                // send all other peers a start message
                MatchManager.sendData({
                    playerIdMap: playerIdMap,
                }, 'startMatch');

                if (onMatchStart) {
                    onMatchStart({});
                }

                // quest event for host
                EventSystem.fire('quest-online', {
                    count: 1
                });
            }
        });

        return entity;
    };
});