/**
 * Make emotes in multiplayer
 * @moduleName EmoteMenu
 */
bento.define('gui/main/emotemenu', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'bento/gui/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'gui/iconbutton',
    'gui/headerbox',
    'modules/matchmanager',
    'gui/speechballoon',
    'components/anchor',
    'managers/save',
    'components/eventlistener'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    IconButton,
    HeaderBox,
    MatchManager,
    SpeechBalloon,
    Anchor,
    SaveManager,
    EventListener
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var emotesUnlocked = ['happy', 'wink', 'whatever', 'sad', 'angry', 'help', 'good', 'bad'];
        var emoteAmountHalved = Math.ceil(emotesUnlocked.length / 2);
        var emoteButtons = [];
        var emoteContainer = new Entity({
            name: 'emoteContainer',
            position: new Vector2(5, 0),
        });
        var viewport = Bento.getViewport();
        var busy = false;
        var opened = false;
        var moving = false;
        var offset = 0;
        var leftHanded = SaveManager.load('leftHanded');

        var alignMenu = function () {
            offset = Anchor.getOffset().x;
            emoteContainer.position.x = leftHanded ? -5 : 5;
            if (!moving) {
                if (opened) {
                    openButton.position.x = -8 - offset;
                    windowBox.position.x = 24 - offset;
                } else {
                    openButton.position.x = 8 - offset;
                    windowBox.position.x = -40 - offset;
                }
            }

            //align Y
            var windowY = Math.min((viewport.height - 52) / 2 + 52 - entity.position.y, 0);
            windowBox.position.y = windowY;
        };

        var anchor = new Anchor({
            offset: new Vector2(0, 120),
            onComplete: alignMenu
        });

        var openButton = new ClickButton({
            z: 0,
            name: 'openButton',
            sfx: 'sfx_button',
            imageName: 'gui2/emote/open',
            frameCountX: 1,
            frameCountY: 1,
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [0]
                }
            },
            position: new Vector2(8, 0),
            updateWhenPaused: 0,
            float: false,
            onClick: function () {
                // open if not already open
                if (!opened) {
                    openEmoteWindow();
                }
            }
        });

        // box that holds the emotes (-40 and 24)
        var windowBox = new HeaderBox({
            name: 'windowBox',
            size: new Rectangle(0, 0, 48 + 4 + 28, 22 * emoteAmountHalved + 4 * (emoteAmountHalved - 1) + 28),
            position: new Vector2(-40, 0),
            windowImage: 'gui2/misc/splash_background_9slice',
        });

        var clickable = Clickable({
            pointerDown: function (data) {
                if (opened && !moving) {
                    var bb = windowBox.dimension.offset(entity.position.add(windowBox.position));
                    if (!bb.hasPosition(data.position)) {
                        closeEmoteWindow();
                    }
                }
            }
        });

        var openEmoteWindow = function () {
            // lock orientation?
            opened = true;
            moving = true;
            new Tween({
                from: -40,
                to: 24,
                in: 30,
                delay: 0,
                ease: 'easeOutBack',
                onUpdate: function (v, t) {
                    windowBox.position.x = v - offset;
                },
                onComplete: function () {
                    moving = false;
                }
            });

            // remove openbutton
            new Tween({
                from: 8,
                to: -8,
                in: 5,
                delay: 0,
                ease: 'easeInQuad',
                onUpdate: function (v, t) {
                    openButton.position.x = v - offset;
                }
            });
            EventSystem.fire('disableControls');
        };

        var closeEmoteWindow = function () {
            moving = true;
            EventSystem.fire('enableControls');
            new Tween({
                from: 24,
                to: -40,
                in: 6,
                delay: 0,
                ease: 'easeInQuad',
                onUpdate: function (v, t) {
                    windowBox.position.x = v - offset;
                },
                onComplete: function () {
                    opened = false;
                    moving = false;
                }
            });

            // remove openbutton
            new Tween({
                from: -8,
                to: 8,
                in: 5,
                delay: 12,
                ease: 'easeOutQuad',
                onUpdate: function (v, t) {
                    openButton.position.x = v - offset;
                }
            });
        };

        var createEmoteButton = function (emote) {
            var button = new ClickButton({
                z: 0,
                name: 'emoteButton',
                sfx: 'sfx_button',
                imageName: 'gui2/emote/' + emote,
                frameCountX: 1,
                frameCountY: 1,
                onClick: function () {
                    var playerId = Math.max(MatchManager.localPlayer.playerId, 0);
                    if (!busy && !moving) {
                        createEmoteBalloon(emote, 'player-' + playerId);
                        closeEmoteWindow();
                        // tell other players to make this emote
                        var msg = {
                            p: playerId,
                            e: emote
                        };
                        MatchManager.sendData(msg, 'emotes');

                        // make sure cannot spam emotes
                        busy = true;
                        Utils.timeout(120, function () {
                            busy = false;
                        });
                    }
                }
            });
            return button;
        };

        var fillButtonContainer = function () {
            Utils.forEach(emotesUnlocked, function (emote, i, l, breakLoop) {
                var emoteButton = createEmoteButton(emote);
                emoteButtons.push(emoteButton);
                emoteContainer.attach(emoteButton);
                emoteButton.position.x = -13 + i % 2 * 26;
                emoteButton.position.y = -(26 * emoteAmountHalved - 4) / 2 + 13 + Math.floor(i / 2) * 26;
            });
            windowBox.attach(emoteContainer);
        };

        var receiveMessage = function (msg) {
            createEmoteBalloon(msg.e, 'player-' + msg.p);
            return true;
        };

        var createEmoteBalloon = function (emote, target) {
            // play emote
            var balloon = new SpeechBalloon({
                typewriter: false,
                touchControl: false,
                text: ' ',
                position: new Vector2(0, 0),
                bottomPosition: -5,
                bottomFlip: false,
                size: new Rectangle(0, 0, 36, 36),
                z: Globals.Layers.GUI + 4,
            });
            var t = Bento.objects.getByName(target);
            balloon.attach({
                name: 'follower',
                update: function () {
                    if (t && t[0]) {
                        var balloonPosition = new Vector2(0, 0);
                        var balloonDimension = balloon.dimension;
                        var margin = new Vector2(4 + balloonDimension.width / 2, balloonDimension.height / 2);
                        balloon.position.x = t[0].position.x;
                        balloon.position.y = t[0].position.y - 48 - balloonDimension.height / 2;
                    }
                }
            });
            new Tween({
                from: 0,
                to: 1,
                in: 20,
                delay: 0,
                ease: 'elastic',
                decay: 5,
                oscillations: 1,
                onUpdate: function (v, tt) {
                    balloon.scale.x = v;
                    balloon.scale.y = v;
                },
            });
            Bento.objects.attach(balloon);

            // attach emote to balloon
            var emoji = new Entity({
                name: 'animatedEmoji',
                position: new Vector2(0, 0),
                components: [
                    new Sprite({
                        spriteSheet: 'emote/' + emote
                    })
                ]
            });
            balloon.attach(emoji);

            // remove balloon after a while
            new Tween({
                from: 1,
                to: 0,
                in: 10,
                delay: 90,
                ease: 'linear',
                onUpdate: function (v, tt) {
                    balloon.alpha = v;
                },
                onComplete: function () {
                    balloon.removeSelf();
                }
            });
        };

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                fillButtonContainer();
                MatchManager.onMessage('emotes', receiveMessage);
            },
            destroy: function (data) {
                MatchManager.offMessage('emotes', receiveMessage);
            },
            update: function (data) {},
            draw: function (data) {}
        };
        var removeOnVictory = new EventListener({
            name: 'bossWin',
            onEvent: function (data) {
                entity.removeSelf();
            }
        });
        var entity = new Entity({
            z: Globals.Layers.GUI,
            name: 'emoteMenu',
            family: ['gui'],
            position: settings.position,
            updateWhenPaused: 0,
            components: [
                openButton,
                windowBox,
                behavior,
                anchor,
                clickable,
                removeOnVictory
            ],
            float: true
        });
        return entity;
    };
});