/**
 * Health bar
 * @moduleName HealthBar
 */
bento.define('gui/main/healthbar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'bento/components/nineslice',
    'components/anchor',
    'managers/character',
    'modules/localization',
    'managers/save',
    'gui/shadowtext',
    'gui/main/combobar'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    NineSlice,
    Anchor,
    CharacterManager,
    Localization,
    SaveManager,
    ShadowText,
    ComboBar
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var character = CharacterManager.getSelected();
        var portrait = new Entity({
            name: 'portrait',
            position: new Vector2(5, 5),
            components: [
                new Sprite({
                    imageName: 'gui2/characterchange/hud_' + character
                }),
                new Clickable({
                    onClickUp: function (data) {
                        if (Bento.isDev() || Bento.saveState.load('showDeveloperCheats', false)) {
                            window.killEveryone(true);
                        }
                    }
                })
            ]
        });
        var healthSlice = new Sprite({
            imageName: 'gui2/hud/health_bar_red'
        });
        // overwrite draw
        healthSlice.width = 86;
        healthSlice.setWidth = function (value) {
            healthSlice.width = value;
        };
        healthSlice.draw = function (data) {
            var entity = data.entity;
            var width = Math.max(healthSlice.width, 0);
            if (width <= 0) {
                return;
            }

            if (!this.currentAnimation || !this.visible) {
                return;
            }

            this.updateFrame();

            data.renderer.drawImage(
                this.spriteImage,
                this.sourceX,
                this.sourceY,
                width,
                this.frameHeight,
                (-this.origin.x) | 0,
                (-this.origin.y) | 0,
                width,
                this.frameHeight
            );
        };

        var healthbar = new Entity({
            name: 'healthbar',
            position: new Vector2(41, 13),
            components: [
                healthSlice
            ]
        });

        var hud = new Entity({
            name: 'overlay',
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'gui2/hud/hud'
                }),
                new Sprite({
                    imageName: 'gui2/hud/padding',
                    originRelative: new Vector2(0, 1)
                })
            ]
        });

        var levelText = new ShadowText({
            text: Localization.getText('lv.'),
            font: 'font',
            fontSize: 8,
            fontColor: '#fffd00',
            align: 'right',
            textBaseline: 'bottom',
            position: new Vector2(52, 40),
            maxWidth: 12,
            maxHeight: 16,
            linebreaks: false
        });
        var levelNumberText = new ShadowText({
            text: SaveManager.load('level-' + character).toString(),
            font: 'font',
            fontSize: 12,
            fontColor: '#fffd00',
            align: 'left',
            textBaseline: 'bottom',
            position: new Vector2(52, 41),
            maxWidth: 16,
            maxHeight: 16,
            linebreaks: false,
            drawDebug: false
        });

        // combo bar for super attacks
        var combobar = new ComboBar({});

        var behavior = {
            name: 'behavior',
            fraction: 1,
            start: function (data) {
                settings.target.getComponent('punchable', function (punchable) {
                    punchable.onHealthChange(function (params) {
                        behavior.fraction = params.relativeHealth;
                    });
                });
                // set initial value if possible
                if (settings.health && settings.maxHealth) {
                    entity.setPercentage(settings.health / settings.maxHealth * 100);
                }
            },
            destroy: function (data) {},
            draw: function (data) {
                healthSlice.setWidth(Math.max(this.fraction * 86, 0.01));
            }
        };

        var entity = new Entity({
            z: Globals.Layers.GUI,
            name: 'healthBar',
            position: new Vector2(0, 0),
            float: true,
            family: ['gui'],
            components: [
                hud,
                healthbar,
                combobar,
                portrait,
                levelText,
                levelNumberText,
                behavior
            ]
        });
        entity.extend({
            shake: function () {
                new Tween({
                    from: 4,
                    to: 0,
                    in: 40,
                    ease: 'elastic',
                    decay: 2,
                    oscillations: 5,
                    onUpdate: function (v, t) {
                        entity.transform.x = v;
                    },
                    onComplete: function () {
                        entity.transform.x = 0;
                    }
                });
            },
            setPercentage: function (percent) {
                behavior.fraction = percent / 100;
            },
            removeComboBar: function (percent) {
                new Tween({
                    from: 1,
                    to: 0,
                    in: 15,
                    ease: 'easeInQuad',
                    onUpdate: function (v, t) {
                        combobar.alpha = v;
                    },
                    onComplete: function () {
                        combobar.visible = false;
                    }
                });
            },
            addComboPoint: function () {
                combobar.add(1);
            },
            canSpendComboPoints: function (amount) {
                return combobar.canSpend(amount);
            },
            depleteComboPoints: function (amount) {
                return combobar.deplete(amount);
            },
            depleteComboBar: function () {
                combobar.depleteAll();
            },
            fillCombobar: function (amount) {
                combobar.fillUp(amount);
            },
            flashCombo: function () {
                combobar.flash();
            }
        });

        entity.attach(new Anchor({
            screenPosition: new Vector2(0, 0),
            offset: new Vector2(0, 0),
            onComplete: function () {
                if (Anchor.isWideScreen()) {
                    if (!Anchor.isPortrait()) {
                        entity.position.x -= Anchor.getOffset().x;
                        entity.position.y += 2;
                    }
                }
            }
        }));

        return entity;
    };
});