/**
 * Beach event menu
 * @moduleName BeachEventMenu
 */
bento.define('gui/menus/beachevent', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/components/sprite',
    'components/anchor',
    'globals',
    'components/modal',
    'gui/iconbutton',
    'components/backbutton',
    'bento/gui/text',
    'gui/buttons/startbutton',
    'gui/shadowtext',
    'modules/localization',
    'managers/save',
    'managers/time',
    'managers/currency',
    'bento/components/nineslice',
    'gui/tokenbar',
    'gui/timertab',
    'gui/iap/buyticketspopup',
    'gui/equipmentslots',
    'managers/inventory'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Sprite,
    Anchor,
    Globals,
    Modal,
    IconButton,
    BackButton,
    Text,
    StartButton,
    ShadowText,
    Localization,
    SaveManager,
    TimeManager,
    CurrencyManager,
    NineSlice,
    TokenBar,
    TimerTab,
    BuyTicketsPopup,
    EquipmentSlots,
    InventoryManager
) {
    'use strict';
    /**
     * Schedule next ticket into the future
     */
    var scheduleNextTicket = function () {
        var nextTicket = SaveManager.load('nextTicket');
        var tickets = SaveManager.load('tickets');
        var now = TimeManager.getTime();
        var maxTickets = Globals.maxTickets;

        // schedule next ticket (if nextTicket is not already in the future)
        if (tickets < maxTickets && nextTicket.time <= now) {
            // schedule X minutes into the future
            nextTicket.time = now + 1000 * 60 * Globals.minutesForTicket;
        }
        SaveManager.save('nextTicket', nextTicket);
    };
    /**
     * Check for any regenerated tickets (call this every second)
     */
    var checkTickets = function () {
        var maxTickets = Globals.maxTickets;
        var now = TimeManager.getTime();
        var nextTicket = SaveManager.load('nextTicket');
        var tickets = SaveManager.load('tickets');
        var timeFrame = 1000 * 60 * Globals.minutesForTicket;
        var didChange;

        // the time passed to get nextTicket
        while (tickets < maxTickets && now > nextTicket.time) {
            tickets += 1;
            // schedule next ticket
            nextTicket.time += timeFrame;
            didChange = true;
        }

        // nextTicket is scheduled in the future while tickets is already full
        if (tickets >= maxTickets && nextTicket.time > now) {
            nextTicket.time = now;
        }

        SaveManager.save('tickets', tickets);
        SaveManager.save('nextTicket', nextTicket);

        return didChange;
    };
    var isBeachEvent = function () {
        var today = Math.floor(TimeManager.getTime() / 1000 / 60 / 60 / 24);
        if (today >= 17521 && today <= 17526) {
            return true;
        }
        if (Bento.isDev() && today <= 17526) {
            // console.log('Testing rad event');
            return true;
        }
        if (Bento.saveState.load('radEventTest', false)) {
            return true;
        }
        return false;
    };
    var getRemainingEventTimeText = function () {
        var now = TimeManager.getTime();
        var endTime = 17527 * 1000 * 60 * 60 * 24;
        var timeLeft;
        var minutes;
        var hours;

        // set time text
        timeLeft = Math.max(0, endTime - now);
        minutes = Math.floor(timeLeft / 1000 / 60) % 60;
        hours = Math.floor(timeLeft / 1000 / 60 / 60);
        return (Localization.getText('offerTime').replace('{HOURS}', hours).replace('{MINUTES}', minutes));
    };

    var Beach = function (settings) {
        var viewport = Bento.getViewport();
        var beachEvent = SaveManager.load('beachEventData');
        var eventMissionsCleared = SaveManager.load('eventMissionsCleared');
        var currentMission = eventMissionsCleared % 5;
        var isNight = Math.floor((eventMissionsCleared % 10) / 5);
        var currentTickets = SaveManager.load('tickets');
        var maxTickets = Globals.maxTickets;

        /**
         * Background
         */
        var bgEntity = new Entity({
            z: 0,
            name: 'beachMap',
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'gui2/beachevent/map',
                    originRelative: new Vector2(0.5, 0.5),
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 0.5),
                        offset: new Vector2(0, 0),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.5, 0.6),
                        offset: new Vector2(0, 0),
                    },
                    onComplete: function () {}
                })
            ]
        });
        /**
         * PLAY
         */
        var ticketCost = 1;
        var costText = new Text({
            position: new Vector2(15, 5),
            text: ticketCost.toString(),
            font: 'font',
            fontSize: 12,
            fontColor: '#ffe773',
            align: 'left',
            textBaseline: 'middle',
            lineWidth: 1.5,
            strokeStyle: '#81532d',
            pixelStroke: true,
        });
        var costIcon = new Entity({
            name: 'costIcon',
            position: new Vector2(-12 - 24, 0),
            components: [
                new Sprite({
                    imageName: 'gui2/beachevent/ticket',
                    originRelative: new Vector2(0, 0.5)
                }),
                costText
            ]
        });
        var playText = new ShadowText({
            position: new Vector2(8, 0),
            originRelative: new Vector2(0, 0.5),
            text: Localization.getText('play'),
            font: 'font',
            fontSize: 22,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            strokeStyle: '#000000',
            maxWidth: 44,
            maxHeight: 22
        });
        var container = new Entity({
            name: 'icon',
            components: [
                costIcon,
                playText
            ]
        });
        var buttonPlay = new IconButton({
            name: 'startButton',
            sfx: 'sfx_button',
            icon: container,
            iconAlpha: 0.5,
            imageName: 'gui2/misc/button_medium_primary',
            frameCountX: 1,
            frameCountY: 3,
            onClick: function () {
                //start beach event
                var startButton = new StartButton({
                    isEvent: true,
                    mission: (isNight ? 'beachnight' : 'beach') + currentMission,
                    chosenDifficulty: 0,
                    difficulty: 0.2 + currentMission * 0.1,
                    cost: 0,
                    unlocks: [],
                    rewards: []
                });
                if (CurrencyManager.canSpendTickets(ticketCost)) {
                    CurrencyManager.spendTickets(ticketCost);
                    scheduleNextTicket();

                    startButton.start();
                } else {
                    notEnoughTickets(1, buttonPlay);
                }
            },
            components: [
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(0, -44),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(68, -44),
                    },
                    onComplete: function () {

                    }
                })
            ]
        });
        /**
         * Close
         */
        var buttonClose = new IconButton({
            name: 'buttonClose',
            sfx: 'sfx_button',
            imageName: 'gui2/misc/button_smaller',
            iconName: 'gui2/misc/button_icon_back_small',
            frameCountX: 1,
            frameCountY: 3,
            onClick: function () {
                entity.removeSelf();
                if (settings.onComplete) {
                    settings.onComplete();
                }
            },
            components: [
                new BackButton({
                    removeSelf: true,
                    onPressed: function (data) {
                        data.entity.doCallback();
                    }
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(-64, -60 + 16)
                    },
                    landscape: {
                        screenPosition: new Vector2(0, 0),
                        offset: new Vector2(20, 20)
                    }
                })
            ]
        });

        /**
         * Stages
         */
        var bounceTween;
        var currentStage = currentMission + 1;
        var stagePosition = 32.5 * (currentStage - 1) - 65.5;
        var stagesEntity = new Entity({
            z: 0,
            name: 'beachStages',
            position: new Vector2(0, 0),
            components: [
                new Sprite({
                    imageName: 'gui2/beachevent/stages',
                    originRelative: new Vector2(0.5, 0),
                }),
                new Entity({
                    z: 0,
                    name: 'beachStagesMarker',
                    position: new Vector2(stagePosition, -10),
                    components: [
                        new Sprite({
                            imageName: 'gui2/beachevent/marker',
                            originRelative: new Vector2(0.5, 0, 5),
                        }),
                    ]
                }),
                new Entity({
                    name: 'rewardBalloon',
                    position: new Vector2(stagePosition + 5, -32),
                    components: [
                        new NineSlice({
                            imageName: 'gui/speechbub-map',
                            width: 36,
                            height: 36,
                            originRelative: new Vector2(0.5, 0.5)
                        }),
                        new Entity({
                            name: 'radToken',
                            position: new Vector2(1, -1),
                            components: [
                                new Sprite({
                                    spriteSheet: 'playertokens/rad',
                                })
                            ]
                        })
                    ]
                }),
                new Entity({
                    name: 'bottom',
                    position: new Vector2(stagePosition + 11, -18),
                    scale: new Vector2(-1, 1),
                    components: [
                        new Sprite({
                            imageName: 'gui/speechbub-map-bottom',
                            originRelative: new Vector2(0.5, 0)
                        })
                    ]
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 0.4),
                        offset: new Vector2(0, 0),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.5, 0.55),
                        offset: new Vector2(0, 0),
                    },
                    onComplete: function () {
                        if (typeof bounceTween == "object") {
                            bounceTween.removeSelf();
                        }
                        bounceTween = bounceMarker();
                    }
                })
            ]
        });
        var bounceMarker = function () {
            return new Tween({
                from: -10,
                to: -15,
                in: 20,
                delay: 0,
                applyOnDelay: 0,
                ease: 'easeOutQuad',
                decay: 1,
                oscillations: 1,
                onStart: function () {},
                onUpdate: function (v, t) {
                    stagesEntity.components[1].position.y = v;
                },
                onComplete: function () {
                    bounceTween = new Tween({
                        from: -15,
                        to: -10,
                        in: 20,
                        delay: 0,
                        applyOnDelay: 0,
                        ease: 'easeInQuad',
                        decay: 1,
                        oscillations: 1,
                        onStart: function () {},
                        onUpdate: function (v, t) {
                            stagesEntity.components[1].position.y = v;
                        },
                        onComplete: function () {
                            bounceTween = bounceMarker();
                        }
                    });
                }
            });
        };

        /**
         * Ticket counter
         */
        var timerTab = new TimerTab({
            saveId: 'nextTicket',
            useLocal: false,
            position: new Vector2(-3, 7)
        });
        var ticketsText = new ShadowText({
            text: currentTickets.toString() + "/" + maxTickets.toString(),
            font: 'font',
            fontSize: 48,
            align: 'center',
            textBaseline: 'middle',
            linebreaks: false,
            fontColor: (currentTickets > 0) ? '#ffffff' : '#9442b3',
            maxWidth: 60,
            maxHeight: 20,
            components: [
                new Anchor({
                    offset: new Vector2(-2, 10),
                    onComplete: function (parent) {}
                })
            ]
        });
        var ticketCounter = new Entity({
            name: 'ticketCounter',
            components: [
                new Sprite({
                    imageName: 'gui2/beachevent/ticketcounter',
                    originRelative: new Vector2(0.7, 0, 5),
                }),
                ticketsText,
                timerTab,
                new IconButton({
                    position: new Vector2(32, 11),
                    name: 'addTickets',
                    sfx: 'sfx_button',
                    imageName: 'gui2/misc/buttontiny',
                    iconName: 'gui2/misc/button_icon_up',
                    iconAlpha: 0.15,
                    frameCountX: 1,
                    frameCountY: 3,
                    animations: {
                        up: {
                            frames: [0]
                        },
                        down: {
                            frames: [1]
                        },
                        inactive: {
                            frames: [2]
                        }
                    },
                    components: [],
                    onClick: function () {
                        notEnoughTickets();
                    }
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(0, -100),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(-68, -54),
                    },
                    onComplete: function () {
                        if (typeof bounceTween == "object") {
                            bounceTween.removeSelf();
                        }
                        bounceTween = bounceMarker();
                    }
                })
            ]
        });

        /**
         * Title
         */
        var rescue = Localization.getText('rescue');
        var titleText = rescue ? Localization.getText('name-rad').toUpperCase() : Localization.getText('radEvent');
        var subtitleText = rescue.toUpperCase();
        var titleEntity = new Entity({
            name: 'titleEntity',
            components: [
                new ShadowText({
                    text: titleText,
                    font: 'font',
                    fontSize: 48,
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: rescue ? false : true,
                    fontColor: '#ffffff',
                    maxWidth: 80,
                    maxHeight: rescue ? 35 : 48,
                    components: [
                        new Anchor({
                            portrait: {
                                screenPosition: new Vector2(0.75, 0),
                                offset: new Vector2(0, rescue ? 30 : 38),
                            },
                            landscape: {
                                screenPosition: new Vector2(0.5, 0),
                                offset: new Vector2(68, 32),
                            },
                            onComplete: function (parent) {
                                if (!Anchor.isPortrait()) {
                                    parent.setText(titleText + " " + subtitleText, {
                                        maxWidth: 160
                                    });
                                } else {
                                    parent.setText(titleText, {
                                        maxWidth: 80
                                    });
                                }
                            }
                        })
                    ]
                }),

                new ShadowText({
                    text: subtitleText, //Localization.getText('beachEventSubTitle'),
                    font: 'font',
                    fontSize: 48,
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: false,
                    fontColor: '#ffffff',
                    maxWidth: 80,
                    maxHeight: 26,
                    components: [
                        new Anchor({
                            portrait: {
                                screenPosition: new Vector2(0.75, 0),
                                offset: new Vector2(0, 55),
                            },
                            landscape: {
                                screenPosition: new Vector2(0.75, 0),
                                offset: new Vector2(0, -100), //TODO: find out how to hide properly
                            },
                        })
                    ]
                })
            ]
        });

        /**
         * Tokens needed
         */
        var character = "rad";
        var tokenBack = new Entity({
            name: 'tokenBack',
            position: new Vector2(0, -24),
            components: [
                new NineSlice({
                    imageName: 'gui2/characterchange/tokens_required_9slice',
                    width: 96,
                    height: 40,
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });
        var tokenSprite = new Sprite({
            spriteSheet: 'playertokens/' + character
        });
        var tokenIcon = new Entity({
            name: 'tokenIcon',
            position: new Vector2(-20, 0),
            components: [
                tokenSprite
            ]
        });
        var tokenBar = new TokenBar({
            position: new Vector2(14, 0),
            character: character,
            type: 'playertokens'
        });
        var tokensRequired = new Entity({
            name: 'tokensRequired',
            position: new Vector2(60, 30),
            components: [
                tokenBar,
                tokenIcon,
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.25, 0),
                        offset: new Vector2(0, 32),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.25, 0),
                        offset: new Vector2(0, 32),
                    },
                    onComplete: function () {}
                })
            ]
        });

        /**
         * End time
         */
        var timeLeftText = new ShadowText({
            text: getRemainingEventTimeText(),
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            maxWidth: 80,
            maxHeight: 16,
            linebreaks: false,
            components: [
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.25, 0),
                        offset: new Vector2(0, 52),
                    },
                    landscape: {
                        screenPosition: new Vector2(0.25, 0),
                        offset: new Vector2(0, 52),
                    },
                    onComplete: function () {}
                })
            ]
        });

        var notEnoughTickets = function (ticketsToBuy, buttonToClick) {
            //open popup to buy tickets
            var popup = new BuyTicketsPopup({
                z: entity.z + 1,
                ticketsToBuy: ticketsToBuy || Utils.clamp(0, Globals.maxTickets - currentTickets, 5),
                buttonToClick: buttonToClick,
                onComplete: function (didBuy) {
                    if (didBuy) {
                        currentTickets = SaveManager.load('tickets');
                        if (currentTickets >= Globals.maxTickets) {
                            // no need anymore to wait for next ticket
                            SaveManager.save('nextTicket', {
                                time: TimeManager.getTime()
                            });
                        }
                        updateText();
                    }
                }
            });
            Bento.objects.attach(popup);
        };
        var updateText = function () {
            ticketsText.setText(currentTickets.toString() + "/" + maxTickets.toString(), {
                fontColor: (currentTickets > 0) ? '#ffffff' : '#9442b3'
            });
        };
        var autoSelectItems = function () {
            var consumables = new EquipmentSlots({
                iconImage: 'gui/startlevel/inventory',
                type: 'consumable',
                slotSize: 'new',
                // hideSlot2: true,
                position: new Vector2(0, 4),
            });
            var itemsInInventory;
            var slot1 = consumables.getSlot(0);
            var slot2 = consumables.getSlot(1);
            var randomItem;

            itemsInInventory = InventoryManager.getAllItems(['consumables', 'weapons']);

            // fill slot 1
            if (!slot1 && itemsInInventory.length > 0) {
                randomItem = itemsInInventory[Utils.getRandom(itemsInInventory.length)];
                consumables.setSlot(0, randomItem.key);
            }

            // refresh inventory
            itemsInInventory = InventoryManager.getAllItems(['consumables', 'weapons']);

            // fill slot 2
            if (!slot2 && itemsInInventory.length > 0) {
                randomItem = itemsInInventory[Utils.getRandom(itemsInInventory.length)];
                consumables.setSlot(1, randomItem.key);
            }

        };
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                //hide bottom tab backbutton on this screen
                Bento.objects.get('buttonBack', function (buttonBack) {
                    buttonBack.alpha = 0;
                });

                if (!isBeachEvent()) {
                    // event is over
                    buttonClose.doCallback();
                }

                if (SaveManager.load('autoSelect')) {
                    autoSelectItems();
                }
            },
            destroy: function (data) {
                Bento.objects.get('buttonBack', function (buttonBack) {
                    buttonBack.alpha = 1;
                });
            },
            update: function (data) {
                var didChange;
                if (entity.timer % 60 === 0) {
                    // check for ticket regeneration every second
                    // console.log('Tickets: ' + SaveManager.load('tickets'));
                    didChange = checkTickets();
                    if (didChange) {
                        currentTickets = SaveManager.load('tickets');
                        updateText();
                    }
                }
            },
            draw: function (data) {}
        };

        var entity = new Entity({
            z: Globals.WorldSelectLayers.TOWERCHALLENGE,
            name: 'BeachEventMenu',
            position: new Vector2(0, 5, 0, 5),
            float: true,
            components: [
                new Modal({}),
                bgEntity,
                behavior,
                buttonPlay,
                buttonClose,
                tokensRequired,
                timeLeftText,
                titleEntity,
                stagesEntity,
                ticketCounter
            ]
        });
        return entity;
    };

    Beach.isBeachEvent = isBeachEvent;

    return Beach;
});