/**
 * "Container" for starting a new level, adds blur background and topbar
 * The actual menus will be spawned within
 * @moduleName MissionDetails
 */
bento.define('gui/menus/missiondetails', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'gui/blurbackground',
    'globals',
    'components/modal',
    'gui/worldselect/topbarsimple',
    'components/anchor',
    'gui/levelstart/playercount',
    'gui/levelstart/rewardspopup',
    'gui/buttons/startbutton',
    'gui/menus/onlinecreate',
    'gui/menus/onlinejoin',
    'components/viewanalytics',
    'modules/matchmanager',
    'managers/resolution',
    'components/orientationlock',
    'gui/dialog',
    'modules/localization',
    'managers/save',
    'managers/stats'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    BlurBackground,
    Globals,
    Modal,
    TopBarSimple,
    Anchor,
    PlayerCountMenu,
    RewardsPopup,
    StartButton,
    OnlineCreate,
    OnlineJoin,
    ViewAnalytics,
    MatchManager,
    ResolutionManager,
    OrientationLock,
    Dialog,
    Localization,
    SaveManager,
    StatsManager
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var invitation = settings.invitation;
        var onClosed = settings.onClosed;
        var blurBackground = new BlurBackground({
            blurriness: 8,
            darken: 0.3,
        });
        var modal = new Modal({});
        var topbar = new TopBarSimple({});
        var buttonClose = new ClickButton({
            name: 'buttonClose',
            position: new Vector2(20, 20),
            sfx: 'sfx_button',
            imageName: 'gui/buttons/exit',
            frameCountX: 2,
            frameCountY: 1,
            onClick: function () {
                if (rewardMenu) {
                    rewardMenu.closeMenu();
                    return;
                }
                if (playerCountMenu && playerCountMenu.isShowingOnline()) {
                    playerCountMenu.closeOnline();
                    return;
                }
                new Tween({
                    from: 1,
                    to: 0,
                    in: 6,
                    ease: 'linear',
                    onUpdate: function (v, t) {
                        entity.alpha = v;
                    },
                    onComplete: function () {
                        this.removeByName('modal');
                        Bento.objects.remove(entity);
                        if (joinMenu && joinMenu.isAdded) {
                            // joinMenu may be hidden when trying to close
                            joinMenu.disconnect();
                            joinMenu.removeSelf();
                        }
                        if (onClosed) {
                            onClosed();
                        }
                    }
                }).attach(new Modal({}));
            }
        });
        var playerCountMenu;
        var joinMenu;
        var rewardMenu;
        var behavior = {
            name: 'behaviorComponent',
            start: function () {
                var createJoinMenu = function () {
                    // show numberpad before showing reward info
                    joinMenu = new OnlineJoin({
                        onFoundRoom: function (missionData, onSuccess) {
                            // disable modal
                            joinMenu.removeByName('modal');
                            // hide lobbywindow
                            joinMenu.hide();
                            // joinMenu has a blurbackground with a hidebelow component, remove joinMenu and later re attach it
                            Bento.objects.remove(joinMenu);

                            // roomCode was inserted and room was found, show mission/reward info before actually joining
                            showRewardsMenu({
                                isSingle: false,
                                isJoining: true,
                                missionData: missionData,
                                joinRoom: function () {
                                    onSuccess();
                                    // restore modal
                                    joinMenu.attach(new Modal({}));
                                    // show lobbywindow
                                    joinMenu.show();
                                    Bento.objects.attach(joinMenu);
                                }
                            });
                        },
                        onClose: function () {
                            if (joinMenu.hasJoinedRoom()) {
                                // pressing the close button presses the upper close button
                                buttonClose.doCallback();
                            }
                        }
                    });
                    Bento.objects.attach(joinMenu);

                    if (invitation) {
                        joinMenu.showLobbyWindow(invitation, true);
                    }
                };
                var showPlayersMenu = function () {
                    playerCountMenu = new PlayerCountMenu({
                        cost: settings.cost,
                        mission: settings.mission,
                        unlocks: settings.unlocks,
                        rewards: settings.rewards,
                        onSinglePlayer: function () {
                            // carry on to reward info and player settings
                            showRewardsMenu({
                                isSingle: true,
                                isJoining: false,
                                missionData: settings
                            });
                        },
                        onMultiPlayerCreate: function () {
                            // show reward info and player settings before OnlineCreate menu
                            showRewardsMenu({
                                isSingle: false,
                                isJoining: false,
                                missionData: settings
                            });
                        },
                        onMultiPlayerJoin: function () {
                            createJoinMenu();
                        }
                    });
                    playerCountMenu.attach(new Anchor({
                        screenPosition: new Vector2(0.5, 0.5),
                        offset: new Vector2(0, 8),
                        onComplete: function (parent, isPortrait) {
                            if (!isPortrait) {
                                playerCountMenu.position.y += 8;
                            }
                        }
                    }));
                    entity.attach(playerCountMenu);

                    // tween in
                    new Tween({
                        from: 0,
                        to: 1,
                        in: 30,
                        ease: 'elastic',
                        decay: 8,
                        oscillations: 1,
                        onUpdate: function (v, t) {
                            playerCountMenu.scale.x = v;
                            playerCountMenu.scale.y = v;
                        }
                    });
                };
                var showRewardsMenu = function (data) {
                    var isSingle = data.isSingle;
                    var isJoining = data.isJoining;
                    var missionData = data.missionData;
                    var joinRoom = data.joinRoom;
                    var isTweening = false;
                    var tweenBack = function () {
                        // when 2nd close button is pressed
                        if (isTweening) {
                            return;
                        }
                        // disconnect when joining is closed to prevent a bug
                        if (isJoining) {
                            MatchManager.disconnect();
                        }
                        isTweening = true;
                        new Tween({
                            from: rewardMenu.position.x,
                            to: rewardMenu.position.x + viewport.width * (isSingle ? -1 : 1),
                            in: 10,
                            ease: 'exponential',
                            decay: 4,
                            onUpdate: function (v, t) {
                                rewardMenu.position.x = v;
                            },
                            onComplete: function () {
                                rewardMenu.removeSelf();
                                rewardMenu = null;
                                isTweening = true;
                            }
                        });
                        new Tween({
                            from: playerCountMenu.position.x,
                            to: viewport.width / 2,
                            in: 10,
                            ease: 'exponential',
                            decay: 4,
                            onUpdate: function (v, t) {
                                playerCountMenu.position.x = v;
                            },
                            onComplete: function () {
                                playerCountMenu.enableButtons();
                            }
                        });
                    };

                    // tween out playerCount
                    if (!invitation) {
                        playerCountMenu.removeByName('anchor');
                        playerCountMenu.disableButtons();
                        new Tween({
                            from: playerCountMenu.position.x,
                            to: playerCountMenu.position.x - viewport.width * (isSingle ? -1 : 1),
                            in: 10,
                            ease: 'exponential',
                            decay: 4,
                            onUpdate: function (v, t) {
                                playerCountMenu.position.x = v;
                            }
                        });
                    }

                    // rewards + player details
                    rewardMenu = new RewardsPopup({
                        isOnline: !isSingle,
                        mission: missionData.mission,
                        cost: missionData.cost,
                        unlocks: missionData.unlocks,
                        enemyLevel: missionData.enemyLevel,
                        rewards: missionData.rewards,
                        invitation: invitation,
                        onClick: function () {
                            var dialog;
                            var currentCharacter;
                            // single player: start the game
                            if (isSingle) {
                                currentCharacter = SaveManager.load('selectedCharacter');
                                // check if the enemy level is 4 higher
                                if (StatsManager.getLevel(currentCharacter) + 4 <= missionData.enemyLevel) {
                                    // check if mission is not too difficult
                                    dialog = new Dialog({
                                        name: 'tooDifficult',
                                        type: 'yesno',
                                        title: Localization.getText('warningTitle'),
                                        text: Localization.getText('warnTooDifficult'),
                                        canClose: false,
                                        onYes: function () {
                                            startGame(missionData);
                                        }
                                    });
                                    Bento.objects.attach(dialog);
                                } else {
                                    startGame(missionData);
                                }
                            } else {
                                if (!isJoining) {
                                    // go to lobby (this will start the game later)
                                    var createMenu = new OnlineCreate({
                                        enemyLevel: settings.enemyLevel,
                                        cost: settings.cost,
                                        mission: settings.mission,
                                        unlocks: settings.unlocks,
                                        rewards: settings.rewards
                                    });
                                    Bento.objects.attach(createMenu);
                                } else {
                                    // show windowlobby
                                    joinRoom();
                                }
                            }
                        }
                    });
                    rewardMenu.attach(new Anchor({
                        screenPosition: new Vector2(0.5, 0.5),
                        offset: new Vector2(0, 8),
                        onComplete: function (parent, isPortrait) {
                            if (!isPortrait) {
                                rewardMenu.position.y += 8;
                            }
                        }
                    }));
                    entity.attach(rewardMenu);
                    rewardMenu.extend({
                        closeMenu: function () {
                            tweenBack();
                        }
                    });

                    // tween in
                    new Tween({
                        from: rewardMenu.position.x + viewport.width * (isSingle ? -1 : 1),
                        to: rewardMenu.position.x,
                        in: 10,
                        ease: 'exponential',
                        decay: 4,
                        onUpdate: function (v, t) {
                            rewardMenu.position.x = v;
                        }
                    });

                    // 2nd close button
                    /*var buttonClose2 = new ClickButton({
                        name: 'buttonClose2',
                        position: new Vector2(20, 20),
                        sfx: 'sfx_button',
                        imageName: 'gui/buttons/exit',
                        frameCountX: 2,
                        frameCountY: 1,
                        onClick: function () {
                            entity.attach(buttonClose);
                            buttonClose2.removeSelf();
                            tweenBack();
                        }
                    });

                    if (!joinMenu) {
                        entity.attach(buttonClose2);
                        buttonClose.removeSelf();
                    }*/
                };
                var startGame = function (missionData) {
                    var startButton = new StartButton({
                        enemyLevel: missionData.enemyLevel,
                        cost: missionData.cost,
                        mission: missionData.mission,
                        unlocks: missionData.unlocks,
                        rewards: missionData.rewards
                    });
                    startButton.getComponent('buttonPlay').doCallback();
                };

                new Tween({
                    from: 0,
                    to: 1,
                    in: 15,
                    onUpdate: function (v, t) {
                        blurBackground.alpha = v;
                    }
                });

                // spawn in single/multiplayer menu
                if (!invitation) {
                    showPlayersMenu();
                } else {
                    createJoinMenu();
                }
            },
            destroy: function (data) {},
            update: function (data) {},
            draw: function (data) {}
        };
        var entity = new Entity({
            z: Globals.WorldSelectLayers.LEVELDETAILS,
            name: 'missionDetails',
            float: true,
            components: [
                modal,
                blurBackground,
                topbar,
                buttonClose,
                behavior,
                // don't rotate during mission details at all
                new OrientationLock({
                    name: 'missionDetails'
                })
            ]
        });
        return entity;
    };
});