/**
 * Notification popup
 * @moduleName Notification
 */
bento.define('gui/notification', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'bento/components/nineslice',
    'gui/shadowtext',
    'components/anchor',
    'globals',
    'components/debug',
    'modules/localization'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    NineSlice,
    ShadowText,
    Anchor,
    Globals,
    Debug,
    Localization
) {
    'use strict';
    var Notification = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var width = 140;
        var height = 40;
        var icon = settings.icon;

        var text = new ShadowText({
            position: new Vector2(0, -2),
            text: settings.text,
            font: 'font',
            fontSize: 16,
            fontColor: '#ffffff',
            align: 'center',
            textBaseline: 'middle',
            maxWidth: width - 12,
            maxHeight: height - 12
        });
        var offset = new Vector2(20, -height / 2 - 4);
        var anchor = new Anchor({
            screenPosition: new Vector2(0.5, 0),
            offset: offset
        });
        var iconBack = new Entity({
            name: 'iconBack',
            position: new Vector2(-10 - width / 2, 0),
            components: [
                new Sprite({
                    imageName: 'gui2/misc/notification_alert',
                    originRelative: new Vector2(0.5, 0.5)
                })
            ]
        });
        // window
        var nineSlice = new NineSlice({
            originRelative: new Vector2(0.5, 0.5),
            imageName: 'gui2/misc/notification_9slice',
            width: width,
            height: height
        });
        var windowBox = new Entity({
            name: 'windowBox',
            components: [
                iconBack,
                nineSlice,
                text,
                anchor
            ]
        });

        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                var clickable = new Clickable({
                    onClick: function () {
                        tweenOut();
                    }
                });
                var tweenIn = function () {
                    // drop down
                    new Tween({
                        global: true,
                        from: -height / 2 - 8,
                        to: height + 8,
                        in: 30,
                        ease: 'exponential',
                        decay: 1,
                        updateWhenPaused: Globals.PauseLevels.SYSTEM,
                        ignoreGameSpeed: true,
                        onUpdate: function (v, t) {
                            offset.y = v;
                            entity.position.y = v;
                        },
                        onComplete: function () {
                            entity.attach(clickable);
                            new Tween({
                                global: true,
                                in: 120,
                                updateWhenPaused: Globals.PauseLevels.SYSTEM,
                                ignoreGameSpeed: true,
                                onComplete: tweenOut
                            });
                        }
                    });
                };
                var isTweeningOut = false;
                var tweenOut = function (time) {
                    if (isTweeningOut) {
                        return;
                    }
                    isTweeningOut = true;
                    new Tween({
                        global: true,
                        from: height + 8,
                        to: -height / 2 - 8,
                        in: time || 30,
                        ease: 'exponential',
                        decay: 1,
                        ignoreGameSpeed: true,
                        updateWhenPaused: Globals.PauseLevels.SYSTEM,
                        onUpdate: function (v, t) {
                            offset.y = v;
                            entity.position.y = v;
                        },
                        onComplete: function () {
                            // remove self
                            entity.removeSelf();
                            currentPopup = null;

                            // call next popup
                            showNext();
                        }
                    });
                };
                tweenIn();
            }
        };
        var entity = new Entity({
            z: Globals.Layers.NOTIFICATIONS,
            name: 'notificationPopup',
            position: settings.position,
            float: true,
            global: true,
            updateWhenPaused: Globals.PauseLevels.SYSTEM,
            boundingBox: new Rectangle(0, -height, width, height),
            components: [
                behavior,
                windowBox
            ]
        });

        // spawn standard exclamation for icon
        if (!icon) {
            icon = new Entity({
                name: 'icon',
                components: [
                    new Sprite({
                        imageName: 'gui2/misc/button_icon_alert',
                        originRelative: new Vector2(0.5, 0.5)
                    })
                ]
            });
        }
        icon.position.x -= 4;
        iconBack.attach(icon);

        return entity;
    };

    // create popup and schedule it to show
    Notification.add = function (text, icon) {
        var popup = new Notification({
            text: text,
            icon: icon
        });
        popups.push(popup);

        if (!currentPopup) {
            showNext();
        }
        return popup;
    };

    // show next popup in line
    var showNext = function () {
        if (currentPopup) {
            return;
        }

        var popup = popups.shift();
        if (popup) {
            currentPopup = popup;
            Bento.objects.attach(currentPopup);
        }
    };
    var popups = [];
    var currentPopup = null;

    // catch events for notifications
    EventSystem.on('Notification', function (data) {
        if (data.textId) {
            data.text = Localization.getText(data.textId);
        }
        Notification.add(data.text, data.icon);
    });

    return Notification;
});