/**
 * Scrolling list of items
 * @moduleName ScrollingList
 * @snippet ScrollingList.snippet
ScrollingList({
    z: 0,
    position: new Vector2(0, 0),
    updateWhenPaused: 0,
    area: new Rectangle(0, 0, 0, 0), // area that user can grab
    direction: 'x', // 'x' or 'y'
    spacing: 0, // additional spacing between items
    maxOffset: 0, // additional max scrolling distance
    minOffset: 0, // additional min scrolling distance
    snap: false, // snap items (cancels maxOffset and minOffset)
    onSnap: function (item, index) {}, // called when snap focus changed
    onLoseFocus: function (index) {}, // called when snap focus changed
    onScrollStart: function () {}, // callback when scrolling has started
    snapMinSlideSpeed: 1, // minimum slidespeed before starting to snap
    snapSpeed: 0.3, // speed of snapping
    damping: 0.9 // speed decrease,
});
 */
bento.define('gui/scrollinglist', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    return function (settings) {
        /*settings = {
            z: Number
            position: Vector2
            updateWhenPaused: Boolean
            subPixel: Boolean
            area: Rectangle // area that user can grab
            direction: String // 'x' or 'y'
            spacing: Number // additional spacing between items
            maxOffset: Number // additional max scrolling distance
            minOffset: Number // additional min scrolling distance
            snap: Boolean // snap items (cancels maxOffset and minOffset)
            onSnap: Function // called when snap focus changed
            onLoseFocus: Function // called when snap focus changed
            onScrollStart: Function // callback when scrolling has started
            snapMinSlideSpeed: Number // minimum slidespeed before starting to snap
            snapSpeed: Number // speed of snapping
            damping: Number // speed decrease,
        }*/
        var viewport = Bento.getViewport(),
            items = [],

            // scroll variables
            originalPos = 0,
            holdPos = 0,
            grabX = 0,
            diffX = 0,
            isHolding = null,
            speedX = 0,
            avgSpeedX = 0,
            slideSpeed = 0,
            previousSelection = 0,
            currentSelection = 0,
            release = false,
            dir = settings.direction || 'y', // 'x' or 'y'
            dim = (dir === 'x' ? 'width' : 'height'),
            otherDim = (dir === 'x' ? 'height' : 'width'),
            spacing = settings.spacing || 0,
            snap = settings.snap || false,
            maxOffset = snap ? 0 : settings.maxOffset || 0, // snap and extra offset doesn't make sense
            minOffset = snap ? 0 : settings.minOffset || 0,
            clamp = settings.clamp || false,
            snapMinSlideSpeed = settings.snapMinSlideSpeed || 1,
            snapSpeed = settings.snapSpeed || 0.3,
            damping = settings.damping || 0.9,
            forcedSnap = -1,
            totalSize = 0,
            itemWidth = 0,
            previousPos = 0,
            currentPos = 0,
            active = true,
            // main entity
            entity = new Entity({
                z: settings.z,
                name: 'scrollingList',
                position: settings.position || new Vector2(0, 0),
                updateWhenPaused: settings.updateWhenPaused || 0,
                components: [

                    new Clickable({
                        pointerDown: function (evt) {
                            if (!active) {
                                return;
                            }
                            if (isHolding !== null) {
                                // already scrolling
                                return;
                            }
                            var y = container.position[dir];
                            grabX = evt.position[dir];

                            holdPos = evt.position[dir];
                            originalPos = container.position[dir];
                            // TEMP don't scroll if only 1 item (there should be a nicer solution for this)
                            if (items.length <= 1) {
                                return;
                            }
                            if (settings.area) {
                                if (settings.area.hasPosition(evt.localPosition)) {
                                    isHolding = evt.id;
                                    if (settings.onScrollStart) {
                                        settings.onScrollStart();
                                    }
                                }
                            } else {
                                isHolding = evt.id;
                                if (settings.onScrollStart) {
                                    settings.onScrollStart();
                                }
                            }
                            // if (evt.position.y > 48 && evt.position.y < viewport.height - 64) {
                            //     isHolding = true;
                            // }
                        },
                        pointerMove: function (evt) {
                            if (isHolding === evt.id) {
                                // measure speed
                                speedX = evt.position[dir] - grabX;
                                if (Math.abs(speedX) > Math.abs(avgSpeedX)) {
                                    avgSpeedX = speedX;
                                }
                                diffX = evt.position[dir] - grabX;
                                // container.position[dir] += diffX;
                                container.position[dir] = originalPos + (evt.position[dir] - holdPos);
                                grabX = evt.position[dir];
                                if (settings.onScroll) {
                                    settings.onScroll(container.position[dir]);
                                }
                            }
                        },
                        pointerUp: function (evt) {
                            if (isHolding === evt.id) {
                                isHolding = null;
                                slideSpeed = avgSpeedX;
                            }
                        }
                    }), {
                        pointers: [],
                        start: function () {
                            this.pointers = Bento.input.getPointers();
                        },
                        update: function () {
                            if (isHolding !== null && this.pointers.length === 0) {
                                // missed a pointerUp somehow
                                isHolding = null;
                            }
                        }
                    }
                ].concat(settings.components || [])
            }),
            // container
            container = new Entity({
                name: 'container',
                updateWhenPaused: settings.updateWhenPaused || 0,
                components: [

                ]
            }),
            setItemPos = function () {
                var i, l,
                    item,
                    size = 0;

                totalSize = 0;
                for (i = 0, l = items.length; i < l; ++i) {
                    item = items[i];
                    item.position[dir] = size;
                    if (i !== l - 1) { // don't count last part
                        itemWidth = item.dimension[dim];
                        size += itemWidth + spacing;
                    }

                }
                totalSize = size;
            };

        entity.attach(container);

        // scroll behavior
        container.attach({
            update: function () {
                var // diff position
                    containerX = container.position[dir],
                    //
                    size = (itemWidth + spacing),
                    maxSel = items.length - 1,
                    minSel = 0,
                    selection = Utils.clamp(-maxSel, size === 0 ? 0 : Math.round(containerX / size), minSel),
                    index,
                    // focus
                    target = selection * size;

                if (release) {
                    release = false;
                    isHolding = null;
                    slideSpeed = 0;
                    return;
                }

                if (previousSelection !== selection) {
                    if (settings.onLoseFocus) {
                        settings.onLoseFocus(items[Utils.clamp(0, -previousSelection, items.length - 1)]);
                    }
                    previousSelection = selection;
                    currentSelection = -selection;
                    if (settings.onSnap) {
                        index = Utils.clamp(0, currentSelection, items.length - 1);
                        settings.onSnap(items[index], index);
                    }
                }

                // force snap
                if (forcedSnap >= 0 && selection !== -forcedSnap) {
                    target = -forcedSnap * size;
                } else {
                    forcedSnap = -1;
                }

                avgSpeedX = avgSpeedX * 0.6;

                if (isHolding === null) {
                    // snap
                    if (snap) {
                        if (Math.abs(slideSpeed) < snapMinSlideSpeed) {
                            container.position[dir] += (target - containerX) * snapSpeed;
                        } else {
                            container.position[dir] += slideSpeed;
                        }
                    }

                    // move
                    container.position[dir] += slideSpeed;

                    // limits
                    if (containerX <= -totalSize - maxOffset) {
                        target = -totalSize - maxOffset;
                        container.position[dir] += (target - containerX) * 0.1;
                        slideSpeed /= 2;
                    } else if (containerX > minOffset) {
                        target = minOffset;
                        container.position[dir] += (target - containerX) * 0.1;
                        slideSpeed /= 2;
                    }
                }

                // set item positions, because they can change due to expanding/collapsing
                setItemPos();

                // damping
                if (Math.abs(slideSpeed) < 2) {
                    slideSpeed *= damping;
                } else {
                    slideSpeed *= damping;
                }

                if (clamp) {
                    if (container.position[dir] > minOffset) {
                        container.position[dir] = minOffset;
                        slideSpeed = 0;
                    } else if (container.position[dir] <= -totalSize - maxOffset) {
                        container.position[dir] = -totalSize - maxOffset;
                        slideSpeed = 0;
                    }
                }

                currentPos = container.position[dir];
            }
        });

        // public
        entity.extend({
            didMove: function (tolerance) {
                return Math.abs(currentPos - originalPos) > tolerance;
            },
            iterate: function (callback) {
                var i = 0;
                var l = items.length;
                var stop = false;
                var breakLoop = function () {
                    stop = true;
                };
                for (i = 0; i < l; ++i) {
                    callback(items[i], i, l, breakLoop);
                    if (stop) {
                        return;
                    }
                }
            },
            addItem: function (item) {
                // item.position[dir] = (item.dimension[dim] + spacing) * items.length;
                items.push(item);
                container.attach(item);
                setItemPos();
            },
            removeItem: function (index) {
                var item = items.splice(index, 1);
                container.remove(item[0]);
                setItemPos();
            },
            forceTo: function (value) {
                container.position[dir] = value;
                entity.update();
            },
            getScroll: function () {
                return container.position[dir];
            },
            getItems: function () {
                return items;
            },
            getTotalSize: function () {
                return totalSize;
            },
            setArea: function (area) {
                settings.area = area;
            },
            snapTo: function (index) {
                forcedSnap = Utils.clamp(0, index, items.length - 1);
            },
            getSelection: function () {
                return currentSelection;
            },
            cancel: function () {
                release = true;
            },
            isScrolling: function () {
                return Math.abs(slideSpeed) > 0.2;
            },
            setMinOffset: function (value) {
                minOffset = value;
            },
            setMaxOffset: function (value) {
                maxOffset = value;
            },
            setSpacing: function (size) {
                spacing = size;
                setItemPos();
            },
            stopSliding: function () {
                slideSpeed = 0;
            },
            setActive: function (value) {
                active = value;
            },
            length: function () {
                return items.length;
            }
        });

        return entity;
    };
});