/**
 * Sensei pops up in front
 * Will respond to the event "dismissSensei"
 * @moduleName SenseiPopup
 */
bento.define('gui/senseipopup', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/anchor',
    'gui/speechballoon',
    'modules/localization',
    'gui/menus/shop',
    'globals',
    'components/eventlistener',
    'components/modal',
    'components/fillscreen'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Anchor,
    SpeechBalloon,
    Localization,
    ShopMenu,
    Globals,
    EventListener,
    Modal,
    FillScreen
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var textArray = settings.text || [];
        var onPageEnd = settings.onPageEnd;
        var modal = settings.modal || false;
        var darken = settings.darken || false;
        var onNoMoreText = settings.onNoMoreText || false;
        var fillScreen = new FillScreen({
            color: [0, 0, 0, 0]
        });
        var index = -1;
        var canChangePage = false;
        var speechBalloon = new SpeechBalloon({
            text: '',
            size: new Rectangle(0, 0, 112, 54),
            position: new Vector2(88, -40),
            typewriter: true,
            typeSpeed: 1,
            fromLeft: true,
            onComplete: function () {
                if (index === -1) {
                    return;
                }
                if (onPageEnd) {
                    onPageEnd.call(entity, index);
                }
                // wait for button input to go to next page
                canChangePage = true;
                if (textArray[index + 1]) {
                    // has another page
                    triangle.visible = true;
                }
            }
        });
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                var tweenOutDelay = 60;
                speechBalloon.scale.x = 0;
                speechBalloon.scale.y = 0;

                // settings.text should be an array
                if (!Utils.isArray(textArray)) {
                    textArray = [textArray];
                }

                // tween in speech balloon
                speechBalloon.setTyping(false);
                new Tween({
                    from: 0,
                    to: 1,
                    in: 20,
                    delay: 10,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        speechBalloon.scale.x = v;
                        speechBalloon.scale.y = v;
                    },
                    onComplete: function () {
                        speechBalloon.setTyping(true);
                        nextPage();
                        speechBalloon.attach(triangle);
                    }
                });
                // tween in from the right
                new Tween({
                    from: -64,
                    to: 0,
                    in: 30,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        sensei.position.x = v;
                    }
                });
                new Tween({
                    from: -Math.PI / 4,
                    to: 0,
                    in: 30,
                    ease: 'elastic',
                    decay: 5,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        sensei.rotation = v;
                    }
                });

                // darken
                if (darken) {
                    new Tween({
                        from: 0,
                        to: 0.5,
                        in: 30,
                        onUpdate: function (v, t) {
                            fillScreen.setColor([0, 0, 0, v]);
                        },
                        onComplete: function () {

                        }
                    });
                }
            },
            destroy: function (data) {},
            update: function (data) {},
            draw: function (data) {}
        };
        var isFading = false;
        var fadeOut = function () {
            isFading = true;
            // tween out
            new Tween({
                from: sensei.position.y,
                to: sensei.position.y + 128,
                in: 30,
                ease: 'easeInBack',
                updateWhenPaused: Globals.PauseLevels.ALWAYS,
                onUpdate: function (v, t) {
                    sensei.position.y = v;
                },
                onComplete: function () {
                    // remove self
                    entity.removeSelf();
                    if (settings.onComplete) {
                        settings.onComplete();
                    }
                }
            });
            new Tween({
                from: 1,
                to: 0,
                in: 30,
                ease: 'linear',
                updateWhenPaused: Globals.PauseLevels.ALWAYS,
                onUpdate: function (v, t) {
                    speechBalloon.alpha = v;
                },
                onComplete: function () {
                    speechBalloon.removeSelf();
                }
            });
        };
        var nextPage = function () {
            index += 1;
            var currentText = textArray[index];
            triangle.visible = false;
            if (currentText) {
                speechBalloon.setText(currentText);
                canChangePage = false;
            } else {
                // no more text
            }
        };
        var anchor = new Anchor({
            screenPosition: new Vector2(0, 1),
            offset: new Vector2(32, 8),
            onComplete: function (parent, isPortrait) {}
        });
        var sprite = new Sprite({
            imageName: settings.imageName || 'gui/sensei',
            originRelative: new Vector2(0.5, 1)
        });
        var sensei = new Entity({
            name: 'sensei',
            scale: new Vector2(2, 2),
            components: [
                behavior,
                sprite
            ]
        });
        var eventListener = new EventListener({
            name: 'dismissSensei',
            onEvent: function () {
                fadeOut();
                senseiContainer.remove(eventListener);
            }
        });
        var senseiContainer = new Entity({
            name: 'senseiContainer',
            components: [
                sensei,
                speechBalloon,
                anchor,
                eventListener
            ]
        });
        var clickable = new Clickable({
            onClick: function (data) {
                if (canChangePage) {
                    if (!textArray[index + 1]) {
                        // no more pages
                        if (onNoMoreText) {
                            onNoMoreText.call(entity);
                            onNoMoreText = null;
                        }
                    } else {
                        nextPage();
                    }
                }
            }
        });
        var triangle = new Entity({
            name: 'triangle',
            position: new Vector2(46, 22),
            components: [
                new Sprite({
                    imageName: 'gui/triangle',
                    originRelative: new Vector2(0.5, 1)
                }), //
                {
                    name: 'bounce',
                    update: function () {
                        triangle.position.y = 22 - 2 * Math.abs(Math.sin(triangle.timer / 20));
                    },
                    destroy: function () {

                    }
                }
            ]
        });
        var entity = new Entity({
            z: settings.z || Globals.WorldSelectLayers.SENSEI,
            name: 'senseiPopup',
            position: settings.position,
            float: true,
            updateWhenPaused: Globals.PauseLevels.ALWAYS,
            boundingBox: new Rectangle(0, 0, viewport.width, viewport.height),
            components: [
                modal ? new Modal({}) : {},
                darken ? fillScreen : {},
                senseiContainer,
                clickable
            ]
        });

        entity.extend({
            setDarken: function (value) {
                if (!value) {
                    fillScreen.setColor([0, 0, 0, 0]);
                } else {
                    fillScreen.setColor([0, 0, 0, 0.5]);
                }
            },
            dismiss: function () {
                if (!isFading) {
                    fadeOut();
                    senseiContainer.remove(eventListener);
                }
            }
        });

        return entity;
    };
});