/**
 * Speech balloon
 * @moduleName SpeechBalloon
 */
bento.define('gui/speechballoon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'bento/components/nineslice',
    'gui/shadowtext'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    NineSlice,
    ShadowText
) {
    'use strict';
    var SpeechBalloon = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
            typewriter: Boolean // if the text should be shown in a typewriter way
            typeSpeed: number // the speed at which typewriter adds characters
            size: rectangle // the size of the speech bubble
        }*/
        var viewport = Bento.getViewport();

        var size = settings.size || new Rectangle(0, 0, 64, 32);
        var fitToText = settings.fitToText || false;
        var currentText = settings.typewriter ? '' : settings.text;
        // decide size based on character length

        var nineSlice;
        var padding = new Vector2(15, 8);
        // if settings.typewriter exists add text to typewriter instead;
        var typewriter = settings.typewriter ? settings.text : '';
        var shouldType = true;
        // speed to type text with
        var typeSpeed = settings.typeSpeed || 1;
        var fromLeft = settings.fromLeft || false;
        // if a character should be added
        var forceSpeed = 1;
        var addCharacter = 0;

        var bottomPosition;

        var textEntity;

        // internal text to check linebreaks and fontsize
        var internalText;

        // internal fontsize
        var fontSize = 16;

        var touchControl = settings.touchControl || false;
        var touchPhase = 0;
        var touchRelease = false;

        var checkLineBreak = false;
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                var textLenght = Math.max(currentText.length, typewriter.length);
                if (!settings.size) {
                    if (textLenght <= 10) {
                        size = new Rectangle(0, 0, 80, 30);
                    } else if (textLenght > 10 && textLenght <= 16) {
                        size = new Rectangle(0, 0, 96, 48);
                    } else if (textLenght > 16 && textLenght <= 48) {
                        size = new Rectangle(0, 0, 140, 48);
                    } else if (textLenght > 48 && textLenght <= 64) {
                        size = new Rectangle(0, 0, 160, 64);
                    } else {
                        size = new Rectangle(0, 0, 160, 96);
                    }
                }
                if (fitToText) {
                    // if we're fitting the balloon to the text, the height doesn't matter much
                    size.height = 96;
                }

                entity.boundingBox = size;

                // set horizontal bottom position
                if (!bottomPosition && bottomPosition !== 0) {
                    bottomPosition = size.width / 2;
                }
                if (Utils.isDefined(settings.bottomPosition)) {
                    bottomPosition = settings.bottomPosition;
                }

                // bottom flip
                if (settings.bottomFlip) {
                    bottom.scale.x = -1;
                }

                internalText = new Text({
                    position: new Vector2(0, 0),
                    text: settings.typewriter ? '' : settings.text,
                    font: 'font',
                    fontSize: 16,
                    fontColor: '#ffffff',
                    align: 'left',
                    textBaseline: 'top',
                    maxWidth: size.width - padding.x,
                    maxHeight: size.height - padding.y - 2
                });

                checkText();

                // re-fit height to text after splitting strings
                // size.height = 12;
                // Utils.forEach(internalText.getStrings(), function () {
                //     size.height += 16;
                // });
                if (fitToText) {
                    size = internalText.dimension.clone();
                    size.width += 16;
                    size.height += 8;
                }

                nineSlice = new NineSlice({
                    imageName: 'gui/speechbub',
                    width: size.width,
                    height: size.height,
                    originRelative: new Vector2(0.5, 0.5)
                });
                entity.attach(nineSlice);
                bottom.position.x = Math.floor(Utils.clamp(size.width / -2 + 13, bottomPosition, size.width / 2 - 13));
                bottom.position.y = Math.floor(size.height / 2) - 4;
                entity.attach(bottom);

                textEntity = new Text({
                    // drawDebug: true,
                    position: new Vector2((size.width - padding.x) / -2, (size.height - padding.y) / -2 - 1),
                    text: currentText,
                    font: 'font',
                    fontSize: fontSize,
                    fontColor: '#2b2e48',
                    shadowColor: '#b7bace',
                    align: 'left',
                    textBaseline: 'top',
                    // maxWidth: size.width - padding.x,
                    // maxHeight: size.height - padding.y - 2
                });
                entity.attach(textEntity);

                // add clickable if touchcontrol is enabled
                if (touchControl) {
                    entity.attach(clickable);
                }

                if (fromLeft) {
                    bottomSprite.setup({
                        imageName: 'gui/speechbub-left',
                        originRelative: new Vector2(1, 0.5)
                    });
                    bottom.position.x = size.width / -2 + 3;
                    bottom.position.y = 0;
                }
            },
            destroy: function (data) {},
            update: function (data) {
                if (typewriter.length > 0 && shouldType) {
                    // speed up text
                    if (touchPhase === 1) {
                        typeSpeed = forceSpeed;
                    }

                    // write the characters to the balloon
                    addCharacter -= typeSpeed;
                    var updateText = false;
                    while (addCharacter <= 0 && typewriter.length > 0) {
                        currentText += typewriter[0];
                        typewriter = typewriter.substr(1, typewriter.length - 1);
                        updateText = true;
                        addCharacter++;
                    }

                    // if characters have been added update shadowtext
                    if (updateText) {
                        // jumpy text fix
                        var txt;
                        if (endsWith(currentText, ' ')) {
                            // trim() removes whitespace at the start and end of strings
                            // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/Trim
                            txt = currentText.trim();
                        } else {
                            txt = currentText;
                        }

                        textEntity.setText(txt, {
                            fontSize: fontSize,
                            // maxWidth: size.width - padding.x,
                            // maxHeight: size.height - padding.y - 2
                        });
                    }

                    if (typewriter.length === 0) {
                        if (touchPhase !== 0) {
                            touchPhase = 3;
                        }
                    }
                } else if (touchControl) {
                    entity.waitingForInput = true;
                    if (touchPhase > 0 && touchRelease) {
                        touchControl = false;
                        if (!entity.complete && settings.onComplete) {
                            settings.onComplete();
                        }
                        entity.complete = true;
                    }
                } else {
                    if (!entity.complete && settings.onComplete) {
                        settings.onComplete();
                    }
                    entity.complete = true;
                }

                //reset touch id
                if (touchRelease) {
                    touchPhase = 0;
                    touchRelease = false;
                }
            },
            draw: function (data) {}
        };

        //tap logic
        var clickable = new Clickable({
            pointerDown: function (data) {
                // 0 is unpressed, 1 is pressed, 2 is down
                if (touchPhase < 2) {
                    touchPhase++;
                }
            },
            pointerUp: function (data) {
                // in combination with touchPhase can make sure that holding before and releasing during cutscene doesnt work.
                touchRelease = true;
            }
        });

        var setBottomPosition = function (value) {
            bottomPosition = value;
            bottom.position.x = Math.floor(Utils.clamp(size.width / -2 + 13, bottomPosition, size.width / 2 - 13));
        };

        var checkText = function () {
            var internalStrings;
            var formattedText = '';
            // get new text and fontsize
            fontSize = internalText.setText(typewriter);
            internalStrings = internalText.getStrings();

            // set up single strings
            Utils.forEach(internalStrings, function (str) {
                var line = str.string;
                formattedText += line + '\n';
            });
            formattedText = formattedText.substr(0, formattedText.length - 1);
            typewriter = formattedText;
        };

        var endsWith = function (txt) {
            var length = txt.length;
            if (length-- > 0) {
                return txt[length] === ' ';
            }
            return false;
        };

        var bottomSprite = new Sprite({
            imageName: 'gui/speechbub-bottom',
            originRelative: new Vector2(0.5, 0)
        });
        var bottom = new Entity({
            name: 'bottom',
            components: [
                bottomSprite
            ]
        });
        var entity = new Entity({
            z: settings.z,
            name: settings.name || 'speechBalloon',
            position: settings.position,
            originRelative: new Vector2(0.5, 0.5),
            float: settings.float,
            updateWhenPaused: settings.updateWhenPaused,
            components: [
                behavior
            ]
        }).extend({
            complete: false,
            waitingForInput: false,
            setBottomPosition: setBottomPosition,
            getBottom: function () {
                return bottom;
            },
            setTyping: function (value) {
                shouldType = value;
            },
            setText: function (str) {
                addCharacter = 0;
                entity.complete = false;
                currentText = settings.typewriter ? '' : str;
                typewriter = settings.typewriter ? str : '';

                internalText.setText(settings.typewriter ? '' : str);
                checkText();
                textEntity.setText(currentText, {
                    fontSize: fontSize
                });
            }
        });
        return entity;
    };

    return SpeechBalloon;
});