/**
 * Container showing the stats of the current character + stats for the next level
 * @moduleName Stats
 */
bento.define('gui/upgrade/stats', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'managers/save',
    'managers/character',
    'managers/stats',
    'components/debug',
    'gui/shadowtext',
    'components/shadow',
    'modules/localization',
    'bento/components/nineslice'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    SaveManager,
    CharacterManager,
    StatsManager,
    Debug,
    ShadowText,
    Shadow,
    Localization,
    NineSlice
) {
    'use strict';
    var Stats = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var showBackground = settings.showBackground;
        var character = settings.character || CharacterManager.getSelected();
        var displayMode = settings.displayMode || Stats.DISPLAY_POSSIBLE;
        var characterStats = StatsManager.getAllDetailedStats(character);
        var possibleStats = {};

        var levelText = new Text({
            text: Localization.getText('level').replace('{NUMBER}', SaveManager.load('level-' + character)),
            font: 'font',
            fontSize: 10,
            fontColor: '#23243a',
            align: 'right',
            textBaseline: 'top',
            position: new Vector2(0, -20),
            maxWidth: 24,
            maxHeight: 10
        });
        var levelNumberText = new Text({
            text: SaveManager.load('level-' + character).toString(),
            font: 'font',
            fontSize: 10,
            fontColor: '#ffdf00',
            align: 'left',
            textBaseline: 'top',
            position: new Vector2(4, -20),
            maxWidth: 8,
            maxHeight: 10
        });
        var levelMaxText = new Text({
            text: '/' + CharacterManager.getMaxLevel(character),
            font: 'font',
            fontSize: 10,
            fontColor: '#23243a',
            align: 'left',
            textBaseline: 'top',
            position: new Vector2(12, -20)
        });
        var titleText = new ShadowText({
            position: new Vector2(0, -28),
            text: Localization.getText('name-' + character),
            font: 'font',
            fontSize: 20,
            fontColor: '#ffdf00',
            align: 'center',
            textBaseline: 'middle',
            maxWidth: 72,
            maxHeight: 20
        });

        var getStatText = function (stat) {
            var text = '';
            var possible = 0;
            if (displayMode === Stats.DISPLAY_COMBINED) {
                text = characterStats[stat].total + '';
            } else if (displayMode === Stats.DISPLAY_POSSIBLE) {
                // total stat + possible, "+ 0" is not displayed
                possible = (possibleStats[stat] || 0);
                if (possible) {
                    text = characterStats[stat].total + ' + ' + possible;
                } else {
                    text = characterStats[stat].total;
                }
            } else if (displayMode === Stats.DISPLAY_EQUIPMENT) {
                // bare stat + equipment
                text = characterStats[stat].bare + '';
                if (characterStats[stat].equipped) {
                    text += ' + ' + characterStats[stat].equipped;
                }
            } else if (displayMode === Stats.DISPLAY_BARE) {
                text = characterStats[stat].bare + '';
            } else {
                Utils.log('unknown stat display mode!');
                text = characterStats[stat].total + '';
            }

            return text;
        };
        var setStatText = function (stat) {
            var parent;
            var plus;
            var fontSize = 12;

            var text = '';
            var possible = 0;

            if (stat === 'health') {
                parent = healthText;
                plus = healthTextPlus;
            } else if (stat === 'attack') {
                parent = attackText;
                plus = attackTextPlus;
            } else if (stat === 'skill') {
                parent = defenseText;
                plus = defenseTextPlus;
            }

            parent.position.x = plus.position.x;
            plus.setText('');

            if (displayMode === Stats.DISPLAY_COMBINED) {
                text = characterStats[stat].total + '';
                parent.setText(text, {
                    fontSize: 12
                });
            } else if (displayMode === Stats.DISPLAY_POSSIBLE) {
                // total stat + possible, "+ 0" is not displayed
                possible = (possibleStats[stat] || 0);
                if (possible) {
                    text = characterStats[stat].total + '+' + possible;
                    // measure text first
                    fontSize = measureText.setText(text);
                    parent.setText(characterStats[stat].total, {
                        fontSize: fontSize
                    });
                    plus.setText('+' + possible, {
                        fontSize: fontSize
                    });
                    parent.position.x = plus.position.x - plus.dimension.width;

                } else {
                    text = characterStats[stat].total;
                    parent.setText(text, {
                        fontSize: 12
                    });
                }
            } else if (displayMode === Stats.DISPLAY_EQUIPMENT) {
                // bare stat + equipment
                text = characterStats[stat].bare + '';
                if (characterStats[stat].equipped) {
                    text += '+' + characterStats[stat].equipped;

                    // measure text first
                    fontSize = measureText.setText(text);
                    parent.setText(characterStats[stat].bare, {
                        fontSize: fontSize
                    });
                    plus.setText('+' + characterStats[stat].equipped, {
                        fontSize: fontSize
                    });
                    parent.position.x = plus.position.x - plus.dimension.width;
                } else {
                    parent.setText(text, {
                        fontSize: 12
                    });
                }
            } else if (displayMode === Stats.DISPLAY_BARE) {
                text = characterStats[stat].bare + '';
                parent.setText(text, {
                    fontSize: 12
                });
            } else {
                Utils.log('unknown stat display mode!');
                text = characterStats[stat].total + '';
                parent.setText(text, {
                    fontSize: 12
                });
            }

            return text;
        };

        var spacing = 4;
        /**
         * Health
         */
        var healthText = new Text({
            position: new Vector2(22, -18),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#ff4b38',
            align: 'right',
            textBaseline: 'middle'
        });
        var healthTextPlus = new Text({
            position: new Vector2(22, -18),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#ffff00',
            align: 'right',
            textBaseline: 'middle',
            margin: new Vector2(0, 8)
        });
        /**
         * Attack
         */
        var attackText = new Text({
            position: new Vector2(22, -18 + 14),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#23ff23',
            align: 'right',
            textBaseline: 'middle',
            margin: new Vector2(0, 8)
        });
        var attackTextPlus = new Text({
            position: new Vector2(22, -18 + 14),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#ffff00',
            align: 'right',
            textBaseline: 'middle',
            margin: new Vector2(0, 8)
        });
        /**
         * Defense
         */
        var defenseText = new Text({
            position: new Vector2(22, -18 + 14 * 2),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#52e9ff',
            align: 'right',
            textBaseline: 'middle',
            margin: new Vector2(0, 8)
        });
        var defenseTextPlus = new Text({
            position: new Vector2(22, -18 + 14 * 2),
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#ffff00',
            align: 'right',
            textBaseline: 'middle',
            margin: new Vector2(0, 8)
        });
        /**
         * Measure text to predict what the actual font size will be
         */
        var measureText = new Text({
            text: '',
            font: 'font',
            fontSize: 12,
            fontColor: '#ffffff',
            align: 'right',
            textBaseline: 'middle',
            maxWidth: 28,
            maxHeight: 12
        });

        /**
         * Container
         */
        var containerSprite = new NineSlice({
            imageName: 'gui2/misc/splash_background_9slice',
            width: 80,
            height: 96,
            originRelative: new Vector2(0.5, 0.5)
        });
        var statsSprite = new Sprite({
            imageName: 'gui2/characterchange/stat_background',
            originRelative: new Vector2(0.5, 0.5),
        });
        var statsSpriteContainer = new Entity({
            name: 'container',
            position: new Vector2(0, 16),
            components: [
                statsSprite,
                attackText,
                healthText,
                defenseText,
                attackTextPlus,
                healthTextPlus,
                defenseTextPlus
            ]
        });
        var statsContainer = new Entity({
            name: 'stats',
            position: settings.position,
            components: [
                //containerSprite,
                statsSpriteContainer,
                levelText,
                levelNumberText,
                levelMaxText,
                titleText
            ].concat(settings.components || [])
        }).extend({
            updateStats: function () {
                characterStats = StatsManager.getAllDetailedStats(character);

                // healthText.setText(getStatText('health'));
                // attackText.setText(getStatText('attack'));
                // defenseText.setText(getStatText('skill'));
                setStatText('health');
                setStatText('attack');
                setStatText('skill');

                levelNumberText.setText(SaveManager.load('level-' + character).toString());
                levelMaxText.setText('/' + CharacterManager.getMaxLevel(character));
                titleText.setText(Localization.getText('name-' + character));
            },
            resetPossible: function () {
                possibleStats = {};
            },
            addPossible: function (name, value) {
                possibleStats[name] = possibleStats[name] || 0;
                possibleStats[name] += value;
            },
            setCharacter: function (name) {
                character = name;
                statsContainer.updateStats();
            },
            glowPlusText: function () {
                var createBehavior = function () {
                    return {
                        name: 'pulsingBehavior',
                        update: function (data) {
                            data.entity.alpha = 0.7 + 0.25 * Math.sin(data.entity.timer / 15);
                        }
                    };
                };
                healthTextPlus.attach(createBehavior());
                attackTextPlus.attach(createBehavior());
                defenseTextPlus.attach(createBehavior());
            },
            setDisplayMode: function (value) {
                displayMode = value;
                statsContainer.updateStats();
            }
        });

        statsContainer.updateStats();

        if (showBackground) {
            statsContainer.attach(containerSprite);
            statsContainer.moveComponentTo(containerSprite, 0);
        }

        return statsContainer;
    };

    Stats.DISPLAY_COMBINED = 0;
    Stats.DISPLAY_EQUIPMENT = 1;
    Stats.DISPLAY_BARE = 2;
    // "possible" displays stats that are not applied but could happen if user chooses to continue
    Stats.DISPLAY_POSSIBLE = 3;

    return Stats;
});