/**
 * Bar at the bottom of world select screen with character, equipment, upgrade and store buttons.
 * @moduleName Bottombar
 */
bento.define('gui/worldselect/bottombar', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'components/anchor',
    'globals',
    'managers/character',
    'components/eventlistener',
    'gui/menus/iap',
    'gui/badge',
    'managers/save',
    'managers/currency',
    'gui/centeredlist',
    'managers/stats',
    'gui/iap/offerpack',
    'gui/taphand',
    'bento/gui/togglebutton',
    'gui/menus/loadout',
    'gui/menus/shop',
    'gui/menus/dailyquest',
    'managers/quest',
    'modules/ktplay',
    'modules/analytics',
    'components/glowsprite',
    'components/backbutton',
    'gui/menus/towerchallenge',
    'gui/menus/beachevent'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Anchor,
    Globals,
    CharacterManager,
    EventListener,
    IAPMenu,
    Badge,
    SaveManager,
    CurrencyManager,
    CenteredList,
    StatsManager,
    OfferPack,
    TapHand,
    ToggleButton,
    LoadOutMenu,
    ShopMenu,
    DailyQuestMenu,
    QuestManager,
    KTPlayModule,
    Analytics,
    GlowSprite,
    BackButton,
    TowerMenu,
    BeachEventMenu
) {
    'use strict';

    /**
     * ClickButton refactored as a toggle button
     * Reason for using a ClickButton is to prevent overlapping buttons in the menu
     * ClickButtons already have built-in functionality to rpevent this
     */
    var Toggle = function (settings) {
        var entity;
        var spriteReal;
        var spriteToggle = new Sprite({
            imageName: settings.imageName,
            frameCountX: 3,
            frameCountY: 1,
            originRelative: new Vector2(0.5, 0.5),
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [2]
                }
            }
        });
        var setSprite = function () {
            if (entity.toggled) {
                spriteToggle.setAnimation('down');
            } else {
                spriteToggle.setAnimation('up');
            }
        };

        settings.onClick = function () {
            entity.toggled = !entity.toggled;
            if (settings.onToggle) {
                settings.onToggle.apply(entity);
            }
            setSprite();
        };

        entity = new ClickButton(settings);
        spriteReal = entity.getComponent('sprite');
        spriteReal.visible = false;
        entity.extend({
            toggled: settings.toggled || false,
            isToggled: function () {
                return entity.toggled;
            },
            toggle: function (value) {
                entity.toggled = value;
                setSprite();
            }
        });
        entity.attach(spriteToggle);
        setSprite();
        return entity;
    };

    var BottomBar = function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var target = settings.target; // additional target screen
        var background = new Entity({
            name: 'background',
            components: [
                new Sprite({
                    imageName: 'gui2/bottom_tabs/bottom_tabs_backboard',
                    originRelative: new Vector2(0.5, 1)
                }),
                new Sprite({
                    imageName: 'gui2/bottom_tabs/padding',
                    originRelative: new Vector2(0.5, 0)
                })
            ]
        });
        var anchor = new Anchor({
            screenPosition: new Vector2(0.5, 1),
            offset: new Vector2(0, 0),
            onComplete: function () {
                if (Anchor.isWideScreen()) {
                    if (Anchor.isPortrait()) {
                        entity.position.y += Anchor.getOffset().y / 2;
                    } else {}
                }
            }
        });
        var currentMenu = null;
        var menus = {
            buttonPlay: LoadOutMenu,
            buttonChest: IAPMenu,
            buttonDailyShop: ShopMenu,
            buttonDailyQuests: DailyQuestMenu
        };
        var showMenu = function (MenuConstructor) {

            var menu = new MenuConstructor({
                isPopUp: false,
                bottomBar: entity,
                tweenIn: settings.tweenIn
            });
            // tween in only once (only used for loadout)
            settings.tweenIn = undefined;
            currentMenu = menu;
            Bento.objects.attach(menu);
            // take over pause level
            if (currentMenu) {
                entity.updateWhenPaused = currentMenu.updateWhenPaused;
            }

            if (SaveManager.load('tutorialWorldSelect') > 0) {
                Analytics.trackView('worldMap-' + menu.name);
            }

        };
        // toggles on 1 button and untoggles the rest
        var toggleOn = function (button, openMenu) {
            // untoggle all
            buttonPlay.toggle(false);
            buttonChest.toggle(false);
            buttonDailyShop.toggle(false);
            buttonDailyQuests.toggle(false);

            button.toggle(true);

            if (openMenu) {
                // hide current menu
                if (currentMenu) {
                    currentMenu.removeSelf();
                    currentMenu = null;
                }
                showMenu(menus[button.name]);
            }
        };
        var buttonY = -14 + 23 / 2;
        /**
         * Play
         */
        var buttonPlay = new Toggle({
            name: 'buttonPlay',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs__play',
            frameCountX: 3,
            frameCountY: 1,
            toggled: true,
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [2]
                }
            },
            position: new Vector2(0, buttonY - 2),
            onToggle: function () {
                var previousToggle = !this.isToggled();
                if (previousToggle) {
                    // already toggled on
                    toggleOn(this);
                    return;
                }
                toggleOn(this, true);

            }
        });
        /**
         * Chest
         */
        var badgeChest = new Badge({
            pulsate: true,
            exclamation: true,
            imageName: 'gui2/misc/badge'
        });
        var glowChest = new GlowSprite({
            position: new Vector2(0, -4)
        });
        var buttonChest = new Toggle({
            name: 'buttonChest',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs2_chests',
            frameCountX: 3,
            frameCountY: 1,
            toggled: false,
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [2]
                }
            },
            position: new Vector2(71, buttonY),
            onToggle: function () {
                var previousToggle = !this.isToggled();
                if (previousToggle) {
                    // already toggled on
                    toggleOn(this);
                    return;
                }
                toggleOn(this, true);
                EventSystem.fire('chestButtonTapped');
            }
        });
        /**
         * DailyShop
         */
        var buttonDailyShop = new Toggle({
            name: 'buttonDailyShop',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs2_shop',
            frameCountX: 3,
            frameCountY: 1,
            toggled: false,
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [2]
                }
            },
            position: new Vector2(39, buttonY),
            onToggle: function () {
                var previousToggle = !this.isToggled();
                if (previousToggle) {
                    // already toggled on
                    toggleOn(this);
                    return;
                }
                toggleOn(this, true);
            }
        });
        /**
         * DailyShop
         */

        var badgeQuests = new Badge({
            pulsate: true,
            exclamation: true,
            imageName: 'gui2/misc/badge'
        });
        var buttonDailyQuests = new Toggle({
            name: 'buttonDailyQuests',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs2_dailyquests',
            frameCountX: 3,
            frameCountY: 1,
            toggled: false,
            animations: {
                up: {
                    frames: [0]
                },
                down: {
                    frames: [1]
                },
                inactive: {
                    frames: [2]
                }
            },
            position: new Vector2(-38, buttonY),
            onToggle: function () {
                var previousToggle = !this.isToggled();
                if (previousToggle) {
                    // already toggled on
                    toggleOn(this);
                    return;
                }
                toggleOn(this, true);
            }
        });
        /**
         * Community
         */
        var badgeCommunity = new Badge({
            pulsate: true,
            exclamation: true,
            imageName: 'gui2/misc/badge'
        });
        var buttonCommunity = new ClickButton({
            z: Globals.WorldSelectLayers.BOTTOMBAR,
            name: 'buttonBack',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs2_ktplay',
            frameCountX: 3,
            frameCountY: 1,
            toggled: false,
            float: true,
            position: new Vector2(-70, buttonY),
            onButtonDown: function () {
                EventSystem.fire('preventScroll');
            },
            onClick: function () {
                SaveManager.save('pressedKTPlay', true);
                badgeCommunity.removeSelf();
                KTPlayModule.getLoginStatus(function (isLoggedIn) {
                    //if we're logged in show the ktplay community
                    if (isLoggedIn) {
                        KTPlayModule.showCommunity(function () {
                            Analytics.event("KTPlay", "opened");
                        }, function (e) {
                            console.log(e);
                        });
                    } else {
                        // if we're not logged in try auto log in again
                        Utils.KTPlayAutoLogin(function () {
                            // if we auto login, then show the community
                            KTPlayModule.showCommunity(function () {
                                Analytics.event("KTPlay", "opened");
                            }, function (e) {
                                console.log(e);
                            });
                        }, function (error) {
                            // we didn't auto login fallback on ktplay login
                            KTPlayModule.showLogin(function () {}, function (e) {
                                console.log(e);
                            });
                            console.log(error);
                        });
                    }
                });
            }
        });
        /**
         * Back from worldmap
         */
        var buttonBackWorldMap = new ClickButton({
            z: Globals.WorldSelectLayers.BOTTOMBAR,
            name: 'buttonBack',
            sfx: 'sfx_button',
            imageName: 'gui2/bottom_tabs/bottom_tabs__back',
            frameCountX: 2,
            frameCountY: 1,
            toggled: false,
            float: true,
            sort: true, // gives priority over mission buttons
            onButtonDown: function () {
                EventSystem.fire('preventScroll');
            },
            onClick: function () {
                // entity.show();
                var bottomBar = new BottomBar({
                    tweenIn: true
                });
                Bento.objects.attach(bottomBar);
                buttonBackWorldMap.removeSelf();
            },
            components: [
                new BackButton({
                    removeSelf: true,
                    onPressed: function (data) {
                        data.entity.doCallback();
                    }
                }),
                new Anchor({
                    portrait: {
                        screenPosition: new Vector2(0.5, 1),
                        offset: new Vector2(-48 + (Anchor.isWideScreen() ? 4 : 0), -14 + (Anchor.isWideScreen() ? Anchor.getOffset().y : 0))
                    },
                    landscape: {
                        screenPosition: new Vector2(0.25, 1),
                        offset: new Vector2(-44, -12)
                    }
                })
            ]
        });
        //
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                // start with loadout screen
                showMenu(menus.buttonPlay);

                // attach badges, we put them on the entity, so they all go over the buttons
                entity.attach(badgeChest);
                entity.attach(glowChest);
                entity.attach(badgeQuests);

                if (SaveManager.load('unlockedLevels').length > 5 && !SaveManager.load('pressedKTPlay')) {
                    badgeCommunity.position = buttonCommunity.position.add(new Vector2(12, -24));
                    entity.attach(badgeCommunity);
                }

                // show taphands
                showTutorial();


                if (target === 'upgrade') {
                    // open character select
                    if (currentMenu && currentMenu.openCharacterSelect) {
                        currentMenu.openCharacterSelect();
                    }
                } else if (target === 'worldmap') {
                    entity.hide();
                } else if (target === 'tower') {
                    entity.hide();
                    var towerMenu = new TowerMenu({});
                    Bento.objects.attach(towerMenu);
                } else if (target === 'beach') {
                    entity.hide();
                    var beachMenu = new BeachEventMenu({});
                    Bento.objects.attach(beachMenu);
                }
            },
            destroy: function () {
                badgeCommunity.removeSelf();
                badgeChest.removeSelf();
                glowChest.removeSelf();
                badgeQuests.removeSelf();
            },
            update: function () {
                // check if badges should be shown
                if (
                    CurrencyManager.canSpendKeys(Globals.chests.classic.cost) ||
                    CurrencyManager.canSpendGems(Globals.chests.special.cost)
                ) {
                    badgeChest.visible = true;
                    glowChest.visible = true;
                    badgeChest.position = buttonChest.position.add(new Vector2(12, -24 - 2 * buttonChest.isToggled()));
                } else {
                    badgeChest.visible = false;
                    glowChest.visible = false;
                }
                // quests
                if (QuestManager.canClaim()) {
                    badgeQuests.visible = true;
                    badgeQuests.position = buttonDailyQuests.position.add(new Vector2(12, -24 - 2 * buttonDailyQuests.isToggled()));
                } else {
                    badgeQuests.visible = false;
                }
            }
        };
        // event for chest tutorial
        var eventListener = new EventListener({
            name: 'giftClaimed',
            onEvent: function (data) {
                showTutorial();
            }
        });
        var buttons = [
            buttonCommunity,
            buttonDailyQuests,
            buttonDailyShop,
            buttonChest,
            buttonPlay
        ];
        var entity = new Entity({
            z: Globals.WorldSelectLayers.BOTTOMBAR,
            name: 'bottomBar',
            float: true,
            updateWhenPaused: 0,
            components: [
                anchor,
                background,
                buttonCommunity,
                buttonDailyQuests,
                buttonDailyShop,
                buttonChest,
                buttonPlay,
                behavior,
                eventListener
            ]
        }).extend({
            setActive: function (value) {},
            /**
             * Hide bottom bar to show the worldmap
             */
            hide: function () {
                if (currentMenu) {
                    currentMenu.removeSelf();
                }
                currentMenu = null;
                entity.removeSelf();
                // add a back button
                Bento.objects.attach(buttonBackWorldMap);
                new Tween({
                    from: buttonBackWorldMap.position.y + 32,
                    to: buttonBackWorldMap.position.y,
                    in: 20,
                    ease: 'elastic',
                    decay: 8,
                    oscillations: 1,
                    onUpdate: function (v, t) {
                        buttonBackWorldMap.position.y = v;
                    }
                });
            }
        });
        var showTutorial = function () {
            var tutorialChest = SaveManager.load('tutorialChest');
            if (!tutorialChest &&
                CurrencyManager.canSpendKeys(Globals.chests.classic.cost) &&
                !buttonChest.getComponent('tapHand')
            ) {
                var chestTapHand = new TapHand({
                    position: new Vector2(0, -8)
                });
                buttonChest.attach(chestTapHand);
                // remove self user presses chest button
                chestTapHand.attach(new EventListener({
                    name: 'chestButtonTapped',
                    onEvent: function () {
                        chestTapHand.removeSelf();
                    }
                }));
            }
        };

        return entity;
    };
    return BottomBar;
});