/**
 * Roadblock: there are 2 requirements to open a roadblock
 * -A specific level needs to be unlocked
 * -Player needs to have an amount of stars
 * @moduleName RoadBlock
 */
bento.define('gui/worldselect/roadblock', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'globals',
    'gui/shadowtext',
    'managers/save',
    'components/sort',
    'components/blink',
    'gui/sparkleburst',
    'gui/taphand'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Globals,
    ShadowText,
    SaveManager,
    Sort,
    Blink,
    SparkleBurst,
    TapHand
) {
    'use strict';
    return function (settings) {
        /*settings = {
            // describe your settings object parameters
            position: Vector2 // positions the entity
        }*/
        var viewport = Bento.getViewport();
        var sprite = new Sprite({
            imageName: 'gui/worldselect/roadblock-locked',
            originRelative: new Vector2(0.5, 0.5)
        });
        var tiled = (settings.tiled || {});
        var object = tiled.object || {};
        var nameProperties = Utils.nameToJson(object.name);
        var stars = nameProperties.stars || 0;
        var unlockedBy = nameProperties.unlockedBy;
        var next = nameProperties.next;
        var isUnlocked = false;
        var text;
        var totalStars = SaveManager.load('totalStars');
        var clickable = new Clickable({
            onClick: function (data) {
                // is locked
                if (!isUnlocked) {
                    return;
                }
                // enough stars to unlock?
                var unlocked = SaveManager.load('unlockedLevels');
                var recentlyUnlocked = SaveManager.load('recentlyUnlocked');
                if (totalStars >= stars) {
                    // unlock next mission
                    unlocked.push(next);
                    recentlyUnlocked.push(next);
                    SaveManager.save('unlockedLevels', unlocked);
                    SaveManager.save('recentlyUnlocked', recentlyUnlocked);

                    // dont click again
                    entity.removeByName('clickable');
                    entity.removeByName('tapHand');

                    // sparkle
                    Bento.audio.stopSound('sfx_itemget');
                    Bento.audio.playSound('sfx_itemget');
                    entity.attach(new Blink({}));
                    entity.attach(new SparkleBurst({}));
                    new Tween({
                        from: 1,
                        to: 0,
                        in: 30,
                        delay: 15,
                        ease: 'easeInBack',
                        onUpdate: function (v, t) {
                            entity.scale.x = v;
                            entity.scale.y = v;
                        },
                        onComplete: function () {
                            entity.removeSelf();
                            // send event to let worldselect know
                            EventSystem.fire('roadBlockRemoved');

                        }
                    });
                } else {
                    EventSystem.fire('shakeStars');
                    Bento.audio.stopSound('sfx_error');
                    Bento.audio.playSound('sfx_error');
                    new Tween({
                        from: 2,
                        to: 0,
                        in: 40,
                        ease: 'elastic',
                        decay: 2, // for 'exponential' and 'elastic'
                        oscillations: 5, // for 'elastic', 'sin' and 'cos'
                        onUpdate: function (v, t) {
                            entity.transform.x = v;
                        },
                        onComplete: function () {
                            entity.transform.x = 0;
                        }
                    });
                }
            }
        });
        var behavior = {
            name: 'behaviorComponent',
            start: function (data) {
                var unlocked = SaveManager.load('unlockedLevels');
                var missionStars = SaveManager.load('missionStars-' + unlockedBy);

                // already unlocked what it's supposed to unlock?
                if (unlocked.indexOf(next) >= 0) {
                    entity.removeSelf();
                    return;
                }

                // was required level played?
                if (missionStars >= 0) {
                    sprite.setup({
                        imageName: 'gui/worldselect/roadblock',
                        originRelative: new Vector2(0.5, 0.5)
                    });
                    isUnlocked = true;
                }

                // can open
                if (isUnlocked && totalStars >= stars) {
                    entity.attach(new TapHand({
                        position: new Vector2(-12, -12),
                        rotation: -Math.PI / 4
                    }));
                }

                // attach text after knowing the unlock stats
                text = new ShadowText({
                    position: new Vector2(8, -8),
                    text: stars.toString(),
                    font: 'font',
                    fontSize: 16,
                    fontColor: '#ffffff',
                    align: 'center',
                    textBaseline: 'middle',
                    maxWidth: 18,
                    maxHeight: 16,
                    lineWidth: isUnlocked ? 1 : 0,
                    strokeStyle: '#dfd009',
                    pixelStroke: true
                });
                entity.attach(text);
            }
        };
        var entity = new Entity({
            z: Globals.WorldSelectLayers.LEVELS,
            name: 'roadblock',
            position: settings.position,
            updateWhenPaused: 0,
            components: [
                sprite,
                behavior,
                clickable,
                new Sort({
                    baseLayer: Globals.WorldSelectLayers.LEVELS
                })
            ]
        });
        return entity;
    };
});