/**
 * Initialization of system and cordova modules
 */
bento.define('init', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/tween',
    'bento/eventsystem',
    'utils',
    'modules/tapticengine',
    'managers/resolution',
    'managers/time',
    'modules/matchmanager',
    'modules/videoadtimer',
    'modules/analytics',
    'modules/firebase',
    'managers/save',
    'modules/notifications',
    'modules/store',
    'modules/tenjin',
    'modules/cloudsave',
    'modules/gamecenter',
    'modules/googleplay',
    'modules/localization',
    'dummy',
    'modules/ktplay',
    'modules/googleplaylogin'
], function (
    Bento,
    Vector2,
    Rectangle,
    Tween,
    EventSystem,
    Utils,
    TapticEngine,
    ResolutionManager,
    TimeManager,
    MatchManager,
    VideoAdTimer,
    Analytics,
    FireBase,
    SaveManager,
    Notifications,
    Store,
    Tenjin,
    CloudSave,
    GameCenter,
    GooglePlay,
    Localization,
    Dummy,
    KTPlayModule,
    GooglePlayLogin
) {
    'use strict';
    return function () {
        /**
         * Catch uncaught errors here
         * Note: this code could be added to Bento
         */
        var catchErrors = function () {
            // attach a global timer
            var globalTimer = {
                z: 0,
                name: 'globalTimer',
                timer: 0,
                global: true,
                update: function () {
                    globalTimer.timer += 1;
                },
                destroy: function () {
                    // re-add self
                    Bento.objects.attach(globalTimer);
                }
            };
            Bento.objects.attach(globalTimer);

            // event listener for errors
            window.onerror = function (errorMsg, url, lineNumber, column, errorObj) {
                // Cocoon: column and errorObj are undefined, url is an empty string
                var globalTime = globalTimer.timer;
                var summary;
                var isFatal = false;
                var currentObject = Bento.objects.getCurrentObject();

                summary = 'line ' + lineNumber + ': ' + errorMsg;

                // detect if fatal: wait 1 second to see if globalTimer moved
                window.setTimeout(function () {
                    if (globalTimer.isAdded && globalTime === globalTimer.timer) {
                        isFatal = true;
                        console.log('Fatal error detected');
                    } else {
                        console.log('Non-fatal error detected');
                    }
                    // send event to perform game specific action
                    EventSystem.fire('UncaughtError', {
                        isFatal: isFatal,
                        line: lineNumber,
                        errorMsg: errorMsg,
                        summary: summary,
                        currentObject: currentObject
                    });
                }, 1000);

                // window.alert('Error: ' + errorMsg + ' Script: ' + url + ' Line: ' + lineNumber +
                // ' Column: ' + column + ' StackTrace: ' + JSON.stringify(errorObj));
            };
        };
        /**
         * Clears screen with black every tick (android only)
         */
        var clearScreen = function () {
            var canvasDimension = Bento.getViewport();
            var clear = function (data) {
                data.renderer.begin();
                data.renderer.fillRect([0, 0, 0, 1], 0, 0, canvasDimension.width, canvasDimension.height);
                data.renderer.flush();
            };
            if (Utils.isNativeAndroid()) {
                EventSystem.on('preDraw', clear);
            }
        };
        /**
         * Turn off antialiasing for pixel art
         */
        var antiAliasing = function () {
            if (Utils.isCocoonJS() && window.Cocoon) {
                window.Cocoon.Utils.setAntialias(false);
            }
        };
        /**
         * Event listeners
         */
        var startEventListeners = function () {
            document.addEventListener('resume', function () {
                //Resume timemanager
                TimeManager.resumeEvent();

                // Standard Firebase event: By logging this event when an App is moved to the foreground, developers can understand how often users leave and return during the course of a Session
                FireBase.logEvent(FireBase.Events.APP_OPEN);
                EventSystem.fire('onResume');

                if (SaveManager.load('notificationsAllowed')) {
                    // Note: the badge number appears to be glitchy, so let's just reset it upon opening app
                    Notifications.resetBadgeNumber();
                }

            });
            document.addEventListener('pause', function () {
                // pause the game if pausebutton exists
                if (!Bento.objects.isPaused()) {
                    Bento.objects.get('pauseButton', function (buttonPause) {
                        buttonPause.doCallback();
                    });
                }

                // fire pause event
                EventSystem.fire('onPause');
            });

            // iOS: willResignActive is fired when the app switcher is called
            document.addEventListener('resign', function () {
                // pause the game if pausebutton exists
                if (!Bento.objects.isPaused()) {
                    Bento.objects.get('pauseButton', function (buttonPause) {
                        buttonPause.doCallback();
                    });
                }
            });

            if (Utils.isCocoonJs() && window.Cocoon) {
                window.Cocoon.App.on("suspended", function () {
                    // fire event
                    EventSystem.fire('cocoonSuspended');

                    // save to cloud asap!
                    // CloudSave.saveToCloud();

                    // android only: may need to force close an online connection because android
                    // won't disconnect on its own while suspended and leaving a CCU open
                    if (Utils.isNativeAndroid() && MatchManager.shouldDisconnect) {
                        MatchManager.disconnect();
                    }
                });
            }
        };
        /**
         * Save the user's device id. Good for checking Android apks involving save hacks.
         */
        var saveFingerPrint = function () {
            if (
                Utils.isCocoonJs() &&
                window.Cocoon &&
                window.Cocoon.Device &&
                Bento.saveState.load('fingerPrint', null) === null
            ) {
                var fingerPrint = window.Cocoon.Device.getDeviceId();
                Bento.saveState.save('fingerPrint', fingerPrint);
            }
        };
        /**
         * Haptic feedback: every button should give feedback
         */
        var startHapticFeedback = function () {
            if (!TapticEngine) {
                console.log('WARNING: TapticEngine does not exist in Utils');
                return;
            }
            EventSystem.on('clickButton-onButtonDown', function (data) {
                TapticEngine.impact({
                    style: 'light'
                });
            });
            EventSystem.on('toggleButton-onToggle', function (data) {
                TapticEngine.selection();
            });
            // for testing purposes: replace the official API with unofficial for iphone6
            if (Bento.isDev() && window.device && window.device.model === 'iPhone8,1') {
                TapticEngine.initUnofficial();
            }
            // Utils.toggleTapticEngine(Bento.saveState.load(Utils.SaveKeys.TAPTICENGINE, true));
            TapticEngine.enable(!Bento.saveState.load('disableTapticEngine', false));
        };
        /**
         * Initialize GameServices (GameCenter/GooglePlay)
         * While doing so, try to retrieve the player name
         */
        var initGameServices = function () {
            if (Utils.isCocoonJS()) {
                // only on cocoon
                if (Utils.isNativeIos()) {
                    // iOS only: Game Center
                    GameCenter.login(function (hasLoggedIn, response) {
                        if (hasLoggedIn && response.alias) {
                            // set the player alias here
                            if (SaveManager.load('localPlayerName', 'Guest') === 'Guest') {
                                SaveManager.save('localPlayerName', response.alias);
                            }
                            // set the player id here
                            if (SaveManager.load('gameServiceID', null) === null) {
                                SaveManager.save('gameServiceID', response.playerID);
                            }
                        }
                    });
                } else if (Utils.isNativeAndroid()) {
                    // Google Play is initialized if user agreed to login in the past
                    if (Bento.saveState.load('loginAttemptGooglePlay', false)) {
                        GooglePlayLogin(function () {});
                    }
                }
            }
        };
        /**
         * Initialize KTplay
         */
        var initKTplay = function () {
            //set this function beforehand as KTPlay can log in as soon as initialized
            KTPlayModule.setLoginOutHandler(function (data) {
                //if we just logged in
                if (data.loggedin) {
                    KTPlayModule.getCurrentUser(function (user) {
                        // get user ID integer
                        var userID = parseInt(user.userID);
                        //double check the number is an integer
                        if (!isNaN(userID)) {
                            console.log('Logged in to KTPlay with User ID: ' + userID);
                            // first time logged in with this account
                            if (Bento.saveState.load('sentKTUserID', null) === null) {
                                Bento.saveState.save('sentKTUserID', userID);
                                Tenjin.sendEventAndValue('KTuserID', userID);
                            } else {
                                //we've just changed account
                                if (Bento.saveState.load('sentKTUserID', userID) !== userID) {
                                    Bento.saveState.save('sentKTUserID', userID);
                                    Tenjin.sendEventAndValue('KTuserID', userID);
                                }
                            }
                        } else {
                            console.log('ERROR: User ID is not a number: ' + userID);
                        }
                    }, function (error) {
                        console.log(error);
                    });
                }
            });
            //if the community window closed we could have changed the nick name, so we re save it again anyway
            KTPlayModule.setWindowHandler(function (data) {
                var now = Date.now();
                var timeDiff;
                var timeDiffMin;
                if (!data.communitywindow) {
                    // closed
                    Utils.changeLocalNameToKTNickName();

                    timeDiff = now - Bento.saveState.load('KTPlayOpenTime', now);
                    // conver to minutes
                    timeDiffMin = timeDiff / (1000 * 60);
                    // send analytics
                    Analytics.event('KTPlay', 'AverageMinutes', timeDiffMin);

                    // send event for >1, 2, 5 minutes
                    if (timeDiffMin > 5) {
                        Analytics.event('KTPlay', 'Over5Minutes', timeDiffMin);
                    }
                    if (timeDiffMin > 2) {
                        Analytics.event('KTPlay', 'Over2Minutes', timeDiffMin);
                    }
                    if (timeDiffMin > 1) {
                        Analytics.event('KTPlay', 'Over1Minute', timeDiffMin);
                    } else {
                        Analytics.event('KTPlay', 'LessThan1Minute', timeDiffMin);
                    }
                } else {
                    // opened
                    Bento.saveState.save('KTPlayOpenTime', now);
                }
            });
            //initialize
            KTPlayModule.init({
                key: "koqvzjm9X",
                secret: "1672065ee2ea34b998efefaf09bea0f79c277638",
            });
        };
        /**
         * Initialize Analytics
         */
        var initAnalytics = function () {
            Analytics.init();
            if (Bento.saveState.load('showDeveloperCheats', false)) {
                console.log('Analytics turned off for developer');
                Analytics.setActive(false);
            }
            FireBase.init();
            Tenjin.sendEvent('bootup');

            // use events to send analytics
            EventSystem.on('Analytics', function (data) {
                if (!data.category || !data.action) {
                    // need at least category and action
                    Utils.log('Error: wrong Analytics parameters');
                    return;
                }
                Analytics.event(data.category, data.action, data.value, data.label);
            });
            EventSystem.on('Tenjin', function (data) {
                var eventName = data.category || data.action || data.name || data.event;
                if (!eventName) {
                    Utils.log('Error: wrong Analytics parameters');
                    return;
                }
                if (data.value) {
                    Tenjin.sendEventAndValue(eventName, data.value);
                } else {
                    Tenjin.sendEvent(eventName);
                }
            });
        };
        var initNotifications = function () {
            var gameHasInitialized = false;
            var scheduleNotificationInactive = function () {
                if (!gameHasInitialized) {
                    return;
                }
                var savedNotifications = SaveManager.load('notifications');
                var deltaTime = 48;

                // reset badge
                Notifications.resetBadgeNumber();

                // cancel old notification if needed
                var oldId = savedNotifications['inactive'];
                var oldIdWeek = savedNotifications['inactiveWeek'];
                var oldIdWeek2 = savedNotifications['inactiveWeek2'];
                var oldIdWeek3 = savedNotifications['inactiveWeek3'];
                if (oldId) {
                    Notifications.cancel(oldId);
                }
                if (oldIdWeek) {
                    Notifications.cancel(oldIdWeek);
                }
                if (oldIdWeek2) {
                    Notifications.cancel(oldIdWeek2);
                }
                if (oldIdWeek3) {
                    Notifications.cancel(oldIdWeek3);
                }

                // schedule notification
                var notifications = Utils.getNotifications({
                    Localization: Localization,
                    time: deltaTime * 60 * 60 * 1000
                });
                var id = Notifications.send(notifications.inactive);
                // save id so it can be cancelled if needed
                savedNotifications['inactive'] = id;


                // repeat for weeks
                notifications = Utils.getNotifications({
                    Localization: Localization,
                    time: 7 * 24 * 60 * 60 * 1000
                });
                id = Notifications.send(notifications.inactive);
                // save id so it can be cancelled if needed
                savedNotifications['inactiveWeek'] = id;

                notifications = Utils.getNotifications({
                    Localization: Localization,
                    time: 2 * 7 * 24 * 60 * 60 * 1000
                });
                id = Notifications.send(notifications.inactive);
                // save id so it can be cancelled if needed
                savedNotifications['inactiveWeek2'] = id;

                notifications = Utils.getNotifications({
                    Localization: Localization,
                    time: 3 * 7 * 24 * 60 * 60 * 1000
                });
                id = Notifications.send(notifications.inactive);
                // save id so it can be cancelled if needed
                savedNotifications['inactiveWeek3'] = id;

                SaveManager.save('notifications', savedNotifications);
            };
            var energyNotification = function (energyData) {
                if (!gameHasInitialized) {
                    return;
                }
                if (!energyData) {
                    return;
                }
                var savedNotifications = SaveManager.load('notifications');
                var savedNotification;
                var isAllowed = SaveManager.load('notificationsAllowed'); // notifications could be initialized mid-game
                var longestDelta = 0;

                if (isAllowed) {
                    // cancel old notification
                    savedNotification = savedNotifications['energy'];
                    if (savedNotification) {
                        Notifications.cancel(savedNotification);
                        savedNotification = null;
                    }

                    // schedule new one into the future, whose energy takes the longest
                    Utils.forEach(energyData, function (energyTimeData, name) {
                        if (energyTimeData && energyTimeData.full > longestDelta) {
                            longestDelta = energyTimeData.full;
                        }
                    });
                    if (longestDelta > 0) {
                        // schedule notification
                        var notifications = Utils.getNotifications({
                            Localization: Localization,
                            time: longestDelta
                        });
                        savedNotification = Notifications.send(notifications.energy);
                    }
                    // save id so it can be cancelled if needed
                    if (savedNotification) {
                        savedNotifications['energy'] = savedNotification;
                    }
                    SaveManager.save('notifications', savedNotifications);
                }
            };

            // wait for the preloader is finished and loaded the localization
            EventSystem.on('gameInit', function () {
                gameHasInitialized = true;
            });

            // if user has allowed notifications, initialize notification system
            if (SaveManager.load('notificationsAllowed')) {
                Notifications.init();

                // Note: the badge number appears to be glitchy, so let's just reset it upon opening
                Notifications.resetBadgeNumber();

                // schedule for inactive users
                EventSystem.on('gameInit', scheduleNotificationInactive);
                // reschedule on resume
                EventSystem.on('onResume', scheduleNotificationInactive);
            } else {
                // wait until user allows notifications to schedule inactive
                EventSystem.on('allowedNotifications', function () {
                    scheduleNotificationInactive();
                    // reschedule on resume
                    EventSystem.on('onResume', scheduleNotificationInactive);
                    // done
                    EventSystem.clear('allowedNotifications');
                });
            }

            // global notification planner whenever energy changes
            EventSystem.on('energySubtracted', energyNotification);

        };
        var initIAP = function () {
            Store.init(Utils.IAP);
        };
        var addInGameTimer = function () {
            Bento.objects.attach(VideoAdTimer);
        };
        var checkVersion = function () {
            // for both iOS and Android we check the version on the App Store
            var link = 'http://itunes.apple.com/lookup?bundleId=com.luckykat.beatstreet';
            var xhr = new window.XMLHttpRequest();
            xhr.open("GET", link);
            xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
            xhr.send();
            xhr.onreadystatechange = function () {
                var response;
                var data;
                var storeVersion;
                var currentVersion;
                var getVersionNumber = function (strArray) {
                    return parseInt(strArray[0]) * 10000 + parseInt(strArray[1]) * 100 + parseInt(strArray[2]);
                };
                if (xhr.readyState === 4) {
                    if (xhr.status >= 200 && xhr.status < 300) {
                        try {
                            response = JSON.parse(xhr.responseText);
                            if (response.results && response.results.length) {
                                data = response.results[0];
                                storeVersion = getVersionNumber(data.version.split('.'));
                                currentVersion = getVersionNumber(window.VERSION.split('.'));
                                // store version in Utils
                                if (storeVersion > currentVersion) {
                                    Utils.newVersion = data.version;
                                }
                            }
                        } catch (e) {
                            // do nothing
                        }
                    } else {
                        // do nothing
                    }
                }
            };
        };
        var detectJailBreak = function () {
            // detect jailbreak or rooted devices
            if (window.IRoot) {
                window.IRoot.isRooted(function (isJailbroken) {
                    if (isJailbroken) {
                        Bento.saveState.save('jailbreak', true);
                        // in consideration: are they cheat suspects?
                        // Bento.saveState.save('cheatSuspect', true);
                        Analytics.event('Jailbreak', 'jailbreak', 1);
                    } else {
                        Analytics.event('Jailbreak', 'jailbreak', 0);
                    }
                }, function () {});
            }
            if (Bento.saveState.load('showDeveloperCheats', false) || Bento.isDev()) {
                Bento.saveState.save('jailbreak', false);
                Bento.saveState.save('fakeIAP', false);
                Bento.saveState.save('cheatSuspect', false);
            }
        };

        // android: always initialize notifications
        if (Utils.isNativeAndroid()) {
            SaveManager.save('notificationsAllowed', true);
        }

        catchErrors();
        clearScreen();
        antiAliasing();
        startEventListeners();
        saveFingerPrint();
        startHapticFeedback();
        initGameServices();
        initAnalytics();
        initKTplay();
        if (!Bento.isDev()) {
            // devapp crashes on notifications
            initNotifications();
        } else {
            console.log('Notifications not initialized during dev mode.');
        }
        detectJailBreak();
        initIAP();
        addInGameTimer();
        // check for new updates asap
        checkVersion();

        ResolutionManager.init();

        /**
         * Start preloader
         */
        Bento.assets.load('jpg-preloader', function (err) {
            if (err) {
                console.log(err);
            }
            Bento.assets.load('preloader', function () {
                Bento.screens.show('screens/preloader');
            });
        });
    };
});