/**
 * Ai manager. Keeps track of all the players and all the enemies in a stage. Gives broad instructions to enemyconrollers (who to target, what to do)
 * @moduleName AiManager
 */
bento.define('managers/ai', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    MatchManager
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */

    var enemies = [];

    var players = [];

    var isListening = false;

    var registerEnemy = function (enemyController) {
        var enemyData = {
            id: Utils.checksum(enemyController.parent.name),
            controller: enemyController,
            entity: enemyController.parent,
            target: null
        };
        enemies.push(enemyData);

        distributeAiTasks();
    };

    var unregisterEnemy = function (enemyController) {
        var index = -1;
        var enemy;
        var ii;
        for (ii = 0; ii < enemies.length; ii++) {
            if (enemies[ii].controller === enemyController) {
                index = ii;
                break;
            }
        }
        if (index !== -1) {
            enemy = enemies[ii];

            if (enemy.target) {
                untargetPlayer(enemy, enemy.target);
            }
            enemies.splice(index, 1);

            distributeAiTasks();
        }
    };

    var onScreenChange = function () {
        enemies = [];
        players = [];
    };

    var targetPlayer = function (enemyData, playerData, side) {
        enemyData.target = playerData;

        if (side && side !== 'none') {
            enemyData.controller.setTarget(playerData.entity, side === 'left' ? -1 : 1);
            enemyData.controller.aiState = AiManager.AiStates.CHASE;

            // occupy the side
            playerData[side] = enemyData;
        } else {
            // it is possible to target a player, but not be chasing
            enemyData.controller.setTarget(playerData.entity);
            enemyData.controller.aiState = AiManager.AiStates.IDLE;
        }
    };

    var getSide = function (value) {
        if (value === -1) {
            return 'attackLeft';
        } else if (value === 1) {
            return 'attackRight';
        }
        return '';
    };

    var applyForAttackSide = function (data) {
        if (data.target.indexOf('player') === -1) {
            data.succes();
        } else {
            // Note: in a perfect world we'd sync this with other players

            // check side directly
            var playerId = data.target.split('-')[1];
            var side = getSide(data.side);
            if (players[playerId] && players[playerId][side] && players[playerId][side] !== data.id) {
                data.fail();
            } else {
                // console.log('aiManager assigned', side, data.id, players[playerId][side]);
                data.succes();
                if (players[playerId]) {
                    players[playerId][side] = data.id;
                }
            }
        }
    };

    var forceAttackSide = function (data) {
        if (data.target.indexOf('player') === -1) {
            data.succes();
        } else {
            if (MatchManager.isMaster) {} else {
                var playerId = data.target.split('-')[1];
                var side = getSide(data.side);
                if (players[playerId]) {
                    players[playerId][side] = data.id;
                }
            }
        }
    };

    var revokeAttackSide = function (data) {
        // console.log('AiManager-revoke')
        if (data.target.indexOf('player') > -1) {
            // Note: in a perfect world we'd sync this with other players
            // if (MatchManager.isMaster) {
            // } else {
            //     // revoke at master
            // }
            // check side directly
            var playerId = data.target.split('-')[1];
            if (!players[playerId]) {
                return;
            }
            var side = getSide(data.side);

            if (players[playerId][side] !== data.id) {
                // console.log('aiManager revoke fail', side, data.id, players[playerId][side]);
                return;
            }
            // console.log('aiManager revoke succes', side, data.id, players[playerId][side]);
            players[playerId][side] = null;
        }
    };

    var getLocalAttackSide = function (data) {
        if (data.target.indexOf('player') === -1) {
            data.succes();
        } else {
            // check side directly
            var playerId = data.target.split('-')[1];
            var side = getSide(data.side);
            if (!players[playerId]) {
                // line 162: TypeError: undefined is not an object (evaluating 'players[playerId][side]') Object: character-108
                if (data.fail) {
                    data.fail();
                }
                return;
            }
            if (players[playerId][side] === data.id) {
                if (data.succes) {
                    data.succes();
                }
            } else {
                data.fail();
            }
        }
    };

    var untargetPlayer = function (enemyData, playerData) {
        if (playerData.right === enemyData) {
            playerData.right = null;
        }
        if (playerData.left === enemyData) {
            playerData.left = null;
        }
        enemyData.controller.setTarget(null);
        enemyData.target = null;
        enemyData.controller.aiState = AiManager.AiStates.IDLE;
    };

    var findPlayers = function () {
        var playerList = Bento.objects.getByFamily('player');
        var ii;
        var playerData;

        // reset players
        players = [];

        for (ii = 0; ii < playerList.length; ii++) {
            playerData = {
                id: playerList[ii].playerId,
                entity: playerList[ii],
                left: null,
                right: null,
                attackLeft: null,
                attackRight: null,
            };
            players.push(playerData);
        }
    };

    var distributeAiTasks = function () {
        //this functions rids every enemyController of it's current target, and figures out which enemies should chase which player
        var ii;
        var scores = [];
        var idx;

        // collect players
        // TODO: check if alive etc
        findPlayers();

        // reset enemies
        Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
            if (enemy.target) {
                untargetPlayer(enemy, enemy.target);
                scores[ii] = {
                    idx: ii,
                    score: 0
                };
            }
            enemy.controller.aiState = AiManager.AiStates.IDLE;
        });

        if (MatchManager.isOnline) {
            // there should be enemies targeting the local player
            distributeEnemiesOnline();
        } else {
            distributeEnemiesOffline();
        }

    };
    var distributeEnemiesOffline = function () {
        // double pass targeting: first try to assign 1 enemy per player
        // then loop through left over enemies to assign a 2nd enemy to players
        Utils.forEach(players, function (playerData) {
            var closestEnemies = Utils.copyArray(enemies);
            // sort array by closest
            closestEnemies.sort(function (enemyDataA, enemyDataB) {
                var diffA = enemyDataA.entity.position.subtract(playerData.entity.position);
                var diffB = enemyDataB.entity.position.subtract(playerData.entity.position);
                var distA = diffA.sqrMagnitude();
                var distB = diffB.sqrMagnitude();
                return distA - distB;
            });

            // target player
            Utils.forEach(closestEnemies, function (enemyData, i, l, breakLoop) {
                // check if enemy is not "inactive"
                var skip = false;
                enemyData.entity.getComponent('enemyController', function (enemyController) {
                    if (!enemyController.active) {
                        skip = true;
                    }
                });
                if (skip) {
                    return;
                }
                target(enemyData, playerData, breakLoop);
            });
        });

        // 2nd pass: inverse check
        Utils.forEach(enemies, function (enemyData) {
            // check if enemy is not "inactive"
            var skip = false;
            enemyData.entity.getComponent('enemyController', function (enemyController) {
                if (!enemyController.active) {
                    skip = true;
                }
            });
            // during 2nd pass: skip enemies that are already targeting
            if (enemyData.target) {
                skip = true;
            }
            if (skip) {
                return;
            }

            var closestPlayers = Utils.copyArray(players);
            // sort array by closest
            closestPlayers.sort(function (playerDataA, playerDataB) {
                var diffA = playerDataA.entity.position.subtract(enemyData.entity.position);
                var diffB = playerDataB.entity.position.subtract(enemyData.entity.position);
                var distA = diffA.sqrMagnitude();
                var distB = diffB.sqrMagnitude();
                return distA - distB;
            });

            // target player
            Utils.forEach(closestPlayers, function (playerData, i, l, breakLoop) {
                target(enemyData, playerData, breakLoop);
            });
        });

        // randomize the rest
        Utils.forEach(enemies, function (enemyData) {
            if (enemyData.target) {
                return;
            }
            // target a random player (deterministic)
            var index = Utils.modulo(enemyData.id, players.length);
            var playerData = players[index];
            if (playerData) {
                targetPlayer(enemyData, playerData, 'none');
            }
        });
    };
    var distributeEnemiesOnline = function () {
        var skip = false;
        var localId;
        var localPlayer;
        var localPlayerData = MatchManager.localPlayer;
        var playerState;
        if (localPlayerData) {
            localId = localPlayerData.playerId;
        }
        // find local player
        Utils.forEach(players, function (playerData, i, l, breakLoop) {
            if (playerData.entity.playerId === localId) {
                localPlayer = playerData;
            }
        });
        if (!localPlayer) {
            return;
        }

        playerState = localPlayer.entity.getComponent('stateMachine');

        // enemy controller should have a targetPlayerId
        Utils.forEach(enemies, function (enemyData, i, l, breakLoop) {
            enemyData.entity.getComponent('enemyController', function (enemyController) {
                var targetPlayerId = enemyController.targetPlayerId;
                if (enemyController.active) {
                    if (enemyController.isLocallyControlled) {
                        // set local
                        enemyData.entity.getComponent('online', function (online) {
                            online.setRemote(false);
                        });

                        if (playerState.stateName !== 'dead') {
                            // player should not be dead!
                            target(enemyData, localPlayer, function () {});
                        }
                    } else {
                        // set remote
                        enemyData.entity.getComponent('online', function (online) {
                            online.setRemote(true);
                        });
                    }
                }
            });
        });

        if (playerState.stateName === 'dead') {
            distributeEnemiesOffline();
        }
    };
    var target = function (enemyData, playerData, breakLoop) {
        // there is no place left to target player: try next in list
        // the player should not be dead
        var stateMachine = playerData.entity.getComponent('stateMachine');
        if (stateMachine && stateMachine.stateName === 'dead') {
            return;
        }

        // try preferred side
        if (playerData.right === null && enemyData.entity.position.x > playerData.entity.position.x) {
            targetPlayer(enemyData, playerData, 'right');
            breakLoop();
        } else if (playerData.left === null && enemyData.entity.position.x < playerData.entity.position.x) {
            targetPlayer(enemyData, playerData, 'left');
            breakLoop();
        } else {
            // try to take whatever is left
            if (!playerData.right) {
                targetPlayer(enemyData, playerData, 'right');
            } else if (!playerData.left) {
                targetPlayer(enemyData, playerData, 'left');
            } else {
                var rndSide = 'left';
                if (Utils.getRandom(2)) {
                    rndSide = 'right';
                }
                targetPlayer(enemyData, playerData, rndSide);
            }

            breakLoop();
        }
    };
    var assignNewTask = function () {
        // TODO: allow an enemy to reassign itself (for example when its being attacked)
    };

    EventSystem.clear('screenHidden');
    EventSystem.on('screenHidden', onScreenChange);

    var AiManager = {
        registerEnemy: registerEnemy,
        unregisterEnemy: unregisterEnemy,
        distributeAiTasks: distributeAiTasks,
        assignNewTask: assignNewTask,
        applyForAttackSide: applyForAttackSide,
        revokeAttackSide: revokeAttackSide,
        getLocalAttackSide: getLocalAttackSide,
        forceAttackSide: forceAttackSide
    };

    // enum
    AiManager.AiStates = {
        IDLE: 0,
        CHASE: 1
    };

    // debugging: show ai manager state
    if (Bento.isDev()) {
        window.debugAiManager = function () {
            console.log('enemyData', enemies);
            console.log('playerData', players);
        };
    }

    return AiManager;
});