/**
 * Allies manager
 * @moduleName AlliesManager
 */
bento.define('managers/allies', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'managers/save',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    SaveManager,
    Globals
) {
    'use strict';
    var hasInitialized = false;
    var availableAllies = [
        'rat',
        'bigrat',
        'weasel',
        'raccoon',
        'ratguy',
        'ratgirl',
        'pig',
        'gangster',
        'punker',
        'punkgirl',
        'mrpigeon',
        'dog',
        'cat',
        'twerka',
        'ratboss',
        'pigchef',
        'gangstaboss'
    ];
    var maxLevels = {
        'rat': 50,
        'bigrat': 50,
        'weasel': 50,
        'raccoon': 45,
        'ratguy': 50,
        'ratgirl': 50,
        'pig': 50,
        'gangster': 25,
        'punker': 40,
        'punkgirl': 50,
        'mrpigeon': 50,
        'cat': 50,
        'dog': 50,
        'twerka': 45,
        'ratboss': 10,
        'pigchef': 50,
        'gangstaboss': 50
    };
    var getAvailable = function () {
        // we check available through the existing enemytokens
        var assets = Bento.assets.getAssets();
        var jsonAssets = assets.json;

        Utils.forEach(jsonAssets, function (item, key, l, breakLoop) {
            var keys = key.split('/');
            if (keys[0] === 'enemytokens' && availableAllies.indexOf(keys[1]) < 0) {
                availableAllies.push(keys[1]);
            }
        });
        hasInitialized = true;
    };

    var getEffectiveness = function (allyName, localizedText) {
        var level = AlliesManager.getLevel(allyName);
        var lv = level - 1;
        var effect = {
            value: 0
        };

        if (localizedText) {
            effect.string = localizedText;
        }
        switch (allyName) {
        case 'rat':
            // max health 
            effect.value = 5 + lv * 5;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'bigrat':
            // attack 
            effect.value = Math.ceil(2 + lv * 0.75);
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'weasel':
            // power up seconds
            effect.value = 15 + lv * 5;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'raccoon':
            // weapon throw damage
            effect.value = 2 + lv * 2;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'pig':
            // extra coins
            effect.value = 10 + lv * 1;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value));
            }
            break;
        case 'ratguy':
            // skill
            effect.value = 1 + lv * 1;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'ratgirl':
            // weapon durability %
            // up to 100%
            effect.value = 0.10 + Utils.clamp(0, lv, 90) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'gangster':
            // % decrease of obstacle + projectile damage
            // up to 75%
            effect.value = 0.25 + Utils.clamp(0, lv, 25) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'punker':
            // Start with extra combo meter
            // Up to 50%
            effect.value = 0.10 + Utils.clamp(0, lv, 40) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'punkgirl':
            // weapon attack
            effect.value = 2 + lv * 1;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        case 'mrpigeon':
            // % critical hit dash attack
            // up to 50%
            effect.value = 0.05 + Utils.clamp(0, lv, 45) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'cat':
            // % critical hit combo attack
            // up to 50%
            effect.value = 0.05 + Utils.clamp(0, lv, 45) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'dog':
            // % critical hit charge attack
            // up to 50%
            effect.value = 0.05 + Utils.clamp(0, lv, 45) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'twerka':
            // % chance of getting up once
            // up to 50%
            effect.value = 0.10 + Utils.clamp(0, lv, 45) * 0.01;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', Math.round(effect.value * 100));
            }
            break;
        case 'ratboss':
            // Combo meter fills up faster
            effect.value = 1 + lv * 0.5;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value.toFixed(1)); // show 1 decimal
            }
            break;
        case 'pigchef':
            // X seconds longer invincibility
            effect.value = 60 * 2 + lv * 30;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', (effect.value / 60).toFixed(1)); // show 1 decimal
            }
            break;
        case 'gangstaboss':
            // 0 damage for X hits
            effect.value = 1 + lv * 1;
            if (effect.string) {
                effect.string = effect.string.replace('{NUMBER}', effect.value);
            }
            break;
        default:
            break;
        }
        return effect;
    };

    var AlliesManager = {
        /**
         * Get names of all available allies
         * @return {Array} Array of Strings
         */
        getAvailable: function () {
            // initialize if needed (note: don't call this before assets are loaded)
            if (!hasInitialized) {
                getAvailable();
            }
            return availableAllies;
        },
        /**
         * Whether an ally was unlocked
         * @param  {String}  allyName Id of ally
         * @return {Boolean}          Unlocked or not
         */
        isUnlocked: function (allyName) {
            var allies = SaveManager.load('unlockedAllies');
            return allies.indexOf(allyName) >= 0;
        },
        /**
         * Unlock an ally
         * @param  {String} allyName Id of Ally
         */
        unlock: function (allyName) {
            var allies = SaveManager.load('unlockedAllies');
            if (allies.indexOf(allyName) >= 0) {
                // already unlocked
                return;
            }
            allies.push(allyName);
            SaveManager.save('unlockedAllies', allies);
        },
        /**
         * Get amount of coins needed to level up
         * @param  {String} allyName Id of ally
         * @return {Number}          Number of coins neeeded
         */
        getUpgradeCostCoins: function (allyName, diff) {
            // TODO should this be the same as playable characters?
            var level = AlliesManager.getLevel(allyName);
            return Globals.getLevelUpCost(level, diff);
        },
        /**
         * Get amount of tokens needed to level up
         * @param  {String} allyName Id of ally
         * @return {Number}          Number of tokens neeeded
         */
        getUpgradeCost: function (allyName, diff) {
            // TODO should this be the same as playable characters?
            var level = AlliesManager.getLevel(allyName);
            return Globals.getTokenCost(level, diff);
        },
        /**
         * Get amount of tokens needed to unlock
         * @param  {Strign} allyName Id of ally
         * @return {Number}          Number of tokens needed
         */
        getUnlockCost: function (allyName) {
            // unlock cost is located in the token json
            var json = Bento.assets.getJson('enemytokens/' + allyName);
            return json.unlockCost;
        },
        /**
         * Get level of particular ally
         * @param  {String} allyName Id of ally
         * @return {Number}          Level of ally
         */
        getLevel: function (allyName) {
            return SaveManager.load('level-' + allyName);
        },
        /**
         * Get max level of particular ally
         * @param  {String} allyName Id of ally
         * @return {Number}          Level of ally
         */
        getMaxLevel: function (allyName) {
            return maxLevels[allyName];
        },
        /**
         * Level up ally
         * @param  {String} allyName Id of ally
         */
        addLevel: function (allyName) {
            SaveManager.add('level-' + allyName, 1);
        },
        /**
         * Get currently selected allies
         * @return {Array}
         */
        getAllies: function () {
            return SaveManager.load('allies');
        },
        /**
         * Get selected ally in slot
         * @param  {Number} index Index number
         * @return {String}       Name of ally
         */
        getAlly: function (index) {
            return SaveManager.load('allies')[index];
        },
        /**
         * Get effect per ally per level
         * @param  {String} allyName Id of ally
         * @param  {String} Localized string of effect, where "{NUMBER}" needs to be replaced
         * @return {Object} effect
         *                  effect.value - value of effect
         *                  effect.string - localized string of effect
         */
        getEffectiveness: getEffectiveness
    };
    return AlliesManager;
});