/**
 * Character manager, manages character energy, and wether they are unlocked or not
 *
 * @moduleName CharacterManager
 */
bento.define('managers/character', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'managers/save',
    'managers/time',
    'entities/players/boy',
    'entities/players/fitgirl',
    'entities/players/krillin',
    'entities/players/boxer',
    'entities/players/jun',
    'entities/players/val',
    'entities/players/pat',
    'entities/players/oni',
    'entities/players/rad'
], function (
    Bento,
    Vector2,
    Rectangle,
    EventSystem,
    Utils,
    Tween,
    SaveManager,
    TimeManager,
    PlayerBoy,
    PlayerFitGirl,
    PlayerKrillin,
    PlayerBoxer,
    PlayerJun,
    PlayerVal,
    PlayerPat,
    PlayerOni,
    PlayerRad
) {
    'use strict';
    /**
     * Describe your settings object parameters
     * @param {Object} settings
     */
    var CHARACTER_ORDER = [
        'boy',
        'krillin',
        'boxer',
        'val',
        'jun',
        'pat',
        'fitgirl',
        'oni',
        'rad'
    ];
    var characterConstructors = {
        'boy': PlayerBoy,
        'krillin': PlayerKrillin,
        'fitgirl': PlayerFitGirl,
        'boxer': PlayerBoxer,
        'jun': PlayerJun,
        'val': PlayerVal,
        'pat': PlayerPat,
        'oni': PlayerOni,
        'rad': PlayerRad,
    };
    var minimumLevels = {
        'boy': 1,
        'krillin': 1,
        'boxer': 1,
        'fitgirl': 1,
        'jun': 1,
        'val': 1,
        'pat': 1,
        'oni': 1,
        'rad': 1
    };

    // evolution and unlock data
    var characterData = {
        boy: {
            maxLevel: 10,
            evolveItems: 20, // DEPRECATED evolve items needed
            maxTotal: Infinity,
            unlocksAt: 0,
            tokensToUnlock: 10
        },
        krillin: {
            maxLevel: 10,
            evolveItems: 20,
            maxTotal: Infinity,
            unlocksAt: 5,
            tokensToUnlock: 5
        },
        boxer: {
            maxLevel: 20,
            evolveItems: 40,
            maxTotal: Infinity,
            unlocksAt: 10,
            tokensToUnlock: 11
        },
        val: {
            maxLevel: 20,
            evolveItems: 40,
            maxTotal: Infinity,
            unlocksAt: 20,
            tokensToUnlock: 20
        },
        jun: {
            maxLevel: 30,
            evolveItems: 80,
            maxTotal: Infinity,
            unlocksAt: 30,
            tokensToUnlock: 32
        },
        pat: {
            maxLevel: 30,
            evolveItems: 80,
            maxTotal: Infinity,
            unlocksAt: 40,
            tokensToUnlock: 44
        },
        fitgirl: {
            maxLevel: 40,
            evolveItems: 160,
            maxTotal: Infinity,
            unlocksAt: 50,
            tokensToUnlock: 50
        },
        oni: {
            maxLevel: 40,
            evolveItems: 160,
            maxTotal: Infinity,
            unlocksAt: 1,
            tokensToUnlock: 62
        },
        rad: {
            maxLevel: 40,
            evolveItems: 160,
            maxTotal: Infinity,
            unlocksAt: 1,
            tokensToUnlock: 150
        }
    };
    var MAX_STAMINA = 20;
    var REGENTIME = 1000 * 60 * 10; //10 minutes

    var started = false;
    var updateEnergy = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = SaveManager.load('energy-' + char);
        var hasChanged = false;

        var currentTime = TimeManager.getTime();
        var timeDifference = 0;

        if (charData.energy < charData.maximum) {
            if (charData.regen === null) {
                charData.regen = TimeManager.getTime();
                hasChanged = true;
            }
            timeDifference = currentTime - charData.regen;

            //add energy until the time has caught up
            while (timeDifference >= REGENTIME) {
                charData.energy += 1;
                charData.regen += REGENTIME;
                hasChanged = true;
                if (charData.energy === charData.maximum) {
                    break;
                }
                timeDifference -= REGENTIME;
            }
        }

        if (hasChanged) {
            if (charData.energy === charData.maximum) {
                charData.regen = null;
            }
            SaveManager.save('energy-' + char, charData);
        }

        return charData;
    };


    var isUnlocked = function (char) {
        if (CHARACTER_ORDER.indexOf(char) === -1) {
            return false;
        }

        if (SaveManager.load('unlockedCharacters').indexOf(char) === -1) {
            return false;
        }

        return true;
    };

    var unlock = function (char, force) {
        if (CHARACTER_ORDER.indexOf(char) === -1) {
            Utils.log('WARNING: trying to unlock non-existing character');
            return false;
        }

        var unlocked = SaveManager.load('unlockedCharacters');

        if (unlocked.indexOf(char) !== -1) {
            return false;
        }

        unlocked.push(char);
        SaveManager.save('unlockedCharacters', unlocked);

        // Post achievement for if they haven't got the 'Double Trouble' achievement yet
        Utils.postAchievement("CgkI9NqdkoAGEAIQDA");

        var charData = createCharacter(char);
        SaveManager.save('energy-' + char, charData);
    };

    var loadCharacter = function (char) {
        var charData = SaveManager.load('energy-' + char);
        if (charData) {
            return;
        }
        charData = createCharacter(char);
        SaveManager.save('energy-' + char, charData);
    };

    var createCharacter = function () {
        //load JSON default variables for the character
        var energy = 20;
        var maximum = 20;
        var charData = {
            energy: energy,
            maximum: energy,
            regen: null
        };
        return charData;
    };

    var addEnergy = function (amount) {
        var charData = updateEnergy('boy');
        charData.energy += amount;
        SaveManager.save('energy-boy', charData);
        return charData.energy;
    };

    var subtractEnergy = function (amount) {
        var charData = updateEnergy('boy');
        if (amount > 0) {
            if (charData.energy === charData.maximum) {
                charData.regen = TimeManager.getTime();
            }
        }
        charData.energy -= amount;
        if (charData.energy < 0) {
            charData.energy = 0;
        }
        SaveManager.save('energy-boy', charData);

        EventSystem.fire('quest-energy', {
            count: amount
        });

        // pass all character energy data to event
        var unlocked = SaveManager.load('unlockedCharacters');
        var energyData = {};
        Utils.forEach(unlocked, function (character) {
            energyData[character] = getEnergyTime(character);
        });
        EventSystem.fire('energySubtracted', energyData);

        return charData.energy;
    };

    var setEnergy = function (_char, amount) {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = updateEnergy(char);
        charData.energy = amount;
        SaveManager.save('energy-' + char, charData);
    };

    var getEnergy = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = updateEnergy(char);
        return charData.energy;
    };

    var hasEnergy = function (amount) {
        if (getEnergy('boy') >= amount) {
            return true;
        }
        return false;
    };

    var getMaximum = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = updateEnergy(char);
        return charData.maximum;
    };

    var setEnergyTime = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = SaveManager.load('energy-' + char);
        charData.regen = TimeManager.getTime();
    };

    var getEnergyTime = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        var charData = updateEnergy(char);

        if (charData.energy >= charData.maximum) {
            return {
                next: 0,
                full: 0
            };
        }

        var now = TimeManager.getTime();
        var timePassed = now - charData.regen;
        var online = TimeManager.online;
        if (Utils.isDev()) {
            online = true;
        }

        return {
            next: (REGENTIME - timePassed) % REGENTIME,
            full: (charData.maximum - charData.energy) * REGENTIME - timePassed,
            online: online
        };
    };

    var getSelected = function () {
        return SaveManager.load('selectedCharacter');
    };

    var setSelected = function (char) {
        if (!isUnlocked(char)) {
            return;
        }
        SaveManager.save('selectedCharacter', char);
        EventSystem.fire('updateCharacter', char);
    };

    var getEnergyData = function () {
        var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
        return SaveManager.load('energy-' + char);
    };


    var CharacterManager = {
        isUnlocked: isUnlocked,
        unlock: unlock,

        addEnergy: addEnergy,
        subtractEnergy: subtractEnergy,
        setEnergy: setEnergy,
        getEnergy: getEnergy,
        getEnergyRelative: function () {
            var char = 'boy'; // NOTE: character per energy is deprecated, point everything to boy
            return getEnergy(char) / getMaximum(char);
        },
        hasEnergy: hasEnergy,
        getEnergyTime: getEnergyTime,

        getMaximum: getMaximum,

        getSelected: getSelected,
        setSelected: setSelected,
        getEnergyData: getEnergyData,

        getAvailableCharacters: function () {
            return [].concat(CHARACTER_ORDER);
        },
        getUnlockedCharacters: function () {
            return SaveManager.load('unlockedCharacters');
        },
        getAvailable: function () {
            return CharacterManager.getAvailableCharacters();
        },
        getUnlocked: function () {
            return CharacterManager.getUnlockedCharacters();
        },
        getCharacterConstructor: function (name) {
            if (Utils.isDev() && !characterConstructors[name]) {
                Utils.log("Character " + name + " does not exist. Add in character CharacterManager and restart.");
            }
            return characterConstructors[name];
        },
        restoreAllEnergy: function () {
            // only used for cheating
            var characters = SaveManager.load('unlockedCharacters');
            Utils.forEach(characters, function (char, i, l, breakLoop) {
                var charData = SaveManager.load('energy-' + char);
                charData.energy = charData.maximum;
                SaveManager.save('energy-' + char, charData);
            });
        },
        /**
         * Amount of tokens needed to unlock
         */
        getUnlockCost: function (name) {
            return characterData[name].tokensToUnlock || 10;
        },
        /**
         * Amount of items needed to evolve
         */
        getEvolveItems: function (name) {
            var maxLevel = SaveManager.load('maxLevel-' + name);
            // maxLevel 10 -> 20 gems
            // maxLevel 20 -> 40 gems etc.
            // maxLevel 30 -> 60 gems etc.
            return Math.max(20, maxLevel * 2);
        },
        /**
         * Misc character data
         */
        getCharacterData: function (name) {
            return characterData[name];
        },
        /**
         * Level of a character
         */
        getLevel: function (name) {
            if (!name) {
                name = getSelected();
            }
            return SaveManager.load('level-' + name);
        },
        /**
         * Maxlevel of a character
         */
        getMaxLevel: function (name) {
            return SaveManager.load('maxLevel-' + name);
        },
        /**
         * Max maxlevel of a character
         */
        getMaxTotal: function (name) {
            return characterData[name].maxTotal;
        },
        /**
         * Amount of gems needed to evolve
         */
        getEvolveCost: function (name) {
            var maxLevel = SaveManager.load('maxLevel-' + name);
            // maxLevel 10 -> 200 gems
            // maxLevel 20 -> 200 gems etc.
            // maxLevel 30 -> 300 gems etc.
            return Math.max(200, maxLevel * 10);
        },
        init: function () {
            // setup minimum levels
            Utils.forEach(minimumLevels, function (level, char) {
                var currentLevel = SaveManager.load('level-' + char);
                if (!isUnlocked(char)) {
                    SaveManager.save('level-' + char, level);
                }
            });
            // setup minimum maxLevel
            Utils.forEach(CHARACTER_ORDER, function (char) {
                var currentLevel = SaveManager.load('level-' + char);
                var currentMaxLevel = SaveManager.load('maxLevel-' + char);
                if (currentMaxLevel < characterData[char].maxLevel) {
                    currentMaxLevel = characterData[char].maxLevel;
                    SaveManager.save('maxLevel-' + char, currentMaxLevel);
                }
                // backwards compatibility: if user is already over the max, we match the max level to that level
                if (currentLevel > currentMaxLevel) {
                    currentMaxLevel = Math.floor(currentLevel / 10) * 10;
                    SaveManager.save('maxLevel-' + char, currentMaxLevel);
                }
            });
        }
    };

    if (Utils.isDev()) {
        window.characterManager = CharacterManager;
    }

    return CharacterManager;
});