/**
 * Manages the amount of coins collected during a stage by all players (local and remote).
 * If a local player collects a coin, it's added to the total.
 * Remote players do not add to the total, instead we wait for a message.
 * All coins should have a unique, deterministic name. They can be collected only once.
 * @moduleName CoinCollector
 */
bento.define('managers/coincollector', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    MatchManager
) {
    'use strict';
    // private
    var collected = {};
    var amount = 0;
    var collect = function (coinId, value) {
        var hasCollectedBefore;
        // update coins amount if not collected yet
        if (!collected[coinId]) {
            amount += value;
            hasCollectedBefore = false;
        } else {
            hasCollectedBefore = true;
        }

        // save to collected
        collected[coinId] = true;

        return hasCollectedBefore;
    };

    // public
    var CoinCollector = {
        name: 'coinCollector',
        /**
         * Prepare for new stage
         */
        reset: function () {
            // reset the collected object
            collected = {};
            amount = 0;
        },
        /**
         * Returns amount of coins collected
         */
        getCollected: function () {
            return amount;
        },
        /**
         * Adds a coin to collected
         * Should be called by local player only!
         */
        collect: function (coinId, value) {
            var hasCollected = false;
            if (!value) {
                value = 1;
            }
            hasCollected = collect(coinId, value);

            // send message to remotes
            MatchManager.sendData({
                id: coinId,
                v: value
            }, 'coinCollector');

            return hasCollected;
        }
    };

    // listen to incoming coins
    var onMessage = function (msg) {
        var coinId = msg.id;
        var value = msg.v || 1;
        collect(coinId, value);
        return true;
    };

    // initialize listener
    // NOTE: goes wrong after quick reload because this listener keeps existing
    MatchManager.onMessage('coinCollector', onMessage);

    // debug
    if (Bento.isDev()) {
        window.showCoinsCollected = function () {
            console.log(CoinCollector.getCollected());
        };
    }

    return CoinCollector;
});