/**
 * The enemy activator switches the EnemyController.activate to true when a player comes close (when they come into view)
 * This only applies to the local camera, but it will send/receive messages also so that remote
 * players can activate enemies as well.
 * Only activates enemies left of the active blockade.
 * This manager must be attached to the game at the start of screens/main
 * @moduleName EnemyActivator
 */
bento.define('managers/enemyactivator', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'entities/triggers/blockade',
    'modules/matchmanager',
    'managers/ai'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Blockade,
    MatchManager,
    AiManager
) {
    'use strict';
    var EnemyActivator = function () {
        var viewport = Bento.getViewport();
        var enemies = Bento.getByFamily('enemy');
        var cutscene = false;
        var onMessage = function (msg) {
            var activeBlockade = Blockade.active;
            if (!activeBlockade) {
                // we remove the activator here because there is no use for it anymore
                // when there is no activeblockade, it will continue to check for enemies to activate
                Bento.objects.remove(activator);
            }
            EnemyActivator.activateEnemies();
            return true;
        };

        var disableControls = function () {
            cutscene = true;
            deactivateEnemies();
        };

        var enableControls = function () {
            cutscene = false;
        };

        var deactivateEnemies = function () {
            // search for active enemies and deactivate
            Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                enemy.getComponent('enemyController', function (enemyController) {
                    if (enemyController.active) {
                        enemyController.active = false;
                    }
                    enemy.processAction({
                        type: 'idle'
                    });
                });
            });
            var activeBlockade = Blockade.active;
            if (activeBlockade) {
                activeBlockade.hasActivated = false;
            }
        };

        var activator = {
            name: 'enemyActivator',
            timer: 0,
            start: function (data) {
                // listen to incoming messages
                MatchManager.onMessage('activateEnemies', onMessage);
                // turn on cutscene events
                EventSystem.on('startCutscene', disableControls);
                EventSystem.on('endCutscene', enableControls);
            },
            destroy: function (data) {
                // stop listening
                MatchManager.offMessage('activateEnemies', onMessage);
                // remove cutscene events
                EventSystem.off('startCutscene', disableControls);
                EventSystem.off('endCutscene', enableControls);
            },
            update: function (data) {
                // only checks every 30 frames to save some cpu power
                activator.timer += 1;
                if (activator.timer % 30) {
                    return;
                }

                // deactivate if cutscene is active
                if (cutscene) {
                    return;
                }

                // check if any enemy comes into the camera
                // any enemy will activate the rest
                Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
                    if (viewport.hasPosition(enemy.position)) {
                        EnemyActivator.activateEnemies();
                        breakLoop();
                    }
                });
            }
        };
        return activator;
    };

    /**
     * Activates all enemies left of the active blocked
     */
    EnemyActivator.activateEnemies = function () {
        var activeBlockade = Blockade.active;
        var enemies = Bento.getByFamily('enemy');

        if (activeBlockade) {
            if (activeBlockade.hasActivated) {
                // already activated
                return;
            } else {
                activeBlockade.hasActivated = true;
            }
        } else {
            // hacky fix to make sure that levels with no blockades (bosslevel) don't keep on distributing AI tasks after activating
            if (Blockade.done === 1) {
                Blockade.done = 2;
            } else {
                return;
            }
        }

        // search enemies left of blockade
        Utils.forEach(enemies, function (enemy, i, l, breakLoop) {
            if (activeBlockade && enemy.position.x > activeBlockade.position.x) {
                // note that if no blockade exists, all enemies will activate
                return;
            }
            enemy.getComponent('enemyController', function (enemyController) {
                if (!enemyController.active) {
                    enemyController.active = true;
                }
            });
        });

        EventSystem.fire('activateEnemies');

        AiManager.distributeAiTasks();

        // notify remotes
        MatchManager.sendData({}, 'activateEnemies');
    };

    return EnemyActivator;
});