/**
 * Inventory Manager
 *
 * Keeps track of all the items in the players inventory, as well as the 'equipped' items.
 * Items are saved as an object containing a key and an amount.
 * 		-The key is a string constructed as "[category]/[name]" e.g. "weapon/pipe" or "consumable/apple".
 * Items that have an amount of 0 should not be saved.
 * Items are sorted by category and by conditions within that category
 * Keeps track of upgrade tokens
 *
 * @moduleName InventoryManager
 */
bento.define('managers/inventory', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'managers/save',
    'modules/firebase'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    SaveManager,
    FireBase
) {
    'use strict';
    //items in the inventory
    var inventory = [];

    /**
     * Save the current inventory
     */
    var save = function () {
        sanitize();
        SaveManager.save('inventory', inventory);
    };

    /**
     * Load the inventory
     */
    var load = function () {
        inventory = SaveManager.load('inventory');
    };

    /**
     * Finds the index of an item with a certain key
     * @param  {String} key 	The key of the item to find
     * @return {Int}     		The index of the item, or -1 if it does not exist
     */
    var find = function (key) {
        //TODO : keep track of where each category starts and ends, to narrow down the searching area
        var ii;
        for (ii = 0; ii < inventory.length; ii++) {
            if (inventory[ii].key === key) {
                return ii;
            }
        }

        return -1;
    };

    /**
     * Finds the index of an item, or creates an entry for it if it's not in the inventroy yet
     * * @param {String} key    	The key of the item to find
     */
    var findOrCreate = function (key) {
        //TODO : keep track of where each category starts and ends, to insert the item at an appropriate place instantly
        var index = find(key);

        if (index !== -1) {
            return index;
        }

        inventory.push({
            key: key,
            amount: 0
        });

        return inventory.length - 1;
    };

    /**
     * Add a certain value to the amount of items with the given key
     * @param {String} key    	The key of the item to add to
     * @param {Int} amount 		The amount to add
     */
    var addItem = function (key, amount) {
        setItem(key, getItem(key) + amount);

        // if we just got a scroll
        if (key.indexOf('scrolls/') !== -1) {
            //check if we have all of the scrolls
            if (
                getItem('scrolls/super') &&
                getItem('scrolls/dash') && 
                getItem('scrolls/charge') &&
                getItem('scrolls/jump')
            ) {
                // We just got the last scroll so post the 'Kung Fu Master' Achievement
                Utils.postAchievement("CgkI9NqdkoAGEAIQCw");
            }
        }
    };

    /**
     * Set the amount of items with a certain key
     * @param {String} key    	The key of the item to change the amount of
     * @param {Int} amount 		The new amount of items
     */
    var setItem = function (key, amount) {
        if (isNaN(amount) || Utils.isUndefined(amount) || amount === null) {
            Utils.log('WARNING: amount is not a valid value ' + amount);
            return;
        }
        var index = findOrCreate(key);
        var oldAmount = inventory[index].amount;
        inventory[index].amount = amount;

        // special case: getting a new equipment lits up the equipment button
        var category = key.split('/')[0];
        var inEquipment = false;
        if (category === 'equipment' && !oldAmount) {
            // should be in nobody's equipment
            Utils.forEach(SaveManager.load('unlockedCharacters'), function (character) {
                var equipment = SaveManager.load('equipped-' + character);
                Utils.forEach(equipment, function (item) {
                    if (item === key) {
                        inEquipment = true;
                    }
                });
            });
            if (!inEquipment) {
                SaveManager.save('hasNewEquipment', true);
            }
        }

        save();
    };

    /**
     * Get the value of an item with a certain key.
     * @param  {String} key 	The key of the item to get the amount of
     * @return {Int}     		The amount the player owns of this item, defaults to 0
     */
    var getItem = function (key) {
        var index = find(key);

        if (index === -1) {
            return 0;
        } else {
            return inventory[index].amount || 0;
        }
    };

    /**
     *  get all items the player has in their inventory
     * a filter can be applied
     * @param  {String or [String]} filter      a string or array of strings. Only items from these categories will be returned
     * @return {Array}                          An array of items
     */
    var getAllItems = function (filter) {
        sanitize();
        sortItems();
        var result = [].concat(inventory);
        // 2nd pass sanitize: hide items in that appear in the hidden array
        sanitize(result);

        if (!filter || !filter[0]) {
            return result;
        }

        filter = [].concat(filter);
        var ii;
        for (ii = result.length - 1; ii >= 0; ii--) {
            var key = inventory[ii].key;
            var category = key.split('/')[0];
            if (filter.indexOf(category) === -1) {
                result.splice(ii, 1);
            }
        }

        return result;
    };

    /**
     * get the equipped items for the current character
     */
    var getEquipped = function (char) {
        var equipped = SaveManager.load('equipped-' + char);
        var parsed = [];
        var ii;
        var count = 2; //TODO? hardcoded amount of equipable items
        for (ii = 0; ii < count; ii++) {
            parsed[ii] = undefined;
            if (equipped[ii]) {
                parsed[ii] = {
                    key: equipped[ii],
                    amount: 1
                };
            }
        }
        return parsed;
    };
    /**
     * get equipped current consumables
     */
    var getEquippedConsumables = function () {
        var equipped = SaveManager.load('equippedConsumables');
        var parsed = [];
        var ii;
        var count = 2; //TODO? hardcoded amount of equipable items
        for (ii = 0; ii < count; ii++) {
            parsed[ii] = undefined;
            if (equipped[ii]) {
                parsed[ii] = {
                    key: equipped[ii],
                    amount: 1
                };
            }
        }
        return parsed;
    };


    /**
     * Set the equiped item in the provided slot
     * @param {Int} index [description]
     * @param {String} key   [description]
     */
    var setEquipped = function (index, key, char) {
        var equipped = SaveManager.load('equipped-' + char);
        equipped[index] = key;
        SaveManager.save('equipped-' + char, equipped);
    };
    var setEquippedConsumable = function (index, key) {
        var equipped = SaveManager.load('equippedConsumables');
        equipped[index] = key;
        SaveManager.save('equippedConsumables', equipped);
    };
    /**
     * Sort all items, based on category and parameters within the category
     */
    var sortItems = function () {
        //TODO
    };

    /**
     * Sanitize item array, removing all items that have an 'amount' of 0
     * @param {Array} reference to the array to edit
     */
    var sanitize = function (array) {
        var ii;
        var key;
        var amount;

        if (!array) {
            array = inventory;
        }
        for (ii = array.length - 1; ii >= 0; ii--) {
            amount = array[ii].amount;
            key = array[ii].key;

            // don't hide from the original inventory
            if (array !== inventory) {
                // subtract the hidden amount
                amount -= countHidden(key);
            }

            if (amount <= 0) {
                array.splice(ii, 1);
            } else {
                //remove any unnecessary data
                array[ii] = {
                    key: key,
                    amount: amount
                };
            }
        }
    };

    var countHidden = function (key) {
        var count = 0;
        Utils.forEach(manager.hidden, function (item, i, l, breakLoop) {
            if (item === key) {
                count += 1;
            }
        });
        return count;
    };

    var manager = {
        save: save,
        load: load,
        addItem: addItem,
        setItem: setItem,
        getItem: getItem,
        getAllItems: getAllItems,
        setEquipped: setEquipped,
        getEquipped: getEquipped,
        getEquippedConsumables: getEquippedConsumables,
        setEquippedConsumable: setEquippedConsumable,

        // tokens
        addTokens: function (key, amount) {
            addItem(key, amount);
        },
        getTokens: function (key) {
            return getItem(key);
        },
        spendTokens: function (key, amount) {
            var tokens = getItem(key);
            if (amount > tokens) {
                Utils.log("WARNING: player is trying to spend more than amount of tokens");
            }
            addItem(key, -amount);
        },
        canSpendTokens: function (key, amount) {
            var tokens = getItem(key);
            return tokens >= amount; 
        },

        /*
         * Array of temporarily hidden items (Array of String keys)
         */
        hidden: []
    };

    // for dev only: reload the items because refresh will clear it
    if (Bento.isDev()) {
        load();
    }

    return manager;
});