/**
 * Resolution manager
 * @moduleName ResolutionManager
 */
bento.define('managers/resolution', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'bento/autoresize',
    'modules/matchmanager',
    'components/eventlistener',
    'bento/components/sprite'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    AutoResize,
    MatchManager,
    EventListener,
    Sprite
) {
    'use strict';
    var pixelSize = 3;
    var portraitDimensions = {
        minHeight: 320,
        maxHeight: 320,
        minWidth: 180,
        maxWidth: 240
    };
    var landscapeDimensions = {
        minHeight: 180,
        maxHeight: 240,
        minWidth: 320,
        maxWidth: 320
    };
    var orientation = 'unknown';
    var userLock = Bento.saveState.load('orientationLock', false);
    var systemLocks = {};
    var isSystemLocked = function () {
        var isLocked = false;
        Utils.forEach(systemLocks, function (value, name, l, breakLoop) {
            if (value) {
                isLocked = true;
                breakLoop();
            }
        });
        return isLocked;
    };
    var playerDimensions = {};
    /**
     * Pushes the view down so it's centered
     * Draws borders
     */
    var viewOffset = new Entity({
        z: 1000000,
        name: 'viewOffset',
        global: true,
        float: true,
        position: new Vector2(0, 0),
        components: [
            new EventListener({
                name: 'preDrawLoop',
                onEvent: function (data) {
                    data.renderer.save();
                    data.renderer.translate(viewOffset.position.x, viewOffset.position.y);
                }
            }),
            new EventListener({
                name: 'postDrawLoop',
                onEvent: function (data) {
                    data.renderer.restore();
                }
            }),
            new EventListener({
                name: 'assetsLoaded',
                onEvent: function () {
                    var viewport = Bento.getViewport();
                    var topBorder = new Sprite({
                        imageName: 'gui/border',
                        originRelative: new Vector2(0.5, 0)
                    });
                    var topLeft = new Entity({
                        name: 'topLeft',
                        rotation: orientation === 'landscape' ? Math.PI / 2 : Math.PI,
                        position: orientation === 'landscape' ?
                            new Vector2(-viewOffset.position.x, viewport.height / 2) : new Vector2(viewport.width / 2, -viewOffset.position.y),
                        components: [
                            topBorder,
                            new EventListener({
                                name: 'orientationChanged',
                                onEvent: function (data) {
                                    if (data.orientation === 'portrait') {
                                        topLeft.rotation = Math.PI;
                                        topLeft.position.x = viewport.width / 2;
                                        topLeft.position.y = -viewOffset.position.y;
                                    } else {
                                        topLeft.rotation = Math.PI / 2;
                                        topLeft.position.x = -viewOffset.position.x;
                                        topLeft.position.y = viewport.height / 2;
                                    }
                                }
                            })
                        ]
                    });
                    var bottomBorder = new Sprite({
                        imageName: 'gui/border',
                        originRelative: new Vector2(0.5, 0)
                    });
                    var bottomRight = new Entity({
                        name: 'bottomRight',
                        rotation: orientation === 'landscape' ? -Math.PI / 2 : 0,
                        position: orientation === 'landscape' ?
                            new Vector2(viewport.width - viewOffset.position.x, viewport.height / 2) : new Vector2(viewport.width / 2, viewport.height - viewOffset.position.y),
                        components: [
                            bottomBorder,
                            new EventListener({
                                name: 'orientationChanged',
                                onEvent: function (data) {
                                    if (data.orientation === 'portrait') {
                                        bottomRight.rotation = 0;
                                        bottomRight.position.x = viewport.width / 2;
                                        bottomRight.position.y = viewport.height - viewOffset.position.y;
                                    } else {
                                        bottomRight.rotation = -Math.PI / 2;
                                        bottomRight.position.x = viewport.width - viewOffset.position.x;
                                        bottomRight.position.y = viewport.height / 2;
                                    }
                                }
                            })
                        ]
                    });
                    viewOffset.attach(topLeft);
                    viewOffset.attach(bottomRight);

                    // make borders optional
                    viewOffset.extend({
                        showBorders: function (bool) {
                            topBorder.visible = bool;
                            bottomBorder.visible = bool;
                        }
                    });
                    // start invisible
                    viewOffset.showBorders(false);
                }
            })
        ]
    });
    var isWideScreen = false;
    var wideScreenSize = new Rectangle(0, 0, 0, 0);

    var createPortraitResize = function (dimensions) {
        return new AutoResize(
            new Rectangle(0, 0, dimensions.minWidth, dimensions.maxHeight), // base size
            dimensions.minHeight, // minimum height
            dimensions.maxHeight // maximum height
        );
    };

    var createLandscapeResize = function (dimensions) {
        return new AutoResize(
            new Rectangle(0, 0, dimensions.maxWidth, dimensions.minHeight), // base size
            dimensions.minWidth, // minimum width
            dimensions.maxWidth, // maximum width
            true //is landscape
        );
    };

    var onResize = function () {
        var viewport = Bento.getViewport();
        var renderer = Bento.getRenderer();
        var canvas = document.getElementById('canvas');
        var context;
        var wentOverMinimum = false;
        var screenRatio = 1;
        var cutOff = 0;

        var ratio = window.innerWidth / window.innerHeight;
        var newOrientation = (ratio < 1) ? 'portrait' : 'landscape';

        var canvasDimension;

        if (!canvas) {
            return;
        }
        context = renderer.getContext();

        if (newOrientation === 'portrait') {
            canvasDimension = createPortraitResize(portraitDimensions);
            // max/min width
            if (canvasDimension.width > portraitDimensions.maxWidth) {
                canvasDimension.width = portraitDimensions.maxWidth;
            }
            if (Math.ceil(canvasDimension.width) < portraitDimensions.minWidth) {
                canvasDimension.width = portraitDimensions.minWidth;
                wentOverMinimum = true;
                isWideScreen = true;
            } else {
                isWideScreen = false;
            }

            canvas.width = canvasDimension.width * pixelSize;
            canvas.height = canvasDimension.height * pixelSize;
            viewport.width = canvasDimension.width;
            viewport.height = canvasDimension.height;
            wideScreenSize = viewport.clone();

            // pad extra pixels (?)
            if (wentOverMinimum) {
                screenRatio = window.innerWidth / window.innerHeight;
                // append to height
                canvasDimension.height = Math.round(canvasDimension.width / screenRatio);
                canvas.height = canvasDimension.height * pixelSize;
                viewport.height = canvasDimension.height;
                wideScreenSize.height = canvasDimension.height;
            }

            // fit to height
            canvas.style.height = window.innerHeight + 'px';
            canvas.style.width = (viewport.width / viewport.height * window.innerHeight) + 'px';
        } else {
            canvasDimension = createLandscapeResize(landscapeDimensions);
            // max/min height
            if (canvasDimension.height > landscapeDimensions.maxHeight) {
                canvasDimension.height = landscapeDimensions.maxHeight;
            }
            if (Math.ceil(canvasDimension.height) < landscapeDimensions.minHeight) {
                canvasDimension.height = landscapeDimensions.minHeight;
                wentOverMinimum = true;
                isWideScreen = true;
            } else {
                isWideScreen = false;
            }

            canvas.width = canvasDimension.width * pixelSize;
            canvas.height = canvasDimension.height * pixelSize;
            viewport.width = canvasDimension.width;
            viewport.height = canvasDimension.height;
            wideScreenSize = viewport.clone();

            // iphoneX: pad extra pixels
            if (wentOverMinimum) {
                screenRatio = window.innerWidth / window.innerHeight;
                // append to height
                canvasDimension.width = Math.round(canvasDimension.height * screenRatio);
                canvas.width = canvasDimension.width * pixelSize;
                viewport.width = canvasDimension.width;
                wideScreenSize.width = canvasDimension.width;
            }

            // fit to width
            canvas.style.width = window.innerWidth + 'px';
            canvas.style.height = (viewport.height / viewport.width * window.innerWidth) + 'px';
        }


        // Bento input updates its viewport before this event, so call this manually
        if (Bento.input) {
            Bento.input.updateCanvas();
        }

        // cut off the extra pixels so the viewport becomes 2:1
        if (newOrientation === 'portrait') {
            if (viewport.height > viewport.width * 2) {
                cutOff = viewport.height - viewport.width * 2;
                // cut off bottom
                viewport.height -= cutOff;
                // offset view
                viewOffset.position.x = 0;
                viewOffset.position.y = cutOff / 2;
                Bento.input.setOffset(new Vector2(0, -cutOff / 2));
            } else {
                // revert
                viewOffset.position.x = 0;
                viewOffset.position.y = 0;
                Bento.input.setOffset(new Vector2(0, 0));
            }
        } else {
            if (viewport.width > viewport.height * 2) {
                cutOff = viewport.width - viewport.height * 2;
                // cut off right
                viewport.width -= cutOff;
                // offset view
                viewOffset.position.x = cutOff / 2;
                viewOffset.position.y = 0;
                Bento.input.setOffset(new Vector2(-cutOff / 2, 0));
            } else {
                // revert
                viewOffset.position.x = 0;
                viewOffset.position.y = 0;
                Bento.input.setOffset(new Vector2(0, 0));
            }
        }

        if (newOrientation !== orientation) {
            orientation = newOrientation;
            EventSystem.fire('orientationChanged', {
                orientation: orientation,
                resolution: viewport.clone()
            });
        }

        EventSystem.fire('resolutionChanged', {
            orientation: orientation,
            resolution: viewport.clone()
        });

        if (MatchManager.isOnline) {
            MatchManager.sendData({
                type: 'resolutionChange',
                data: {
                    w: viewport.width,
                    h: viewport.height
                }
            }, 'general');
            resolutionManager.setPlayerDimension(MatchManager.localPlayer.playerId, viewport.width, viewport.height);
        }

        // prevent the canvas being blurry after resizing
        if (context.imageSmoothingEnabled) {
            context.imageSmoothingEnabled = false;
        }
        if (context.webkitImageSmoothingEnabled) {
            context.webkitImageSmoothingEnabled = false;
        }
        if (context.mozImageSmoothingEnabled) {
            context.mozImageSmoothingEnabled = false;
        }
        if (context.msImageSmoothingEnabled) {
            context.msImageSmoothingEnabled = false;
        }

        resolutionManager.canvasDimension = canvasDimension;
    };


    var resolutionManager = {
        onResize: onResize,
        init: function () {
            window.addEventListener('resize', onResize, false);
            window.addEventListener('orientationchange', onResize, false);

            onResize();

            resolutionManager.setUserLock(userLock);

            Bento.objects.attach(viewOffset);
        },
        pixelSize: pixelSize,
        canvasDimension: createPortraitResize(portraitDimensions),
        lock: function (dimension) {
            if (Utils.isCocoonJs()) {
                if (window.screen.orientation) {
                    if (Utils.isUndefined(dimension)) {
                        // lock to whatever is the current 
                        dimension = orientation;
                    }
                    window.screen.orientation.lock(dimension);
                } else if (window.screen.lockOrientation) {
                    if (Utils.isUndefined(dimension)) {
                        // lock to whatever is the current 
                        dimension = window.orientation;
                    }
                    window.screen.lockOrientation(dimension);
                }
                onResize();
            }
        },
        unlock: function () {
            if (Utils.isCocoonJs()) {
                if (window.screen.unlockOrientation) {
                    window.screen.unlockOrientation();
                } else if (window.screen.orientation) {
                    window.screen.orientation.unlock();
                }
                onResize();
            }
        },
        getOrientation: function () {
            return orientation;
        },
        isPortrait: function () {
            return orientation === 'portrait';
        },
        getMaxDimensions: function () {
            var width = 0;
            var height = 0;

            var ii;
            var dims;

            for (ii = 0; ii < 2; ii++) {
                dims = [portraitDimensions, landscapeDimensions][ii];

                width = Math.max(width, dims.maxWidth);
                height = Math.max(height, dims.maxHeight);
            }

            return {
                width: width,
                height: height
            };
        },
        setPlayerDimension: function (id, width, height) {
            playerDimensions[id] = {
                width: width,
                height: height
            };
        },
        getPlayerDimension: function (id) {
            if (playerDimensions.hasOwnProperty(id)) {
                return playerDimensions[id];
            } else {
                //assume the 'worst'
                return resolutionManager.getMaxDimensions();
            }
        },
        // lock and unlock by user
        setUserLock: function (value) {
            userLock = value;
            Bento.saveState.save('orientationLock', value);

            if (userLock || isSystemLocked()) {
                resolutionManager.lock();
            } else if (!value) {
                resolutionManager.unlock();
            }
        },
        // use this to lock orientation during tweens and such
        setSystemLock: function (value, name) {
            if (!name) {
                Utils.log('ERROR: you must identify a system lock with a name');
                return;
            }
            systemLocks[name] = value;

            if (userLock || isSystemLocked()) {
                resolutionManager.lock();
            } else if (!value) {
                resolutionManager.unlock();
            }
        },
        isWideScreen: function () {
            return isWideScreen;
        },
        getOffset: function () {
            return viewOffset.position;
        },
        getWideScreenSize: function () {
            return wideScreenSize;
        }
    };

    return resolutionManager;
});