/**
 * Reward manager: randomizes rewards once per day
 * @moduleName RewardManager
 */
bento.define('managers/reward', [
    'bento',
    'bento/eventsystem',
    'bento/utils',
    'managers/time',
    'modules/randomizer',
    'managers/save',
    'managers/character',
    'managers/allies'
], function (
    Bento,
    EventSystem,
    Utils,
    TimeManager,
    Randomizer,
    SaveManager,
    CharacterManager,
    AlliesManager
) {
    'use strict';
    var cachedDay = -1;
    var rewards = {};
    var missionsJson;
    var randomizer;
    var randomizeRewards = function (day) {
        // reset
        rewards = {};

        // create new randomized based on day
        cachedDay = day;
        randomizer = new Randomizer({
            seed: day
        });
        // we need the missions json 
        if (!missionsJson) {
            missionsJson = Bento.assets.getJson('missions/missions');
        }

        // (do we randomize all missions or only the unlocked ones?)
        Utils.forEach(missionsJson, function (missionData, mission) {
            var reward = getRandomizedReward();

            rewards[mission] = {
                reward: reward.string,
                amount: 1
            };
        });

        // spread out the tokens/evolve items you need
        var lockedCharacters = CharacterManager.getAvailableCharacters();
        var unlockedCharacters = CharacterManager.getUnlockedCharacters() || [];
        var j;
        var ii;
        for (j = unlockedCharacters.length - 1; j >= 0; j--) {
            ii = lockedCharacters.indexOf(unlockedCharacters[j]);
            if (ii !== -1) {
                lockedCharacters.splice(ii, 1);
            }
        }
        var availableAllies = AlliesManager.getAvailable();
        var unlockedAllies = SaveManager.load('unlockedAllies');
        var unlockedMissions = SaveManager.load('unlockedLevels');
        // make an array of tokens
        var tokens = [];
        // 3 player tokens of each unlocked
        Utils.forEach(unlockedCharacters, function (character) {
            tokens.push('playertokens/' + character);
            tokens.push('playertokens/' + character);
            tokens.push('playertokens/' + character);
        });
        // 1-3 evolve item of each unlocked
        Utils.forEach(unlockedCharacters, function (character) {
            tokens.push('evolve/' + character);
            // 50% of more evolve items
            if (randomizer.getRandom(2)) {
                tokens.push('evolve/' + character);
            }
            if (randomizer.getRandom(2)) {
                tokens.push('evolve/' + character);
            }
        });
        // 1 enemy token of each unlocked
        Utils.forEach(unlockedAllies, function (character) {
            tokens.push('enemytokens/' + character);
        });
        // 1 token of locked characters
        if (lockedCharacters.length) {
            // only the 1st in the list
            tokens.push('playertokens/' + lockedCharacters[0]);
        }
        // 1 locked enemy token
        Utils.forEach(availableAllies, function (ally, i, l, breakLoop) {
            // only the one next in line
            if (unlockedAllies.indexOf(ally) >= 0) {
                // already have this one
                return;
            } else {
                tokens.push('enemytokens/' + ally);
                breakLoop();
            }
        });

        // shuffle the list of missions and clone it
        unlockedMissions = shuffle(unlockedMissions, true);

        // more rewards than unlocked missions: 
        // add missions so when user unlocks them, they will get meaningful new rewards
        while (tokens.length > unlockedMissions.length) {
            unlockedMissions.push(unlockedMissions.length + 1);
        }

        // shuffle again
        unlockedMissions = shuffle(unlockedMissions, true);

        // overwrite the rewards
        Utils.forEach(unlockedMissions, function (mission, i, l, breakLoop) {
            // no more tokens
            if (tokens.length === 0) {
                breakLoop();
                return;
            }
            if (mission === 'dojo') {
                return;
            }
            var token = tokens.splice(randomizer.getRandom(tokens.length), 1);
            rewards[mission] = {
                reward: token[0],
                amount: 1
            };
        });
        // uncomment this to check what rewards are placed on missions
        // Utils.forEach(unlockedMissions, function (mission, i, l, breakLoop) {
        //     SaveManager.save('rewards-' + mission, [true, true, true]);
        // });

    };
    /*
     * Copy of Utils.shuffle, but need randomizer here
     */
    var shuffle = function (arr, clone) {
        clone = clone || false;

        var ii;
        var idx;
        var temp;

        if (clone) {
            arr = [].concat(arr);
        }

        for (ii = arr.length - 1; ii >= 0; ii--) {
            idx = Math.floor(randomizer.getRandomFloat(1) * (ii + 1));
            temp = arr[ii];
            arr[ii] = arr[idx];
            arr[idx] = temp;
        }

        return arr;
    };
    /*
     * Chance x out of y
     */
    var chanceOutOf = function (value, max) {
        var chance = randomizer.getRandomFloat(max);
        return chance > value;
    };
    /*
     * Pick a random item from an array
     */
    var pickRandomArray = function (array) {
        return array[randomizer.getRandom(array.length)];
    };
    /*
     * Pick a random item, weighted by its tier
     */
    var pickRandomItem = function (array) {
        var weightedItems = [];
        var winner;
        Utils.forEach(array, function (item, i, l, breakLoop) {
            var weight = 1;
            if (item.tier) {
                if (item.tier === 2) {
                    weight = 0.25;
                } else if (item.tier > 2) {
                    weight = 0.12;
                }
            }
            weightedItems.push({
                key: item.key,
                item: item.item,
                weight: weight
            });
        });
        winner = weightedRandom(weightedItems);
        return winner;
    };

    /*
     * Weighted randomization
     * array = [{value: Something, weight: Number }, ...]
     */
    var weightedRandom = function (array) {
        // calculate sum of weights
        var weightTotal = 0;
        Utils.forEach(array, function (item, i, l, breakLoop) {
            weightTotal += item.weight || 0;
        });
        // randomize a number
        var random = randomizer.getRandomRangeFloat(0, weightTotal);
        // console.log(weightTotal, random);
        // find the winner
        var winner;
        var weightIncr = 0;
        Utils.forEach(array, function (item, i, l, breakLoop) {
            if (item.weight <= 0) {
                return;
            }
            weightIncr += item.weight;
            if (weightIncr > random) {
                winner = item;
                breakLoop();
                return;
            }
        });
        // console.log(winner);
        return winner;
    };
    /*
     * Returns a single randomized reward
     */
    var getRandomizedReward = function () {
        var category = pickRandomArray(categories);
        var items = possibleItems[category];
        var item = pickRandomItem(items).key;

        return {
            category: category,
            string: item
        };
    };
    /*
     *
     */
    var possibleItems = {};
    var categories = [];
    var initialized = false;
    var initPossibleItems = function () {
        // this function is run only once
        if (initialized) {
            return;
        }
        var assets = Bento.assets.getAssets();
        var jsonAssets = assets.json;
        var cachedCategories = {};

        Utils.forEach(jsonAssets, function (item, key, l, breakLoop) {
            // weapons count as consumables too
            var keys = key.split('/');
            var category = keys[0];
            var array = category;
            if (
                // only weapons and consumables
                category === 'weapons' ||
                category === 'consumables'
            ) {
                possibleItems[array] = possibleItems[array] || [];
                possibleItems[array].push({
                    key: key,
                    item: item,
                    category: category,
                    tier: (item.inventory || {}).tier || 0
                });

                cachedCategories[category] = category;
            }

            // save the categories in an array
            Utils.forEach(cachedCategories, function (cat) {
                categories.push(cat);
            });
        });
        initialized = true;
    };

    return {
        /**
         * Returns reward of a certain mission, an object with reward string + amount
         */
        getReward: function (mission) {
            // DEPRECATED: rewardmanager will now simply give reward1 from missionsjson
            var now = TimeManager.getTime();
            var today = Math.floor(now / (1000 * 60 * 60 * 24));
            var reward;

            if (!mission) {
                Utils.log('ERROR: Pass mission as parameter');
            }

            // init if needed
            initPossibleItems();

            // refresh rewards if needed
            if (today !== cachedDay) {
                randomizeRewards(today);
            }

            reward = rewards[mission];

            // in the event this reward does not exist -> randomize single reward(?)
            if (!reward) {
                reward = getRandomizedReward();
                rewards[mission] = reward;
            }

            return reward;
            /*if (!missionsJson) {
                missionsJson = Bento.assets.getJson('missions/missions');
            }
            var missionRewards = missionsJson[mission].rewards;
            return {
                reward: missionRewards.reward1,
                amount: missionRewards.amount1
            };*/
        },
        /**
         * Get energy needed for mission
         * @param  {Number} mission Mission to check
         * @return {Number}         Amount of energy
         */
        getEnergyCost: function (mission) {
            return missionsJson[mission].cost;
        },
        /**
         * Pick a random item from the consumable (category must be consumables or weapons)
         */
        getRandomConsumable: function (category) {
            var items;
            var item;
            // init if needed
            initPossibleItems();

            // consumable or weapon
            items = possibleItems[category || 'consumables'];
            item = items[Utils.getRandom(items.length)];

            return item;
        }
    };
});