/**
 * Save Manager.
 * Wrapper for Bento's SaveState.
 * New savekeys need to be defined here, paired with a default value. When using the SaveManager, it will check wether the provided key is defined, and throw a warning otherwise.
 * It will also use the paired default value, ensuring consistency between all places in the codebase where a value is loaded.
 *
 * @moduleName SaveManager
 * @snippet SaveManager.load
SaveManager.load('$1');
 * @snippet SaveManager.save
SaveManager.save('$1', ${2:value});
 */
bento.define('managers/save', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/cloudsave'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    CloudSave
) {
    'use strict';
    /**
     * Save keys
     */
    var Keys = {};
    // Strings - Last version played
    Keys.lastVersionPlayed = '';
    // Array of Strings
    Keys.inventory = [];
    // Number - WARNING: coins is managed by CurrencyManager
    Keys.coins = 0;
    // Number - WARNING: gems is managed by CurrencyManager
    Keys.gems = 0;
    // Number - WARNING: keys is managed by CurrencyManager
    Keys.keys = 0;
    // Position in the worldmap, use Utils.toVector2
    Keys.worldmapPosition = { // DEPRECATED
        x: 0,
        y: 0,
        portrait: true
    };
    Keys.worldmapPositionVertical = {
        x: 76,
        y: 192,
        portrait: true
    };
    // Boolean - place consumables and super attack button on the other side
    Keys.leftHanded = false;
    // String - whether the player is in a test group
    Keys.testGroup = '';
    // Number - sensitivity factor for swiping, lower is more sensitive
    Keys.sensitivityFactor = 1;
    // Number - sensitivity factor for flicking, higher is more sensitive
    Keys.flickSensitivity = 1;
    // Boolean - did player revive before?
    Keys.hasRevived = false;
    // Boolean - did player call backup?
    Keys.calledBackup = false;
    // Number/String - current mission
    Keys.currentMission = 1;
    // Number/String - current mission cost
    Keys.currentCost = 0;
    // Array of Strings/Numbers - These missions exempt from energy because player watched an ad
    Keys.missionsExemptFromEnergy = [];
    // Number - current section of the mission
    Keys.currentSection = 1;
    // Boolean - if the section has a boss or not
    Keys.hasBoss = false;
    // Number - difficulty of enemies in current mission, used for hesitation
    Keys.difficulty = 0;
    // Number - difficulty chosen by player in the startlevel screen
    Keys.chosenDifficulty = 0;
    // Number - level of enemies in current mission
    Keys.enemyLevel = 1;
    // Number - cached level of enemies (enemyLevel may change)
    Keys.originalEnemyLevel = 1;
    // Number - cached character
    Keys.originallySelectedCharacter = 'boy';
    // Number - number of coins collected during mission
    Keys.coinsCollected = 0;
    // Array of Numbers - containing mission(s) to unlock if current mission is cleared
    Keys.toUnlock = [];
    // Number - amount of stored up combo points
    Keys.comboMeter = 0;
    // Number - amount combo points waiting to be added
    Keys.preComboMeter = 0;
    // Number - saved health of a citizen
    Keys.citizenHealth = 4;
    // Number - saved maxHealth of a citizen
    Keys.citizenMaxHealth = 4;
    // Boolean - if you have failed to protect, starts gameover in next section of a mission
    Keys.citizenFail = false;
    // Object with possible rewards that can be obtained after mission is cleared
    Keys.possibleRewards = {};
    // Array of Numbers - Most recent levels unlocked
    Keys.recentlyUnlocked = [1];
    // Array of Numbers - unlocked levels
    Keys.unlockedLevels = [1];
    // Array of strings - unlocked characters
    Keys.unlockedCharacters = ['boy'];
    // ID of the character the player has currently selected
    Keys.selectedCharacter = 'boy';
    // Array of Strings - equipped consumables (global, for all characters)
    Keys.equippedConsumables = [];
    // Array with objects
    Keys.dailyQuests = [];
    // Number - current day
    Keys.questDay = -1;
    // Number - claimed daily quests
    Keys.totalQuestsClaimed = 0;
    // Object - local player's weapon data (note: must be serializable as JSON)
    Keys.currentWeapon = {
        type: '',
        durability: 0
    };
    // Number - Total amount of stars
    Keys.totalStars = 0;
    // Object - Social Media liked/followed
    Keys.socialMedia = {
        facebook: false,
        twitter: false,
        instagram: false,
        weibo: false
    };
    // Object - Chests in player's posession
    Keys.chests = {
        classic: 0,
        special: 0,
        mythic: 0
    };
    // Object - Classic chests opened
    Keys.classicChestsOpened = 0;
    // String - Last known name of the local player
    Keys.localPlayerName = 'Guest';
    // String - First ID given from game services
    Keys.gameServiceID = null;
    // String - KTPlay user name, saved only for reference
    Keys.ktPlayUserName = null;
    // Number - Current tutorial
    Keys.tutorial = 0;
    // Number - Death tutorial,
    //          0 = never died, 1 = ready for tutorial, 2 = saw tutorial, 3 = ready for 2nd tutorial, 4 = saw 2nd
    Keys.tutorial1stTimeDeath = 0;
    // Number - Item tutorial (shows hand in main game)
    Keys.tutorialItem = 0;
    // Number - Super Attack tutorial (shows hand in main game)
    Keys.tutorialSuperAttack = 0;
    // Number - Steps for daily shop + equipment tutorial
    Keys.tutorialDailyShopItems = 0;
    // Number - Steps for equipment
    Keys.tutorialEquipment = 0;
    // Number - Steps for first free gift tutorial
    Keys.tutorialFreeGift = 0;
    // Number - Steps for first daily chest
    Keys.tutorialDailyChest = 0;
    // Number - Steps for chest tutorial
    Keys.tutorialChest = 0;
    // Number - Steps for character upgrade tutorial
    Keys.tutorialCharacter = 0;
    // Number - Worldselect tutorial complete
    Keys.tutorialWorldSelect = 0;
    // Number - Worldselect tutorial complete
    Keys.tutorialQuest = 0;
    // Number - First time quickloot tutorial complete
    Keys.tutorialQuickLoot = 0;
    // Number - First time quickloot (sensei giving some info about it)
    Keys.tutorialQuickLootInfo = 0;
    // Number - First time getting 3 stars ever
    Keys.firstTime3Stars = false;
    // Number - First time allies tutorial complete
    Keys.tutorialAllies = 0;
    // Boolean - Has broken something before?
    Keys.firstBreakable = true;
    // Boolean - Has new equipment?
    Keys.hasNewEquipment = false;
    // Object - Daily shop data
    Keys.dailyShop = {
        firstTime: true, // very first time there will be tutorial items
        time: 0, // time when shop refreshes in milliseconds
        items: [], // items of today
        bought: [0, 0, 0] // count how many times bought (increases price)
    };
    // Object - Daily ad refresh
    Keys.dailyAdRefresh = {
        adsWatched: 0,
        day: 0
    };
    // Object - Ad for energy
    Keys.adsForEnergy = {
        adsWatched: 0,
        day: 0
    };
    // Object - Ads for coins
    Keys.adsForCoins = {
        adsWatched: 0,
        day: 0,
        time: 0
    };
    // Number - Last known server time
    Keys.lastKnownTime = 0;
    // Number - difference between user's time and server timne
    Keys.timeOffset = null;
    // Object - contains tower challenge time / number of rewards collected / highscore
    Keys.tower24 = {
        time: 0,
        day: 0
    };    
    // Object - contains current 24h gift reward
    Keys.gift24 = {
        reward: 'none',
        amount: 0,
        time: 0
    };
    // Object - contains current 12h gift reward
    Keys.gift12 = {
        reward: 'none',
        amount: 0,
        time: 0
    };
    // Object - contains current 8h gift reward
    Keys.gift8 = {
        reward: 'none',
        amount: 0,
        time: 0
    };
    // Object - The amount of ads viewed and timer to reset
    Keys.giftAd = {
        views: 0,
        time: 'unset'
    };
    // Number - Free gift rotation of object by using an incrementing number
    Keys.giftRotation = 0;
    // Boolean - Automatically select items in start level screen
    Keys.autoSelect = true;
    // Boolean - Accepted notifications
    Keys.notificationsAllowed = false;
    // Number - Asked for notifications
    Keys.notificationsAsked = 0;
    // Object - Saved identifiers for local notifications, can be used to cancel them
    Keys.notifications = {
        /*
        [name, defined by us]: [identifier (String), defined by Cocoon]
        */
    };
    // Number - Total number of rewarded ads watched
    Keys.rewardedAdsWatched = 0;
    // Number - Total number of chests opened
    Keys.chestsOpenedTotal = 0;
    // Number - Total number of classic chests opened
    Keys.chestsClassicTotal = 0;
    // Number - Total number of special chests opened
    Keys.chestsSpecialTotal = 0;
    // Number - Total number of mythic chests opened
    Keys.chestsMythicTotal = 0;

    // Boolean - Ready to ask for review
    Keys.shouldAskForReview = false;
    // Boolean - Asked for review
    Keys.hasReviewed = false;

    // Object - contains offerData, see offerpack.js
    Keys.offers = {};
    // Object - bought offers in the past
    Keys.offersBought = {};
    // Number - counter for bootups
    Keys.bootupsForOffer = 0;
    // Number - counter for opening the worldmap
    Keys.worldMapForOffer = 0;
    // Boolean - force the starter pack
    Keys.showStarterPack = false;
    // String - force an offer
    Keys.showOffer = '';
    // Number - used weapon in mission: higher chance of showing weapon pack, 0: unused, 1: used, 2: cleared mission
    Keys.usedWeapon = 0;
    // Number - used any consumable, 0: unused, 1: used, 2: cleared mission
    Keys.usedConsumable = 0;
    // Number - Ads that were presented to the user but user chose to ignore it, resets when player watches an ad
    Keys.ignoredAds = 0;
    // Number - Last time that user watched any ad
    Keys.lastTimeWatchedAd = 0;
    // Number - Last time that user saw crosspromo
    Keys.lastTimeSawCrossPromo = 0;
    // Number - Last time that user saw interstitial
    Keys.lastTimeSawInterstitial = 0;
    // Number - Is the current mission started with online multiplayer?
    Keys.isMissionOnline = false;
    // Object - stuff to show in the worldselect
    Keys.worldSelectItems = {
        stars: 0,
        item: '',
        lastMission: ''
    };
    // Array of Strings - selected enemy allies
    Keys.allies = [null, null];
    // Array of Strings - Unlocked allies
    Keys.unlockedAllies = [];
    // Object - data for daily chest
    Keys.dailyChestData = {
        day: 0, // last known day chest claimed
        claimed: 0 // claimed chests
    };
    // Boolean - is currently playing tower
    Keys.isTower = false;
    // Number - Tower challenge floor
    Keys.towerFloor = 0;
    // Number - Which section of the tower has been loaded
    Keys.towerSection = '';
    // Object - Collected gachas
    Keys.towerGachas = {};
    // Boolean - Did get first gacha ever
    Keys.towerFirstGacha = false;
    // Number - Score for tower (kills)
    Keys.towerScore = 0;
    // Number - Highest score today
    Keys.towerHighestScoreToday = 0;
    // Number - Highest floor ever
    Keys.towerHighestScoreEver = 0;
    // Number - Highest floor today
    Keys.towerHighestFloorToday = 0;
    // Number - Highest floor ever
    Keys.towerHighestFloorEver = 0;
    // Number - Keys collected from tower gacha
    Keys.towerGotKeys = 0;
    // Number - The (server) time at which the player started playing this game
    Keys.initialStartingTime = 0;
    // Number - time scheduled into future to reset the backups
    Keys.backupResetTime = 0;
    // Boolean - is the player joining another one's room? Joiners do not gain stars or rewards
    Keys.isJoining = false;
    // Number - Asked about getting notifications for energy
    Keys.askNotificationEnergy = 0;
    // Number - Daily gems left to give
    Keys.dailyGems = 0;
    // Number - Coin sink, total spendage
    Keys.coinSink = 0;
    // Number - Gem sink, total spendage
    Keys.gemSink = 0;
    // Boolean - Did user ever press the ktplay button
    Keys.pressedKTPlay = false;
    // Number - Reputation points for random online play
    Keys.reputationPoints = 0;
    // Boolean - The current mission is played for reputation
    Keys.wantsReputation = 0;

    //Beach Event data
    Keys.beachEventData = {
        currentMission : 0,
        tickets:  {
            numberOfTickets: 20,
            regen: null
        }
    };
    // Event ticket timer
    Keys.eventTicket = {
        time: 0,
        day: 0
    };    
    // Boolean - is currently playing (rad) event
    Keys.isEvent = false;
    // Number - Number of tickets to play event
    Keys.tickets = 0;
    // Object - Time when new ticket is generated, constructed as object to make compatible with TimerTab
    Keys.nextTicket = {
        time: 0
    };
    // Number - Amount of event missions cleared
    Keys.eventMissionsCleared = 0;

    /**
     * Base keys. These allow to set up constructed SaveKeys.
     * For example, instead of saving the upgrades for each character in a big monolithic object, we can create a new save key for each character,
     * like 'upgrades-boy'
     */
    var Base = {};
    //Array of strings
    Base.equipped = [];
    //Object, containing current stamina amount and data to calculate new stamina amounts over time
    Base.stamina = {
        amount: 20,
        fillTime: -1,
        setTime: -1,
    };
    //Array with energy objects
    Base.energy = {
        energy: 20,
        maximum: 20,
        regen: null
    };
    // String - level of a character
    Base.level = 1;
    // String - max level of a character
    Base.maxLevel = 10;
    // Int amount of stars for a level, NOTE: default value of -1 means level has not been played yet
    Base.missionStars = -1;
    // Array containing reward data
    Base.rewards = [false, false, false];
    Base.rewardsOnline = [false];
    // Number - time scheduled into the future until reward becomes available again, per mission
    Base.rewardCooldown = 0;
    //Object containing upgrade tokens
    Base.token = {
        health: 0,
        attack: 0,
        skill: 0
    };
    // Number - Offers shown
    Base.offersShown = 0;
    // Boolean - Showed scroll before
    Base.openedScroll = false;
    // Number - Number of backups used on character
    Base.backups = false;

    /**
     * Private functions
     */
    var getParsedDefault = function (def) {
        if (def !== null) {
            if (def.constructor === Array) {
                def = [].concat(def);
            } else if (typeof def === 'object') {
                def = JSON.parse(JSON.stringify(def));
            }
        }
        return def;
    };

    var checkKey = function (key) {
        var base;
        var def;
        if (Keys.hasOwnProperty(key)) {
            return true;
        } else {
            //check if this is a constructed key
            if (key.indexOf('-') !== -1) {
                base = key.split('-')[0];
                if (key === base + '-') {
                    Utils.log('only base part of key provided!');
                    return false;
                }
                if (Base.hasOwnProperty(base)) {
                    //we now know this to be a valid (constructed) key, so add it to the Keys list
                    def = getParsedDefault(Base[base]);
                    Keys[key] = def;
                    return true;
                }
            }
            Utils.log('WARNING: save key ' + key + ' is not defined!');
            return false;
        }
    };

    var save = function (key, value) {
        if (!checkKey(key)) {
            return;
        }
        Bento.saveState.save(key, value);
    };

    var excludedKeys = CloudSave.getExcluded();
    var load = function (key) {
        if (!checkKey(key)) {
            return undefined;
        }
        var def = getParsedDefault(Keys[key]);

        return Bento.saveState.load(key, def);
    };

    /**
     * Public functions
     */
    var manager = {
        save: save,
        load: load,
        add: function (key, value) {
            var current = load(key);
            current += value;
            save(key, current);
        },
        saveToCloud: function () {
            // attempt to save to cloud
            // cloud saving starts after mission 5
            var unlockedMissionsCount = load('unlockedLevels').length;
            if (unlockedMissionsCount > 5) {
                CloudSave.saveToCloud();
            }
        }
    };
    return manager;
});