/**
 * Tower manager
 * @moduleName Tower
 */
bento.define('managers/tower', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween',
    'modules/randomizer',
    'managers/save',
    'modules/matchmanager',
    'globals',
    'components/boss',
    'managers/reward',
    'components/eventlistener',
    'modules/localization',
    'components/anchor',
    'gui/shadowtext',
    'modules/gamecenter',
    'modules/googleplay'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween,
    Randomizer,
    SaveManager,
    MatchManager,
    Globals,
    Boss,
    RewardManager,
    EventListener,
    Localization,
    Anchor,
    ShadowText,
    GameCenter,
    GooglePlay
) {
    'use strict';
    var Tower = function (settings) {
        var startGameAsync = settings.startGameAsync;
        var seed = Globals.towerInfo.seed; // should be set during the tower screen
        var randomizer = new Randomizer({
            seed: seed
        });
        var onHasBoss = settings.onHasBoss || function () {};
        var onComplete = settings.onComplete || function () {};
        var towerJson = Bento.assets.getJson('towerchallenge');
        var setConsistentSeed = function () {
            // we check all online player's their reported seed
            // to stay consistent we can simply pick the highest seed
            // (we could also try to find who is the master)
            var highestSeed = Globals.towerInfo.seed;
            if (MatchManager.isOnline) {
                // find highest level
                Utils.forEach(MatchManager.otherPlayers, function (other, i, l, breakLoop) {
                    if (other && other.seed > highestSeed) {
                        highestSeed = other.seed;
                    }
                });
            }
            seed = highestSeed;
            Globals.towerInfo.seed = highestSeed;
            randomizer.setSeed(seed);
        };

        var tower = {
            name: 'towerChallenge',
            /**
             * Setup tower
             */
            setup: function () {
                var bestFloorEver = SaveManager.load('towerHighestFloorEver');
                var bestFloorToday = SaveManager.load('towerHighestFloorToday');
                var currentMission = SaveManager.load('currentMission');
                var currentSection = SaveManager.load('currentSection');
                var spawnPoints = [];
                var spawners = [];
                var enemyList = getEnemyList();
                // tower floor
                var currentFloor = SaveManager.load('towerFloor') + 1;
                var previousSection = SaveManager.load('towerSection'); // can be an empty string
                var randomSection = previousSection.length > 0 ? previousSection.substr(1) : '0';
                var scoreText = new ShadowText({
                    z: Globals.Layers.GUI,
                    text: Localization.getText('score').replace('{NUMBER}', SaveManager.load('towerScore')),
                    font: 'font',
                    fontSize: 16,
                    fontColor: '#ffffff',
                    align: 'center',
                    textBaseline: 'middle',
                    linebreaks: false,
                    drawDebug: false,
                    float: true,
                    components: [
                        new Anchor({
                            screenPosition: new Vector2(0.5, 1),
                            offset: new Vector2(0, -12),
                            onComplete: function () {}
                        })
                    ]
                });
                var setSeed = function (amount) {
                    randomizer.setSeed(seed + currentFloor);
                };
                /*
                 * Increase enemy level and difficulty based on the floor
                 * Note that this has to be done before the enemies are spawned
                 */
                var setEnemyLevel = function () {
                    // level starts at player level - 4 and increases every floor
                    var enemyLevel;
                    var highestLevel = SaveManager.load('level-' + SaveManager.load('selectedCharacter'));
                    if (MatchManager.isOnline) {
                        // find highest level
                        Utils.forEach(MatchManager.otherPlayers, function (other, i, l, breakLoop) {
                            if (other && other.level > highestLevel) {
                                highestLevel = other.level;
                            }
                        });
                    }
                    highestLevel = Utils.clamp(1, highestLevel - 4, Infinity);

                    enemyLevel = highestLevel + Math.round(Math.pow(currentFloor, MatchManager.isOnline ? 0.5 : 0.67)) - 1;
                    enemyLevel = Utils.clamp(1, enemyLevel, Infinity);

                    SaveManager.save('enemyLevel', enemyLevel);
                    SaveManager.save('difficulty', Math.min(0.2 + Math.floor((currentFloor - 1) / 4 * 0.1), 0.8));
                };
                var addEnemiesToFloor = function (mapHasBoss) {
                    setSeed(2);
                    var enemiesToLoad = 0;
                    var enemiesLoaded = 0;
                    var floorAddition = Utils.clamp(0, Math.floor(currentFloor / 2), 6);
                    var maxGroupSize = 3 + randomizer.getRandom(floorAddition);

                    if (MatchManager.isOnline) {
                        maxGroupSize *= 1.5;
                        maxGroupSize = Math.floor(maxGroupSize);
                    }

                    var checkDone = function () {
                        if (enemiesToLoad > 0 && enemiesToLoad === enemiesLoaded) {
                            startGameAsync();
                        }
                    };

                    if (!mapHasBoss) {
                        Utils.forEach(spawnPoints, function (spawnGroup, i, l, breakGroupLoop) {
                            var enemies = 0;
                            Utils.forEach(spawnGroup, function (spawnPoint, ii, ll, breakLoop) {
                                // random needs to be inside for each function because of asynchronous nature
                                var random = randomizer.getRandom(enemyList.length);
                                enemies += 1;
                                enemiesToLoad += 1;
                                bento.require([enemyList[random].module], function (Module) {
                                    var spawnEnemy = new Module({
                                        rootPath: enemyList[random].rootPath
                                    });
                                    spawnEnemy.position = spawnPoint.clone();
                                    spawnEnemy.name = 'character-rnd-' + i + '-' + ii;
                                    Bento.objects.attach(spawnEnemy);
                                    enemiesLoaded += 1;
                                    checkDone();
                                });
                            });

                            var randomSpawner;
                            var randomEnemy;
                            var wave = 0;
                            var waveObjects = [];
                            setSeed(5);
                            while (enemies < maxGroupSize) {
                                if (spawners[i] && spawners[i].length) {
                                    randomSpawner = randomizer.getRandom(spawners[i].length);
                                    waveObjects[randomSpawner] = waveObjects[randomSpawner] || {};
                                    // generate waves
                                    randomEnemy = randomizer.getRandom(enemyList.length);
                                    waveObjects[randomSpawner]['wave0' + wave] = 'wave-' + wave + ':' + enemyList[randomEnemy].module + ',rootPath:' + enemyList[randomEnemy].rootPath;
                                    wave += 1;
                                }
                                enemies += 1;
                            }

                            Utils.forEach(waveObjects, function (waveObj, ii, ll, breakLoop) {
                                if (waveObj) {
                                    if (spawnGroup[0]) {
                                        spawners[i][ii].activateX = spawnGroup[0].x;
                                    }
                                    spawners[i][ii].addWavesAtStart(waveObj);
                                }
                            });
                        });
                    } else {
                        var spawnPointsLeft = [0, 1, 2, 3];
                        var bossAmount = 3;
                        var bossId = [];
                        var bossSpawn = [];
                        var nameProps = 'addLv:1';
                        bossAmount += randomizer.getRandom(2);

                        if (bossAmount === 3) {
                            nameProps += ',setHp:2';
                        } else {
                            nameProps += 'setHp:1.5';
                        }

                        while (bossAmount > 0) {
                            var random;
                            var spawnRnd;
                            var currentSpawn = 0;
                            random = randomizer.getRandom(enemyList.length);
                            spawnRnd = randomizer.getRandom(spawnPointsLeft.length);
                            bossId.push(random);
                            bossSpawn.push(spawnPointsLeft[spawnRnd]);
                            spawnPointsLeft.splice(spawnRnd, 1);
                            --bossAmount;
                        }

                        Utils.forEach(bossId, function (id, i, l, breakLoop) {
                            enemiesToLoad += 1;
                            bento.require([enemyList[id].module], function (Module) {
                                var spawnEnemy = new Module({
                                    rootPath: enemyList[id].rootPath,
                                    tiled: {
                                        object: {
                                            name: nameProps
                                        }
                                    }
                                });
                                spawnEnemy.position = spawnPoints[0][bossSpawn[i]].clone();
                                spawnEnemy.name = 'character-boss-' + i;
                                var bossComponent = new Boss({});
                                bossComponent.onDeath = bossDeath;
                                spawnEnemy.attach(bossComponent);
                                Bento.objects.attach(spawnEnemy);
                                enemiesLoaded += 1;
                                checkDone();
                            });
                        });

                    }
                };
                // slowmo on lastboss but no level end screen
                var bossDeath = function () {
                    var bossCount = Boss.currentBosses.length;
                    Utils.forEach(Boss.currentBosses, function (boss, ii, ll, breakLoopBossCount) {
                        boss.getComponent('punchable', function (punchableComponent) {
                            if (punchableComponent.isDead) {
                                bossCount -= 1;
                            }
                        });
                    });

                    if (Bento.isDev() && window.skipBossIntro || bossCount > 0) {
                        // faster bosses during dev
                        return;
                    }
                    new Tween({
                        from: 0.1,
                        to: 0.05,
                        in: 120,
                        ignoreGameSpeed: true,
                        ease: Tween.LINEAR,
                        onUpdate: function (v, t) {
                            Bento.setGameSpeed(v);
                        },
                        onComplete: function () {
                            // restore speed
                            new Tween({
                                from: 0.05,
                                to: 1,
                                'in': 60,
                                ease: Tween.LINEAR,
                                ignoreGameSpeed: true,
                                onUpdate: function (v, t) {
                                    Bento.setGameSpeed(v);
                                },
                                onComplete: function () {
                                    EventSystem.fire('slowmoEnd', {});
                                }
                            });
                        }
                    });
                };
                tower.onObjectSpawnPoint = function (object) {
                    var nameProperties = Utils.nameToJson(object.name);
                    var spawnGroup = parseInt(nameProperties.group);
                    // make a new array if the group doesnt have on yet
                    if (!spawnPoints[spawnGroup]) {
                        spawnPoints[spawnGroup] = [];
                    }
                    // add the spawnpoint to the array on the predefined position
                    spawnPoints[spawnGroup][parseInt(nameProperties.position)] = new Vector2(object.x + object.width / 2, object.y + object.height / 2);
                };
                tower.onSpawnTower = function (instance, object, extraData, nameProperties) {
                    if (instance.name.indexOf('spawner') > -1) {
                        // add spawners to group so that they can be populated with enemies
                        if (Utils.isDefined(nameProperties.group) && Utils.isDefined(nameProperties.position)) {
                            if (!spawners[nameProperties.group]) {
                                spawners[nameProperties.group] = [];
                            }
                            spawners[nameProperties.group][nameProperties.position] = instance;
                        } else {
                            Utils.log('WARNING: ' + instance.name + " doesn't have a group (" + nameProperties.group + ') or position (' + nameProperties.position + ') set at section ' + Bento.saveState.load('towerSection'));
                        }
                    }
                };
                tower.onSpawnCompleteTower = function (mapHasBoss) {
                    setEnemyLevel();
                    addEnemiesToFloor(mapHasBoss);
                };

                // before starting, sync the seed
                if (MatchManager.isOnline) {
                    setConsistentSeed();
                }

                SaveManager.save('towerFloor', currentFloor);

                setSeed();

                if (currentFloor % 4 !== 0) {
                    // normal floor
                    while ('t' + randomSection === previousSection) {
                        randomSection = randomizer.getRandom(Globals.towerSections);
                    }
                    currentSection = 't' + randomSection;
                } else {
                    // boss floor
                    currentSection = 't_boss' + randomizer.getRandom(Globals.towerBossSections);
                    onHasBoss();
                }
                SaveManager.save('towerSection', currentSection);
                SaveManager.save('currentSection', currentSection);
                SaveManager.save('currentMission', 'tower');

                if (currentFloor < 7) {
                    tower.preloadNextEnemies();
                }

                // track score
                if (currentFloor > bestFloorEver) {
                    SaveManager.save('towerHighestFloorEver', currentFloor);
                }
                if (currentFloor > bestFloorToday) {
                    SaveManager.save('towerHighestFloorToday', currentFloor);
                }
                Bento.objects.attach(new EventListener({
                    name: 'destroyEnemyController',
                    onEvent: function (data) {
                        var score = SaveManager.load('towerScore');
                        var bestScoreEver = SaveManager.load('towerHighestScoreEver');
                        var bestScoreToday = SaveManager.load('towerHighestScoreToday');
                        score += 1;
                        // update score text
                        scoreText.setText(Localization.getText('score').replace('{NUMBER}', score));

                        if (score > bestScoreEver) {
                            SaveManager.save('towerHighestScoreEver', score);
                        }
                        if (score > bestScoreToday) {
                            SaveManager.save('towerHighestScoreToday', score);
                        }
                        SaveManager.save('towerScore', score);
                    }
                }));
                Bento.objects.attach(scoreText);
                Bento.objects.attach(new EventListener({
                    name: 'endTower',
                    onEvent: function (data) {
                        // send score on game over
                        // consideration: score is kills or floors?
                        var score = SaveManager.load('towerScore');
                        var floor = SaveManager.load('towerFloor');
                        if (Utils.isNativeIos()) {
                            GameCenter.post(score, 'CgkI9NqdkoAGEAIQAQ', function () {});
                        } else if (Utils.isNativeAndroid()) {
                            GooglePlay.post(score, 'CgkI9NqdkoAGEAIQAQ', function () {});
                        }

                        EventSystem.fire('Analytics', {
                            category: 'Tower',
                            action: 'score',
                            value: score
                        });
                        EventSystem.fire('Analytics', {
                            category: 'Tower',
                            action: 'floor',
                            value: floor
                        });
                    }
                }));

                onComplete({
                    currentSection: currentSection,
                    currentMission: 'tower',
                    currentMap: 'missions/tower/' + currentSection
                });
            },
            /**
             * Preload enemies for next group
             * Point of interest: maybe keep 2 asset groups and load and unload these groups all the time
             */
            isLoading: false,
            preloadNextEnemies: function () {
                var assetGroups = Bento.assets.getAssetGroups();
                var city = Utils.assetGroups.city;
                var towerEnemies = assetGroups['towerEnemies'] || {
                    spritesheets: {},
                    path: city.path
                };
                var enemyList = towerJson.enemyList;
                var missingEnemies = [];
                var enemiesToLoad = [];
                var pickRandom = function () {
                    // pick a random enemy form the missing list
                    if (!missingEnemies.length) {
                        return;
                    }
                    var index = randomizer.getRandom(missingEnemies.length);
                    var enemyToLoad = missingEnemies.splice(index, 1);
                    enemiesToLoad = enemiesToLoad.concat(enemyToLoad);
                };
                var getSpriteSheets = function () {
                    // collect all the spritesheets from the city asset group
                    var newSpriteSheets = [];
                    Utils.forEach(enemiesToLoad, function (data, i, l, breakLoop) {
                        var rootPath = data.rootPath;
                        Utils.forEach(city.spritesheets, function (citySpriteSheet) {
                            if (citySpriteSheet.indexOf(rootPath) === 0) {
                                if (city.spritesheets[citySpriteSheet]) {
                                    towerEnemies.spritesheets[citySpriteSheet] = citySpriteSheet;
                                }
                            }
                        });
                    });
                };

                // collect which enemies are missing based on what's loaded already
                Utils.forEach(enemyList, function (data, i, l, breakLoop) {
                    if (!Utils.hasAsset('spritesheets', data.rootPath + 'idle')) {
                        missingEnemies.push(data);
                    }
                });

                // pick random 3 enemies to be loaded
                pickRandom();
                pickRandom();
                pickRandom();

                if (enemiesToLoad.length > 0) {
                    tower.isLoading = true;
                    // collect all spritesheets of enemies
                    getSpriteSheets();

                    assetGroups['towerEnemies'] = towerEnemies;
                    Bento.assets.load('towerEnemies', function () {
                        // complete
                        tower.isLoading = false;
                    }, function (assetsLoaded, assetCount, name) {
                        // console.log('loaded ' + name);
                    });
                }

            },
            /**
             * Remove enemy assets from tower
             */
            unloadAssets: function () {
                if (tower.isLoading) {
                    console.log('WARNING: Tower is currently loading');
                    return;
                }
                Bento.assets.unload('towerEnemies', true);
                var city = Utils.assetGroups.city;
                var assetGroups = Bento.assets.getAssetGroups();
                assetGroups['towerEnemies'] = {
                    spritesheets: {},
                    path: city.path
                };
            },
            // overwritten during setup
            onObjectSpawnPoint: function () {},
            onSpawnTower: function () {},
            onSpawnCompleteTower: function () {}
        };
        /**
         * Get list of enemies based on what is loaded
         */
        var getEnemyList = function () {
            var enemyList = towerJson.enemyList;
            var listOut = [];
            Utils.forEach(enemyList, function (data, i, l, breakLoop) {
                if (Utils.hasAsset('spritesheets', data.rootPath + 'idle')) {
                    listOut.push(data);
                }
            });
            return listOut;
            // return enemyList;
        };

        return tower;
    };
    Tower.getRandomReward = function () {
        // gives a random reward for the gacha
        var categories = ['currency', 'currency', 'currency', 'consumables', 'weapons'];
        var category = categories[Utils.getRandom(categories.length)];
        var item = 'currency/coin';
        var amount = 1;

        if (category === 'weapons' || category === 'consumables') {
            item = RewardManager.getRandomConsumable(category).key;
        } else if (category === 'currency') {
            // gem (20%) or coin
            item = 'currency/' + (Utils.getRandom(5) ? 'coin' : 'gem');
            // amount
            if (item === 'currency/coin') {
                amount = (100 + Utils.getRandom(3) * 50);
            } else {
                amount = 2 + Utils.getRandom(4);
            }
        }

        // if user didnt collect a key yet: 20% chance of a key
        if (SaveManager.load('towerGotKeys') < 1 && Utils.getRandom(5) === 0) {
            category = 'currency';
            item = 'currency/key';
            amount = 1;
            SaveManager.add('towerGotKeys', 1);
        }

        // first time: overwrite gacha result
        if (!SaveManager.load('towerFirstGacha')) {
            SaveManager.save('towerFirstGacha', true);
            category = 'currency';
            item = 'currency/coin';
            amount = 200;
        }

        return {
            category: category,
            item: item,
            amount: amount
        };
    };
    Tower.shouldShowGacha = function () {
        var currentFloor = SaveManager.load('towerFloor');
        var collectedGachas = SaveManager.load('towerGachas');
        var hasGacha = false;
        var collectedCount = Utils.getKeyLength(collectedGachas);
        // is there a gacha at this floor?
        if (Globals.towerGachaFloors.indexOf(currentFloor) >= 0) {
            hasGacha = true;
        }
        // did player claim the gacha already at this floor?
        if (collectedGachas[currentFloor]) {
            hasGacha = false;
        }
        return hasGacha;
    };
    Tower.countCollectedGachas = function () {
        var collectedGachas = SaveManager.load('towerGachas');
        return Utils.getKeyLength(collectedGachas);
    };
    return Tower;
});