/**
 * Ask a yes/no question to all remote players, 3 way communication:
 * step 1: send question to all remotes
 * step 2: receive a response from all remotes (or any?)
 * step 3: confirm the joint decision with all remotes
 *
 * All remotes need to respond through onMessage
 *
 * @param {Object} settings 
 * @param {String} settings.name - unique id for agreement
 * @param {Function} settings.onResponses - handle all responses
 * @param {Function} settings.onResponse - callback for a single response
 * @param {Function} settings.onDisconnect - what to do if one of the remotes disconnects during step 2
 * @param {Function} settings.onConflict - what to do if one of the remotes sends an agreement as well
 * @moduleName Agreement
 */
bento.define('modules/agreement', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/matchmanager'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    MatchManager
) {
    'use strict';
    // all agreements need to have some unique ID
    var uniqueId = 0;
    var Agreement = function (settings) {
        var agreementId = uniqueId + (MatchManager.localPlayer.playerId << 8); // reserving 8 bits for agreementIds
        var name = settings.name;
        var messageBody = settings.messageBody; // extra data to send
        var onResponses = settings.onResponses;
        var onResponse = settings.onResponse;
        var onDisconnect = settings.onDisconnect;
        var onConflict = settings.onConflict;

        // message names
        var recipientOut = name + '-out';
        var recipientIn = name + '-in';
        var recipientEnd = name + '-confirm';
        var responses = {};

        // possible actions to take
        var actions = {
            cancel: function (reason) {
                // cancel the agreement
                // for example when any remote responds "no" to an invitation or a conflict occurred
                MatchManager.sendData({
                    agreementId: agreementId,
                    response: reason
                }, recipientEnd);
                cleanUp();
            },
            removeFromAgreement: function (playerId) {
                // remove this remote player from the agreement 
                checkDone();
            },
            sendConfirmation: function (message) {
                MatchManager.sendData({
                    agreementId: agreementId,
                    response: message
                }, recipientEnd);
                cleanUp();
            }
        };

        // listeners
        var responseListener = function (msg) {
            var remoteId = msg.sender;

            console.log('get response');

            // what to do if the agreementId does not match?
            // -could be a conflict
            // -remnant of an old agreement
            if (msg.agreementId !== agreementId) {
                // for now let's ignore it
                return true;
            }

            // update responses
            responses[remoteId] = {
                state: 'received',
                msg: msg
            };
            // callback
            if (onResponse) {
                onResponse(msg, actions);
            }

            checkDone();
        };
        var disconnectListener = function (data) {
            // user disconnected, one of the actions MUST be called
            onDisconnect(data, actions);
        };
        var conflictListener = function () {
            // a conflict occurred
            actions.cancel('conflict');
            if (onConflict) {
                onConflict();
            }
        };
        var checkDone = function () {
            var done = true;
            // did we receive all responses?
            Utils.forEach(responses, function (response, i, l, breakLoop) {
                // for any waiting status the agreement is not done
                if (response.state === 'waiting') {
                    done = false;
                    breakLoop();
                }
            });

            if (done) {
                console.log('all responses received');
                if (onResponses) {
                    // game must either cancel or confirm
                    onResponses(responses, actions);
                }
                cleanUp();
            }
        };
        var cleanUp = function () {
            // delete self
            Bento.objects.remove(agreement);
            // clear out agreement
            delete Agreement.outgoing[name];
        };

        // agreements begins when it's attached
        var agreement = {
            name: 'agreement-' + name,
            start: function () {
                // cache outgoing agreements
                Agreement.outgoing[name] = agreement;

                // get ready to receive responses from remotes
                responses = {};
                Utils.forEach(MatchManager.otherPlayers, function (playerData, i, l, breakLoop) {
                    responses[playerData.playerId] = {
                        state: 'waiting',
                        msg: ''
                    };
                });

                // clean up the message queue
                var messages = MatchManager.getMessages(recipientIn);
                messages.length = 0;

                // possible responses:
                // 1. normal response by remotes
                MatchManager.onMessage(recipientIn, responseListener);
                // 2. remotes disconnects
                EventSystem.on(MatchManager.Events.PLAYERDISCONNECTS, disconnectListener);
                // 3. remote sends a similar outgoing agreement: conflict
                // listener will call agreement.conflict


                // send to others
                MatchManager.sendData({
                    agreementId: agreementId,
                    messageBody: messageBody
                }, recipientOut);
            },
            destroy: function () {
                // clean up events
                MatchManager.offMessage(recipientIn, responseListener);
                EventSystem.off(MatchManager.Events.PLAYERDISCONNECTS, disconnectListener);
            },
            conflict: conflictListener
        };

        uniqueId += 1;
        uniqueId %= (1 << 8);

        return agreement;
    };

    // current outgoing agreements
    Agreement.outgoing = {};

    // listen to outgoing agreements from remotes
    Agreement.Listener = function (settings) {
        var name = settings.name;
        var onComplete = settings.onComplete;
        var onReceiveAgreement = settings.onReceiveAgreement;
        var onReceiveConfirmation = settings.onReceiveConfirmation;
        var isWaitingForConfirmation = false;
        var isWaitingForGame = false;
        // these messages are coming from remotes that are sending outgoing agreements
        var recipientOut = name + '-out';
        var recipientIn = name + '-in';
        var recipientEnd = name + '-confirm';
        var agreementListener = function (msg) {
            // if a new agreement comes in while we are either waiting for confirmation or response (by game/user)
            // this is considered to be a conflict, it's likely a (cancel) confirmation will be sent shortly after
            if (isWaitingForConfirmation) {
                return;
            }
            // player isnt able to respond to agreements while there is an outgoing agreement is active
            if (Agreement.outgoing[name]) {
                Agreement.outgoing[name].conflict();
                return;
            }

            var remoteId = msg.sender;
            var agreementId = msg.agreementId;
            var actions = {
                respondWith: function (response) {
                    MatchManager.sendData({
                        agreementId: agreementId,
                        response: response
                    }, recipientIn);

                    // we now wait for the confirmation of the agreement sender
                    isWaitingForGame = false;
                }
            };
            console.log('receive agreement');

            // we now wait for a response action by the game
            isWaitingForGame = true;
            if (onReceiveAgreement) {
                onReceiveAgreement(msg, actions);
            }

            // we already start waiting for a response by the agreement starter
            isWaitingForConfirmation = true;
            MatchManager.onMessage(recipientEnd, confirmListener);
        };
        var confirmListener = function (msg) {
            if (isWaitingForGame) {
                // we were still waiting for the game/user input, but the remote has already confirmed the outcome
                // possibly an agreement was not met or a conflict occurred.

                // not sure what to do here, probably just listen to the outcome
            }
            isWaitingForGame = false;
            isWaitingForConfirmation = false;

            if (onComplete) {
                onComplete(msg);
            }
            cleanUp();
        };
        var cleanUp = function () {
            MatchManager.offMessage(recipientEnd, confirmListener);
        };

        var listener = {
            name: 'agreementListener-' + name,
            start: function () {
                MatchManager.onMessage(recipientOut, agreementListener);
            },
            destroy: function () {
                MatchManager.offMessage(recipientOut, agreementListener);
            }
        };

        return listener;
    };

    return Agreement;
});