/**
 * Cloud save module using AWS Lambda to be used with cloudsave.js
 * Modeled after Google Play Saved Games
 * @moduleName AwsCloudSave
 */
bento.define('modules/awscloudsave', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var didSendWarning = false;
    var apiKey = 'rbnKKhXeYR5d8yihRar3oKVWIXHBbpq6wHFNwU5j';
    /**
     * Get id from GameCenter or GooglePlay
     */
    var getId = function (onSuccess, onError) {
        // we will use gameServiceID, which must be stored previously
        var gameServiceId = Bento.saveState.load('gameServiceID', null);
        if (!gameServiceId) {
            if (Bento.isDev()) {
                // use "dev" as id
                onSuccess('dev');
                return;
            }
            onError("Game service ID not found.");
            if (!didSendWarning) {
                didSendWarning = true;
                // send warning about failed cloud save
                EventSystem.fire('Notification', {
                    textId: 'cloudSaveError'
                });
                if (Utils.isNativeIos()) {
                    EventSystem.fire('Notification', {
                        text: 'Cloud saving requires Game Center.'
                    });
                }
                if (Utils.isNativeAndroid()) {
                    EventSystem.fire('Notification', {
                        text: 'Cloud saving requires Google Play.'
                    });
                }
            }
            return null;
        }
        if (Utils.isNativeIos()) {
            gameServiceId = 'gc_' + gameServiceId;
        }
        if (Utils.isNativeAndroid()) {
            gameServiceId = 'gp_' + gameServiceId;
        }
        onSuccess(gameServiceId);
        return gameServiceId;
    };
    /**
     * Send savefile to aws
     */
    var sendToAws = function (snapshot, onSuccess, onError) {
        var ktPlayId = Bento.saveState.load('sentKTUserID', null);
        var userName = Bento.saveState.load('ktPlayUserName', null) || Bento.saveState.load('localPlayerName', 'Guest');
        var saveData = JSON.stringify(snapshot);
        var id = snapshot.identifier;
        var body = {
            id: id,
            saveFile: saveData,
            checksum: Utils.checksum(snapshot.identifier + saveData + 'btstrt'),
            ktPlayId: ktPlayId,
            userName: userName
        };
        var xhr = new window.XMLHttpRequest();
        xhr.open("POST", "https://scctd5apj0.execute-api.us-east-1.amazonaws.com/prod/beatstreet-cloudsave");
        xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
        xhr.setRequestHeader("X-Api-Key", apiKey);
        xhr.onreadystatechange = function () {
            if (xhr.readyState === 4) {
                if (xhr.status >= 200 && xhr.status < 300) {
                    onSuccess();
                } else {
                    if (onError) {
                        Utils.log('ERROR: bad server response. ' + xhr.responseText);
                        if (!didSendWarning) {
                            didSendWarning = true;
                            // send warning about failed cloud save
                            EventSystem.fire('Notification', {
                                textId: 'cloudSaveError'
                            });
                            EventSystem.fire('Notification', {
                                textId: 'Server response error ' + xhr.status
                            });
                        }
                        onError('Bad server response');
                    }
                }
            }
        };
        xhr.send(JSON.stringify(body));
    };
    /**
     * Download savefile
     */
    var downloadFromAws = function (id, onSuccess, onError) {
        var body = {
            id: id,
            checksum: Utils.checksum(id + 'btstrt')
        };
        var xhr = new window.XMLHttpRequest();
        xhr.open("POST", "https://nl9f1hkji7.execute-api.us-east-1.amazonaws.com/prod/beatstreet-cloudload");
        xhr.setRequestHeader("Content-Type", "application/json;charset=UTF-8");
        xhr.setRequestHeader("X-Api-Key", apiKey);
        xhr.onreadystatechange = function () {
            var res;
            var snapshot;
            if (xhr.readyState === 4) {
                if (xhr.status >= 200 && xhr.status < 300) {
                    try {
                        res = JSON.parse(xhr.responseText);
                        snapshot = JSON.parse(res.Item.saveFile);
                        onSuccess(snapshot);
                    } catch (e) {
                        if (onError) {
                            Utils.log('ERROR: could not parse data');
                            onError(xhr.responseText);
                        }
                    }
                } else {
                    if (onError) {
                        Utils.log('ERROR: bad server response. ' + xhr.responseText);
                        onError('Bad server response');
                    }
                }
            }
        };
        xhr.send(JSON.stringify(body));
    };
    return {
        writeSavedGame: function (snapshot, onSuccess, onError) {
            // get id first
            getId(function (id) {
                // overwite id with service id
                snapshot.identifier = id;
                sendToAws(snapshot, onSuccess, onError);
            }, onError);
        },
        loadSavedGame: function (identifier, onSuccess, onError) {
            // get id first (ignore the identifier provided)
            getId(function (id) {
                downloadFromAws(id, onSuccess, onError);
            }, onError);
        }
    };
});