/**
 * Interface for the Cordova File plugin.
 * Can read and write JSON data (objects passed should be stringify-able to JSON)
 * See https://cordova.apache.org/docs/en/latest/reference/cordova-plugin-file/
 * @moduleName File
 */
bento.define('modules/file', [
    'bento',
    'bento/utils'
], function (
    Bento,
    Utils
) {
    'use strict';
    var directory = 'documentsDirectory';
    var dataDirectory;
    var onError = function () {};
    var errorHandler = function (fileName, e) {
        var msg = '';

        switch (e.code) {
        case window.FileError.QUOTA_EXCEEDED_ERR:
            msg = 'Storage quota exceeded';
            break;
        case window.FileError.NOT_FOUND_ERR:
            msg = 'File not found';
            break;
        case window.FileError.SECURITY_ERR:
            msg = 'Security error';
            break;
        case window.FileError.INVALID_MODIFICATION_ERR:
            msg = 'Invalid modification';
            break;
        case window.FileError.INVALID_STATE_ERR:
            msg = 'Invalid state';
            break;
        default:
            msg = 'Unknown error';
            break;
        }

        console.log('Error (' + fileName + '): ' + msg);
        onError(e.code);
    };
    var readFromFile = function (fileName, onSuccess, onFail) {
        var pathToFile = dataDirectory + fileName;

        window.resolveLocalFileSystemURL(pathToFile, function (fileEntry) {
            fileEntry.file(function (file) {
                var reader = new window.FileReader();

                reader.onloadend = function (e) {
                    var result;
                    try {
                        result = JSON.parse(this.result);
                        if (onSuccess) {
                            onSuccess(result);
                        }
                    } catch (error) {
                        console.log("Error parsing JSON file");
                        console.log(error);
                    }
                    return result;
                };
                reader.readAsText(file);
            }, errorHandler.bind(null, fileName));
        }, errorHandler.bind(null, fileName));
    };
    var writeToFile = function (fileName, data, onSuccess, onFail) {
        data = JSON.stringify(data, null, '\t');
        window.resolveLocalFileSystemURL(dataDirectory, function (directoryEntry) {
            directoryEntry.getFile(fileName, {
                create: true
            }, function (fileEntry) {
                fileEntry.createWriter(function (fileWriter) {
                    fileWriter.onwriteend = function (e) {
                        // for real-world usage, you might consider passing a success callback
                        console.log('Write of file "' + fileName + '"" completed.');
                        if (onSuccess) {
                            onSuccess();
                        }
                    };

                    fileWriter.onerror = function (e) {
                        // you could hook this up with our global error handler, or pass in an error callback
                        console.log('Write failed: ' + e.toString());
                        if (onFail) {
                            onFail(e.toString());
                        }
                    };

                    fileWriter.write(data);

                }, errorHandler.bind(null, fileName));
            }, errorHandler.bind(null, fileName));
        }, errorHandler.bind(null, fileName));
    };

    if (window.cordova && window.cordova.file) {
        if (Utils.isNativeAndroid && Utils.isNativeAndroid()) {
            directory = 'externalDataDirectory';
        }
        dataDirectory = window.cordova.file[directory];
    }

    return {
        setDataDirectory: function (dir) {
            dataDirectory = dir;
            console.log('Current data directory: ' + dataDirectory);
        },
        read: readFromFile,
        write: writeToFile,
        setErrorCallback: function (fn) {
            onError = fn;
        }
    };
});