/**
 * Module for using KTplay functions in Bento
 * @moduleName KTPlayModule
 */
bento.define('modules/ktplay', [
    'bento',
    'bento/utils',
    'bento/eventsystem'
], function (
    Bento,
    Utils,
    EventSystem
) {
    // Does NOTHING
    var nothing = function () {};

    // Fallbacks
    var KTPlayModule = {
        init: nothing,
        setLoginOutHandler: nothing,
        setWindowHandler: nothing,
        setRewardHandler: nothing,
        showLogin: nothing,
        customLogin: nothing,
        getLoginStatus: function () {
            return false;
        },
        getCurrentUser: nothing,
        getUser: nothing,
        updateUserProfile: nothing,
        showCommunity: nothing,
        hideCommunity: nothing,
        postToCommunity: nothing,
        preloadInterstitial: nothing,
        showInterstitial: nothing,
        openDeeplink: nothing,
        deeplinkCallbacks: {},
        addARLDeeplink: nothing,
        removeARLDeeplink: nothing
    };

    if (window.plugins && window.plugins.ktplay) {
        var KTPlay = window.plugins.ktplay;

        // function for controlling how audio is handled with 
        var audioHandler = function (data) {
            if (data.sound) {
                //sound is playing in KTplay so mute music and sound
                Bento.audio.muteMusic(true, true);
                Bento.audio.muteSound(true);
            } else {
                //sound is not playing in KTplay so un-mute music and sound
                Bento.audio.muteMusic(false, true);
                Bento.audio.muteSound(false);
            }
        };

        var loginHandler = function (user) {
            console.log('LOGIN DATA:' + JSON.stringify(user));
        };

        var windowHandler = function (data) {
            console.log('WINDOW DATA:' + JSON.stringify(data.communitywindow));
        };

        var rewardHandler = function (data) {
            console.log('REWARD DATA:' + JSON.stringify(data.rewards));
        };

        KTPlayModule = {

            /**
            * Initializes the SDK and creates the events
            * @param {Object} settings -
                settings = {
                    key: 'key' string from KTplay console
                    avatarPath: 'secret' string from KTplay console
                    audioHandler: (optional) function that controls how the audio is handled receives a data object with the 'sound' parameter which is true if sound is playing in KTplay
                }
            */
            init: function (settings) {
                // initiate the sdk
                KTPlay.initSdk({
                    key: settings.key,
                    secret: settings.secret
                });

                // add the deeplink event listener
                document.addEventListener('deeplink', function (data) {
                    var fn = KTPlayModule.deeplinkCallbacks[data.deeplink];
                    if (fn) {
                        fn();
                    }
                    EventSystem.fire('KTPlayDeepLink', data);
                }, false);

                // add the handlesound event listener
                document.addEventListener('handlesound', settings.audioHandler || audioHandler, false);

                console.log("KTPlay is initialized!");
            },

            /**
             * Sets the login/logout handler callback
             * @param {function} callback - sets the function that will be used when a reward is received
             */
            setLoginOutHandler: function (callback) {
                // add the reward event listener
                document.addEventListener('logoutLogin', callback, false);
            },

            /**
             * Sets the reward handler callback
             * @param {function} callback - sets the function that will be used when a reward is received
             */
            setWindowHandler: function (callback) {
                // add the reward event listener
                document.addEventListener('communitywindow', callback, false);
            },

            /**
             * Sets the reward handler callback
             * @param {function} callback - sets the function that will be used when a reward is received
             */
            setRewardHandler: function (callback) {
                // add the reward event listener
                document.addEventListener('rewards', callback, false);
            },

            /**
             * Shows and attempts to log in directly with KTPlay
             * @param {function} success - success callback with user information object as the argument
             * @param {function} error - error callback with the error string as the argument
             */
            showLogin: function (success, error) {
                KTPlay.showLoginView(success, error);
            },

            /**
             * Attempts to log in using a custom 'gamerId' string, this can be taken from Game Center or Google Play Games
             * @param {function} success - success callback with user information object as the argument
             * @param {function} error - error callback with the error string as the argument
             */
            customLogin: function (gamerId, success, error) {
                KTPlay.customCommunityLogin(gamerId, success, error);
            },

            /**
             * Is the player logged in or not?
             * @param {function} callback - function called with status boolean as argument
             */
            getLoginStatus: function (callback) {
                KTPlay.getUserLoginStatus(callback);
            },

            /**
             * Gets the currently logged in user
             * @param {function} success - success callback with user information object as the argument
             * @param {function} error - error callback with the error string as the argument
             */
            getCurrentUser: function (success, error) {
                KTPlay.getCurrentLoggedInUser(success, error);
            },

            /**
             * Gets specific user by id
             * @param {String} userId - string of the user to be retrieved
             * @param {function} success - success callback with user information object as the argument
             * @param {function} error - error callback with the error string as the argument
             */
            getUser: function (userId, success, error) {
                KTPlay.getUserProfile(userId, success, error);
            },

            /**
            * Update profile information for the current user
            * @param {Object} userData - 
                userData = {
                    nickname: 'NewNickname' or null for no change,
                    avatarPath: 'path/to/the/file.png' or null for no change,
                    gender: 0 no change, 1 male, 2 female, 3 hidden
                }
            * @param {function} success - success callback
            * @param {function} error - error callback with the error string as the argument
            */
            updateUserProfile: function (userData, success, error) {
                KTPlay.updateUserProfile(userData, success, error);
            },

            /**
             * Show the KTPlay community window
             * @param {function} success - success callback
             * @param {function} error - error callback with the error string as the argument
             */
            showCommunity: function (success, error) {
                KTPlay.showCommunityWindow(success, error);
            },

            /**
             * Hide the KTPlay community window
             * @param {function} success - success callback
             * @param {function} error - error callback with the error string as the argument
             */
            hideCommunity: function (success, error) {
                KTPlay.hideCommunityWindow(success, error);
            },

            /**
            * Make a post within the KTplay community with code defined parameters
            * @param {Object} userPost -
                userPost = {
                    type: 'image' or 'video',
                    filePath: 'path/to/the/file.png',
                    title: 'Title of the Post',
                    message: 'The message for the post that's made
                }
            * @param {function} success - success callback
            * @param {function} failure - failure callback with the error string as the argument
            */
            postToCommunity: function (userPost, success, failure) {
                KTPlay.postToCommunity(userPost, success, failure);
            },

            /**
             * Prepares an interstitial message
             * @param {String} locationId - string for the in-app location ID
             * @param {function} success - success callback
             * @param {function} failure - error callback with the error string as the argument
             */
            preloadInterstitial: function (locationId, success, failure) {
                KTPlay.preloadInterstitialMessage(locationId, success, failure);
            },

            /**
             * Shows an interstitial message, or prepare an interstitial and immediately show it if one is not prepared
             * @param {String} locationId - string for the in-app location ID
             * @param {function} success - success callback
             */
            showInterstitial: function (locationId, success) {
                KTPlay.showInterstitialMessage(locationId, success);
            },

            /**
             * Different to the ARL callback functions, this executes a deeplink event on KT's side, allowing you to perform specific actions
             * Such as opening KT play settings, specific topics, or specific user profiles etc
             * @param {String} deeplinkID - string for deeplink id that will be received
             * @param {function} callback - the callback that will be called when the deeplink is received
             */
            openDeeplink: function (deeplinkID, success) {
                KTPlay.openDeeplink(deeplinkID, success);
            },

            //object containing the deeplink functions
            ARLdeeplinkCallbacks: {},

            /**
             * Adds a callback to be executed on deeplink event from KTplay
             * @param {String} deeplinkID - string for deeplink id that will be received
             * @param {function} callback - the callback that will be called when the deeplink is received
             */
            addARLDeeplink: function (deeplinkID, callback) {
                KTPlayModule.ARLdeeplinkCallbacks[deeplinkID] = callback;
            },

            /**
             * Removes the callback
             * @param {String} deeplinkID - string for deeplink id that will be received
             */
            removeARLDeeplink: function (deeplinkID) {
                delete KTPlayModule.ARLdeeplinkCallbacks[deeplinkID];
            }
        };
        console.log("KTPlay Ready!");
    } else {
        console.log("WARNING: KTPlay plugin does not exist.");
    }

    return KTPlayModule;
});