/**
 * A match is basically a session between players
 * This module handles the Cocoon.Multiplayer.Match instance
 * @moduleName MatchManager
 */
bento.define('modules/matchmanager', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'modules/onlineservice/appwarp',
    'modules/onlineservice/photon'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    AppWarpService,
    PhotonService
) {
    'use strict';

    /*var onlineService = AppWarpService;
    if (Utils.isCocoonJs()) {
        onlineService = CocoonService;
    }*/
    var onlineService = PhotonService;

    var MatchManager = {};
    var setActiveMatch = function (match, onStart) {
        // already processing a match
        if (!Bento.isDev() && MatchManager.currentMatch && MatchManager.currentMatch !== match) {
            Utils.log("Already processing a session.");
            EventSystem.fire('matchDuplicate', {
                error: "Already processing a session."
            });
            return;
        }

        MatchManager.reset();
        MatchManager.currentMatch = match;

        match.onDataRecieved = function (data) {
            if (data.playerData.uniqueId === MatchManager.localPlayer.uniqueId) {
                // skip own messages
                return;
            }

            var list = data.list || [];
            var ii;
            var dataObj;
            for (ii = 0; ii < list.length; ii++) {
                dataObj = list[ii];

                dataObj.msg = dataObj.msg || {};
                dataObj.msg.sender = data.playerId;
                dataObj.msg.uniqueId = data.uniqueId;

                EventSystem.fire('receiveData', dataObj);
                // MatchManager will collect all the received messages in an array so they can be processed later
                MatchManager.handleMessage(dataObj);
            }
        };

        // player state changes (connects or disconnects)
        match.onStateChanged = function (stateObj) {
            // see modules/onlineservice/cocoon
            var state = stateObj.state;

            if (state === 1) {
                // fire event
                EventSystem.fire(MatchManager.Events.PLAYERCONNECTS, stateObj);
            } else if (state > 1) {
                // find corresponding player
                Utils.forEach(MatchManager.otherPlayers, function (playerData, i, l, breakLoop) {
                    if (playerData.uniqueId === stateObj.playerData.uniqueId) {
                        // remove from players list
                        MatchManager.otherPlayers.splice(i, 1);

                        // fire event
                        EventSystem.fire(MatchManager.Events.PLAYERDISCONNECTS, {
                            playerData: stateObj.playerData,
                            playerId: playerData.playerId,
                            userName: playerData.userName,
                            uniqueId: playerData.uniqueId,
                            reason: state === 2 ? 'disconnected' : 'failure'
                        });
                        breakLoop();
                    }

                });
            }
        };

        match.onConnectionWithPlayerFailed = function (failureObj) {
            // let's not use this, see onStateChanged
        };

        match.onReady(function (error) {
            if (error) {
                // report back with callback
                if (onStart) {
                    onStart({
                        success: false,
                        match: match
                    });
                }
                return;
            }
            var ii;

            MatchManager.localPlayer = match.localPlayer;
            MatchManager.isMaster = match.playerHasLowestId();
            MatchManager.isOnline = true;

            console.log('isMaster', MatchManager.isMaster);

            if (match.otherPlayers) {
                MatchManager.otherPlayers = match.otherPlayers;
            } else {
                for (ii = 0; ii < match.players.length; ii++) {
                    if (match.players[ii].uniqueId !== MatchManager.localPlayer.uniqueId) {
                        MatchManager.otherPlayers.push(match.players[ii]);
                    }
                }
            }

            match.start();
            // fire event
            if (onStart) {
                onStart({
                    success: true,
                    match: match
                });
            }
            EventSystem.fire(MatchManager.Events.MATCHSTART, match);
        });
    };
    /**
     * Static: Reference to current Cocoon.Multiplayer.Match object
     */
    MatchManager.currentMatch = null;
    /**
     * Static: Own playerdata object (contains ids)
     */
    MatchManager.localPlayer = {
        // numerical id
        playerId: -1,
        // id according to online service (may not be numerical)
        uniqueId: -1,
        // username
        userName: ''
    };
    /**
     * Static: other player objects
     */
    MatchManager.otherPlayers = [];
    /**
     * Static: are we playing online?
     */
    MatchManager.isOnline = false;
    /**
     * Turn of syncing, is used during cutscenes
     */
    MatchManager.noSync = false;
    /**
     * Static: messages to process
     */
    MatchManager.messages = {};
    /*
     * Listeners for specific messages
     */
    var messageListeners = {};

    /**
     * Static: is current player the master?
     * default to true, so that we are master if playing singleplayer
     */
    MatchManager.isMaster = true;
    /**
     * All data that should be send at the end of this frame
     */
    var frameData = [];
    var frameDataUnreliable = [];
    /**
     * sending of all frame data
     */
    var onFrameEnd = function () {
        var message = '[';
        var ii;
        for (ii = 0; ii < frameData.length; ii++) {
            if (message.length + frameData[ii].length > 250) {
                message += ']';
                onlineService.sendData(MatchManager.currentMatch, message, MatchManager.SendModes.RELIABLE);
                message = '[';
            }
            if (message.length > 1) {
                message += ',';
            }
            message += frameData[ii];

        }
        if (message.length > 1) {
            message += ']';
            onlineService.sendData(MatchManager.currentMatch, message, MatchManager.SendModes.RELIABLE);
        }

        frameData.length = 0;

        // repeat for frameDataUnreliable
        message = '[';
        for (ii = 0; ii < frameDataUnreliable.length; ii++) {
            if (message.length + frameDataUnreliable[ii].length > 250) {
                message += ']';
                onlineService.sendData(MatchManager.currentMatch, message, MatchManager.SendModes.UNRELIABLE);
                message = '[';
            }
            if (message.length > 1) {
                message += ',';
            }
            message += frameDataUnreliable[ii];

        }
        if (message.length > 1) {
            message += ']';
            onlineService.sendData(MatchManager.currentMatch, message, MatchManager.SendModes.UNRELIABLE);
        }
        frameDataUnreliable.length = 0;
    };
    /**
     * Static: Send data to all players
     */
    MatchManager.sendData = function (msg, recipient, sendMode) {
        var data;
        if (!MatchManager.currentMatch) {
            return;
        }
        data = JSON.stringify({
            recipient: recipient || '',
            msg: msg
        });

        if (sendMode === MatchManager.SendModes.UNRELIABLE) {
            frameDataUnreliable.push(data);
        } else {
            frameData.push(data);
        }

        // onlineService.sendData(MatchManager.currentMatch, data, sendMode);
    };
    /**
     * Static: Send data to single player
     */
    MatchManager.sendDataTo = function (uniqueId, msg, recipient, sendMode) {
        var data;
        if (!MatchManager.currentMatch) {
            return;
        }
        data = JSON.stringify({
            recipient: recipient || '',
            msg: msg
        });

        onlineService.sendDataTo(uniqueId, MatchManager.currentMatch, data, sendMode);
    };
    /**
     * Static: Handle a recieved message
     */
    MatchManager.handleMessage = function (dataObj) {
        var msg = dataObj.msg || {};
        var recipient = dataObj.recipient || 'unaddressed';

        var listeners = messageListeners[recipient] || [];
        var ii;
        for (ii = 0; ii < listeners.length; ii++) {
            if (listeners[ii].call(undefined, msg)) {
                //handler used the message and returned true, indicating further usage of the message is not needed
                return;
            }
        }

        MatchManager.messages[recipient] = MatchManager.messages[recipient] || [];
        MatchManager.messages[recipient].push(msg);

        if (Bento.isDev()) {
            if (MatchManager.messages[recipient].length > 50 && !MatchManager.messages[recipient].warned) {
                Utils.log("WARNING: potential memory leak by messages " + recipient);
                MatchManager.messages[recipient].warned = true;
            }
        }
    };
    /**
     * Static: register a message handler
     */
    MatchManager.onMessage = function (recipient, handler) {
        var listeners = messageListeners[recipient];
        if (!listeners) {
            messageListeners[recipient] = [];
            listeners = messageListeners[recipient];
        }

        if (listeners.indexOf(handler) !== -1) {
            return;
        }

        listeners.push(handler);
    };
    /**
     * Static: unregister a message handler
     */
    MatchManager.offMessage = function (recipient, handler) {
        var listeners = messageListeners[recipient];
        var idx = -1;
        if (!listeners) {
            return;
        }
        idx = listeners.indexOf(handler);
        if (idx !== -1) {
            listeners.splice(idx);
        }
    };

    /**
     * Static: clean up
     */
    MatchManager.reset = function () {
        EventSystem.off('postUpdate', onFrameEnd);

        if (MatchManager.currentMatch) {
            // clean up
            MatchManager.currentMatch.destroy();
        }
        MatchManager.currentMatch = null;
        MatchManager.messages = {
            unaddressed: []
        };
        MatchManager.isMaster = true;
        MatchManager.localPlayer = null;
        MatchManager.otherPlayers = [];
        MatchManager.isOnline = false;

        EventSystem.on('postUpdate', onFrameEnd);
    };
    /*
     * Static : clear all unused messages
     */
    MatchManager.clear = function () {
        var key;
        for (key in MatchManager.messages) {
            if (MatchManager.messages.hasOwnProperty(key)) {
                MatchManager.messages[key] = [];
            }
        }
    };

    MatchManager.getMessages = function (recipient) {
        if (!recipient) {
            return MatchManager.messages['unaddressed'];
        }
        MatchManager.messages[recipient] = MatchManager.messages[recipient] || [];
        return MatchManager.messages[recipient];
    };

    /**
     * Static: find a match
     */
    MatchManager.findMatch = function (playersCount, playerGroup, callback, otherData) {
        if (MatchManager.currentMatch) {
            // clean up
            MatchManager.currentMatch.destroy();
        }
        MatchManager.currentMatch = null;

        // handle match start
        onlineService.onMatchFound = function (match, error) {
            if (match) {
                setActiveMatch(match, callback);
                // wait for the match start
                // callback({
                //     success: true
                // });
            } else {
                // matching failed
                callback({
                    success: false,
                    error: error
                });
            }
        };

        // handle

        onlineService.findMatch(playersCount, playersCount, undefined, playerGroup, undefined, otherData);
    };
    /*
     * Static: log in to the online service
     */
    MatchManager.login = function (callback, extraParams) {
        onlineService.login(callback, extraParams);

        //TODO handle invitations
    };
    /*
     * Static: is logged in? (connected)
     */
    MatchManager.isLoggedIn = function () {
        return onlineService.isLoggedIn();
    };
    /*
     * Static: disconnect
     */
    MatchManager.disconnect = function () {
        if (MatchManager.currentMatch) {
            MatchManager.currentMatch.disconnect();
            MatchManager.reset();
            messageListeners = {};
            EventSystem.fire(MatchManager.Events.MATCHEND);
        } else if (onlineService.disconnect) {
            onlineService.disconnect();
            MatchManager.reset();
            messageListeners = {};
        }
    };
    /*
     * Re-determine the master
     */
    MatchManager.determineMaster = function () {
        var match = MatchManager.currentMatch;
        var localId = MatchManager.localPlayer.playerId;

        MatchManager.isMaster = true;

        // if any other player has a lower id than you, then local player is not the master!
        Utils.forEach(MatchManager.otherPlayers, function (playerData, i, l, breakLoop) {
            if (playerData.playerId < localId) {
                MatchManager.isMaster = false;
            }
        });

    };
    MatchManager.clearMessages = function () {
        MatchManager.messages = {};
    };
    // some callback functions as replacement for events
    MatchManager.onMatchStart = function (players) {};

    // enum events fired by MatchManager
    MatchManager.Events = {
        MATCHSTART: 'onMatchStart',
        MATCHEND: 'onMatchEnd', // local player disconnects (or all remote players left??)
        PLAYERDISCONNECTS: 'onRemotePlayerDisconnect',
        PLAYERCONNECTS: 'onRemotePlayerConnect'
    };
    MatchManager.SendModes = {
        RELIABLE: 0,
        UNRELIABLE: 1
    };

    // Android only: will disconnect upon pressing the home button, except when player is trying to host a game
    MatchManager.shouldDisconnect = true;

    // received an error by service: disconnect
    EventSystem.on('OnlineService.Error', function (msg) {
        MatchManager.disconnect();
        // EventSystem.fire('Notification', {
        //     text: msg
        // });
    });

    return MatchManager;
});