/**
 * Cocoon multiplayer service
 * @moduleName CocoonService
 */
bento.define('modules/onlineservice/cocoon', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    if (!window.Cocoon) {
        return {};
    }

    var CocoonMatch = function (match) {
        this._match = match;
        this.localPlayer = {
            // numerical id
            playerId: -1,
            // id according to online service (may not be numerical)
            uniqueId: -1,
            // username
            userName: ''
        };
        this.players = [];

        this.onDataRecieved = function () {};
        this.onStateChanged = function () {};
        this.onConnectionWithPlayerFailed = function () {};
        this.onFailed = function () {};

        this._onReadyCallback = null;

        this._isReady = false;

        var self = this;

        // maps uniqueId -> playerId
        var playerIdMap = {};

        var requestPlayersCallback = function (players, err) {
            if (err) {
                // window.alert("requestPlayersInfo:" + err.message);
                EventSystem.fire('matchFailed', err);
                return;
            }

            var ownId = match.getLocalPlayerID();

            self.players = players;

            //sort players based on their player ID, we need all players to have the same numerical ID, but cocoon provides us string ids in an arbitrary order
            self.players.sort(function (a, b) {
                return a.userID.localeCompare(b.userID, 'en-en');
                // return a.userID < b.userID;
            });

            Utils.forEach(self.players, function (playerData, i, l, breakLoop) {
                playerData.userName = playerData.userName;
                playerData.uniqueId = playerData.userID;
                playerData.playerId = i;

                if (playerData.uniqueId === ownId) {
                    self.localPlayer = playerData;
                }

                playerIdMap[playerData.uniqueId] = playerData.playerId;
            });

            if (!self._isReady) {
                self._isReady = true;
                if (self._onReadyCallback) {
                    self._onReadyCallback();
                    self._onReadyCallback = null;
                }
            }
        };

        var setUpEvents = function () {
            match.on("match", {
                dataReceived: function (matchObj, data, uniqueId) {
                    var dataObj;
                    var numericalId = playerIdMap[uniqueId];
                    //console.log("received Data: " + data + " from Player: " + playerID);
                    try {
                        dataObj = {
                            match: matchObj,
                            list: JSON.parse(data),
                            playerId: numericalId
                        };
                    } catch (e) {
                        Utils.log("ERROR: could not parse data!!");
                        return;
                    }
                    self.onDataRecieved(dataObj);
                },
                stateChanged: function (matchObj, uniqueId, state) {
                    console.log("onMatchStateChanged: " + uniqueId + " " + state);
                    // state 1 = connected
                    // state 2 = disconnected
                    // state 3 = failure (see connectionWithPlayerFailed)

                    if (!self._isReady) {
                        // starting process: wait until all players are connected
                        if (match.getExpectedPlayerCount() === 0) {
                            // all players are connected, get their player info before starting
                            match.requestPlayersInfo(requestPlayersCallback);
                        } else if (state === 2) {
                            // person disconnected on match startup??
                            // best thing to do is to quit the entire startup process
                            if (self._onReadyCallback) {
                                self._onReadyCallback("A player disconnected prematurely");
                                self._onReadyCallback = null;
                            }
                        }
                    } else {
                        // we are only interested in players leaving the room AFTER the match started
                        self.onStateChanged({
                            match: matchObj,
                            playerId: playerIdMap[uniqueId],
                            state: state
                        });
                    }
                },
                connectionWithPlayerFailed: function (matchObj, uniqueId, err) {
                    // not sure when this is called

                    console.log("onMatchConnectionWithPlayerFailed: " + uniqueId + " " + err);
                    if (!self._isReady) {
                        if (self._onReadyCallback) {
                            self._onReadyCallback("A player disconnected prematurely");
                            self._onReadyCallback = null;
                        }
                    } else {
                        // self.onConnectionWithPlayerFailed({
                        //     match: matchObj,
                        //     player: player,
                        //     error: err
                        // });

                        // we handle this as player disconnecting, but with state 3 to indicate a connection failure
                        self.onStateChanged({
                            match: matchObj,
                            playerId: playerIdMap[uniqueId],
                            state: 3
                        });
                    }
                },
                failed: function (matchObj, err) {
                    console.error("onMatchFailed " + err);
                    if (!self._isReady) {
                        if (self._onReadyCallback) {
                            self._onReadyCallback("Could not start connection. " + err);
                            self._onReadyCallback = null;
                        }
                    }
                }
            });
        };

        if (match.getExpectedPlayerCount() === 0) {
            match.requestPlayersInfo(requestPlayersCallback);
        }


        setUpEvents();
    };

    CocoonMatch.prototype.playerHasLowestId = function () {
        var lowestId = this.localPlayer.playerId;
        Utils.forEach(this.players, function (player, i, l, breakLoop) {
            if (player.playerId < lowestId) {
                lowestId = player.playerId;
            }
        });
        if (lowestId === this.localPlayer.playerId) {
            return true;
        } else {
            return false;
        }
    };

    CocoonMatch.prototype.onReady = function (callback) {
        if (this._isReady) {
            callback();
        } else {
            this._onReadyCallback = callback;
        }
    };

    CocoonMatch.prototype.start = function () {
        this._match.start();
    };

    CocoonMatch.prototype.disconnect = function () {
        this._match.disconnect();
    };

    /**
     *
     *
     *
     */

    var socialService;
    var multiplayerService;
    var usingGameCenter = false;
    var waitingLogin = false;

    if (window.Cocoon.getPlatform() === 'ios') {
        var gc = window.Cocoon.Social.GameCenter;
        socialService = gc.getSocialInterface();
        multiplayerService = gc.getMultiplayerInterface();
        usingGameCenter = true;
    } else if (window.Cocoon.getPlatform() === 'android') {
        var gp = window.Cocoon.Social.GooglePlayGames;
        gp.init({
            clientId: '206196665716-4dlouvv33sasa8ta7tbpp0o85g5q7nke.apps.googleusercontent.com',
            // defaultLeaderboard: "CgkI_P7xs4EREAIQBg",
            showAchievementNotifications: true,
            /**
             * WARNING: Be sure to also enable the scopes on console.developers.google.com,
             * otherwise you will get an 'INVALID_SCOPE' login error.
             */
            scopes: [
                // 'https://www.googleapis.com/auth/drive.appdata',
                // 'appfolder',
                'drive.appfolder',
                // 'drive.appdata'
            ]
        }, function () {});
        socialService = gp.getSocialInterface();
        multiplayerService = gp.getMultiplayerInterface();
    }
    if (socialService) {
        // DEPRECATED: can probably be removed altogether
        socialService.on("loginStatusChanged", function (loggedIn, error) {
            // pass to eventsystem
            EventSystem.fire('loginStatusChanged', {
                success: loggedIn,
                error: error
            });
        });
    }

    var loginSocialService = function (autoLogin, callback) {
        console.log('Login to service');
        if (!socialService) {
            Utils.log("Could not find social service");
            return;
        }

        if (!waitingLogin) {
            waitingLogin = true;
            socialService.login(function (loggedIn, error) {
                if (!loggedIn || error) {
                    Utils.log("Login failed: " + JSON.stringify(error));
                    //Tell the user that Game Center is Disabled
                    if (!autoLogin && error.code == 2 && usingGameCenter) {
                        window.Cocoon.Dialog.confirm({
                            title: "Game Center Disabled",
                            message: "Sign in with the Game Center application to enable it",
                            confirmText: "Ok",
                            cancelText: "Cancel"
                        }, function (accepted) {
                            if (accepted) {
                                window.Cocoon.App.openURL("gamecenter:");
                            }
                        });
                    }
                    callback({
                        success: false,
                        error: error
                    });
                } else {
                    console.log('Login complete');
                    callback({
                        success: true
                    });
                }
                waitingLogin = false;
            });

        }
    };

    var onMatchFound = function (match, error) {
        if (error) {
            Utils.log(error);
            return;
        }

        // Possible errors: no match came out
        if (!match) {
            console.log(error ? error.message : "match canceled");
            if (module.onMatchFound) {
                module.onMatchFound(null, error);
            }
            EventSystem.fire('matchCancel', {
                error: error
            });
            return;
        }

        //create a new Match object
        var cm = new CocoonMatch(match);

        if (module.onMatchFound) {
            module.onMatchFound(cm);
        }
    };

    var module = {
        findMatch: function (minPlayers, maxPlayers, playersToInvite, playerGroup, playerAttributes) {
            var request = new window.Cocoon.Multiplayer.MatchRequest(minPlayers, maxPlayers, playersToInvite, playerGroup, playerAttributes);
            multiplayerService.findMatch(request, onMatchFound);
        },
        onMatchFound: null,
        onInvitation: null,
        onPlayerDisconnects: null,
        sendData: function (match, data, sendMode) {
            match._match.sendDataToAllPlayers(data, sendMode || 0);
        },
        sendDataTo: function (playerId, match, data, sendMode) {
            var uniqueId; // need to find this id
            Utils.forEach(match.players, function (playerData, i, l, breakLoop) {
                if (playerData.playerId === playerId) {
                    uniqueId = playerData.uniqueId;
                    breakLoop();
                }
            });

            match._match.sendData(data, [uniqueId], sendMode || 0);
        },
        onDataRecieved: null,
        login: function (onLogin) {
            if (multiplayerService) {
                multiplayerService.on("invitation", {
                    received: function () {
                        //TODO
                    },
                    loaded: function (match, error) {
                        //TODO
                        if (error) {
                            Utils.log(error);
                        }
                    }
                });
            }

            //Social Service Login and Score Listeners
            if (socialService) {
                //login
                loginSocialService(true, onLogin);
            }
        },
        isLoggedIn: function () {
            return socialService.isLoggedIn();
        }
    };

    return module;
});