/**
 * Pedestrian states
 * @moduleName PedestrianStates
 */
bento.define('states/pedestrian', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/entity',
    'bento/eventsystem',
    'bento/utils',
    'bento/tween'
], function (
    Bento,
    Vector2,
    Rectangle,
    Entity,
    EventSystem,
    Utils,
    Tween
) {
    'use strict';
    var PedestrianStates = {
        idle: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                data.idleTime = Utils.getRandomRange(30, 240);
            },
            update: function (data) {

                data.idleTime--;
                if (data.idleTime <= 0) {
                    data.setState('walk', {
                        direction: Math.round(Math.random()) * 2 - 1 //-1 or 1
                    });
                }

            },
            end: function (data) {}
        },
        walk: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'walk');
                data.direction = transitionData.direction;
                data.time = Utils.getRandomRange(60, 180);


                data.lastX = data.entity.position.x;
                data.stuckTimer = 0;
            },
            update: function (data) {
                if (data.movable.velocity.x > 0) {
                    data.multiScale.direction.x = 1;
                } else if (data.movable.velocity.x < 0) {
                    data.multiScale.direction.x = -1;
                }

                data.movable.velocity.x = data.direction;

                data.time--;
                if (data.time <= 0) {
                    data.setState('idle');
                }


                if (Math.abs(data.lastX - data.entity.position.x) < 1) {
                    data.stuckTimer++;
                    if (data.stuckTimer >= 15) {
                        data.setState('idle');
                        console.log('stuck!');
                    }
                } else {
                    data.stuckTimer = 0;
                    data.lastX = data.entity.position.x;
                }
                //TODO check if we are stuck (position didn't change for half a second?), if so, cut back to idle state
            },
            end: function (data) {}
        },
        hurt: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'hurt');
                data.hurt = 20;

                data.punchDirection = Utils.sign(data.movable.velocity.x) * -1 || data.multiScale.direction.x * -1;

                new Tween({
                    from: 0.4,
                    to: 0,
                    'in': 20,
                    ease: Tween.ELASTIC,
                    alpha: -5,
                    beta: 2,
                    onUpdate: function (v, t) {
                        data.multiScale.hurt.x = 1 + v;
                        data.multiScale.hurt.y = 1 - v;
                    }
                });
            },
            update: function (data) {
                data.hurt--;
                data.multiScale.direction.x = data.punchDirection;
                if (data.hurt <= 0 && data.movable.grounded) {
                    data.setState('flee');
                }
            },
            end: function (data) {}
        },
        shocked: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'shocked');

                data.movable.grounded = false;
                data.movable.velocityZ = -2;
            },
            update: function (data) {
                if (data.movable.grounded) {
                    data.setState('flee');
                }
            },
            end: function (data) {}
        },
        flee: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'flee');
                data.viewport = Bento.getViewport();
                if (!data.fleeSpeed) {
                    data.fleeSpeed = 0.2;
                }
                if (!data.maxFlee) {
                    data.maxFlee = 3.8;
                }
                if (!data.startFlee) {
                    data.startFlee = 0.75;
                }
            },
            update: function (data) {
                var side = Utils.sign(data.entity.position.x - (data.viewport.x + data.viewport.width / 2));

                if (Math.abs(data.movable.velocity.x) < data.startFlee) {
                    //set starting velocity when fleeing
                    data.movable.velocity.x = side * data.startFlee;
                } else {
                    //accelerate velocity til max speed has been reached
                    data.movable.velocity.x = Utils.approach(data.movable.velocity.x, data.maxFlee * side, data.fleeSpeed);
                }

                if (data.movable.velocity.x > 0) {
                    data.multiScale.direction.x = 1;
                } else if (data.movable.velocity.x < 0) {
                    data.multiScale.direction.x = -1;
                }

                //TODO not working!:
                var visibleArea = data.viewport.clone().grow(Math.max(data.entity.dimension.width, data.entity.dimension.height) * 2);

                if (!visibleArea.hasPosition(data.entity.position)) {
                    Bento.objects.remove(data.entity);
                }
            },
            end: function (data) {}
        }
    };

    return PedestrianStates;
});