/**
 * Standard (walking) enemy states
 * @moduleName StandardEnemyStates
 */
bento.define('states/standardenemy', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/punch',
    'modules/tapticengine',
    'entities/effect',
    'modules/matchmanager',
    'entities/freezecontroller',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Punch,
    TapticEngine,
    Effect,
    MatchManager,
    FreezeController,
    Globals
) {
    'use strict';
    var StandardEnemyStates = {
        idle: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                data.punchable.isInvincible = false;
            },
            update: function (data) {},
            move: function (data, direction) {
                data.setState('walk');
            },
            action: function (data, action) {
                if (action && action.data) {
                    data.setState('punch', {
                        target: action.data.target.name
                    });
                }
            },
            end: function (data) {

            },
            canGrab: true
        },
        walk: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'walk');
                data.punchable.isInvincible = false;
            },
            update: function (data) {},
            move: function (data, direction) {
                var viewport = Bento.getViewport().clone();
                data.movable.velocity = direction.scalarMultiply(0.8);
                data.movable.velocity.y *= 0.8;

                if (data.movable.velocity.sqrMagnitude() > 1) {
                    data.sprite.setSpriteSheet(data.rootPath + 'run');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'walk');
                }
            },
            action: function (data, action) {
                //ai controller wants to punch, cancel walk and switch to idle
                data.setState('idle');
            },
            canGrab: true
        },
        jump: {
            ignoreDirection: true,
            start: function (data, transitionData) {
                // TODO: add jump animation
                data.sprite.setSpriteSheet(data.rootPath + 'jump');
                data.punchable.isInvincible = true;
                // controlling player should sync velocity
                data.sync({
                    velocityX: transitionData.jumpX,
                    velocityY: transitionData.jumpY,
                    velocityZ: data.movable.velocityZ,
                    target: transitionData.target ? transitionData.target : null,
                });

                data.movable.velocity.x = data.syncData.velocityX;
                data.movable.velocity.y = data.syncData.velocityY;
                data.movable.velocityZ = data.syncData.velocityZ;
            },
            update: function (data) {
                var player;
                // check if there is a player to target
                if (data.syncData.target) {
                    player = Utils.getPlayerById(data.syncData.target.substr(-1));
                }

                if (player) {
                    data.multiScale.direction.x = Utils.sign(player.position.x - data.entity.position.x) || data.multiScale.direction.x;
                } else {
                    if (data.movable.velocity.x > 0) {
                        data.multiScale.direction.x = 1;
                    } else if (data.movable.velocity.x < 0) {
                        data.multiScale.direction.x = -1;
                    }
                }

                if (data.movable.grounded) {
                    data.setState('idle');
                }
            },
            end: function (data) {
                data.punchable.isInvincible = false;
            },
            canGrab: false
        },
        punch: {
            ignoreBlock: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                data.punchAnimName = data.punchAnimName || 'punch'; // attack sprite is called "punchX"
                data.punchAnims = data.punchAnims || 2; // default behavior is having 2 punches

                data.punchAnim = (data.punchAnim || 0) + 1;
                data.punchAnimNumber = '';
                if (data.punchAnims > 1) {
                    data.punchAnimNumber = (data.punchAnim % 2 + 1);
                }
                data.sprite.setSpriteSheet(data.rootPath + data.punchAnimName + data.punchAnimNumber, function () {
                    data.setState('idle');
                });

                data.punchState = 1;

                //turn towards player
                var target = Bento.objects.getByName(transitionData.target)[0];
                if (target) {
                    data.multiScale.direction.x = Utils.sign(target.position.x - data.entity.position.x);
                }
                data.punchable.isInvincible = false;
            },
            update: function (data) {
                if (data.punchState === 1 && data.sprite.currentFrame >= 2) {
                    var punchObject = new Punch({
                        target: ['player', 'citizen'],
                        attack: data.attack || 1,
                        stamina: data.stamina || 0.334,
                        skill: data.stats.skill,
                        direction: data.multiScale.direction.x,
                        pushBack: new Vector2(0.5, 1.5),
                        parent: data.entity,
                        boundingBox: new Rectangle(0, -5, data.entity.punchRange || 32, 10),
                        position: data.entity.position.add(new Vector2((data.multiScale.direction.x < 0 ? -data.entity.punchRange - 4 : 4), 0)),
                        onHit: Punch.getDefaultCallback({
                            entity: data.entity,
                            effect: 'effects/hitfx3-small',
                            sfx: 'sfx_punch1',
                            taptic: 'medium'
                        }),
                        debug: true
                    });
                    Bento.objects.attach(punchObject);

                    data.punchState = 2;
                }
            }
        },
        hurt: {
            ignoreSync: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                EventSystem.fire('enemyHurt', data.entity.name);
                data.hurtFrame = (data.hurtFrame || 0) + 1;
                data.sprite.setSpriteSheet(data.rootPath + 'hurt' + (data.hurtFrame % 2 + 1));
                data.hurt = 30;
                // set standard gravity
                data.movable.gravity = 0.15;

                // decrease vertical speed
                data.movable.velocity.y /= 2;

                data.punchDirection = Utils.sign(data.movable.velocity.x) * -1 || data.multiScale.direction.x * -1;

                new Tween({
                    from: 0.5,
                    to: 0,
                    'in': 30,
                    ease: Tween.ELASTIC,
                    alpha: -5,
                    beta: 2,
                    onUpdate: function (v, t) {
                        data.multiScale.hurt.x = 1 + v;
                        data.multiScale.hurt.y = 1 - v;
                    }
                });

                // right now this is specific to the bat,
                // but assuming hitting an enemy would stop anyone from flying
                data.isFlying = false;
            },
            update: function (data) {
                data.hurt--;
                data.multiScale.direction.x = data.punchDirection;
                if (data.hurt <= 0) {
                    data.setState('idle');
                }
            },
            end: function (data) {

            }
        },
        faint: {
            ignoreSync: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                // decrease vertical speed
                data.movable.velocity.y /= 2;

                data.movable.velocityZ = Math.min(data.movable.velocityZ, -1) - 1;
                // set standard gravity
                data.movable.gravity = 0.15;

                // add any extra velocity
                if (transitionData.velocity) {
                    data.movable.velocity.x += transitionData.velocity.x;
                    data.movable.velocity.y += transitionData.velocity.y;
                    data.movable.velocityZ = transitionData.velocity.z;
                }

                // standardize velocity
                if (data.movable.velocity.sqrMagnitude() > 0 && Math.abs(data.movable.velocity.x) < 1.2) {
                    data.movable.velocity.normalize();
                    data.movable.velocity.x *= 1.2;
                }


                data.sprite.setSpriteSheet(data.rootPath + 'flyup');

                if (Utils.getRandom(2) && data.previousState !== 'faint') {
                    Bento.audio.stopSound(data.entity.sfxBase + 'sfx_faint');
                    Bento.audio.playSound(data.entity.sfxBase + 'sfx_faint');
                }

                data.faintState = 1;
                data.faintTime = transitionData.faintTime || 40;

                data.punchDirection = Utils.sign(data.movable.velocity.x) * -1 || data.multiScale.direction.x * -1;

                // right now this is specific to the bat,
                // but assuming hitting an enemy would stop anyone from flying
                data.isFlying = false;
            },
            update: function (data) {
                if (data.movable.grounded) {
                    // lying on the ground
                    data.sprite.setSpriteSheet(data.rootPath + 'liedown');

                    data.movable.velocity.x = 0;
                    data.movable.velocity.y = 0;

                    // cant hit while on the ground
                    data.punchable.isInvincible = true;

                    // get up after a while
                    --data.faintTime;
                    if (data.faintTime <= 0) {
                        data.setState('getUp');
                    }

                } else if (data.movable.velocityZ >= 0) {
                    // flying down
                    data.sprite.setSpriteSheet(data.rootPath + 'flydown');
                }

                if (data.punchDirection) {
                    data.multiScale.direction.x = data.punchDirection;
                }
            }
        },
        getUp: {
            ignoreDirection: true,
            start: function (data, transitionData) {
                data.getUpTimer = 15;
                data.sprite.setSpriteSheet(data.rootPath + 'getup');
                // data.punchable.addStamina(2);
                data.punchable.resetStamina();
            },
            update: function (data) {
                data.getUpTimer--;
                if (data.getUpTimer <= 0) {
                    data.setState('jumpBack');
                }
            }
        },
        jumpBack: {
            start: function (data, transitionData) {
                // random jumping away (50%)
                data.sync({
                    shouldJump: Utils.getRandom(2),
                });
                if (!data.syncData.shouldJump) {
                    data.punchable.isInvincible = false;
                    data.setState('idle');
                    return;
                }

                if (data.enemyController.isLocallyControlled) {
                    // only jump when player is nearby
                    var angles = [];
                    Utils.forEach(Bento.objects.getByFamily('player'), function (player, i, l, breakLoop) {
                        if (player.position.distance(data.entity.position) < 40) {
                            var pPos = player.position;
                            var ePos = data.entity.position;
                            var rad = Math.atan2(ePos.y - pPos.y, ePos.x - pPos.x);
                            angles.push(new Vector2(Math.cos(rad), Math.sin(rad)));
                        }
                    });

                    if (angles.length) {
                        var angle = Math.atan2(angles[0].y, angles[0].x);
                        // angle -= atan2(b.y, b.x);
                        for (var i = 1; i < angles.length; ++i) {
                            angle = Math.atan2(angles[i].y, angles[i].x);
                        }

                        // add to velocity
                        data.movable.velocity.x = Math.cos(angle) * 2.2;
                        data.movable.velocity.y = Math.sin(angle) * 2.2;
                        data.movable.velocityZ = -0.8;
                        data.sprite.setSpriteSheet(data.rootPath + 'jump');
                    } else {
                        data.punchable.isInvincible = false;
                        data.sprite.setSpriteSheet(data.rootPath + 'idle');
                        data.setState('idle');
                    }
                }
            },
            update: function (data) {
                if (data.movable.grounded) {
                    data.setState('idle');
                    data.punchable.isInvincible = false;
                }
            }
        },
        die: {
            ignoreSync: true,
            ignoreDirection: true,
            start: function (data, transitionData) {
                data.hurtFrame = (data.hurtFrame || 0) + 1;
                data.sprite.setSpriteSheet(data.rootPath + 'hurt' + (data.hurtFrame % 2 + 1));

                // play a random deathsound based of available sounds, only one death sound can be played at the same time.
                Bento.audio.stopSound(Globals.deathSound);
                Globals.deathSound = data.entity.sfxBase + 'sfx_death' + Utils.getRandom(Globals.availableDeathSounds[data.entity.sfxBase]);
                Bento.audio.playSound(Globals.deathSound);

                Bento.audio.stopSound('sfx_death-hit');
                Bento.audio.playSound('sfx_death-hit');

                FreezeController.freeze(8, 2);
                // screenshake
                Bento.objects.get('camera', function (camera) {
                    camera.shake(8, 4);
                    camera.updateWhenPaused = Globals.PauseLevels.FREEZE;
                });

                var signX = Utils.sign(data.movable.velocity.x);
                var signY = Utils.sign(data.movable.velocityZ);
                var velocityX = transitionData.direction * 5 || (signX * 5);

                if (velocityX) {
                    data.movable.velocity.x = velocityX;
                } else {
                    data.bounce = true;
                }
                data.movable.velocityZ = -5;

                data.movable.ignoreCollisions = true;
            },
            update: function (data) {
                var gameSpeed = Bento.getGameSpeed();
                data.entity.rotation += Utils.sign(data.movable.velocity.x) * 0.05 * gameSpeed;

                if (data.bounce && data.movable.grounded) {
                    data.bounce = false;
                    data.movable.velocity.x = data.multiScale.direction.x * 5;
                    data.movable.velocityZ = -3;
                }
            }
        },
        block: {
            ignoreDirection: true,
            start: function (data) {
                data.timer = 0;
                data.sprite.setSpriteSheet(data.rootPath + 'block');

                // shake a bit
                data.shake.shake(15);

                // cant be hit while blocking
                if (data.punchable) {
                    data.punchable.setInvincible(true);
                }
                // release grab if needed
                if (data.grabable) {
                    data.grabable.setGrabbed(null);
                }

                // dont jump as far
                data.movable.velocityZ /= 4;

                // show blocked
                data.damageCounter.block();

                var effect = new Effect({
                    position: new Vector2(
                        data.entity.position.x,
                        data.entity.position.y - 4
                    ),
                    spriteSheet: 'effects/block'
                });
                Bento.objects.attach(effect);

                Bento.audio.stopSound('sfx_block');
                Bento.audio.playSound('sfx_block');
            },
            update: function (data) {
                // end after 15 frames
                data.timer += 1;
                if (data.timer > 10) {
                    data.setState('idle');
                }

                // stand still
                data.movable.velocity.x *= 0.9;
                data.movable.velocity.y *= 0.5;
            },
            end: function (data) {
                if (data.punchable) {
                    data.punchable.setInvincible(false);
                }
            }
        },
        grabbed: {
            ignoreDirection: true,
            start: function (data, transitionData) {
                var grabber = transitionData.grabber;
                if (!Utils.isDefined(grabber)) {
                    // invalid state! --> caused by punchsyncer
                    // try to get the last grabber id
                    grabber = data.lastGrabbed;
                }
                var entity = data.entity;

                data.lastGrabbed = grabber;

                // set spritesheet
                if (Utils.hasAsset('spritesheets', data.rootPath + 'grabbed')) {
                    data.sprite.setSpriteSheet(data.rootPath + 'grabbed');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'flyup');
                }

                // cancel any previous movement
                data.movable.velocity.x = 0;
                data.movable.velocity.y = 0;

                data.grabTimer = 0;
            },
            update: function (data) {
                var grabber = data.lastGrabbed;
                var player = Utils.getPlayerById(grabber);
                if (!player) {
                    Utils.log('Grab bug');
                    data.movable.velocity.x = 0;
                    data.movable.velocity.y = 0;
                    data.setState('idle');
                    return;
                }
                var playerMultiScale = player.getComponent('multiScale');
                var entity = data.entity;
                var movable = data.movable;
                var grabDistance = 16;
                var playerStateMachine = player.getComponent('stateMachine');

                // change grabdistance if defined
                if(playerStateMachine) {
                    if(Utils.isDefined(playerStateMachine.data.grabDistance)) {
                        grabDistance = playerStateMachine.data.grabDistance;
                    }
                }
                var target = player.position.add(new Vector2(playerMultiScale.direction.x * grabDistance, -1));

                // look at player
                data.multiScale.direction.x = -playerMultiScale.direction.x;

                // move towards player
                movable.velocity.x = (target.x - entity.position.x) / 4;
                movable.velocity.y = (target.y - entity.position.y) / 4;

                // quick fix: check the local grabbing player's state, should be in grab or grabAttack
                if (!player.isRemote &&
                    playerStateMachine.stateName !== 'grab' &&
                    playerStateMachine.stateName !== 'grabAttack' &&
                    playerStateMachine.stateName !== 'specialGrab'
                ) {
                    // should have been released
                    // console.log('warning: grab bug');
                    movable.velocity.x = 0;
                    movable.velocity.y = 0;
                    data.setState('idle');
                    return;
                }

                // after a while, the enemy should release itself
                data.grabTimer += 1;
                if (data.grabTimer > data.grabDuration - 60) {
                    // shake after 2 sec
                    entity.transform.x = Utils.getRandomRangeFloat(-1, 1);
                    entity.transform.y = Utils.getRandomRangeFloat(-1, 1);
                }
                if (data.grabTimer > data.grabDuration) {
                    data.addVelocity = 1;
                    data.grabable.release(player);
                    movable.velocity.x = playerMultiScale.direction.x * 1;
                    movable.velocityZ = -1;
                }

            },
            end: function (data) {
                // stop shaking
                data.entity.transform.x = 0;
                data.entity.transform.y = 0;
            }
        },
        grabReleased: { // this state prevents player grabbing immediately after release
            ignoreDirection: true,
            start: function (data) {
                data.grabReleaseTimer = 0;
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
            },
            update: function (data) {
                data.grabReleaseTimer += 1;
                // simple cooldown window of a few frames
                if (data.grabReleaseTimer > 30) {
                    data.setState('idle');
                }
            }
        },
        puppet: {
            ignoreSync: true,
            start: function (data, transitionData) {
                var anim = transitionData.animation;
                if (!anim) {
                    anim = 'idle';
                }
                data.sprite.setSpriteSheet(data.rootPath + anim);
            }
        },
        targetSuperAttack: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                data.punchable.setInvincible(true);
            }
        },
        punchIntent: {
            start: function (data) {
                data.walkingEnemy.noHurt = true;
                data.walkingEnemy.noFaint = true;
            }
        }
    };
    return StandardEnemyStates;
});