/**
 * Standard Player states
 * @moduleName StandardPlayerStates
 */
bento.define('states/standardplayer', [
    'bento',
    'bento/math/vector2',
    'bento/math/rectangle',
    'bento/components/sprite',
    'bento/components/clickable',
    'bento/entity',
    'bento/eventsystem',
    'bento/gui/clickbutton',
    'bento/gui/counter',
    'modules/text',
    'bento/utils',
    'bento/tween',
    'entities/effect',
    'modules/tapticengine',
    'components/sort',
    'managers/ai',
    'entities/punch',
    'managers/buff',
    'modules/matchmanager',
    'globals'
], function (
    Bento,
    Vector2,
    Rectangle,
    Sprite,
    Clickable,
    Entity,
    EventSystem,
    ClickButton,
    Counter,
    Text,
    Utils,
    Tween,
    Effect,
    TapticEngine,
    Sort,
    AiManager,
    Punch,
    BuffManager,
    MatchManager,
    Globals
) {
    'use strict';
    var doPunch = function (data, action) {
        // is the player standing on a weapon? -> pick that up first
        var isPickingUp = false;
        var hasWeapon = data.entity.getComponent('weapon');

        var checkWeapon = function () {
            if (hasWeapon) {
                return;
            }
            data.entity.collidesWith({
                family: 'weapons',
                onCollide: function (weapon) {
                    weapon.getComponent('weapon', function (weaponComponent) {
                        // picking up a weapon may fail, for example because it's already picked up
                        isPickingUp = weaponComponent.pickUp(data.entity);
                    });
                }
            });
        };


        // not always a punch, but a very common player action for many states
        switch (action.type) {
        case 'tap':
            checkWeapon();
            if (!isPickingUp) {
                // has weapon?
                if (hasWeapon) {
                    data.setState('weapon');
                } else {
                    data.setState('punch');
                }
            } else {
                data.setState('pickup');
            }
            return true;
        case 'hold':
            if (action.data.duration >= 5) {
                checkWeapon();
                if (!isPickingUp) {
                    if (!hasWeapon && data.entity.canCharge) {
                        data.setState('charge');
                    }
                } else {
                    data.setState('pickup');
                }
            }
            if (action.data.duration >= 20 && !isPickingUp && hasWeapon) {
                data.setState('prepareThrow');
            }
            return true;
        }
        // returns false to allow other types
        return false;
    };
    var StandardPlayerStates = {
        idle: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
            },
            update: function (data) {
                var movable = data.movable;
                movable.velocity.x = Utils.approach(movable.velocity.x, 0, 0.2);
                movable.velocity.y = Utils.approach(movable.velocity.y, 0, 0.2);

                // punch again?
                if (data.shouldPunchAgain) {
                    data.shouldPunchAgain = false;
                    doPunch(data, {
                        type: 'tap'
                    });
                }
            },
            move: function (data, dir) {
                data.movable.velocity = dir.scalarMultiply(2);
                data.setState('walk');
            },
            action: function (data, action) {
                // doPunch(data, action);
                if (!doPunch(data, action)) {
                    if (action.type === 'swipe') {
                        if (action.data.direction.x && data.entity.canDash) {
                            // horizontal swipe
                            data.setState('running', action.data);
                        } else {
                            if (action.data.direction.y < 0) {
                                // up swipe
                                var canJump = false;
                                Bento.objects.get('healthBar', function (healthBar) {
                                    var comboPoints = Globals.attackComboPoints[data.rootPath.slice(0, -1)].jump;
                                    canJump = healthBar.depleteComboPoints(comboPoints);
                                });
                                if (canJump) {
                                    if (!data.isRemote) {
                                        EventSystem.fire('quest-doJumpAttack', {
                                            count: 1
                                        });
                                    }
                                    data.setState('jumpAttack', action.data);
                                } else {
                                    data.setState('jump', action.data);
                                }
                            } else {
                                // down swipe
                            }
                        }
                    }
                }
            }
        },
        walk: {
            start: function (data) {
                data.isWalking = true;
                data.sprite.setSpriteSheet(data.rootPath + 'run');
                // release punch cache
                data.shouldPunchAgain = false;
                data.walked = 0;
            },
            update: function (data) {
                var entity = data.entity;
                if (!data.isWalking && (!data.isRemote || MatchManager.noSync)) {
                    data.setState('idle');
                    return;
                }
                data.isWalking = false;

                if (data.movable.velocity.x > 0) {
                    data.multiScale.direction.x = 1;
                } else if (data.movable.velocity.x < 0) {
                    data.multiScale.direction.x = -1;
                }

                if (!data.isRemote) {
                    data.walked += data.movable.velocity.x + data.movable.velocity.x;
                    var didWalk = false;
                    while (data.walked > 50) {
                        data.walked -= 50;
                        didWalk = true;
                    }
                    if (didWalk) {
                        EventSystem.fire('quest-walk', {
                            count: 5
                        });
                    }
                }

                if (data.entity.timer % 8 === 0 && data.movable.velocity.sqrMagnitude() > 1.5 * 1.5) {
                    var puff = new Effect({
                        position: data.entity.position.add(new Vector2(
                            data.multiScale.direction.x * -8,
                            0
                        )),
                        spriteSheet: 'effects/walkcloud'
                    });
                    puff.scale.x = data.multiScale.direction.x;
                    puff.attach(new Sort({}));
                    Bento.objects.attach(puff);
                }

                // grab enemy when moving into them
                // not while player is holding a weapon
                if (!entity.isRemote && !(entity.getComponent('weapon'))) {
                    entity.collidesWith({
                        family: 'enemy',
                        firstOnly: false,
                        onCollide: function (other) {
                            var grabable = other.getComponent('grabable');
                            if (!grabable) {
                                return;
                            }
                            if (data.currentlyGrabbed) {
                                // already grabbing
                                return;
                            }
                            // only grab if the enemy is in front of the player
                            var direction = data.multiScale.direction.x;
                            if (direction > 0 && other.position.x < entity.position.x) {
                                return;
                            } else if (direction < 0 && other.position.x > entity.position.x) {
                                return;
                            }
                            // the player must be moving forward
                            var velocity = data.movable.velocity;
                            if (Math.abs(velocity.x) < Math.abs(velocity.y)) {
                                return;
                            }
                            // other must be grounded
                            var isGrounded = true;
                            other.getComponent('movable', function (movable) {
                                isGrounded = movable.grounded;
                            });
                            if (!isGrounded) {
                                return;
                            }

                            other.getComponent('stateMachine', function (stateMachine) {
                                // there is not a lot of states where you can grab an enemy (probably)
                                if (!stateMachine.state.canGrab) {
                                    return;
                                }
                                // should not be invincible
                                var punchable = other.getComponent('punchable');
                                if (punchable.isInvincible) {
                                    return;
                                }
                                data.setState('grab', {
                                    grabbed: other.name
                                });
                                stateMachine.setState('grabbed', {
                                    grabber: data.entity.playerId
                                });
                            });
                        }
                    });
                }

            },
            move: function (data, dir) {
                var mag = dir.magnitude();
                if (mag === 0) {
                    return;
                }
                dir.normalize();

                data.movable.velocity = dir.scalarMultiply((mag > 0.667 ? 2 * BuffManager.current.walkSpeed : 1 * BuffManager.current.walkSpeed)); //move at one of two velocities
                data.movable.velocity.y *= 0.85;
                data.isWalking = true;
                if (data.movable.velocity.sqrMagnitude() > 1.5 * 1.5) {
                    data.sprite.setSpriteSheet(data.rootPath + 'run');
                } else {
                    data.sprite.setSpriteSheet(data.rootPath + 'walk');
                }

            },
            end: function (data) {
                data.isWalking = false;
            },
            action: function (data, action) {
                if (!doPunch(data, action)) {
                    if (action.type === 'swipe') {
                        if (action.data.direction.x && data.entity.canDash) {
                            // horizontal swipe
                            data.setState('running', action.data);
                        } else {
                            if (action.data.direction.y < 0) {
                                // up swipe
                                var canJump = false;
                                Bento.objects.get('healthBar', function (healthBar) {
                                    var comboPoints = Globals.attackComboPoints[data.rootPath.slice(0, -1)].jump;
                                    canJump = healthBar.depleteComboPoints(comboPoints);
                                });
                                if (canJump) {
                                    if (!data.isRemote) {
                                        EventSystem.fire('quest-doJumpAttack', {
                                            count: 1
                                        });
                                    }
                                    data.setState('jumpAttack', action.data);
                                } else {
                                    data.setState('jump', action.data);
                                }
                            } else {
                                // down swipe
                            }
                        }
                    }
                }
            }
        },
        pickup: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'pickup', function () {
                    data.setState('idle');
                });
                // release punch cache
                data.shouldPunchAgain = false;
            }
        },
        weapon: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'weapon', function () {
                    data.setState('idle');
                });

                Bento.audio.stopSound('sfx_fist');
                Bento.audio.playSound('sfx_fist');

                var weapon = data.entity.getComponent('weapon');

                if (!weapon) {
                    console.log('WARNING: no weapon exists. Possibly due to syncing error.');
                    return;
                }

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: data.stats.attack + BuffManager.current.weaponPower,
                    multiplier: 1 + weapon.damage,
                    //TODO: add weapon stamina
                    stamina: weapon.stamina || 0.4,
                    skill: data.stats.skill,
                    pushBack: new Vector2(0.75, 1.45),
                    boundingBox: new Rectangle(0, -5, weapon.range, 10),
                    offset: new Vector2(8, 0),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1-small',
                        sfx: 'sfx_punch',
                        taptic: 'medium',
                        onComplete: function () {
                            weapon.onHit();
                        }
                    }),
                    onMiss: function () {
                        data.combo = 0;
                    },
                    type: 'weapon'
                });
            },
            update: function (data) {
                if (data.punch && data.sprite.getCurrentFrame() > 2) {
                    data.movable.velocity.x = 1 * data.multiScale.direction.x;
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                }
            }
        },

        punch: {
            start: function (data) {
                data.sync({
                    combo: data.combo || 0
                });
                data.combo = data.syncData.combo || 0;

                // get resetframe
                if (data.punchReset) {
                    data.resetPunch = data.punchReset[data.combo];
                } else {
                    data.resetPunch = 4;
                }
                var originalCombo = data.combo;
                var baseAttack = data.stats.attack;

                if (data.combo === 4) {
                    data.setState('comboEnd');
                    return;
                }

                data.punchTimer = 0;

                // set hashit to false
                data.hasHit = false;

                data.sprite.setSpriteSheet(data.rootPath + 'punch' + (1 + data.combo % 4), function () {
                    data.setState('idle');
                });
                data.sprite.setFrame(0);
                data.sprite.setCurrentSpeed(data.sprite.getCurrentSpeed() * BuffManager.current.punchSpeed);

                var sfx = 0;
                if (data.combo >= 4) {
                    sfx = 3; // hard punch
                } else {
                    sfx = Utils.getRandom(data.combo) % 3; // random pitch
                }
                Bento.audio.stopSound('sfx_fist' + sfx);
                Bento.audio.playSound('sfx_fist' + sfx);
                if (data.combo >= 3) {
                    Bento.audio.stopSound(data.rootPath + 'sfx_attack');
                    Bento.audio.playSound(data.rootPath + 'sfx_attack');
                }

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    skill: data.stats.skill,
                    stamina: 0.2,
                    multiplier: 1,
                    pushBack: new Vector2(0.4 * BuffManager.rollCritical('knockback'), 0.8),
                    freeze: data.standardFreeze || 2,
                    boundingBox: new Rectangle(0, -5, data.punchLength[originalCombo], 10),
                    offset: new Vector2(data.punchDistance[originalCombo], 0),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1-small',
                        sfx: 'sfx_punch' + sfx,
                        taptic: 'light',
                        onComplete: function () {
                            data.combo = originalCombo + 1;
                            data.hasHit = true;
                        }
                    }),
                    onMiss: function () {
                        data.combo = 0;
                    }
                });

                if (!data.isRemote) {
                    EventSystem.fire('quest-punch', {
                        count: 1
                    });
                }
            },
            update: function (data) {
                data.punchTimer++;
                if (data.punch && data.sprite.getCurrentFrame() > 2) {
                    data.movable.velocity.x = 0.8 * data.multiScale.direction.x;
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                }
            },
            action: function (data, action) {
                // punch again if requirements are met
                if (action.type === 'tap' && data.sprite.getCurrentFrame() > data.resetPunch) {
                    // data.setState('punch', null, true);
                    data.shouldPunchAgain = true;
                }
            }
        },
        charge: {
            start: function (data) {
                // TODO
                data.sprite.setSpriteSheet(data.rootPath + 'chargeup', function () {
                    data.sprite.setSpriteSheet(data.rootPath + 'chargeup-loop');
                });
                data.specialStage = 0;
            },
            end: function (data) {
                Bento.audio.stopSound('sfx_charge');
            },
            update: function (data) {

            },
            move: function (data, dir) {
                // TEMP
                data.entity.getComponent('touchController', function (touchController) {
                    if (touchController.isActive()) {
                        data.movable.velocity = dir.scalarMultiply(2);
                        data.setState('walk');
                    }
                });
            },
            action: function (data, action) {
                var canChargeAttack = false;
                //if player inputs a second punch before this one ends, schedule it?
                switch (action.type) {
                case 'tap':
                    if (data.specialStage > 0) {
                        Bento.objects.get('healthBar', function (healthBar) {
                            var comboPoints = Globals.attackComboPoints[data.rootPath.slice(0, -1)].charge;
                            canChargeAttack = healthBar.depleteComboPoints(comboPoints);
                        });
                        if (canChargeAttack) {
                            EventSystem.fire('playerChargeAttack');
                            if (!data.isRemote) {
                                EventSystem.fire('quest-doChargeAttack', {
                                    count: 1
                                });
                            }
                            data.setState('chargeAttack', {
                                specialStage: data.specialStage
                            });
                        } else {
                            data.setState('comboEnd');
                        }
                    } else {
                        // has weapon?
                        if (data.entity.getComponent('weapon')) {
                            data.setState('weapon');
                        } else {
                            data.setState('punch');
                        }
                    }
                    break;
                case 'hold':
                    var prevStage = data.specialStage;
                    if (action.data.duration > data.chargeDuration && data.specialStage < 1) {
                        data.specialStage = 1;
                    }

                    // the sound needs to be played slightly later,
                    // because on mobile it's easy to accidentally start the charge
                    if (action.data.duration === 15 && data.specialStage < 1) {
                        Bento.audio.stopSound('sfx_charge');
                        Bento.audio.playSound('sfx_charge', true);
                    }
                    // if (action.data.duration > 80 && data.specialStage < 2) {
                    //     data.specialStage = 2;
                    // }

                    if (prevStage !== data.specialStage) {
                        new Tween({
                            from: 1.2,
                            to: 1,
                            'in': 20,
                            ease: Tween.EASEINBACK,
                            onStart: function () {
                                TapticEngine.impact({
                                    style: 'light'
                                });
                            },
                            onUpdate: function (v, t) {
                                data.multiScale.charge.x = v;
                                data.multiScale.charge.y = 1 / v;
                            }
                        });
                        data.sprite.setSpriteSheet(data.rootPath + 'chargeup-ready');
                        EventSystem.fire('playerCharge');
                        Bento.objects.attach(new Effect({
                            // z: data.entity.z,
                            spriteSheet: 'effects/charge',
                            position: data.entity.position.add(new Vector2(-4 * data.multiScale.direction.x, -16))
                        }));
                        Bento.audio.stopSound('sfx_charge');
                        Bento.audio.stopSound('sfx_chargeready');
                        Bento.audio.playSound('sfx_chargeready');

                    }
                    break;
                }
            }
        },
        prepareThrow: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'preparethrow');
            },
            move: function (data, dir) {
                data.movable.velocity = dir.scalarMultiply(2);
                data.setState('walk');
            },
            action: function (data, action) {
                if (action.type === 'tap') {
                    data.setState('throw');
                }
            }
        },
        throw: {
            start: function (data) {
                data.weapon = data.entity.getComponent('weapon');
                data.sprite.setSpriteSheet(data.rootPath + 'weapon', function () {
                    data.setState('idle');
                });
                Bento.audio.stopSound('sfx_throw');
                Bento.audio.playSound('sfx_throw');
            },
            update: function (data) {
                if (data.sprite.getCurrentFrame() > 2 && data.weapon) {
                    if (!data.isRemote) {
                        //throw events are send seperately, clients should not throw 'for' other players
                        data.weapon.throw(data.multiScale.direction.x);
                    }
                    data.weapon = null;
                }
            }
        },
        comboEnd: {
            start: function (data) {
                var baseAttack = data.stats.attack;
                data.sprite.setSpriteSheet(data.rootPath + 'kick', function () {
                    data.setState('idle');
                });

                data.combo = 0;

                Bento.audio.stopSound('sfx_kick');
                Bento.audio.playSound('sfx_kick');

                //punches are created first for auto-target purposes, and added later for animation purposes
                data.punch = data.createPunch({
                    target: ['enemy', 'destructible'],
                    attack: baseAttack,
                    skill: data.stats.skill,
                    stamina: 1,
                    multiplier: BuffManager.rollCritical('comboEnd'),
                    pushBack: new Vector2(0.75 * BuffManager.rollCritical('knockback'), 2.0),
                    freeze: 5,
                    boundingBox: new Rectangle(0, -5, 18, 10),
                    offset: new Vector2(8, 0),
                    onHit: Punch.getDefaultCallback({
                        entity: data.entity,
                        effect: 'effects/hitfx1',
                        sfx: 'sfx_punch3',
                        taptic: 'medium',
                        onComplete: function () {
                            data.movable.velocity.x = 1 * data.multiScale.direction.x;
                        }
                    })
                });
            },
            update: function (data) {
                if (data.punch && data.sprite.getCurrentFrame() > 3) {
                    Bento.objects.attach(data.punch);
                    data.punch = null;
                    data.movable.velocity.x = 1.5 * data.multiScale.direction.x;
                }
            }
        },
        hurt: {
            start: function (data, transitionData) {
                data.hurt = 40;
                data.hurtFrame = (data.hurtFrame || 0) + 1;
                data.sprite.setSpriteSheet(data.rootPath + 'hurt' + (1 + data.hurtFrame % 2));

                // decrease vertical speed
                data.movable.velocity.y /= 2;

                data.punchDirection = transitionData.punch.direction * -1;
                // release punch cache
                data.shouldPunchAgain = false;

                new Tween({
                    from: 0.5,
                    to: 0,
                    'in': 30,
                    ease: Tween.ELASTIC,
                    alpha: -5,
                    beta: 2,
                    onUpdate: function (v, t) {
                        data.multiScale.hurt.x = 1 + v;
                        data.multiScale.hurt.y = 1 - v;
                    }
                });

                if (!data.isRemote) {
                    EventSystem.fire('quest-hurt', {
                        count: 1
                    });
                }
            },
            update: function (data) {
                data.hurt--;

                data.multiScale.direction.x = data.punchDirection;
                if (data.hurt <= 0) {
                    data.setState('idle');
                }
            }
        },
        faint: {
            start: function (data, transitionData) {
                // decrease vertical speed
                data.movable.velocity.y /= 2;
                data.movable.velocityZ = Math.min(data.movable.velocityZ, -1) - 1;
                data.faintTimeStart = Math.max(data.prevFaintTime || 0, transitionData.faintTime);
                data.faintTime = 30;
                data.sprite.setSpriteSheet(data.rootPath + 'flyup');
                data.faintState = 1;
                // release punch cache
                data.shouldPunchAgain = false;

                data.punchDirection = transitionData.punch.direction * -1;

                if (!data.isRemote) {
                    data.punchable.setInvincible(true);
                }
            },
            update: function (data) {
                if (data.faintState === 1 && data.movable.velocityZ >= 0) {
                    data.sprite.setSpriteSheet(data.rootPath + 'flydown');
                    data.faintState = 2;
                }

                if (data.faintState === 2 && data.movable.grounded) {
                    data.sprite.setSpriteSheet(data.rootPath + 'liedown');
                    data.faintState = 3;
                    data.movable.velocity.x = 0;
                    data.movable.velocity.y = 0;

                    data.faintTime = 30;
                }

                if (data.movable.grounded) {
                    --data.faintTime;
                }

                if (data.faintState === 3 && data.faintTime <= 0) {
                    data.setState('getUp');
                }

                data.multiScale.direction.x = data.punchDirection;
            },
            end: function (data) {
                // data.entity.isDown = false;
                // data.punchable.addStamina(1.5);
            }
        },
        dead: {
            start: function (data, transitionData) {
                Bento.audio.stopSound(data.rootPath + 'sfx_death');
                Bento.audio.playSound(data.rootPath + 'sfx_death');
                data.sprite.setSpriteSheet(data.rootPath + 'liedown');
                data.movable.velocity.x = transitionData ? transitionData.direction || 0 : 0;
                data.movable.velocity.y = 0;
                // release punch cache
                data.shouldPunchAgain = false;

                if (transitionData && transitionData.direction) {
                    // the player might have been punched to death
                    data.multiScale.direction.x = transitionData.direction * -1;
                }
                if (!data.isRemote) {
                    data.punchable.setInvincible(true);
                }

                // re-target if a player dies
                AiManager.distributeAiTasks();
            },
            end: function () {
                // re-target if a player revives
                AiManager.distributeAiTasks();
            }
        },
        escortFailed: {
            start: function (data, transitionData) {
                Bento.audio.stopSound('sfx_charge');
                data.sprite.setSpriteSheet(data.rootPath + 'idle');
                data.movable.velocity.x = transitionData ? transitionData.direction || 0 : 0;
                data.movable.velocity.y = 0;
                // release punch cache
                data.shouldPunchAgain = false;

                // set statename to make the host trigger see that character is not active(dead)
                // delay this otherwise this state gets sent via syncing
                Utils.timeout(5, function () {
                    data.stateMachine.stateName = 'dead';
                });
            },
            action: function (data, transitionData) {
                data.setState('idle');
            },
            move: function (data, dir) {
                data.setState('idle');
            }
        },
        getUp: {
            start: function (data, transitionData) {
                if (data.punchable) {
                    data.punchable.setInvincible(true);
                }
                data.getUpTimer = 30;
                data.sprite.setSpriteSheet(data.rootPath + 'getup');
                // data.punchable.addStamina(2);
                if (!data.isRemote) {
                    data.punchable.resetStamina();
                }
            },
            update: function (data) {
                data.getUpTimer--;
                if (data.getUpTimer <= 0) {
                    data.setState('idle');
                    data.entity.isDown = false;
                }
            },
            end: function (data) {
                if (!data.isRemote) {
                    data.punchable.setInvincibleTime(60 + BuffManager.current.invincibleTime);
                }
            }
        },
        running: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'rundash');
                data.movable.velocity = Utils.toVector2(transitionData.direction).scalarMultiply(2.5 * BuffManager.current.dashSpeed);
                data.isDashing = true;

                EventSystem.fire('playerDash');

                var puff = new Effect({
                    position: data.entity.position.add(new Vector2(
                        Utils.sign(data.movable.velocity.x) * -8,
                        0
                    )),
                    spriteSheet: 'effects/runcloud'
                });
                puff.scale.x = Utils.sign(data.movable.velocity.x);
                puff.attach(new Sort({}));
                Bento.objects.attach(puff);

                Bento.audio.stopSound('sfx_startdash');
                Bento.audio.playSound('sfx_startdash');

                data.ran = 0;
            },
            update: function (data) {
                var movable = data.movable;

                if (data.movable.velocity.x > 0) {
                    data.multiScale.direction.x = 1;
                } else if (data.movable.velocity.x < 0) {
                    data.multiScale.direction.x = -1;
                }

                if (!data.isRemote) {
                    data.ran += data.movable.velocity.x + data.movable.velocity.x;
                    var didRun = false;
                    while (data.ran > 50) {
                        data.ran -= 50;
                        didRun = true;
                    }
                    if (didRun) {
                        EventSystem.fire('quest-run', {
                            count: 5
                        });
                    }
                }
            },
            end: function (data) {
                data.isDashing = false;
            },
            action: function (data, action) {
                if (action.type === 'tap') {
                    if (data.entity.getComponent('weapon')) {
                        data.setState('weapon');
                    } else {
                        var proceed = true;
                        Bento.objects.get('healthBar', function (healthBar) {
                            var comboPoints = Globals.attackComboPoints[data.rootPath.slice(0, -1)].dash;
                            proceed = healthBar.depleteComboPoints(comboPoints);
                        });
                        if (!proceed) {
                            data.setState('punch');
                            return;
                        } else {
                            data.specialStage = 1;
                            EventSystem.fire('playerDashAttack');
                            if (!data.isRemote) {
                                EventSystem.fire('quest-doDashAttack', {
                                    count: 1
                                });
                            }
                            data.setState('runAttack', {
                                specialStage: data.specialStage
                            });
                        }

                    }
                } else if (action.type === 'hold') {
                    // ignore
                } else {
                    data.setState('idle');
                }
            },
            move: function (data) {
                // cancel rundash
                data.setState('walk');
            }
        },
        victory: {
            start: function (data) {
                if (data.punchable) {
                    // remotes may not have a punchable
                    data.punchable.setInvincible(true);
                }
                data.sprite.setSpriteSheet(data.rootPath + 'victory');

                // drop any weapon in hand
                data.entity.getComponent('weapon', function (weapon) {
                    weapon.getComponent('weapon', function (weaponComponent) {
                        weaponComponent.drop(0, true);
                    });
                });
            }
        },
        grab: {
            start: function (data, transitionData) {
                data.sprite.setSpriteSheet(data.rootPath + 'grabbing');
                // stop moving
                data.movable.velocity.x = 0;
                data.movable.velocity.y = 0;

                // store the grabbed
                data.currentlyGrabbed = Bento.objects.get(transitionData.grabbed);
                data.punch = null;
                data.isPunching = null;
                data.hasHit = false;
                data.didRelease = false;
                EventSystem.fire('playerGrab');

                if (!data.isRemote) {
                    EventSystem.fire('quest-grabEnemy', {
                        count: 1
                    });
                }
            },
            update: function (data) {
                var grabbed = data.currentlyGrabbed;
                var stateMachine;

                if (grabbed) {
                    // check if enemy was not forced out of grabbed state
                    // this is usually called if enemy suddenly blocks
                    stateMachine = grabbed.getComponent('stateMachine');
                    if (!grabbed.isAdded || (stateMachine.stateName !== 'grabbed' && stateMachine.stateName !== 'grabbedHurt')) {
                        // enemy was released
                        data.currentlyGrabbed = null;
                        data.setState('idle');
                    }
                } else if (!grabbed) {
                    data.setState('idle');
                }
            },
            action: function (data, action) {
                var grabbed = data.currentlyGrabbed;

                // needs to have released at least once
                if (action.type === 'release' && !data.didRelease) {
                    data.didRelease = true;
                    return;
                }

                // we won't make any punch object because this is a special case
                // -the enemy should not go into another state or move
                // -other enemies should not be able to get hit
                if (action.type === 'tap' && !data.isPunching && grabbed) {
                    data.setState('grabAttack', {
                        grabbed: grabbed.name
                    });
                } else if (
                    (action.type === 'swipe' && !data.isPunching) ||
                    (action.type === 'release' && Math.abs(action.data.joystickPosition.x) > 0.5)
                ) {
                    Bento.audio.stopSound(data.rootPath + 'sfx_throw');
                    Bento.audio.playSound(data.rootPath + 'sfx_throw');
                    if (action.data.joystickPosition) {
                        action.data.direction = new Vector2(
                            action.data.joystickPosition.x >= 0 ? 1 : -1,
                            0
                        );
                    }
                    if (grabbed) {
                        data.currentlyGrabbed = null; // TODO use release?

                        var throwFail = false;
                        grabbed.getComponent('stats', function (stats) {
                            var playerSkill = data.stats.skill;
                            var enemySkill = stats.skill;
                            var failChance = 0;
                            if (playerSkill < stats.skill) {
                                // decrease 10% with each skill difference
                                failChance = Math.abs(playerSkill - stats.skill) * 0.1;
                                failChance = Utils.clamp(0, failChance, 1);
                            }
                            // throw dice
                            if (Utils.getRandomFloat(1) < failChance) {
                                throwFail = true;
                            }
                        });
                        grabbed.getComponent('stateMachine', function (stateMachine) {
                            if (throwFail) {
                                // chance grabthrow fails
                                grabbed.getComponent('movable', function (movable) {
                                    movable.velocityZ = -3;
                                });
                                stateMachine.setState('jump', {
                                    jumpX: (action.data.direction.x || data.multiScale.direction.x) * 2,
                                    jumpY: 0
                                });
                                return;
                            }
                            stateMachine.setState('grabbedThrow', {
                                faintTime: 10,
                                velocity: {
                                    x: 3 * Utils.sign(action.data.direction.x || data.multiScale.direction.x),
                                    y: 0,
                                    z: -2
                                },
                            });
                        });
                        // also turn the player
                        data.multiScale.direction.x = Utils.sign(action.data.direction.x) || data.multiScale.direction.x || 1;
                    }
                    data.setState('grabThrow');

                }

            },
            end: function (data) {
                // release grabbed (unless next state is an attack)
                var nextState = data.stateMachine.stateName;
                var grabbed = data.currentlyGrabbed;
                if (grabbed && nextState !== 'grabAttack' &&
                    nextState !== 'specialGrab' && !data.entity.isRemote
                ) {
                    // do not change the state actively for remotes: there will be messages arriving for that
                    grabbed.getComponent('grabable', function (grabable) {
                        grabable.release(data.entity);
                    });
                }

                data.isPunching = null;
                data.punch = null;
                data.hasHit = false;
            }
        },
        grabAttack: {
            start: function (data, transitionData) {
                // we need to sync the attack stats, otherwise the local player's stats will be used
                var syncData = {
                    attack: data.stats.attack,
                    skill: data.stats.skill,
                    direction: data.multiScale.direction.x
                };
                data.currentlyGrabbed = Bento.objects.get(transitionData.grabbed);
                data.isPunching = true;
                data.hasHit = false;
                data.sprite.setSpriteSheet(data.rootPath + 'grabattack', function () {
                    data.isPunching = null;
                    if (data.currentlyGrabbed) {
                        data.setState('grab', {
                            grabbed: data.currentlyGrabbed.name
                        });
                    }
                });
                EventSystem.fire('playerGrabAttack');
                // release punch cache
                data.shouldPunchAgain = false;

                data.sync(syncData);
            },
            update: function (data) {
                var grabbed = data.currentlyGrabbed;
                var onPunch = Punch.getDefaultCallback({
                    entity: data.entity,
                    effect: 'effects/hitfx1-small',
                    sfx: 'sfx_punch',
                    taptic: 'medium',
                    onComplete: function () {}
                });

                if (grabbed && !data.hasHit && data.sprite.getCurrentFrame() >= 2 && !data.entity.isRemote) {
                    data.hasHit = true;

                    // punch other
                    var punch = new Punch({
                        type: 'grabAttack',
                        target: ['enemy'],
                        attack: data.syncData.attack,
                        skill: data.syncData.skill,
                        stamina: 0.334,
                        multiplier: 1,
                        pushBack: new Vector2(2, 1),
                        boundingBox: new Rectangle(0, 0, 1, 1),
                        offset: new Vector2(0, 0),
                        direction: data.syncData.direction,
                        onHit: onPunch,
                        attacker: data.entity
                    });

                    grabbed.getComponent('punchable', function (punchable) {
                        punchable.hit(punch);
                        if (punchable.isDead && !data.isRemote) {
                            EventSystem.fire('quest-grabAttack', {
                                count: 1
                            });
                        }
                    });

                    // simulate event for combo manager
                    // (can't send event because there are unrelated listeners)
                    /*if (!data.comboManager) {
                        // cache combo manager
                        data.comboManager = Bento.objects.get('comboManager');
                    }
                    if (!data.entity.isRemote && data.comboManager) {
                        data.comboManager.onPunchHit({
                            punch: {
                                target: ['enemy']
                            },
                            attacker: data.entity
                        });
                    }*/
                }
            },
            end: function (data) {
                data.hasHit = false;
                data.isPunching = null;

                // release grabbed (unless next state is back to grab)
                var nextState = data.stateMachine.stateName;
                var grabbed = data.currentlyGrabbed;
                if (grabbed && nextState !== 'grab' && !data.entity.isRemote) {
                    // do not change the state actively for remotes: there will be messages arriving for that
                    grabbed.getComponent('grabable', function (grabable) {
                        grabable.release(data.entity);
                    });
                }
            }
        },
        grabThrow: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'grabthrow', function () {
                    data.setState('idle');
                });
                data.sync({
                    throwDir: data.multiScale.direction.x
                });

                Bento.audio.stopSound('sfx_kick');
                Bento.audio.playSound('sfx_kick');
                Bento.audio.stopSound(data.rootPath + 'sfx_attack');
                Bento.audio.playSound(data.rootPath + 'sfx_attack');

                if (!data.isRemote) {
                    EventSystem.fire('quest-throwEnemy', {
                        count: 1
                    });
                }
            },
            update: function (data) {
                data.multiScale.direction.x = data.syncData.throwDir;
            }
        },
        superAttack: {
            start: function (data) {
                data.setState('idle');
            }
        },
        superAttackDud: {
            start: function (data) {
                data.sprite.setSpriteSheet(data.rootPath + 'grabbing');
                data.movable.velocity.x = data.multiScale.direction.x * 4;
                data.dud = 0;
                if (data.missText) {
                    data.missText.showMiss();
                }
                Bento.objects.get('healthBar', function (healthBar) {
                    healthBar.shake();
                });
            },
            update: function (data) {
                if (++data.dud >= 40) {
                    data.setState('idle');
                }
            }
        },
        jumpAttack: {
            start: function (data, transitionData) {
                var direction = transitionData.cacheDirection;
                data.multiScale.direction.x = direction;

                data.movable.velocityZ = -2.3;
                data.movable.velocity.x = data.multiScale.direction.x * 4;
                data.movable.velocity.y = 0;
                EventSystem.fire('playerJumpAttack');
            },
            update: function (data) {
                if (data.movable.grounded) {
                    data.setState('idle');
                }
            }
        },
        jump: {
            start: function (data, transitionData) {
                var direction = transitionData.cacheDirection;
                data.multiScale.direction.x = direction;
                data.movable.velocityZ = -2.5;
                data.movable.velocity.x = data.multiScale.direction.x * 2.2;
                data.movable.velocity.y = 0;
                EventSystem.fire('playerJump');
                data.sprite.setSpriteSheet(data.rootPath + 'jump');
            },
            update: function (data) {
                if (data.movable.grounded) {
                    data.sprite.setSpriteSheet(data.rootPath + 'idle');
                    data.setState('idle');
                } else {
                    if (data.movable.velocityZ > 0) {
                        data.sprite.setSpriteSheet(data.rootPath + 'jump-down');
                    }
                }
            }
        }
    };

    return StandardPlayerStates;
});